Package: fect
Type: Package
Title: Fixed Effects Counterfactual Estimators
Version: 2.4.5
Date: 2026-05-29
Authors@R: 
    c(person("Yiqing", "Xu", , "yiqingxu@stanford.edu", role = c("aut", "cre")),
      person("Licheng", "Liu", , "lichengl@stanford.edu", role = c("aut")), 
      person("Ziyi", "Liu", , "zyliu2023@berkeley.edu", role = c("aut")),
      person("Ye", "Wang", , "yezhehuzhi@gmail.com", role = c("aut")),       
      person("Tianzhu", "Qin", , "tianzhu@stanford.edu", role = c("aut")),
      person("Shiyun", "Hu", , "hushiyun@pku.edu.cn", role = c("aut")),
      person("Rivka", "Lipkovitz", , "rivkal@mit.edu", role = c("aut")))
Author: Yiqing Xu [aut, cre],
  Licheng Liu [aut],
  Ziyi Liu [aut],
  Ye Wang [aut],
  Tianzhu Qin [aut],
  Shiyun Hu [aut],
  Rivka Lipkovitz [aut]
Maintainer: Yiqing Xu <yiqingxu@stanford.edu>
Description: Provides tools for estimating causal effects in panel data using counterfactual methods, as well as other modern DID estimators. It is designed for causal panel analysis with binary treatments under the parallel trends assumption. The package supports scenarios where treatments can switch on and off and allows for limited carryover effects. It includes several imputation estimators, such as Gsynth (Xu 2017), linear factor models, and the matrix completion method. Detailed methodology is described in Liu, Wang, and Xu (2024) <doi:10.48550/arXiv.2107.00856> and Chiu et al. (2025) <doi:10.48550/arXiv.2309.15983>. Optionally integrates with the "HonestDiDFEct" package for sensitivity analyses compatible with imputation estimators. "HonestDiDFEct" is not on CRAN but can be obtained from <https://github.com/lzy318/HonestDiDFEct>.
URL: https://yiqingxu.org/packages/fect/,
        https://github.com/xuyiqing/fect
BugReports: https://github.com/xuyiqing/fect/issues
NeedsCompilation: yes
License: MIT + file LICENSE
Imports: Rcpp (>= 0.12.3), ggplot2 (>= 2.1.0), GGally (>= 1.0.1),
        doParallel (>= 1.0.10), doFuture, foreach (>= 1.4.3), abind (>=
        1.4-0), codetools, MASS, gridExtra, grid, fixest, doRNG,
        future, parallelly, mvtnorm, dplyr, future.apply, reshape2,
        rlang, scales, splines
Suggests: panelView, testthat (>= 3.0.0), did, DIDmultiplegtDYN,
        ggrepel, HonestDiDFEct, withr
Depends: R (>= 4.1.0)
LinkingTo: Rcpp, RcppArmadillo
RoxygenNote: 7.3.3
Packaged: 2026-05-29 18:08:06 UTC; yiqingxu
Encoding: UTF-8
Config/testthat/edition: 3
LazyData: true
Repository: CRAN
Date/Publication: 2026-05-30 17:10:08 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-30 18:05:07 UTC; unix
Archs: fect.so.dSYM

useDynLib(fect, .registration=TRUE)
##exportPattern("^[[:alpha:]]+")
importFrom(Rcpp, evalCpp)
importFrom("stats", "na.omit", "quantile", "sd", "var", "cov", "pchisq", "lm",
"as.formula", "median", "pnorm", "predict", "qnorm", "reshape",
"dnorm", "pf", "rbinom", "loess", "aggregate", "pt", "qt", "setNames", "time")
importFrom("foreach","foreach","%dopar%")
importFrom("parallel", "detectCores", "stopCluster", "makeCluster")
importFrom("doRNG","registerDoRNG")
importFrom("MASS", "ginv")
importFrom("utils", "setTxtProgressBar", "txtProgressBar", "tail")
importFrom("ggplot2", "ggplot_build", "geom_boxplot", "geom_density", "geom_tile",
           "geom_point", "labs", "theme_bw", "scale_fill_manual",
           "geom_hline", "geom_line", "geom_ribbon", "geom_vline", "geom_text",
           "ggplot", "coord_cartesian", "coord_flip", "element_text",
           "element_blank", "scale_x_discrete", "scale_x_continuous", "scale_y_continuous",
           "scale_linetype_manual", "scale_size_manual", "theme", "geom_area",
           "scale_colour_manual", "aes", "aes_string", "annotate", "ggtitle", "geom_rect",
           "scale_color_discrete", "scale_color_manual", "ggplotGrob",
           "guide_legend", "margin", "geom_jitter", "geom_pointrange",
           "guides","xlab","ylab","element_rect","geom_errorbar","geom_linerange",
           "scale_linewidth_manual","geom_polygon","geom_histogram","after_stat")
importFrom("abind", "abind")
importFrom("mvtnorm", "rmvnorm")
importFrom("GGally", "ggpairs")
importFrom("graphics", "plot")
importFrom("graphics", "hist")
importFrom("gridExtra", "grid.arrange", "arrangeGrob")
importFrom("grid", "textGrob", "gpar")
importFrom("grDevices", "adjustcolor")
importFrom("fixest", "feols")
importFrom(dplyr, "%>%", "group_by", "mutate")
importFrom("future.apply", "future_lapply")
importFrom("parallelly", "availableCores")
importFrom("reshape2", "melt")
importFrom("rlang", "sym", ".data")
importFrom("scales", "pretty_breaks")

export(esplot)
export(get.cohort)
export(interFE)
export(fect)
export(fect_mspe)
export(r.cv.rolling)
export(effect)
export(plot.fect)
export(did_wrapper)
export(fect_sens)
export(fect_iden)
export(att.cumu)
export(imputed_outcomes)
export(estimand)
S3method("fect", "default")
S3method("fect", "formula")
S3method("print", "fect")
S3method("plot", "fect")
S3method("print", "fect_plot_return")
S3method("interFE", "default")
S3method("interFE", "formula")
S3method("print", "interFE")

###################################################################
## CV Helper Functions — single canonical implementation of one
## (r, fold) or (lambda, fold) scoring iteration for each call site.
##
## All functions are internal (prefix ".") and not exported.
## Used by cv.R (notyettreated path) and fect_nevertreated.R (nevertreated path).
###################################################################

## Threshold (Nco * TT) above which CV parallelism auto-enables for each method.
## Used when parallel = TRUE (auto mode); overridden when parallel = "cv".
## CFE value is a placeholder — empirical tuning is deferred to Phase 3.
.CV_PARALLEL_THRESH <- list(ife = 20000L, mc = 20000L, cfe = 60000L)


## Build a PSOCK cluster that inherits the parent's `.libPaths()` so that
## workers can find `fect` (and other user-installed packages) regardless
## of how the parent R session was launched (e.g. Quarto render, RStudio,
## bare Rscript). The default `future::plan(future::multisession)` does
## not propagate libPaths, which causes "no package called 'fect'" worker
## init errors in non-default R session contexts. Use this helper anywhere
## the package would otherwise call `future::plan(future::multisession,
## workers = cores)`. Returns the cluster object; caller is responsible
## for calling `future::plan(future::cluster, workers = cl)` and the
## subsequent on.exit cleanup.
.fect_make_future_cluster <- function(cores) {
    cl <- parallelly::makeClusterPSOCK(
        workers      = cores,
        rscript_libs = .libPaths(),
        autoStop     = TRUE
    )
    ## Pre-load packages on workers with messages + warnings suppressed
    ## (v2.4.2+). Without this, each worker fires "package was built under
    ## R version X.Y.Z" warnings from the user's local R / package skew on
    ## first use of mvtnorm / future / etc., once per worker. These are
    ## informational, not actionable, but they clutter user-facing output
    ## (especially under parametric bootstrap where mvtnorm::rmvnorm
    ## fires per worker).
    parallel::clusterEvalQ(cl, {
        suppressPackageStartupMessages({
            suppressWarnings({
                ## Pre-load packages used inside parallel paths.
                requireNamespace("mvtnorm",       quietly = TRUE)
                requireNamespace("future",        quietly = TRUE)
                requireNamespace("future.apply",  quietly = TRUE)
                requireNamespace("doParallel",    quietly = TRUE)
                requireNamespace("foreach",       quietly = TRUE)
            })
        })
        ## Suppress "built under R version" warnings for any subsequent
        ## library() / requireNamespace() inside this worker.
        options(warn.conflicts = FALSE)
    })
    cl
}


## Run `expr` with "package was built under R version" warnings
## suppressed (they fire from parallel workers loading packages
## compiled against a slightly newer R; informational, not actionable).
## Other warnings pass through normally. Use to wrap parallel-execution
## entry points (future_lapply / foreach calls) so user-facing output
## isn't cluttered. Added v2.4.2.
.fect_with_quiet_pkg_warnings <- function(expr) {
    withCallingHandlers(
        expr,
        warning = function(w) {
            msg <- conditionMessage(w)
            if (grepl("was built under R version", msg, fixed = TRUE)) {
                invokeRestart("muffleWarning")
            }
        }
    )
}


## Normalize the cv.method argument across all CV entry points
## (fect_cv, fect_binary_cv, fect_nevertreated, fect_mspe).
##
## Adds `"block"` as a forward-looking alias for `"all_units"` (added in
## v2.3.0; the names will be unified in v2.4.0 alongside a new `cv.units`
## parameter). Emits a one-time deprecation `message()` when the user
## passes the legacy `"all_units"` or `"treated_units"` strings, then
## maps `"block"` -> `"all_units"` so all internal code paths stay
## unchanged in this release.
##
## Returns the canonical (internal) cv.method string. `loo` and
## `rolling` pass through unchanged.
.fect_normalize_cv_method <- function(cv.method,
                                      allowed = c("rolling", "block",
                                                  "all_units",
                                                  "treated_units",
                                                  "loo")) {
    cv.method <- match.arg(cv.method, allowed)
    if (cv.method == "all_units") {
        message('Note: cv.method = "all_units" is being deprecated; ',
                'use cv.method = "block" instead. ',
                'Both names produce identical behavior in this release.')
    }
    if (cv.method == "block") {
        cv.method <- "all_units"  # internal dispatch alias
    }
    if (cv.method == "treated_units") {
        message('Note: cv.method = "treated_units" will be deprecated ',
                'when the cv.units parameter is added (v2.4.0). ',
                'The legacy name still works for now.')
    }
    cv.method
}


## --------------------------------------------------------------------------
## 2a. IFE all_units variant (notyettreated context)
##     Called from cv.R IFE loop (both serial and parallel branches)
## --------------------------------------------------------------------------
.fect_cv_score_one_ife_all <- function(
    ii,            # fold index (integer, 1..k)
    r,             # candidate rank (integer)
    YY,            # TT×N outcome matrix (masked-zero version)
    Y0CV,          # TT×N×k array of initial Y0 per fold
    X,             # TT×N×p covariate array
    II,            # TT×N observation indicator
    W.use,         # TT×N weight matrix or scalar 0
    WW,            # TT×N weight matrix (unmasked) or NULL
    beta0CV,       # p×1×k or 1×0×k array of initial beta0 per fold
    rmCV,          # list of k integer vectors (masked indices)
    estCV,         # list of k integer vectors (scored indices)
    T.on,          # TT×N matrix of event-time indicators
    force,         # integer: force FE type
    cv_tol,        # numeric tolerance for CV fits
    max.iteration, # integer
    use_weight     # integer: 0 or 1
) {
    II.cv <- II
    II.cv[rmCV[[ii]]] <- 0
    YY.cv <- YY
    YY.cv[rmCV[[ii]]] <- 0
    if (use_weight) {
        W.use2 <- W.use
        W.use2[rmCV[[ii]]] <- 0
    } else {
        W.use2 <- as.matrix(0)
    }
    est.cv.fit <- inter_fe_ub(
        YY.cv, as.matrix(Y0CV[, , ii]), X, II.cv,
        W.use2, as.matrix(beta0CV[, , ii]),
        r, force, cv_tol, max.iteration
    )$fit
    resid_ii <- YY[estCV[[ii]]] - est.cv.fit[estCV[[ii]]]
    idx_ii <- as.character(T.on[estCV[[ii]]])
    idx_ii[which(is.na(idx_ii))] <- "Control"
    obs_w_ii <- if (use_weight == 1) WW[estCV[[ii]]] else NULL
    list(resid = resid_ii, time_idx = idx_ii, obs_w = obs_w_ii)
}


## --------------------------------------------------------------------------
## 2b. MC all_units variant (notyettreated context)
##     Helper defined for symmetry; parallel branch wired in Phase 2.
## --------------------------------------------------------------------------
.fect_cv_score_one_mc_all <- function(
    ii,
    lambda_i,
    YY,
    Y0CV,
    X,
    II,
    W.use,
    WW,
    beta0CV,
    rmCV,
    estCV,
    T.on,
    force,
    cv_tol,
    max.iteration,
    use_weight
) {
    II.cv <- II; II.cv[rmCV[[ii]]] <- 0
    YY.cv <- YY; YY.cv[rmCV[[ii]]] <- 0
    W.use2 <- if (use_weight) {
        W2 <- W.use; W2[rmCV[[ii]]] <- 0; W2
    } else {
        as.matrix(0)
    }
    est.cv.fit <- inter_fe_mc(
        YY.cv, as.matrix(Y0CV[, , ii]), X, II.cv,
        W.use2, as.matrix(beta0CV[, , ii]),
        1L, lambda_i, force, cv_tol, max.iteration
    )$fit
    resid_ii <- YY[estCV[[ii]]] - est.cv.fit[estCV[[ii]]]
    idx_ii <- as.character(T.on[estCV[[ii]]]); idx_ii[is.na(idx_ii)] <- "Control"
    obs_w_ii <- if (use_weight == 1) WW[estCV[[ii]]] else NULL
    list(resid = resid_ii, time_idx = idx_ii, obs_w = obs_w_ii)
}


## --------------------------------------------------------------------------
## 2c. nevertreated IFE all_units variant
##     Replaces the inline body in fect_nevertreated.R IFE all_units foreach block
## --------------------------------------------------------------------------
.fect_cv_score_one_ife_nt_all <- function(
    ii,
    YY.co,
    Y0CV.co,
    X.co,
    II.co,
    W.use,
    W,
    beta0CV.co,
    rmCV,
    estCV,
    r,
    force,
    cv_tol,
    max.iteration
) {
    II.co.cv <- II.co; II.co.cv[rmCV[[ii]]] <- 0
    YY.co.cv <- YY.co; YY.co.cv[rmCV[[ii]]] <- 0
    W.cv <- if (!is.null(W)) {
        W2 <- W.use; W2[rmCV[[ii]]] <- 0; W2
    } else {
        as.matrix(0)
    }
    est.cv.fit <- inter_fe_ub(
        YY.co.cv, as.matrix(Y0CV.co[, , ii]), X.co, II.co.cv,
        W.cv, as.matrix(beta0CV.co[, , ii]),
        r, force, cv_tol, max.iteration
    )$fit
    resid_ii <- YY.co[estCV[[ii]]] - est.cv.fit[estCV[[ii]]]
    list(
        resid    = resid_ii,
        time_idx = rep("Control", length(resid_ii)),
        obs_w    = if (!is.null(W)) W.use[estCV[[ii]]] else NULL
    )
}


## --------------------------------------------------------------------------
## 2d. nevertreated IFE treated_units variant
##     Uses precomputed F.hat, U.tr, pre from the outer r-loop.
## --------------------------------------------------------------------------
.fect_cv_score_one_ife_nt_tr <- function(
    ii,
    U.tr,
    F.hat,
    pre,
    r,
    force,
    rmCV.tr,
    estCV.tr,
    W.tr,
    T.on,
    tr,
    TT,
    Ntr
) {
    pre.cv <- pre; pre.cv[rmCV.tr[[ii]]] <- 0
    if (r == 0) {
        if (force %in% c(1, 3)) {
            alpha.cv <- rep(0, Ntr)
            for (j in 1:Ntr) {
                pre_t <- which(pre.cv[, j] == 1)
                if (length(pre_t) > 0) alpha.cv[j] <- mean(U.tr[pre_t, j])
            }
            fitted.cv <- matrix(alpha.cv, TT, Ntr, byrow = TRUE)
        } else {
            fitted.cv <- matrix(0, TT, Ntr)
        }
    } else {
        fitted.cv <- matrix(0, TT, Ntr)
        for (j in 1:Ntr) {
            pre_t <- which(pre.cv[, j] == 1)
            if (length(pre_t) >= ncol(F.hat)) {
                F.pre <- as.matrix(F.hat[pre_t, , drop = FALSE])
                lam <- try(
                    solve(t(F.pre) %*% F.pre) %*% t(F.pre) %*% U.tr[pre_t, j],
                    silent = TRUE
                )
                if (!"try-error" %in% class(lam)) fitted.cv[, j] <- F.hat %*% lam
            }
        }
    }
    resid_ii <- U.tr[estCV.tr[[ii]]] - fitted.cv[estCV.tr[[ii]]]
    obs_w_ii <- if (!is.null(W.tr)) W.tr[estCV.tr[[ii]]] else NULL
    idx_ii <- NULL
    if (!is.null(T.on)) {
        T.on.tr <- T.on[, tr, drop = FALSE]
        idx_ii <- as.character(T.on.tr[estCV.tr[[ii]]])
        idx_ii[is.na(idx_ii)] <- "Control"
    }
    list(resid = resid_ii, time_idx = idx_ii, obs_w = obs_w_ii)
}


## --------------------------------------------------------------------------
## 2e. nevertreated CFE all_units variant
##     Same structure as 2c but calls complex_fe_ub with full CFE argument list
## --------------------------------------------------------------------------
.fect_cv_score_one_cfe_nt_all <- function(
    ii,
    YY.co,
    Y0CV.co,
    X.co,
    II.co,
    W.use,
    W,
    beta0CV.co,
    X.extra.FE.co.B,
    X.Z.co,
    X.Q.co,
    X.gamma.co,
    X.kappa.co,
    Zgamma.id,
    kappaQ.id,
    rmCV,
    estCV,
    r,
    force,
    cv_tol,
    max.iteration
) {
    II.co.cv <- II.co; II.co.cv[rmCV[[ii]]] <- 0
    YY.co.cv <- YY.co; YY.co.cv[rmCV[[ii]]] <- 0
    W.cv <- if (!is.null(W)) {
        W2 <- W.use; W2[rmCV[[ii]]] <- 0; W2
    } else {
        as.matrix(0)
    }
    est.cv.co <- complex_fe_ub(
        YY.co.cv, as.matrix(Y0CV.co[, , ii]), X.co,
        X.extra.FE.co.B, X.Z.co, X.Q.co, X.gamma.co, X.kappa.co,
        Zgamma.id, kappaQ.id,
        II.co.cv, W.cv, as.matrix(beta0CV.co[, , ii]),
        r, force = force, cv_tol, max.iteration
    )
    resid_ii <- YY.co[estCV[[ii]]] - est.cv.co$fit[estCV[[ii]]]
    list(
        resid    = resid_ii,
        time_idx = rep("Control", length(resid_ii)),
        obs_w    = if (!is.null(W)) W.use[estCV[[ii]]] else NULL
    )
}


## --------------------------------------------------------------------------
## 2f. nevertreated CFE treated_units variant
##     Same structure as 2d: F.hat, U.tr, pre passed in as arguments.
##     Body is identical to IFE treated_units — factor loadings computation
##     is independent of whether the control fit used IFE or CFE.
## --------------------------------------------------------------------------
.fect_cv_score_one_cfe_nt_tr <- function(
    ii,
    U.tr,
    F.hat,
    pre,
    r,
    force,
    rmCV.tr,
    estCV.tr,
    W.tr,
    T.on,
    tr,
    TT,
    Ntr
) {
    pre.cv <- pre; pre.cv[rmCV.tr[[ii]]] <- 0
    if (r == 0) {
        if (force %in% c(1, 3)) {
            alpha.cv <- rep(0, Ntr)
            for (j in 1:Ntr) {
                pre_t <- which(pre.cv[, j] == 1)
                if (length(pre_t) > 0) alpha.cv[j] <- mean(U.tr[pre_t, j])
            }
            fitted.cv <- matrix(alpha.cv, TT, Ntr, byrow = TRUE)
        } else {
            fitted.cv <- matrix(0, TT, Ntr)
        }
    } else {
        fitted.cv <- matrix(0, TT, Ntr)
        for (j in 1:Ntr) {
            pre_t <- which(pre.cv[, j] == 1)
            if (length(pre_t) >= ncol(F.hat)) {
                F.pre <- as.matrix(F.hat[pre_t, , drop = FALSE])
                lam <- try(
                    solve(t(F.pre) %*% F.pre) %*% t(F.pre) %*% U.tr[pre_t, j],
                    silent = TRUE
                )
                if (!"try-error" %in% class(lam)) fitted.cv[, j] <- F.hat %*% lam
            }
        }
    }
    resid_ii <- U.tr[estCV.tr[[ii]]] - fitted.cv[estCV.tr[[ii]]]
    obs_w_ii <- if (!is.null(W.tr)) W.tr[estCV.tr[[ii]]] else NULL
    idx_ii <- NULL
    if (!is.null(T.on)) {
        T.on.tr <- T.on[, tr, drop = FALSE]
        idx_ii <- as.character(T.on.tr[estCV.tr[[ii]]])
        idx_ii[is.na(idx_ii)] <- "Control"
    }
    list(resid = resid_ii, time_idx = idx_ii, obs_w = obs_w_ii)
}


## --------------------------------------------------------------------------
## 3. Rolling-window mask builder
##    Shared helper for cv.method = "rolling" across fect_cv,
##    fect_nevertreated, fect_binary_cv, and fect_mspe.
##
##    Returns a list of length `k`, each element a list with the same
##    shape as cv.sample()'s output:
##        list(cv.id  = integer,   ## flat column-major indices to mask
##             est.id = integer)   ## flat column-major indices to score
##
##    Per fold (uses set.seed(seed + fold_id) when `seed` is supplied):
##      1. Sample max(1L, round(cv.prop * n_eligible)) units from the
##         eligible pool. Eligible = controls (all observed times) +
##         treated (observed times strictly before treatment onset).
##         A unit is eligible only if it has at least min.T0 + cv.nobs
##         observations in its eligible window.
##      2. For each sampled unit, draw a random anchor index a_idx in
##         seq.int(min.T0 + 1L, n_obs - cv.nobs + 1L). The mask
##         partitions cells into:
##           - holdout: obs_t[a_idx:(a_idx + cv.nobs - 1L)]
##                      -> added to BOTH cv.id (masked) and est.id (scored)
##           - past-side buffer (if cv.buffer > 0):
##             obs_t[max(1, a_idx - cv.buffer):(a_idx - 1L)]
##                      -> added to cv.id only (masked but not scored)
##           - drop-future: obs_t[(a_idx + cv.nobs):n_obs] (if any)
##                      -> added to cv.id only (rolling-window step)
##
##    (t, j) is mapped to vec(II) position via (j - 1L) * TT + t.
##
##    II : TT x N indicator matrix (II[t, j] = 1 means unit j observed at t)
##    D  : TT x N treatment indicator (used to derive treatment-onset times)
## --------------------------------------------------------------------------
.build_cv_mask_rolling <- function(II, D, k, cv.nobs, cv.buffer, cv.prop,
                                    min.T0, seed = NULL) {
    II <- as.matrix(II)
    D  <- as.matrix(D)
    if (!all(dim(II) == dim(D))) {
        stop(".build_cv_mask_rolling: II and D must have the same shape.")
    }
    TT <- nrow(II)
    N  <- ncol(II)
    k         <- as.integer(k);         if (k         < 1L) stop("k must be >= 1.")
    cv.nobs   <- as.integer(cv.nobs);   if (cv.nobs   < 1L) stop("cv.nobs must be >= 1.")
    cv.buffer <- as.integer(cv.buffer); if (cv.buffer < 0L) stop("cv.buffer must be >= 0.")
    min.T0    <- as.integer(min.T0);    if (min.T0    < 1L) stop("min.T0 must be >= 1.")
    if (!is.finite(cv.prop) || cv.prop <= 0 || cv.prop > 1) {
        stop(".build_cv_mask_rolling: cv.prop must satisfy 0 < cv.prop <= 1.")
    }

    ## Per-unit treatment-onset time (NA when never treated).
    onset <- rep(NA_integer_, N)
    for (j in seq_len(N)) {
        treated_t <- which(D[, j] >= 1)
        if (length(treated_t) > 0L) onset[j] <- min(treated_t)
    }

    ## Per-unit eligible time vector:
    ##   - controls: all observed t (II == 1)
    ##   - treated:  observed t strictly before onset
    elig_times <- vector("list", N)
    for (j in seq_len(N)) {
        obs_t <- which(II[, j] == 1L)
        if (!is.na(onset[j])) obs_t <- obs_t[obs_t < onset[j]]
        elig_times[[j]] <- obs_t
    }

    ## Filter to units with enough eligible observations.
    elig_lengths <- vapply(elig_times, length, integer(1))
    eligible_units <- which(elig_lengths >= (min.T0 + cv.nobs))
    n_eligible <- length(eligible_units)

    if (n_eligible == 0L) {
        stop(".build_cv_mask_rolling: no eligible units have enough ",
             "observations (need >= min.T0 + cv.nobs = ",
             min.T0 + cv.nobs, ").")
    }

    n_sample_per_fold <- max(1L, as.integer(round(cv.prop * n_eligible)))

    folds <- vector("list", k)
    for (fold_id in seq_len(k)) {
        if (!is.null(seed)) set.seed(as.integer(seed) + fold_id)

        sampled <- if (n_sample_per_fold >= n_eligible) {
            eligible_units
        } else {
            eligible_units[sort(sample.int(n_eligible, n_sample_per_fold))]
        }

        cv_id_acc  <- integer(0)
        est_id_acc <- integer(0)

        for (j in sampled) {
            obs_t <- elig_times[[j]]
            n_obs <- length(obs_t)
            valid <- seq.int(min.T0 + 1L, n_obs - cv.nobs + 1L)
            if (length(valid) == 0L) next
            a_idx <- valid[sample.int(length(valid), 1L)]

            holdout_t <- obs_t[a_idx:(a_idx + cv.nobs - 1L)]
            buf_t <- if (cv.buffer > 0L) {
                bs <- max(1L, a_idx - cv.buffer)
                obs_t[bs:(a_idx - 1L)]
            } else integer(0)
            drop_t <- if ((a_idx + cv.nobs) <= n_obs) {
                obs_t[(a_idx + cv.nobs):n_obs]
            } else integer(0)

            base <- (j - 1L) * TT
            holdout_idx <- base + as.integer(holdout_t)
            buf_idx     <- if (length(buf_t)  > 0L) base + as.integer(buf_t)  else integer(0)
            drop_idx    <- if (length(drop_t) > 0L) base + as.integer(drop_t) else integer(0)

            cv_id_acc  <- c(cv_id_acc, holdout_idx, buf_idx, drop_idx)
            est_id_acc <- c(est_id_acc, holdout_idx)
        }

        folds[[fold_id]] <- list(
            cv.id  = sort(unique(cv_id_acc)),
            est.id = sort(unique(est_id_acc))
        )
    }

    folds
}

####
## diagtest
# diagnostic tests for no pre-trend, placebo effect and carryover effect
diagtest <- function(
    x, # fect object
    proportion = 0.3,
    pre.periods = NULL,
    f.threshold = NULL,
    tost.threshold = NULL,
    N_bar = NULL) {
    # get equivalence p values for two-one-sided-t tests
    tost <- function(coef, se, range) {
        z <- coef / se
        p1 <- 1 - pnorm((coef - range[1]) / se, lower.tail = TRUE) # left bound
        p2 <- 1 - pnorm((range[2] - coef) / se, lower.tail = TRUE) # right bound
        tost.p <- max(c(p1, p2))
        return(tost.p)
    }

    if (is.null(tost.threshold) == TRUE) {
        tost.threshold <- 0.36 * sqrt(x$sigma2.fect)
    }
    if (is.null(f.threshold) == TRUE) {
        f.threshold <- 0.6
    }

    # placebo test
    if (x$placeboTest == TRUE) {
        est.out <- x$est.placebo
        placebo.equiv.p <- tost(est.out[1], est.out[2], c(-tost.threshold, tost.threshold))
        out <- list(placebo.p = est.out[5], placebo.equiv.p = placebo.equiv.p)
    } # end of placebo test

    # carryover test
    if (x$carryoverTest == TRUE) {
        est.out <- x$est.carryover
        carryover.equiv.p <- tost(est.out[1], est.out[2], c(-tost.threshold, tost.threshold))
        out <- list(carryover.p = est.out[5], carryover.equiv.p = carryover.equiv.p)
    } # end of carryover test

    if (is.null(proportion) == TRUE) {
        proportion <- 0
    }

    max.pre.periods <- pre.pos <- NULL

    if (x$placeboTest == FALSE && x$carryoverTest == FALSE) {
        max.count <- max(x$count)
        max.pre.periods <- x$time[which(x$count >= max.count * proportion & x$time <= 0)]
        if (is.null(pre.periods) == TRUE) {
            pre.periods <- max.pre.periods
        } else if(length(pre.periods)>0) {
            pre.periods <- intersect(pre.periods[1]:pre.periods[length(pre.periods)], max.pre.periods)
        } else{
            pre.periods <- NA
        }
        pre.term <- pre.periods
        N_bar <- max(x$count[which(x$time %in% pre.periods)])
    }

    # testing no pre-trend
    if (x$placeboTest == FALSE & x$carryoverTest == FALSE & x$loo == TRUE) {
        max.pre.periods <- sum(x$time <= 0)
        pre.pos <- intersect(c(1:dim(x$pre.att.boot)[1]), which(x$time %in% pre.periods))
        res_boot <- x$pre.att.boot
        ## Preserve matrix shape when only one bootstrap column passes
        ## the all-non-NA filter; otherwise res_boot collapses to a
        ## vector and the row-subset on the next branch errors with
        ## "incorrect number of dimensions".
        res_boot <- res_boot[, which(apply(!is.na(res_boot), 2, all)),
                             drop = FALSE]
        if (length(pre.pos) == max.pre.periods) {
            pre.pos <- pre.pos[-1]
            #message("Cannot use full pre-treatment periods in F-test. The first period is removed.\n")
        }
        if (length(pre.pos) > 1) {
            res_boot <- res_boot[pre.pos, ]
        } else {
            res_boot <- t(as.matrix(res_boot[pre.pos, ]))
        }

        D <- as.matrix(x$pre.est.att[pre.pos, 1])
        coef_mat <- res_boot
        S <- cov(t(coef_mat)) ## * N_bar
        psi <- try(as.numeric(t(D) %*% solve(S) %*% D), silent = TRUE)
        
        if ("try-error" %in% class(psi)) {
            message("\n")
            #message("The estimated covariance matrix is irreversible.")
            message("F-test could not be computed: the estimated covariance matrix is singular.")
            message("\n")
            f.stat <- f.p <- f.equiv.p <- f.threshold <- NA
        } else {
            scale <- (N_bar - length(pre.pos)) / ((N_bar - 1) * length(pre.pos))
            ## F statistic

            if (scale <= 0) {
                message("Can't calculate the F statistic because of insufficient treated units.\n")
                f.stat <- NA
                f.p <- NA
                f.equiv.p <- NA
            } else {
                f.stat <- psi * scale
                f.p <- pf(f.stat,
                    df1 = length(pre.pos), df2 = N_bar - length(pre.pos),
                    lower.tail = FALSE
                )

                ## Equivalent F test
                f.equiv.p <- pf(f.stat,
                    df1 = length(pre.pos), df2 = N_bar - length(pre.pos),
                    ncp = N_bar * f.threshold
                )
            }
        }

        # TOST
        est.att <- x$pre.est.att[, c(1:2)]
        tost.equiv.p <- sapply(1:nrow(est.att), function(i) {
            return(tost(est.att[i, 1], est.att[i, 2], c(-tost.threshold, tost.threshold)))
        }) # keep the maximum p value

        tost.equiv.p <- max(tost.equiv.p)
        out <- list(
            f.stat = f.stat,
            f.p = f.p,
            f.threshold = f.threshold,
            f.equiv.p = f.equiv.p,
            df1 = length(pre.pos),
            df2 = N_bar - length(pre.pos),
            N_bar = N_bar,
            tost.equiv.p = tost.equiv.p,
            tost.threshold = tost.threshold
        )
    }

    if (x$placeboTest == FALSE & x$carryoverTest == FALSE & x$loo == FALSE) {
        max.pre.periods <- sum(x$time <= 0)
        pre.pos <- which(x$time %in% pre.periods)
        if (length(pre.pos) == max.pre.periods) {
            pre.pos <- pre.pos[-1]
            #message("Cannot use full pre-treatment periods in the F test. The first period is removed.\n")
        }

        res_boot <- x$att.boot
        ## drop = FALSE keeps res_boot a matrix even when only one
        ## bootstrap column survives the all-non-NA filter (can occur
        ## when small nboots interacts with NA-filled iterations from
        ## parallel-worker errors).
        res_boot <- res_boot[, which(apply(!is.na(res_boot), 2, all)),
                             drop = FALSE]
        nboots <- ncol(res_boot)
        if (length(pre.pos) > 1) {
            res_boot <- res_boot[pre.pos, ]
        } else {
            res_boot <- t(as.matrix(res_boot[pre.pos, ]))
        }

        D <- as.matrix(x$est.att[pre.pos, 1])
        coef_mat <- res_boot

        S <- cov(t(coef_mat)) ## * N_bar

        psi <- try(as.numeric(t(D) %*% solve(S) %*% D), silent = TRUE)
        if ("try-error" %in% class(psi)) {
            message("\n")
            #message("The estimated covariance matrix is irreversible.")
            message("F-test could not be computed: the estimated covariance matrix is singular.")
            message("\n")
            f.stat <- f.p <- f.equiv.p <- f.threshold <- NA
        } else {
            scale <- (N_bar - length(pre.pos)) / ((N_bar - 1) * length(pre.pos))
            ## F statistic
            if (scale <= 0) {
                message("Can't calculate the F statistic because of insufficient treated units.\n")
                f.stat <- NA
                f.p <- NA
                f.equiv.p <- NA
            } else {
                f.stat <- psi * scale
                f.p <- pf(f.stat,
                    df1 = length(pre.pos), df2 = N_bar - length(pre.pos),
                    lower.tail = FALSE
                )

                ## Equivalent F test
                f.equiv.p <- pf(f.stat,
                    df1 = length(pre.pos), df2 = N_bar - length(pre.pos),
                    ncp = N_bar * f.threshold
                )
            }
        }

        # TOST

        est.att <- x$est.att[, c(1:2)]
        pos.zero <- which(x$time == 0)
        est.att <- est.att[pre.pos, , drop = FALSE]
        if (dim(est.att)[1] > 0) {
            tost.equiv.p <- sapply(1:nrow(est.att), function(i) {
                return(tost(est.att[i, 1], est.att[i, 2], c(-tost.threshold, tost.threshold)))
            }) # keep the maximum p value
            tost.equiv.p <- max(tost.equiv.p)
        } else {
            tost.equiv.p <- NA
        }


        out <- list(
            f.stat = f.stat,
            f.p = f.p,
            f.threshold = f.threshold,
            f.equiv.p = f.equiv.p,
            df1 = length(pre.pos),
            df2 = N_bar - length(pre.pos),
            N_bar = N_bar,
            tost.equiv.p = tost.equiv.p,
            tost.threshold = tost.threshold
        )
    }

    # end of testing no pre-trend

    return(out)
}

## an auxiliary function for estimating cumulative or sub-group treatment effect based on fect

## 1. cumulative effect for a specified event window
## 2. averaging treatment effect in a sub-group

effect <- function(x, ## a fect object
                   cumu = TRUE, ## whether to calculate cumulative effect
                   id = NULL, ## units to be averaged on
                   period = NULL, ## event window
                   plot = FALSE,
                   count = TRUE,
                   xlab = NULL,
                   ylab = NULL,
                   main = NULL) {
  ## v2.4.0: soft-deprecated in favor of estimand(). Continues to work
  ## byte-identical to today; one-time-per-session message points users
  ## at the unified API. Removal not before v3.0.0. Suppressed when
  ## called internally from .compute_att_cumu_event_time().
  .fect_deprecation_message_once("effect")

  if (is.null(x$eff.boot)){
    stop("No bootstrap/jackknife results available. Choose keep.sims = TRUE in fect().")
  }
  if (x$hasRevs) {
    warning("Cumulative effects are not well-defined for panels with treatment reversals.")
    return() # Return NA
  }

  # Select units for analysis
  if (is.null(id)) {
    # If no specific units provided, select all treated units
    mask <- (colSums(x$D.dat) > 0)
  } else {
    # Otherwise, select specified units
    mask <- (colnames(x$eff) %in% id)
  }

  # Extract relevant matrices for selected units
  eff <- x$eff[, mask]      # Treatment effects
  D <- x$D.dat[, mask]      # Treatment indicators
  I <- x$I.dat[, mask]      # Inclusion (non-missing) indicators
  inference <- x$call$vartype  # Inference type
  method <- x$method        # Method

  # Get dimensions of data
  TT <- dim(eff)[1]  # Number of time periods
  N <- dim(eff)[2]   # Number of units

  # Replace NA values in treatment indicators with zeros
  if (sum(is.na(c(D))) > 0) {
    D[which(is.na(D))] <- 0
  }

  # Store original treatment matrix
  D.old <- D

  # Calculate cumulative treatment for determining time periods
  D <- apply(D, 2, function(vec) {
    cumsum(vec)
  })

  # Find minimum pre-treatment period across all units
  minT0 <- min(apply(D, 2, function(vec) {
    sum(vec == 0)
  }))

  # Restore original treatment matrix
  D <- D.old

  # Set default period range if not provided
  period.raw <- c(1, TT - minT0)
  if (!is.null(period)) {
    # Validate user-provided period
    if (period[2] > period.raw[2]) {
      stop(paste("Ending period should not be greater than ", period.raw[2], sep = ""))
    }
  } else {
    period <- period.raw
  }

  # Calculate cumulative average treatment effect
  catt <- getEffect(D, I, eff, cumu, period)

  # Initialize bootstrap results
  catt.boot <- NULL

  # Check if uncertainty estimates are available
  if (is.null(x$est.avg)) {
    cat("No uncertainty estimates.")
  } else {
    # Determine nboots from the 3D arrays
    if (!is.null(x$eff.boot) && length(dim(x$eff.boot)) == 3) {
      nboots <- dim(x$eff.boot)[3]
    } else {
      stop("eff.boot is not a valid 3D array. Ensure keep.sims = TRUE in fect().")
    }
    if (nboots == 0) {
      stop("All bootstrap iterations failed. Cannot compute cumulative effects.")
    }

    # Validate D.boot and I.boot exist with matching dimensions
    has.D.boot <- !is.null(x$D.boot) && length(dim(x$D.boot)) == 3 && dim(x$D.boot)[3] >= nboots
    has.I.boot <- !is.null(x$I.boot) && length(dim(x$I.boot)) == 3 && dim(x$I.boot)[3] >= nboots

    catt.boot <- matrix(NA, period[2] - period[1] + 1, nboots)

    # Calculate treatment effect for each bootstrap sample
    for (i in 1:nboots) {
      # Extract bootstrap matrices
      if (has.D.boot) {
        D.boot <- x$D.boot[, , i]
        I.boot <- x$I.boot[, , i]
      } else {
        # Fallback: use original D.dat and I.dat (less accurate but prevents crash)
        D.boot <- x$D.dat
        I.boot <- x$I.dat
      }
      eff.boot <- x$eff.boot[, , i]

      # Select treated units in bootstrap sample
      if (is.null(id)){
        mask.boot <- (colSums(D.boot) > 0)
      } else {
        mask.boot <- (x$id[unlist(x$colnames.boot[i])] %in% id)
      }

      # Extract relevant matrices for selected units
      Itr.boot <- I.boot[, mask.boot]
      Dtr.boot <- D.boot[, mask.boot]
      eff.tr.boot <- eff.boot[, mask.boot]

      # Calculate treatment effect for this bootstrap sample
      catt.boot[, i] <- getEffect(
        as.matrix(Dtr.boot),
        as.matrix(Itr.boot),
        as.matrix(eff.tr.boot),
        cumu, period)
    }
  }

  # Function to calculate p-values for non-parametric test
  get.pvalue <- function(vec) {
    if (NaN %in% vec | NA %in% vec) {
      # Handle NaN and NA values
      nan.pos <- is.nan(vec)
      na.pos <- is.na(vec)
      pos <- c(which(nan.pos), which(na.pos))
      vec.a <- vec[-pos]
      # Calculate two-sided p-values
      a <- sum(vec.a >= 0) / (length(vec) - sum(nan.pos | na.pos)) * 2
      b <- sum(vec.a <= 0) / (length(vec) - sum(nan.pos | na.pos)) * 2
    } else {
      # Calculate two-sided p-values for complete data
      a <- sum(vec >= 0) / length(vec) * 2
      b <- sum(vec <= 0) / length(vec) * 2
    }
    # Return minimum p-value, capped at 1
    return(min(as.numeric(min(a, b)), 1))
  }

  # Calculate confidence intervals and statistics if bootstrap results exist
  if (!is.null(catt.boot)) {
    # Check if inference method is jackknife
    is_jackknife <- !is.null(inference) && inference == "jackknife"
    is_parametric <- !is.null(inference) && inference == "parametric"

    # Calculate standard errors with proper scaling for jackknife
    if (is_jackknife) {
      # For jackknife, scale by sqrt(N-1)
      N_samples <- ncol(catt.boot)
      jackknife_scale <- sqrt(N_samples - 1)
      se.att <- apply(catt.boot, 1, function(vec) sd(vec, na.rm = TRUE) * jackknife_scale)
    } else {
      # Standard calculation for bootstrap
      se.att <- apply(catt.boot, 1, function(vec) sd(vec, na.rm = TRUE))
    }

    # Calculate 95% confidence intervals
    if (is_jackknife || is_parametric) {
      # For jackknife, use t-distribution with N-1 degrees of freedom
      N_samples <- ncol(catt.boot)
      t_critical <- stats::qt(0.975, df = N_samples - 1)
      CI.att <- t(apply(cbind(catt, se.att), 1, function(row) {
        c(row[1] - t_critical * row[2], row[1] + t_critical * row[2])
      }))
    } else {
      # For bootstrap, use empirical quantiles
      CI.att <- t(apply(catt.boot, 1, function(vec) {
        quantile(vec, c(0.025, 0.975), na.rm = TRUE)
      }))
    }

    # Calculate p-values
    if (is_jackknife || is_parametric) {
      # For jackknife, use t-distribution for p-values
      N_samples <- ncol(catt.boot)
      pvalue.att <- sapply(1:nrow(catt.boot), function(i) {
        t_stat <- catt[i] / se.att[i]
        2 * stats::pt(-abs(t_stat), df = N_samples - 1)
      })
    } else {
      # For bootstrap, use empirical distribution
      pvalue.att <- apply(catt.boot, 1, get.pvalue)
    }
    # Combine results into a matrix
    est.catt <- cbind(catt, se.att, CI.att, pvalue.att)
    colnames(est.catt) <- c("ATT", "S.E.", "CI.lower", "CI.upper", "p.value")
    rownames(est.catt) <- period[1]:period[2]
  }

  # Plot
  if (plot) {
    # Create a data frame for plotting
    plot_data <- data.frame(
      time = as.numeric(rownames(est.catt)),
      catt = est.catt[, "ATT"],
      ci_lower = est.catt[, "CI.lower"],
      ci_upper = est.catt[, "CI.upper"]
    )

    # Add count data for the bar chart at the bottom
    time_range <- unique(plot_data$time)
    # Calculate treated units at each relative time point
    # First re-create the relative time matrix (similar to what getEffect does)
    D_rel <- D.old  # Start with original treatment matrix
    for (i in 1:ncol(D_rel)) {
      cumD <- cumsum(D_rel[, i])
      t0 <- sum(cumD == 0)  # Find time of first treatment
      if (t0 > 0) {  # Only process treated units
        D_rel[, i] <- 1:nrow(D_rel) - t0  # Convert to relative time
      } else {
        D_rel[, i] <- NA  # Not treated
      }
    }

    # Count treated units at each time point
    count_data <- data.frame(
      time = time_range,
      count = sapply(time_range, function(t) {
        sum(D_rel == t, na.rm = TRUE)
      })
    )

    # Calculate rectangle dimensions for count display with more space between bars and main plot
    y_range <- diff(range(plot_data$ci_lower, plot_data$ci_upper, na.rm = TRUE))
    rect.min <- min(plot_data$ci_lower, na.rm = TRUE) - 0.3 * y_range
    rect.length <- 0.15 * y_range

    T.start <- time_range - 0.25
    T.end <- time_range + 0.25
    ymin <- rep(rect.min, length(time_range))
    ymax <- rect.min + rect.length * count_data$count / max(count_data$count)

    data.toplot <- data.frame(
      xmin = T.start,
      xmax = T.end,
      ymin = ymin,
      ymax = ymax
    )

    # xlab, ylab, and main
    if (is.null(xlab) == TRUE) {xlab <- "Time"}
    if (is.null(ylab) == TRUE) {
      if (cumu == TRUE) {
        ylab <- "Cumulative ATT"
      } else {
        ylab <- "ATT"
      }
    }
    if (is.null(main) == TRUE) {
      if (cumu == TRUE) {
        main <- "Average Cumulative Treatment Effects on the Treated"
      } else {
        main <- "Average Treatment Effects on the Treated"
      }
    }

    # Create the plot
    p <- ggplot() +
      geom_point(data = plot_data, aes(x = .data$time, y = .data$catt), size = 2) +
      geom_hline(yintercept = 0, linetype = "dashed", color = "gray50") +
      geom_linerange(data = plot_data,
                     aes(x = .data$time, ymin = .data$ci_lower, ymax = .data$ci_upper),
                     size = 0.5) +
      labs(x = xlab, y = ylab, title = main) +
      theme_bw() +
      theme(
        panel.grid.minor = element_blank(),
        panel.grid.major.x = element_blank(),
        axis.title = element_text(size = 11),
        axis.title.x = element_text(margin = margin(t = 8, r = 0, b = 0, l = 0)),
        axis.title.y = element_text(margin = margin(t = 0, r = 8, b = 0, l = 0)),
        axis.text = element_text(color = "black", size = 10),
        plot.title = element_text(size = 14, hjust = 0.5, face = "bold", margin = margin(10, 0, 10, 0))
      ) +
      # Force integer breaks on x-axis
      scale_x_continuous(breaks = function(x) seq(floor(x[1]), ceiling(x[2]), by = 1))

    if (count == TRUE) {
      p <- p + geom_rect(data = data.toplot,
                         aes(xmin = .data$xmin, xmax = .data$xmax, ymin = .data$ymin, ymax = .data$ymax),
                         fill = "gray50", alpha = 0.3, size = 0.3, color = "black") +
        annotate("text",
                 x = time_range[which.max(count_data$count)],
                 y = max(data.toplot$ymax) + 0.05 * y_range,
                 label = max(count_data$count),
                 size = 3, hjust = 0.5)
    }
    # Print the plot
    print(p)
  }
  # Prepare output
  out = x
  out$effect.est.avg <- catt
  if (!is.null(catt.boot)) {
    # Add estimation results to output (commented out line would also include bootstrap samples)
    out$effect.est.att = est.catt
  }
  class(out) <- "fect"
  return(out)
}


getEffect <- function(D,           # Treatment indicator matrix
                      I,           # Inclusion indicator matrix
                      eff,         # Effect matrix
                      cumu,        # Logical: whether to calculate cumulative effect
                      period) {    # Event window range: c(start, end)
  # Initialize output vector with NAs
  aeff <- rep(NA, period[2] - period[1] + 1)

  # Return empty result if all D values are NA
  if (sum(is.na(D)) == length(c(D))) {
    return(aeff)
  }

  # Replace NA values in D with 0
  if (sum(is.na(c(D))) > 0) {
    D[which(is.na(D))] <- 0
  }

  # Calculate cumulative sum of treatment for each unit
  D <- apply(D, 2, function(vec) {
    cumsum(vec)
  })

  TT <- dim(D)[1]  # Number of time periods
  N <- dim(D)[2]   # Number of units

  # Calculate relative time to treatment for each unit
  for (i in 1:N) {
    subd <- D[, i]
    t0 <- sum(subd == 0)  # Pre-treatment periods
    D[, i] <- 1:TT - t0   # Convert to relative time
  }

  # Set treatment indicators to NA where units are not included
  if (sum(c(I) == 0) > 0) {
    D[which(I == 0)] <- NA
  }

  # Flatten matrices to vectors for processing
  vd <- c(D)       # Vector of relative times
  veff <- c(eff)   # Vector of effects

  # Remove NA entries
  if (sum(is.na(vd)) > 0) {
    vd.rm <- which(is.na(vd))
    vd <- vd[-vd.rm]
    veff <- veff[-vd.rm]
  }

  # Get unique relative time periods
  uniT <- sort(unique(vd))

  # Define effective time range
  ts <- period[1]  # Start period
  te <- min(period[2], max(vd))  # End period (capped by available data)
  effT <- ts:te    # Effective time range

  if (cumu == TRUE) {
    # Calculate cumulative treatment effect
    if (sum(!(effT %in% uniT)) == 0) {
      pos <- c()
      for (i in 1:length(effT)) {
        # Accumulate positions for all periods up to current one
        pos <- c(pos, which(vd == effT[i]))
        # Calculate cumulative effect as mean effect times number of periods
        aeff[i] <- mean(veff[pos]) * i
      }
    }
  } else {
    # Calculate average treatment effect for each period
    ave <- as.numeric(tapply(veff, vd, mean))
    effT2 <- period[1]:period[2]
    for (i in 1:length(effT2)) {
      if (effT2[i] %in% uniT) {
        aeff[i] <- ave[which(uniT == effT2[i])]
      }
    }
  }
  return(aeff)  # Return vector of treatment effects
}

## ---------------------------------------------------------------
## Tests for utility functions: fect_mspe,
## esplot (parameter variations), att.cumu, and effect.
## ---------------------------------------------------------------

## Shared fixtures (file-scope) — fitted once, reused across blocks.
suppressWarnings(data("simdata", package = "fect"))

## Basic fect output (no SE, fast)
out_base <- suppressWarnings(suppressMessages(
  fect::fect(
    Y ~ D + X1 + X2,
    data    = simdata,
    index   = c("id", "time"),
    method  = "ife",
    r       = 1,
    CV      = FALSE,
    se      = FALSE,
    parallel = FALSE
  )
))

## Fect output with bootstrap simulations (needed for att.cumu / effect / esplot CI)
out_boot <- suppressWarnings(suppressMessages(
  fect::fect(
    Y ~ D + X1 + X2,
    data      = simdata,
    index     = c("id", "time"),
    method    = "ife",
    r         = 1,
    CV        = FALSE,
    se        = TRUE,
    nboots    = 20,
    keep.sims = TRUE,
    parallel  = FALSE
  )
))

## Subset without treatment reversals — needed for effect() which
## refuses to compute cumulative effects with reversals.
no_rev_ids <- {
  splits <- split(simdata$D, simdata$id)
  as.integer(names(which(vapply(splits, function(x) all(diff(x) >= 0), logical(1)))))
}
simdata_norev <- simdata[simdata$id %in% no_rev_ids, ]

out_norev <- suppressWarnings(suppressMessages(
  fect::fect(
    Y ~ D + X1 + X2,
    data      = simdata_norev,
    index     = c("id", "time"),
    method    = "ife",
    r         = 1,
    CV        = FALSE,
    se        = TRUE,
    nboots    = 20,
    keep.sims = TRUE,
    parallel  = FALSE
  )
))

## -----------------------------------------------------------------
## 1. fect_mspe
## -----------------------------------------------------------------
test_that("fect_mspe returns correct structure and values", {

  skip_on_cran()
  ## Basic invocation (cv.sample-based API — no hide_n / n_rep / hide_mask)
  result <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42)
  ))

  expect_type(result, "list")
  expect_true(all(c("summary", "records", "fits", "criterion", "scores") %in% names(result)))
  expect_s3_class(result$summary, "data.frame")
  expect_true(all(c("Model", "Hidden_N", "RMSE", "Bias", "MSPE") %in% names(result$summary)))
  expect_s3_class(result$records, "data.frame")
  expect_true(all(c("Rep", "Model", "Hidden_N", "RMSE", "Bias", "MSPE") %in% names(result$records)))
  expect_true(result$summary$RMSE > 0)

  ## Reproducibility: same seed => same RMSE
  r1 <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42)
  ))
  r2 <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42)
  ))
  expect_equal(r1$summary$RMSE, r2$summary$RMSE, tolerance = 1e-10)

  ## Multiple models
  multi <- suppressWarnings(suppressMessages(
    fect_mspe(list(ife = out_base, ife2 = out_base), seed = 42)
  ))
  expect_equal(nrow(multi$summary), 2)
  expect_true(all(c("ife", "ife2") %in% multi$summary$Model))
})


## -----------------------------------------------------------------
## 2. esplot parameter variations
## -----------------------------------------------------------------
test_that("esplot handles parameter variations correctly", {

  skip_on_cran()
  ## Manual test data
  es_data <- data.frame(
    time = -3:3,
    ATT  = c(-0.1, 0.05, -0.02, 0, 0.5, 0.8, 1.2),
    SE   = rep(0.2, 7),
    count = c(100, 100, 100, 100, 80, 60, 40)
  )

  ## Connected mode with SE-derived CIs
  p1 <- suppressMessages(
    esplot(es_data, Period = "time", Estimate = "ATT", SE = "SE", connected = TRUE)
  )
  expect_s3_class(p1, "ggplot")

  ## SE-derived CI (no CI.lower/CI.upper columns in data)
  p2 <- suppressMessages(
    esplot(es_data, Period = "time", Estimate = "ATT", SE = "SE")
  )
  expect_s3_class(p2, "ggplot")

  ## show.count
  p3 <- suppressMessages(
    esplot(es_data, Period = "time", Estimate = "ATT", SE = "SE",
           Count = "count", show.count = TRUE)
  )
  expect_s3_class(p3, "ggplot")

  ## highlight.periods
  p4 <- suppressMessages(
    esplot(es_data, Period = "time", Estimate = "ATT", SE = "SE",
           highlight.periods = c(-1, 1), highlight.colors = c("red", "blue"))
  )
  expect_s3_class(p4, "ggplot")

  ## fill.gap — create data with a gap (skip time = 0)
  es_gap <- es_data[es_data$time != 0, ]
  p5 <- suppressMessages(
    esplot(es_gap, Period = "time", Estimate = "ATT", SE = "SE", fill.gap = TRUE)
  )
  expect_s3_class(p5, "ggplot")
  ## fill.gap should have added a row for the missing period
  expect_true(nrow(p5$data) > nrow(es_gap))

  ## start0
  p6 <- suppressMessages(
    esplot(es_data, Period = "time", Estimate = "ATT", SE = "SE", start0 = TRUE)
  )
  expect_s3_class(p6, "ggplot")

  ## only.pre
  p7 <- suppressMessages(
    esplot(es_data, Period = "time", Estimate = "ATT", SE = "SE", only.pre = TRUE)
  )
  expect_s3_class(p7, "ggplot")

  ## only.post
  p8 <- suppressMessages(
    esplot(es_data, Period = "time", Estimate = "ATT", SE = "SE", only.post = TRUE)
  )
  expect_s3_class(p8, "ggplot")

  ## From fect output
  p9 <- suppressMessages(
    esplot(out_boot$est.att, Estimate = "ATT")
  )
  expect_s3_class(p9, "ggplot")
})


## -----------------------------------------------------------------
## 4. att.cumu and effect — cumulative treatment effects
## -----------------------------------------------------------------
test_that("att.cumu and effect produce valid cumulative ATT", {

  skip_on_cran()
  ## ---- att.cumu (works with reversals) ----
  c1 <- att.cumu(out_boot, period = c(1, 3), plot = FALSE)
  expect_true(is.matrix(c1))
  expect_equal(nrow(c1), 3)
  expect_true("catt" %in% colnames(c1))
  ## All start values equal period[1]
  expect_true(all(c1[, "start"] == 1))
  ## end column goes from 1 to 3
  expect_equal(as.numeric(c1[, "end"]), 1:3)
  ## Values are finite
  expect_true(all(is.finite(c1[, "catt"])))

  ## SE columns present (since out_boot has se=TRUE)
  expect_true(all(c("S.E.", "CI.lower", "CI.upper", "p.value") %in% colnames(c1)))

  ## att.cumu with different period
  c2 <- att.cumu(out_boot, period = c(1, 5), plot = FALSE)
  expect_equal(nrow(c2), 5)

  ## ---- effect (requires no treatment reversals) ----
  ## effect() warns and returns NULL with reversals
  expect_warning(
    e_rev <- effect(out_boot, cumu = TRUE, period = c(1, 3), plot = FALSE),
    "reversals"
  )
  expect_null(e_rev)

  ## Use no-reversal subset for effect() tests
  e1 <- suppressWarnings(
    effect(out_norev, cumu = TRUE, period = c(1, 3), plot = FALSE)
  )
  expect_s3_class(e1, "fect")
  expect_true(!is.null(e1$effect.est.avg))
  expect_equal(length(e1$effect.est.avg), 3)
  expect_true(all(is.finite(e1$effect.est.avg)))
  expect_true(!is.null(e1$effect.est.att))
  expect_true(all(c("ATT", "S.E.", "CI.lower", "CI.upper", "p.value") %in%
                    colnames(e1$effect.est.att)))
  expect_true(all(is.finite(e1$effect.est.att)))

  ## effect non-cumulative
  e2 <- suppressWarnings(
    effect(out_norev, cumu = FALSE, period = c(1, 3), plot = FALSE)
  )
  expect_s3_class(e2, "fect")
  expect_true(!is.null(e2$effect.est.avg))
  expect_true(all(is.finite(e2$effect.est.avg)))

  ## Both att.cumu and effect produce finite output for the same window
  c_norev <- att.cumu(out_norev, period = c(1, 3), plot = FALSE)
  expect_true(all(is.finite(c_norev[, "catt"])))
  expect_true(all(is.finite(e1$effect.est.avg)))

  ## effect with unit subset
  treated_ids <- out_norev$id[colSums(out_norev$D.dat) > 0]
  if (length(treated_ids) >= 5) {
    e3 <- suppressWarnings(
      effect(out_norev, id = treated_ids[1:5], period = c(1, 3), plot = FALSE)
    )
    expect_s3_class(e3, "fect")
    expect_true(!is.null(e3$effect.est.avg))
  }
})

## ---------------------------------------------------------------
## Tests for soft-deprecation messages on effect() and att.cumu().
##
## Locked behavior per ref/po-estimands-contract.md §7:
## - Calling effect() or att.cumu() emits a one-time-per-session
##   message pointing at ?estimand.
## - Subsequent calls in the same session do NOT re-emit.
## - Calling estimand("att.cumu", ...) (which delegates internally)
##   does NOT trigger the message.
## ---------------------------------------------------------------

.fit_no_reversal <- function(nboots = 30, keep.sims = TRUE) {
  set.seed(1); N <- 30; TT <- 20
  df <- expand.grid(id = 1:N, time = 1:TT)
  treat_start <- sample(c(NA, 8:15), N, replace = TRUE)
  df$D <- ifelse(is.na(treat_start[df$id]) | df$time < treat_start[df$id],
                 0, 1)
  df$Y <- rnorm(nrow(df)) + 0.5 * df$D + 0.1 * df$time
  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "fe", force = "two-way",
      se = TRUE, nboots = nboots, parallel = FALSE,
      keep.sims = keep.sims
    )
  ))
}


## Reset the once-per-session flags before each test so messaging
## behavior is reproducible regardless of test ordering.
.reset_dep_flags <- function() {
  options(
    fect.deprecated_warned_effect   = FALSE,
    fect.deprecated_warned_att.cumu = FALSE,
    fect.suppress_estimand_deprecation = FALSE
  )
}


## -- DEP.1  effect() emits a deprecation message on first call -------

test_that("DEP.1: effect() emits a one-time deprecation message", {

  skip_on_cran()

  .reset_dep_flags()
  fit <- .fit_no_reversal()
  expect_message(
    out <- fect::effect(fit, cumu = TRUE, plot = FALSE),
    "soft-deprecated",
    fixed = FALSE
  )
  ## Subsequent call: no message (one-time gate).
  expect_no_message(
    out2 <- fect::effect(fit, cumu = TRUE, plot = FALSE)
  )
})


## -- DEP.2  att.cumu() emits a deprecation message on first call -----

test_that("DEP.2: att.cumu() emits a one-time deprecation message", {

  skip_on_cran()

  .reset_dep_flags()
  fit <- .fit_no_reversal()
  expect_message(
    out <- fect::att.cumu(fit, period = c(1, 5), plot = FALSE),
    "soft-deprecated",
    fixed = FALSE
  )
  expect_no_message(
    out2 <- fect::att.cumu(fit, period = c(1, 5), plot = FALSE)
  )
})


## -- DEP.3  estimand("att.cumu") does NOT trigger deprecation msg ----

test_that("DEP.3: estimand('att.cumu') is silent (no internal deprecation)", {

  skip_on_cran()

  .reset_dep_flags()
  fit <- .fit_no_reversal()
  ## estimand() delegates to effect()/att.cumu() internally, with the
  ## suppression flag set. No message expected.
  expect_no_message(
    fect::estimand(fit, "att.cumu", "event.time")
  )
  expect_no_message(
    fect::estimand(fit, "att.cumu", "overall", window = c(1, 5))
  )

  ## After estimand() returns, the suppression flag is reset, so a
  ## direct effect() call DOES emit the message (one-time gate intact).
  expect_message(
    fect::effect(fit, cumu = TRUE, plot = FALSE),
    "soft-deprecated",
    fixed = FALSE
  )
})


## -- DEP.4  Numerical equality preserved post-deprecation ------------

test_that("DEP.4: deprecation does not change numerical output", {

  skip_on_cran()

  .reset_dep_flags()
  fit <- .fit_no_reversal()
  out_eff <- suppressMessages(fect::effect(fit, cumu = TRUE, plot = FALSE))
  out_acc <- suppressMessages(
    fect::att.cumu(fit, period = c(1, 5), plot = FALSE)
  )

  expect_false(is.null(out_eff$effect.est.att))
  expect_false(is.null(out_acc))
  ## Post-deprecation outputs are still byte-identical to estimand().
  est <- fect::estimand(fit, "att.cumu", "event.time")
  est_overall <- fect::estimand(fit, "att.cumu", "overall",
                                 window = c(1, 5))

  shared_et <- intersect(est$event.time,
                         as.numeric(rownames(out_eff$effect.est.att)))
  expect_true(length(shared_et) > 0)
})

## ---------------------------------------------------------------
## Tests for estimand(fit, "att.cumu", ...).
##
## Locked invariants per ref/po-estimands-contract.md §2:
##   estimand(fit, "att.cumu", "event.time") matches effect(fit) on
##     catt / S.E. / CI.lower / CI.upper.
##   estimand(fit, "att.cumu", "overall", window = period) matches the
##     final row of att.cumu(fit, period = period) on catt / S.E. / CIs.
##
## simdata has treatment reversals so effect() / att.cumu() warn or
## error on it. We construct a small canonical no-reversal panel
## inline for the byte-equality tests.
## ---------------------------------------------------------------

.canonical_no_reversal_panel <- function(N = 30, TT = 20, seed = 1) {
  set.seed(seed)
  df <- expand.grid(id = 1:N, time = 1:TT)
  treat_start <- sample(c(NA, 8:15), N, replace = TRUE)
  df$D <- ifelse(is.na(treat_start[df$id]) | df$time < treat_start[df$id],
                 0, 1)
  df$Y <- rnorm(nrow(df)) + 0.5 * df$D + 0.1 * df$time
  df
}

.fit_no_reversal <- function(nboots = 50, keep.sims = TRUE) {
  df <- .canonical_no_reversal_panel()
  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "fe", force = "two-way",
      se = TRUE, nboots = nboots, parallel = FALSE,
      keep.sims = keep.sims
    )
  ))
}


## -- AC.1  Schema: returns documented columns -----------------------

test_that("AC.1: estimand(fit, 'att.cumu', 'event.time') has the documented schema", {

  skip_on_cran()

  fit <- .fit_no_reversal()
  est <- fect::estimand(fit, "att.cumu", "event.time")

  expect_s3_class(est, "data.frame")
  expected_cols <- c("event.time", "estimate", "se",
                     "ci.lo", "ci.hi", "n_cells", "vartype")
  expect_setequal(names(est), expected_cols)

  expect_true(all(est$event.time >= 1))
  expect_true(is.numeric(est$estimate))
})


## -- AC.2  Backcompat: matches effect(fit) numerically (LOCKED INVARIANT) --

test_that("AC.2: estimand(fit, 'att.cumu', 'event.time') matches effect(fit)", {

  skip_on_cran()

  fit <- .fit_no_reversal()
  ## ci.method = "percentile" replicates the raw-quantile CI used by the
  ## legacy effect() function. v2.4.2 changed the default to "basic"
  ## (reflected pivot CI per Davison-Hinkley 1997 §5.2.1), so byte-equality
  ## with effect() requires the explicit override.
  est <- fect::estimand(fit, "att.cumu", "event.time",
                        ci.method = "percentile")
  out_eff <- suppressWarnings(suppressMessages(
    fect::effect(fit, cumu = TRUE, plot = FALSE)
  ))

  ## effect() returns the augmented fit with the cumulative matrix at
  ## $effect.est.att (columns ATT / S.E. / CI.lower / CI.upper / p.value,
  ## rownames = event time).
  expect_false(is.null(out_eff))
  expect_false(is.null(out_eff$effect.est.att))
  M <- out_eff$effect.est.att
  eff_df <- as.data.frame(M)
  eff_df$event.time <- as.numeric(rownames(M))
  eff_df <- eff_df[order(eff_df$event.time), ]
  est_sorted <- est[order(est$event.time), ]

  ## Match on the post-treatment event times that effect() reports.
  shared <- intersect(eff_df$event.time, est_sorted$event.time)
  expect_true(length(shared) > 0)

  eff_sub <- eff_df[eff_df$event.time %in% shared, ]
  est_sub <- est_sorted[est_sorted$event.time %in% shared, ]

  expect_equal(est_sub$estimate, unname(eff_sub$ATT),      tolerance = 1e-8)
  expect_equal(est_sub$se,       unname(eff_sub$S.E.),     tolerance = 1e-8)
  expect_equal(est_sub$ci.lo,    unname(eff_sub$CI.lower), tolerance = 1e-8)
  expect_equal(est_sub$ci.hi,    unname(eff_sub$CI.upper), tolerance = 1e-8)
})


## -- AC.3  Backcompat: matches att.cumu(fit, period) on overall window --

test_that("AC.3: estimand(fit, 'att.cumu', 'overall', window) matches att.cumu()", {

  skip_on_cran()

  fit <- .fit_no_reversal()
  period <- c(1, 5)

  ## ci.method = "percentile" replicates the raw-quantile CI used by the
  ## legacy att.cumu() function (see AC.2 comment for context).
  est <- fect::estimand(fit, "att.cumu", "overall", window = period,
                        ci.method = "percentile")
  out_acc <- suppressWarnings(suppressMessages(
    fect::att.cumu(fit, period = period, plot = FALSE)
  ))

  ## att.cumu() returns a matrix; final row is the cumulative through
  ## period[2] starting at period[1].
  final_row <- out_acc[nrow(out_acc), ]

  expect_equal(est$estimate, unname(final_row["catt"]),     tolerance = 1e-8)
  expect_equal(est$se,       unname(final_row["S.E."]),     tolerance = 1e-8)
  expect_equal(est$ci.lo,    unname(final_row["CI.lower"]), tolerance = 1e-8)
  expect_equal(est$ci.hi,    unname(final_row["CI.upper"]), tolerance = 1e-8)
})


## -- AC.4  Reversal panels: explicit error -------------------------

test_that("AC.4: cumulative on reversal panels errors with the contract message", {

  skip_on_cran()

  ## simdata has reversal.
  data("simdata", package = "fect")
  set.seed(42)
  fit_rev <- suppressWarnings(suppressMessages(
    fect::fect(Y ~ D, data = simdata, index = c("id", "time"),
               method = "fe", force = "two-way",
               se = TRUE, nboots = 20, parallel = FALSE,
               keep.sims = TRUE)
  ))
  expect_true(isTRUE(fit_rev$hasRevs == 1))
  expect_error(
    fect::estimand(fit_rev, "att.cumu", "event.time"),
    "treatment reversals",
    fixed = FALSE
  )
})


## -- AC.5  keep.sims = FALSE: clear error path ---------------------

test_that("AC.5: cumulative without keep.sims errors with the locked wording", {

  skip_on_cran()

  fit_no_keep <- .fit_no_reversal(keep.sims = FALSE)
  ## Event-time cumulative delegates to effect(), which requires
  ## eff.boot. Without keep.sims the fit has no eff.boot; we surface
  ## the canonical fect error wording.
  expect_error(
    fect::estimand(fit_no_keep, "att.cumu", "event.time"),
    "keep\\.sims = TRUE",
    fixed = FALSE
  )
})


## -- AC.6  n_cells populated from fit$est.att[, "count"] -----------

test_that("AC.6: cumulative event.time n_cells matches fit$est.att counts", {

  skip_on_cran()

  fit <- .fit_no_reversal()
  est <- fect::estimand(fit, "att.cumu", "event.time")

  expect_true(is.integer(est$n_cells))
  expect_false(any(is.na(est$n_cells)))
  expect_true(all(est$n_cells > 0L))

  ## Each event-time row's n_cells equals the corresponding count in
  ## fit$est.att (same per-period treated-cell count as the level ATT
  ## --- esplot's bottom-bar panel reads from this column).
  count_lookup <- as.integer(fit$est.att[, "count"])
  names(count_lookup) <- rownames(fit$est.att)
  expected <- unname(count_lookup[as.character(est$event.time)])
  expect_equal(est$n_cells, expected)
})


test_that("AC.6b: cumulative overall n_cells equals window-summed count", {

  skip_on_cran()

  fit <- .fit_no_reversal()
  period <- c(1, 5)

  est <- fect::estimand(fit, "att.cumu", "overall", window = period)

  expect_true(is.integer(est$n_cells))
  expect_false(any(is.na(est$n_cells)))

  et <- as.numeric(rownames(fit$est.att))
  in_window <- !is.na(et) & et >= period[1] & et <= period[2]
  expected <- as.integer(sum(fit$est.att[in_window, "count"], na.rm = TRUE))
  expect_equal(est$n_cells, expected)
})

#####################################
## goodness of fit test: wild
#####################################

fect_test <- function(
    out, # from fect
    Y,
    X,
    D, ## input
    I,
    II,
    T.on,
    T.off = NULL,
    method = "ife",
    degree = 2,
    knots = NULL, 
    cl = NULL,
    r = 0,
    lambda = Inf,
    force,
    hasRevs = 0,
    tol,
    norm.para,
    pre.period = NULL,
    off.period = NULL,
    nboots,
    parallel = TRUE,
    cores = NULL) {
    

    x <- y <- na.pos <- NULL
    TT <- dim(Y)[1]
    N <- dim(Y)[2]
    if (is.null(X) == FALSE) {
        p <- dim(X)[3]
    } else {
        p <- 0
    }
    
    ## output
    eff <- out$eff
    eff[which(is.na(eff))] <- 0

    ## fitted values Y0
    Y.f <- out$est$fit

    if (is.null(pre.period)) {
        pre.period <- c(min(c(T.on), na.rm = TRUE), 0)
    }

    if (is.null(off.period) & hasRevs == TRUE) {
        off.period <- c(0, max(c(T.off), na.rm = TRUE))
    }

    pre.pos <- which(T.on <= pre.period[2] & T.on >= pre.period[1] & I == 1)

    ## message("\n OK3 \n")

    ## demean pre-treatment tr eff under H0 for wild bootstrap
    #eff[pre.pos] <- eff[pre.pos] - mean(eff[pre.pos])

    ## observed F
    eff.pre <- as.data.frame(out$eff.pre)
    eff.pre <- eff.pre[which(eff.pre[,"period"] >= pre.period[1] & eff.pre[,"period"] <= pre.period[2]),]

    lm.fit <- lm(eff ~ factor(period), data = eff.pre)
    #res <- lm.fit$residuals
    #f <- ((sum(eff.pre[,"eff"]^2) - sum(res^2))/length(unique(eff.pre[,"period"])))/(sum(res^2)/(dim(eff.pre)[1] - length(unique(eff.pre[,"period"]))))
    f <- summary(lm.fit)$f[1]
    names(f) <- NULL

    ## message("\n OK4 \n")

    f2 <- NULL
    if (hasRevs) {
        ## switch-off 
        eff.off <- as.data.frame(out$eff.off)
        eff.off <- eff.off[which(eff.off[,"period"] >= off.period[1] & eff.off[,"period"] <= off.period[2]),]
        lm.fit2 <- lm(eff ~ factor(period), data = eff.off)

        f2 <- summary(lm.fit2)$f[1]
        names(f2) <- NULL
    }

    ## bootstrapped F under H0
    f.boot <- rep(NA, nboots)
    f2.boot <- NULL
    if (hasRevs) {
        f2.boot <- rep(NA, nboots)
    }

    ## message("\n OK1 \n")

    #if (method == "ife") {
        one.nonpara <- function() {
                
            if (is.null(cl)) {
                res.p <- matrix(sample(c(-1, 1), N*TT, replace = TRUE), TT, N)
            } else {
                res.p <- matrix(rep(sample(c(-1, 1), N, replace = TRUE), each = TT), TT, N)
            }
            Y.boot <- Y.f + res.p * eff
            Y.boot[which(I == 0)] <- 0
            if (method %in% c("ife", "fe")) {
                boot <- try(fect_fe(Y = Y.boot, X = X, D = D, I = I, II = II, 
                                T.on = T.on, T.off = NULL,
                                r.cv = r, force = force, hasRevs = 0, 
                                tol = tol, boot = 1,
                                norm.para = norm.para, 
                                placebo.period = NULL, placeboTest = 0), silent = TRUE)
            }
            else if (method == "mc") {
                boot <- try(fect_mc(Y = Y.boot, X = X, D = D, 
                                I = I, II = II, hasF = out$validF,
                                T.on = T.on, T.off = NULL, 
                                lambda.cv = lambda, force = force, hasRevs = 0, 
                                tol = tol, boot = 1,
                                norm.para = norm.para,
                                placebo.period = NULL, placeboTest = 0), silent = TRUE)
            }
            
            if ('try-error' %in% class(boot)) {
                ## message("NA")
                return(list(f.boot = NA, f.boot.2 = NA))
            } else {
                ## message(boot$niter)
                ## message("\n")
                data <- as.data.frame(boot$eff.pre)
                #colnames(data) <- c("period", "eff")
                data <- data[which(data[,"period"] <= pre.period[2] & data[,"period"] >= pre.period[1]),]
                lm.fit.boot <- lm(eff~factor(period), data = data)
                #f.boot <- ((sum(data[,"eff"]^2) - sum((lm.fit.boot$residuals)^2))/length(unique(data[,"period"])))/(sum((lm.fit.boot$residuals)^2)/(dim(data[1]-unique(data[,"period"]))))
                f.boot <- summary(lm.fit.boot)$f[1]
                ## if (r.cv == 0) {

                f.boot.2 <- NULL
                if (hasRevs) {
                    ## switch-off 
                    data2 <- as.data.frame(out$eff.off)
                    data2 <- data2[which(data2[,"period"] >= off.period[1] & data2[,"period"] <= off.period[2]),]
                    lm.fit.boot2 <- lm(eff ~ factor(period), data = data2)

                    f.boot.2 <- summary(lm.fit.boot2)$f[1]
                }
                if (hasRevs) {
                    return(list(f.boot = f.boot, f.boot.2 = f.boot.2))
                } else {
                    return(list(f.boot = f.boot))
                }
                
                ## } else {
                ##     return(list(f.boot = f.boot, factor.boot = boot$factor))
                ## } 
            }    
            
        } 
            
    #} else { ## mc
    #    one.nonpara <- function() {
            
    #        if (is.null(cl)) {
    #            res.p <- matrix(sample(c(-1, 1), N*TT, replace = TRUE), TT, N)
    #        } else {
    #            res.p <- matrix(rep(sample(c(-1, 1), N, replace = TRUE), each = TT), TT, N)
    #        }
    #        Y.boot <- Y.f + res.p * eff
    #        Y.boot[which(I == 0)] <- 0
    #        boot <- try(fect.mc(Y = Y.boot, X = X, D = D, 
    #                            I = I, II = II, hasF = out$validF,
    #                            T.on = T.on, T.off = NULL, 
    #                            lambda.cv = lambda, force = force, hasRevs = 0, 
    #                            tol = tol, boot = 1,
    #                            norm.para = norm.para,
    #                            placebo.period = NULL, placeboTest = 0), silent = TRUE)

    #        if ('try-error' %in% class(boot)) {
                ## message("NA")
    #            return(list(f.boot = NA, f.boot.2 = NA))
    #        } else {
                ## message(boot$niter)
                ## message("\n")
    #            data <- as.data.frame(boot$eff.pre)
                #colnames(data) <- c("period", "eff")
    #            data <- data[which(data[,"period"] <= pre.period[2] & data[,"period"] >= pre.period[1]),]
    #            lm.fit.boot <- lm(eff~factor(period), data = data)
                #f.boot <- ((sum(data[,"eff"]^2) - sum((lm.fit.boot$residuals)^2))/length(unique(data[,"period"])))/(sum((lm.fit.boot$residuals)^2)/(dim(data[1]-unique(data[,"period"]))))
    #            f.boot <- summary(lm.fit.boot)$f[1]
                ## if (r.cv == 0) {
                
    #            f.boot.2 <- NULL
    #            if (hasRevs) {
                    ## switch-off 
    #                data2 <- as.data.frame(out$eff.off)
    #                data2 <- data2[which(data2[,"period"] >= off.period[1] & data2[,"period"] <= off.period[2]),]
    #                lm.fit.boot2 <- lm(eff ~ factor(period), data = data2)

    #                f.boot.2 <- summary(lm.fit.boot2)$f[1]
    #            }
    #            if (hasRevs) {
    #                return(list(f.boot = f.boot, f.boot.2 = f.boot.2))
    #            } else {
    #                return(list(f.boot = f.boot))
    #            }

                ## } else {
                ##     return(list(f.boot = f.boot, factor.boot = boot$factor))
                ## } 
    #        }
    #    } 
    #}
    
    ## computing
    if (!identical(parallel, FALSE)) {
        boot.out <- foreach(j=1:nboots, 
                            .inorder = FALSE,
                            .export = c("fect_fe", "fect_mc", "get_term"),
                            .packages = c("fect")
                            ) %dopar% {
                                return(one.nonpara())
                            }

        for (j in 1:nboots) { 
            f.boot[j] <- boot.out[[j]]$f.boot
        } 
    } else {
        for (j in 1:nboots) { 
            one.boot <- one.nonpara()
            f.boot[j] <- one.boot$f.boot
            ## report progress
            if (j%%100 == 0)  {
                message(".")   
            }  
        }  
    } 

    ## message("\n OK2 \n")

    ## end of bootstrapping
    if (sum(is.na(f.boot)) > 0) {
        f.boot <- f.boot[-which(is.na(f.boot))]
    }
    message(length(f.boot), " runs\n", sep = "")
    
    f.q <- quantile(f.boot, probs = c(0.025, 0.975))
    f.p <- sum(f.boot > f)/length(f.boot)
    wald.on <- list(stat = f, 
                    p = f.p,
                    quantile_lower = f.q[1], 
                    quantile_upper = f.q[2],
                    boot = f.boot)


    wald.off <- NULL
    if (hasRevs) {
        if (sum(is.na(f2.boot)) > 0) {
            f2.boot <- f2.boot[-which(is.na(f2.boot))]
        }
        f.q2 <- quantile(f2.boot, probs = c(0.025, 0.975))
        f.p2 <- sum(f2.boot > f2)/length(f2.boot)
        wald.off <- list(stat = f2, 
                         p = f.p2,
                         quantile_lower = f.q2[1], 
                         quantile_upper = f.q2[2],
                         boot = f2.boot)
    }

    if (hasRevs) {
        return(list(wald.on = wald.on, wald.off = wald.off))
    } else {
        return(list(wald.on = wald.on))
    }
    
} ## end of test

#####################################
## goodness of fit test: non para 
#####################################

#fect.test2 <- function(Y,
#                       X,
#                       D, ## input
#                       I,
#                       II,
#                       T.on,
#                       T.off = NULL,
#                       method = "ife",
#                       r = 0,
#                       lambda = Inf,
#                       force,
#                       hasRevs = 0,
#                       tol,
#                       norm.para,
#                       pre.period = NULL,
#                       nboots,
#                       parallel = TRUE,
#                       cores = NULL)  {
    
    
#    x <- y <- na.pos <- NULL
#    TT <- dim(Y)[1]
#    N <- dim(Y)[2]
#    if (is.null(X) == FALSE) {
#        p <- dim(X)[3]
#    } else {
#        p <- 0
#    }

#    if (hasRevs == 1) {
        ## D.fake : check reversals
#        D.fake <- apply(D, 2, function(vec){cumsum(vec)})
#        D.fake <- ifelse(D.fake > 0, 1, 0)
#        D.fake[which(I == 0)] <- 0

#        rev <- which(apply(D.fake == D, 2, sum) != TT)
#        co <- which(apply(D, 2, sum) == 0)
#        tr.all <- which(apply(D, 2, sum) > 0)
#        tr <- tr.all[which(!tr.all %in% rev)]

#        Nrev <- length(rev)
#        Ntr <- length(tr)
#        Nco <- length(co)
#    } else {
        ## treatement indicator
#        tr <- which(apply(D, 2, sum) > 0)
#        co <- which(apply(D, 2, sum) == 0)

#        Ntr <- length(tr)
#        Nco <- length(co)
#    }
    
    ## estimation
#    if (method == "ife") {
#        out <- fect.fe(Y = Y, X = X, D = D, I = I, II = II, 
#                       T.on = T.on, T.off = NULL,
#                       r.cv = r, force = force, hasRevs = 0, 
#                       tol = tol, boot = 1,
#                       norm.para = norm.para, 
#                       placebo.period = NULL, placeboTest = 0)
#    } else {
#        out <- fect.mc(Y = Y, X = X, D = D, I = I, II = II,
#                       T.on = T.on, T.off = NULL, 
#                       lambda.cv = lambda, force = force, hasRevs = 0, 
#                       tol = tol, boot = 1,
#                       norm.para = norm.para,
#                       placebo.period = NULL, placeboTest = 0)
#    }

    ## output
#    eff <- out$eff
#    eff[which(is.na(eff))] <- 0

    ## fitted values Y0
#    Y.f <- out$est$fit

#    if (is.null(pre.period)) {
#        pre.period <- c(min(c(T.on), na.rm = TRUE), 0)
#    }

#    pre.pos <- which(T.on <= pre.period[2] & T.on >= pre.period[1] & I == 1)

    ## demean pre-treatment tr eff under H0 for wild bootstrap
    #eff[pre.pos] <- eff[pre.pos] - mean(eff[pre.pos])

    ## observed F
#    eff.pre <- as.data.frame(out$eff.pre)
#    eff.pre <- eff.pre[which(eff.pre[,"period"] >= pre.period[1] & eff.pre[,"period"] <= pre.period[2]),]

#    eff.pre[,"period"] <- as.factor(eff.pre[,"period"])
#    lm.fit <- lm(eff ~ period, data = eff.pre)
    #res <- lm.fit$residuals
    #f <- ((sum(eff.pre[,"eff"]^2) - sum(res^2))/length(unique(eff.pre[,"period"])))/(sum(res^2)/(dim(eff.pre)[1] - length(unique(eff.pre[,"period"]))))
#    f <- summary(lm.fit)$f[1]


    ## bootstrapped F under H0
#    f.boot <- rep(NA, nboots)

#    message("\rBootstrapping ...\n")

#    if (method == "ife") {
#        one.nonpara <- function() {

#            if (hasRevs == 0) {
#                if (Nco > 0) {
#                    repeat{
#                        fake.co <- sample(co, Nco, replace=TRUE)
#                        fake.tr <- sample(tr, Ntr, replace=TRUE)
#                        boot.id <- c(fake.tr, fake.co)
#                        if (sum(apply(as.matrix(I[,boot.id]),1,sum)>=1)==TT) {
#                            break
#                        }
#                    }
#                } else {
#                    repeat{
#                        boot.id <- sample(tr, Ntr, replace=TRUE)
#                        if (sum(apply(as.matrix(I[,boot.id]),1,sum)>=1)==TT) {
#                            break
#                        }
#                    }
#                }
#            } else {
#                if (Ntr > 0) {
#                    if (Nco > 0) {
#                        repeat{
#                            fake.co <- sample(co, Nco, replace=TRUE)
#                            fake.tr <- sample(tr, Ntr, replace=TRUE)
#                            fake.rev <- sample(rev, Nrev, replace=TRUE)
#                            boot.id <- c(fake.rev, fake.tr, fake.co)
#                            if (sum(apply(as.matrix(I[,boot.id]),1,sum)>=1)==TT) {
#                                break
#                            }
#                        }
#                    } else {
#                        repeat{
#                            fake.tr <- sample(tr, Ntr, replace=TRUE)
#                            fake.rev <- sample(rev, Nrev, replace=TRUE)
#                            boot.id <- c(fake.rev, fake.tr)
#                            if (sum(apply(as.matrix(I[,boot.id]),1,sum)>=1)==TT) {
#                                break
#                            }
#                        }
#                    }
#                } else {
#                    if (Nco > 0) {
#                        repeat{
#                            fake.co <- sample(co, Nco, replace=TRUE)
#                            fake.rev <- sample(rev, Nrev, replace=TRUE)
#                            boot.id <- c(fake.rev, fake.co)
#                            if (sum(apply(as.matrix(I[,boot.id]),1,sum)>=1)==TT) {
#                                break
#                            }
#                        }
#                    } else {
#                        repeat{
#                            boot.id <- sample(rev, Nrev, replace=TRUE)
#                            if (sum(apply(as.matrix(I[,boot.id]),1,sum)>=1)==TT) {
#                                break
#                            }
#                        }
#                    }
#                }
#            }
                
#            X.boot <- X[,boot.id,,drop=FALSE]

                
#            boot <- try(fect.fe(Y = Y[, boot.id], X = X.boot, D = D[, boot.id],
#                                I = I[, boot.id], II = II[, boot.id], 
#                                T.on = T.on[, boot.id], T.off = NULL, 
#                                r.cv = r, force = force,
#                                hasRevs = 0, tol = tol, boot = 1,
#                                norm.para = norm.para,
#                                time.on.seq = NULL, time.off.seq = NULL,
#                                placebo.period = NULL, 
#                                placeboTest = 0), silent = TRUE)
#            if ('try-error' %in% class(boot)) {
#                return(NA)
#            } else {
#                data <- as.data.frame(boot$eff.pre)
                #colnames(data) <- c("period", "eff")
#                data <- data[which(data[,"period"] <= pre.period[2] & data[,"period"] >= pre.period[1]),]
#                data[,"period"] <- as.factor(data[,"period"])

#                eff.fit <- predict(lm.fit, data)
                ## remove pre-treatment effect
#                data[,"eff0"] <- data[,"eff"] - eff.fit

#                lm.fit.boot <- lm(eff0~period, data = data)
                #f.boot <- ((sum(data[,"eff"]^2) - sum((lm.fit.boot$residuals)^2))/length(unique(data[,"period"])))/(sum((lm.fit.boot$residuals)^2)/(dim(data[1]-unique(data[,"period"]))))
#                f.boot <- summary(lm.fit.boot)$f[1]
                ## if (r.cv == 0) {
#                return(list(f.boot = f.boot))
                ## } else {
                ##     return(list(f.boot = f.boot, factor.boot = boot$factor))
                ## } 
#            }
#        } 
            
#    } else { ## mc
#        one.nonpara <- function() {

#            if (hasRevs == 0) {
#                if (Nco > 0) {
#                    repeat{
#                        fake.co <- sample(co, Nco, replace=TRUE)
#                        fake.tr <- sample(tr, Ntr, replace=TRUE)
#                        boot.id <- c(fake.tr, fake.co)
#                        if (sum(apply(as.matrix(I[,boot.id]),1,sum)>=1)==TT) {
#                            break
#                        }
#                    }
#                } else {
#                    repeat{
#                        boot.id <- sample(tr, Ntr, replace=TRUE)
#                        if (sum(apply(as.matrix(I[,boot.id]),1,sum)>=1)==TT) {
#                            break
#                        }
#                    }
#                }
#            } else {
#                if (Ntr > 0) {
#                    if (Nco > 0) {
#                        repeat{
#                            fake.co <- sample(co, Nco, replace=TRUE)
#                            fake.tr <- sample(tr, Ntr, replace=TRUE)
#                            fake.rev <- sample(rev, Nrev, replace=TRUE)
#                            boot.id <- c(fake.rev, fake.tr, fake.co)
#                            if (sum(apply(as.matrix(I[,boot.id]),1,sum)>=1)==TT) {
#                                break
#                            }
#                        }
#                    } else {
#                        repeat{
#                            fake.tr <- sample(tr, Ntr, replace=TRUE)
#                            fake.rev <- sample(rev, Nrev, replace=TRUE)
#                            boot.id <- c(fake.rev, fake.tr)
#                            if (sum(apply(as.matrix(I[,boot.id]),1,sum)>=1)==TT) {
#                                break
#                            }
#                        }
#                    }
#                } else {
#                    if (Nco > 0) {
#                        repeat{
#                            fake.co <- sample(co, Nco, replace=TRUE)
#                            fake.rev <- sample(rev, Nrev, replace=TRUE)
#                            boot.id <- c(fake.rev, fake.co)
#                            if (sum(apply(as.matrix(I[,boot.id]),1,sum)>=1)==TT) {
#                                break
#                            }
#                        }
#                    } else {
#                        repeat{
#                            boot.id <- sample(rev, Nrev, replace=TRUE)
#                            if (sum(apply(as.matrix(I[,boot.id]),1,sum)>=1)==TT) {
#                                break
#                            }
#                        }
#                    }
#                }
#            }
                
#            X.boot <- X[,boot.id,,drop=FALSE]
            
#            boot <- try(fect.mc(Y = Y[,boot.id], X = X.boot, D = D[,boot.id],
#                                I = I[,boot.id], II = II[,boot.id],
#                                T.on = T.on[,boot.id], T.off = NULL, 
#                                lambda.cv = lambda, force = force, 
#                                hasF = out$validF, hasRevs = 0, 
#                                tol = tol, boot = 1,
#                                norm.para = norm.para,
#                                time.on.seq = NULL, time.off.seq = NULL,
#                                placebo.period = NULL, 
#                                placeboTest = 0), silent = TRUE)
            
#            if ('try-error' %in% class(boot)) {
#                return(NA)
#            } else {
#                data <- as.data.frame(boot$eff.pre)
                #colnames(data) <- c("period", "eff")
#                data <- data[which(data[,"period"] <= pre.period[2] & data[,"period"] >= pre.period[1]),]
#                data[,"period"] <- as.factor(data[,"period"])

#                eff.fit <- predict(lm.fit, data)
#                data[,"eff0"] <- data[,"eff"] - eff.fit

#                lm.fit.boot <- lm(eff0~period, data = data)
                #f.boot <- ((sum(data[,"eff"]^2) - sum((lm.fit.boot$residuals)^2))/length(unique(data[,"period"])))/(sum((lm.fit.boot$residuals)^2)/(dim(data[1]-unique(data[,"period"]))))
#                f.boot <- summary(lm.fit.boot)$f[1]
                ## if (r.cv == 0) {
#                return(list(f.boot = f.boot))
                ## } else {
                ##     return(list(f.boot = f.boot, factor.boot = boot$factor))
                ## }
#            }
            
#        } 
#    }

    ## computing
#    if (parallel == TRUE) { 
#        boot.out <- foreach(j=1:nboots, 
#                            .inorder = FALSE,
#                            .export = c("fect.fe", "fect.mc", "get_term"),
#                            .packages = c("fect")
#                            ) %dopar% {
#                                return(one.nonpara())
#                            }

#        for (j in 1:nboots) { 
#            f.boot[j] <- boot.out[[j]]$f.boot
#        } 
#    } else {
#        for (j in 1:nboots) { 
#            one.boot <- one.nonpara()
#            f.boot[j] <- one.boot$f.boot
#            ## report progress
#            if (j%%100 == 0)  {
#                message(".")   
#            }  
#        }  
#    } 
    ## end of bootstrapping
#    message("\r")

    
#    f.q <- quantile(f.boot, probs = c(0.025, 0.975))
#    f.p <- sum(f.boot > f)/nboots

#    title <- "Empirical distribution of F under H0"
    ## plot
#    f.data <- cbind.data.frame(f.boot = f.boot)
#    p <- ggplot(f.data, aes(f.boot)) + geom_density() + geom_vline(xintercept = f, colour="red",size = 0.5)

#    d <- ggplot_build(p)$data[[1]]

#    if (f < max(d[,"x"])) {
#        p <- p + geom_area(data = subset(d, x > f), aes(x=x, y=y), fill="red", alpha = 0.2)
#    }

#    p <- p + annotate("text", x = 0.9*max(d[,"x"]), y = 0.9*max(d[,"y"]), label = paste("P value: ", f.p, sep=""))

#    p <- p + ggtitle(title) +  theme(plot.title = element_text(size=20,
#                                                               hjust = 0.5,
#                                                               margin = margin(10, 0, 10, 0)))

    ## suppressWarnings(print(p))
  
    ## store results
#    Ftest <- matrix(NA, 1, 4)
#    Ftest[1, ] <- c(f, f.q, f.p)

#    colnames(Ftest) <- c("stat", "quantile_lower", "quantile_upper", "p_value")
    
#    return(list(stat = Ftest, boot = f.boot, graph = p))
    
#} ## end of test

utils::globalVariables(c("x", "y", "legend_id"))

esplot <- function(data,  # time, ATT, CI.lower, CI.upper, count, ...
                   Period = NULL,
                   Estimate = "ATT",
                   SE = NULL,
                   CI.lower = "CI.lower", # Default name to look for
                   CI.upper = "CI.upper", # Default name to look for
                   Count = NULL,

                   proportion = 0.3,
                   est.lwidth = NULL,  # line thickness
                   est.pointsize = NULL,    # point size if show.points=TRUE
                   show.points = TRUE,   # show points at integer times
                   fill.gap = TRUE,      # fill missing times with 0
                   start0 = FALSE,       # if TRUE => vertical dashed line at x=-0.5
                   only.pre = FALSE,
                   only.post = FALSE,
                   show.count = TRUE,    # whether to show count bars
                   stats = NULL,         # numeric p-values
                   stats.labs = NULL,    # labels for each p-value

                   highlight.periods = NULL,  # numeric vector of times to highlight
                   highlight.colors  = NULL,    # color for each highlight time
                   highlight.shapes  = NULL,    # integer shape for each highlight time (17=triangle, 18=diamond, etc.)
                   highlight.fill    = FALSE,   # if TRUE, draw a lightened-tone background rectangle behind each highlighted period

                   lcolor = NULL,
                   lwidth = NULL,
                   ltype = NULL,
                   connected = FALSE,    # if TRUE => line + ribbon
                   ci.outline = FALSE,
                   main      = NULL,
                   xlim      = NULL,
                   ylim      = NULL,
                   xlab      = NULL,
                   ylab      = NULL,
                   gridOff   = FALSE,
                   stats.pos = NULL,
                   theme.bw  = TRUE,
                   legacy.style = FALSE,
                   cex.main  = NULL,
                   cex.axis  = NULL,
                   cex.lab   = NULL,
                   cex.text  = NULL,
                   axis.adjust = FALSE,
                   color       = "#000000",
                   pre.color   = NULL,
                   post.color  = NULL,
                   count.color = "gray70",
                   count.alpha = 0.4,
                   count.outline.color = "grey69",
                   xangle    = NULL,
                   yangle    = NULL,
                   xbreaks   = NULL,
                   ybreaks   = NULL,
                   legendOff = FALSE,
                   cex.legend = NULL,
                   pre.label  = "Pre-treatment",
                   post.label = "Post-treatment"
)
{
  ## Three visual modes:
  ##  - modern    (theme.bw = TRUE,  legacy.style = FALSE): 2.3.1 default
  ##  - classic   (theme.bw = TRUE,  legacy.style = TRUE):  pre-2.3.1 white-panel default
  ##  - gray      (theme.bw = FALSE):                       legacy gray-panel look
  use_modern_recipe <- isTRUE(theme.bw) && !isTRUE(legacy.style)

  ## Default ltype: modern uses dashed vline at treatment onset; legacy uses solid.
  if (is.null(ltype)) {
    ltype <- if (use_modern_recipe) c("solid", "dashed") else c("solid", "solid")
  }

  if (is.null(est.lwidth) || is.null(est.pointsize)) {
    default_est_lwidth <- .8
    default_est_pointsize <- 3

    if (!connected) {
      default_est_lwidth <- 0.6
      default_est_pointsize <- 2
    } else {
      if (show.points) {
        default_est_lwidth <- 0.7
        default_est_pointsize <- 1.2
      } else {
        default_est_lwidth <- 1.2
        default_est_pointsize <- 3
      }
    }
    if(is.null(est.lwidth)) est.lwidth <- default_est_lwidth
    if(is.null(est.pointsize)) est.pointsize <- default_est_pointsize
  }

  all_integer_like <- function(x) {
    x_chr <- if (is.factor(x)) as.character(x) else x
    all(grepl("^[-+]?[0-9]+$", x_chr))
  }
  # If input is a fect object, extract the event-study data
  if (inherits(data, "fect")) {
    if (is.null(data$est.att)) {
      stop("The fect object does not contain 'est.att'. Run fect() with se = TRUE.")
    }
    fect_att <- as.data.frame(data$est.att)
    fect_att$Period <- as.numeric(rownames(fect_att))
    if (is.null(Period)) Period <- "Period"
    if (is.null(Count) && "count" %in% names(fect_att)) Count <- "count"
    if (is.null(Count) && "count.on" %in% names(fect_att)) {
      Count <- "count.on"
    }
    data <- fect_att
  } else if (inherits(data, "did_wrapper")) {
    data <- data$est.att
    if (is.null(Count) && "count" %in% names(data)) Count <- "count"
    if (is.null(Count) && "count.on" %in% names(data)) Count <- "count.on"
  }
  data <- as.data.frame(data)

  ## Resolve pre/post colors. Both modern and legacy keep a lightness contrast
  ## (pre = "grey50", post = color) so pre-treatment and post-treatment points
  ## are visually distinguishable. fect convention: with start0 = FALSE
  ## (default), period 0 is the LAST pre-treatment period; first post is t=1.
  ##
  ## Exception: estimand() with test = "placebo" / "carryover" attaches a
  ## "fect_test" attribute to the returned data frame; in those modes ALL
  ## event-times are special (no pre-vs-post contrast to convey), so we
  ## use a single uniform color (= `color`) across all points and skip the
  ## gray pre-treatment shade.
  fect_test <- attr(data, "fect_test")
  if (is.null(post.color)) post.color <- color
  if (is.null(pre.color)) {
    pre.color <- if (!is.null(fect_test) &&
                     fect_test %in% c("placebo", "carryover")) {
      color
    } else {
      "gray50"
    }
  }

  # Identify time/period column
  if (is.null(Period)) {
    common_period_names <- c("time", "Time", "period", "Period", "event.time", "event_time", "rel_time")
    found_period <- common_period_names[common_period_names %in% names(data)]
    if (length(found_period) > 0) {
      Period <- found_period[1]
    } else if (all_integer_like(rownames(data))) {
      data$Period_generated <- as.numeric(rownames(data))
      Period <- "Period_generated"
    } else {
      stop("Period column not specified and could not be inferred. Please specify the 'Period' argument.")
    }
  }
  if (!(Period %in% names(data))) stop(paste0("Period column '", Period, "' not found in data."))


  # Ensure Estimate column exists and is numeric
  if (!(Estimate %in% names(data))) stop(paste("Estimate column '", Estimate, "' not found in data."))
  data[[Estimate]] <- as.numeric(data[[Estimate]])

  # Ensure SE column is numeric if it exists and is specified
  if (!is.null(SE) && SE %in% names(data)) {
    data[[SE]] <- as.numeric(data[[SE]])
  }

  # --- START: Centralized CI calculation from SE ---
  # Check for CI.lower. If not present, try to calculate from SE.
  # CI.lower is the parameter name (e.g., "CI.lower" by default).
  if (!(CI.lower %in% names(data))) {
    if (!is.null(SE) && SE %in% names(data)) {
      calculated_ci_lower <- data[[Estimate]] - 1.96 * data[[SE]]
      if (any(!is.na(calculated_ci_lower))) {
        message(paste0("Column '", CI.lower, "' not found in input data. Calculating values for '", CI.lower, "' using '", Estimate, "' and '", SE, "'."))
      }
      data[[CI.lower]] <- calculated_ci_lower
    } else {
      data[[CI.lower]] <- NA_real_ # Create the column with NAs
    }
  } else { # CI.lower column exists in input data
    data[[CI.lower]] <- as.numeric(data[[CI.lower]]) # Ensure it's numeric
  }

  # Check for CI.upper. If not present, try to calculate from SE.
  if (!(CI.upper %in% names(data))) {
    if (!is.null(SE) && SE %in% names(data)) {
      calculated_ci_upper <- data[[Estimate]] + 1.96 * data[[SE]]
      if (any(!is.na(calculated_ci_upper))) {
        message(paste0("Column '", CI.upper, "' not found in input data. Calculating values for '", CI.upper, "' using '", Estimate, "' and '", SE, "'."))
      }
      data[[CI.upper]] <- calculated_ci_upper
    } else {
      data[[CI.upper]] <- NA_real_
    }
  } else { # CI.upper column exists in input data
    data[[CI.upper]] <- as.numeric(data[[CI.upper]]) # Ensure it's numeric
  }
  # --- END: Centralized CI calculation from SE ---
  # Now, 'data' is guaranteed to have columns named by CI.lower and CI.upper parameters,
  # containing original CIs, CIs from SE, or NAs. These columns are also numeric.

  # Possibly build an interpolated dataset if connected=TRUE
  if (connected) {
    min_period_val <- min(data[[Period]], na.rm = TRUE)
    max_period_val <- max(data[[Period]], na.rm = TRUE)

    if (!is.finite(min_period_val) || !is.finite(max_period_val)) {
      stop("Cannot interpolate with non-finite Period values.")
    }

    new_periods <- seq(
      from = min_period_val,
      to   = max_period_val,
      by   = 0.5
    )

    # SE for interpolation: use data[[SE]] if SE column exists, otherwise NA
    # data[[SE]] is already numeric or NA due to upfront logic.
    se_for_interp <- if (!is.null(SE) && SE %in% names(data)) {
      data[[SE]]
    } else {
      rep(NA_real_, nrow(data))
    }

    # CI columns are guaranteed to exist in 'data' and be numeric (or NA_real_).
    ci_lower_for_interp <- data[[CI.lower]]
    ci_upper_for_interp <- data[[CI.upper]]

    count_for_interp <- if (!is.null(Count) && Count %in% names(data)) {
      data[[Count]]
    } else {
      rep(NA_real_, nrow(data))
    }

    temp_data_for_interp <- data.frame(
      Period_val = data[[Period]],
      Estimate_val = data[[Estimate]],      # Already numeric
      SE_val = se_for_interp,              # Numeric or NA
      CI.lower_val = ci_lower_for_interp,  # Numeric or NA
      CI.upper_val = ci_upper_for_interp,  # Numeric or NA
      Count_val = count_for_interp
    )

    interp_col_safe <- function(y_vals) {
      if (all(is.na(y_vals))) return(rep(NA_real_, length(new_periods)))
      idx_valid_approx <- !is.na(temp_data_for_interp$Period_val) & !is.na(y_vals)
      if(sum(idx_valid_approx) < 2) return(rep(NA_real_, length(new_periods)))
      stats::approx(x = temp_data_for_interp$Period_val[idx_valid_approx],
                    y = y_vals[idx_valid_approx],
                    xout = new_periods, rule = 2)$y
    }

    df_interp <- data.frame(Period_placeholder = new_periods)
    df_interp[[Estimate]] <- interp_col_safe(temp_data_for_interp$Estimate_val)

    # Interpolate SE if SE parameter was provided
    if (!is.null(SE)) {
      # temp_data_for_interp$SE_val contains original SEs or NAs if SE column didn't exist/was all NA.
      # interp_col_safe will handle all-NA y_vals correctly.
      df_interp[[SE]] <- interp_col_safe(temp_data_for_interp$SE_val)
    }

    # Interpolate CIs. CI.lower and CI.upper are the names of the columns.
    df_interp[[CI.lower]] <- interp_col_safe(temp_data_for_interp$CI.lower_val)
    df_interp[[CI.upper]] <- interp_col_safe(temp_data_for_interp$CI.upper_val)


    if (!is.null(Count)) {
      df_interp[[Count]] <- NA_real_
      orig_t <- temp_data_for_interp$Period_val
      orig_c <- temp_data_for_interp$Count_val
      for (i in seq_len(nrow(df_interp))) {
        pval <- df_interp$Period_placeholder[i]
        if (abs(pval - round(pval)) < 1e-8) {
          idx_match <- which(abs(orig_t - pval) < 1e-8 & !is.na(orig_c))
          if (length(idx_match) >= 1) {
            df_interp[[Count]][i] <- orig_c[idx_match[1]]
          }
        }
      }
    }
    names(df_interp)[names(df_interp) == "Period_placeholder"] <- Period
    data <- df_interp
  } # End of if(connected)

  # Ensure relevant columns are numeric after potential interpolation.
  # Upfront logic and interpolation should handle this, but this is a safeguard.
  if (!is.null(SE) && SE %in% names(data)) data[[SE]] <- as.numeric(data[[SE]])
  # CI.lower and CI.upper columns are guaranteed to exist by upfront logic.
  data[[CI.lower]] <- as.numeric(data[[CI.lower]])
  data[[CI.upper]] <- as.numeric(data[[CI.upper]])


  # Check "show.count": default TRUE only if count data is available
  if (isTRUE(show.count) && (is.null(Count) || !(Count %in% names(data)))) {
    show.count <- FALSE
  }
  if (is.null(show.count)) {
    show.count <- (!is.null(Count) && Count %in% names(data))
  }
  if (!is.logical(show.count) && !is.numeric(show.count)) {
    stop("\"show.count\" must be TRUE/FALSE or NULL.")
  }
  if (show.count && is.null(Count)) {
    stop("\"Count\" column name is not specified via 'Count' parameter, but show.count is TRUE.")
  }
  if (show.count && !(Count %in% names(data))) {
    stop(paste0("Specified \"Count\" column '", Count, "' is not in the data, but show.count is TRUE."))
  }

  # Check other logical flags
  if (!is.logical(gridOff) && !gridOff %in% c(0, 1)) {
    stop("\"gridOff\" must be TRUE/FALSE.")
  }
  if (!is.logical(fill.gap) && !fill.gap %in% c(0, 1)) {
    stop("\"fill.gap\" must be TRUE/FALSE.")
  }
  if (!is.logical(axis.adjust) && !axis.adjust %in% c(0, 1)) {
    stop("\"axis.adjust\" must be TRUE/FALSE.")
  }

  # Title & label sizing (values are in pt; defaults match plot.fect)
  if (!is.null(main)) {
    if (!is.character(main)) stop("\"main\" is not a string.")
    main <- main[1]
  }
  if (!is.null(cex.main)) {
    if (!is.numeric(cex.main)) stop("\"cex.main\" must be numeric.")
  } else {
    cex.main <- if (use_modern_recipe) 11 else 16
  }

  if (!is.null(xlab) && !is.character(xlab)) stop("\"xlab\" is not a string.")
  if (!is.null(ylab) && !is.character(ylab)) stop("\"ylab\" is not a string.")

  if (!is.null(cex.lab)) {
    if (!is.numeric(cex.lab)) stop("\"cex.lab\" must be numeric.")
  } else {
    cex.lab <- if (use_modern_recipe) 10 else 15
  }

  if (!is.null(cex.axis)) {
    if (!is.numeric(cex.axis)) stop("\"cex.axis\" must be numeric.")
  } else {
    cex.axis <- if (use_modern_recipe) 9 else 15
  }

  if (!is.null(cex.text)) {
    if (!is.numeric(cex.text)) stop("\"cex.text\" must be numeric.")
  } else {
    cex.text <- if (use_modern_recipe) 3.0 else 5
  }

  # Stats
  if (!is.null(stats)) {
    stats <- c(stats)
    if (!is.numeric(stats)) {
      stop("The \"stats\" option must be numeric. Calling function should extract numeric values.")
    }
    stats_display <- formatC(stats, digits = 3, format = "f", flag = "#") # Keep as is

    n.stats <- length(stats_display)
    if (is.null(stats.labs)) {
      stats.labs <- rep("", n.stats)
    } else if (length(stats.labs) != n.stats) {
      stop("The \"stats.labs\" option should have the same length as \"stats\".")
    }
  }
  # Validation for stats.pos (already present and correct)
  if (!is.null(stats.pos)) {
    if (length(stats.pos) != 2) {
      stop("\"stats.pos\" must be of length 2.")
    }
    if (!is.numeric(stats.pos[1]) || !is.numeric(stats.pos[2])) {
      stop("Elements of \"stats.pos\" must be numeric.")
    }
  }


  # Axis rotation
  if (axis.adjust) {
    angle <- 45
    x.v <- 1 # vjust for x-axis text
    x.h <- 1 # hjust for x-axis text
  } else {
    angle <- 0
    x.v <- 0.5
    x.h <- 0.5
  }

  # xlim, ylim checks (user-provided)
  user_xlim <- xlim
  user_ylim <- ylim
  if (!is.null(user_xlim)) {
    if (!is.numeric(user_xlim) || length(user_xlim) != 2) {
      stop("\"xlim\" must be a numeric vector of length 2.")
    }
    if(user_xlim[1] > user_xlim[2]) user_xlim <- rev(user_xlim) # Ensure min < max
  }
  if (!is.null(user_ylim)) {
    if (!is.numeric(user_ylim) || length(user_ylim) != 2) {
      stop("\"ylim\" must be a numeric vector of length 2.")
    }
    if(user_ylim[1] > user_ylim[2]) user_ylim <- rev(user_ylim) # Ensure min < max
  }

  # Default axis labels
  if (is.null(xlab)) {
    xlab <- "Time Relative to Treatment"
  } else if (xlab == "") {
    xlab <- NULL # ggplot prefers NULL for no label
  }
  if (is.null(ylab)) {
    ylab <- "Effect on Y"
  } else if (ylab == "") {
    ylab <- NULL
  }

  if (is.null(lcolor)) {
    if (use_modern_recipe) {
      lcolor <- "grey75"
    } else if (isTRUE(theme.bw)) {
      lcolor <- "#AAAAAA70"  # pre-2.3.1 default for theme.bw=TRUE (classic mode)
    } else {
      lcolor <- "white" # This might be invisible on a white background if not theme_bw
    }
  }

  if (length(as.vector(lcolor)) == 1) {
    lcolor <- rep(lcolor, 2)
  } else if (length(as.vector(lcolor)) != 2) {
    stop("\"lcolor\" must be a numeric vector of length 1 or 2.")
  }
  if (is.null(lwidth)) {
    if (use_modern_recipe) {
      lwidth <- 0.4
    } else if (isTRUE(theme.bw)) {
      lwidth <- 1.5  # pre-2.3.1 default for theme.bw=TRUE (classic mode)
    } else {
      lwidth <- 2
    }
  }
  if (length(as.vector(lwidth)) == 1) {
    lwidth <- rep(lwidth, 2)
  } else if (length(as.vector(lwidth)) != 2) {
    stop("\"lwidth\" must be a numeric vector of length 1 or 2.")
  }

  # Prepare data vectors for plotting logic
  current_time_vec <- data[[Period]]
  current_att_vec  <- data[[Estimate]]

  # CI.lower and CI.upper columns are now guaranteed to exist in 'data'
  # and be numeric (or NA_real_). Their names are given by the CI.lower and CI.upper parameters.
  CI.lower.val <- data[[CI.lower]]
  CI.upper.val <- data[[CI.upper]]

  current_count_vec <- if (!is.null(Count) && Count %in% names(data)) {
    data[[Count]]
  } else {
    rep(NA_real_, length(current_time_vec))
  }

  # Determine initial final_xlim for plotting
  final_xlim <- user_xlim
  if (is.null(final_xlim)) {
    if (show.count && !is.null(Count) && Count %in% names(data) && any(!is.na(data[[Count]]))) {
      finite_counts <- data[[Count]][is.finite(data[[Count]])]
      if (length(finite_counts) > 0) {
        maxC <- max(finite_counts, na.rm = TRUE)
        if (is.finite(maxC) && maxC > 0) {
          threshold <- maxC * proportion
          idx_valid_count_times <- data[[Count]] >= threshold & !is.na(data[[Count]]) & is.finite(data[[Period]])
          valid_times <- data[[Period]][idx_valid_count_times]

          if (length(valid_times) > 0) {
            final_xlim <- range(valid_times, na.rm = TRUE)
          } else {
            idx_finite_period <- is.finite(data[[Period]])
            final_xlim <- range(data[[Period]][idx_finite_period], na.rm = TRUE)
          }
        } else {
          idx_finite_period <- is.finite(data[[Period]])
          final_xlim <- range(data[[Period]][idx_finite_period], na.rm = TRUE)
        }
      } else {
        idx_finite_period <- is.finite(data[[Period]])
        final_xlim <- range(data[[Period]][idx_finite_period], na.rm = TRUE)
      }
    } else {
      idx_finite_period <- is.finite(data[[Period]])
      final_xlim <- range(data[[Period]][idx_finite_period], na.rm = TRUE)
    }
    if (any(!is.finite(final_xlim))) {
      idx_finite_period <- is.finite(data[[Period]])
      valid_times <- data[[Period]][idx_finite_period]
      if(length(valid_times) > 0) final_xlim <- range(valid_times) else final_xlim <- c(0,1)
    }
    if (final_xlim[1] == final_xlim[2]) final_xlim <- final_xlim + c(-0.5, 0.5)
  }


  # fill.gap logic
  time_vec_for_fillgap <- current_time_vec
  att_vec_for_fillgap  <- current_att_vec
  ci_l_vec_for_fillgap <- CI.lower.val # These are from data[[CI.lower]]
  ci_u_vec_for_fillgap <- CI.upper.val # These are from data[[CI.upper]]
  count_vec_for_fillgap<- current_count_vec

  if (fill.gap && !connected) {
    idx_finite_time <- is.finite(time_vec_for_fillgap)
    min_time_fill <- min(time_vec_for_fillgap[idx_finite_time], na.rm = TRUE)
    max_time_fill <- max(time_vec_for_fillgap[idx_finite_time], na.rm = TRUE)

    if (is.finite(min_time_fill) && is.finite(max_time_fill) && min_time_fill <= max_time_fill) {
      idx_finite_unique_times <- !is.na(time_vec_for_fillgap) & is.finite(time_vec_for_fillgap)
      unique_times_present <- unique(time_vec_for_fillgap[idx_finite_unique_times])


      diff_times <- diff(sort(unique_times_present))
      idx_valid_diffs <- is.finite(diff_times) & diff_times > 1e-9
      valid_diffs <- diff_times[idx_valid_diffs]

      step_by <- if(length(valid_diffs) > 0 && all(abs(valid_diffs - round(valid_diffs)) < 1e-9)) {
        1
      } else if (length(valid_diffs) > 0) {
        min(valid_diffs, na.rm = TRUE)
      } else {
        1
      }

      if(!is.finite(step_by) || step_by <= 1e-9) step_by <- 1

      if ( (max_time_fill - min_time_fill)/step_by + 1 > length(unique_times_present) ) {
        full_seq <- seq(min_time_fill, max_time_fill, by = step_by)
        time.add <- setdiff(full_seq, unique_times_present)
        if (length(time.add) > 0) {
          att_vec_for_fillgap  <- c(att_vec_for_fillgap, rep(0, length(time.add)))
          ci_l_vec_for_fillgap <- c(ci_l_vec_for_fillgap, rep(0, length(time.add)))
          ci_u_vec_for_fillgap <- c(ci_u_vec_for_fillgap, rep(0, length(time.add)))
          count_vec_for_fillgap<- c(count_vec_for_fillgap, rep(NA, length(time.add)))
          time_vec_for_fillgap <- c(time_vec_for_fillgap, time.add)
        }
      }
    }
  }

  plot_data <- data.frame(
    time = time_vec_for_fillgap
  )
  plot_data[[Estimate]] <- att_vec_for_fillgap
  plot_data[[CI.lower]] <- ci_l_vec_for_fillgap # Use the parameter name as col name
  plot_data[[CI.upper]] <- ci_u_vec_for_fillgap # Use the parameter name as col name
  if(!is.null(Count)) plot_data[[Count]] <- count_vec_for_fillgap
  else plot_data[["count_placeholder"]] <- count_vec_for_fillgap

  names(plot_data)[names(plot_data) == "time"] <- Period


  # Apply only.pre/only.post modifications to final_xlim
  if (!is.null(final_xlim) && all(is.finite(final_xlim))){
    if (only.pre) {
      final_xlim[2] <- min(final_xlim[2], 0, na.rm = TRUE)
      if (start0) final_xlim[2] <- min(final_xlim[2], -0.5, na.rm=TRUE)
    } else if (only.post) {
      final_xlim[1] <- max(final_xlim[1], 0, na.rm = TRUE)
      if (!start0) final_xlim[1] <- max(final_xlim[1], 0.5, na.rm=TRUE)
    }
    if(final_xlim[1] > final_xlim[2]) final_xlim <- rev(final_xlim)
  }

  if (!is.null(final_xlim)) {
    needs_resolve_min <- is.na(final_xlim[1]) || !is.finite(final_xlim[1])
    needs_resolve_max <- is.na(final_xlim[2]) || !is.finite(final_xlim[2])

    idx_finite_times_plot_data <- is.finite(plot_data[[Period]])
    finite_times_in_plot_data <- plot_data[[Period]][idx_finite_times_plot_data]
    data_derived_min_time <- if(length(finite_times_in_plot_data) > 0) min(finite_times_in_plot_data, na.rm=TRUE) else 0
    data_derived_max_time <- if(length(finite_times_in_plot_data) > 0) max(finite_times_in_plot_data, na.rm=TRUE) else 1
    if (data_derived_min_time > data_derived_max_time) { data_derived_min_time <- 0; data_derived_max_time <- 1; }


    if (needs_resolve_min) final_xlim[1] <- data_derived_min_time
    if (needs_resolve_max) final_xlim[2] <- data_derived_max_time


    if (final_xlim[1] > final_xlim[2]) {
      final_xlim <- range(c(data_derived_min_time, data_derived_max_time, final_xlim))
    }
    if (final_xlim[1] == final_xlim[2]) {
      final_xlim <- final_xlim + c(-0.5, 0.5)
    }
    idx <- plot_data[[Period]] >= final_xlim[1] &
      plot_data[[Period]] <= final_xlim[2]
    plot_data <- plot_data[idx, ]

  }


  if(nrow(plot_data) == 0 && !(is.null(final_xlim) && is.null(user_ylim))) {
    warning("No data points remaining after applying xlim/filters. Plot will be empty or may error.")
    empty_df_structure <- data.frame(temp_period = numeric(0), temp_est = numeric(0), temp_cil = numeric(0), temp_ciu = numeric(0))
    names(empty_df_structure) <- c(Period, Estimate, CI.lower, CI.upper)
    if(!is.null(Count)) empty_df_structure[[Count]] <- numeric(0)
    plot_data <- empty_df_structure
  }

  # Determine final ylim for plotting
  final_ylim <- user_ylim
  if (is.null(final_ylim)) {
    y_min_series <- plot_data[[Estimate]]
    y_max_series <- plot_data[[Estimate]]
    # CI.lower and CI.upper columns are guaranteed to be in plot_data
    idx_finite_cil <- is.finite(plot_data[[CI.lower]])
    y_min_series <- c(y_min_series, plot_data[[CI.lower]][idx_finite_cil])

    idx_finite_ciu <- is.finite(plot_data[[CI.upper]])
    y_max_series <- c(y_max_series, plot_data[[CI.upper]][idx_finite_ciu])


    idx_finite_ymin <- is.finite(y_min_series)
    y_min_data <- min(y_min_series[idx_finite_ymin], na.rm = TRUE)
    idx_finite_ymax <- is.finite(y_max_series)
    y_max_data <- max(y_max_series[idx_finite_ymax], na.rm = TRUE)


    if (!is.finite(y_min_data) && !is.finite(y_max_data) && nrow(plot_data)==0) {
      y_min_data <- 0; y_max_data <- 1;
    } else {
      if (!is.finite(y_min_data)) y_min_data <- if(is.finite(y_max_data)) y_max_data -1 else 0
      if (!is.finite(y_max_data)) y_max_data <- if(is.finite(y_min_data)) y_min_data +1 else 1
    }
    if (y_min_data == y_max_data) {
      y_min_data <- y_min_data - 0.5
      y_max_data <- y_max_data + 0.5
    }

    range_val <- y_max_data - y_min_data
    if (!is.finite(range_val) || range_val == 0) range_val <- 1

    ## Modern recipe: small top padding so the highest CI doesn't graze the panel
    ## ceiling, larger top padding when stats annotation is drawn at top-left.
    ## Legacy: keep zero top padding (preserve the pre-2.3.1 layout).
    top_expand_factor <- if (use_modern_recipe) 0.08 else 0
    bot_expand_factor <- 0

    p_label_text_for_ylim_calc <- function(stats_vals, labs_vals) {
      text <- ""
      s_display <- formatC(stats_vals, digits = 3, format = "f", flag = "#")
      for (k in seq_along(s_display)) {
        text <- paste0(text, labs_vals[k], ifelse(labs_vals[k]=="", "", ": "), s_display[k], "\n")
      }
      sub("\n$", "", text)
    }
    if (!is.null(stats)) {
      num_stat_lines <- ceiling(length(unlist(strsplit(p_label_text_for_ylim_calc(stats, stats.labs), "\n"))))
      ## Reserve room for the stats annotation block at the top.
      ## Per-line factor accounts for the 1.25x text size used below;
      ## doubled from 0.06 to 0.12 so stats text doesn't graze the
      ## leftmost CIs on data with high pre-treatment estimates.
      if (use_modern_recipe) {
        top_expand_factor <- top_expand_factor + 0.12 * num_stat_lines
      }
    }
    if (show.count && !is.null(Count) && Count %in% names(plot_data)) {
      idx_valid_count_plot_data <- !is.na(plot_data[[Count]]) & plot_data[[Count]] > 0
      if (any(idx_valid_count_plot_data)) {
        bot_expand_factor <- bot_expand_factor + 0.35
      }
    } else if (use_modern_recipe) {
      ## No count bars => still reserve a small bottom margin for breathing room
      bot_expand_factor <- bot_expand_factor + 0.06
    }

    final_ylim <- c(
      y_min_data - bot_expand_factor * range_val,
      y_max_data + top_expand_factor * range_val
    )
    if(any(!is.finite(final_ylim))) {
      final_ylim <- c(y_min_data, y_max_data)
      if(final_ylim[1] == final_ylim[2]) final_ylim <- final_ylim + c(-0.5, 0.5)
    }
  }


  p <- ggplot(data = plot_data)


  if (use_modern_recipe) {
    p <- p + .modern_theme(base_size = 11)
  } else if (isTRUE(theme.bw)) {
    p <- p + ggplot2::theme_bw()
  }
  if (gridOff) {
    p <- p + theme(
      panel.grid.major = element_blank(),
      panel.grid.minor = element_blank()
    )
  }

  ## Modern: peach (or auto-lightened) rectangle behind highlight periods,
  ## drawn before data layers so points/lines paint on top. Opt-in via
  ## `highlight.fill = TRUE`. Default is FALSE: the colored glyph alone is
  ## sufficient signal that a period is part of a test window, and the
  ## glyph-only look survives grayscale printing without surprise.
  if (use_modern_recipe && isTRUE(highlight.fill) &&
      !is.null(highlight.periods) && length(highlight.periods) > 0) {
    hl_n <- length(highlight.periods)
    hl_pt_colors <- if (!is.null(highlight.colors) && length(highlight.colors) == hl_n) {
      highlight.colors
    } else {
      rep(.MODERN_PLACEBO_PT, hl_n)
    }
    for (i in seq_len(hl_n)) {
      hl_fill_i <- .lighten_color(hl_pt_colors[i], amount = 0.70)
      p <- p + annotate("rect",
                        xmin = as.numeric(highlight.periods[i]) - 0.5,
                        xmax = as.numeric(highlight.periods[i]) + 0.5,
                        ymin = -Inf, ymax = Inf,
                        fill = hl_fill_i, alpha = 0.55)
    }
  }

  p <- p + geom_hline(yintercept = 0, colour = lcolor[1], linewidth = lwidth[1], linetype = ltype[1])

  vline_pos <- if (start0) -0.5 else 0.5
  show_vline <- TRUE
  if (only.pre || only.post) show_vline <- FALSE
  if (!is.null(final_xlim) && all(is.finite(final_xlim))) {
    if (vline_pos < final_xlim[1] || vline_pos > final_xlim[2]) {
      show_vline <- FALSE
    }
  } else if (is.null(final_xlim)) {
    if (nrow(plot_data) > 0 && Period %in% names(plot_data)) {
      idx_finite_period_plot_data <- is.finite(plot_data[[Period]])
      if (any(idx_finite_period_plot_data)) {
        data_period_range <- range(plot_data[[Period]][idx_finite_period_plot_data], na.rm = TRUE)
        if (vline_pos < data_period_range[1] || vline_pos > data_period_range[2]) show_vline <- FALSE
      } else {
        show_vline <- FALSE
      }
    } else {
      show_vline <- FALSE
    }
  }


  if (show_vline) {
    p <- p + geom_vline(
      xintercept = vline_pos, colour = lcolor[2],
      linewidth = lwidth[2], linetype = ltype[2]
    )
  }


  intervals <- NULL
  if (!is.null(highlight.periods) && length(highlight.periods) > 0) {
    if (is.null(highlight.colors) || length(highlight.colors) != length(highlight.periods)) {
      warning("highlight.colors not provided or mismatched length; using default rainbow colors.")
      highlight.colors <- grDevices::rainbow(length(highlight.periods))
    }
    if (is.null(highlight.shapes) || length(highlight.shapes) != length(highlight.periods)) {
      highlight.shapes <- rep(19, length(highlight.periods))
    }
    intervals <- data.frame(
      start = highlight.periods - 0.5,
      end   = highlight.periods + 0.5,
      color = highlight.colors,
      shape = highlight.shapes
    )
  }

  remove_strictly_inside <- function(df, intervals_df, estimate_col, ci_l_col, ci_u_col, period_col) {
    if (is.null(intervals_df) || nrow(intervals_df) == 0) return(df)
    out_df <- df
    for (i in seq_len(nrow(intervals_df))) {
      s <- intervals_df$start[i]
      e <- intervals_df$end[i]
      idx_to_na <- which(out_df[[period_col]] > s & out_df[[period_col]] < e)
      if(length(idx_to_na) > 0) {
        out_df[[estimate_col]][idx_to_na] <- NA
        if(ci_l_col %in% names(out_df)) out_df[[ci_l_col]][idx_to_na] <- NA
        if(ci_u_col %in% names(out_df)) out_df[[ci_u_col]][idx_to_na] <- NA
      }
    }
    return(out_df)
  }

  ## Helper: add connected layers (ribbon + line + optional points) for a data subset
  .add_connected_layers <- function(p, df, col, lt, Period, Estimate, CI.lower, CI.upper,
                                    est.lwidth, est.pointsize, show.points, ci.outline,
                                    point.shape = 19) {
    if (nrow(df) == 0) return(p)
    outline_col <- ifelse(ci.outline, grDevices::adjustcolor(col, offset = c(0.3,0.3,0.3,0)), NA)
    p <- p +
      geom_ribbon(data = df, aes(x = .data[[Period]], ymin = .data[[CI.lower]], ymax = .data[[CI.upper]]),
                  fill = col, alpha = 0.2, inherit.aes = FALSE, na.rm = TRUE, color = outline_col) +
      geom_line(data = df, aes(x = .data[[Period]], y = .data[[Estimate]]),
                linewidth = est.lwidth, color = col, linetype = lt, inherit.aes = FALSE, na.rm = TRUE)
    if (show.points) {
      idx_int <- abs(as.numeric(df[[Period]]) - round(as.numeric(df[[Period]]))) < 1e-8 & !is.na(df[[Estimate]])
      df_int <- df[idx_int, ]
      if (nrow(df_int) > 0) {
        ## Non-circle shapes (triangles, diamonds) appear smaller at the same
        ## nominal size; scale up by 1.5× so they match circle visual weight
        ## Triangle (17) and diamond (18) glyphs render at a larger visual
        ## area than circle (19) at the same size aesthetic. Scale them down
        ## by 0.85 so highlight overlays look slightly smaller (not larger)
        ## than the base circles in the same plot.
        pt_size <- if (point.shape != 19) est.pointsize * 0.85 else est.pointsize
        p <- p + geom_point(data = df_int, aes(x = .data[[Period]], y = .data[[Estimate]]),
                            size = pt_size, color = col, shape = point.shape,
                            inherit.aes = FALSE, na.rm = TRUE)
      }
    }
    p
  }

  ## Split plot_data into pre-treatment and post-treatment
  ## When start0 = TRUE, period 0 is the first post-treatment period
  if (start0) {
    is_pre  <- plot_data[[Period]] < 0
    is_post <- plot_data[[Period]] >= 0
  } else {
    is_pre  <- plot_data[[Period]] <= 0
    is_post <- plot_data[[Period]] > 0
  }
  data_pre  <- plot_data[is_pre, ]
  data_post <- plot_data[is_post, ]

  if (connected) {
    data_base <- plot_data
    if (!is.null(intervals) && nrow(intervals) > 0) {
      data_base <- remove_strictly_inside(data_base, intervals, Estimate, CI.lower, CI.upper, Period)
    }
    if (start0) {
      data_base_pre  <- data_base[data_base[[Period]] < 0, ]
      data_base_post <- data_base[data_base[[Period]] >= 0, ]
    } else {
      data_base_pre  <- data_base[data_base[[Period]] <= 0, ]
      data_base_post <- data_base[data_base[[Period]] > 0, ]
    }

    ## Pre-treatment: long-dashed line
    p <- .add_connected_layers(p, data_base_pre, pre.color, "longdash",
                               Period, Estimate, CI.lower, CI.upper,
                               est.lwidth, est.pointsize, show.points, ci.outline)
    ## Post-treatment: solid line
    p <- .add_connected_layers(p, data_base_post, post.color, "solid",
                               Period, Estimate, CI.lower, CI.upper,
                               est.lwidth, est.pointsize, show.points, ci.outline)

    ## Bridge segment: connect the last pre point to the first post point
    ## When start0=FALSE: pre ends at 0, post starts at 0.5 → bridge 0→0.5
    ## When start0=TRUE:  pre ends at -0.5, post starts at 0 → bridge -0.5→0
    if (start0) {
      bridge_from <- -0.5; bridge_to <- 0
    } else {
      bridge_from <- 0; bridge_to <- 0.5
    }
    idx_from <- which(abs(data_base[[Period]] - bridge_from) < 1e-8)
    idx_to   <- which(abs(data_base[[Period]] - bridge_to)   < 1e-8)
    if (length(idx_from) > 0 && length(idx_to) > 0) {
      bridge_df <- data_base[c(idx_from[1], idx_to[1]), ]
      p <- .add_connected_layers(p, bridge_df, pre.color, "longdash",
                                 Period, Estimate, CI.lower, CI.upper,
                                 est.lwidth, est.pointsize, FALSE, ci.outline)
    }

    ## Highlight interval overlays (e.g., placebo, carryover)
    if (!is.null(intervals) && nrow(intervals) > 0) {
      for (i in seq_len(nrow(intervals))) {
        sub_color <- intervals$color[i]
        sub_shape <- if (!is.null(intervals$shape)) intervals$shape[i] else 19
        idx_sub <- plot_data[[Period]] >= intervals$start[i] & plot_data[[Period]] <= intervals$end[i]
        sub.data <- plot_data[idx_sub, ]
        p <- .add_connected_layers(p, sub.data, sub_color, "solid",
                                   Period, Estimate, CI.lower, CI.upper,
                                   est.lwidth, est.pointsize, show.points, ci.outline,
                                   point.shape = sub_shape)
      }
    }
  } else { # Not connected
    ## When highlight overlays will be drawn, exclude those periods from
    ## the base pre/post layer so we don't render two stacked symbols
    ## (a base circle peeking out from under the highlight triangle / diamond).
    base_pre  <- data_pre
    base_post <- data_post
    if (!is.null(highlight.periods) && length(highlight.periods) > 0) {
      hp_set <- as.numeric(highlight.periods)
      base_pre  <- base_pre[!(as.numeric(base_pre[[Period]])  %in% hp_set), ]
      base_post <- base_post[!(as.numeric(base_post[[Period]]) %in% hp_set), ]
    }
    ## Pre-treatment points
    if (nrow(base_pre) > 0) {
      p <- p + geom_pointrange(
        data = base_pre,
        aes(x = .data[[Period]], y = .data[[Estimate]], ymin = .data[[CI.lower]], ymax = .data[[CI.upper]]),
        linewidth = est.lwidth, color = pre.color, fill = pre.color,
        size = est.pointsize / 6,
        inherit.aes = FALSE, na.rm = TRUE)
    }
    ## Post-treatment points
    if (nrow(base_post) > 0) {
      p <- p + geom_pointrange(
        data = base_post,
        aes(x = .data[[Period]], y = .data[[Estimate]], ymin = .data[[CI.lower]], ymax = .data[[CI.upper]]),
        linewidth = est.lwidth, color = post.color, fill = post.color,
        size = est.pointsize / 6,
        inherit.aes = FALSE, na.rm = TRUE)
    }

    ## Highlight overlays
    if (!is.null(highlight.periods) && length(highlight.periods) > 0) {
      for (i in seq_along(highlight.periods)) {
        hp_period <- highlight.periods[i]
        hp_color <- highlight.colors[i]
        hp_shape <- if (!is.null(highlight.shapes)) highlight.shapes[i] else 19
        idx_sub <- as.numeric(plot_data[[Period]]) == hp_period
        sub_data_point <- plot_data[idx_sub, ]
        if (nrow(sub_data_point) > 0) {
          if (hp_shape != 19) {
            # Use linerange + point to support custom shapes
            p <- p + geom_linerange(
              data = sub_data_point,
              aes(x = .data[[Period]], ymin = .data[[CI.lower]], ymax = .data[[CI.upper]]),
              linewidth = est.lwidth, color = hp_color, inherit.aes = FALSE, na.rm = TRUE)
            p <- p + geom_point(
              data = sub_data_point,
              aes(x = .data[[Period]], y = .data[[Estimate]]),
              ## Diamonds (shape 18) have less glyph area than triangles (17)
              ## at the same nominal size, so scale them up to match the
              ## triangle's visual weight.
              size = if (hp_shape == 18) est.pointsize * 1.3 else est.pointsize * 0.9,
              color = hp_color, shape = hp_shape,
              inherit.aes = FALSE, na.rm = TRUE)
          } else {
            p <- p + geom_pointrange(
              data = sub_data_point,
              aes(x = .data[[Period]], y = .data[[Estimate]], ymin = .data[[CI.lower]], ymax = .data[[CI.upper]]),
              linewidth = est.lwidth, color = hp_color, fill = hp_color,
              size = est.pointsize / 6,
              inherit.aes = FALSE, na.rm = TRUE)
          }
        }
      }
    }
  }

  # Build shape legend when highlight shapes differ from base circle.
  # Modern theme suppresses this legend when the highlight rect is drawn ---
  # the peach/blue background + colored points self-document, and the
  # pre/post legend entries are misleading (both render in grey20).
  modern_suppresses_legend <- use_modern_recipe &&
    !is.null(highlight.periods) && length(highlight.periods) > 0
  if (!is.null(highlight.shapes) && any(highlight.shapes != 19) &&
      !legendOff && !modern_suppresses_legend) {
    unique_shapes <- unique(highlight.shapes[highlight.shapes != 19])

    # Standard shape label map
    shape_label_map <- c("17" = "Placebo Test", "18" = "Carryover Test",
                          "15" = "Square", "8" = "Asterisk")

    # Build legend entries conditionally:
    # - Include "Pre-treatment" only if pre-treatment data is shown
    # - Include "Post-treatment" only if post-treatment data is shown (not only.pre)
    legend_ids    <- c()
    legend_shapes <- c()
    legend_colors <- c()
    legend_labels <- c()

    has_pre  <- nrow(data_pre) > 0
    has_post <- nrow(data_post) > 0 && !only.pre

    if (has_pre) {
      legend_ids    <- c(legend_ids, "pre")
      legend_shapes <- c(legend_shapes, 19)
      legend_colors <- c(legend_colors, pre.color)
      legend_labels <- c(legend_labels, pre.label)
    }
    if (has_post) {
      legend_ids    <- c(legend_ids, "post")
      legend_shapes <- c(legend_shapes, 19)
      legend_colors <- c(legend_colors, post.color)
      legend_labels <- c(legend_labels, post.label)
    }

    # Map each unique highlight shape to its color from highlight.colors
    for (s in unique_shapes) {
      s_chr <- as.character(s)
      lbl <- if (s_chr %in% names(shape_label_map)) shape_label_map[s_chr] else paste("Shape", s)
      idx <- which(highlight.shapes == s)[1]
      hcol <- if (!is.null(highlight.colors) && !is.na(idx)) highlight.colors[idx] else color
      legend_ids    <- c(legend_ids, s_chr)
      legend_shapes <- c(legend_shapes, s)
      legend_colors <- c(legend_colors, hcol)
      legend_labels <- c(legend_labels, lbl)
    }
    names(legend_shapes) <- legend_ids
    names(legend_colors) <- legend_ids
    names(legend_labels) <- legend_ids

    legend_df <- data.frame(
      x = rep(NA_real_, length(legend_ids)),
      y = rep(NA_real_, length(legend_ids)),
      legend_id = legend_ids,
      stringsAsFactors = FALSE
    )

    # Use 2 rows when 4+ legend entries to avoid overflow
    legend_nrow <- if (length(legend_ids) >= 4) 2 else 1

    p <- p + geom_point(
      data = legend_df,
      aes(x = x, y = y, shape = legend_id, color = legend_id),
      na.rm = TRUE, inherit.aes = FALSE
    ) +
      ggplot2::scale_shape_manual(
        values = legend_shapes,
        labels = legend_labels,
        name = NULL
      ) +
      ggplot2::scale_color_manual(
        values = legend_colors,
        labels = legend_labels,
        name = NULL
      ) +
      guides(shape = guide_legend(nrow = legend_nrow, title = NULL),
             color = guide_legend(nrow = legend_nrow, title = NULL)) +
      theme(legend.position = "bottom",
            legend.box = "vertical",
            legend.margin = margin(0, 0, 0, 0),
            legend.text = element_text(size = if (!is.null(cex.legend)) cex.legend * 0.6 else 8),
            plot.margin = margin(5, 5, 15, 5))
  }

  ## Title styling: modern recipe is plain + left-aligned; legacy is bold + centered.
  modern_title_face  <- if (use_modern_recipe) "plain" else "bold"
  modern_title_hjust <- if (use_modern_recipe) 0     else 0.5

  p <- p +
    xlab(xlab) + ylab(ylab) +
    theme(
      axis.title  = element_text(size = cex.lab),
      axis.text   = element_text(color = "black", size = cex.axis),
      axis.text.x = element_text(angle = angle, vjust = x.v, hjust = x.h),
      plot.title  = element_text(size = cex.main, hjust = modern_title_hjust, face = modern_title_face)
    )



  if (show.count && !is.null(Count) && Count %in% names(plot_data)) {
    idx_rect_data <- !is.na(plot_data[[Count]]) & plot_data[[Count]] > 0
    rect_data <- plot_data[idx_rect_data, ]

    if(nrow(rect_data) > 0) {
      rect_data[[Period]] <- as.numeric(rect_data[[Period]])

      rect_data[,"xmin"] <- rect_data[[Period]] - 0.2
      rect_data[,"xmax"] <- rect_data[[Period]] + 0.2

      plot_range_y <- final_ylim[2] - final_ylim[1]
      if(!is.finite(plot_range_y) || plot_range_y <=0) plot_range_y <- 1

      ## Divide the bot_expand space into bars (bottom 50%) and gap (top 50%)
      ## This guarantees a visible gap without feedback loops
      bot_space <- if (exists("y_min_data") && is.finite(y_min_data)) {
        y_min_data - final_ylim[1]
      } else {
        plot_range_y * 0.35
      }
      rect_bar_max_h <- bot_space * 0.50
      rect.min_yval  <- final_ylim[1]

      max_rect_count <- max(rect_data[[Count]], na.rm = TRUE)

      if (is.finite(max_rect_count) && max_rect_count > 0) {
        rect_data[,"ymin"] <- rect.min_yval
        rect_data[,"ymax"] <- rect.min_yval + (rect_data[[Count]] / max_rect_count) * rect_bar_max_h
        p <- p + geom_rect(
          data = rect_data, aes(xmin = .data$xmin, xmax = .data$xmax, ymin = .data$ymin, ymax = .data$ymax),
          fill = count.color, linewidth = 0.2, inherit.aes = FALSE, alpha = count.alpha, color = count.outline.color)

        max_count_val <- max_rect_count
        idx_max_count <- which(rect_data[[Count]] == max_count_val)
        max_count_time_pos <- rect_data[[Period]][idx_max_count[1]]

        if (length(idx_max_count) > 1) {
          preferred_pos <- if (start0) -1 else 0
          idx_preferred_pos_in_max <- rect_data[[Period]][idx_max_count] == preferred_pos
          if (any(idx_preferred_pos_in_max)) {
            max_count_time_pos <- preferred_pos
          } else {
            if(!is.null(final_xlim) && all(is.finite(final_xlim))){
              xlim_center <- mean(final_xlim)
              max_count_time_pos <- rect_data[[Period]][idx_max_count[which.min(abs(rect_data[[Period]][idx_max_count] - xlim_center))]]
            }
          }
        }
        p <- p + annotate("text",
                          x = max_count_time_pos,
                          y = rect.min_yval + rect_bar_max_h + 0.02 * plot_range_y,
                          label = format(max_count_val, big.mark=","),
                          size = cex.text * 0.7,
                          hjust = 0.5)
      }
    }
  }

  if (!is.null(stats)) {
    actual_stats_pos <- stats.pos
    stats_hjust_val <- 0
    stats_vjust_val <- 1

    if (is.null(actual_stats_pos)) {
      x_min_for_stats <- if(!is.null(final_xlim) && is.finite(final_xlim[1])) final_xlim[1] else {
        if(nrow(plot_data) > 0 && Period %in% names(plot_data)) {
          idx_finite_period_plot_data <- is.finite(plot_data[[Period]])
          if(any(idx_finite_period_plot_data)) min(plot_data[[Period]][idx_finite_period_plot_data], na.rm = TRUE) else 0
        } else 0
      }
      x_max_for_stats <- if(!is.null(final_xlim) && is.finite(final_xlim[2])) final_xlim[2] else {
        if(nrow(plot_data) > 0 && Period %in% names(plot_data)) {
          idx_finite_period_plot_data <- is.finite(plot_data[[Period]])
          if(any(idx_finite_period_plot_data)) max(plot_data[[Period]][idx_finite_period_plot_data], na.rm = TRUE) else 1
        } else 1
      }
      if (x_min_for_stats >= x_max_for_stats) {
        x_max_for_stats <- x_min_for_stats + 1
      }


      y_min_for_stats <- if(is.finite(final_ylim[1])) final_ylim[1] else {
        y_coords <- c(plot_data[[Estimate]], plot_data[[CI.lower]], plot_data[[CI.upper]])
        idx_finite_y_coords <- is.finite(y_coords)
        y_coords <- y_coords[idx_finite_y_coords]
        if(length(y_coords) > 0) min(y_coords, na.rm = TRUE) else 0
      }
      y_max_for_stats <- if(is.finite(final_ylim[2])) final_ylim[2] else {
        y_coords <- c(plot_data[[Estimate]], plot_data[[CI.lower]], plot_data[[CI.upper]])
        idx_finite_y_coords <- is.finite(y_coords)
        y_coords <- y_coords[idx_finite_y_coords]
        if(length(y_coords) > 0) max(y_coords, na.rm = TRUE) else 1
      }
      if (y_min_for_stats >= y_max_for_stats) {
        y_max_for_stats <- y_min_for_stats + 1
      }

      current_x_range <- x_max_for_stats - x_min_for_stats
      if (!is.finite(current_x_range) || current_x_range <= 1e-9) current_x_range <- 1

      current_y_range <- y_max_for_stats - y_min_for_stats
      if (!is.finite(current_y_range) || current_y_range <= 1e-9) current_y_range <- 1

      ## Modern: symmetric inset from the top-left panel corner so the
      ## stats block looks anchored, not floating mid-panel. Both factors
      ## are 2.5% of their respective ranges; the slight pixel asymmetry
      ## from non-square panels (width:height ~ 1.33 for typical fect
      ## plots) is small enough to read as visually symmetric. Legacy
      ## keeps its pre-2.3.1 0.02/0.02 placement.
      padding_x_factor <- if (use_modern_recipe) 0.025 else 0.02
      padding_y_factor <- if (use_modern_recipe) 0.025 else 0.02

      actual_stats_pos <- c(
        x_min_for_stats + padding_x_factor * current_x_range,
        y_max_for_stats - padding_y_factor * current_y_range
      )

      if(!is.finite(actual_stats_pos[1])) actual_stats_pos[1] <- x_min_for_stats
      if(!is.finite(actual_stats_pos[2])) actual_stats_pos[2] <- y_max_for_stats
    }

    p_label_text <- ""
    for (ii in seq_along(stats_display)) {
      p_label_text <- paste0(p_label_text, stats.labs[ii], ifelse(stats.labs[ii]=="", "", ": "), stats_display[ii], "\n")
    }
    p_label_text <- sub("\n$", "", p_label_text)

    p <- p + annotate("text",
                      x = actual_stats_pos[1], y = actual_stats_pos[2],
                      label = p_label_text,
                      size = if (use_modern_recipe) cex.text * 1.0 else cex.text,
                      hjust = stats_hjust_val,
                      vjust = stats_vjust_val,
                      lineheight = .8)
  }


  if (is.null(main)) {
    p <- p + ggtitle("Estimated Dynamic Treatment Effects")
  } else if (main != "") {
    p <- p + ggtitle(main)
  }
  p <- p + coord_cartesian(xlim = c(final_xlim[1]-0.2,final_xlim[2]+0.2), ylim = final_ylim)
  p <- p + scale_x_continuous(
    breaks = function(lims) {
      ## 1. ggplot’s own suggestion (needs {scales})
      br <- scales::pretty_breaks()(lims)

      ## 2. keep only integers
      br <- br[abs(br - round(br)) < 1e-8]

      ## 3. if none left (very narrow panel or weird limits), make 5 integers
      if (length(br) == 0) {
        br <- round(seq(lims[1], lims[2], length.out = 5))
      }
      br
    }
  )


  if (!is.null(xbreaks)) {
    p <- p + scale_x_continuous(breaks = xbreaks)
  }
  if (!is.null(ybreaks)) {
    p <- p + scale_y_continuous(breaks = ybreaks)
  }
  if (!is.null(xangle)) {
    p <- p + theme(axis.text.x = element_text(angle = xangle, hjust = 1))
  }
  if (!is.null(yangle)) {
    p <- p + theme(axis.text.y = element_text(angle = yangle))
  }
  if (isTRUE(legendOff)) {
    p <- p + theme(legend.position = "none")
  }
  if (!is.null(cex.legend)) {
    p <- p + theme(legend.text = element_text(size = cex.legend))
  }

  return(p)
}

## ---------------------------------------------------------
## Tests for plot refactoring: pre/post colors, esplot(fect),
## harmonized defaults
## ---------------------------------------------------------

## Shared fixture
out_fect <- NULL
setup_once <- function() {
    if (is.null(out_fect)) {
        set.seed(9001)
        data(simdata, package = "fect")
        out_fect <<- fect::fect(Y ~ D + X1 + X2, data = simdata,
                                index = c("id", "time"), method = "fe",
                                se = TRUE, nboots = 30, parallel = FALSE)
    }
}

## =========================================================
## Section A: Pre/Post Color Distinction
## =========================================================

test_that("A1: default gap plot has pre.color=gray50, post.color=black", {

  skip_on_cran()
    setup_once()
    p <- plot(out_fect, type = "gap")
    expect_true(inherits(p, "gg") || inherits(p, "ggplot"))
    ## Extract the geom layers — should have separate pre and post pointranges
    layers <- p$layers
    expect_true(length(layers) >= 3,
                info = "Gap plot should have at least hline + pre-pointrange + post-pointrange")
})

test_that("A2: connected gap plot uses dashed pre-treatment line", {

  skip_on_cran()
    setup_once()
    p <- plot(out_fect, type = "gap", connected = TRUE)
    expect_true(inherits(p, "gg"))
    layers <- p$layers
    ## Should have ribbon + line layers for both pre and post
    expect_true(length(layers) >= 4,
                info = "Connected gap should have hline + pre-ribbon/line + post-ribbon/line")
})

test_that("A3: custom pre.color and post.color", {

  skip_on_cran()
    setup_once()
    p <- plot(out_fect, pre.color = "steelblue", post.color = "firebrick")
    expect_true(inherits(p, "gg"))
})

test_that("A4: pre.color = post.color = color reproduces single-color behavior", {

  skip_on_cran()
    setup_once()
    p <- plot(out_fect, pre.color = "red", post.color = "red", color = "red")
    expect_true(inherits(p, "gg"))
})

## =========================================================
## Section B: esplot() Accepts fect Objects
## =========================================================

test_that("B1: esplot(fect_object) produces a valid plot", {

  skip_on_cran()
    setup_once()
    p <- esplot(out_fect, main = "esplot(fect)")
    expect_true(inherits(p, "gg"))
})

test_that("B2: esplot(fect_object) matches plot.fect gap output structure", {

  skip_on_cran()
    setup_once()
    p1 <- plot(out_fect, type = "gap")
    p2 <- esplot(out_fect)
    ## Both should be ggplot objects with similar layer count
    expect_true(inherits(p1, "gg"))
    expect_true(inherits(p2, "gg"))
    ## Both should have the same number of data points
    d1 <- suppressWarnings(ggplot2::ggplot_build(p1)$data)
    d2 <- suppressWarnings(ggplot2::ggplot_build(p2)$data)
    ## At least the point data should have similar row counts
    ## (exact match may differ due to count bars, but should be close)
    expect_true(length(d1) > 0)
    expect_true(length(d2) > 0)
})

test_that("B3: esplot(fect_object) with custom colors", {

  skip_on_cran()
    setup_once()
    p <- esplot(out_fect, pre.color = "blue", post.color = "darkred")
    expect_true(inherits(p, "gg"))
})

test_that("B4: esplot(fect_object) with connected = TRUE", {

  skip_on_cran()
    setup_once()
    p <- esplot(out_fect, connected = TRUE)
    expect_true(inherits(p, "gg"))
})

## =========================================================
## Section C: esplot() Backward Compatibility (data.frame input)
## =========================================================

test_that("C1: esplot(data.frame) still works", {

  skip_on_cran()
  df <- data.frame(
        Period = -3:3,
        ATT = c(-0.1, 0.05, -0.02, 0.01, 1.5, 2.0, 2.5),
        CI.lower = c(-0.1, 0.05, -0.02, 0.01, 1.5, 2.0, 2.5) - 0.5,
        CI.upper = c(-0.1, 0.05, -0.02, 0.01, 1.5, 2.0, 2.5) + 0.5
    )
    p <- esplot(df, show.count = FALSE)
    expect_true(inherits(p, "gg"))
})

test_that("C2: esplot(data.frame) with pre/post colors", {

  skip_on_cran()
  df <- data.frame(
        Period = -3:3,
        ATT = c(-0.1, 0.05, -0.02, 0.01, 1.5, 2.0, 2.5),
        CI.lower = c(-0.1, 0.05, -0.02, 0.01, 1.5, 2.0, 2.5) - 0.5,
        CI.upper = c(-0.1, 0.05, -0.02, 0.01, 1.5, 2.0, 2.5) + 0.5
    )
    p <- esplot(df, show.count = FALSE, pre.color = "gray70", post.color = "black")
    expect_true(inherits(p, "gg"))
})

test_that("C3: esplot(data.frame) connected with pre/post", {

  skip_on_cran()
  df <- data.frame(
        Period = -3:3,
        ATT = c(-0.1, 0.05, -0.02, 0.01, 1.5, 2.0, 2.5),
        CI.lower = c(-0.1, 0.05, -0.02, 0.01, 1.5, 2.0, 2.5) - 0.5,
        CI.upper = c(-0.1, 0.05, -0.02, 0.01, 1.5, 2.0, 2.5) + 0.5
    )
    p <- esplot(df, connected = TRUE, show.count = FALSE)
    expect_true(inherits(p, "gg"))
})

## =========================================================
## Section D: Harmonized Defaults
## =========================================================

test_that("D1: esplot show.points defaults to TRUE", {

  skip_on_cran()
  expect_equal(formals(esplot)$show.points, TRUE)
})

test_that("D2: esplot show.count defaults to TRUE", {

  skip_on_cran()
  expect_equal(formals(esplot)$show.count, TRUE)
})

test_that("D3: plot.fect count.color uses 'gray' not 'grey'", {

  skip_on_cran()
    ## Check that the default is set consistently
    ## (This is a static check; the actual default is resolved at runtime)
    setup_once()
    ## Just verify the plot works with default colors
    p <- plot(out_fect, type = "gap")
    expect_true(inherits(p, "gg"))
})

## =========================================================
## Section E: Highlight / Placebo Overlay on Pre/Post
## =========================================================

test_that("E1: placebo test plot with pre/post colors", {

  skip_on_cran()
    data(simdata, package = "fect")
    out_p <- suppressWarnings(fect::fect(
        Y ~ D, data = simdata, index = c("id", "time"),
        method = "fe", se = TRUE, nboots = 30, parallel = FALSE,
        placeboTest = TRUE, placebo.period = c(-2, 0), CV = FALSE
    ))
    p <- plot(out_p)
    expect_true(inherits(p, "gg"))
})

test_that("E2: esplot highlight.periods overlays on pre/post", {

  skip_on_cran()
  df <- data.frame(
        Period = -3:3,
        ATT = c(-0.1, 0.05, -0.02, 0.01, 1.5, 2.0, 2.5),
        CI.lower = c(-0.1, 0.05, -0.02, 0.01, 1.5, 2.0, 2.5) - 0.5,
        CI.upper = c(-0.1, 0.05, -0.02, 0.01, 1.5, 2.0, 2.5) + 0.5
    )
    p <- esplot(df, show.count = FALSE,
                highlight.periods = c(-2, -1, 0),
                highlight.colors = rep("blue", 3))
    expect_true(inherits(p, "gg"))
})

## =========================================================
## Section F: Edge Cases
## =========================================================

test_that("F1: all-pre data (only.pre = TRUE)", {

  skip_on_cran()
    setup_once()
    p <- esplot(out_fect, only.pre = TRUE)
    expect_true(inherits(p, "gg"))
})

test_that("F2: all-post data (only.post = TRUE)", {

  skip_on_cran()
    setup_once()
    p <- esplot(out_fect, only.post = TRUE)
    expect_true(inherits(p, "gg"))
})

test_that("F3: esplot errors on fect object without se", {

  skip_on_cran()
    data(simdata, package = "fect")
    out_nose <- fect::fect(Y ~ D, data = simdata, index = c("id", "time"),
                           method = "fe", se = FALSE, CV = FALSE)
    expect_error(esplot(out_nose))
})

test_that("F4: equiv plot with pre/post colors", {

  skip_on_cran()
    setup_once()
    p <- plot(out_fect, type = "equiv")
    expect_true(inherits(p, "gg"))
})

test_that("cumu/att.cumu/esplot run without error on fect output", {
  skip_on_cran()
  suppressWarnings(try(data("simdata", package = "fect"), silent = TRUE))
  expect_true(exists("simdata"))
  out <- fect::fect(
    Y ~ D + X1 + X2,
    data = simdata,
    index = c("id", "time"),
    method = "ife",
    r = 1,
    CV = FALSE,
    se = FALSE,
    parallel = FALSE
  )
  # cumu (att.cumu)
  # att.cumu requires uncertainty objects from keep.sims=TRUE
  out_keep <- fect::fect(
    Y ~ D + X1 + X2,
    data = simdata,
    index = c("id", "time"),
    method = "ife",
    r = 1,
    CV = FALSE,
    se = TRUE,
    nboots = 20,
    keep.sims = TRUE,
    parallel = FALSE
  )
  c <- att.cumu(out_keep, period = c(1, 3), plot = FALSE)
  expect_true(is.matrix(c) || is.data.frame(c))
  # esplot
  p <- esplot(out_keep$est.att, Estimate = "ATT")
  expect_s3_class(p, "ggplot")
})

###################################################################
## Matrix Completion Function
###################################################################
fect_mc <- function(Y, # Outcome variable, (T*N) matrix
                    X, # Explanatory variables:  (T*N*p) array
                    D,
                    W,
                    I,
                    II,
                    T.on,
                    T.off = NULL,
                    T.on.carry = NULL,
                    T.on.balance = NULL,
                    balance.period = NULL,
                    lambda.cv = 1e10,
                    force,
                    hasF = 1,
                    hasRevs = 1,
                    tol, # tolerance level
                    max.iteration = 1000,
                    boot = FALSE, # bootstrapped sample
                    norm.para = NULL,
                    placeboTest = 0,
                    placebo.period = NULL,
                    carryoverTest = 0,
                    carryover.period = NULL,
                    calendar.enp.seq = NULL,
                    time.on.seq = NULL,
                    time.off.seq = NULL,
                    time.on.carry.seq = NULL,
                    time.on.balance.seq = NULL,
                    time.on.seq.W = NULL,
                    time.off.seq.W = NULL,
                    group.level = NULL,
                    group = NULL,
                    time.on.seq.group = NULL,
                    time.off.seq.group = NULL,
                    W.in.fit = TRUE,
                    fit.init = NULL) { ## warm-start matrix for inter_fe_mc (v2.4.2+)
    ## -------------------------------##
    ## Parsing data
    ## -------------------------------##
    carryover.pos <- placebo.pos <- na.pos <- NULL

    ## unit id and time
    TT <- dim(Y)[1]
    N <- dim(Y)[2]
    if (is.null(X) == FALSE) {
        p <- dim(X)[3]
    } else {
        p <- 0
    }

    ## replicate data
    YY <- Y
    YY[which(II == 0)] <- 0 ## reset to 0
    t.on <- c(T.on)

    ## initial fit using fastplm
    data.ini <- matrix(NA, (TT * N), (2 + 1 + p))
    data.ini[, 2] <- rep(1:N, each = TT) ## unit fe
    data.ini[, 3] <- rep(1:TT, N) ## time fe
    data.ini[, 1] <- c(Y) ## outcome
    if (p > 0) { ## covar
        for (i in 1:p) {
            data.ini[, (3 + i)] <- c(X[, , i])
        }
    }

    if (is.null(W) || !W.in.fit) {
        ## When W.in.fit = FALSE (W.agg supplied alone), the outcome model
        ## is fit unweighted; W is still used at the aggregation step below.
        W.use <- as.matrix(0)
    } else {
        W.use <- W
        W.use[which(II == 0)] <- 0
    }

    ## observed Y0 indicator:
    oci <- which(c(II) == 1)
    if (!is.null(W)) {
        initialOut <- initialFit(data = data.ini, force = force, w = c(W), oci = oci)
    } else {
        initialOut <- initialFit(data = data.ini, force = force, w = NULL, oci = oci)
    }
    Y0 <- initialOut$Y0
    beta0 <- initialOut$beta0
    if (p > 0 && sum(is.na(beta0)) > 0) {
        beta0[which(is.na(beta0))] <- 0
    }

    ## -------------------------------##
    ## ----------- Main Algorithm ----------- ##
    ## -------------------------------##

    lambda.norm <- eigen.all <- NULL
    if (boot == FALSE) {
        Y.lambda <- YY - Y0
        Y.lambda[which(II == 0)] <- 0
        eigen.all <- svd(Y.lambda / (TT * N))$d
        lambda.norm <- lambda.cv / max(eigen.all)
    }

    validX <- 1 ## no multi-colinearity
    ## matrix completion
    est.best <- inter_fe_mc(YY, Y0, X, II, W.use, beta0, hasF, lambda.cv, force, tol, max.iteration,
                            fit_init = fit.init)
    validX <- est.best$validX
    validF <- est.best$validF
    est.fect <- NULL
    if (boot == FALSE) {
        est.fect <- inter_fe_ub(YY, Y0, X, II, W.use, beta0, 0, force = force, tol, max.iteration)
    }

    ## ------------------------------##
    ## ----------- Summarize -------------- ##
    ## ------------------------------##

    ## -------------------------------##
    ##   ATT and Counterfactuals     ##
    ## -------------------------------##

    ## we first adjustment for normalization
    if (!is.null(norm.para)) {
        Y <- Y * norm.para[1]

        ## output of estimates
        est.best$mu <- est.best$mu * norm.para[1]
        if (force %in% c(1, 3)) {
            est.best$alpha <- est.best$alpha * norm.para[1]
        }
        if (force %in% c(2, 3)) {
            est.best$xi <- est.best$xi * norm.para[1]
        }
        # if (p>0) {
        #    est.best$beta <- est.best$beta * norm.para[1]
        # }
        est.best$residuals <- est.best$residuals * norm.para[1]
        est.best$fit <- est.best$fit * norm.para[1]
        if (boot == FALSE) {
            est.fect$fit <- est.fect$fit * norm.para[1]
            est.fect$sigma2 <- est.fect$sigma2 * (norm.para[1]^2)
        }
    }

    ## 0. revelant parameters
    if (p > 0) {
        na.pos <- is.nan(est.best$beta)
        beta <- est.best$beta
        if (sum(na.pos) > 0) {
            beta[na.pos] <- NA
        }
    } else {
        beta <- NA
    }

    ## 1. estimated att and counterfactuals
    eff <- Y - est.best$fit
    complete.index <- which(!is.na(eff))
    missing.index <- which(is.na(eff))
    if (length(missing.index) > 0) {
        I[missing.index] <- 0
        II[missing.index] <- 0
    }
    if (0 %in% I) {
        eff[which(I == 0)] <- NA
    }
    complete.index <- which(!is.na(eff))
    denom <- sum(D[complete.index])
    att.avg <- if (denom > 0) {
        sum(eff[complete.index] * D[complete.index]) / denom
    } else {
        NA
    }

    att.avg.balance <- NA
    if (!is.null(balance.period)) {
        complete.index2 <- which(!is.na(T.on.balance))
        denom.balance <- sum(D[complete.index2])
        att.avg.balance <- if (denom.balance > 0) {
            sum(eff[complete.index2] * D[complete.index2]) / denom.balance
        } else {
            NA
        }
    }

    # weighted effect
    att.avg.W <- NA
    if (!is.null(W)) {
        denom.W <- sum(D[complete.index] * W[complete.index])
        att.avg.W <- if (denom.W > 0) {
            sum(eff[complete.index] * D[complete.index] * W[complete.index]) / denom.W
        } else {
            NA
        }
    }

    ## att.avg.unit
    tr.pos <- which(apply(D, 2, sum) > 0)
    att.unit <- sapply(1:length(tr.pos), function(vec) {
        d <- sum(D[, tr.pos[vec]])
        if (d > 0) {
            return(sum(eff[, tr.pos[vec]] * D[, tr.pos[vec]]) / d)
        } else {
            return(NA)
        }
    })
    att.avg.unit <- mean(att.unit, na.rm = TRUE)


    equiv.att.avg <- eff.equiv <- NULL
    if (boot == FALSE) {
        eff.equiv <- Y - est.fect$fit
        if (0 %in% I) {
            eff.equiv[which(I == 0)] <- NA
        }
        complete.index <- which(!is.na(eff.equiv))
        denom.equiv <- sum(D[complete.index])
        equiv.att.avg <- if (denom.equiv > 0) {
            sum(eff.equiv[complete.index] * D[complete.index]) / denom.equiv
        } else {
            NA
        }
    }

    ## 2. rmse for treated units' observations under control
    tr <- which(apply(D, 2, sum) > 0)
    co <- which(apply(D, 2, sum) == 0)
    tr.co <- which((as.matrix(1 - D[, tr]) * as.matrix(II[, tr])) == 1)
    eff.tr <- as.matrix(eff[, tr])
    v.eff.tr <- eff.tr[tr.co]
    rmse <- sqrt(mean(v.eff.tr^2))

    ## 3. unbalanced output
    Y.ct.full <- est.best$fit
    if (0 %in% I) {
        eff[which(I == 0)] <- NA
        est.best$fit[which(I == 0)] <- NA
    }
    res.full <- est.best$residuals
    est.best$residuals[which(II == 0)] <- NA

    est.best$sigma2 <- mean(c(est.best$residuals[which(II == 1)])^2) ## mean squared error of residuals


    ## 4. dynamic effects
    t.on <- c(T.on)
    eff.v <- c(eff) ## a vector
    rm.pos1 <- which(is.na(eff.v))
    rm.pos2 <- which(is.na(t.on))
    eff.v.use1 <- eff.v
    t.on.use <- t.on
    n.on.use <- rep(1:N, each = TT)

    eff.equiv.v <- NULL
    if (boot == FALSE) {
        eff.equiv.v <- c(eff.equiv)
    }

    if (NA %in% eff.v | NA %in% t.on) {
        eff.v.use1 <- eff.v[-c(rm.pos1, rm.pos2)]
        t.on.use <- t.on[-c(rm.pos1, rm.pos2)]
        n.on.use <- n.on.use[-c(rm.pos1, rm.pos2)]
        if (boot == FALSE) {
            eff.equiv.v <- eff.equiv.v[-c(rm.pos1, rm.pos2)]
        }
    }

    pre.pos <- which(t.on.use <= 0)
    eff.pre <- cbind(eff.v.use1[pre.pos], t.on.use[pre.pos], n.on.use[pre.pos])
    colnames(eff.pre) <- c("eff", "period", "unit")

    pre.sd <- eff.pre.equiv <- NULL
    if (boot == FALSE) {
        eff.pre.equiv <- cbind(eff.equiv.v[pre.pos], t.on.use[pre.pos], n.on.use[pre.pos])
        colnames(eff.pre.equiv) <- c("eff.equiv", "period", "unit")

        pre.sd <- tapply(eff.pre.equiv[, 1], eff.pre.equiv[, 2], sd)
        pre.sd <- cbind(pre.sd, sort(unique(eff.pre.equiv[, 2])), table(eff.pre.equiv[, 2]))
        colnames(pre.sd) <- c("sd", "period", "count")
    }


    time.on <- sort(unique(t.on.use))
    att.on <- as.numeric(tapply(eff.v.use1, t.on.use, mean)) ## NA already removed
    count.on <- as.numeric(table(t.on.use))

    if (!is.null(time.on.seq)) {
        count.on.med <- att.on.med <- rep(NA, length(time.on.seq))
        att.on.med[which(time.on.seq %in% time.on)] <- att.on
        count.on.med[which(time.on.seq %in% time.on)] <- count.on
        att.on <- att.on.med
        count.on <- count.on.med
        time.on <- time.on.seq
    }

    ## weighted treatment effect

    if (!is.null(W)) {
        W.v <- c(W)
        rm.pos.W <- which(is.na(W))
        if (NA %in% eff.v | NA %in% t.on | NA %in% W.v) {
            eff.v.use.W <- eff.v[-c(rm.pos1, rm.pos2, rm.pos.W)]
            W.v.use <- W.v[-c(rm.pos1, rm.pos2, rm.pos.W)]
            t.on.use.W <- t.on[-c(rm.pos1, rm.pos2, rm.pos.W)]
            n.on.use.W <- n.on.use[-c(rm.pos1, rm.pos2, rm.pos.W)]
        } else {
            eff.v.use.W <- eff.v.use1
            t.on.use.W <- t.on.use
            n.on.use.W <- n.on.use
            W.v.use <- W.v
        }
        time.on.W <- sort(unique(t.on.use.W))
        att.on.sum.W <- as.numeric(tapply(eff.v.use.W * W.v.use, t.on.use.W, sum)) ## NA already removed
        W.on.sum <- as.numeric(tapply(W.v.use, t.on.use.W, sum))
        att.on.W <- att.on.sum.W / W.on.sum
        count.on.W <- as.numeric(table(t.on.use.W))

        if (!is.null(time.on.seq.W)) {
            att.on.sum.med.W <- W.on.sum.med <- count.on.med.W <- att.on.med.W <- rep(NA, length(time.on.seq.W))
            att.on.sum.med.W[which(time.on.seq.W %in% time.on.W)] <- att.on.sum.W
            att.on.med.W[which(time.on.seq.W %in% time.on.W)] <- att.on.W
            count.on.med.W[which(time.on.seq.W %in% time.on.W)] <- count.on.W
            W.on.sum.med[which(time.on.seq.W %in% time.on.W)] <- W.on.sum
            att.on.sum.W <- att.on.sum.med.W
            att.on.W <- att.on.med.W
            count.on.W <- count.on.med.W
            time.on.W <- time.on.seq.W
            W.on.sum <- W.on.sum.med
        }
    } else {
        att.on.sum.med.W <- att.on.sum.W <- count.on.med.W <- att.on.med.W <- W.on.sum.med <- att.on.W <- count.on.W <- time.on.W <- W.on.sum <- NULL
    }

    ## 4.1 carryover effect
    carry.att <- NULL
    if (!is.null(T.on.carry)) {
        t.on.carry <- c(T.on.carry)
        rm.pos4 <- which(is.na(t.on.carry))
        t.on.carry.use <- t.on.carry

        if (NA %in% eff.v | NA %in% t.on.carry) {
            eff.v.use3 <- eff.v[-c(rm.pos1, rm.pos4)]
            t.on.carry.use <- t.on.carry[-c(rm.pos1, rm.pos4)]
        }

        carry.time <- sort(unique(t.on.carry.use))
        carry.att <- as.numeric(tapply(eff.v.use3, t.on.carry.use, mean)) ## NA already removed

        if (!is.null(time.on.carry.seq)) {
            carry.att.med <- rep(NA, length(time.on.carry.seq))
            carry.att.med[which(time.on.carry.seq %in% carry.time)] <- carry.att
            carry.att <- carry.att.med
            carry.time <- time.on.carry.seq
        }
    }

    ## 4.2 balance effect
    balance.att <- NULL
    if (!is.null(balance.period)) {
        t.on.balance <- c(T.on.balance)
        rm.pos4 <- which(is.na(t.on.balance))
        t.on.balance.use <- t.on.balance

        if (NA %in% eff.v | NA %in% t.on.balance) {
            eff.v.use3 <- eff.v[-c(rm.pos1, rm.pos4)]
            t.on.balance.use <- t.on.balance[-c(rm.pos1, rm.pos4)]
        }

        balance.time <- sort(unique(t.on.balance.use))
        balance.att <- as.numeric(tapply(eff.v.use3, t.on.balance.use, mean)) ## NA already removed
        balance.count <- as.numeric(table(t.on.balance.use))

        if (!is.null(time.on.balance.seq)) {
            balance.att.med <- rep(NA, length(time.on.balance.seq))
            balance.count.med <- rep(0, length(time.on.balance.seq))
            balance.att.med[which(time.on.balance.seq %in% balance.time)] <- balance.att
            if (length(balance.count) > 0) {
                balance.count.med[which(time.on.balance.seq %in% balance.time)] <- balance.count
            }
            balance.count <- balance.count.med
            balance.att <- balance.att.med
            balance.time <- time.on.balance.seq
        }

        # placebo for balanced samples
        if (!is.null(placebo.period) && placeboTest == 1) {
            if (length(placebo.period) == 1) {
                balance.placebo.pos <- which(balance.time == placebo.period)
                balance.att.placebo <- balance.att[balance.placebo.pos]
            } else {
                balance.placebo.pos <- which(balance.time >= placebo.period[1] & balance.time <= placebo.period[2])
                balance.att.placebo <- sum(balance.att[balance.placebo.pos] * balance.count[balance.placebo.pos]) / sum(balance.count[balance.placebo.pos])
            }
        }
    }

    ## 5. placebo effect, if placeboTest == 1
    if (!is.null(placebo.period) && placeboTest == 1) {
        if (length(placebo.period) == 1) {
            placebo.pos <- which(time.on == placebo.period)
            att.placebo <- att.on[placebo.pos]
        } else {
            placebo.pos <- which(time.on >= placebo.period[1] & time.on <= placebo.period[2])
            att.placebo <- sum(att.on[placebo.pos] * count.on[placebo.pos]) / sum(count.on[placebo.pos])
        }

        if (!is.null(W)) {
            if (length(placebo.period) == 1) {
                placebo.pos.W <- which(time.on.W == placebo.period)
                att.placebo.W <- att.on.W[placebo.pos.W]
            } else {
                placebo.pos.W <- which(time.on.W >= placebo.period[1] & time.on.W <= placebo.period[2])
                att.placebo.W <- sum(att.on.sum.W[placebo.pos.W]) / sum(W.on.sum[placebo.pos.W])
            }
        }
    }

    eff.off.equiv <- off.sd <- eff.off <- NULL

    ## 6. switch-off effects
    if (hasRevs == 1) {
        t.off <- c(T.off)
        rm.pos3 <- which(is.na(t.off))
        eff.v.use2 <- eff.v
        t.off.use <- t.off
        n.off.use <- rep(1:N, each = TT)

        if (NA %in% eff.v | NA %in% t.off) {
            eff.v.use2 <- eff.v[-c(rm.pos1, rm.pos3)]
            t.off.use <- t.off[-c(rm.pos1, rm.pos3)]
            n.off.use <- n.off.use[-c(rm.pos1, rm.pos3)]
        }

        off.pos <- which(t.off.use > 0)
        eff.off <- cbind(eff.v.use2[off.pos], t.off.use[off.pos], n.off.use[off.pos])
        colnames(eff.off) <- c("eff", "period", "unit")

        if (boot == FALSE) {
            eff.off.equiv <- cbind(eff.equiv.v[off.pos], t.off.use[off.pos], n.off.use[off.pos])
            colnames(eff.off.equiv) <- c("off.equiv", "period", "unit")

            off.sd <- tapply(eff.off.equiv[, 1], eff.off.equiv[, 2], sd)
            off.sd <- cbind(off.sd, sort(unique(eff.off.equiv[, 2])), table(eff.off.equiv[, 2]))
            colnames(off.sd) <- c("sd", "period", "count")
        }

        time.off <- sort(unique(t.off.use))
        att.off <- as.numeric(tapply(eff.v.use2, t.off.use, mean)) ## NA already removed
        count.off <- as.numeric(table(t.off.use))

        if (!is.null(time.off.seq)) {
            count.off.med <- att.off.med <- rep(NA, length(time.off.seq))
            att.off.med[which(time.off.seq %in% time.off)] <- att.off
            count.off.med[which(time.off.seq %in% time.off)] <- count.off
            att.off <- att.off.med
            count.off <- count.off.med
            time.off <- time.off.seq
        }

        if (!is.null(W)) {
            if (NA %in% eff.v | NA %in% t.off | NA %in% W.v) {
                eff.v.use2.W <- eff.v[-c(rm.pos1, rm.pos3, rm.pos.W)]
                W.v.use2 <- W.v[-c(rm.pos1, rm.pos3, rm.pos.W)]
                t.off.use.W <- t.off[-c(rm.pos1, rm.pos3, rm.pos.W)]
            } else {
                eff.v.use2.W <- eff.v.use2
                t.off.use.W <- t.off.use
                W.v.use2 <- W.v
            }

            time.off.W <- sort(unique(t.off.use.W))
            att.off.sum.W <- as.numeric(tapply(eff.v.use2.W * W.v.use2, t.off.use.W, sum))
            W.off.sum <- as.numeric(tapply(W.v.use2, t.off.use.W, sum))
            att.off.W <- att.off.sum.W / W.off.sum ## NA already removed
            count.off.W <- as.numeric(table(t.off.use.W))

            if (!is.null(time.off.seq.W)) {
                att.off.sum.med.W <- W.off.sum.med <- count.off.med.W <- att.off.med.W <- rep(NA, length(time.off.seq.W))
                att.off.sum.med.W[which(time.off.seq.W %in% time.off.W)] <- att.off.sum.W
                att.off.med.W[which(time.off.seq.W %in% time.off.W)] <- att.off.W
                count.off.med.W[which(time.off.seq.W %in% time.off.W)] <- count.off.W
                W.off.sum.med[which(time.off.seq.W %in% time.off.W)] <- W.off.sum
                att.off.sum.W <- att.off.sum.med.W
                att.off.W <- att.off.med.W
                count.off.W <- count.off.med.W
                time.off.W <- time.off.seq.W
                W.off.sum <- W.off.sum.med
            }
        } else {
            W.off.sum.med <- W.off.sum <- att.off.sum.W <- att.off.sum.med.W <- count.off.med.W <- att.off.med.W <- count.off.med.W <- att.off.W <- count.off.W <- time.off.W <- NULL
        }
    }

    ## 7. carryover effects
    if (!is.null(carryover.period) && carryoverTest == 1 && hasRevs) {
        if (length(carryover.period) == 1) {
            carryover.pos <- which(time.off == carryover.period)
            att.carryover <- att.off[carryover.pos]
        } else {
            carryover.pos <- which(time.off >= carryover.period[1] & time.off <= carryover.period[2])
            att.carryover <- sum(att.off[carryover.pos] * count.off[carryover.pos]) / sum(count.off[carryover.pos])
        }
        if (!is.null(W)) {
            if (length(carryover.period) == 1) {
                carryover.pos.W <- which(time.off.W == carryover.period)
                att.carryover.W <- att.off.W[carryover.pos.W]
            } else {
                carryover.pos.W <- which(time.off.W >= carryover.period[1] & time.off.W <= carryover.period[2])
                att.carryover.W <- sum(att.off.sum.W[carryover.pos.W]) / sum(W.off.sum[carryover.pos.W])
            }
        }
    }

    ## 9. loess HTE by time
    D.missing <- D
    D.missing[which(D == 0)] <- NA
    eff.calendar <- apply(eff * D.missing, 1, mean, na.rm = TRUE)
    N.calendar <- apply(!is.na(eff * D.missing), 1, sum)
    T.calendar <- c(1:TT)
    if (sum(!is.na(eff.calendar)) > 1) {
        # loess fit
        if (!is.null(calendar.enp.seq)) {
            if (length(calendar.enp.seq) == 1 & is.na(calendar.enp.seq)) {
                calendar.enp.seq <- NULL
            }
        }
        if (is.null(calendar.enp.seq)) {
            loess.fit <- suppressWarnings(try(loess(eff.calendar ~ T.calendar, weights = N.calendar), silent = TRUE))
        } else {
            loess.fit <- suppressWarnings(try(loess(eff.calendar ~ T.calendar, weights = N.calendar, enp.target = calendar.enp.seq), silent = TRUE))
        }
        if ("try-error" %in% class(loess.fit)) {
            eff.calendar.fit <- eff.calendar
            calendar.enp <- NULL
        } else {
            eff.calendar.fit <- eff.calendar
            eff.calendar.fit[which(!is.na(eff.calendar))] <- loess.fit$fit
            calendar.enp <- loess.fit$enp
        }
    } else {
        eff.calendar.fit <- eff.calendar
        calendar.enp <- NULL
    }

    ## 8. cohort effects
    if (!is.null(group)) {
        cohort <- cbind(c(group), c(D), c(eff.v))
        rm.pos <- unique(c(rm.pos1, which(cohort[, 2] == 0)))
        cohort <- cohort[-rm.pos, ]

        g.level <- sort(unique(cohort[, 1]))
        raw.group.att <- as.numeric(tapply(cohort[, 3], cohort[, 1], mean))

        group.att <- rep(NA, length(group.level))
        group.att[which(group.level %in% g.level)] <- raw.group.att

        # by-group dynamic effects
        group.level.name <- names(group.level)

        group.output <- list()
        for (i in c(1:length(group.level))) {
            sub.group <- group.level[i]
            sub.group.name <- group.level.name[i]

            ## by-group dynamic effects
            t.on.sub <- c(T.on[which(group == sub.group)])
            eff.v.sub <- c(eff[which(group == sub.group)]) ## a vector
            rm.pos1.sub <- which(is.na(eff.v.sub))
            rm.pos2.sub <- which(is.na(t.on.sub))
            eff.v.use1.sub <- eff.v.sub
            t.on.use.sub <- t.on.sub
            if (NA %in% eff.v.sub | NA %in% t.on.sub) {
                eff.v.use1.sub <- eff.v.sub[-c(rm.pos1.sub, rm.pos2.sub)]
                t.on.use.sub <- t.on.sub[-c(rm.pos1.sub, rm.pos2.sub)]
            }
            if (length(t.on.use.sub) > 0) {
                time.on.sub <- sort(unique(t.on.use.sub))
                att.on.sub <- as.numeric(tapply(
                    eff.v.use1.sub,
                    t.on.use.sub,
                    mean
                )) ## NA already removed
                count.on.sub <- as.numeric(table(t.on.use.sub))
            } else {
                time.on.sub <- att.on.sub <- count.on.sub <- NULL
            }

            if (!is.null(time.on.seq.group)) {
                count.on.med.sub <- att.on.med.sub <- rep(NA, length(time.on.seq.group[[sub.group.name]]))
                time.on.seq.sub <- time.on.seq.group[[sub.group.name]]
                att.on.med.sub[which(time.on.seq.sub %in% time.on.sub)] <- att.on.sub
                count.on.med.sub[which(time.on.seq.sub %in% time.on.sub)] <- count.on.sub
                att.on.sub <- att.on.med.sub
                count.on.sub <- count.on.med.sub
                time.on.sub <- time.on.seq.sub
            }
            suboutput <- list(
                att.on = att.on.sub,
                time.on = time.on.sub,
                count.on = count.on.sub
            )

            ## placebo effect, if placeboTest == 1
            if (!is.null(placebo.period) && placeboTest == 1) {
                if (length(placebo.period) == 1) {
                    placebo.pos.sub <- which(time.on.sub == placebo.period)
                    if (length(placebo.pos.sub) > 0) {
                        att.placebo.sub <- att.on.sub[placebo.pos.sub]
                    } else {
                        att.placebo.sub <- NULL
                    }
                } else {
                    placebo.pos.sub <- which(time.on.sub >= placebo.period[1] & time.on.sub <= placebo.period[2])
                    if (length(placebo.pos.sub) > 0) {
                        att.placebo.sub <- sum(att.on.sub[placebo.pos.sub] * count.on.sub[placebo.pos.sub]) / sum(count.on.sub[placebo.pos.sub])
                    } else {
                        att.placebo.sub <- NULL
                    }
                }
                suboutput <- c(suboutput, list(att.placebo = att.placebo.sub))
            }

            ## T.off
            if (hasRevs == 1) {
                t.off.sub <- c(T.off[which(group == sub.group)])
                rm.pos3.sub <- which(is.na(t.off.sub))
                eff.v.use2.sub <- eff.v.sub
                t.off.use.sub <- t.off.sub
                if (NA %in% eff.v.sub | NA %in% t.off.sub) {
                    eff.v.use2.sub <- eff.v.sub[-c(rm.pos1.sub, rm.pos3.sub)]
                    t.off.use.sub <- t.off.sub[-c(rm.pos1.sub, rm.pos3.sub)]
                }
                if (length(t.off.use.sub) > 0) {
                    time.off.sub <- sort(unique(t.off.use.sub))
                    att.off.sub <- as.numeric(tapply(eff.v.use2.sub, t.off.use.sub, mean)) ## NA already removed
                    count.off.sub <- as.numeric(table(t.off.use.sub))
                } else {
                    time.off.sub <- att.off.sub <- count.off.sub <- NULL
                }

                if (!is.null(time.off.seq.group)) {
                    count.off.med.sub <- att.off.med.sub <- rep(NA, length(time.off.seq.group[[sub.group.name]]))
                    time.off.seq.sub <- time.off.seq.group[[sub.group.name]]
                    att.off.med.sub[which(time.off.seq.sub %in% time.off.sub)] <- att.off.sub
                    count.off.med.sub[which(time.off.seq.sub %in% time.off.sub)] <- count.off.sub
                    att.off.sub <- att.off.med.sub
                    count.off.sub <- count.off.med.sub
                    time.off.sub <- time.off.seq.sub
                }
                suboutput <- c(suboutput, list(
                    att.off = att.off.sub,
                    count.off = count.off.sub,
                    time.off = time.off.sub
                ))

                if (!is.null(carryover.period) && carryoverTest == 1) {
                    if (length(carryover.period) == 1) {
                        carryover.pos.sub <- which(time.off.sub == carryover.period)
                        if (length(carryover.pos.sub) > 0) {
                            att.carryover.sub <- att.off.sub[carryover.pos.sub]
                        } else {
                            att.carryover.sub <- NULL
                        }
                    } else {
                        carryover.pos.sub <- which(time.off.sub >= carryover.period[1] & time.off.sub <= carryover.period[2])
                        if (length(carryover.pos.sub) > 0) {
                            att.carryover.sub <- sum(att.off.sub[carryover.pos.sub] * count.off.sub[carryover.pos.sub]) / sum(count.off.sub[carryover.pos.sub])
                        } else {
                            att.carryover.sub <- NULL
                        }
                    }
                    suboutput <- c(suboutput, list(att.carryover = att.carryover.sub))
                }
            }
            group.output[[sub.group.name]] <- suboutput
        }
    }

    ## -------------------------------##
    ## Storage
    ## -------------------------------##
    out <- list(
        ## main results
        method = "mc",
        T.on = T.on,
        Y.ct = est.best$fit,
        Y.ct.full = Y.ct.full,
        eff = eff,
        I = I,
        II = II,
        D = D,
        Y = Y,
        X = X,
        att.avg = att.avg,
        att.avg.unit = att.avg.unit,
        ## supporting
        force = force,
        T = TT,
        N = N,
        tr = tr,
        co = co,
        p = p,
        lambda.cv = lambda.cv,
        lambda.norm = lambda.norm,
        eigen.all = eigen.all,
        beta = beta,
        est = est.best,
        mu = est.best$mu,
        validX = validX,
        validF = validF,
        niter = est.best$niter,
        time = time.on,
        att = att.on,
        count = count.on,
        eff.calendar = eff.calendar,
        N.calendar = N.calendar,
        eff.calendar.fit = eff.calendar.fit,
        calendar.enp = calendar.enp,
        eff.pre = eff.pre,
        eff.pre.equiv = eff.pre.equiv,
        pre.sd = pre.sd,
        rmse = rmse,
        res = est.best$residuals,
        sigma2 = est.best$sigma2,
        sigma2.fect = est.fect$sigma2
    )

    if (hasRevs == 1) {
        out <- c(out, list(
            time.off = time.off,
            att.off = att.off,
            count.off = count.off,
            eff.off = eff.off,
            eff.off.equiv = eff.off.equiv,
            off.sd = off.sd
        ))
    }

    if (!is.null(T.on.carry)) {
        out <- c(out, list(carry.att = carry.att, carry.time = carry.time))
    }

    if (!is.null(balance.period)) {
        out <- c(out, list(
            balance.att = balance.att, balance.time = balance.time,
            balance.count = balance.count, balance.avg.att = att.avg.balance
        ))
        if (!is.null(placebo.period) && placeboTest == 1) {
            out <- c(out, list(balance.att.placebo = balance.att.placebo))
        }
    }

    if (!is.null(W)) {
        out <- c(out, list(
            W = W,
            att.avg.W = att.avg.W,
            att.on.sum.W = att.on.sum.W,
            att.on.W = att.on.W,
            count.on.W = count.on.W,
            time.on.W = time.on.W,
            W.on.sum = W.on.sum
        ))
        if (hasRevs == 1) {
            out <- c(out, list(
                att.off.sum.W = att.off.sum.W,
                att.off.W = att.off.W,
                count.off.W = count.off.W,
                time.off.W = time.off.W,
                W.off.sum = W.off.sum
            ))
        }
        if (!is.null(placebo.period) && placeboTest == 1) {
            out <- c(out, list(att.placebo.W = att.placebo.W))
        }

        if (!is.null(carryover.period) && carryoverTest == 1) {
            out <- c(out, list(att.carryover.W = att.carryover.W))
        }
    }

    # if (boot == FALSE) {
    #    out <- c(out, list(equiv.att.avg = equiv.att.avg))
    # }

    if (force == 1) {
        out <- c(out, list(alpha = est.best$alpha))
    } else if (force == 2) {
        out <- c(out, list(xi = est.best$xi))
    } else if (force == 3) {
        out <- c(out, list(alpha = est.best$alpha, xi = est.best$xi))
    }

    if (!is.null(placebo.period) && placeboTest == 1) {
        out <- c(out, list(att.placebo = att.placebo))
    }

    if (!is.null(carryover.period) && carryoverTest == 1) {
        out <- c(out, list(att.carryover = att.carryover))
    }

    if (!is.null(group)) {
        out <- c(out, list(
            group.att = group.att,
            group.output = group.output
        ))
    }
    return(out)
} ## mc functions ends

## Ensure datasets are available in tests across installed/build contexts
if (!exists("simdata", inherits = TRUE)) {
  suppressWarnings(try(utils::data("simdata", package = "fect"), silent = TRUE))
}
if (!exists("simdata", inherits = TRUE)) {
  f <- system.file("data", "fect.RData", package = "fect")
  if (nzchar(f) && file.exists(f)) {
    try(load(f, envir = environment()), silent = TRUE)
  }
}
if (!exists("simdata", inherits = TRUE)) {
  f <- file.path("data", "fect.RData")
  if (file.exists(f)) {
    try(load(f, envir = environment()), silent = TRUE)
  }
}
if (!exists("simgsynth", inherits = TRUE)) {
  suppressWarnings(try(utils::data("simgsynth", package = "fect"), silent = TRUE))
}

###################################################################
## Shared scoring function for cross-validation and MSPE diagnostics
###################################################################

.score_residuals <- function(resid,
                             obs_weights = NULL,
                             time_index = NULL,
                             count_weights = NULL,
                             norm.para = NULL) {

    ## ---- input validation ---- ##
    n <- length(resid)
    if (n == 0L) stop("No residuals to score.")
    if (!is.null(obs_weights) && length(obs_weights) != n) {
        stop("obs_weights must have the same length as resid.")
    }
    if (!is.null(time_index) && length(time_index) != n) {
        stop("time_index must have the same length as resid.")
    }

    ## ---- Step 1: squared residuals and period weights ---- ##
    e2 <- resid^2

    if (!is.null(count_weights) && !is.null(time_index)) {
        w_period <- count_weights[time_index]
        names(w_period) <- NULL
    } else {
        w_period <- rep(1, n)
    }

    ## ---- Step 2: accumulations ---- ##
    if (is.null(obs_weights)) {
        ## unweighted path
        SSE    <- sum(e2)
        WSSE   <- sum(w_period * e2)
        GSSE   <- sum(log(e2))
        ll     <- w_period * e2
        ll_pos <- ll[which(ll > 0)]
        WGSSE  <- sum(log(ll_pos))
        ll_len <- length(ll_pos)
        mad_list <- e2
        denom  <- n
    } else {
        ## weighted path
        SSE    <- sum(obs_weights * e2)
        WSSE   <- sum(obs_weights * w_period * e2)
        GSSE   <- sum(obs_weights * log(e2))
        ll     <- obs_weights * w_period * e2
        ll_pos <- ll[which(ll > 0)]
        WGSSE  <- sum(log(ll_pos))
        ll_len <- length(ll_pos)
        mad_list <- obs_weights * e2
        denom  <- sum(obs_weights)
    }

    ## ---- Step 3: criteria ---- ##
    MSPE   <- SSE / denom
    WMSPE  <- WSSE / denom
    GMSPE  <- exp(GSSE / denom)
    WGMSPE <- if (ll_len > 0) exp(WGSSE / ll_len) else Inf
    MAD    <- median(abs(mad_list - median(mad_list)))

    ## ---- Step 4: Moment and GMoment ---- ##
    if (!is.null(time_index)) {
        resid_mean   <- tapply(resid, time_index, mean)
        resid_mean   <- abs(resid_mean)

        gm_mean <- function(x) exp(sum(log(x)) / length(x))
        resid_g_mean <- tapply(abs(resid), time_index, gm_mean)

        if (!is.null(count_weights)) {
            w_moment   <- count_weights[names(resid_mean)]
            names(w_moment) <- NULL
            w_gmoment  <- count_weights[names(resid_g_mean)]
            names(w_gmoment) <- NULL
            Moment     <- sum(w_moment * resid_mean) / sum(w_moment)
            GMoment    <- sum(w_gmoment * resid_g_mean) / sum(w_gmoment)
        } else {
            Moment  <- mean(resid_mean)
            GMoment <- mean(resid_g_mean)
        }
    } else {
        Moment  <- NA_real_
        GMoment <- NA_real_
    }

    ## ---- Step 5: normalization ---- ##
    if (!is.null(norm.para)) {
        scale <- norm.para[1]^2
        MSPE    <- MSPE * scale
        WMSPE   <- WMSPE * scale
        GMSPE   <- GMSPE * scale
        WGMSPE  <- WGMSPE * scale
        MAD     <- MAD * scale
        Moment  <- Moment * scale
        GMoment <- GMoment * scale
    }

    ## ---- Step 6: convenience scores and return ---- ##
    RMSE <- sqrt(MSPE)
    Bias <- mean(resid)
    c(MSPE = MSPE, WMSPE = WMSPE, GMSPE = GMSPE, WGMSPE = WGMSPE,
      MAD = MAD, Moment = Moment, GMoment = GMoment,
      RMSE = RMSE, Bias = Bias)
}

## ---------------------------------------------------------------
## Tests for fect_mspe: criterion (S3), cv.method (S4), weights (S6),
## norm.para (S7), return structure (S8), input validation (S9), and
## fect_mspe simplification (Section D, NEW for Phase 2).
##
## Originally part of test-score-unify.R; split out 2026-05-03 for
## progress visibility. Shared fixtures live in helper-score-unify.R.
## ---------------------------------------------------------------

## =================================================================
## S3: fect_mspe criterion support (updated for Phase 2)
## =================================================================

test_that("S3.1: Default criterion='mspe' matches old RMSE", {

  skip_on_cran()
  res_new <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 123, criterion = "mspe")
  ))

  # RMSE = sqrt(MSPE) invariant (P1)
  if ("MSPE" %in% names(res_new$summary)) {
    expect_equal(res_new$summary$RMSE, sqrt(res_new$summary$MSPE),
                 tolerance = 1e-10)
  }
})

test_that("S3.2: All 7 criteria produce finite, positive scores", {

  skip_on_cran()
  for (crit in c("mspe", "wmspe", "gmspe", "wgmspe", "mad",
                 "moment", "gmoment")) {
    res <- suppressWarnings(suppressMessages(
      fect_mspe(out_base, seed = 123, criterion = crit)
    ))
    crit_upper <- toupper(crit)
    if (crit_upper %in% names(res$summary)) {
      expect_true(all(is.finite(res$summary[[crit_upper]])),
                  info = paste("Criterion", crit, "should be finite"))
      expect_true(all(res$summary[[crit_upper]] > 0),
                  info = paste("Criterion", crit, "should be positive"))
    }
  }
})


## =================================================================
## S4: fect_mspe with cv.method (Phase 2 — replaces mask.method)
## =================================================================

test_that("S4.1: cv.method='treated_units' masking runs without error", {

  skip_on_cran()
  res <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, cv.method = "treated_units",
              k = 3, cv.prop = 0.1,
              cv.nobs = 3, cv.donut = 1, min.T0 = 3,
              criterion = "mspe")
  ))
  expect_true("summary" %in% names(res))
  if ("MSPE" %in% names(res$summary)) {
    expect_true(all(is.finite(res$summary$MSPE)))
    expect_true(all(res$summary$MSPE > 0))
  }
  expect_true(res$summary$RMSE > 0)
})

test_that("S4.2: cv.method='all_units' masking runs without error", {

  skip_on_cran()
  res_cv <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, cv.method = "all_units",
              k = 3, criterion = "mspe")
  ))
  expect_true("summary" %in% names(res_cv))
  expect_true(res_cv$summary$RMSE > 0)
})

test_that("S4.3: cv.method='treated_units' with k=1 (single fold)", {

  skip_on_cran()
  res <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, cv.method = "treated_units",
              k = 1, criterion = "mspe")
  ))
  expect_true(res$summary$RMSE > 0)
})


## =================================================================
## S6: fect_mspe with observation weights (W) — updated for Phase 2
## =================================================================

test_that("S6.1: W parameter produces different scores than unweighted", {

  skip_on_cran()
  TT <- nrow(out_base$Y.dat)
  NN <- ncol(out_base$Y.dat)
  set.seed(7)
  W_mat <- matrix(runif(TT * NN, 0.5, 1.5), nrow = TT, ncol = NN)
  res_unw <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, criterion = "mspe",
              W = NULL)
  ))
  res_w <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, criterion = "mspe",
              W = W_mat)
  ))
  expect_false(identical(res_unw$summary$RMSE, res_w$summary$RMSE))
})

test_that("S6.2: Uniform W equals unweighted", {

  skip_on_cran()
  TT <- nrow(out_base$Y.dat)
  NN <- ncol(out_base$Y.dat)
  W_uniform <- matrix(1, nrow = TT, ncol = NN)
  res_u <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, criterion = "mspe",
              W = W_uniform)
  ))
  res_n <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, criterion = "mspe",
              W = NULL)
  ))
  if ("MSPE" %in% names(res_u$summary) && "MSPE" %in% names(res_n$summary)) {
    expect_equal(res_u$summary$MSPE, res_n$summary$MSPE, tolerance = 1e-10)
  }
})


## =================================================================
## S7: fect_mspe with norm.para — updated for Phase 2
## =================================================================

test_that("S7.1: norm.para scales scores", {

  skip_on_cran()
  np <- c(2.0, 0.0)
  res_raw <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, criterion = "mspe")
  ))
  res_norm <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, criterion = "mspe",
              norm.para = np)
  ))
  if ("MSPE" %in% names(res_raw$summary) &&
      "MSPE" %in% names(res_norm$summary)) {
    expect_equal(res_norm$summary$MSPE, res_raw$summary$MSPE * 4.0,
                 tolerance = 1e-10)
  }
})


## =================================================================
## S8: Return structure (updated for Phase 2 — hide_mask removed)
## =================================================================

test_that("S8.3: Return structure has summary and records", {

  skip_on_cran()
  res <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42)
  ))
  expect_true("summary" %in% names(res))
  expect_true("records" %in% names(res))
  expect_true("fits" %in% names(res))
  expect_true("RMSE" %in% names(res$summary))
  expect_true("Bias" %in% names(res$summary))
})


## =================================================================
## S9: Input validation (updated for Phase 2)
## =================================================================

test_that("S9.1: Invalid criterion rejected", {

  skip_on_cran()
  invalid_result <- tryCatch({
    suppressWarnings(suppressMessages(
      fect_mspe(out_base, criterion = "invalid", seed = 42)
    ))
    "no_error"
  }, error = function(e) "error_thrown")

  expect_equal(invalid_result, "error_thrown",
    info = paste("fect_mspe(criterion='invalid') should throw an error,",
                 "but it completed without error. Builder must add",
                 "criterion validation."))
})

test_that("S9.2: Invalid cv.method rejected by fect_mspe", {

  skip_on_cran()
  expect_error(
    fect_mspe(out_base, cv.method = "invalid"),
    "cv.method|arg"
  )
})

test_that("S9.3: W wrong dimensions", {

  skip_on_cran()
  W_bad <- matrix(1, nrow = 5, ncol = 5)
  expect_error(
    fect_mspe(out_base, W = W_bad),
    "dimension|W"
  )
})

test_that("S9.5: .score_residuals() empty input", {

  skip_on_cran()
  score_fn <- tryCatch(
    getFromNamespace(".score_residuals", "fect"),
    error = function(e) NULL
  )
  skip_if(is.null(score_fn), ".score_residuals() not yet implemented")

  expect_error(score_fn(numeric(0)), "No residuals")
})


## =================================================================
## Section D: fect_mspe simplification (NEW for Phase 2)
## =================================================================

test_that("MSPE1: Simplified fect_mspe with cv.method='all_units'", {

  skip_on_cran()
  res <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, cv.method = "all_units", criterion = "mspe")
  ))
  expect_true("summary" %in% names(res))
  expect_true("records" %in% names(res))
  expect_true("MSPE" %in% names(res$summary))
  expect_true("RMSE" %in% names(res$summary))
  expect_true("Bias" %in% names(res$summary))
  expect_true(res$summary$RMSE > 0)
  expect_equal(res$summary$RMSE, sqrt(res$summary$MSPE), tolerance = 1e-10)
})

test_that("MSPE2: Simplified fect_mspe with cv.method='treated_units'", {

  skip_on_cran()
  res <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, cv.method = "treated_units",
              criterion = "mspe")
  ))
  expect_true("summary" %in% names(res))
  expect_true("records" %in% names(res))
  expect_true(res$summary$RMSE > 0)
})

test_that("MSPE3: Removed parameters rejected", {

  skip_on_cran()
  expect_error(fect_mspe(out_base, mask.method = "random"))
  expect_error(fect_mspe(out_base, hide_mask = matrix(TRUE, 10, 10)))
  expect_error(fect_mspe(out_base, n_rep = 3))
  expect_error(fect_mspe(out_base, pre.trend = TRUE))
  expect_error(fect_mspe(out_base, actual = out_base$Y.ct.full))
  expect_error(fect_mspe(out_base, control.only = FALSE))
  expect_error(fect_mspe(out_base, hide_n = 20))
})

test_that("MSPE4: Invalid cv.method rejected", {

  skip_on_cran()
  expect_error(
    fect_mspe(out_base, cv.method = "loo"),
    "cv.method|arg"
  )
})

test_that("MSPE5: Multi-model comparison with cv.method", {

  skip_on_cran()
  res <- suppressWarnings(suppressMessages(
    fect_mspe(list(m1 = out_base, m2 = out_base),
              seed = 42, cv.method = "all_units")
  ))
  expect_equal(nrow(res$summary), 2)
  expect_true(all(c("m1", "m2") %in% res$summary$Model))
})

test_that("MSPE6: Seed reproducibility", {

  skip_on_cran()
  r1 <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, cv.method = "all_units")
  ))
  r2 <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, cv.method = "all_units")
  ))
  expect_identical(r1$summary, r2$summary)
})

test_that("MSPE7: fect_mspe with observation weights", {

  skip_on_cran()
  TT <- nrow(out_base$Y.dat)
  NN <- ncol(out_base$Y.dat)
  W_mat <- matrix(1, nrow = TT, ncol = NN)
  res <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, cv.method = "all_units", W = W_mat)
  ))
  if ("MSPE" %in% names(res$summary)) {
    expect_true(all(is.finite(res$summary$MSPE)))
    expect_true(all(res$summary$MSPE > 0))
  }
})

test_that("MSPE8: fect_mspe with norm.para", {

  skip_on_cran()
  res_raw <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, cv.method = "all_units")
  ))
  res_norm <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, cv.method = "all_units",
              norm.para = c(2.0))
  ))
  if ("MSPE" %in% names(res_raw$summary) &&
      "MSPE" %in% names(res_norm$summary)) {
    # With norm.para, MSPE scaled by norm.para[1]^2 = 4.0
    expect_equal(res_norm$summary$MSPE, res_raw$summary$MSPE * 4.0,
                 tolerance = 1e-10)
  }
})

## ---------------------------------------------------------------
## Tests for .score_residuals() unit behavior + property-based
## invariants + edge cases.
##
## Originally Section S1 + Property + Edge of test-score-unify.R;
## split out 2026-05-03 for progress visibility under reporter = "summary".
## Shared fixtures (make_factor_data, ntdata, out_base) live in
## helper-score-unify.R.
## ---------------------------------------------------------------

## =================================================================
## S1: .score_residuals() unit tests with known inputs
## =================================================================

test_that("S1.1: Basic unweighted scoring", {

  skip_on_cran()
  score_fn <- tryCatch(
    getFromNamespace(".score_residuals", "fect"),
    error = function(e) NULL
  )
  skip_if(is.null(score_fn), ".score_residuals() not yet implemented")

  resid <- c(1.0, -2.0, 3.0, -1.0, 2.0)
  result <- score_fn(resid)

  # MSPE = mean(resid^2) = (1+4+9+1+4)/5 = 3.8
  expect_equal(result[["MSPE"]], 3.8, tolerance = 1e-10)

  # GMSPE = exp(mean(log(resid^2)))
  expected_gmspe <- exp(mean(log(c(1, 4, 9, 1, 4))))
  expect_equal(result[["GMSPE"]], expected_gmspe, tolerance = 1e-10)

  # MAD = median(|e2 - median(e2)|) where e2=c(1,4,9,1,4), median=4
  # deviations = c(3,0,5,3,0), MAD = 3
  expect_equal(result[["MAD"]], 3.0, tolerance = 1e-10)

  # RMSE = sqrt(3.8)
  expect_equal(result[["RMSE"]], sqrt(3.8), tolerance = 1e-10)

  # Bias = mean(resid) = 0.6
  expect_equal(result[["Bias"]], 0.6, tolerance = 1e-10)

  # Moment and GMoment = NA (no time_index)
  expect_true(is.na(result[["Moment"]]))
  expect_true(is.na(result[["GMoment"]]))

  # WMSPE = MSPE with uniform weights = 3.8
  expect_equal(result[["WMSPE"]], 3.8, tolerance = 1e-10)

  # WGMSPE = GMSPE with uniform weights
  expect_equal(result[["WGMSPE"]], expected_gmspe, tolerance = 1e-10)
})

test_that("S1.2: With observation weights", {

  skip_on_cran()
  score_fn <- tryCatch(
    getFromNamespace(".score_residuals", "fect"),
    error = function(e) NULL
  )
  skip_if(is.null(score_fn), ".score_residuals() not yet implemented")

  resid <- c(1.0, -2.0, 3.0)
  obs_weights <- c(0.5, 1.0, 0.5)
  result <- score_fn(resid, obs_weights = obs_weights)

  # MSPE = (0.5*1 + 1.0*4 + 0.5*9) / (0.5+1.0+0.5) = 9/2 = 4.5
  expect_equal(result[["MSPE"]], 4.5, tolerance = 1e-10)

  # RMSE = sqrt(4.5)
  expect_equal(result[["RMSE"]], sqrt(4.5), tolerance = 1e-10)

  # Bias = mean(resid) = 2/3
  expect_equal(result[["Bias"]], 2 / 3, tolerance = 1e-10)
})

test_that("S1.3: With time_index and count_weights (Moment/GMoment)", {

  skip_on_cran()
  score_fn <- tryCatch(
    getFromNamespace(".score_residuals", "fect"),
    error = function(e) NULL
  )
  skip_if(is.null(score_fn), ".score_residuals() not yet implemented")

  resid <- c(1.0, -1.0, 2.0, -3.0, 0.5, -0.5)
  time_index <- c("-2", "-2", "-1", "-1", "Control", "Control")
  count_weights <- c("-2" = 1.5, "-1" = 2.0, "Control" = 1.0)

  result <- score_fn(resid, time_index = time_index,
                     count_weights = count_weights)

  # Moment:
  # resid_mean per group: "-2"->0, "-1"->-0.5, "Control"->0
  # abs: c(0, 0.5, 0)
  # Moment = (1.5*0 + 2.0*0.5 + 1.0*0) / (1.5+2.0+1.0) = 1.0/4.5
  expected_moment <- 1.0 / 4.5
  expect_equal(result[["Moment"]], expected_moment, tolerance = 1e-10)

  # GMoment:
  # geometric mean of abs(resid) per group:
  # "-2" -> exp((log(1)+log(1))/2) = 1.0
  # "-1" -> exp((log(2)+log(3))/2) = sqrt(6)
  # "Control" -> exp((log(0.5)+log(0.5))/2) = 0.5
  # GMoment = (1.5*1.0 + 2.0*sqrt(6) + 1.0*0.5) / (1.5+2.0+1.0)
  expected_gmoment <- (1.5 * 1.0 + 2.0 * sqrt(6) + 1.0 * 0.5) / 4.5
  expect_equal(result[["GMoment"]], expected_gmoment, tolerance = 1e-10)
})

test_that("S1.4: With norm.para", {

  skip_on_cran()
  score_fn <- tryCatch(
    getFromNamespace(".score_residuals", "fect"),
    error = function(e) NULL
  )
  skip_if(is.null(score_fn), ".score_residuals() not yet implemented")

  resid <- c(1.0, -2.0, 3.0, -1.0, 2.0)
  result <- score_fn(resid, norm.para = c(2.0))

  # All 7 scores multiplied by 4.0 (norm.para[1]^2)
  expect_equal(result[["MSPE"]], 3.8 * 4.0, tolerance = 1e-10)
  expect_equal(result[["RMSE"]], sqrt(3.8 * 4.0), tolerance = 1e-10)

  # Bias unchanged
  expect_equal(result[["Bias"]], 0.6, tolerance = 1e-10)
})

test_that("S1.5: Edge case - single residual", {

  skip_on_cran()
  score_fn <- tryCatch(
    getFromNamespace(".score_residuals", "fect"),
    error = function(e) NULL
  )
  skip_if(is.null(score_fn), ".score_residuals() not yet implemented")

  result <- score_fn(c(2.5))

  expect_equal(result[["MSPE"]], 6.25, tolerance = 1e-10)
  expect_equal(result[["RMSE"]], 2.5, tolerance = 1e-10)
  expect_equal(result[["Bias"]], 2.5, tolerance = 1e-10)
  expect_equal(result[["MAD"]], 0.0, tolerance = 1e-10)
  expect_equal(result[["GMSPE"]], 6.25, tolerance = 1e-10)
})

test_that("S1.6: Edge case - zero residual", {

  skip_on_cran()
  score_fn <- tryCatch(
    getFromNamespace(".score_residuals", "fect"),
    error = function(e) NULL
  )
  skip_if(is.null(score_fn), ".score_residuals() not yet implemented")

  resid <- c(0.0, 1.0, -1.0)
  result <- score_fn(resid)

  # MSPE = 2/3
  expect_equal(result[["MSPE"]], 2 / 3, tolerance = 1e-10)

  # GMSPE = exp(mean(log(c(0,1,1)))) = exp(-Inf) = 0
  expect_true(result[["GMSPE"]] <= 1e-300 || !is.finite(log(result[["GMSPE"]])))

  # WGMSPE: zero filtered out, only c(1,1) remain. WGMSPE = 1.0
  expect_equal(result[["WGMSPE"]], 1.0, tolerance = 1e-10)
})

test_that("S1.7: Edge case - empty residuals", {

  skip_on_cran()
  score_fn <- tryCatch(
    getFromNamespace(".score_residuals", "fect"),
    error = function(e) NULL
  )
  skip_if(is.null(score_fn), ".score_residuals() not yet implemented")

  expect_error(score_fn(numeric(0)), "No residuals")
})


## =================================================================
## Property-based invariants (updated for Phase 2)
## =================================================================

test_that("P1: RMSE = sqrt(MSPE) invariant", {

  skip_on_cran()
  res <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, criterion = "mspe")
  ))
  if ("MSPE" %in% names(res$summary)) {
    expect_equal(res$summary$RMSE, sqrt(res$summary$MSPE), tolerance = 1e-10)
  }
})

test_that("P2: MSPE >= 0", {

  skip_on_cran()
  res <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42, criterion = "mspe")
  ))
  if ("MSPE" %in% names(res$summary)) {
    expect_true(all(res$summary$MSPE >= 0))
  }
})

test_that("P5: Moment = 0 when all per-group mean residuals are 0", {

  skip_on_cran()
  score_fn <- tryCatch(
    getFromNamespace(".score_residuals", "fect"),
    error = function(e) NULL
  )
  skip_if(is.null(score_fn), ".score_residuals() not yet implemented")

  # Construct residuals where each group mean is exactly 0
  resid <- c(1, -1, 2, -2)
  time_index <- c("A", "A", "B", "B")
  count_weights <- c("A" = 1.0, "B" = 1.0)

  result <- score_fn(resid, time_index = time_index,
                     count_weights = count_weights)
  expect_equal(result[["Moment"]], 0.0, tolerance = 1e-10)
})

test_that("P7: Seed reproducibility", {

  skip_on_cran()
  r1 <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42)
  ))
  r2 <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42)
  ))
  expect_identical(r1$summary, r2$summary)
})


## =================================================================
## Edge cases (updated for Phase 2)
## =================================================================

test_that("E3: Single model in fect_mspe (not a list)", {

  skip_on_cran()
  res <- suppressWarnings(suppressMessages(
    fect_mspe(out_base, seed = 42)
  ))
  expect_equal(nrow(res$summary), 1)
  expect_true(res$summary$RMSE > 0)
})

test_that("E4: Multiple models in fect_mspe", {

  skip_on_cran()
  multi <- suppressWarnings(suppressMessages(
    fect_mspe(list(m1 = out_base, m2 = out_base),
              seed = 42)
  ))
  expect_equal(nrow(multi$summary), 2)
  expect_true(all(c("m1", "m2") %in% multi$summary$Model))
})

library(testthat)
library(fect)

## For non-interactive / captured runs (local devtools::test() and GitHub
## Actions), use a MultiReporter that preserves per-test output.
## LocationReporter prints failures as file:line (readable); JunitReporter
## writes test-results.xml that CI tooling surfaces as PR annotations.
## Skipped on CRAN (NOT_CRAN unset, GITHUB_ACTIONS unset) to keep CRAN's
## default behavior unchanged.
if (!interactive() &&
    (identical(Sys.getenv("NOT_CRAN"), "true") ||
     nzchar(Sys.getenv("GITHUB_ACTIONS")))) {
  options(testthat.default_reporter = testthat::MultiReporter$new(list(
    testthat::LocationReporter$new(),
    testthat::JunitReporter$new(file = "test-results.xml")
  )))
}

## On CRAN (NOT_CRAN unset and not on GitHub Actions) run ONLY the minimal
## smoke tests in test-cran.R to keep the check budget small. Locally and in CI
## (NOT_CRAN=true or GITHUB_ACTIONS set) run the full regression suite.
if (!identical(Sys.getenv("NOT_CRAN"), "true") && !nzchar(Sys.getenv("GITHUB_ACTIONS"))) {
  test_check("fect", filter = "^cran$")
} else {
  test_check("fect")
}

##########
## Print
##########
## a fect object

print.fect <- function(x,
                       switch.on = TRUE,
                       switch.off = FALSE,
                       time.on.lim = NULL,
                       time.off.lim = NULL,
                       ...) {


    if (!is.null(x$effect.est.avg)) {
      cat("\nOverall cumulative effect:\n")
      print(x$effect.est.avg, digits = 4)
      if (!is.null(x$effect.est.att)) {
        cat("\nPeriod-by-period cumulative effect:\n")
        print(x$effect.est.att, digits = 4)
      }
      return()
    }
    cat("Call:\n")
    print(x$call, digits = 4)

    ## Estimator + Fixed effects + Cluster SE lines (D8 in design memo).
    ## Surfaces the actual FE composition so users can verify their model
    ## without re-reading the call args. force is stored as integer 0/1/2/3.
    .fe_label <- function(force_int, index, group.fe) {
        parts <- character(0)
        if (length(index) >= 1 && force_int %in% c(1, 3)) {
            parts <- c(parts, paste0(index[1], " (unit)"))
        }
        if (length(index) >= 2 && force_int %in% c(2, 3)) {
            parts <- c(parts, paste0(index[2], " (time)"))
        }
        if (!is.null(group.fe) && length(group.fe) > 0) {
            parts <- c(parts, paste(group.fe, collapse = " + "))
        }
        if (length(parts) == 0) "none" else paste(parts, collapse = " + ")
    }
    fe.line <- .fe_label(x$force,
                         if (!is.null(x$call$index)) eval(x$call$index) else NULL,
                         x$group.fe)
    cat("\nEstimator:    ", x$method, "\n", sep = "")
    cat("Fixed effects: ", fe.line, "\n", sep = "")
    if (!is.null(x$cl.label)) {
        cat("Cluster SE:   ", x$cl.label, "\n", sep = "")
    }

    if (switch.on == TRUE) {
        if (!is.null(time.on.lim)) {

            if (is.numeric(time.on.lim)==FALSE) {
                stop("Some element in \"time.on.lim\" is not numeric.")
            } else {
                if (length(time.on.lim)!=2) {
                    stop("time.on.lim must be of length 2.")
                }
            }

            seq.on.min <- min(which(x$time >= time.on.lim[1]))
            seq.on.max <- max(which(x$time <= time.on.lim[2]))
            seq.on <- seq.on.min:seq.on.max
        } else {
            seq.on <- 1:length(x$time)
        }
    }

    if (switch.off == TRUE & is.null(x$att.off) == FALSE) {
        if (!is.null(time.off.lim)) {

            if (is.numeric(time.off.lim)==FALSE) {
                stop("Some element in \"time.off.lim\" is not numeric.")
            } else {
                if (length(time.off.lim)!=2) {
                    stop("time.off.lim must be of length 2.")
                }
            }

            seq.off.min <- min(which(x$time.off >= time.off.lim[1]))
            seq.off.max <- max(which(x$time.off <= time.off.lim[2]))
            seq.off <- seq.off.min:seq.off.max
        } else {
            seq.off <- 1:length(x$time.off)
        }
    }

    if (is.null(x$est.avg) == TRUE) { # no uncertainties
        cat("\nATT:\n")
        att.out <- rbind.data.frame(x$att.avg, x$att.avg.unit)
        colnames(att.out) <- c("ATT")
        rownames(att.out) <- c(
            "Tr obs. equally weighted",
            "Tr units equally weighted")
        print(att.out, digits = 4)
        # if (switch.on == TRUE) {
        #     cat("\n   ~ by Period:\n")
        #     print(x$att[seq.on], digits = 4)
        # }
        # if (switch.off == TRUE & is.null(x$att.off) == FALSE) {
        #     cat("\n   ~ Switch-off by Period:\n")
        #     print(x$att.off[seq.off], digits = 4)
        # }
        if (is.null(x$X) == FALSE) {
            cat("\nCovariates:\n")
            print(x$beta, digits = 4)
        }
        cat("\nUncertainty estimates not available.\n")
    } else {
        cat("\nATT:\n")
        att.out <- rbind.data.frame(c(x$est.avg), c(x$est.avg.unit))
        colnames(att.out) <- c("ATT", "S.E.", "CI.lower", "CI.upper", "p.value")
        ## When W is supplied AND enters the aggregation, the obs-level row
        ## reports the sample-weighted aggregate. Otherwise (no W, or W only
        ## entered the outcome-model fit), the row is the unweighted average.
        first.row <- if (isTRUE(x$W.in.agg)) {
            "Tr obs sample-weighted (W)"
        } else {
            "Tr obs equally weighted"
        }
        rownames(att.out) <- c(
            first.row,
            "Tr units equally weighted")
        print(att.out, digits = 4)
        # if (switch.on == TRUE) {
        #     cat("\n   ~ Switch-on by Period:\n")
        #     print(x$est.att[seq.on,], digits = 4)
        # }
        # if (switch.off == TRUE & is.null(x$att.off) == FALSE) {
        #     cat("\n   ~ Switch-on by Period:\n")
        #     print(x$est.att.off[seq.off,], digits = 4)
        # }
        if (is.null(x$X) == FALSE) {
            cat("\nCovariates:\n")
            print(x$est.beta, digits = 4)
        }
    }

    if (!is.null(x$est.placebo)) {
        cat("\nPlacebo effect for pre-treatment periods:\n")
        print(x$est.placebo, digits = 4)
    }
}

## Minimal CRAN smoke tests for fect.
##
## This is the ONLY test file that runs on CRAN: tests/testthat.R filters the
## suite to this file when NOT_CRAN is unset (see that driver). Keep it fast and
## dependency-light -- a single core fit and the print method. The full
## regression suite (every other test-*.R) runs locally and in CI, where
## NOT_CRAN=true.

library(testthat)
data(simdata, package = "fect")

test_that("core fect() fit runs and returns the expected structure", {
  out <- fect::fect(Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
                    method = "fe", force = "two-way", se = FALSE, parallel = FALSE)
  expect_s3_class(out, "fect")
  expect_true(is.matrix(out$eff))
  expect_true(is.numeric(out$att.avg))
  expect_true(length(out$time) >= 1L)
})

test_that("print.fect() runs without error", {
  out <- fect::fect(Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
                    method = "fe", force = "two-way", se = FALSE, parallel = FALSE)
  printed <- capture.output(print(out))
  expect_type(printed, "character")
})

# Treated-unit counts by event period
get_event_counts <- function(data, time_to_treat_var, unit_id) {
  time_sym <- rlang::sym(time_to_treat_var)
  unit_sym <- rlang::sym(unit_id)

  data %>%
    dplyr::filter(!is.na(!!time_sym)) %>%
    dplyr::group_by(!!time_sym, .drop = FALSE) %>%
    dplyr::summarise(
      count = dplyr::n_distinct(!!unit_sym),
      .groups = "drop"
    ) %>%
    dplyr::rename(Period = !!time_sym)
}


# Build event-study data frame
make_es_df <- function(Period_vec, est_vec, se_vec = NULL,
                       count_df = NULL, method = method) {
  df <- data.frame(ATT = est_vec, check.names = FALSE)

  if (!is.null(se_vec)) {
    df[["S.E."]] <- se_vec
    df[["CI.lower"]] <- df$ATT - 1.96 * se_vec
    df[["CI.upper"]] <- df$ATT + 1.96 * se_vec
    df[["p.value"]] <- 2 * (1 - stats::pnorm(abs(df$ATT / se_vec)))
  } else {
    df[["S.E."]] <- NA
    df[["CI.lower"]] <- NA
    df[["CI.upper"]] <- NA
    df[["p.value"]] <- NA
  }

  ## attach period counts if supplied
  if (!is.null(count_df)) {
    df$Period <- Period_vec + as.numeric(method != "didm")
    df <- df |>
      dplyr::left_join(
        count_df |>
          dplyr::mutate(Period = .data$Period + 1), # Assuming Period in count_df is 0-indexed relative to event
        by = "Period"
      )
    # If Period_vec can have NAs, this might cause issues with rownames
    # Ensure Period_vec used for rownames is clean or handle NAs appropriately
    valid_periods <- !is.na(df$Period) # Or use original Period_vec if df$Period might be modified by join
    if (any(valid_periods)) {
      rownames(df)[valid_periods] <- df$Period[valid_periods]
    }
    df$Period <- NULL
  } else {
    # Ensure Period_vec used for rownames is clean
    valid_rownames <- Period_vec + as.numeric(method != "didm")
    if (length(valid_rownames) == nrow(df) && !any(is.na(valid_rownames))) {
      rownames(df) <- valid_rownames
    }
    df$count <- NA
  }

  df <- df[, c("ATT", "S.E.", "CI.lower", "CI.upper", "p.value", "count")]
  df
}

# ── Main wrapper ───────────────────────────────────────────────────────────

did_wrapper <- function(
    data, Y, D, X = NULL,
    index, # c("unit_id","time_id")
    method = c("twfe", "st", "iw", "cs_never", "cs_notyet", "didm"),
    se = c("default", "boot", "bootstrap", "jackknife"),
    nboots = 200,
    parallel = TRUE,
    core = NULL,
    ## TWFE / stacked-DID options
    time_to_treat_var = "Time_to_Treatment",
    treat_indicator = "treat",
    ## Callaway-Sant'Anna options
    csdid.base_period = "universal",
    ## DIDmultiplegtDYN options
    didm.effects = NA,
    didm.placebo = NA) {
  if (method == "bootstrap") method <- "boot"
  method <- match.arg(method)
  se <- match.arg(se)
  unit_id <- index[1]
  time_id <- index[2]

  # Define symbols for unit_id and D here for use throughout the function, including run_estimator_once
  unit_sym <- rlang::sym(unit_id)
  D_sym <- rlang::sym(D)
  # Y_sym    <- rlang::sym(Y) # If Y is used with dplyr NSE

  if (method == "didm" && is.na(didm.effects)) {
    stop('For "didm", you must supply "didm.effects" and "didm.placebo".')
  }

  ## ── Pre-processing ──────────────────────────────────────────────────────
  original_units <- length(unique(data[[unit_id]])) # Base R subsetting, fine

  data <- data %>%
    dplyr::group_by(!!unit_sym) %>% # Uses pre-defined unit_sym
    dplyr::mutate(treatment_mean = mean(!!D_sym, na.rm = TRUE)) %>% # Uses pre-defined D_sym
    dplyr::ungroup()
  data <- data[data$treatment_mean < 1, ] # Base R subsetting, fine
  message(
    "Dropped ", original_units - length(unique(data[[unit_id]])), # Base R subsetting, fine
    " always-treated units."
  )

  ## define cohorts & event time
  # Ensure fect::get.cohort doesn't have issues with column names if they contain special characters
  data <- fect::get.cohort(data, D = D, index = c(unit_id, time_id), start0 = TRUE)
  count_df <- get_event_counts(data, time_to_treat_var, unit_id)

  ## ── Core estimation routine (one sample) ───────────────────────────────
  run_estimator_once <- function(dd) { # dd is the data for this run

    out_ATT <- out_SE <- out_lower <- out_upper <- NA
    es_df <- data.frame()

    # ── 1) TWFE ───────────────────────────────────────────────────────────
    if (method == "twfe") {
      if (!treat_indicator %in% names(dd)) {
        dd <- dd |>
          dplyr::group_by(!!unit_sym) |> # MODIFIED: Use unit_sym from parent scope
          dplyr::mutate(tmp = as.numeric(mean(!!D_sym, na.rm = TRUE) > 0)) |> # MODIFIED: Use D_sym from parent scope
          dplyr::ungroup()
        names(dd)[names(dd) == "tmp"] <- treat_indicator
      }

      X_part <- if (!is.null(X)) paste(X, collapse = " + ") else "1"

      fit_att <- fixest::feols(
        stats::as.formula(
          paste0(Y, " ~ ", D, " + ", X_part, " | ", unit_id, " + ", time_id)
        ),
        data = dd,
        cluster = unit_id, # fixest can take string for cluster
        notes = FALSE
      )
      co <- fixest::coeftable(fit_att)
      est <- co[D, "Estimate"]
      se_ <- co[D, "Std. Error"]
      out_ATT <- est
      out_SE <- se_
      out_lower <- est - 1.96 * se_
      out_upper <- est + 1.96 * se_

      if (!time_to_treat_var %in% names(dd)) {
        dd <- fect::get.cohort(dd, D = D, index = c(unit_id, time_id), start0 = TRUE)
      }

      # Ensure time_to_treat_var column exists and handle NAs before using in i()
      if (time_to_treat_var %in% names(dd)) {
        dd[[time_to_treat_var]][is.na(dd[[time_to_treat_var]])] <- 0 # Or other suitable placeholder for fixest::i()
      } else {
        stop(paste("Column", time_to_treat_var, "not found in data for TWFE event study."))
      }
      # Ensure treat_indicator column exists
      if (!treat_indicator %in% names(dd)) {
        stop(paste("Column", treat_indicator, "not found in data for TWFE event study."))
      }


      fit_es <- fixest::feols(
        stats::as.formula(
          paste0(
            Y, " ~ i(", time_to_treat_var, ", ", treat_indicator, ", ref=-1) | ",
            unit_id, " + ", time_id
          )
        ),
        data = dd,
        cluster = unit_id, # fixest can take string for cluster
        notes = FALSE
      )

      ctab <- as.data.frame(fit_es$coeftable)
      # Robustly extract period numbers, handle cases where parsing fails
      period_nums <- vapply(
        rownames(ctab),
        function(x) {
          sp <- strsplit(x, "::")[[1]]
          if (length(sp) < 2) {
            return(NA_real_)
          }
          val_str <- strsplit(sp[2], ":")[[1]][1]
          suppressWarnings(as.numeric(val_str)) # Suppress warnings for non-numeric if any
        },
        numeric(1)
      )
      # Filter out NAs that might result from parsing non-coefficient rows or failed parsing
      valid_coeffs <- !is.na(period_nums)
      es_df <- make_es_df(
        period_nums[valid_coeffs], ctab$Estimate[valid_coeffs],
        ctab$`Std. Error`[valid_coeffs], count_df, method
      )


      # ── 2) Stacked DID ("st") ─────────────────────────────────────────────
    } else if (method == "st") {
      if (!"Cohort" %in% names(dd)) {
        dd <- fect::get.cohort(dd, D = D, index = c(unit_id, time_id), start0 = TRUE)
      }

      target.cohorts <- setdiff(unique(dd$Cohort), "Control")
      df.st_list <- vector("list", length(target.cohorts)) # Pre-allocate list
      for (i in seq_along(target.cohorts)) {
        c <- target.cohorts[i]
        # Use a temporary data frame to avoid modifying `dd` in a loop if it's large
        temp_df <- dd[dd$Cohort %in% c(c, "Control"), ]
        temp_df$stack_id <- i # Use loop index for stack_id
        df.st_list[[i]] <- temp_df
      }
      df.st <- do.call(rbind, df.st_list)


      df.st$st_unit <- as.numeric(factor(paste0(df.st$stack_id, "-", df.st[[unit_id]])))
      df.st$st_time <- as.numeric(factor(paste0(df.st$stack_id, "-", df.st[[time_id]])))

      fit_st <- fixest::feols(
        stats::as.formula(paste0(Y, " ~ ", D, " | st_unit + st_time")),
        data = df.st,
        cluster = "st_unit", # fixest can take string for cluster
        notes = FALSE,
        warn = FALSE
      )
      co <- fixest::coeftable(fit_st)
      est <- co[D, "Estimate"]
      se_ <- co[D, "Std. Error"]
      out_ATT <- est
      out_SE <- se_
      out_lower <- est - 1.96 * se_
      out_upper <- est + 1.96 * se_

      ## event study (same routine as above)
      if (!treat_indicator %in% names(df.st)) {
        df.st <- df.st |>
          dplyr::group_by(!!unit_sym) |> # MODIFIED: Use unit_sym from parent scope
          dplyr::mutate(tmp = as.numeric(mean(!!D_sym, na.rm = TRUE) > 0)) |> # MODIFIED: Use D_sym from parent scope
          dplyr::ungroup()
        names(df.st)[names(df.st) == "tmp"] <- treat_indicator
      }

      if (!time_to_treat_var %in% names(df.st)) {
        df.st <- fect::get.cohort(df.st, D = D, index = c(unit_id, time_id), start0 = TRUE)
      }

      # Ensure time_to_treat_var column exists and handle NAs
      if (time_to_treat_var %in% names(df.st)) {
        df.st[[time_to_treat_var]][is.na(df.st[[time_to_treat_var]])] <- 999999 # Or other placeholder
      } else {
        stop(paste("Column", time_to_treat_var, "not found in data for Stacked DID event study."))
      }
      if (!treat_indicator %in% names(df.st)) {
        stop(paste("Column", treat_indicator, "not found in data for Stacked DID event study."))
      }


      fit_es <- fixest::feols(
        stats::as.formula(
          paste0(
            Y, " ~ i(", time_to_treat_var, ", ", treat_indicator,
            ", ref=-1) | st_unit + st_time"
          )
        ),
        data = df.st,
        cluster = "st_unit" # fixest can take string for cluster
      )
      ctab <- as.data.frame(fit_es$coeftable)
      per <- vapply(
        rownames(ctab),
        function(x) {
          sp <- strsplit(x, "::")[[1]]
          if (length(sp) < 2) {
            return(NA_real_)
          }
          val_str <- strsplit(sp[2], ":")[[1]][1]
          suppressWarnings(as.numeric(val_str))
        },
        numeric(1)
      )
      valid_coeffs <- !is.na(per)
      es_df <- make_es_df(
        per[valid_coeffs], ctab$Estimate[valid_coeffs],
        ctab$`Std. Error`[valid_coeffs], count_df, method
      )


      # ── 3) IW (Sun-&-Abraham)  ────────────────────────────────────────────
    } else if (method == "iw") {
      if (!"FirstTreat" %in% names(dd)) { # 'FirstTreat' is typically generated by get.cohort or similar
        stop("Method 'iw' requires a FirstTreat column. Ensure get.cohort() was run or the column exists.")
      }

      dd$FirstTreat[is.na(dd$FirstTreat)] <- 1000 # Or other large value indicating never/not-yet treated for sunab
      X_part <- if (!is.null(X)) paste(X, collapse = " + ") else "1"

      fit_iw <- fixest::feols(
        stats::as.formula(
          # Ensure time_id is a valid column name for sunab
          paste0(
            Y, " ~ sunab(FirstTreat,", time_id, ") + ", X_part,
            " | ", unit_id, " + ", time_id
          )
        ),
        data = dd,
        cluster = unit_id # fixest can take string for cluster
      )
      att_sum <- summary(fit_iw, agg = "ATT")
      out_ATT <- att_sum$coeftable["ATT", "Estimate"]
      out_SE <- att_sum$coeftable["ATT", "Std. Error"]
      out_lower <- out_ATT - 1.96 * out_SE
      out_upper <- out_ATT + 1.96 * out_SE

      ctab <- as.data.frame(fixest::coeftable(fit_iw))
      # Robustly extract offsets for sunab
      offsets <- suppressWarnings(
        as.numeric(sub("^.*::", "", sub(":cohort::.*", "", rownames(ctab))))
      )
      valid <- !is.na(offsets)
      es_df <- make_es_df(
        offsets[valid], ctab$Estimate[valid],
        ctab$`Std. Error`[valid], count_df, method
      )


      # ── 4) Callaway-Sant'Anna (never-treated) ─────────────────────────────
    } else if (method == "cs_never") {
      if (!"FirstTreat" %in% names(dd)) { # 'FirstTreat' is gname for did::att_gt
        dd <- fect::get.cohort(dd, D = D, index = c(unit_id, time_id), start0 = TRUE)
      }

      # For 'nevertreated', control units usually have FirstTreat = 0 or Inf
      # Check did package documentation for exact requirement for gname with nevertreated
      dd$FirstTreat[is.na(dd$FirstTreat)] <- 0 # Assuming NA means never treated and coded as 0

      cs.out <- did::att_gt(
        yname = Y, gname = "FirstTreat", # Ensure 'FirstTreat' is the correct group identifier
        idname = unit_id, tname = time_id,
        xformla = if (!is.null(X)) stats::as.formula(paste("~", paste(X, collapse = "+"))) else ~1,
        control_group = "nevertreated",
        allow_unbalanced_panel = TRUE, # Consider implications
        data = as.data.frame(dd), # did package often prefers data.frame
        est_method = "reg", # Default, consider "dr" for doubly robust
        base_period = csdid.base_period
      )

      simple <- tryCatch(did::aggte(cs.out, type = "simple", na.rm = TRUE),
        error = function(e) {
          warning("Error in did::aggte (simple, cs_never): ", e$message)
          NULL
        }
      )
      if (!is.null(simple)) {
        out_ATT <- simple$overall.att
        out_SE <- simple$overall.se
        out_lower <- out_ATT - 1.96 * out_SE
        out_upper <- out_ATT + 1.96 * out_SE
      }

      dyn <- tryCatch(
        did::aggte(cs.out, type = "dynamic", na.rm = TRUE, cband = FALSE), # cband=FALSE for point CIs
        error = function(e) {
          warning("Error in did::aggte (dynamic, cs_never): ", e$message)
          NULL
        }
      )
      es_df <- if (is.null(dyn) || length(dyn$egt) == 0) { # Check if dyn$egt is empty
        make_es_df(numeric(0), numeric(0), numeric(0), count_df, method)
      } else {
        make_es_df(dyn$egt, dyn$att.egt, dyn$se.egt, count_df, method)
      }


      # ── 5) Callaway-Sant'Anna (not-yet-treated) ───────────────────────────
    } else if (method == "cs_notyet") {
      if (!"FirstTreat" %in% names(dd)) {
        dd <- fect::get.cohort(dd, D = D, index = c(unit_id, time_id), start0 = TRUE)
      }

      # For 'notyettreated', FirstTreat should be year of treatment or Inf/0 for never treated
      # Ensure NA handling for FirstTreat is appropriate for did::att_gt
      # dd$FirstTreat[is.na(dd$FirstTreat)] <- Inf # Or appropriate coding for never treated
      
      dd$FirstTreat[is.na(dd$FirstTreat)] <- 0

      cs.out <- did::att_gt(
        yname = Y, gname = "FirstTreat",
        idname = unit_id, tname = time_id,
        xformla = if (!is.null(X)) stats::as.formula(paste("~", paste(X, collapse = "+"))) else ~1,
        control_group = "notyettreated",
        allow_unbalanced_panel = TRUE,
        data = as.data.frame(dd), # did package often prefers data.frame
        est_method = "reg",
        base_period = csdid.base_period
      )

      simple <- tryCatch(did::aggte(cs.out, type = "simple", na.rm = TRUE),
        error = function(e) {
          warning("Error in did::aggte (simple, cs_notyet): ", e$message)
          NULL
        }
      )
      if (!is.null(simple)) {
        out_ATT <- simple$overall.att
        out_SE <- simple$overall.se
        out_lower <- out_ATT - 1.96 * out_SE
        out_upper <- out_ATT + 1.96 * out_SE
      }

      dyn <- tryCatch(
        did::aggte(cs.out, type = "dynamic", na.rm = TRUE, cband = FALSE),
        error = function(e) {
          warning("Error in did::aggte (dynamic, cs_notyet): ", e$message)
          NULL
        }
      )
      es_df <- if (is.null(dyn) || length(dyn$egt) == 0) {
        make_es_df(numeric(0), numeric(0), numeric(0), count_df, method)
      } else {
        make_es_df(dyn$egt, dyn$att.egt, dyn$se.egt, count_df, method)
      }


      # ── 6) DIDmultiplegtDYN Dyn ──────────────────────────────────────────────
    } else if (method == "didm") {
      res <- DIDmultiplegtDYN::did_multiplegt_dyn(
        df        = as.data.frame(dd), # DIDmultiplegtDYN prefers data.frame
        outcome   = Y,
        group     = unit_id,
        time      = time_id,
        treatment = D,
        controls  = X, # Pass X directly if it's a character vector of control names
        effects   = didm.effects,
        placebo   = didm.placebo,
        cluster   = unit_id, # Ensure this column exists and is appropriate for clustering
        graph_off = TRUE
      )

      # Check structure of res$results for safety
      if (!is.null(res$results) && "ATE" %in% names(res$results) && length(res$results$ATE) >= 2) {
        out_ATT <- res$results$ATE[1]
        out_SE <- res$results$ATE[2]
        if (length(res$results$ATE) >= 4) { # Check if CI is provided
          out_lower <- res$results$ATE[3]
          out_upper <- res$results$ATE[4]
        } else {
          out_lower <- out_ATT - 1.96 * out_SE
          out_upper <- out_ATT + 1.96 * out_SE
        }
      } else {
        warning("Results from did_multiplegt_dyn ATE not found or in unexpected format.")
      }


      Placebos <- res$results$Placebos
      Effects <- res$results$Effects
      T.pre <- if (!is.null(Placebos) && inherits(Placebos, "matrix")) nrow(Placebos) else 0
      T.post <- if (!is.null(Effects) && inherits(Effects, "matrix")) nrow(Effects) else 0


      if (T.pre + T.post == 0) {
        es_df <- make_es_df(numeric(0), numeric(0), numeric(0), count_df, method)
      } else {
        est_placebo <- if (T.pre > 0 && "Estimate" %in% colnames(Placebos)) Placebos[, "Estimate"] else numeric(0)
        est_effect <- if (T.post > 0 && "Estimate" %in% colnames(Effects)) Effects[, "Estimate"] else numeric(0)
        est <- c(est_placebo, est_effect)

        se_placebo <- if (T.pre > 0 && "SE" %in% colnames(Placebos)) Placebos[, "SE"] else numeric(0)
        se_effect <- if (T.post > 0 && "SE" %in% colnames(Effects)) Effects[, "SE"] else numeric(0)
        ses <- c(se_placebo, se_effect)

        # Ensure lengths match before creating data frame
        if (length(est) != length(ses)) {
          warning("Mismatch in length of estimates and standard errors from did_multiplegt_dyn.")
          # Fallback to empty or handle error appropriately
          es_df <- make_es_df(numeric(0), numeric(0), numeric(0), count_df, method)
        } else {
          peri <- c(
            if (T.pre > 0) seq(-T.pre, -1) else numeric(0),
            if (T.post > 0) seq(0, T.post - 1) else numeric(0)
          ) # didm effects often 0-indexed

          peri <- c(
            if (T.pre > 0) seq(-T.pre, -1) else numeric(0),
            if (T.post > 0) seq(1, T.post) else numeric(0)
          )

          es_df <- make_es_df(peri, est, ses, count_df, method)
        }
      }
    }

    list(
      ATT = out_ATT, ATT_se = out_SE,
      CI_lower = out_lower, CI_upper = out_upper,
      es = es_df
    )
  }

  ## ── SE / bootstrap handling ────────────────────────────────────────────
  if (se == "default") {
    out_full <- run_estimator_once(data)
    # Ensure ATT_se is not zero or NA before calculating p-value
    overall_p <- if (!is.na(out_full$ATT_se) && out_full$ATT_se != 0) {
      2 * (1 - stats::pnorm(abs(out_full$ATT / out_full$ATT_se)))
    } else {
      NA_real_
    }
  } else { # bootstrap or jackknife

    cluster_ids <- unique(data[[unit_id]]) # Base R, fine
    n_clusters <- length(cluster_ids)

    # Ensure nboots is sensible for jackknife
    if (se == "jackknife" && nboots != n_clusters) {
      message(paste0("For jackknife, nboots should be equal to the number of clusters (", n_clusters, "). Setting nboots to ", n_clusters, "."))
      nboots <- n_clusters
    }


    boot_fun <- function(b) { # b is the iteration number
      if (se %in% c("boot", "bootstrap")) {
        # Stratified bootstrap if applicable, or simple cluster bootstrap
        samp_indices <- sample(seq_along(cluster_ids), n_clusters, replace = TRUE)
        samp_cluster_names <- cluster_ids[samp_indices]
        # Efficiently create bootstrap sample by selecting rows
        # This assumes unit_id column is a factor or character
        d_b <- data[data[[unit_id]] %in% samp_cluster_names, ]
        # To handle duplicated clusters correctly if units within clusters are distinct in d_b:
        # This requires more care if unit IDs are not unique across original clusters
        # A common way is to replicate data for chosen clusters and assign new unique IDs if needed by downstream.
        # However, fixest handles repeated clusters in cluster argument if data is just rbind-ed.
        # The current approach data[data[[unit_id]] %in% samp, ] is standard for cluster bootstrap.
      } else { # jackknife
        omit_cluster_name <- cluster_ids[b] # b is 1 to n_clusters
        d_b <- data[data[[unit_id]] != omit_cluster_name, ]
      }
      run_estimator_once(d_b)
    }

    if (!identical(parallel, FALSE)) {
      if (is.null(core)) core <- parallelly::availableCores(omit = 1)
      ## v2.4.2+: route through helper to pre-load packages silently.
      future::plan(future::cluster, workers = .fect_make_future_cluster(core))
      # Ensure seed is handled correctly for parallel processing
      rep_list <- .fect_with_quiet_pkg_warnings(
        future.apply::future_lapply(seq_len(nboots), boot_fun,
          future.seed = TRUE
        )
      )
      future::plan(future::sequential) # Reset plan
    } else {
      rep_list <- lapply(seq_len(nboots), boot_fun)
    }

    att_reps <- vapply(rep_list, `[[`, numeric(1), "ATT")
    # Remove NAs from bootstrap replications if any estimator failed
    att_reps_clean <- att_reps[!is.na(att_reps)]
    if (length(att_reps_clean) < length(att_reps)) {
      warning(paste(length(att_reps) - length(att_reps_clean), "bootstrap/jackknife replications resulted in NA ATT values and were dropped."))
    }
    if (length(att_reps_clean) < 2) { # Need at least 2 for sd()
      warning("Too few successful bootstrap/jackknife replications to calculate SE/CI.")
      out_full <- run_estimator_once(data) # Get point estimate
      out_full$ATT_se <- NA_real_
      out_full$CI_lower <- NA_real_
      out_full$CI_upper <- NA_real_
      overall_p <- NA_real_
    } else {
      out_full <- run_estimator_once(data) # Get point estimate from full sample
      theta_hat <- out_full$ATT
      out_full$CI_lower <- 2 * theta_hat - stats::quantile(att_reps_clean, 0.975, na.rm = TRUE)
      out_full$CI_upper <- 2 * theta_hat - stats::quantile(att_reps_clean, 0.025, na.rm = TRUE)
      out_full$ATT_se <- stats::sd(att_reps_clean, na.rm = TRUE)
      p1 <- if (theta_hat > 0) {
        mean(att_reps_clean <= 0, na.rm = TRUE)
      } else {
        mean(att_reps_clean >= 0, na.rm = TRUE)
      }
      overall_p <- min(2 * p1, 1)
    }

    if (nrow(out_full$es) > 0) {
      periods_from_es <- as.numeric(rownames(out_full$es))


      boot_mat_list <- lapply(rep_list, function(x) {
        if (is.null(x$es) || nrow(x$es) == 0) {
          return(rep(NA_real_, length(periods_from_es)))
        }
        # Match based on rownames (periods)
        es_rep_periods <- as.numeric(rownames(x$es))
        matched_att <- x$es$ATT[match(periods_from_es, es_rep_periods)]
        # If some periods are not in x$es, matched_att will have NAs, which is fine.
        return(matched_att)
      })
      boot_mat <- do.call(rbind, boot_mat_list) # Rows are bootstrap reps, columns are periods

      # Filter out columns (periods) that are all NA if any
      valid_cols <- apply(boot_mat, 2, function(col) !all(is.na(col)))
      boot_mat_filtered <- boot_mat[, valid_cols, drop = FALSE]
      periods_filtered <- periods_from_es[valid_cols]
      original_att_filtered <- out_full$es$ATT[valid_cols]

      if (ncol(boot_mat_filtered) > 0) {
        boot_se <- apply(boot_mat_filtered, 2, stats::sd, na.rm = TRUE)
        ci_quantiles_upper <- apply(boot_mat_filtered, 2, stats::quantile, probs = 0.975, na.rm = TRUE)
        ci_quantiles_lower <- apply(boot_mat_filtered, 2, stats::quantile, probs = 0.025, na.rm = TRUE)

        pvals_es <- vapply(
          seq_len(ncol(boot_mat_filtered)),
          function(j) {
            est_j <- original_att_filtered[j]
            reps_j <- boot_mat_filtered[, j]
            reps_j_clean <- reps_j[!is.na(reps_j)]
            if (length(reps_j_clean) < 2) {
              return(NA_real_)
            }
            p1 <- if (est_j > 0) {
              mean(reps_j_clean <= 0, na.rm = TRUE)
            } else {
              mean(reps_j_clean >= 0, na.rm = TRUE)
            }
            min(2 * p1, 1)
          },
          numeric(1)
        )

        # Assign back to the original out_full$es structure, potentially for a subset of periods
        out_full$es$`S.E.` <- NA_real_
        out_full$es$CI.lower <- NA_real_
        out_full$es$CI.upper <- NA_real_
        out_full$es$p.value <- NA_real_

        out_full$es$`S.E.`[valid_cols] <- boot_se
        out_full$es$CI.lower[valid_cols] <- 2 * original_att_filtered - ci_quantiles_upper
        out_full$es$CI.upper[valid_cols] <- 2 * original_att_filtered - ci_quantiles_lower
        out_full$es$p.value[valid_cols] <- pvals_es
      } else {
        warning("No valid data for bootstrapping event study standard errors.")
      }
    }
  }

  ## ── Return object ───────────────────────────────────────────────────────
  res <- list(
    est.avg = data.frame(
      ATT.avg = out_full$ATT,
      S.E. = out_full$ATT_se,
      CI.lower = out_full$CI_lower,
      CI.upper = out_full$CI_upper,
      p.value = overall_p,
      row.names = NULL # Ensure no rownames for this single-row data frame
    ),
    est.att = out_full$es # This should have periods as rownames from make_es_df
  )
  class(res) <- "did_wrapper"
  res
}

## ---------------------------------------------------------------
## Tests for estimand(fit, "log.att", "event.time").
##
## logATT_t = mean_{(t,i): D=1} (log(Y) - log(Y0_hat))
##
## Validates against a manual computation.
## ---------------------------------------------------------------

## simdata's Y can be negative, which makes log undefined for many
## cells. Build a small synthetic positive-Y panel for testing.

.fit_positive_Y <- function(nboots = 30, keep.sims = TRUE, seed = 7) {
  set.seed(seed)
  N <- 30; TT <- 20
  df <- expand.grid(id = 1:N, time = 1:TT)
  treat_start <- sample(c(NA, 8:15), N, replace = TRUE)
  df$D <- ifelse(is.na(treat_start[df$id]) | df$time < treat_start[df$id],
                 0, 1)
  ## Higher intercept (2.0) and smaller noise (0.1) keep Y comfortably
  ## positive so the v2.4.2 cell-drop hard-error doesn't fire on benign
  ## bootstrap noise.
  df$Y <- exp(2.0 + 0.05 * df$time + 0.3 * df$D + rnorm(nrow(df), sd = 0.1))
  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "fe", force = "two-way",
      se = TRUE, nboots = nboots, parallel = FALSE,
      keep.sims = keep.sims
    )
  ))
}


## -- LA.1  Schema -----------------------------------------------------

test_that("LA.1: estimand(fit, 'log.att', 'event.time') has the documented schema", {

  skip_on_cran()

  fit <- .fit_positive_Y()
  est <- suppressWarnings(fect::estimand(fit, "log.att", "event.time"))

  expect_s3_class(est, "data.frame")
  expected_cols <- c("event.time", "estimate", "se",
                     "ci.lo", "ci.hi", "n_cells", "vartype")
  expect_setequal(names(est), expected_cols)

  expect_true(is.numeric(est$estimate))
})


## -- LA.2  Manual recipe match ----------------------------------------

test_that("LA.2: estimate matches manual log-ATT", {

  skip_on_cran()

  fit <- .fit_positive_Y()
  est <- suppressWarnings(fect::estimand(fit, "log.att", "event.time"))

  Don <- !is.na(fit$D.dat) & fit$D.dat == 1 & !is.na(fit$T.on)
  ets <- sort(unique(fit$T.on[Don]))

  manual <- vapply(ets, function(et) {
    m <- Don & fit$T.on == et
    eff_t <- fit$eff[m]
    Y_t   <- fit$Y.dat[m]
    Y0_t  <- Y_t - eff_t
    ok <- !is.na(Y_t) & !is.na(Y0_t) & Y_t > 0 & Y0_t > 0
    if (sum(ok) == 0) return(NA_real_)
    mean(log(Y_t[ok]) - log(Y0_t[ok]), na.rm = TRUE)
  }, numeric(1))

  est_sorted <- est[order(est$event.time), ]
  expect_equal(est_sorted$estimate, manual, tolerance = 1e-10)
})


## -- LA.3  keep.sims = FALSE error path -------------------------------

test_that("LA.3: log.att without keep.sims errors with the locked wording", {

  skip_on_cran()

  fit_no_keep <- .fit_positive_Y(keep.sims = FALSE)
  expect_error(
    fect::estimand(fit_no_keep, "log.att", "event.time"),
    "keep\\.sims = TRUE",
    fixed = FALSE
  )
})


## -- LA.4  Negative Y cells trigger the v2.4.2 hard-error -----------

test_that("LA.4: negative Y cells trigger point-level hard-error (v2.4.2+)", {

  skip_on_cran()

  ## simdata has negative Y values, so log-ATT triggers the v2.4.2
  ## point-estimate-level hard-error before any bootstrap work. The
  ## bootstrap-level "log-ATT bootstrap is unreliable" path still
  ## exists in R/po-estimands.R for strictly-positive panels where
  ## Y0_b crosses zero only in some bootstrap replicates.
  data("simdata", package = "fect")
  set.seed(42)
  fit_neg <- suppressWarnings(suppressMessages(
    fect::fect(Y ~ D, data = simdata, index = c("id", "time"),
               method = "fe", force = "two-way",
               se = TRUE, nboots = 20, parallel = FALSE,
               keep.sims = TRUE)
  ))

  expect_error(
    fect::estimand(fit_neg, "log.att", "event.time"),
    "log\\.att requires Y > 0"
  )
})

test_that("did_wrapper twfe runs and returns structure", {
  skip_on_cran()
  skip_if_not_installed("fixest")
  suppressWarnings(try(data("simdata", package = "fect"), silent = TRUE))
  expect_true(exists("simdata"))

  # create a binary ever-treated indicator for TWFE event study helper
  sim <- simdata
  sim$treat <- ave(sim$D, sim$id, FUN = function(z) as.numeric(mean(z, na.rm = TRUE) > 0))

  res <- did_wrapper(
    data = sim,
    Y = "Y",
    D = "D",
    X = c("X1", "X2"),
    index = c("id", "time"),
    method = "twfe",
    se = "default",
    parallel = FALSE
  )

  expect_s3_class(res, "did_wrapper")
  expect_true(is.data.frame(res$est.avg))
  expect_true(is.data.frame(res$est.att))
  expect_true(all(colnames(res$est.avg) %in% c("ATT.avg", "S.E.", "CI.lower", "CI.upper", "p.value")))
})

###############################################################################
## Post-hoc estimands API
##
## Public surface (v2.4.0):
##   imputed_outcomes(fit, cells, replicates, direction)
##   estimand(fit, type, by, cells, weights, direction, vartype, ...)
##
## This file owns the slot contract documented in
##   statsclaw-workspace/fect/ref/po-estimands-contract.md
## See section 3 (Slot contract on the fit object) for the invariants enforced by
## .validate_po_contract().
###############################################################################


## ---------------------------------------------------------------------------
## Soft-deprecation helper for v2.4.0
## ---------------------------------------------------------------------------

## Internal: emit a one-time-per-session message that the named function
## is soft-deprecated, pointing users at the unified estimand() API.
## Suppressed during internal calls from estimand() itself.
.fect_deprecation_message_once <- function(fn) {
    if (isTRUE(getOption("fect.suppress_estimand_deprecation", FALSE))) {
        return(invisible(NULL))
    }
    key <- paste0("fect.deprecated_warned_", fn)
    if (isTRUE(getOption(key, FALSE))) {
        return(invisible(NULL))
    }
    message(
        sprintf(
            "`%s()` is soft-deprecated as of fect 2.4.0; prefer the unified ",
            fn
        ),
        "`estimand(fit, \"att.cumu\", ...)` API for the same numerics. ",
        "See `?estimand` and the \"Alternative estimands\" vignette section. ",
        "This function will continue to work; removal not before v3.0.0."
    )
    do.call(options, setNames(list(TRUE), key))
    invisible(NULL)
}


## Internal: evaluate `expr` with the deprecation messages on effect()
## and att.cumu() temporarily suppressed. Used by .compute_att_cumu_*()
## so that calling estimand("att.cumu", ...) does not surface the
## deprecation message that fires only on user-facing calls.
.with_no_deprecation <- function(expr) {
    old <- getOption("fect.suppress_estimand_deprecation", FALSE)
    options(fect.suppress_estimand_deprecation = TRUE)
    on.exit(options(fect.suppress_estimand_deprecation = old), add = TRUE)
    force(expr)
}


## ---------------------------------------------------------------------------
## Slot contract
## ---------------------------------------------------------------------------

## Internal: validate that `fit` satisfies the slot contract.
## Called at the top of every public-API entry point. Raises an error
## that points at the contract on violation.
.validate_po_contract <- function(fit, require_replicates = FALSE) {
    if (!inherits(fit, "fect") && !is.list(fit)) {
        stop("`fit` must be a fect object (output of fect()).", call. = FALSE)
    }

    required <- c("Y.dat", "D.dat", "I.dat", "T.on",
                  "eff", "id", "rawtime")
    missing_slots <- setdiff(required, names(fit))
    if (length(missing_slots) > 0) {
        stop(
            "fit object missing required slots: ",
            paste(missing_slots, collapse = ", "),
            ". The new post-hoc estimand API requires fects produced by ",
            "v2.3.x or later. Refit with the current package version.",
            call. = FALSE
        )
    }

    if (!is.matrix(fit$Y.dat)) {
        stop("Slot contract: fit$Y.dat must be a matrix.", call. = FALSE)
    }
    TT <- nrow(fit$Y.dat)
    N  <- ncol(fit$Y.dat)

    for (slot in c("D.dat", "I.dat", "T.on", "eff")) {
        v <- fit[[slot]]
        if (!is.matrix(v) || !identical(dim(v), c(TT, N))) {
            stop(
                "Slot contract: fit$", slot,
                " must be a TT x N matrix matching fit$Y.dat (",
                TT, " x ", N, ").",
                call. = FALSE
            )
        }
    }

    if (length(fit$id) != N) {
        stop(
            "Slot contract: length(fit$id) (", length(fit$id),
            ") must equal ncol(fit$Y.dat) (", N, ").",
            call. = FALSE
        )
    }
    if (length(fit$rawtime) != TT) {
        stop(
            "Slot contract: length(fit$rawtime) (", length(fit$rawtime),
            ") must equal nrow(fit$Y.dat) (", TT, ").",
            call. = FALSE
        )
    }

    ## eff is the cell-level score per the contract:
    ##   eff[t, i] = Y_obs[t, i] - Y_hat[t, i]
    ## where Y_hat is the prediction from the control-fit. At treated
    ## cells this is the cell-level treatment effect estimate; at
    ## non-treated cells it is the model residual (typically small but
    ## not exactly zero). All ATT-style estimands aggregate eff over
    ## treated cells only.

    ## eff.boot shape depends on vartype.
    ##   bootstrap / wild / parametric: TT x N x B
    ##   jackknife:                     TT x (N-1) x N
    ##     The third dimension is the dropped-unit index; the second is N-1
    ##     because each leave-one-out fit runs on a panel with one unit
    ##     removed. The post-hoc estimand API uses eff.boot only through the
    ##     dedicated .jackknife_tukey_se_*() helpers, never by treating its
    ##     second dimension as N.
    if (!is.null(fit$eff.boot)) {
        eb <- fit$eff.boot
        if (!is.array(eb) || length(dim(eb)) != 3) {
            stop(
                "Slot contract: fit$eff.boot must be a 3D array.",
                call. = FALSE
            )
        }
        is_jack <- isTRUE(fit$vartype == "jackknife")
        if (is_jack) {
            if (dim(eb)[1] != TT) {
                stop(
                    "Slot contract (jackknife): fit$eff.boot[1] = ",
                    dim(eb)[1], " must equal TT = ", TT, ".",
                    call. = FALSE
                )
            }
            if (dim(eb)[3] != N) {
                stop(
                    "Slot contract (jackknife): fit$eff.boot[3] = ",
                    dim(eb)[3], " must equal N = ", N,
                    " (one leave-one-out replicate per unit).",
                    call. = FALSE
                )
            }
            ## dim(eb)[2] == N-1 is the normal case; warn if unexpected.
            if (dim(eb)[2] != N - 1L) {
                warning(
                    "Slot contract (jackknife): fit$eff.boot[2] = ",
                    dim(eb)[2], "; expected N-1 = ", N - 1L,
                    ". Proceeding, but results may be incorrect.",
                    call. = FALSE
                )
            }
        } else {
            if (!identical(dim(eb)[1:2], c(TT, N))) {
                stop(
                    "Slot contract: fit$eff.boot first two dimensions (",
                    dim(eb)[1], " x ", dim(eb)[2],
                    ") must match TT x N (", TT, " x ", N, ").",
                    call. = FALSE
                )
            }
        }
    }

    ## eff_debias, if populated, matches eff shape.
    if (!is.null(fit$eff_debias)) {
        eb <- fit$eff_debias
        if (!is.matrix(eb) || !identical(dim(eb), c(TT, N))) {
            stop(
                "Slot contract: fit$eff_debias must be a TT x N matrix ",
                "(or NULL).",
                call. = FALSE
            )
        }
    }

    if (require_replicates && is.null(fit$eff.boot)) {
        stop(
            "No bootstrap/jackknife results available. ",
            "Choose keep.sims = TRUE in fect().",
            call. = FALSE
        )
    }

    invisible(TRUE)
}


## Internal: hard-error if ci.method is not "normal" for a jackknife fit.
## Jackknife produces an SE estimate via the Tukey pseudo-value formula,
## not a sampling distribution of θ̂. The accepted CI is Wald-style:
##   CI = θ̂ ± z_{1-α/2} · SE_jack
## (ci.method = "normal").
##
## Reflection-based methods (basic, percentile) require exchangeable draws
## from the sampling distribution — jackknife leave-one-out values are not
## exchangeable draws (Efron & Tibshirani 1993, §11; Davison & Hinkley
## 1997, §3.2.1). Bias-corrected methods (bc, bca) require a bootstrap
## distribution to compute z₀ — jackknife has no bootstrap distribution.
## Passing jackknife replicates to any of these methods would produce CIs
## with undefined (and likely poor) coverage. Hard-error forces the caller
## to be explicit about this statistical limitation.
## Internal: warn (not error) when a tail-quantile-based CI is requested on a
## bootstrap / parametric fit with fewer than 1000 replicates. The point
## estimate and SE are unaffected; only basic / percentile / bc / bca CIs read
## the empirical tail of the bootstrap distribution and need B large enough for
## the relevant order statistics to be stable. Efron 1987 §3 and DiCiccio &
## Efron 1996 §4 recommend B >= 1000 as the floor for bca; the same threshold
## suits basic / percentile (which read 5th / 195th order statistics at B=200
## vs 25th / 975th at B=1000). Warning rather than error so users can still
## explore on small B during iteration.
.check_tail_ci_replicates <- function(fit, ci.method, vartype) {
    if (vartype == "none") return(invisible(TRUE))
    if (!isTRUE(fit$vartype %in% c("bootstrap", "parametric"))) {
        return(invisible(TRUE))
    }
    if (!ci.method %in% c("basic", "percentile", "bc", "bca")) {
        return(invisible(TRUE))
    }
    if (is.null(fit$eff.boot)) return(invisible(TRUE))
    B <- dim(fit$eff.boot)[3]
    if (is.na(B) || B >= 1000) return(invisible(TRUE))
    if (!identical(Sys.getenv("TESTTHAT"), "true")) {
        warning(
            "estimand() with ci.method = \"", ci.method, "\" on a fit with ",
            "nboots = ", B, " (< 1000): tail-quantile-based CIs may have ",
            "erratic endpoints at this replicate count (Efron 1987 Section 3; ",
            "DiCiccio & Efron 1996 Section 4 recommend B >= 1000). The point ",
            "estimate and SE are unaffected. For publication-grade CIs, refit ",
            "with `fect(..., nboots = 1000)` and re-call estimand().",
            call. = FALSE
        )
    }
    invisible(TRUE)
}


.check_jackknife_ci_method <- function(ci.method) {
    if (ci.method != "normal") {
        stop(
            "ci.method = \"", ci.method, "\" is not supported for fits ",
            "produced with vartype = \"jackknife\".\n\n",
            "Jackknife inference produces a standard error (SE) estimate ",
            "via the Tukey pseudo-value formula, not a sampling distribution ",
            "of the estimator. The only statistically valid CI from jackknife ",
            "is the Wald-style interval: estimate +/- z * SE_jack ",
            "(ci.method = \"normal\").\n\n",
            "  Methods basic and percentile require exchangeable draws from ",
            "the sampling distribution of theta-hat. Jackknife leave-one-out ",
            "values are influence-function-flavored quantities, not ",
            "exchangeable draws (Efron & Tibshirani 1993, Chapter 11; ",
            "Davison & Hinkley 1997, Section 3.2.1).\n\n",
            "  Methods bc and bca require a bootstrap distribution to compute ",
            "the bias-correction z0 = Phi^-1(P(boot < estimate)). Jackknife ",
            "fits have no bootstrap distribution.\n\n",
            "Use ci.method = \"normal\" (the default for type = \"att\") or ",
            "refit with vartype = \"bootstrap\" to access all ci.methods.",
            call. = FALSE
        )
    }
    invisible(TRUE)
}


## ---------------------------------------------------------------------------
## Public: imputed_outcomes()
## ---------------------------------------------------------------------------

#' Imputed potential outcomes accessor
#'
#' Returns the cell-level imputed potential-outcome surface from a fect fit
#' as a long-form data frame. This is the rock-bottom accessor for any
#' post-hoc estimand: read off the columns and aggregate however you like.
#'
#' @param fit A fect object (output of \code{\link{fect}}).
#' @param cells Optional filter on which treated cells to return. Accepts
#'   \code{NULL} (default; all treated cells), a logical vector aligned
#'   with the unfiltered row order, or a one-sided rlang formula
#'   evaluated against the long-form data (e.g.,
#'   \code{~ event.time \%in\% 1:5 & !id \%in\% bad_ids}).
#' @param replicates Logical. \code{FALSE} (default) returns one row per
#'   treated cell with the point-estimate \code{Y0_hat}. \code{TRUE}
#'   expands by bootstrap/jackknife replicate; requires the fit to have
#'   been built with \code{keep.sims = TRUE}.
#' @param direction Either \code{"on"} (default; event time relative to
#'   treatment onset) or \code{"off"} (relative to treatment exit; only
#'   meaningful on reversal panels).
#'
#' @return A data frame with one row per (treated cell) or per
#'   (treated cell, replicate). Columns:
#'   \describe{
#'     \item{\code{id}}{unit identifier from \code{fit$id}.}
#'     \item{\code{time}}{calendar time from \code{fit$rawtime}.}
#'     \item{\code{event.time}}{event time at this cell (relative to onset
#'       or exit per \code{direction}).}
#'     \item{\code{cohort}}{first-treatment calendar time for the unit
#'       (\code{direction = "on"}) or first-exit calendar time
#'       (\code{direction = "off"}).}
#'     \item{\code{treated}}{logical; always \code{TRUE} in the returned
#'       rows (untreated cells are filtered out).}
#'     \item{\code{Y_obs}}{observed outcome at this cell.}
#'     \item{\code{Y0_hat}}{imputed counterfactual outcome
#'       (\code{Y_obs - eff}).}
#'     \item{\code{eff}}{cell-level score: contribution to the ATT
#'       estimator. For imputation estimators
#'       \code{eff = Y_obs - Y0_hat}; for doubly-robust estimators
#'       \code{eff = (Y_obs - Y0_hat) + eff_debias}.}
#'     \item{\code{eff_debias}}{debias correction; 0 for plain imputation
#'       estimators, populated for DR estimators.}
#'     \item{\code{W.agg}}{aggregation weight at this cell; 1 if the fit
#'       was built without \code{W} or \code{W.agg}.}
#'     \item{\code{replicate}}{(only when \code{replicates = TRUE})
#'       1..\code{nboots} for bootstrap, or the dropped-unit index for
#'       jackknife.}
#'   }
#'
#' @section Memory cost:
#'   With \code{replicates = TRUE} the returned data frame has
#'   \code{n_treated_cells * nboots} rows. For typical panels this is
#'   manageable; for large panels (\eqn{TT \times N \ge 50{,}000} and
#'   \eqn{nboots \ge 500}) consider filtering via \code{cells} before
#'   expansion.
#'
#' @seealso \code{\link{estimand}} for the typed dispatcher built on top
#'   of this accessor; \code{\link{fect}} for the fitting interface.
#'
#' @examples
#' \dontrun{
#' library(fect)
#' fit <- fect(Y ~ D, data = simdata, index = c("id", "time"),
#'             method = "fe", force = "two-way",
#'             se = TRUE, nboots = 200, parallel = FALSE,
#'             keep.sims = TRUE)
#'
#' ## Point-estimate long form: one row per treated cell.
#' po <- imputed_outcomes(fit)
#' head(po)
#'
#' ## Filter to first 5 event times.
#' po5 <- imputed_outcomes(fit, cells = ~ event.time \%in\% 1:5)
#'
#' ## Bootstrap replicate expansion (requires keep.sims = TRUE).
#' po_b <- imputed_outcomes(fit, replicates = TRUE)
#' nrow(po_b) == nrow(po) * 200    # one row per (cell, replicate)
#' }
#'
#' @export
imputed_outcomes <- function(fit,
                              cells = NULL,
                              replicates = FALSE,
                              direction = c("on", "off")) {

    direction <- match.arg(direction)
    .validate_po_contract(fit, require_replicates = replicates)

    Y      <- fit$Y.dat
    D      <- fit$D.dat
    eff    <- fit$eff
    eff_db <- if (is.null(fit$eff_debias)) {
                  matrix(0, nrow = nrow(eff), ncol = ncol(eff))
              } else fit$eff_debias

    TT <- nrow(Y)
    N  <- ncol(Y)

    Tev <- if (direction == "on") fit$T.on else fit$T.off
    if (is.null(Tev)) {
        stop(
            "direction = \"off\" requested, but fit$T.off is NULL ",
            "(this fit has no treatment exit). Use direction = \"on\".",
            call. = FALSE
        )
    }

    ## Treated-cell mask: cells where D = 1 (direction = "on") or where the
    ## unit has been treated at least once and is now post-exit (direction
    ## = "off"). Use D.dat == 1 for "on"; T.off non-NA flags exit-relative
    ## cells for "off".
    if (direction == "on") {
        treated_mask <- !is.na(D) & D == 1
    } else {
        treated_mask <- !is.na(Tev)
    }

    ## Cell coordinates of treated cells.
    idx <- which(treated_mask, arr.ind = TRUE)
    if (nrow(idx) == 0) {
        stop("No treated cells found for direction = \"", direction, "\".",
             call. = FALSE)
    }
    t_idx <- idx[, "row"]
    i_idx <- idx[, "col"]

    ## Cohort: calendar time of first treatment ("on") or first exit
    ## ("off") for each unit. NA for units that are never treated /
    ## never exit.
    cohort_per_unit <- if (direction == "on") {
        apply(D == 1 & !is.na(D), 2, function(d) {
            pos <- which(d)
            if (length(pos) == 0) NA_real_ else fit$rawtime[min(pos)]
        })
    } else {
        apply(Tev == 1 & !is.na(Tev), 2, function(d) {
            pos <- which(d)
            if (length(pos) == 0) NA_real_ else fit$rawtime[min(pos)]
        })
    }

    W_mat <- if (is.null(fit$W.agg)) {
                 matrix(1, nrow = TT, ncol = N)
             } else fit$W.agg

    df_point <- data.frame(
        id          = fit$id[i_idx],
        time        = fit$rawtime[t_idx],
        event.time  = Tev[cbind(t_idx, i_idx)],
        cohort      = cohort_per_unit[i_idx],
        treated     = TRUE,
        Y_obs       = Y[cbind(t_idx, i_idx)],
        Y0_hat      = Y[cbind(t_idx, i_idx)] - eff[cbind(t_idx, i_idx)],
        eff         = eff[cbind(t_idx, i_idx)],
        eff_debias  = eff_db[cbind(t_idx, i_idx)],
        W.agg       = W_mat[cbind(t_idx, i_idx)],
        stringsAsFactors = FALSE
    )

    ## Apply cells filter (point-form data) before optional replicate
    ## expansion to keep memory footprint down.
    if (!is.null(cells)) {
        df_point <- .apply_cells_filter(df_point, cells)
    }

    if (!replicates) {
        rownames(df_point) <- NULL
        return(df_point)
    }

    ## Replicate expansion: one row per (cell, b) where b indexes
    ## bootstrap or jackknife replicate. Y_obs and W.agg are constant
    ## across replicates; Y0_hat and eff vary per replicate.

    ## Jackknife guard: eff.boot has shape TT x (N-1) x N for jackknife.
    ## The second dimension is N-1, not N, so indexing eff.boot[t, i, b]
    ## with the full panel's column index i (1..N) accesses incorrect slices
    ## for i > N-1 or returns out-of-bounds values. The replicate expansion
    ## cannot be aligned to the full panel's cell coordinates under jackknife.
    if (replicates && isTRUE(fit$vartype == "jackknife")) {
        stop(
            "imputed_outcomes(replicates = TRUE) is not supported for ",
            "fits with vartype = \"jackknife\". Jackknife replicates drop one ",
            "unit per fit, so each replicate has N-1 columns rather than N; ",
            "the per-cell replicates cannot be aligned to the full panel's cell ",
            "coordinates. Use replicates = FALSE for the point-estimate surface, ",
            "or refit with vartype = \"bootstrap\" to access the replicate ",
            "expansion.",
            call. = FALSE
        )
    }

    eb <- fit$eff.boot
    nboots <- dim(eb)[3]

    n_cells <- nrow(df_point)

    ## Re-derive cell coordinates against the (potentially-filtered)
    ## df_point so we map back into the original eff.boot array.
    t_idx_f <- match(df_point$time, fit$rawtime)
    i_idx_f <- match(df_point$id,   fit$id)

    df_rep <- df_point[rep(seq_len(n_cells), each = nboots), , drop = FALSE]
    df_rep$replicate <- rep(seq_len(nboots), times = n_cells)

    eff_rep <- vapply(
        seq_len(n_cells),
        function(k) eb[t_idx_f[k], i_idx_f[k], ],
        numeric(nboots)
    )
    ## eff_rep is nboots x n_cells; flatten in (cell-major, replicate-minor)
    ## order to match df_rep's row order.
    df_rep$eff    <- as.vector(eff_rep)
    df_rep$Y0_hat <- df_rep$Y_obs - df_rep$eff

    rownames(df_rep) <- NULL
    df_rep
}


## ---------------------------------------------------------------------------
## Public: estimand()
## ---------------------------------------------------------------------------

#' Post-hoc estimand dispatcher
#'
#' Computes a post-hoc estimand from a \code{fect} fit, with bootstrap or
#' jackknife uncertainty. The \code{type} argument selects from a closed
#' enum of mathematically-defined estimands; the \code{by} argument
#' selects the grouping axis. The accessor \code{\link{imputed_outcomes}}
#' is the underlying long-form data source for any estimand the
#' dispatcher does not ship natively.
#'
#' @param fit A fect object (output of \code{\link{fect}}).
#' @param type Estimand type. One of:
#'   \describe{
#'     \item{\code{"att"}}{Per-cell mean treatment effect, aggregated
#'       per group: \eqn{\mathrm{ATT}_g = \mathrm{mean}_{(t,i)\in g, D=1}(Y_{ti} - \widehat Y_{ti}(0))}.}
#'     \item{\code{"att.cumu"}}{Cumulative ATT through each event time.
#'       Replaces \code{\link{effect}} for the unified API.}
#'     \item{\code{"aptt"}}{Average proportional treatment effect on the
#'       treated (Chen & Roth 2024 QJE):
#'       \eqn{\mathrm{APTT}_g = \mathrm{mean}_g(Y - \widehat Y(0)) / \mathrm{mean}_g(\widehat Y(0))}.
#'       Requires \code{keep.sims = TRUE} at fit time.}
#'     \item{\code{"log.att"}}{Mean log-scale treatment effect:
#'       \eqn{\mathrm{logATT}_g = \mathrm{mean}_g(\log Y - \log \widehat Y(0))}.
#'       Requires \code{keep.sims = TRUE}.}
#'   }
#' @param by Grouping axis. One of \code{"event.time"} (default;
#'   per-event-time series), \code{"cohort"}, \code{"calendar.time"},
#'   \code{"overall"} (one row), or any column name resolvable in the
#'   fit's panel data.
#' @param test Selects which subset of cells to aggregate over (v2.4.3+).
#'   \code{"none"} (default) uses standard treated post-treatment
#'   cells. \code{"placebo"} restricts to pre-treatment cells in
#'   \code{fit$placebo.period} that were masked-and-imputed during the
#'   placebo fit; produces e.g.\ a per-event-time placebo APTT series
#'   for credibility checks. Requires \code{placeboTest = TRUE} at fit
#'   time; forces \code{direction = "on"}. \code{"carryover"} is the
#'   analogous extension on the reversal side; requires
#'   \code{carryoverTest = TRUE} + a panel with reversals; forces
#'   \code{direction = "off"}. Both incompatible with
#'   \code{type = "att.cumu"}.
#' @param cells Optional filter on which treated cells to include.
#'   Accepts \code{NULL} (default; all treated cells), a logical vector,
#'   or a one-sided formula. See \code{\link{imputed_outcomes}}.
#' @param weights Aggregation-weight handling. \code{NULL} (default) uses
#'   \code{fit$W.agg} if the fit was built with \code{W} or \code{W.agg};
#'   otherwise uniform.
#' @param window Optional event-time window \code{c(L, R)}; convenience
#'   sugar for \code{cells = ~ event.time >= L & event.time <= R}.
#' @param direction Either \code{"on"} (default) or \code{"off"}; see
#'   \code{\link{imputed_outcomes}}.
#' @param vartype \code{"bootstrap"} (default), \code{"jackknife"},
#'   \code{"parametric"}, or \code{"none"}. Selects which variance method
#'   to source replicates from. The output \code{vartype} column reports
#'   the method actually used at fit time (read from \code{fit$vartype}),
#'   which may differ from this argument value if the fit was produced
#'   with a different setting --- the argument is informational and does
#'   not re-aggregate replicates.
#' @param conf.level Two-sided confidence level. Defaults to 0.95.
#' @param ci.method One of \code{"basic"} (reflected),
#'   \code{"percentile"} (raw bootstrap quantiles), \code{"bc"}
#'   (bias-corrected percentile; Efron 1987 minus the acceleration),
#'   or \code{"normal"} (Wald: \eqn{\hat\theta \pm z \cdot SE}).
#'   Default is \code{NULL}, which triggers a per-type default:
#'   \code{"att"} -> \code{"normal"} (matches what \code{fit$est.att}
#'   already uses), \code{"att.cumu"} -> \code{"percentile"} (matches
#'   what \code{att.cumu()} does internally), \code{"aptt"} ->
#'   \code{"bc"} and \code{"log.att"} -> \code{"bc"} (ratio / log
#'   estimators benefit from bias correction when the bootstrap
#'   distribution is skewed). Pass an explicit value to override.
#'
#' @return A data frame with columns \code{<by_key>}, \code{estimate},
#'   \code{se}, \code{ci.lo}, \code{ci.hi}, \code{n_cells}, and
#'   \code{vartype}. Always tidy regardless of \code{type} or \code{by}.
#'
#' @section Numerical equality with existing slots:
#'   \code{estimand(fit, "att", "event.time")} returns
#'   \code{estimate}, \code{se}, \code{ci.lo}, \code{ci.hi}
#'   byte-identical to columns \code{ATT}, \code{S.E.}, \code{CI.lower},
#'   \code{CI.upper} of \code{fit$est.att}, when default arguments are
#'   used. This invariant is asserted by package tests.
#'
#' @seealso \code{\link{imputed_outcomes}} for the underlying long-form
#'   accessor; \code{\link{fect}} for the fitting interface.
#'
#' @examples
#' \dontrun{
#' library(fect)
#' fit <- fect(Y ~ D, data = simdata, index = c("id", "time"),
#'             method = "fe", force = "two-way",
#'             se = TRUE, nboots = 200, parallel = FALSE)
#'
#' ## Default: per-event-time ATT (matches fit$est.att numerically).
#' est <- estimand(fit, "att", "event.time")
#' head(est)
#' }
#'
#' @export
estimand <- function(fit,
                     type   = c("att", "att.cumu", "aptt", "log.att"),
                     by     = c("event.time", "cohort", "calendar.time",
                                "overall"),
                     test        = c("none", "placebo", "carryover"),
                     cells       = NULL,
                     weights     = NULL,
                     window      = NULL,
                     direction   = c("on", "off"),
                     vartype     = c("bootstrap", "jackknife", "parametric", "none"),
                     conf.level  = 0.95,
                     ci.method   = NULL) {

    type      <- match.arg(type)
    test      <- match.arg(test)
    vartype   <- match.arg(vartype)

    ## test = "placebo" / "carryover": auto-set direction to the semantic
    ## default. Users should not need to remember the pairing.
    if (test == "placebo") {
        direction <- "on"
    } else if (test == "carryover") {
        direction <- "off"
    } else {
        direction <- match.arg(direction)
    }

    ## Cumulative semantics are undefined for placebo / carryover cells.
    if (test != "none" && type == "att.cumu") {
        stop("estimand(type = \"att.cumu\") is incompatible with ",
             "test = \"", test, "\". Cumulative effects are defined ",
             "relative to treatment onset; placebo and carryover cells ",
             "do not have a meaningful cumulative anchor.",
             call. = FALSE)
    }

    ## ci.method = NULL triggers per-type defaults (v2.4.2+).
    ## See statsclaw-workspace/fect/ref/v242-vartype-cimethod-design.md.
    if (is.null(ci.method)) {
        ci.method <- switch(type,
            "att"      = "normal",      ## matches what fit$est.att uses (Wald: theta +- z*SE)
            "att.cumu" = "basic",       ## reflected pivot CI (Davison-Hinkley 1997 §5.2.1; boot::boot.ci(type = "basic"))
            "aptt"     = "bca",         ## ratio: bootstrap-bias + skew -> BCa (Efron 1987)
            "log.att"  = "bca"          ## log: same rationale
        )
    }
    ci.method <- match.arg(ci.method,
                           c("basic", "percentile", "bc", "bca", "normal"))

    by_canon <- c("event.time", "cohort", "calendar.time", "overall")
    if (length(by) > 1L) {
        by <- match.arg(by, by_canon)
    } else {
        if (!is.character(by) || length(by) != 1L) {
            stop("`by` must be a single string.", call. = FALSE)
        }
        if (!(by %in% by_canon)) {
            ## Future: resolve as user column on fit's panel data.
            stop("user-column `by` not yet supported in this release; ",
                 "use one of: ",
                 paste(shQuote(by_canon), collapse = ", "),
                 ".", call. = FALSE)
        }
    }

    if (!is.null(window)) {
        if (!is.null(cells)) {
            stop("Pass either `cells` or `window`, not both. ",
                 "`window = c(L, R)` is sugar for ",
                 "`cells = ~ event.time >= L & event.time <= R`.",
                 call. = FALSE)
        }
        if (!is.numeric(window) || length(window) != 2L) {
            stop("`window` must be a numeric vector of length 2: c(L, R).",
                 call. = FALSE)
        }
        L <- window[1]; R <- window[2]
        cells <- stats::as.formula(
            sprintf("~ event.time >= %s & event.time <= %s", L, R)
        )
    }

    .validate_po_contract(fit)

    ## Jackknife ci.method guard: only "normal" is statistically valid.
    ## See .check_jackknife_ci_method() for the full rationale.
    ## Placed after .validate_po_contract() so fit is known to be a list.
    if (isTRUE(fit$vartype == "jackknife") && vartype != "none") {
        .check_jackknife_ci_method(ci.method)
    }

    ## Tail-CI under-replication warning: bc / bca / percentile / basic CIs
    ## need B >= 1000 for stable tail quantiles. Warning, not error, so users
    ## can still explore on small B during iteration.
    .check_tail_ci_replicates(fit, ci.method, vartype)

    result <- if (type == "att") {
        .estimand_att(fit, by, cells, weights, direction,
                      vartype, conf.level, ci.method, test)
    } else if (type == "att.cumu") {
        .estimand_att_cumu(fit, by, cells, weights, direction,
                           vartype, conf.level, ci.method, window)
    } else if (type == "aptt") {
        .estimand_aptt(fit, by, cells, weights, direction,
                       vartype, conf.level, ci.method, test)
    } else if (type == "log.att") {
        .estimand_log_att(fit, by, cells, weights, direction,
                          vartype, conf.level, ci.method, test)
    } else {
        stop("type = \"", type, "\" is part of the v2.4.0 API surface but ",
             "is not yet implemented at this commit. Stay tuned.",
             call. = FALSE)
    }

    ## Attach a hidden attribute so esplot() can recognize placebo/carryover
    ## frames and adapt its visual defaults (e.g., uniform color across all
    ## event-times since there's no pre-vs-post contrast in those modes).
    if (!is.null(test) && test != "none") {
        attr(result, "fect_test") <- test
    }
    result
}


## Internal: type = "att" dispatcher.
.estimand_att <- function(fit, by, cells, weights, direction,
                          vartype, conf.level, ci.method, test = "none") {

    ## Fast path: by = "event.time" + default args + direction = "on".
    ## Reads directly from fit$est.att for byte-equality with the existing
    ## fect output surface. Triggered when:
    ##   - by = "event.time"
    ##   - cells = NULL (no filter)
    ##   - weights = NULL (use fit's W.agg if any)
    ##   - direction = "on"
    ##   - conf.level = 0.95 (fit$est.att uses 95% by default)
    ##   - ci.method = "normal" (fit$est.att uses Wald: theta +- z*SE)
    ##                          --- this is the v2.4.2 default for type="att"
    is_fast_path <- by == "event.time" &&
                    is.null(cells) &&
                    is.null(weights) &&
                    direction == "on" &&
                    abs(conf.level - 0.95) < 1e-12 &&
                    ci.method == "normal" &&
                    test == "none"

    if (is_fast_path) {
        return(.estimand_att_fast_event_time(fit))
    }

    if (by == "event.time" && test != "none") {
        return(.estimand_att_event_time(fit, cells, weights, direction,
                                        vartype, conf.level, ci.method,
                                        test))
    }

    if (by == "overall") {
        return(.estimand_att_overall(fit, cells, weights, direction,
                                     vartype, conf.level, ci.method,
                                     test))
    }

    stop("estimand(type = \"att\") with by = \"", by, "\" is part of ",
         "the v2.4.0 API surface but is not yet implemented at this ",
         "commit. Default form estimand(fit, \"att\", \"event.time\") ",
         "and estimand(fit, \"att\", \"overall\", window = ...) work.",
         call. = FALSE)
}


## Per-event-time ATT slow path. Used when test = "placebo" /
## "carryover" forces per-cell aggregation.
.estimand_att_event_time <- function(fit, cells, weights, direction,
                                     vartype, conf.level, ci.method,
                                     test) {

    if (!is.null(weights)) {
        stop("estimand(\"att\", \"event.time\", test = \"", test, "\") ",
             "with non-default weights is not yet supported.",
             call. = FALSE)
    }
    if (!is.null(cells)) {
        stop("estimand(\"att\", \"event.time\", test = \"", test, "\") ",
             "with `cells` filter is not yet supported.",
             call. = FALSE)
    }

    ## Jackknife dispatch: eff.boot has shape TT x (N-1) x N; direct
    ## cell-masking with a TT x N mask is incompatible. Route through
    ## the dedicated jackknife helper.
    if (isTRUE(fit$vartype == "jackknife")) {
        return(.estimand_att_event_time_jackknife(fit, cells, weights,
                                                  direction, conf.level,
                                                  vartype, ci.method, test))
    }

    mask_info <- .test_cell_mask(fit, test, direction)
    base_mask <- mask_info$mask
    Tev       <- mask_info$Tev

    ets <- sort(unique(Tev[base_mask]))
    if (length(ets) == 0) {
        stop("No cells satisfy test = \"", test, "\".", call. = FALSE)
    }

    nboots <- if (is.null(fit$eff.boot)) 0L else dim(fit$eff.boot)[3]

    estimate <- numeric(length(ets))
    se_vec   <- rep(NA_real_, length(ets))
    ci_lo    <- rep(NA_real_, length(ets))
    ci_hi    <- rep(NA_real_, length(ets))
    n_cells  <- integer(length(ets))

    for (k in seq_along(ets)) {
        et <- ets[k]
        cell_mask <- base_mask & Tev == et
        n_cells[k] <- sum(cell_mask)

        eff_t <- fit$eff[cell_mask]
        estimate[k] <- mean(eff_t, na.rm = TRUE)

        if (nboots > 0L && vartype != "none") {
            eff_boot_cells <- apply(fit$eff.boot, 3,
                                    function(eb) eb[cell_mask])
            if (!is.matrix(eff_boot_cells)) {
                eff_boot_cells <- matrix(eff_boot_cells,
                                         nrow = sum(cell_mask))
            }
            att_b <- colMeans(eff_boot_cells, na.rm = TRUE)

            ## PARAMETRIC SHIFT (v2.4.2 fix): center att_b at the point estimate.
            ## For parametric vartype, eff.boot is H0-centered; only ci.method="normal"
            ## uses sd() and is unaffected by centering. All other ci.methods require
            ## the distribution to be centered near estimate. The shift preserves sd().
            is_parametric <- isTRUE(fit$vartype == "parametric") ||
                             isTRUE(vartype == "parametric")
            if (is_parametric) {
                att_b <- att_b - mean(att_b, na.rm = TRUE) + estimate[k]
            }

            jack_v <- if (ci.method == "bca") {
                .cell_jackknife("att", eff = eff_t)
            } else NULL

            ci <- .compute_ci(estimate[k], att_b, ci.method, conf.level,
                              jack = jack_v)
            se_vec[k] <- ci$se
            ci_lo[k]  <- ci$ci.lo
            ci_hi[k]  <- ci$ci.hi
        }
    }

    used_vartype <- if (vartype == "none") "none"
                    else if (is.null(fit$vartype)) "bootstrap"
                    else fit$vartype

    data.frame(
        event.time = ets,
        estimate   = estimate,
        se         = se_vec,
        ci.lo      = ci_lo,
        ci.hi      = ci_hi,
        n_cells    = n_cells,
        vartype    = used_vartype,
        stringsAsFactors = FALSE
    )
}


## Per-event-time ATT slow path for jackknife fits.
## Called from .estimand_att_event_time() when fit$vartype == "jackknife".
## Uses the same column-drop masking as .estimand_att_overall_jackknife()
## to handle the TT x (N-1) x N shape of eff.boot.
.estimand_att_event_time_jackknife <- function(fit, cells, weights,
                                                direction, conf.level,
                                                vartype, ci.method, test) {

    mask_info <- .test_cell_mask(fit, test, direction)
    base_mask <- mask_info$mask
    Tev       <- mask_info$Tev

    ets <- sort(unique(Tev[base_mask]))
    if (length(ets) == 0) {
        stop("No cells satisfy test = \"", test, "\".", call. = FALSE)
    }

    nboots <- if (is.null(fit$eff.boot)) 0L else dim(fit$eff.boot)[3]

    estimate <- numeric(length(ets))
    se_vec   <- rep(NA_real_, length(ets))
    ci_lo    <- rep(NA_real_, length(ets))
    ci_hi    <- rep(NA_real_, length(ets))
    n_cells  <- integer(length(ets))

    for (k in seq_along(ets)) {
        et <- ets[k]
        cell_mask <- base_mask & Tev == et
        n_cells[k] <- sum(cell_mask)

        eff_t <- fit$eff[cell_mask]
        estimate[k] <- mean(eff_t, na.rm = TRUE)

        ## Jackknife: per-replicate leave-one-out mean over cells.
        if (nboots > 0L && vartype != "none") {
            theta_j <- numeric(nboots)
            for (j in seq_len(nboots)) {
                cm_j <- cell_mask[, -j, drop = FALSE]
                eb_j <- fit$eff.boot[, , j]
                theta_j[j] <- mean(eb_j[cm_j], na.rm = TRUE)
            }
            theta_valid <- theta_j[is.finite(theta_j)]
            N_eff <- length(theta_valid)
            if (N_eff >= 2L) {
                pseudo     <- nboots * estimate[k] - (nboots - 1) * theta_valid
                se_vec[k]  <- sqrt(var(pseudo) / N_eff)
                z_q        <- stats::qnorm(1 - (1 - conf.level) / 2)
                ci_lo[k]   <- estimate[k] - z_q * se_vec[k]
                ci_hi[k]   <- estimate[k] + z_q * se_vec[k]
            }
        }
    }

    data.frame(
        event.time = ets,
        estimate   = estimate,
        se         = se_vec,
        ci.lo      = ci_lo,
        ci.hi      = ci_hi,
        n_cells    = n_cells,
        vartype    = "jackknife",
        stringsAsFactors = FALSE
    )
}


## Compute overall ATT (single scalar) over treated cells, optionally
## filtered. Reads from fit$eff and fit$D.dat directly; bootstrap from
## fit$eff.boot if available, else delegates to the pre-aggregated
## fit$att.avg.boot when no cells filter is active.
.estimand_att_overall <- function(fit, cells, weights, direction,
                                  vartype, conf.level, ci.method,
                                  test = "none") {

    if (!is.null(weights)) {
        stop("estimand(\"att\", \"overall\") with non-default weights ",
             "is not yet supported in v2.4.0.", call. = FALSE)
    }
    if (test != "none" && !is.null(cells)) {
        stop("estimand(\"att\", \"overall\") with both test = \"", test,
             "\" and `cells` is not supported. The test = ... argument ",
             "already filters cells to the placebo / carryover window.",
             call. = FALSE)
    }

    mask_info <- .test_cell_mask(fit, test, direction)
    treated_mask <- mask_info$mask
    Tev          <- mask_info$Tev

    ## Apply cells filter at the event-time / id level, not via long-form
    ## conversion (faster). Build a per-cell mask matching shape(eff).
    if (!is.null(cells)) {
        po <- imputed_outcomes(fit, direction = direction)
        po_filtered <- .apply_cells_filter(po, cells)
        ## Reconstruct the cell-level mask from the filtered po (id, time).
        keep_idx <- match(
            paste(po_filtered$id, po_filtered$time, sep = "::"),
            paste(fit$id[col(fit$D.dat)], fit$rawtime[row(fit$D.dat)],
                  sep = "::")
        )
        keep_mask <- matrix(FALSE, nrow = nrow(fit$D.dat),
                            ncol = ncol(fit$D.dat))
        keep_mask[keep_idx] <- TRUE
        cell_mask <- treated_mask & keep_mask
    } else {
        cell_mask <- treated_mask
    }

    n_cells <- sum(cell_mask)
    if (n_cells == 0L) {
        stop("No treated cells satisfy the requested filter.",
             call. = FALSE)
    }

    estimate <- mean(fit$eff[cell_mask], na.rm = TRUE)

    se_val <- NA_real_
    ci_lo  <- NA_real_
    ci_hi  <- NA_real_

    if (vartype != "none") {
        if (is.null(fit$eff.boot)) {
            stop("estimand(\"att\", \"overall\") with non-default ",
                 "filter requires keep.sims = TRUE in fect() so the ",
                 "per-cell bootstrap surface is available.",
                 call. = FALSE)
        }

        ## Jackknife: eff.boot has shape TT x (N-1) x N. Each slice eff.boot[,,j]
        ## is the effect matrix for the fit that excluded unit j; it has N-1 columns
        ## mapping to the N-1 retained units. We cannot apply cell_mask (TT x N)
        ## directly to eff.boot[,,j] (TT x (N-1)). Instead, reroute through the
        ## jackknife-specific helper.
        if (isTRUE(fit$vartype == "jackknife")) {
            return(.estimand_att_overall_jackknife(fit, cell_mask, conf.level,
                                                   vartype))
        }

        nboots <- dim(fit$eff.boot)[3]
        att_b  <- vapply(seq_len(nboots), function(b) {
            mean(fit$eff.boot[, , b][cell_mask], na.rm = TRUE)
        }, numeric(1))

        ## PARAMETRIC SHIFT (v2.4.2 fix)
        is_parametric <- isTRUE(fit$vartype == "parametric") ||
                         isTRUE(vartype == "parametric")
        if (is_parametric) {
            att_b <- att_b - mean(att_b, na.rm = TRUE) + estimate
        }

        jack_v <- if (ci.method == "bca") {
            .cell_jackknife("att", eff = fit$eff[cell_mask])
        } else NULL

        ci <- .compute_ci(estimate, att_b, ci.method, conf.level,
                          jack = jack_v)
        se_val <- ci$se
        ci_lo  <- ci$ci.lo
        ci_hi  <- ci$ci.hi
    }

    used_vartype <- if (vartype == "none") "none"
                    else if (is.null(fit$vartype)) "bootstrap"
                    else fit$vartype

    data.frame(
        estimate = estimate,
        se       = se_val,
        ci.lo    = ci_lo,
        ci.hi    = ci_hi,
        n_cells  = as.integer(n_cells),
        vartype  = used_vartype,
        stringsAsFactors = FALSE
    )
}


## Compute overall ATT under jackknife using the Tukey SE from the
## per-event-time att.boot matrix.
##
## Strategy: eff.boot is TT x (N-1) x N for jackknife. Direct cell-masking
## is not applicable because each slice drops one unit column. Instead, use
## the pre-aggregated att.avg.unit.boot (per-unit leave-one-out ATT averages,
## 1 x N) to compute N scalar overall-ATT leave-one-out estimates, then apply
## the Tukey SE formula.
##
## When a cells filter is active, fall back to per-replicate recomputation
## from eff.boot by mapping each replicate's reduced column space back to
## the cell mask of the retained units.
##
## Arguments:
##   fit        - fect object with vartype == "jackknife"
##   cell_mask  - logical TT x N matrix; TRUE at treated cells to include
##   conf.level - numeric confidence level (default 0.95)
##   vartype    - "jackknife" or "none"
##
## Returns a one-row data frame with estimate, se, ci.lo, ci.hi, n_cells,
## vartype (same schema as .estimand_att_overall()).
.estimand_att_overall_jackknife <- function(fit, cell_mask, conf.level,
                                             vartype) {
    if (vartype == "none") {
        ## No SE requested.
        estimate <- mean(fit$eff[cell_mask], na.rm = TRUE)
        return(data.frame(
            estimate = estimate, se = NA_real_,
            ci.lo = NA_real_, ci.hi = NA_real_,
            n_cells = as.integer(sum(cell_mask)),
            vartype = "none",
            stringsAsFactors = FALSE
        ))
    }

    estimate <- mean(fit$eff[cell_mask], na.rm = TRUE)
    n_cells  <- sum(cell_mask)

    ## Determine whether the cell_mask is "all treated cells" (no filter
    ## beyond the standard treated-cell mask). In that case we can use the
    ## pre-aggregated att.avg.unit.boot slot (1 x N matrix) for the
    ## Tukey SE rather than recomputing from eff.boot.
    ## att.avg.unit.boot is a unit-weighted grand average of all post-treatment
    ## cells, which equals mean(eff[treated_mask]) when there is no
    ## additional filter.
    all_treated_mask <- !is.na(fit$D.dat) & fit$D.dat == 1 &
                        !is.na(fit$T.on)
    use_precomp <- identical(cell_mask, all_treated_mask) &&
                   !is.null(fit$att.avg.unit.boot)

    if (use_precomp) {
        ## att.avg.unit.boot: 1 x N matrix (one row; N columns = jackknife
        ## replicates). jackknifed() formula:
        ##   pseudo_j = N * theta_hat - (N-1) * theta_j
        ##   SE = sqrt(var(pseudo_j) / N)
        theta_j <- as.vector(fit$att.avg.unit.boot)
        N_jack  <- length(theta_j)
        theta_j_valid <- theta_j[!is.na(theta_j)]
        N_eff   <- length(theta_j_valid)
        if (N_eff < 2L) {
            return(data.frame(
                estimate = estimate, se = NA_real_,
                ci.lo = NA_real_, ci.hi = NA_real_,
                n_cells = as.integer(n_cells),
                vartype = "jackknife",
                stringsAsFactors = FALSE
            ))
        }
        pseudo  <- N_jack * estimate - (N_jack - 1) * theta_j_valid
        se_val  <- sqrt(var(pseudo) / N_eff)
    } else {
        ## Cells filter is active or att.avg.unit.boot is unavailable.
        ## Recompute per-replicate overall ATT from eff.boot.
        ##
        ## eff.boot[, , j] is TT x (N-1): the effect matrix for the panel
        ## with unit j removed. The column ordering of eff.boot[,,j] matches
        ## fit$id[-j] (all units except unit j, in original order).
        ##
        ## For each replicate j:
        ##   1. Build the corresponding cell_mask_j (TT x (N-1)) by dropping
        ##      column j from cell_mask.
        ##   2. Compute theta_j = mean(eff.boot[,,j][cell_mask_j]).
        ##
        ## This is correct because cell_mask[t, j] is FALSE for unit j at any
        ## time (unit j is only in the full panel, not in a leave-one-out fit
        ## that excludes j). Dropping column j from cell_mask gives the mask
        ## applicable to eff.boot[,,j].
        N_jack <- dim(fit$eff.boot)[3]
        theta_j <- numeric(N_jack)
        for (j in seq_len(N_jack)) {
            ## Drop column j from cell_mask.
            cm_j <- cell_mask[, -j, drop = FALSE]
            eb_j <- fit$eff.boot[, , j]  ## TT x (N-1)
            theta_j[j] <- mean(eb_j[cm_j], na.rm = TRUE)
        }
        theta_j_valid <- theta_j[is.finite(theta_j)]
        N_eff <- length(theta_j_valid)
        if (N_eff < 2L) {
            return(data.frame(
                estimate = estimate, se = NA_real_,
                ci.lo = NA_real_, ci.hi = NA_real_,
                n_cells = as.integer(n_cells),
                vartype = "jackknife",
                stringsAsFactors = FALSE
            ))
        }
        pseudo <- N_jack * estimate - (N_jack - 1) * theta_j_valid
        se_val <- sqrt(var(pseudo) / N_eff)
    }

    z     <- stats::qnorm(1 - (1 - conf.level) / 2)
    ci_lo <- estimate - z * se_val
    ci_hi <- estimate + z * se_val

    data.frame(
        estimate = estimate,
        se       = se_val,
        ci.lo    = ci_lo,
        ci.hi    = ci_hi,
        n_cells  = as.integer(n_cells),
        vartype  = "jackknife",
        stringsAsFactors = FALSE
    )
}


## Fast path implementation: byte-equality with fit$est.att. No
## recomputation; just reshapes the existing matrix into the tidy
## return schema.
.estimand_att_fast_event_time <- function(fit) {

    if (is.null(fit$est.att)) {
        stop("fit$est.att is missing; refit with se = TRUE for SE/CI ",
             "or use vartype = \"none\".",
             call. = FALSE)
    }

    ## Detect vartype actually used at fit time. fect populates a
    ## `vartype` element on the fit object; default to "bootstrap" if
    ## absent.
    used_vartype <- if (is.null(fit$vartype)) "bootstrap" else fit$vartype

    out <- data.frame(
        event.time = as.numeric(rownames(fit$est.att)),
        estimate   = unname(fit$est.att[, "ATT"]),
        se         = unname(fit$est.att[, "S.E."]),
        ci.lo      = unname(fit$est.att[, "CI.lower"]),
        ci.hi      = unname(fit$est.att[, "CI.upper"]),
        n_cells    = unname(fit$est.att[, "count"]),
        vartype    = used_vartype,
        stringsAsFactors = FALSE
    )
    rownames(out) <- NULL
    out
}


## Internal: type = "att.cumu" dispatcher.
.estimand_att_cumu <- function(fit, by, cells, weights, direction,
                               vartype, conf.level, ci.method, window) {

    if (direction != "on") {
        stop("estimand(\"att.cumu\") with direction = \"off\" is not ",
             "supported (cumulative effects are defined relative to ",
             "treatment onset).", call. = FALSE)
    }
    if (!is.null(weights)) {
        stop("estimand(\"att.cumu\") with non-default weights is not ",
             "yet supported in v2.4.0.", call. = FALSE)
    }
    if (!is.null(cells) && by != "overall") {
        stop("estimand(\"att.cumu\") with `cells` is supported only when ",
             "by = \"overall\". For event-time series, use ",
             "by = \"event.time\" with no cells filter.",
             call. = FALSE)
    }

    ## Implementation strategy: derive the cumulative bootstrap
    ## distribution from fit$att.boot (already pre-aggregated per event
    ## time) for the by = "event.time" and by = "overall" paths. These
    ## paths do not require keep.sims = TRUE. Cohort / calendar-time /
    ## user-column paths (deferred) need per-cell eff.boot.

    if (by == "event.time") {
        ## Event-time cumulative series: matches effect(fit)$catt.
        return(.compute_att_cumu_event_time(fit, conf.level, ci.method,
                                            vartype))
    }

    if (by == "overall") {
        ## Single scalar: matches the final row of att.cumu(fit, period).
        if (is.null(window)) {
            ## Default to the maximum post-treatment window.
            available_post <- fit$time[fit$time >= 1]
            if (length(available_post) == 0) {
                stop("No post-treatment event times in fit$time.",
                     call. = FALSE)
            }
            window <- c(min(available_post), max(available_post))
        }
        return(.compute_att_cumu_overall(fit, window, conf.level,
                                         ci.method, vartype))
    }

    stop("estimand(\"att.cumu\") with by = \"", by, "\" is not yet ",
         "supported in v2.4.0. Use by = \"event.time\" or ",
         "by = \"overall\".", call. = FALSE)
}


## Compute event-time cumulative ATT series. Delegates to the existing
## effect() function for the actual numerics (which use a per-cell mean
## approach, not a simple cumsum of fit$att). Reshapes the matrix into
## the tidy estimand return schema. Numerical equality with effect() is
## by construction.
.compute_att_cumu_event_time <- function(fit, conf.level, ci.method,
                                          vartype) {

    if (isTRUE(fit$hasRevs == 1)) {
        stop("Cumulative effects are not well-defined for panels with ",
             "treatment reversals.", call. = FALSE)
    }

    if (is.null(fit$eff.boot)) {
        stop("No bootstrap/jackknife results available. ",
             "Choose keep.sims = TRUE in fect().",
             call. = FALSE)
    }

    out_eff <- .with_no_deprecation(effect(fit, cumu = TRUE, plot = FALSE))
    if (is.null(out_eff) || is.null(out_eff$effect.est.att)) {
        stop("Cumulative effects unavailable for this fit.",
             call. = FALSE)
    }

    M <- out_eff$effect.est.att
    counts <- fit$est.att[match(rownames(M), rownames(fit$est.att)), "count"]
    data.frame(
        event.time = as.numeric(rownames(M)),
        estimate   = unname(M[, "ATT"]),
        se         = unname(M[, "S.E."]),
        ci.lo      = unname(M[, "CI.lower"]),
        ci.hi      = unname(M[, "CI.upper"]),
        n_cells    = as.integer(unname(counts)),
        vartype    = if (is.null(fit$vartype)) "bootstrap" else fit$vartype,
        stringsAsFactors = FALSE
    )
}


## Compute single overall cumulative ATT in a window. Delegates to the
## existing att.cumu() function for the canonical math (count-weighted
## across event times, percentile bootstrap CI). Returns the final row
## reshaped into the tidy estimand schema.
.compute_att_cumu_overall <- function(fit, window, conf.level, ci.method,
                                       vartype) {

    if (isTRUE(fit$hasRevs == 1)) {
        stop("Cumulative effects are not well-defined for panels with ",
             "treatment reversals.", call. = FALSE)
    }

    out_acc <- .with_no_deprecation(att.cumu(fit, period = window,
                                              plot = FALSE))
    final <- out_acc[nrow(out_acc), ]

    ## att.cumu() column name is "catt" when SE is present;
    ## "cumulative att" when no SE. Pick whichever is there.
    catt_col <- if ("catt" %in% names(final)) "catt" else "cumulative att"

    has_se <- "S.E." %in% names(final)

    et <- as.numeric(rownames(fit$est.att))
    in_window <- !is.na(et) & et >= window[1] & et <= window[2]
    n_cells_window <- sum(fit$est.att[in_window, "count"], na.rm = TRUE)

    data.frame(
        estimate = unname(final[catt_col]),
        se       = if (has_se) unname(final["S.E."])     else NA_real_,
        ci.lo    = if (has_se) unname(final["CI.lower"]) else NA_real_,
        ci.hi    = if (has_se) unname(final["CI.upper"]) else NA_real_,
        n_cells  = as.integer(n_cells_window),
        vartype  = if (is.null(fit$vartype)) "bootstrap" else fit$vartype,
        stringsAsFactors = FALSE
    )
}


## Internal: type = "aptt" dispatcher.
##
## APTT (Chen & Roth 2024 QJE):
##   APTT_g = mean_g(Y_obs - Y0_hat) / mean_g(Y0_hat)
## per group g (event time, cohort, etc.). Numerator and denominator are
## both computed per replicate and the ratio is taken inside each
## replicate, so the bootstrap distribution is the distribution of
## ratios, not the ratio of mean distributions.
.estimand_aptt <- function(fit, by, cells, weights, direction,
                           vartype, conf.level, ci.method, test = "none") {

    if (!is.null(weights)) {
        stop("estimand(\"aptt\") with non-default weights is not yet ",
             "supported in v2.4.0.", call. = FALSE)
    }
    if (!is.null(cells)) {
        stop("estimand(\"aptt\") with `cells` filter is not yet ",
             "supported in v2.4.0.", call. = FALSE)
    }
    if (is.null(fit$eff.boot) && vartype != "none") {
        stop("No bootstrap/jackknife results available. ",
             "Choose keep.sims = TRUE in fect().",
             call. = FALSE)
    }

    if (by == "event.time") {
        return(.compute_aptt_event_time(fit, conf.level, ci.method,
                                        vartype, direction, test))
    }

    stop("estimand(\"aptt\") with by = \"", by, "\" is not yet ",
         "supported in v2.4.0. Use by = \"event.time\".",
         call. = FALSE)
}


## Compute per-event-time APTT with bootstrap CI.
.compute_aptt_event_time <- function(fit, conf.level, ci.method, vartype,
                                      direction, test = "none") {

    mask_info <- .test_cell_mask(fit, test, direction)
    treated_mask <- mask_info$mask
    Tev          <- mask_info$Tev

    ets <- sort(unique(Tev[treated_mask]))
    if (length(ets) == 0) {
        stop("No cells satisfy test = \"", test, "\".", call. = FALSE)
    }

    nboots <- if (is.null(fit$eff.boot)) 0L else dim(fit$eff.boot)[3]

    estimate <- numeric(length(ets))
    se_vec   <- rep(NA_real_, length(ets))
    ci_lo    <- rep(NA_real_, length(ets))
    ci_hi    <- rep(NA_real_, length(ets))
    n_cells  <- integer(length(ets))

    alpha <- 1 - conf.level
    probs <- c(alpha / 2, 1 - alpha / 2)

    for (k in seq_along(ets)) {
        et <- ets[k]
        cell_mask <- treated_mask & Tev == et
        n_cells[k] <- sum(cell_mask)

        ## Point estimate.
        eff_t <- fit$eff[cell_mask]
        Y_t   <- fit$Y.dat[cell_mask]
        Y0_t  <- Y_t - eff_t
        num   <- mean(eff_t, na.rm = TRUE)
        den   <- mean(Y0_t, na.rm = TRUE)
        estimate[k] <- num / den

        ## Bootstrap / jackknife distribution per replicate.
        if (nboots > 0L && vartype != "none") {
            if (isTRUE(fit$vartype == "jackknife")) {
                ## Jackknife branch: column-drop masking for TT x (N-1) x N eff.boot.
                ## For each replicate j (unit j dropped), extract the surviving cells
                ## and compute per-replicate APTT.
                keep_cols <- seq_len(ncol(fit$D.dat))
                theta_j <- numeric(nboots)
                for (j in seq_len(nboots)) {
                    ## Build TT x (N-1) masks aligned with eff.boot[,,j].
                    kcols <- keep_cols[-j]
                    cm_j_full <- cell_mask[, kcols, drop = FALSE]
                    eb_j      <- fit$eff.boot[, , j]  ## TT x (N-1)
                    eff_j_v   <- eb_j[cm_j_full]
                    Y_j       <- fit$Y.dat[, kcols, drop = FALSE][cm_j_full]
                    Y0_j_v    <- Y_j - eff_j_v
                    denom_j   <- mean(Y0_j_v, na.rm = TRUE)
                    if (is.finite(denom_j) && abs(denom_j) > 1e-10) {
                        theta_j[j] <- mean(eff_j_v, na.rm = TRUE) / denom_j
                    } else {
                        theta_j[j] <- NA_real_
                    }
                }
                theta_valid <- theta_j[is.finite(theta_j)]
                N_eff <- length(theta_valid)
                if (N_eff >= 2L) {
                    pseudo    <- nboots * estimate[k] - (nboots - 1) * theta_valid
                    se_vec[k] <- sqrt(var(pseudo) / N_eff)
                    z_q       <- stats::qnorm(1 - (1 - conf.level) / 2)
                    ci_lo[k]  <- estimate[k] - z_q * se_vec[k]
                    ci_hi[k]  <- estimate[k] + z_q * se_vec[k]
                }
            } else {
                ## Bootstrap / wild / parametric branch (original code).
                ## eff_boot_cells: rows = cells in this group, cols = replicates.
                eff_boot_cells <- apply(fit$eff.boot, 3, function(eb) eb[cell_mask])
                if (!is.matrix(eff_boot_cells)) {
                    eff_boot_cells <- matrix(eff_boot_cells,
                                             nrow = sum(cell_mask))
                }
                Y0_boot <- Y_t - eff_boot_cells

                ## Hard-error on cell-drop pathology (v2.4.2+).
                ## See .compute_log_att_event_time for the full rationale.
                ## For APTT specifically: when E(Y0_b) crosses zero in any
                ## replicate, the denominator blows up (or flips sign),
                ## producing wildly unstable per-replicate APTT values.
                mean_Y0_per_rep <- colMeans(Y0_boot, na.rm = TRUE)
                n_bad_reps <- sum(abs(mean_Y0_per_rep) < 1e-10 |
                                  is.na(mean_Y0_per_rep))
                if (n_bad_reps > 0L) {
                    pct_bad <- 100 * n_bad_reps / length(mean_Y0_per_rep)
                    stop(sprintf(
                        "APTT bootstrap is unreliable at event time %s\n  (E(Y0_hat) at the point = %.4f, but %d of %d bootstrap replicates\n  (%.1f%%) have E(Y0_b) ~ 0, blowing up the APTT denominator and\n  producing wildly unstable per-replicate ratios).\n\n  Options:\n    1. Filter out cells where E(Y0_hat) is small relative to E(Y):\n         estimand(fit, \"aptt\", \"event.time\",\n                  cells = ~ abs(Y0_hat) > <threshold>)\n    2. Transform the outcome to keep Y0_hat away from zero\n    3. Use a different estimand: estimand(\"att\", ...) does not have\n       this denominator instability",
                        as.character(et), den, n_bad_reps,
                        length(mean_Y0_per_rep), pct_bad
                    ), call. = FALSE)
                }

                aptt_b  <- colMeans(eff_boot_cells, na.rm = TRUE) /
                           mean_Y0_per_rep

                ## PARAMETRIC SHIFT (v2.4.2 fix)
                is_parametric <- isTRUE(fit$vartype == "parametric") ||
                                 isTRUE(vartype == "parametric")
                if (is_parametric) {
                    aptt_b <- aptt_b - mean(aptt_b, na.rm = TRUE) + estimate[k]
                }

                ## Cell-level jackknife for the BCa acceleration parameter.
                ## Only computed when bca is requested (cheap; no model refits).
                jack_v <- if (ci.method == "bca") {
                    .cell_jackknife("aptt", eff = eff_t, Y0 = Y0_t)
                } else NULL

                ci <- .compute_ci(estimate[k], aptt_b, ci.method, conf.level,
                                  jack = jack_v)
                se_vec[k] <- ci$se
                ci_lo[k]  <- ci$ci.lo
                ci_hi[k]  <- ci$ci.hi
            }
        }
    }

    used_vartype <- if (vartype == "none") "none"
                    else if (is.null(fit$vartype)) "bootstrap"
                    else fit$vartype

    data.frame(
        event.time = ets,
        estimate   = estimate,
        se         = se_vec,
        ci.lo      = ci_lo,
        ci.hi      = ci_hi,
        n_cells    = n_cells,
        vartype    = used_vartype,
        stringsAsFactors = FALSE
    )
}


## Internal: type = "log.att" dispatcher.
##
## logATT_g = mean_g(log(Y_obs) - log(Y0_hat)) over treated cells.
## Hard-stops if any cell included in the aggregation has Y_obs <= 0 or
## Y0_hat <= 0 (log undefined). Caller must pre-transform the outcome
## (e.g. log(Y + c) inside fect) so that all imputed and observed
## outcomes are strictly positive before requesting log.att.
.estimand_log_att <- function(fit, by, cells, weights, direction,
                              vartype, conf.level, ci.method,
                              test = "none") {

    if (!is.null(weights)) {
        stop("estimand(\"log.att\") with non-default weights is not ",
             "yet supported in v2.4.0.", call. = FALSE)
    }
    if (!is.null(cells)) {
        stop("estimand(\"log.att\") with `cells` filter is not yet ",
             "supported in v2.4.0.", call. = FALSE)
    }
    if (is.null(fit$eff.boot) && vartype != "none") {
        stop("No bootstrap/jackknife results available. ",
             "Choose keep.sims = TRUE in fect().",
             call. = FALSE)
    }

    ## Hard-stop on Y <= 0 / Y0_hat <= 0 at the point-estimate level.
    ## log.att is mathematically undefined for non-positive imputed or
    ## observed outcomes; silently dropping cells contaminates both the
    ## point estimate and the bootstrap distribution. Force the caller
    ## to pre-transform (typical fix: log(Y + c) for some c > 0 chosen
    ## so all Y0_hat > 0; refit; then re-call estimand).
    mask_info  <- .test_cell_mask(fit, test, direction)
    cells_mask <- mask_info$mask
    Y_chk  <- fit$Y.dat[cells_mask]
    Y0_chk <- Y_chk - fit$eff[cells_mask]
    n_bad_Y  <- sum(!is.na(Y_chk)  & Y_chk  <= 0)
    n_bad_Y0 <- sum(!is.na(Y0_chk) & Y0_chk <= 0)
    if (n_bad_Y + n_bad_Y0 > 0L) {
        min_Y  <- if (n_bad_Y  > 0L) min(Y_chk[!is.na(Y_chk)],   na.rm = TRUE) else NA_real_
        min_Y0 <- if (n_bad_Y0 > 0L) min(Y0_chk[!is.na(Y0_chk)], na.rm = TRUE) else NA_real_
        stop(sprintf(
            "log.att requires Y > 0 and Y0_hat > 0 in all treated cells.\n  Found %d cell(s) with Y <= 0 (min Y = %s) and %d cell(s) with Y0_hat <= 0 (min Y0_hat = %s).\n  log(Y) and log(Y0_hat) are undefined; silent dropping would bias the point estimate.\n\n  Fix: refit fect on a strictly-positive outcome, e.g.\n      data$Y_pos <- data$Y + (abs(min(data$Y)) + 1)\n      fit <- fect(Y_pos ~ D, ...)\n      estimand(fit, \"log.att\", \"event.time\")\n  Then back out the original-scale interpretation as needed.",
            n_bad_Y,  if (is.na(min_Y))  "NA" else sprintf("%.4f", min_Y),
            n_bad_Y0, if (is.na(min_Y0)) "NA" else sprintf("%.4f", min_Y0)
        ), call. = FALSE)
    }

    if (by == "event.time") {
        return(.compute_log_att_event_time(fit, conf.level, ci.method,
                                           vartype, direction, test))
    }

    stop("estimand(\"log.att\") with by = \"", by, "\" is not yet ",
         "supported in v2.4.0. Use by = \"event.time\".",
         call. = FALSE)
}


## Compute per-event-time log-ATT. Drops cells where either Y_obs or
## Y0_hat is non-positive (would give -Inf or NaN under log).
.compute_log_att_event_time <- function(fit, conf.level, ci.method,
                                         vartype, direction,
                                         test = "none") {

    mask_info <- .test_cell_mask(fit, test, direction)
    treated_mask <- mask_info$mask
    Tev          <- mask_info$Tev

    ets <- sort(unique(Tev[treated_mask]))
    if (length(ets) == 0) {
        stop("No cells satisfy test = \"", test, "\".", call. = FALSE)
    }

    nboots <- if (is.null(fit$eff.boot)) 0L else dim(fit$eff.boot)[3]

    estimate <- numeric(length(ets))
    se_vec   <- rep(NA_real_, length(ets))
    ci_lo    <- rep(NA_real_, length(ets))
    ci_hi    <- rep(NA_real_, length(ets))
    n_cells  <- integer(length(ets))

    alpha <- 1 - conf.level
    probs <- c(alpha / 2, 1 - alpha / 2)

    for (k in seq_along(ets)) {
        et <- ets[k]
        cell_mask <- treated_mask & Tev == et

        eff_t <- fit$eff[cell_mask]
        Y_t   <- fit$Y.dat[cell_mask]
        Y0_t  <- Y_t - eff_t

        ## Caller-level hard-stop in .estimand_log_att already guarantees
        ## all cells have Y > 0 and Y0_hat > 0; only NA filtering needed.
        ok <- !is.na(Y_t) & !is.na(Y0_t)
        n_cells[k] <- sum(ok)

        if (sum(ok) == 0L) {
            estimate[k] <- NA_real_
            next
        }

        log_diff <- log(Y_t[ok]) - log(Y0_t[ok])
        estimate[k] <- mean(log_diff, na.rm = TRUE)

        if (nboots > 0L && vartype != "none") {
            if (isTRUE(fit$vartype == "jackknife")) {
                ## Jackknife branch: column-drop masking for TT x (N-1) x N eff.boot.
                ## For each replicate j (unit j dropped), extract the surviving cells
                ## and compute per-replicate log-ATT using log(Y_j) - log(Y0_j).
                keep_cols <- seq_len(ncol(fit$D.dat))
                theta_j <- numeric(nboots)
                for (j in seq_len(nboots)) {
                    kcols      <- keep_cols[-j]
                    cm_j_full  <- cell_mask[, kcols, drop = FALSE]
                    eb_j       <- fit$eff.boot[, , j]  ## TT x (N-1)
                    eff_j_v    <- eb_j[cm_j_full]
                    Y_j        <- fit$Y.dat[, kcols, drop = FALSE][cm_j_full]
                    Y0_j_v     <- Y_j - eff_j_v
                    ok_j       <- !is.na(Y_j) & !is.na(Y0_j_v) &
                                  Y_j > 0 & Y0_j_v > 0
                    if (sum(ok_j) == 0L) {
                        theta_j[j] <- NA_real_
                    } else {
                        theta_j[j] <- mean(log(Y_j[ok_j]) - log(Y0_j_v[ok_j]),
                                           na.rm = TRUE)
                    }
                }
                theta_valid <- theta_j[is.finite(theta_j)]
                N_eff <- length(theta_valid)
                if (N_eff >= 2L) {
                    pseudo    <- nboots * estimate[k] - (nboots - 1) * theta_valid
                    se_vec[k] <- sqrt(var(pseudo) / N_eff)
                    z_q       <- stats::qnorm(1 - (1 - conf.level) / 2)
                    ci_lo[k]  <- estimate[k] - z_q * se_vec[k]
                    ci_hi[k]  <- estimate[k] + z_q * se_vec[k]
                }
            } else {
                ## Bootstrap / wild / parametric branch (original code).
                eff_boot_cells <- apply(fit$eff.boot, 3,
                                        function(eb) eb[cell_mask])
                if (!is.matrix(eff_boot_cells)) {
                    eff_boot_cells <- matrix(eff_boot_cells,
                                             nrow = sum(cell_mask))
                }
                eff_ok    <- eff_boot_cells[ok, , drop = FALSE]
                Y_ok      <- Y_t[ok]
                Y0_b_ok   <- Y_ok - eff_ok

                ## Hard-error on cell-drop pathology (v2.4.2+).
                ##
                ## When a cell used in the point estimate has Y0_b <= 0 in
                ## a non-trivial fraction of bootstrap replicates, log(Y0_b)
                ## returns NaN and colMeans(..., na.rm = TRUE) silently
                ## averages over fewer cells in that replicate, breaking
                ## the basic bootstrap principle and contaminating the
                ## bootstrap distribution.
                ##
                ## Threshold: trigger when the WORST cell has Y0_b <= 0 in
                ## > 5% of replicates. Sub-threshold cells are tolerated
                ## (small dropping is benign at the bootstrap-distribution
                ## scale; >5% indicates a genuinely unstable cell that
                ## needs filtering or a different estimand).
                n_reps_per_cell <- rowSums(Y0_b_ok <= 0, na.rm = TRUE)
                n_total_reps    <- ncol(Y0_b_ok)
                drop_frac       <- n_reps_per_cell / n_total_reps
                worst_idx       <- which.max(drop_frac)
                worst_frac      <- drop_frac[worst_idx]
                if (length(worst_frac) && worst_frac > 0.05) {
                    worst_Y0 <- Y0_t[ok][worst_idx]
                    stop(sprintf(
                        "log-ATT bootstrap is unreliable at event time %s.\n  The worst cell has Y0_hat = %.4f but %d of %d bootstrap replicates\n  (%.1f%%) have Y0_b <= 0 for it, so log(Y0_b) is undefined and the\n  per-replicate average silently drops the cell. This contaminates the\n  bootstrap distribution and yields meaningless inference.\n\n  Options:\n    1. Filter out unstable cells:\n         estimand(fit, \"log.att\", \"event.time\",\n                  cells = ~ Y0_hat > <threshold>)\n    2. Transform the outcome before fect: log(Y + c) for some c > 0\n    3. Use a different estimand: estimand(\"att\", ...) does not have\n       this pathology",
                        as.character(et), worst_Y0,
                        n_reps_per_cell[worst_idx], n_total_reps,
                        100 * worst_frac
                    ), call. = FALSE)
                }

                log_Y0_b  <- log(Y0_b_ok)
                log_Y     <- log(Y_ok)
                log_diff_b <- log_Y - log_Y0_b
                logatt_b  <- colMeans(log_diff_b, na.rm = TRUE)

                ## PARAMETRIC SHIFT (v2.4.2 fix)
                ## Note: log.att with parametric vartype requires strictly positive Y0_hat,
                ## which is only reachable when the outcome has been transformed (e.g. log(Y+c)).
                ## When reachable, eff.boot is H0-centered so logatt_b ≈ 0; the shift
                ## centers it at estimate[k] (the true log-ATT). The shift preserves sd().
                is_parametric <- isTRUE(fit$vartype == "parametric") ||
                                 isTRUE(vartype == "parametric")
                if (is_parametric) {
                    logatt_b <- logatt_b - mean(logatt_b, na.rm = TRUE) + estimate[k]
                }

                ## Cell-level jackknife on the per-cell log-diff vector.
                jack_v <- if (ci.method == "bca") {
                    .cell_jackknife("log.att", log_diff = log_diff)
                } else NULL

                ci <- .compute_ci(estimate[k], logatt_b, ci.method, conf.level,
                                  jack = jack_v)
                se_vec[k] <- ci$se
                ci_lo[k]  <- ci$ci.lo
                ci_hi[k]  <- ci$ci.hi
            }
        }
    }

    used_vartype <- if (vartype == "none") "none"
                    else if (is.null(fit$vartype)) "bootstrap"
                    else fit$vartype

    data.frame(
        event.time = ets,
        estimate   = estimate,
        se         = se_vec,
        ci.lo      = ci_lo,
        ci.hi      = ci_hi,
        n_cells    = n_cells,
        vartype    = used_vartype,
        stringsAsFactors = FALSE
    )
}


## ---------------------------------------------------------------------------
## Internal helpers
## ---------------------------------------------------------------------------

## Compute SE and (lo, hi) confidence-interval bounds from a bootstrap
## distribution `boot` (a numeric vector of replicate-level estimates),
## a point `estimate`, and a chosen `ci.method`.
##
## Supports five methods:
##   - "basic":     ci = (2*est - q_hi, 2*est - q_lo) [reflected]
##   - "percentile": ci = (q_lo, q_hi)
##   - "bc":        bias-corrected percentile (z0 only, no acceleration)
##                  ci = (q_{Phi(2*z0 + z_alpha/2)},
##                        q_{Phi(2*z0 + z_{1-alpha/2})})
##                  where z0 = Phi^-1(mean(boot < est)).
##   - "bca":       bias-corrected accelerated (Efron 1987 full BCa).
##                  Requires `jack` (a numeric vector of leave-one-out
##                  point estimates over the cells in the aggregation
##                  group) to compute the acceleration parameter.
##                  Cutoffs: a_lo = Phi(z0 + (z0+z_alpha/2) /
##                                          (1 - a*(z0+z_alpha/2)))
##                  Handles bootstrap-bias + bootstrap-skew jointly;
##                  default for ratio (aptt) and log (log.att) estimands
##                  where the bootstrap distribution is inherently
##                  skewed and bc alone degenerates at the boundary.
##   - "normal":    ci = est +/- z_{1-alpha/2} * SE   [Wald, symmetric]
##
## `jack` is the per-cell leave-one-out vector of within-group point
## estimates; required for "bca", ignored otherwise. The acceleration
## is a = sum((mean(jack) - jack)^3) / (6 * (sum(...^2))^1.5).
##
## Returns a list with elements `se`, `ci.lo`, `ci.hi`.
.compute_ci <- function(estimate, boot, ci.method, conf.level, jack = NULL) {
    alpha <- 1 - conf.level
    se    <- stats::sd(boot, na.rm = TRUE)

    if (ci.method == "normal") {
        z <- stats::qnorm(1 - alpha / 2)
        return(list(se    = se,
                    ci.lo = estimate - z * se,
                    ci.hi = estimate + z * se))
    }

    probs <- c(alpha / 2, 1 - alpha / 2)
    qs    <- stats::quantile(boot, probs = probs, na.rm = TRUE)

    if (ci.method == "percentile") {
        return(list(se = se, ci.lo = unname(qs[1]), ci.hi = unname(qs[2])))
    }
    if (ci.method == "basic") {
        return(list(se    = se,
                    ci.lo = 2 * estimate - unname(qs[2]),
                    ci.hi = 2 * estimate - unname(qs[1])))
    }
    if (ci.method == "bc") {
        valid <- !is.na(boot)
        if (sum(valid) == 0L) {
            return(list(se = NA_real_, ci.lo = NA_real_, ci.hi = NA_real_))
        }
        ## z0 = bias correction = Phi^-1(P(boot < estimate))
        p_below <- mean(boot[valid] < estimate)
        ## Clamp to avoid +/-Inf at the boundaries
        p_below <- pmin(pmax(p_below, 1e-6), 1 - 1e-6)
        z0      <- stats::qnorm(p_below)
        z_lo    <- stats::qnorm(alpha / 2)
        z_hi    <- stats::qnorm(1 - alpha / 2)
        a_lo    <- stats::pnorm(2 * z0 + z_lo)
        a_hi    <- stats::pnorm(2 * z0 + z_hi)
        bc_qs   <- stats::quantile(boot, probs = c(a_lo, a_hi), na.rm = TRUE)
        ## Same robustness fallback as bca (see comment in the bca block):
        ## when z0 hits the clamp the bc cutoffs collapse or shift off the
        ## estimate. Normal CI is the safe replacement.
        ci_lo_bc <- unname(bc_qs[1])
        ci_hi_bc <- unname(bc_qs[2])
        is_degenerate <- abs(ci_hi_bc - ci_lo_bc) < 1e-10
        is_uncovered  <- (estimate < ci_lo_bc) || (estimate > ci_hi_bc)
        if (is_degenerate || is_uncovered) {
            z <- stats::qnorm(1 - alpha / 2)
            return(list(se    = se,
                        ci.lo = estimate - z * se,
                        ci.hi = estimate + z * se))
        }
        return(list(se = se, ci.lo = ci_lo_bc, ci.hi = ci_hi_bc))
    }
    if (ci.method == "bca") {
        if (is.null(jack)) {
            stop("ci.method = \"bca\" requires the per-cell jackknife ",
                 "vector via the `jack` argument; the caller must compute ",
                 "leave-one-out within-group estimates and pass them.",
                 call. = FALSE)
        }
        valid <- !is.na(boot)
        jack_valid <- !is.na(jack)
        if (sum(valid) == 0L) {
            return(list(se = NA_real_, ci.lo = NA_real_, ci.hi = NA_real_))
        }
        ## Fall back to bc when the cell-level jackknife is degenerate
        ## (e.g., only one treated cell contributing at this event time:
        ## leave-one-out yields an empty vector; the .cell_jackknife()
        ## helper returns rep(NA, n) at n < 2). bc only needs the
        ## bootstrap distribution, so it stays well-defined wherever
        ## sum(valid) > 0. This fallback fires on the staggered tails
        ## of event-time series where event.time t is only reached by
        ## a single treated unit; without it the user sees missing CI
        ## bands on the plot at the extreme event-times.
        if (sum(jack_valid) < 2L) {
            return(.compute_ci(estimate, boot, "bc", conf.level,
                               jack = NULL))
        }
        ## z0: bias correction
        p_below <- mean(boot[valid] < estimate)
        p_below <- pmin(pmax(p_below, 1e-6), 1 - 1e-6)
        z0      <- stats::qnorm(p_below)
        ## a: acceleration via cell-level jackknife
        jack_v   <- jack[jack_valid]
        jack_bar <- mean(jack_v)
        dev      <- jack_bar - jack_v
        num      <- sum(dev^3)
        den      <- 6 * (sum(dev^2))^1.5
        a        <- if (den > 1e-12) num / den else 0
        ## BCa cutoffs: handles z0 -> +/- inf via the (1 - a*z) denominator
        z_lo  <- stats::qnorm(alpha / 2)
        z_hi  <- stats::qnorm(1 - alpha / 2)
        adjust <- function(z_q) {
            denom <- 1 - a * (z0 + z_q)
            ## Guard against a*(z0+z_q) -> 1 (denom -> 0); fall back to bc.
            if (abs(denom) < 1e-8) {
                return(stats::pnorm(2 * z0 + z_q))
            }
            stats::pnorm(z0 + (z0 + z_q) / denom)
        }
        a_lo  <- adjust(z_lo)
        a_hi  <- adjust(z_hi)
        bca_qs <- stats::quantile(boot, probs = c(a_lo, a_hi), na.rm = TRUE)
        ## Robustness fallback to normal CI when bca produces a pathological
        ## interval: (a) both cutoffs collapse to the same bootstrap quantile
        ## (degenerate; happens when z0 hits the clamp at +-qnorm(1e-6)),
        ## or (b) the interval shifts entirely off the point estimate
        ## (happens with heavily skewed bootstraps at small B). bca is
        ## mathematically valid in both cases but the resulting visual is
        ## misleading on per-event-time plots. Normal CI is centered at
        ## estimate by construction so it always covers, and it uses the
        ## already-computed bootstrap SE.
        ci_lo_bca <- unname(bca_qs[1])
        ci_hi_bca <- unname(bca_qs[2])
        is_degenerate <- abs(ci_hi_bca - ci_lo_bca) < 1e-10
        is_uncovered  <- (estimate < ci_lo_bca) || (estimate > ci_hi_bca)
        if (is_degenerate || is_uncovered) {
            z <- stats::qnorm(1 - alpha / 2)
            return(list(se    = se,
                        ci.lo = estimate - z * se,
                        ci.hi = estimate + z * se))
        }
        return(list(se = se, ci.lo = ci_lo_bca, ci.hi = ci_hi_bca))
    }
    stop("Unknown ci.method = \"", ci.method, "\".", call. = FALSE)
}


## Compute the per-cell jackknife vector for a within-group functional T.
## For aptt: T(eff, Y0) = mean(eff) / mean(Y0). leave-one-out:
##   theta_jack[i] = mean(eff[-i]) / mean(Y0[-i])
## For log.att: T(Y, Y0) = mean(log(Y) - log(Y0)). leave-one-out:
##   theta_jack[i] = mean(log(Y[-i]) - log(Y0[-i]))
## For att (level): T = mean(eff). theta_jack[i] = mean(eff[-i]).
##
## Used for the BCa acceleration parameter without requiring model refits
## (the influence is computed at the aggregation step, holding the model
## fixed). This is the standard practice for BCa when leave-one-unit-out
## refits are too expensive.
.cell_jackknife <- function(type, ...) {
    args <- list(...)
    if (type == "aptt") {
        eff <- args$eff; Y0 <- args$Y0
        n   <- length(eff)
        if (n < 2L) return(rep(NA_real_, n))
        sum_eff <- sum(eff, na.rm = TRUE)
        sum_Y0  <- sum(Y0,  na.rm = TRUE)
        ## leave-one-out means: (sum - eff_i) / (n - 1)
        num <- (sum_eff - eff) / (n - 1)
        den <- (sum_Y0  - Y0)  / (n - 1)
        return(num / den)
    }
    if (type == "log.att") {
        ld <- args$log_diff
        n  <- length(ld)
        if (n < 2L) return(rep(NA_real_, n))
        ## leave-one-out mean of log_diff
        sum_ld <- sum(ld, na.rm = TRUE)
        return((sum_ld - ld) / (n - 1))
    }
    if (type == "att") {
        eff <- args$eff
        n   <- length(eff)
        if (n < 2L) return(rep(NA_real_, n))
        sum_eff <- sum(eff, na.rm = TRUE)
        return((sum_eff - eff) / (n - 1))
    }
    stop("Unknown jackknife type = \"", type, "\".", call. = FALSE)
}


## Build the cell-level base mask for a given test (none / placebo /
## carryover) and direction. Returns a list with `mask` (logical
## matrix matching shape(fit$D.dat)) and `Tev` (the relevant event-
## time matrix from fit$T.on or fit$T.off).
##
## test = "none":      treated post-treatment cells (the default ATT
##                     surface): D.dat == 1 with non-NA Tev.
## test = "placebo":   pre-treatment cells masked during the placebo
##                     fit, identified by Tev within fit$placebo.period.
##                     Requires fit$placeboTest == TRUE.
## test = "carryover": early post-reversal cells masked during the
##                     carryover fit, identified by Tev within
##                     fit$carryover.period (Tev = T.off). Requires
##                     fit$carryoverTest == TRUE and hasRevs.
##
## v2.4.3+ (closes issue #131, ajunquera).
.test_cell_mask <- function(fit, test, direction) {

    Tev <- if (direction == "on") fit$T.on else fit$T.off
    if (is.null(Tev)) {
        stop("direction = \"", direction, "\" requested, but fit$T.",
             direction, " is NULL.", call. = FALSE)
    }

    if (test == "none") {
        return(list(
            mask = !is.na(fit$D.dat) & fit$D.dat == 1 & !is.na(Tev),
            Tev  = Tev
        ))
    }

    if (test == "placebo") {
        if (!isTRUE(as.logical(fit$placeboTest)) ||
            is.null(fit$placebo.period)) {
            stop("test = \"placebo\" requires the fit to have been run ",
                 "with placeboTest = TRUE. The placebo estimand is only ",
                 "meaningful when the placebo cells were masked from the ",
                 "fit (out-of-sample predictions); a standard fit's ",
                 "pre-treatment residuals are in-sample and would not be ",
                 "an honest credibility check. Refit with: ",
                 "fect(..., placeboTest = TRUE, placebo.period = c(L, R)).",
                 call. = FALSE)
        }
        pp <- fit$placebo.period
        if (length(pp) == 1L) pp <- c(pp, pp)
        return(list(
            mask = !is.na(Tev) & Tev >= pp[1] & Tev <= pp[2],
            Tev  = Tev
        ))
    }

    if (test == "carryover") {
        if (!isTRUE(as.logical(fit$carryoverTest)) ||
            is.null(fit$carryover.period)) {
            stop("test = \"carryover\" requires the fit to have been run ",
                 "with carryoverTest = TRUE. The carryover estimand is ",
                 "only meaningful when the early post-reversal cells were ",
                 "masked from the fit (out-of-sample predictions). Refit ",
                 "with: fect(..., carryoverTest = TRUE, ",
                 "carryover.period = c(L, R)).",
                 call. = FALSE)
        }
        if (!isTRUE(fit$hasRevs == 1)) {
            stop("test = \"carryover\" requires a panel with treatment ",
                 "reversals (fit$hasRevs == 1).", call. = FALSE)
        }
        cp <- fit$carryover.period
        if (length(cp) == 1L) cp <- c(cp, cp)
        return(list(
            mask = !is.na(Tev) & Tev >= cp[1] & Tev <= cp[2],
            Tev  = Tev
        ))
    }

    stop("Unknown test = \"", test, "\".", call. = FALSE)
}


## Apply a `cells` filter (NULL, logical, or one-sided formula/function)
## against a long-form data frame. Returns the filtered data frame.
.apply_cells_filter <- function(df, cells) {
    if (is.null(cells)) return(df)

    if (is.logical(cells)) {
        if (length(cells) != nrow(df)) {
            stop(
                "cells: logical filter has length ", length(cells),
                " but data has ", nrow(df), " rows.",
                call. = FALSE
            )
        }
        return(df[cells & !is.na(cells), , drop = FALSE])
    }

    if (inherits(cells, "formula")) {
        if (length(cells) != 2L) {
            stop("cells: formula must be one-sided (e.g., `~ event.time > 0`).",
                 call. = FALSE)
        }
        mask <- eval(cells[[2]], envir = df, enclos = parent.frame())
        if (!is.logical(mask) || length(mask) != nrow(df)) {
            stop("cells: formula must evaluate to a logical vector of ",
                 "length nrow(data).", call. = FALSE)
        }
        return(df[mask & !is.na(mask), , drop = FALSE])
    }

    if (is.function(cells)) {
        mask <- cells(df)
        if (!is.logical(mask) || length(mask) != nrow(df)) {
            stop("cells: function must return a logical vector of ",
                 "length nrow(data).", call. = FALSE)
        }
        return(df[mask & !is.na(mask), , drop = FALSE])
    }

    stop("cells must be NULL, a logical vector, a one-sided formula, ",
         "or a function.", call. = FALSE)
}

## -------------------------------------------------------------------
## v2.4.1 adds "parametric" to the vartype enum on estimand(). The
## boot.R machinery already populates fit$eff.boot under
## vartype = "parametric" + keep.sims = TRUE; the v2.4.1 change is just
## the match.arg gate. These tests verify:
##   - the new enum value is accepted
##   - byte-equality with fit$est.att under parametric
##   - all four `type` paths (att, att.cumu, aptt, log.att) work on
##     parametric fits, sourcing from fit$eff.boot
##   - keep.sims = FALSE still errors helpfully on non-fast paths
##   - vartype = "none" returns NA SE/CI under parametric
## -------------------------------------------------------------------

## Helper: cache one parametric fit per session for the heavy tests.
.make_parametric_fit <- local({
  cached <- NULL
  function() {
    if (!is.null(cached)) return(cached)
    skip_on_cran()
    e <- new.env()
    data(sim_linear, package = "fect", envir = e)
    set.seed(42)
    ## suppressWarnings covers the "EM did not converge within
    ## max.iteration = 5000" warning. The fixture is intentionally small
    ## (nboots = 30) and convergence is not what these tests verify.
    fit <- suppressWarnings(
      fect(Y ~ D, data = e$sim_linear, index = c("id", "time"),
           method = "ife", force = "two-way", se = TRUE,
           nboots = 30, r = 2, CV = FALSE, keep.sims = TRUE,
           vartype = "parametric",
           time.component.from = "nevertreated",
           parallel = FALSE)
    )
    cached <<- fit
    fit
  }
})

test_that("vartype = 'BAD' is rejected (enum gate, CRAN-runnable)", {
  ## Cheap arg-validation. match.arg fails at the top of estimand()
  ## before any fit-shape inspection, so a stub list is sufficient.
  stub <- structure(list(), class = "fect")
  expect_error(
    estimand(stub, "att", "event.time", vartype = "BAD"),
    "should be one of"
  )
})

test_that("vartype = 'parametric' is accepted by estimand()", {
  skip_on_cran()
  fit <- .make_parametric_fit()
  res <- estimand(fit, "att", "event.time", vartype = "parametric")
  expect_s3_class(res, "data.frame")
  expect_true(all(res$vartype == "parametric"))
})

test_that("estimand(fit, 'att', 'event.time') byte-equals fit$est.att under parametric", {
  skip_on_cran()
  fit <- .make_parametric_fit()
  res <- estimand(fit, "att", "event.time")
  expect_identical(res$estimate, unname(fit$est.att[, "ATT"]))
  expect_identical(res$se,       unname(fit$est.att[, "S.E."]))
  expect_identical(res$ci.lo,    unname(fit$est.att[, "CI.lower"]))
  expect_identical(res$ci.hi,    unname(fit$est.att[, "CI.upper"]))
  expect_identical(res$n_cells,  unname(fit$est.att[, "count"]))
})

test_that("output 'vartype' column reports fit$vartype, not the argument", {
  skip_on_cran()
  fit <- .make_parametric_fit()
  ## Default arg defaults to "bootstrap"; output should still say
  ## "parametric" because that's what generated the replicates.
  res_default <- estimand(fit, "att", "event.time")
  expect_true(all(res_default$vartype == "parametric"))
  ## Explicit "parametric": same numerical result, same vartype column.
  res_explicit <- estimand(fit, "att", "event.time", vartype = "parametric")
  expect_identical(res_default, res_explicit)
})

test_that("att.cumu / aptt / att-overall work on parametric fits", {
  skip_on_cran()
  fit <- .make_parametric_fit()

  ## att.cumu defaults to ci.method = "basic"; aptt defaults to "bca".
  ## Both are tail-quantile CIs; on this small fit (nboots = 30) they
  ## emit the v2.4.2 .check_tail_ci_replicates warning by design.
  ## Suppress it here -- this test asserts the call runs and produces
  ## non-NA results, not CI quality.
  suppressWarnings(r1 <- estimand(fit, "att.cumu", "event.time"))
  expect_true(all(r1$vartype == "parametric"))
  expect_true(any(!is.na(r1$estimate)))

  expect_silent(r2 <- estimand(fit, "att", "overall", window = c(1, 5)))
  expect_true(r2$vartype == "parametric")
  expect_true(!is.na(r2$estimate))

  suppressWarnings(r3 <- estimand(fit, "aptt", "event.time"))
  expect_true(all(r3$vartype == "parametric"))
  expect_true(any(!is.na(r3$estimate)))
})

test_that("log.att hard-errors on parametric fits with negative Y (v2.4.2+)", {
  skip_on_cran()
  fit <- .make_parametric_fit()
  ## sim_linear has many negative Y cells, so log.att now hard-errors
  ## at point-estimate level (v2.4.2+). The previous v2.4.1 behavior of
  ## silently warning + dropping cells produced meaningless inference;
  ## the hard-error redirects users to the actionable options.
  expect_error(
    estimand(fit, "log.att", "event.time"),
    "log\\.att requires Y > 0"
  )
})

test_that("vartype = 'none' under parametric returns NA SE/CI", {
  skip_on_cran()
  fit <- .make_parametric_fit()
  res <- estimand(fit, "att", "overall", window = c(1, 5),
                  vartype = "none")
  expect_true(!is.na(res$estimate))
  expect_true(is.na(res$se))
  expect_true(is.na(res$ci.lo))
  expect_true(is.na(res$ci.hi))
})

test_that("keep.sims = FALSE under parametric still errors helpfully on non-fast paths", {
  skip_on_cran()
  e <- new.env()
  data(sim_linear, package = "fect", envir = e)
  set.seed(42)
  ## suppressWarnings covers the "EM did not converge within
  ## max.iteration = 5000" warning on this small (nboots = 20) fixture;
  ## the test verifies error messages on non-fast paths, not convergence.
  fit_no_sims <- suppressWarnings(
    fect(Y ~ D, data = e$sim_linear, index = c("id", "time"),
         method = "ife", force = "two-way", se = TRUE,
         nboots = 20, r = 2, CV = FALSE,
         keep.sims = FALSE,
         vartype = "parametric",
         time.component.from = "nevertreated",
         parallel = FALSE)
  )
  expect_null(fit_no_sims$eff.boot)

  ## att / event.time fast path reads fit$est.att, so still works.
  expect_silent(res_fast <- estimand(fit_no_sims, "att", "event.time"))
  expect_true(all(res_fast$vartype == "parametric"))

  ## att.cumu requires eff.boot; should error with locked wording.
  expect_error(
    estimand(fit_no_sims, "att.cumu", "event.time"),
    "No bootstrap/jackknife results available"
  )
})

test_that("eff.boot replicate count integrity under parametric", {
  skip_on_cran()
  fit <- .make_parametric_fit()
  expect_equal(dim(fit$eff.boot)[3], 30L)
  expect_equal(dim(fit$eff.boot)[1], nrow(fit$Y.dat))
  expect_equal(dim(fit$eff.boot)[2], ncol(fit$Y.dat))
})

## ============================================================================
## Full coverage test suite for para.error redesign (test-spec.md scenarios)
## Companion to test-para-error.R (builder's starter file, 10 tests).
## This file covers: T4, T7, T9-T10, T13-T21, E2-E4, E6
##
## Spec: statsclaw-workspace/fect/wild-as-paraerror/test-spec.md
## Branch: feat/v242-completion
## ============================================================================

## ============================================================================
## DGP helpers (spec §2)
## ============================================================================

## DGP-A: additive TWFE, IID Gaussian, true ATT = 3.0, fully observed
##
## Treatment block is FIXED at units 1:Ntr (deterministic). The parametric
## bootstrap targets the conditional variance V_t(Lambda, F, X, D); a coverage
## simulation that re-randomizes the treatment assignment D across replications
## adds Var_{D}[b_t] to the marginal variance the bootstrap is being judged
## against, biasing measured coverage downward (see gsynth-note section 2).
dgp_a <- function(seed) {
    set.seed(seed)
    N <- 40; TT <- 20; T0 <- 12; Ntr <- 12
    alpha_i <- rnorm(N, 0, 2)
    xi_t    <- rnorm(TT, 0, 1)
    eps     <- matrix(rnorm(N * TT, 0, 1), TT, N)
    D <- matrix(0L, TT, N); D[(T0 + 1):TT, 1:Ntr] <- 1L
    Y <- outer(xi_t, rep(1, N)) + outer(rep(1, TT), alpha_i) + 3 * D + eps
    data.frame(
        id   = rep(1:N, each = TT),
        time = rep(1:TT, N),
        Y    = c(Y),
        D    = c(D)
    )
}

## DGP-A8: additive TWFE, AR(1) rho=0.8, true ATT = 3.0
## Treatment block fixed (see dgp_a comment).
dgp_a8 <- function(seed) {
    set.seed(seed)
    N <- 40; TT <- 20; T0 <- 12; Ntr <- 12
    alpha_i <- rnorm(N, 0, 2)
    xi_t    <- rnorm(TT, 0, 1)
    eps <- matrix(NA, TT, N)
    for (i in 1:N) {
        e <- rnorm(TT, 0, 1)
        eps[1, i] <- e[1] / sqrt(1 - 0.64)   # stationary init
        for (t in 2:TT) eps[t, i] <- 0.8 * eps[t - 1, i] + e[t]
    }
    D <- matrix(0L, TT, N); D[(T0 + 1):TT, 1:Ntr] <- 1L
    Y <- outer(xi_t, rep(1, N)) + outer(rep(1, TT), alpha_i) + 3 * D + eps
    data.frame(
        id   = rep(1:N, each = TT),
        time = rep(1:TT, N),
        Y    = c(Y),
        D    = c(D)
    )
}

## DGP-B: positive Y, multiplicative treatment (for aptt / log.att)
## Treatment block fixed (see dgp_a comment).
dgp_b <- function(seed) {
    set.seed(seed)
    N <- 40; TT <- 20; T0 <- 12; Ntr <- 12
    alpha_i <- rnorm(N, 0, 0.5)
    xi_t    <- rnorm(TT, 0, 0.3)
    eps     <- matrix(rnorm(N * TT, 0, 0.5), TT, N)
    D <- matrix(0L, TT, N); D[(T0 + 1):TT, 1:Ntr] <- 1L
    Y0 <- 20 + outer(rep(1, TT), alpha_i) + outer(xi_t, rep(1, N)) + eps
    Y  <- Y0 * ifelse(D == 1, 1.3, 1)
    Y  <- pmax(Y, 0.01)
    data.frame(
        id   = rep(1:N, each = TT),
        time = rep(1:TT, N),
        Y    = c(Y),
        D    = c(D)
    )
}

## DGP-M: missing-data panel (10% rows removed)
dgp_m <- function(seed) {
    df <- dgp_a(seed)
    set.seed(seed + 999)
    mask <- sample(1:nrow(df), floor(0.10 * nrow(df)))
    df <- df[-mask, ]
    df
}

## DGP-REV: panel with treatment reversals (some units turn off treatment)
## N=30: 10 reversing treated, 10 absorbing treated, 10 never-treated controls.
dgp_rev <- function(seed = 77) {
    set.seed(seed)
    N_rev <- 10; N_abs <- 10; N_co <- 10
    N <- N_rev + N_abs + N_co; TT <- 20
    D_vec <- integer(N * TT)
    Y_vec <- numeric(N * TT)
    ## Reversing units: turn on at t=8, turn off at t=14
    for (i in 1:N_rev) {
        for (t in 1:TT) {
            idx <- (i - 1) * TT + t
            D_vec[idx] <- as.integer(t >= 8 && t < 14)
            Y_vec[idx] <- rnorm(1) + 2 * D_vec[idx]
        }
    }
    ## Absorbing treated: turn on at t=8, stay on
    for (i in (N_rev + 1):(N_rev + N_abs)) {
        for (t in 1:TT) {
            idx <- (i - 1) * TT + t
            D_vec[idx] <- as.integer(t >= 8)
            Y_vec[idx] <- rnorm(1) + 2 * D_vec[idx]
        }
    }
    ## Never-treated controls
    for (i in (N_rev + N_abs + 1):N) {
        for (t in 1:TT) {
            idx <- (i - 1) * TT + t
            D_vec[idx] <- 0L
            Y_vec[idx] <- rnorm(1)
        }
    }
    data.frame(
        id   = rep(1:N, each = TT),
        time = rep(1:TT, N),
        Y    = Y_vec,
        D    = D_vec
    )
}

## DGP-STAG: staggered adoption (3 cohorts) for E6
dgp_stag <- function(seed = 55) {
    set.seed(seed)
    N <- 30; TT <- 20; Ntr <- 18
    cohort_times <- c(rep(6, 6), rep(10, 6), rep(14, 6), rep(TT + 1, 12))
    D <- matrix(0, TT, N)
    for (i in 1:Ntr) {
        if (cohort_times[i] <= TT) D[cohort_times[i]:TT, i] <- 1
    }
    alpha_i <- rnorm(N, 0, 2)
    xi_t    <- rnorm(TT, 0, 1)
    eps     <- matrix(rnorm(N * TT, 0, 1), TT, N)
    Y <- outer(xi_t, rep(1, N)) + outer(rep(1, TT), alpha_i) + 2 * D + eps
    data.frame(
        id   = rep(1:N, each = TT),
        time = rep(1:TT, N),
        Y    = c(Y),
        D    = c(D)
    )
}

## Shared parametric call helper
fect_para_full <- function(d, para.error = "auto", vartype = "parametric",
                           nboots = 100, seed = 42, ...) {
    suppressWarnings(suppressMessages(
        fect(
            Y ~ D,
            data   = d,
            index  = c("id", "time"),
            method = "fe",
            force  = "two-way",
            time.component.from = "nevertreated",
            se      = TRUE,
            vartype = vartype,
            para.error = para.error,
            nboots     = nboots,
            parallel   = FALSE,
            keep.sims  = TRUE,
            seed       = seed,
            CV         = FALSE,
            ...
        )
    ))
}

## ============================================================================
## T4: Point estimate byte-stability across all three para.error modes
## ============================================================================

test_that("T4: att.avg is byte-identical across para.error = ar, empirical, wild", {
    skip_on_cran()
    df <- dgp_a(seed = 42)
    fit_ar  <- fect_para_full(df, para.error = "ar",        nboots = 50)
    fit_emp <- fect_para_full(df, para.error = "empirical", nboots = 50)
    fit_wld <- fect_para_full(df, para.error = "wild",      nboots = 50)

    expect_identical(fit_ar$att.avg, fit_emp$att.avg,
        label = "att.avg: ar vs empirical should be byte-identical")
    expect_identical(fit_ar$att.avg, fit_wld$att.avg,
        label = "att.avg: ar vs wild should be byte-identical")
})

## ============================================================================
## T7: para.error = "ar" on missing-data panel succeeds
## ============================================================================

test_that("T7: para.error = 'ar' succeeds on missing-data panel", {
    skip_on_cran()
    df_m <- dgp_m(seed = 1)
    fit <- suppressWarnings(suppressMessages(
        fect(
            Y ~ D,
            data   = df_m,
            index  = c("id", "time"),
            method = "fe",
            force  = "two-way",
            time.component.from = "nevertreated",
            se         = TRUE,
            vartype    = "parametric",
            para.error = "ar",
            na.rm      = FALSE,
            nboots     = 100,
            parallel   = FALSE,
            keep.sims  = TRUE,
            CV         = FALSE
        )
    ))
    expect_equal(fit$para.error, "ar",
        label = "para.error should resolve to 'ar' on missing-data panel")
    ## At most 20% NA in att.avg.boot (spec T7 tolerance)
    na_rate <- mean(is.na(c(fit$att.avg.boot)))
    expect_lte(na_rate, 0.20,
        label = sprintf("NA rate in att.avg.boot = %.2f%%, should be <= 20%%",
                        na_rate * 100))
})

## ============================================================================
## T9, T10 (spec) — vartype = "wild" deprecation alias
## REMOVED: vartype = "wild" was never released as a standalone vartype, so
## there is no caller to deprecate. The redesign exposes para.error directly.
##
## T11 (spec) / covered in builder file — verify fit$para.error stores resolved value
## Already in test-para-error.R (Test 1), not repeated here.

## T12 (spec) / covered in builder file (Test 10), not repeated here.

## ============================================================================
## T13: estimand() location-shift applies to all three para.error modes
## ============================================================================

test_that("T13: estimand() percentile CI contains point estimate for all para.error modes", {
    skip_on_cran()
    df <- dgp_a(seed = 42)
    for (pm in c("ar", "empirical", "wild")) {
        fit <- fect_para_full(df, para.error = pm, nboots = 200)
        est <- estimand(fit, "att", "overall", window = c(1, 8), ci.method = "percentile")
        expect_true(
            is.finite(est$ci.lo) && is.finite(est$ci.hi),
            label = sprintf("T13 %s: CI must be finite", pm)
        )
        expect_true(
            est$ci.lo < fit$att.avg && fit$att.avg < est$ci.hi,
            label = sprintf(
                "T13 %s: point estimate %.4f must be inside CI [%.4f, %.4f]",
                pm, fit$att.avg, est$ci.lo, est$ci.hi
            )
        )
    }
})

## ============================================================================
## T14: Anti-regression — vartype = "bootstrap" unchanged
## ============================================================================

test_that("T14: vartype='bootstrap' produces finite, sensible results (anti-regression)", {
    skip_on_cran()
    data("simdata", package = "fect")
    set.seed(42)
    fit <- suppressMessages(
        fect(Y ~ D, data = simdata, index = c("id", "time"),
             method = "fe", force = "two-way",
             se = TRUE, vartype = "bootstrap",
             nboots = 50, parallel = FALSE, seed = 42, CV = FALSE)
    )
    expect_equal(fit$vartype, "bootstrap")
    expect_true(is.numeric(fit$att.avg) && !is.na(fit$att.avg),
        label = "bootstrap att.avg must be finite")
    ## Anti-regression: att.avg should be near 2.5 for simdata (known fixture)
    expect_true(abs(fit$att.avg) < 10,
        label = sprintf("bootstrap att.avg = %.4f looks reasonable", fit$att.avg))
    ## Ensure para.error is NULL for bootstrap path
    expect_null(fit$para.error,
        label = "para.error should be NULL for bootstrap fits")
})

## ============================================================================
## T15: Anti-regression — vartype = "jackknife" unchanged
## ============================================================================

test_that("T15: vartype='jackknife' produces finite, sensible results (anti-regression)", {
    skip_on_cran()
    ## Small panel so jackknife is fast
    df_small <- dgp_a(seed = 7)
    df_small <- df_small[df_small$id <= 20, ]
    fit <- suppressMessages(
        fect(Y ~ D, data = df_small, index = c("id", "time"),
             method = "fe", force = "two-way",
             se = TRUE, vartype = "jackknife",
             parallel = FALSE, CV = FALSE)
    )
    expect_equal(fit$vartype, "jackknife")
    expect_true(is.numeric(fit$att.avg) && !is.na(fit$att.avg),
        label = "jackknife att.avg must be finite")
    expect_null(fit$para.error,
        label = "para.error should be NULL for jackknife fits")
})

## ============================================================================
## T16: Anti-regression — vartype = "parametric" with para.error = "auto" unchanged
## ============================================================================

test_that("T16: vartype='parametric' para.error='auto' produces same resolved path as before", {
    skip_on_cran()
    df <- dgp_a(seed = 42)
    fit <- fect_para_full(df, para.error = "auto", nboots = 50)
    ## On fully-observed panel, auto should resolve to "empirical"
    expect_equal(fit$para.error, "empirical",
        label = "auto should resolve to empirical on fully-observed panel")
    ## Point estimate should be near 3.0 (true ATT)
    expect_true(abs(fit$att.avg - 3.0) < 1.5,
        label = sprintf("att.avg = %.4f should be near true ATT = 3.0", fit$att.avg))
    ## est.avg is a matrix (class = matrix/array)
    expect_true(is.matrix(fit$est.avg),
        label = "est.avg should be a matrix")
    expect_true(all(is.finite(fit$est.avg[, "ATT.avg"])),
        label = "est.avg[,ATT.avg] should be finite")
})

## ============================================================================
## T17: aptt + para.error = "wild" + bca returns finite CIs (DGP-B)
## ============================================================================

test_that("T17: aptt and log.att estimands work with para.error='wild' on DGP-B", {
    skip_on_cran()
    df <- dgp_b(seed = 99)
    fit <- suppressMessages(
        fect(
            Y ~ D,
            data   = df,
            index  = c("id", "time"),
            method = "fe",
            force  = "two-way",
            time.component.from = "nevertreated",
            se         = TRUE,
            vartype    = "parametric",
            para.error = "wild",
            nboots     = 200,
            parallel   = FALSE,
            keep.sims  = TRUE,
            seed       = 99,
            CV         = FALSE
        )
    )
    ## aptt estimand
    est_aptt <- estimand(fit, "aptt", "event.time", ci.method = "bca")
    expect_true(all(is.finite(est_aptt$ci.lo)) && all(is.finite(est_aptt$ci.hi)),
        label = "T17: aptt CI must be finite for all event times")
    expect_true(all(est_aptt$estimate > 0),
        label = sprintf("T17: aptt estimates should be positive (true APTT>0), got min=%.4f",
                        min(est_aptt$estimate)))

    ## log.att estimand
    est_log <- estimand(fit, "log.att", "event.time", ci.method = "bca")
    expect_true(all(is.finite(est_log$ci.lo)) && all(is.finite(est_log$ci.hi)),
        label = "T17: log.att CI must be finite for all event times")
    ## Each log.att estimate should be within +/-0.1 of log(1.3) = 0.2624
    true_logatt <- log(1.3)
    for (i in seq_len(nrow(est_log))) {
        expect_true(
            abs(est_log$estimate[i] - true_logatt) < 0.15,
            label = sprintf(
                "T17: log.att estimate[%d] = %.4f should be within 0.15 of %.4f",
                i, est_log$estimate[i], true_logatt
            )
        )
    }
})

## ============================================================================
## T18: CI width parity — wild vs empirical within 30% (nboots = 1000)
## ============================================================================

test_that("T18: CI width parity — wild/empirical ratio in [0.70, 1.30]", {
    skip_on_cran()
    df <- dgp_a(seed = 42)
    fit_emp <- fect_para_full(df, para.error = "empirical", nboots = 1000)
    fit_wld <- fect_para_full(df, para.error = "wild",      nboots = 1000)
    fit_ar  <- fect_para_full(df, para.error = "ar",        nboots = 1000)

    est_emp <- estimand(fit_emp, "att", "overall", window = c(1, 8), ci.method = "basic")
    est_wld <- estimand(fit_wld, "att", "overall", window = c(1, 8), ci.method = "basic")
    est_ar  <- estimand(fit_ar,  "att", "overall", window = c(1, 8), ci.method = "basic")

    width_emp <- est_emp$ci.hi - est_emp$ci.lo
    width_wld <- est_wld$ci.hi - est_wld$ci.lo
    width_ar  <- est_ar$ci.hi  - est_ar$ci.lo

    ratio_wld_emp <- width_wld / width_emp
    ratio_ar_emp  <- width_ar  / width_emp

    expect_gte(ratio_wld_emp, 0.70,
        label = sprintf("T18: wild/empirical width ratio = %.4f should be >= 0.70",
                        ratio_wld_emp))
    expect_lte(ratio_wld_emp, 1.30,
        label = sprintf("T18: wild/empirical width ratio = %.4f should be <= 1.30",
                        ratio_wld_emp))
    expect_gte(ratio_ar_emp, 0.50,
        label = sprintf("T18: ar/empirical width ratio = %.4f should be >= 0.50",
                        ratio_ar_emp))
})

## ============================================================================
## T19, T20, T21 (spec) --- coverage validation simulations
##
## MOVED to tests/coverage-study/run_para_error_coverage.R
##
## These three Monte-Carlo studies (T19: 100 reps x 1000 nboots x 15 cells on
## DGP-A IID; T20: same on DGP-A8 AR(1) rho=0.8; T21: 50 reps x 500 nboots
## width parity) take ~30 min wall and are not appropriate for routine
## devtools::test() runs.  They live alongside the coverage-study artifacts.
##
## Run them via:  Rscript tests/coverage-study/run_para_error_coverage.R
##
## When to run: any change to R/boot.R parametric branch, R/po-estimands.R
## location-shift code, the vartype / ci.method / para.error machinery,
## jackknife dispatch, or eff.boot construction.  See the README in that
## directory for full trigger list and acceptance criteria.
## ============================================================================

## ============================================================================
## E2: para.error = "wild" with nboots = 200 completes with >= 95% success rate
## ============================================================================

test_that("E2: para.error='wild' nboots=200 completes with >=95% successful replicates", {
    skip_on_cran()
    df <- dgp_a(seed = 42)
    fit <- fect_para_full(df, para.error = "wild", nboots = 200)
    n_ok <- sum(!is.na(c(fit$att.avg.boot)))
    expect_gte(n_ok, 190L,
        label = sprintf("E2: %d/200 replicates succeeded (need >=190)", n_ok))
    ## est.avg is a matrix
    expect_true(is.matrix(fit$est.avg),
        label = "E2: est.avg should be a matrix")
    expect_true(all(is.finite(fit$est.avg[, "ATT.avg"])),
        label = "E2: est.avg[,ATT.avg] should be finite")
    expect_true(all(is.finite(fit$est.avg[, "CI.lower"])),
        label = "E2: est.avg[,CI.lower] should be finite")
    expect_true(all(is.finite(fit$est.avg[, "CI.upper"])),
        label = "E2: est.avg[,CI.upper] should be finite")
})

## ============================================================================
## E3: para.error = "wild" with treatment reversals routes to hard error
## ============================================================================

test_that("E3: para.error='wild' with treatment reversals fires the parametric reversal gate", {
    skip_on_cran()
    df_rev <- dgp_rev(seed = 77)
    expect_error(
        suppressMessages(
            fect(
                Y ~ D,
                data   = df_rev,
                index  = c("id", "time"),
                method = "fe",
                force  = "two-way",
                time.component.from = "nevertreated",
                se         = TRUE,
                vartype    = "parametric",
                para.error = "wild",
                nboots     = 50,
                parallel   = FALSE,
                CV         = FALSE
            )
        ),
        regexp = "Parametric bootstrap is not valid when treatment reversal",
        fixed  = FALSE,
        label  = "E3: reversal gate should fire for para.error='wild'"
    )
})

## ============================================================================
## E4: para.error = "wild" + time.component.from = "notyettreated" -> hard error
## ============================================================================

test_that("E4: para.error='wild' + time.component.from='notyettreated' fires the notyettreated gate", {
    skip_on_cran()
    df <- dgp_a(seed = 42)
    expect_error(
        suppressMessages(
            fect(
                Y ~ D,
                data   = df,
                index  = c("id", "time"),
                method = "fe",
                force  = "two-way",
                time.component.from = "notyettreated",
                se         = TRUE,
                vartype    = "parametric",
                para.error = "wild",
                nboots     = 50,
                parallel   = FALSE,
                CV         = FALSE
            )
        ),
        regexp = "notyettreated",
        fixed  = FALSE,
        label  = "E4: notyettreated gate should fire for para.error='wild'"
    )
})

## ============================================================================
## E6: para.error = "wild" with staggered adoption completes, <= 10% NA
## ============================================================================

test_that("E6: para.error='wild' with staggered adoption completes, <=10% NA in att.avg.boot", {
    skip_on_cran()
    df_stag <- dgp_stag(seed = 55)
    fit <- suppressMessages(
        fect(
            Y ~ D,
            data   = df_stag,
            index  = c("id", "time"),
            method = "fe",
            force  = "two-way",
            time.component.from = "nevertreated",
            se         = TRUE,
            vartype    = "parametric",
            para.error = "wild",
            nboots     = 200,
            parallel   = FALSE,
            keep.sims  = TRUE,
            CV         = FALSE
        )
    )
    expect_true(!is.null(fit$att.avg.boot),
        label = "E6: att.avg.boot should exist")
    na_rate <- mean(is.na(c(fit$att.avg.boot)))
    expect_lte(na_rate, 0.10,
        label = sprintf("E6: NA rate in att.avg.boot = %.2f%% (should be <= 10%%)",
                        na_rate * 100))
})

## ---------------------------------------------------------------------------
## v2.4.2 post-hoc fix: parametric bootstrap CI calibration
##
## The pre-fix code passed eff.boot (H0-centered) directly to .compute_ci()
## for all ci.methods except "normal".  This caused percentile/basic/bc/bca
## CIs to be anchored near zero instead of near the point estimate, giving
## 0% coverage for non-zero ATTs.
##
## The fix (R/po-estimands.R): for vartype="parametric", shift att_b by
##   att_b <- att_b - mean(att_b) + estimate
## before calling .compute_ci().  This re-centers the distribution while
## preserving sd(), so "normal" CIs are unaffected.
##
## Tests in this file (from test-spec.md §3-§8):
##   P-INV-1  normal CI satisfies Wald structure exactly
##   P-INV-2  point estimate unchanged by fix
##   P-INV-3  se is identical for normal and basic (shift preserves sd)
##   P-INV-4  .compute_ci() with mock distribution: shifted vs unshifted
##   P-INV-5  bca jackknife path produces non-degenerate CI
##   P-EDGE-2 vartype="none" still returns NA for SE/CI
##   P-EDGE-3 bca on very small nboots does not hard-error
##   P-EDGE-4 shift does not affect vartype="bootstrap" fits
##   P-REG-1  pre-fix degenerate coverage is eliminated (smoke test, 5 reps)
##   P-COV-1  coverage >=0.85 for {basic,percentile,bca,normal}, DGP-A, fe
##   P-COV-2  coverage >=0.85 for bca, DGP-G, gsynth
##   P-COV-3  coverage >=0.75 for bca on aptt, DGP-B-positive
##   P-WIDTH-1 CI widths within 15% of bootstrap widths
## ---------------------------------------------------------------------------

## ---- DGP helpers ------------------------------------------------------------

## DGP-A: two-way FE, additive, true ATT = 3.0
make_panel_A <- function(seed) {
  set.seed(seed)
  N <- 40; TT <- 20; T0 <- 12; Ntr <- 12
  alpha_i <- rnorm(N, 0, 2)
  xi_t    <- rnorm(TT, 0, 1)
  D <- matrix(0L, TT, N); D[(T0 + 1):TT, 1:Ntr] <- 1L
  eps <- matrix(rnorm(N * TT, 0, 1), TT, N)
  Y <- outer(xi_t, rep(1, N)) + outer(rep(1, TT), alpha_i) + 3.0 * D + eps
  data.frame(id = rep(1:N, each = TT), time = rep(1:TT, N),
             Y = as.vector(Y), D = as.vector(D))
}

## DGP-G: interactive fixed effects with factor structure, true ATT = 3.0
make_panel_G <- function(seed, r = 2) {
  set.seed(seed)
  N <- 40; TT <- 20; T0 <- 12; Ntr <- 12
  Lambda <- matrix(rnorm(N * r, 0, 1), N, r)
  F_t    <- matrix(rnorm(TT * r, 0, 1), TT, r)
  eps    <- matrix(rnorm(N * TT, 0, 0.5), TT, N)
  Y0     <- F_t %*% t(Lambda) + eps
  D      <- matrix(0L, TT, N); D[(T0 + 1):TT, 1:Ntr] <- 1L
  Y      <- Y0 + 3.0 * D
  data.frame(id = rep(1:N, each = TT), time = rep(1:TT, N),
             Y = as.vector(Y), D = as.vector(D))
}

## DGP-B-positive: multiplicative effect, true APTT ~ 0.3, true log.ATT ~ log(1.3)
make_panel_B_pos <- function(seed) {
  set.seed(seed)
  N <- 40; TT <- 20; T0 <- 12; Ntr <- 12
  alpha_i <- rnorm(N, 0, 0.5)
  xi_t    <- rnorm(TT, 0, 0.3)
  eps     <- matrix(rnorm(N * TT, 0, 0.5), TT, N)
  Y0      <- 20 + outer(xi_t, rep(1, N)) + outer(rep(1, TT), alpha_i) + eps
  D       <- matrix(0L, TT, N); D[(T0 + 1):TT, 1:Ntr] <- 1L
  Y1      <- Y0; Y1[D == 1L] <- Y0[D == 1L] * 1.3
  Y_obs   <- pmax(ifelse(D == 1L, Y1, Y0), 1)
  treated <- which(D == 1L)
  list(
    df = data.frame(id = rep(1:N, each = TT), time = rep(1:TT, N),
                    Y = as.vector(Y_obs), D = as.vector(D)),
    true_aptt   = mean((Y1[treated] - Y0[treated]) / Y0[treated]),
    true_logatt = mean(log(Y1[treated] / Y0[treated]))
  )
}

## ---- Shared parametric fit helper -------------------------------------------
## Uses same fixture as test-estimand-parametric.R (.make_parametric_fit())
## but we define our own per-DGP helper so we can control the seed cleanly.

.make_parafix_fit <- local({
  cached <- NULL
  function() {
    if (!is.null(cached)) return(cached)
    skip_on_cran()
    e <- new.env()
    data(sim_linear, package = "fect", envir = e)
    set.seed(42)
    ## suppressWarnings here covers the "EM did not converge within
    ## max.iteration = 5000" warning. The fixture is intentionally small
    ## (nboots = 30) and convergence is not what these tests verify ---
    ## they check CI-formula invariants on the parametric path.
    suppressMessages(suppressWarnings(
      fect(Y ~ D, data = e$sim_linear, index = c("id", "time"),
           method = "ife", force = "two-way", se = TRUE,
           nboots = 30, r = 2, CV = FALSE, keep.sims = TRUE,
           vartype = "parametric",
           time.component.from = "nevertreated",
           parallel = FALSE)
    ))
  }
})

## ---- P-INV-1: normal CI satisfies Wald structure ----------------------------

test_that("P-INV-1: normal CI is byte-stable (Wald structure: ci = est +/- z*se)", {
  skip_on_cran()
  fit <- .make_parafix_fit()
  res <- fect::estimand(fit, "att", "overall", window = c(1, 5),
                        ci.method = "normal")
  z <- stats::qnorm(0.975)
  expect_equal(res$ci.lo, res$estimate - z * res$se, tolerance = 1e-10,
               label = "ci.lo = estimate - z*se")
  expect_equal(res$ci.hi, res$estimate + z * res$se, tolerance = 1e-10,
               label = "ci.hi = estimate + z*se")
  ## Width = 2*z*se
  expect_equal(res$ci.hi - res$ci.lo, 2 * z * res$se, tolerance = 1e-10,
               label = "CI width = 2*z*se")
  ## Non-degenerate
  expect_gt(res$ci.hi - res$ci.lo, 1e-6, label = "CI width > 1e-6")
})

## ---- P-INV-2: point estimate is unchanged by fix ----------------------------

test_that("P-INV-2: point estimate matches mean(fit$eff) at treated cells (fix-invariant)", {
  skip_on_cran()
  fit <- .make_parafix_fit()
  res <- fect::estimand(fit, "att", "overall", window = c(1, 5),
                        ci.method = "bca")
  ## Manual point estimate: mean over treated post-treatment cells with T.on in 1..5
  mask <- fit$D.dat == 1 &
          !is.na(fit$T.on) &
          fit$T.on >= 1 &
          fit$T.on <= 5
  manual_est <- mean(fit$eff[mask], na.rm = TRUE)
  expect_equal(res$estimate, manual_est, tolerance = 1e-10,
               label = "estimate == mean(eff[treated cells, T.on 1..5])")
})

## ---- P-INV-3: se is identical for normal and basic (shift preserves sd) -----

test_that("P-INV-3: se is identical for ci.method=normal and ci.method=basic on parametric fit", {
  skip_on_cran()
  fit <- .make_parafix_fit()
  res_normal <- fect::estimand(fit, "att", "overall", window = c(1, 5),
                               ci.method = "normal")
  res_basic  <- fect::estimand(fit, "att", "overall", window = c(1, 5),
                               ci.method = "basic")
  expect_equal(res_normal$se, res_basic$se, tolerance = 1e-10,
               label = "se(normal) == se(basic): shift preserves sd()")
})

## ---- P-INV-4: .compute_ci() mock test: shifted vs unshifted distribution ----

test_that("P-INV-4: .compute_ci() shifted distribution contains estimate; unshifted does not", {
  skip_on_cran()
  set.seed(100)
  boot_h0 <- rnorm(500, mean = 0, sd = 0.15)   ## H0-centered
  estimate <- 2.5
  boot_centered <- boot_h0 - mean(boot_h0) + estimate   ## shifted to point estimate

  ci_bc_shifted   <- fect:::.compute_ci(estimate, boot_centered, "bc", 0.95)
  ci_bc_unshifted <- fect:::.compute_ci(estimate, boot_h0,       "bc", 0.95)

  ## Shifted CI contains the estimate
  expect_lte(ci_bc_shifted$ci.lo, estimate,
             label = "shifted ci.lo <= estimate")
  expect_gte(ci_bc_shifted$ci.hi, estimate,
             label = "shifted ci.hi >= estimate")

  ## Shifted CI lower bound is close to estimate - 1.96*sd (Wald-like)
  expected_lo_approx <- estimate - 1.96 * sd(boot_h0)
  expect_equal(ci_bc_shifted$ci.lo, expected_lo_approx,
               tolerance = 0.15,  ## 15% tolerance: bc differs slightly from Wald
               label = "shifted ci.lo near estimate - 1.96*sd(boot_h0)")

  ## Shifted CI is non-degenerate (width > 1e-6)
  expect_gt(ci_bc_shifted$ci.hi - ci_bc_shifted$ci.lo, 1e-6,
            label = "shifted CI width > 1e-6")

  ## Unshifted (H0-centered) bc on this input would be near-degenerate
  ## (pre-fix pathology: tail quantiles all near 0, far from estimate=2.5).
  ## .compute_ci() bc has an `is_uncovered` safety fallback that detects
  ## CI not covering the estimate and substitutes the Wald CI around the
  ## estimate. So the unshifted call now returns a sensible Wald CI even
  ## though the input was pathological. We assert the fallback fires.
  width_unshifted <- ci_bc_unshifted$ci.hi - ci_bc_unshifted$ci.lo
  wald_width      <- 2 * 1.96 * sd(boot_h0)
  expect_equal(width_unshifted, wald_width, tolerance = 0.05,
               label = "H0-centered bc falls back to Wald around estimate")
  expect_true(ci_bc_unshifted$ci.lo <= estimate &&
              estimate <= ci_bc_unshifted$ci.hi,
              label = "fallback Wald CI contains the estimate")
})

## ---- P-INV-5: bca jackknife path is not broken by the shift -----------------

test_that("P-INV-5: bca on parametric fit produces non-degenerate non-NA CI", {
  skip_on_cran()
  fit <- .make_parafix_fit()
  res <- fect::estimand(fit, "att", "overall", window = c(1, 5),
                        ci.method = "bca")
  expect_s3_class(res, "data.frame")
  expect_true(!is.na(res$ci.lo),  label = "bca ci.lo is not NA")
  expect_true(!is.na(res$ci.hi),  label = "bca ci.hi is not NA")
  expect_true(!is.na(res$se),     label = "bca se is not NA")
  ## Non-degenerate CI
  expect_gt(res$ci.hi - res$ci.lo, 0.01,
            label = "bca CI width > 0.01")
})

## ---- P-EDGE-2: vartype="none" still returns NA for SE/CI --------------------

test_that("P-EDGE-2: vartype='none' returns NA se/ci.lo/ci.hi (unchanged by fix)", {
  skip_on_cran()
  fit <- .make_parafix_fit()
  res <- fect::estimand(fit, "att", "overall", window = c(1, 5),
                        vartype = "none")
  expect_true(!is.na(res$estimate), label = "estimate is not NA with vartype='none'")
  expect_true(is.na(res$se),        label = "se is NA with vartype='none'")
  expect_true(is.na(res$ci.lo),     label = "ci.lo is NA with vartype='none'")
  expect_true(is.na(res$ci.hi),     label = "ci.hi is NA with vartype='none'")
})

## ---- P-EDGE-3: bca on tiny nboots does not hard-error -----------------------

test_that("P-EDGE-3: bca with nboots=5 does not hard-error (returns NA CI, not crash)", {
  skip_on_cran()
  ## Fit with very small nboots; bca guard handles sum(valid)==0 or <2 jackknife points
  d <- make_panel_A(seed = 77)
  set.seed(77)
  fit_tiny <- suppressMessages(
    fect(Y ~ D, data = d, index = c("id", "time"),
         method = "fe", force = "two-way", se = TRUE,
         nboots = 5, CV = FALSE, keep.sims = TRUE,
         vartype = "parametric",
         time.component.from = "nevertreated",
         parallel = FALSE)
  )
  ## Should NOT throw an error — guard returns NA CI if needed
  expect_no_error(
    suppressWarnings(
      fect::estimand(fit_tiny, "att", "overall", window = c(1, 8),
                     ci.method = "bca")
    )
  )
  res <- suppressWarnings(
    fect::estimand(fit_tiny, "att", "overall", window = c(1, 8),
                   ci.method = "bca")
  )
  expect_s3_class(res, "data.frame")
  ## estimate is always computable even with tiny nboots
  expect_true(!is.na(res$estimate), label = "estimate non-NA even with tiny nboots")
})

## ---- P-EDGE-4: shift does NOT affect vartype="bootstrap" fits ---------------

test_that("P-EDGE-4: shift guard is inactive for vartype='bootstrap' (fits unaffected)", {
  skip_on_cran()
  d <- make_panel_A(seed = 42)
  set.seed(42)
  fit_boot <- suppressMessages(
    fect(Y ~ D, data = d, index = c("id", "time"),
         method = "fe", force = "two-way", se = TRUE,
         nboots = 50, CV = FALSE, keep.sims = TRUE,
         vartype = "bootstrap",
         parallel = FALSE)
  )
  expect_equal(fit_boot$vartype, "bootstrap",
               label = "fit$vartype is 'bootstrap'")
  ## bc on bootstrap: should produce a sensible CI (no shift applied)
  res <- fect::estimand(fit_boot, "att", "overall", window = c(1, 8),
                        ci.method = "bc")
  expect_s3_class(res, "data.frame")
  expect_true(!is.na(res$ci.lo), label = "bootstrap bc ci.lo non-NA")
  expect_true(!is.na(res$ci.hi), label = "bootstrap bc ci.hi non-NA")
  expect_gt(res$ci.hi - res$ci.lo, 1e-6, label = "bootstrap bc CI non-degenerate")
})

## ---- P-REG-1: pre-fix degenerate coverage is eliminated (5-rep smoke test) --

test_that("P-REG-1: bca CI contains true ATT=3 on parametric fit (smoke: 5 reps, DGP-A)", {
  skip_on_cran()
  true_att <- 3.0
  n_reps   <- 5L
  in_ci    <- logical(n_reps)
  for (r in seq_len(n_reps)) {
    d <- make_panel_A(seed = 42 + r - 1L)
    set.seed(42 + r - 1L + 5000L)
    fit <- suppressMessages(
      fect(Y ~ D, data = d, index = c("id", "time"),
           method = "fe", force = "two-way", se = TRUE,
           nboots = 200, CV = FALSE, keep.sims = TRUE,
           vartype = "parametric",
           time.component.from = "nevertreated",
           parallel = FALSE)
    )
    res <- fect::estimand(fit, "att", "overall", window = c(1, 8),
                          ci.method = "bca")
    in_ci[r] <- !is.na(res$ci.lo) &&
                res$ci.lo <= true_att &&
                true_att  <= res$ci.hi
  }
  ## At least 3 of 5 reps must cover true ATT (catastrophic failure was 0/5 pre-fix)
  expect_gte(sum(in_ci), 3L,
             label = paste0("bca coverage >= 3/5 reps (got ", sum(in_ci), "/5)"))
})

test_that("P-REG-1b: basic CI width > 0.1 on parametric fit (non-degenerate, DGP-A seed=42)", {
  skip_on_cran()
  d <- make_panel_A(seed = 42)
  set.seed(5042)
  fit <- suppressMessages(
    fect(Y ~ D, data = d, index = c("id", "time"),
         method = "fe", force = "two-way", se = TRUE,
         nboots = 200, CV = FALSE, keep.sims = TRUE,
         vartype = "parametric",
         time.component.from = "nevertreated",
         parallel = FALSE)
  )
  res_basic <- fect::estimand(fit, "att", "overall", window = c(1, 8),
                              ci.method = "basic")
  res_bca   <- fect::estimand(fit, "att", "overall", window = c(1, 8),
                              ci.method = "bca")
  ## Pre-fix: basic CI was ~[5.7, 6.3] for true ATT=3 → 0% coverage
  ## Post-fix: basic CI width must be > 0.1
  expect_gt(res_basic$ci.hi - res_basic$ci.lo, 0.1,
            label = "basic CI width > 0.1 (not degenerate)")
  ## bca CI must be non-degenerate
  expect_gt(res_bca$ci.hi - res_bca$ci.lo, 0.1,
            label = "bca CI width > 0.1 (not degenerate)")
  ## basic CI should contain or be near true ATT=3
  true_att <- 3.0
  half_width <- (res_basic$ci.hi - res_basic$ci.lo) / 2
  dist_to_ci <- max(0, res_basic$ci.lo - true_att, true_att - res_basic$ci.hi)
  expect_lte(dist_to_ci, half_width,
             label = "basic CI is within one half-width of true ATT=3")
})

## ---- All 5 ci.methods produce non-NA, non-degenerate CIs --------------------

test_that("All 5 ci.methods produce non-NA, non-degenerate CIs on parametric fit (DGP-A)", {
  skip_on_cran()
  d <- make_panel_A(seed = 42)
  set.seed(5042)
  fit <- suppressMessages(
    fect(Y ~ D, data = d, index = c("id", "time"),
         method = "fe", force = "two-way", se = TRUE,
         nboots = 200, CV = FALSE, keep.sims = TRUE,
         vartype = "parametric",
         time.component.from = "nevertreated",
         parallel = FALSE)
  )
  for (m in c("basic", "percentile", "bc", "bca", "normal")) {
    res <- fect::estimand(fit, "att", "overall", window = c(1, 8),
                          ci.method = m)
    expect_s3_class(res, "data.frame")
    expect_true(!is.na(res$ci.lo),
                label = paste(m, "ci.lo is not NA"))
    expect_true(!is.na(res$ci.hi),
                label = paste(m, "ci.hi is not NA"))
    expect_gt(res$ci.hi - res$ci.lo, 1e-6,
              label = paste(m, "CI width > 1e-6 (not degenerate)"))
  }
})

## ---- gsynth method: all 5 ci.methods produce reasonable CIs -----------------

test_that("gsynth + parametric: all 5 ci.methods produce non-NA, non-degenerate CIs (DGP-G)", {
  skip_on_cran()
  d <- make_panel_G(seed = 42, r = 2)
  set.seed(5042)
  fit_g <- suppressMessages(
    fect(Y ~ D, data = d, index = c("id", "time"),
         method = "gsynth", r = 2, se = TRUE,
         nboots = 200, CV = FALSE, keep.sims = TRUE,
         vartype = "parametric",
         parallel = FALSE)
  )
  expect_equal(fit_g$vartype, "parametric",
               label = "gsynth fit$vartype is 'parametric'")
  for (m in c("basic", "percentile", "bc", "bca", "normal")) {
    res <- fect::estimand(fit_g, "att", "overall", window = c(1, 8),
                          ci.method = m)
    expect_s3_class(res, "data.frame")
    expect_true(!is.na(res$ci.lo),
                label = paste("gsynth", m, "ci.lo non-NA"))
    expect_true(!is.na(res$ci.hi),
                label = paste("gsynth", m, "ci.hi non-NA"))
    expect_gt(res$ci.hi - res$ci.lo, 1e-6,
              label = paste("gsynth", m, "CI non-degenerate"))
    ## Estimate shifted to CI: CI should bracket estimate (or near)
    ci_width <- res$ci.hi - res$ci.lo
    dist_from_est <- max(0, res$ci.lo - res$estimate,
                            res$estimate - res$ci.hi)
    expect_lte(dist_from_est, ci_width,
               label = paste("gsynth", m, "CI within one width of estimate"))
  }
})

## ---- P-COV-1: ATT coverage >= 0.85 for {basic,percentile,bca,normal}, DGP-A --

test_that("P-COV-1: ATT coverage >= 0.85 for basic/percentile/bca/normal (DGP-A, 100 reps)", {
  skip_on_cran()
  true_att <- 3.0
  n_reps   <- 100L
  methods  <- c("basic", "percentile", "bca", "normal")

  coverage <- setNames(numeric(length(methods)), methods)

  for (r in seq_len(n_reps)) {
    d    <- make_panel_A(seed = 1000L + r)
    set.seed(1000L + r + 5000L)
    fit  <- suppressMessages(
      fect(Y ~ D, data = d, index = c("id", "time"),
           method = "fe", force = "two-way", se = TRUE,
           nboots = 200, CV = FALSE, keep.sims = TRUE,
           vartype = "parametric",
           time.component.from = "nevertreated",
           parallel = FALSE)
    )
    for (m in methods) {
      res <- fect::estimand(fit, "att", "overall", window = c(1, 8),
                            ci.method = m)
      if (!is.na(res$ci.lo) && res$ci.lo <= true_att && true_att <= res$ci.hi) {
        coverage[m] <- coverage[m] + 1L
      }
    }
  }
  coverage <- coverage / n_reps

  for (m in methods) {
    expect_gte(coverage[m], 0.85,
               label = paste0("P-COV-1: ", m, " coverage >= 0.85 (got ",
                              round(coverage[m], 3), ")"))
  }

  ## Anti-regression: any method with < 0.50 coverage is a catastrophic failure
  for (m in methods) {
    expect_gte(coverage[m], 0.50,
               label = paste0("P-COV-1 catastrophic: ", m, " coverage < 0.50"))
  }
})

## ---- P-COV-2: bca coverage >= 0.85, DGP-G, gsynth --------------------------

test_that("P-COV-2: ATT bca coverage >= 0.85 for gsynth+parametric (DGP-G, 100 reps)", {
  skip_on_cran()
  true_att <- 3.0
  n_reps   <- 100L
  in_ci    <- logical(n_reps)

  for (r in seq_len(n_reps)) {
    d    <- make_panel_G(seed = 1000L + r, r = 2)
    set.seed(1000L + r + 5000L)
    fit  <- suppressMessages(
      fect(Y ~ D, data = d, index = c("id", "time"),
           method = "gsynth", r = 2, se = TRUE,
           nboots = 200, CV = FALSE, keep.sims = TRUE,
           vartype = "parametric",
           parallel = FALSE)
    )
    res <- fect::estimand(fit, "att", "overall", window = c(1, 8),
                          ci.method = "bca")
    in_ci[r] <- !is.na(res$ci.lo) &&
                res$ci.lo <= true_att &&
                true_att  <= res$ci.hi
  }
  cov_bca <- mean(in_ci)
  expect_gte(cov_bca, 0.85,
             label = paste0("P-COV-2: gsynth bca coverage >= 0.85 (got ",
                            round(cov_bca, 3), ")"))
})

## ---- P-COV-3: APTT bca coverage >= 0.75, DGP-B-positive --------------------

test_that("P-COV-3: APTT bca coverage >= 0.75 (DGP-B-positive, 100 reps)", {
  skip_on_cran()
  n_reps <- 100L
  in_ci  <- logical(n_reps)

  for (r in seq_len(n_reps)) {
    dgp  <- make_panel_B_pos(seed = 1000L + r)
    true_aptt <- dgp$true_aptt
    set.seed(1000L + r + 5000L)
    fit  <- suppressMessages(
      fect(Y ~ D, data = dgp$df, index = c("id", "time"),
           method = "fe", force = "two-way", se = TRUE,
           nboots = 200, CV = FALSE, keep.sims = TRUE,
           vartype = "parametric",
           time.component.from = "nevertreated",
           parallel = FALSE)
    )
    res_aptt <- suppressMessages(
      fect::estimand(fit, "aptt", "event.time", ci.method = "bca")
    )
    ## Check coverage at event.time = 1
    et1 <- res_aptt[res_aptt$event.time == 1, ]
    if (nrow(et1) == 1L && !is.na(et1$ci.lo)) {
      in_ci[r] <- et1$ci.lo <= true_aptt && true_aptt <= et1$ci.hi
    }
  }
  cov_aptt <- mean(in_ci)
  expect_gte(cov_aptt, 0.75,
             label = paste0("P-COV-3: APTT bca coverage >= 0.75 (got ",
                            round(cov_aptt, 3), ")"))
})

## ---- P-WIDTH-1: parametric CI widths within 15% of bootstrap widths ---------

test_that("P-WIDTH-1: parametric CI widths within 15% of bootstrap widths (DGP-A, seed=42)", {
  skip_on_cran()
  d <- make_panel_A(seed = 42)

  set.seed(42)
  fit_par  <- suppressMessages(
    fect(Y ~ D, data = d, index = c("id", "time"),
         method = "fe", force = "two-way", se = TRUE,
         nboots = 200, CV = FALSE, keep.sims = TRUE,
         vartype = "parametric",
         time.component.from = "nevertreated",
         parallel = FALSE)
  )
  set.seed(42)
  fit_boot <- suppressMessages(
    fect(Y ~ D, data = d, index = c("id", "time"),
         method = "fe", force = "two-way", se = TRUE,
         nboots = 200, CV = FALSE, keep.sims = TRUE,
         vartype = "bootstrap",
         parallel = FALSE)
  )

  for (m in c("basic", "percentile", "normal")) {
    res_par  <- fect::estimand(fit_par,  "att", "overall", window = c(1, 8),
                               ci.method = m)
    res_boot <- fect::estimand(fit_boot, "att", "overall", window = c(1, 8),
                               ci.method = m)
    w_par  <- res_par$ci.hi  - res_par$ci.lo
    w_boot <- res_boot$ci.hi - res_boot$ci.lo

    ## Width within 15% of bootstrap width (both > 0)
    expect_gt(w_par, 0, label = paste(m, "parametric CI width > 0"))
    expect_gt(w_boot, 0, label = paste(m, "bootstrap CI width > 0"))
    rel_diff <- abs(w_par - w_boot) / w_boot
    expect_lte(rel_diff, 0.15,
               label = paste0("P-WIDTH-1: ", m,
                              " parametric vs bootstrap width within 15% (rel_diff = ",
                              round(rel_diff, 3), ")"))
  }
})

## ---------------------------------------------------------------
## Tests for estimand(test = "placebo" / "carryover").
##
## Issue #131 (ajunquera): pre-treatment APTT estimates for the
## alternative-estimand API. Generalized to all four shipped estimand
## types (att, aptt, log.att; att.cumu is intentionally disallowed).
## ---------------------------------------------------------------

suppressWarnings(data("simdata", package = "fect"))


## DGP helper duplicated from test-book-claims.R so this file is
## self-contained without cross-file source ordering.
.make_panel <- function(N = 40, TT = 20, T0 = 12, Ntr = 12,
                        tau = 3.0, seed = 9301, reversals = FALSE) {
    set.seed(seed)
    alpha_i <- rnorm(N, 0, 2)
    xi_t    <- rnorm(TT, 0, 1)
    D <- matrix(0L, TT, N)
    D[(T0 + 1):TT, 1:Ntr] <- 1L
    if (reversals) {
        for (i in 1:min(3, Ntr)) D[(TT - 1):TT, i] <- 0L
    }
    eps <- matrix(rnorm(N * TT, 0, 1), TT, N)
    Y <- outer(xi_t, rep(1, N)) +
         outer(rep(1, TT), alpha_i) +
         tau * D + eps
    data.frame(
        id   = rep(1:N, each = TT),
        time = rep(1:TT, N),
        Y    = as.vector(Y),
        D    = as.vector(D)
    )
}


.fit_placebo <- function(nboots = 50) {
    d <- .make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12, seed = 9301)
    suppressWarnings(suppressMessages(
        fect::fect(Y ~ D, data = d, index = c("id", "time"),
                   method = "fe", se = TRUE, nboots = nboots,
                   parallel = FALSE, keep.sims = TRUE,
                   placeboTest = TRUE, placebo.period = c(-2, 0),
                   CV = FALSE)
    ))
}

.fit_carryover <- function(nboots = 50) {
    d <- .make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12,
                     seed = 9302, reversals = TRUE)
    suppressWarnings(suppressMessages(
        fect::fect(Y ~ D, data = d, index = c("id", "time"),
                   method = "fe", se = TRUE, nboots = nboots,
                   parallel = FALSE, keep.sims = TRUE,
                   carryoverTest = TRUE, carryover.period = c(1, 2),
                   CV = FALSE)
    ))
}


## -- PC.1  test argument is documented and validated ---------------

test_that("PC.1: estimand() rejects unknown test values", {

  skip_on_cran()

  fit <- .fit_placebo()
  expect_error(
    fect::estimand(fit, "att", "event.time", test = "bogus"),
    "should be one of"
  )
})


test_that("PC.2: test='placebo' errors when fit had placeboTest=FALSE", {

  skip_on_cran()

  fit_plain <- suppressWarnings(suppressMessages(
    fect::fect(Y ~ D, data = simdata, index = c("id", "time"),
               method = "fe", force = "two-way",
               se = TRUE, nboots = 50, parallel = FALSE,
               keep.sims = TRUE)
  ))
  expect_error(
    fect::estimand(fit_plain, "aptt", "event.time", test = "placebo"),
    "placeboTest = TRUE"
  )
  ## Also verify the message names the actionable refit hint.
  expect_error(
    fect::estimand(fit_plain, "aptt", "event.time", test = "placebo"),
    "Refit with"
  )
})


test_that("PC.3: test='carryover' errors when fit had carryoverTest=FALSE", {

  skip_on_cran()

  fit_plain <- suppressWarnings(suppressMessages(
    fect::fect(Y ~ D, data = simdata, index = c("id", "time"),
               method = "fe", force = "two-way",
               se = TRUE, nboots = 50, parallel = FALSE,
               keep.sims = TRUE)
  ))
  expect_error(
    fect::estimand(fit_plain, "aptt", "event.time", test = "carryover"),
    "carryoverTest = TRUE"
  )
  expect_error(
    fect::estimand(fit_plain, "aptt", "event.time", test = "carryover"),
    "Refit with"
  )
})


## -- PC.4  test='placebo' returns event-time series in placebo window

test_that("PC.4: placebo aptt returns rows in fit$placebo.period range", {

  skip_on_cran()

  fit <- .fit_placebo()
  est <- fect::estimand(fit, "aptt", "event.time", test = "placebo")

  expect_s3_class(est, "data.frame")
  expect_setequal(names(est),
                  c("event.time", "estimate", "se",
                    "ci.lo", "ci.hi", "n_cells", "vartype"))

  pp <- fit$placebo.period
  expect_true(all(est$event.time >= pp[1] & est$event.time <= pp[2]))
  expect_true(all(est$n_cells > 0L))
})


## -- PC.5  test='placebo' covers att (log.att is gated by v2.4.2 hard-error)

test_that("PC.5: placebo att returns rows in placebo window; placebo log.att hard-errors when DGP triggers cell-drop pathology", {

  skip_on_cran()

  fit <- .fit_placebo()

  est_att <- fect::estimand(fit, "att", "event.time", test = "placebo")

  pp <- fit$placebo.period
  expect_true(all(est_att$event.time >= pp[1] & est_att$event.time <= pp[2]))

  ## log.att on simdata triggers v2.4.2's point-level hard-error because
  ## simdata has Y <= 0 cells. This is the EXPECTED behavior --- the
  ## point-level check fires before any bootstrap work, with a clearer
  ## actionable message than the bootstrap-level pathology message.
  expect_error(
    fect::estimand(fit, "log.att", "event.time", test = "placebo"),
    "log\\.att requires Y > 0"
  )
})


## -- PC.6  byte-equality: placebo att rows match fit$est.att rows -----

test_that("PC.6: estimand(att, test=placebo) matches fit$est.att rows", {

  skip_on_cran()

  fit <- .fit_placebo()
  est <- fect::estimand(fit, "att", "event.time", test = "placebo")

  ## fit$est.att is per-event-time over all (treated post + masked
  ## placebo) cells. Restrict to placebo event times and compare.
  ea <- fit$est.att
  et <- as.numeric(rownames(ea))
  pp <- fit$placebo.period
  in_pp <- et >= pp[1] & et <= pp[2]

  ## Sort both by event.time for safety.
  est <- est[order(est$event.time), , drop = FALSE]
  ref <- ea[in_pp, , drop = FALSE]
  ref <- ref[order(as.numeric(rownames(ref))), , drop = FALSE]

  expect_equal(est$event.time,
               as.numeric(rownames(ref)))
  expect_equal(est$estimate,
               unname(ref[, "ATT"]),
               tolerance = 1e-10)
  expect_equal(est$n_cells,
               as.integer(unname(ref[, "count"])))
})


## -- PC.7  type='att.cumu' is rejected with non-none test ------------

test_that("PC.7: type='att.cumu' is incompatible with test != 'none'", {

  skip_on_cran()

  fit <- .fit_placebo()
  expect_error(
    fect::estimand(fit, "att.cumu", "event.time", test = "placebo"),
    "incompatible with test"
  )
})


## -- PC.8  carryover path works on a reversal panel ------------------

test_that("PC.8: carryover aptt returns rows in carryover window", {

  skip_on_cran()

  fit <- .fit_carryover()
  est <- fect::estimand(fit, "aptt", "event.time", test = "carryover")

  cp <- fit$carryover.period
  expect_true(all(est$event.time >= cp[1] & est$event.time <= cp[2]))
  expect_true(all(est$n_cells > 0L))
})


## -- PC.9  test='placebo' auto-overrides direction to 'on' -----------

test_that("PC.9: test='placebo' silently uses direction='on'", {

  skip_on_cran()

  fit <- .fit_placebo()
  ## Even when user passes direction='off', placebo path uses T.on
  ## (so we should get an event-time series, not error on missing T.off).
  est <- fect::estimand(fit, "aptt", "event.time",
                        test = "placebo", direction = "off")
  pp <- fit$placebo.period
  expect_true(all(est$event.time >= pp[1] & est$event.time <= pp[2]))
})


## -- PC.10 default vartype is sourced from fit, not from arg ---------

test_that("PC.10: vartype column reports method used at fit time", {

  skip_on_cran()

  fit <- .fit_placebo()
  est <- fect::estimand(fit, "aptt", "event.time", test = "placebo")
  expect_true(unique(est$vartype) %in%
              c("bootstrap", "jackknife", "parametric"))
})

## Unit tests for para.error argument (spec: wild-as-paraerror)
##
## Tests verify:
##   1. para.error = "auto" default produces correct resolved mode stored on fit
##   2. para.error = "empirical" runs without error on fully-observed panel
##   3. para.error = "wild" runs without error on fully-observed panel, all 5
##      ci.methods return CIs containing the true ATT
##   4. para.error = "wild" hard-errors on missing-data panel
##   5. para.error = "empirical" hard-errors on missing-data panel
##   6. Invalid para.error value produces a clear error message
##   7. para.error is ignored when vartype != "parametric"
##   8. para.error = "ar" works on fully-observed panel
##   9. para.error = "auto" resolves to "ar" on missing-data panel

## -----------------------------------------------------------------------
## Shared test fixture
## -----------------------------------------------------------------------

make_panel <- function(N = 40, TT = 20, T0 = 12, Ntr = 12, seed = 101,
                       add_missing = FALSE, n_missing = 50) {
    set.seed(seed)
    alpha_i <- rnorm(N, 0, 2)
    xi_t    <- rnorm(TT, 0, 1)
    D       <- matrix(0L, TT, N)
    D[(T0 + 1):TT, 1:Ntr] <- 1L
    eps <- matrix(rnorm(N * TT, 0, 1), TT, N)
    Y   <- outer(xi_t, rep(1, N)) + outer(rep(1, TT), alpha_i) + 3.0 * D + eps
    d   <- data.frame(
        id   = rep(1:N, each = TT),
        time = rep(1:TT, N),
        Y    = as.vector(Y),
        D    = as.vector(D)
    )
    if (add_missing) {
        set.seed(seed + 999)
        d$Y[sample(nrow(d), n_missing)] <- NA
    }
    d
}

fect_para <- function(d, para.error = "auto", vartype = "parametric",
                      nboots = 100, seed = 42, ...) {
    set.seed(seed)
    suppressWarnings(suppressMessages(
        fect(
            Y ~ D,
            data   = d,
            index  = c("id", "time"),
            method = "fe",
            force  = "two-way",
            time.component.from = "nevertreated",
            se      = TRUE,
            vartype = vartype,
            para.error = para.error,
            nboots     = nboots,
            parallel   = FALSE,
            keep.sims  = TRUE,
            CV         = FALSE,
            ...
        )
    ))
}

## -----------------------------------------------------------------------
## Test 1: auto default resolves to "empirical" for fully-observed panel
## -----------------------------------------------------------------------

test_that("para.error = 'auto' resolves to 'empirical' on fully-observed panel", {
    skip_on_cran()
    d   <- make_panel()
    fit <- fect_para(d, para.error = "auto")
    expect_equal(fit$para.error, "empirical")
    expect_equal(fit$vartype, "parametric")
})

## -----------------------------------------------------------------------
## Test 2: para.error = "empirical" runs and stores correct label
## -----------------------------------------------------------------------

test_that("para.error = 'empirical' runs on fully-observed panel", {
    skip_on_cran()
    d   <- make_panel()
    fit <- fect_para(d, para.error = "empirical")
    expect_equal(fit$para.error, "empirical")

    ## CI should be sensible (contains true ATT ≈ 3)
    res <- estimand(fit, "att", "overall", window = c(1, 8), ci.method = "bca")
    expect_true(res$ci.lo < 3.0 && res$ci.hi > 3.0,
        info = sprintf("BCA CI [%.3f, %.3f] should contain 3.0", res$ci.lo, res$ci.hi))
})

## -----------------------------------------------------------------------
## Test 3: para.error = "wild" runs and all ci.methods return sensible CIs
## -----------------------------------------------------------------------

test_that("para.error = 'wild' runs on fully-observed panel and all ci.methods work", {
    skip_on_cran()
    d   <- make_panel()
    fit <- fect_para(d, para.error = "wild", nboots = 200)
    expect_equal(fit$para.error, "wild")
    expect_equal(fit$vartype, "parametric")

    for (m in c("normal", "basic", "percentile", "bc", "bca")) {
        res <- estimand(fit, "att", "overall", window = c(1, 8), ci.method = m)
        expect_true(
            !is.na(res$ci.lo) && !is.na(res$ci.hi),
            info = sprintf("ci.method='%s': CI should not be NA", m)
        )
        expect_true(
            res$ci.lo < 3.0 && res$ci.hi > 3.0,
            info = sprintf(
                "ci.method='%s': CI [%.3f, %.3f] should contain 3.0",
                m, res$ci.lo, res$ci.hi
            )
        )
        expect_true(
            res$ci.hi - res$ci.lo > 0,
            info = sprintf("ci.method='%s': CI width should be positive", m)
        )
    }
})

## -----------------------------------------------------------------------
## Test 4: para.error = "wild" hard-errors on missing-data panel
## -----------------------------------------------------------------------

test_that("para.error = 'wild' hard-errors on missing-data panel", {
    skip_on_cran()
    d <- make_panel(add_missing = TRUE)
    expect_error(
        fect(
            Y ~ D,
            data   = d,
            index  = c("id", "time"),
            method = "fe",
            force  = "two-way",
            time.component.from = "nevertreated",
            se      = TRUE,
            vartype = "parametric",
            para.error = "wild",
            na.rm   = FALSE,
            nboots  = 50,
            parallel = FALSE,
            CV = FALSE
        ),
        regexp = "fully-observed panel",
        fixed  = FALSE
    )
})

## -----------------------------------------------------------------------
## Test 6: para.error = "empirical" hard-errors on missing-data panel
## -----------------------------------------------------------------------

test_that("para.error = 'empirical' hard-errors on missing-data panel", {
    skip_on_cran()
    d <- make_panel(add_missing = TRUE)
    expect_error(
        fect(
            Y ~ D,
            data   = d,
            index  = c("id", "time"),
            method = "fe",
            force  = "two-way",
            time.component.from = "nevertreated",
            se      = TRUE,
            vartype = "parametric",
            para.error = "empirical",
            na.rm   = FALSE,
            nboots  = 50,
            parallel = FALSE,
            CV = FALSE
        ),
        regexp = "fully-observed panel",
        fixed  = FALSE
    )
})

## -----------------------------------------------------------------------
## Test 7: Invalid para.error value produces clear error
## -----------------------------------------------------------------------

test_that("invalid para.error value produces clear error", {
    skip_on_cran()
    d <- make_panel()
    expect_error(
        fect(
            Y ~ D,
            data   = d,
            index  = c("id", "time"),
            method = "fe",
            force  = "two-way",
            time.component.from = "nevertreated",
            se      = TRUE,
            vartype = "parametric",
            para.error = "invalid_value",
            nboots  = 50,
            parallel = FALSE,
            CV = FALSE
        ),
        regexp = "para.error",
        fixed  = FALSE
    )
})

## -----------------------------------------------------------------------
## Test 8: para.error is ignored when vartype != "parametric"
## -----------------------------------------------------------------------

test_that("para.error is accepted but ignored for vartype = 'bootstrap'", {
    skip_on_cran()
    d   <- make_panel()
    ## Should run without error even though para.error = "wild" is passed
    ## with vartype = "bootstrap" (non-parametric path ignores it)
    fit <- NULL
    expect_no_error({
        set.seed(42)
        fit <- suppressWarnings(suppressMessages(fect(
            Y ~ D,
            data   = d,
            index  = c("id", "time"),
            method = "fe",
            force  = "two-way",
            time.component.from = "nevertreated",
            se      = TRUE,
            vartype = "bootstrap",
            para.error = "wild",
            nboots  = 50,
            parallel = FALSE,
            CV = FALSE
        )))
    })
    ## vartype is bootstrap, para.error on fit object should be NULL
    ## (fect_boot para.error.resolved is only set in the parametric branch)
    expect_equal(fit$vartype, "bootstrap")
})

## -----------------------------------------------------------------------
## Test 9: para.error = "ar" works on fully-observed panel
## -----------------------------------------------------------------------

test_that("para.error = 'ar' runs on fully-observed panel", {
    skip_on_cran()
    d   <- make_panel()
    fit <- fect_para(d, para.error = "ar")
    expect_equal(fit$para.error, "ar")

    res <- estimand(fit, "att", "overall", window = c(1, 8), ci.method = "bca")
    expect_true(res$ci.lo < 3.0 && res$ci.hi > 3.0,
        info = sprintf("BCA CI [%.3f, %.3f] should contain 3.0", res$ci.lo, res$ci.hi))
})

## -----------------------------------------------------------------------
## Test 10: para.error = "auto" resolves to "ar" on missing-data panel
## -----------------------------------------------------------------------

test_that("para.error = 'auto' resolves to 'ar' on missing-data panel", {
    skip_on_cran()
    d   <- make_panel(add_missing = TRUE)
    ## na.rm = FALSE needed so missing cells are preserved in I
    fit <- suppressWarnings(suppressMessages(fect(
        Y ~ D,
        data   = d,
        index  = c("id", "time"),
        method = "fe",
        force  = "two-way",
        time.component.from = "nevertreated",
        se      = TRUE,
        vartype = "parametric",
        para.error = "auto",
        na.rm   = FALSE,
        nboots  = 100,
        parallel = FALSE,
        keep.sims = TRUE,
        CV = FALSE
    )))
    expect_equal(fit$para.error, "ar")
})

## ---------------------------------------------------------------
## Tests for estimand(fit, "att", ...) and the backcompat invariant
## with fit$est.att.
##
## Locked invariant per ref/po-estimands-contract.md §2:
##   estimand(fit, "att", "event.time") matches fit$est.att columns
##   ATT / S.E. / CI.lower / CI.upper byte-for-byte under default args.
## ---------------------------------------------------------------

suppressWarnings(data("simdata", package = "fect"))


.fit_canonical <- function(nboots = 50, keep.sims = FALSE) {
  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = simdata, index = c("id", "time"),
      method = "fe", force = "two-way",
      se = TRUE, nboots = nboots, parallel = FALSE,
      keep.sims = keep.sims
    )
  ))
}


## -- A.1  Schema: returns documented columns -------------------------

test_that("A.1: estimand(fit, 'att', 'event.time') returns the documented schema", {

  skip_on_cran()

  fit <- .fit_canonical()
  est <- fect::estimand(fit, "att", "event.time")

  expect_s3_class(est, "data.frame")
  expected_cols <- c("event.time", "estimate", "se",
                     "ci.lo", "ci.hi", "n_cells", "vartype")
  expect_setequal(names(est), expected_cols)

  expect_true(is.numeric(est$event.time))
  expect_true(is.numeric(est$estimate))
  expect_true(is.numeric(est$se))
  expect_true(is.numeric(est$n_cells))
})


## -- A.2  Backcompat: byte-equality with fit$est.att (LOCKED INVARIANT)

test_that("A.2: estimand(fit, 'att', 'event.time') matches fit$est.att byte-for-byte", {

  skip_on_cran()

  fit <- .fit_canonical()
  est <- fect::estimand(fit, "att", "event.time")

  ## fit$est.att rows are labeled by event time (rownames); match the
  ## same axis ordering.
  est_att <- as.data.frame(fit$est.att)
  est_att$event.time <- as.numeric(rownames(fit$est.att))
  est_att <- est_att[order(est_att$event.time), ]
  est <- est[order(est$event.time), ]

  expect_equal(est$event.time, est_att$event.time)
  expect_identical(est$estimate, unname(est_att$ATT))
  expect_identical(est$se,       unname(est_att$S.E.))
  expect_identical(est$ci.lo,    unname(est_att$CI.lower))
  expect_identical(est$ci.hi,    unname(est_att$CI.upper))
  expect_identical(est$n_cells,  unname(est_att$count))
})


## -- A.3  Errors: unsupported types in this commit -------------------

test_that("A.3: unimplemented type raises a clear message", {

  skip_on_cran()

  ## "att.cumu", "aptt", "log.att" all implemented; A.3 has nothing
  ## left to assert.
  expect_true(TRUE)
})


## -- A.4  Errors: invalid by ----------------------------------------

test_that("A.4: invalid by raises a useful error", {

  skip_on_cran()

  fit <- .fit_canonical()
  expect_error(
    fect::estimand(fit, "att", "nope"),
    "user-column",
    fixed = FALSE
  )
})


## -- A.5  Errors: bad fit ------------------------------------------

test_that("A.5: passing non-fect to estimand errors clearly", {
  expect_error(
    fect::estimand("not a fit", "att", "event.time"),
    "fect"
  )
})

## ---------------------------------------------------------------
## Tests for estimand() and imputed_outcomes() behavior on jackknife fits.
##
## Covers:
##   J.1  Slot contract: jackknife eff.boot passes .validate_po_contract
##   J.2  Fast path: event.time + normal is byte-equal to fit$est.att
##   J.3  Overall path: returns finite Wald CI
##   J.4  ci.method guard: bca hard-errors with explanation
##   J.5  ci.method guard: percentile hard-errors with explanation
##   J.6  ci.method guard: basic hard-errors
##   J.7  ci.method guard: bc hard-errors
##   J.8  imputed_outcomes(replicates = TRUE) hard-errors
##   J.9  Overall with cells window: returns finite Wald CI
##   J.10 aptt event.time under jackknife
##   J.11 log.att event.time under jackknife
##   J.12 Anti-regression: bootstrap estimand still works after fix
## ---------------------------------------------------------------

## DGP-A: balanced two-way FE, true ATT = 3.0
.make_jackknife_data <- function(seed = 42, N = 40, TT = 20, tr_start = 11,
                                  tau = 3.0) {
  set.seed(seed)
  time_idx  <- rep(1:TT, each = N)
  unit_idx  <- rep(1:N, times = TT)
  is_treated <- unit_idx <= (N / 2)
  D  <- as.integer(is_treated & time_idx >= tr_start)
  alpha_i <- rnorm(N, sd = 1)
  gamma_t <- rnorm(TT, sd = 0.5)
  Y0 <- outer(gamma_t, alpha_i, "+") + rnorm(TT * N, sd = 0.5)
  Y  <- as.vector(Y0) + tau * D
  data.frame(Y = Y, D = D, id = unit_idx, time = time_idx)
}

.fit_jack <- function(simdata, keep.sims = TRUE) {
  set.seed(1)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = simdata, index = c("id", "time"),
      method = "fe", force = "two-way",
      se = TRUE, vartype = "jackknife",
      parallel = FALSE, keep.sims = keep.sims, CV = FALSE
    )
  ))
}

.fit_boot <- function(simdata, nboots = 50, keep.sims = TRUE) {
  set.seed(1)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = simdata, index = c("id", "time"),
      method = "fe", force = "two-way",
      se = TRUE, vartype = "bootstrap", nboots = nboots,
      parallel = FALSE, keep.sims = keep.sims, CV = FALSE
    )
  ))
}

simdata_A <- .make_jackknife_data()


## -- J.1  Slot contract passes for jackknife eff.boot ---------------

test_that("J.1: .validate_po_contract passes for jackknife fit", {
  skip_on_cran()
  fit <- .fit_jack(simdata_A)
  ## Slot contract is called inside estimand(); if it fails we'd get an error.
  ## Calling the fast path (no eff.boot needed) is sufficient to confirm no error.
  expect_no_error(fect::estimand(fit, "att", "event.time", ci.method = "normal"))
  ## Also confirm eff.boot dimensions match jackknife expectation.
  expect_equal(dim(fit$eff.boot)[3], ncol(fit$Y.dat))  ## third dim = N
  expect_equal(dim(fit$eff.boot)[2], ncol(fit$Y.dat) - 1L)  ## second dim = N-1
})


## -- J.2  Fast path: event.time + normal byte-equals fit$est.att ----

test_that("J.2: event.time + normal is byte-equal to fit$est.att", {
  skip_on_cran()
  fit <- .fit_jack(simdata_A)
  est <- fect::estimand(fit, "att", "event.time", ci.method = "normal")

  tol <- 1e-10
  expect_equal(nrow(est), nrow(fit$est.att))
  expect_equal(est$event.time, as.numeric(rownames(fit$est.att)))
  expect_lt(max(abs(est$estimate - unname(fit$est.att[, "ATT"]))), tol)
  expect_lt(max(abs(est$se      - unname(fit$est.att[, "S.E."]))), tol)
  expect_lt(max(abs(est$ci.lo   - unname(fit$est.att[, "CI.lower"]))), tol)
  expect_lt(max(abs(est$ci.hi   - unname(fit$est.att[, "CI.upper"]))), tol)
  expect_true(all(est$vartype == "jackknife"))
})


## -- J.3  Overall path: returns finite Wald CI ----------------------

test_that("J.3: overall + normal returns finite Wald CI", {
  skip_on_cran()
  fit <- .fit_jack(simdata_A)
  est <- fect::estimand(fit, "att", "overall", ci.method = "normal")

  expect_equal(nrow(est), 1L)
  expect_true(is.finite(est$estimate))
  expect_true(is.finite(est$se) && est$se > 0)
  expect_true(est$ci.lo < est$estimate)
  expect_true(est$estimate < est$ci.hi)
  expect_equal(est$vartype, "jackknife")

  ## Wald consistency: ci width == 2 * z * se
  expected_width <- 2 * stats::qnorm(0.975) * est$se
  expect_lt(abs((est$ci.hi - est$ci.lo) - expected_width), 1e-10)

  ## Point estimate matches mean(eff[treated])
  D_mat   <- fit$D.dat
  treated_mean <- mean(fit$eff[!is.na(D_mat) & D_mat == 1], na.rm = TRUE)
  expect_lt(abs(est$estimate - treated_mean), 1e-10)
})


## -- J.4  ci.method guard: bca hard-errors --------------------------

test_that("J.4: ci.method = 'bca' hard-errors for jackknife fit", {
  skip_on_cran()
  fit <- .fit_jack(simdata_A)
  err <- tryCatch(
    fect::estimand(fit, "att", "event.time", ci.method = "bca"),
    error = function(e) conditionMessage(e)
  )
  expect_match(err, "jackknife", ignore.case = TRUE)
  expect_match(err, "standard error|SE|sampling distribution", ignore.case = TRUE)
  expect_match(err, "normal", ignore.case = TRUE)
})


## -- J.5  ci.method guard: percentile hard-errors -------------------

test_that("J.5: ci.method = 'percentile' hard-errors for jackknife fit", {
  skip_on_cran()
  fit <- .fit_jack(simdata_A)
  err <- tryCatch(
    fect::estimand(fit, "att", "event.time", ci.method = "percentile"),
    error = function(e) conditionMessage(e)
  )
  expect_match(err, "jackknife", ignore.case = TRUE)
  expect_match(err, "normal", ignore.case = TRUE)
})


## -- J.6  ci.method guard: basic hard-errors ------------------------

test_that("J.6: ci.method = 'basic' hard-errors for jackknife fit", {
  skip_on_cran()
  fit <- .fit_jack(simdata_A)
  err <- tryCatch(
    fect::estimand(fit, "att", "event.time", ci.method = "basic"),
    error = function(e) conditionMessage(e)
  )
  expect_match(err, "jackknife", ignore.case = TRUE)
})


## -- J.7  ci.method guard: bc hard-errors ---------------------------

test_that("J.7: ci.method = 'bc' hard-errors for jackknife fit", {
  skip_on_cran()
  fit <- .fit_jack(simdata_A)
  err <- tryCatch(
    fect::estimand(fit, "att", "event.time", ci.method = "bc"),
    error = function(e) conditionMessage(e)
  )
  expect_match(err, "jackknife", ignore.case = TRUE)
})


## -- J.8  imputed_outcomes(replicates = TRUE) hard-errors -----------

test_that("J.8: imputed_outcomes(replicates = TRUE) hard-errors for jackknife", {
  skip_on_cran()
  fit <- .fit_jack(simdata_A)
  err <- tryCatch(
    fect::imputed_outcomes(fit, replicates = TRUE),
    error = function(e) conditionMessage(e)
  )
  expect_match(err, "jackknife", ignore.case = TRUE)
  expect_match(err, "column|N-1|dimension", ignore.case = TRUE)
})


## -- J.9  Overall + cells window: returns finite Wald CI ------------

test_that("J.9: overall + window returns finite Wald CI from cells-filter branch", {
  skip_on_cran()
  fit <- .fit_jack(simdata_A)
  est_full   <- fect::estimand(fit, "att", "overall", ci.method = "normal")
  est_window <- fect::estimand(fit, "att", "overall", ci.method = "normal",
                               window = c(1, 5))

  expect_equal(nrow(est_window), 1L)
  expect_true(is.finite(est_window$estimate))
  expect_true(is.finite(est_window$se) && est_window$se > 0)
  expect_true(est_window$ci.lo < est_window$estimate)
  expect_true(est_window$estimate < est_window$ci.hi)

  ## Wald consistency
  expected_width <- 2 * stats::qnorm(0.975) * est_window$se
  expect_lt(abs((est_window$ci.hi - est_window$ci.lo) - expected_width), 1e-10)

  ## Window estimate differs from full estimate (different cell set)
  expect_false(isTRUE(all.equal(est_window$estimate, est_full$estimate)))
})


## -- J.10 aptt event.time under jackknife ---------------------------

test_that("J.10: aptt event.time returns finite SE/CI under jackknife", {
  skip_on_cran()
  ## DGP-B: Y > 0 everywhere for aptt
  set.seed(42)
  simdata_B <- simdata_A
  simdata_B$Y <- abs(simdata_A$Y) + 5
  fit <- .fit_jack(simdata_B)

  est <- fect::estimand(fit, "aptt", "event.time", ci.method = "normal")
  expect_s3_class(est, "data.frame")
  expect_true(all(est$vartype == "jackknife"))

  post_rows <- est[!is.na(est$se) & is.finite(est$se), ]
  expect_gt(nrow(post_rows), 0L)
  ## All finite-SE rows have valid CIs
  expect_true(all(post_rows$ci.lo < post_rows$estimate))
  expect_true(all(post_rows$estimate < post_rows$ci.hi))
  ## Wald consistency for all finite-SE rows
  widths <- post_rows$ci.hi - post_rows$ci.lo
  expected <- 2 * stats::qnorm(0.975) * post_rows$se
  expect_lt(max(abs(widths - expected)), 1e-10)
})


## -- J.11 log.att event.time under jackknife ------------------------

test_that("J.11: log.att event.time returns finite SE/CI under jackknife", {
  skip_on_cran()
  set.seed(42)
  simdata_B <- simdata_A
  simdata_B$Y <- abs(simdata_A$Y) + 5
  fit <- .fit_jack(simdata_B)

  est <- fect::estimand(fit, "log.att", "event.time", ci.method = "normal")
  expect_s3_class(est, "data.frame")
  expect_true(all(est$vartype == "jackknife"))

  post_rows <- est[!is.na(est$se) & is.finite(est$se), ]
  expect_gt(nrow(post_rows), 0L)
  expect_true(all(post_rows$ci.lo < post_rows$estimate))
  expect_true(all(post_rows$estimate < post_rows$ci.hi))
  ## Wald consistency
  widths <- post_rows$ci.hi - post_rows$ci.lo
  expected <- 2 * stats::qnorm(0.975) * post_rows$se
  expect_lt(max(abs(widths - expected)), 1e-10)
})


## -- J.12 Anti-regression: bootstrap estimand still works ----------

test_that("J.12: bootstrap estimand (att, event.time) unaffected by jackknife fix", {
  skip_on_cran()
  fit <- .fit_boot(simdata_A)
  est <- fect::estimand(fit, "att", "event.time")

  tol <- 1e-10
  expect_equal(nrow(est), nrow(fit$est.att))
  expect_lt(max(abs(est$estimate - unname(fit$est.att[, "ATT"]))), tol)
  expect_true(all(est$vartype == "bootstrap"))

  ## overall also works
  est_ov <- fect::estimand(fit, "att", "overall")
  expect_true(is.finite(est_ov$estimate))
  expect_true(is.finite(est_ov$se))
})




## -- S-11 extended  Anti-regression: bootstrap path unchanged -------------------
## Tests the three calls from test-spec.md §13:
##   est_boot_et  (att, event.time, default ci.method)
##   est_boot_ov  (att, overall, default ci.method)
##   est_boot_pct (att, overall, ci.method = "percentile")
##
## Note: estimand(boot, "att", "event.time", ci.method="percentile") currently
## raises "not yet implemented" for by="event.time" + non-normal ci.method on
## a non-placebo/carryover test (pre-existing limitation, not a regression
## introduced by the jackknife fix). We use by="overall" + "percentile" instead,
## which IS supported and exercises the same downstream .compute_ci() code path.

test_that("S-11: bootstrap att/event.time byte-equals fit$est.att (anti-regression)", {
  skip_on_cran()
  fit <- .fit_boot(simdata_A, nboots = 200)
  est_boot_et <- fect::estimand(fit, "att", "event.time")
  expect_s3_class(est_boot_et, "data.frame")
  tol <- 1e-10
  expect_lt(max(abs(est_boot_et$estimate - unname(fit$est.att[, "ATT"]))), tol)
  expect_true(all(est_boot_et$vartype == "bootstrap"))
})

test_that("S-11: bootstrap att/overall is finite (anti-regression)", {
  skip_on_cran()
  fit <- .fit_boot(simdata_A, nboots = 200)
  est_boot_ov <- fect::estimand(fit, "att", "overall")
  expect_true(is.finite(est_boot_ov$estimate))
  expect_true(is.finite(est_boot_ov$se))
  expect_equal(est_boot_ov$vartype, "bootstrap")
})

test_that("S-11: bootstrap att/overall percentile returns valid CI (anti-regression)", {
  skip_on_cran()
  fit <- .fit_boot(simdata_A, nboots = 200)
  est_boot_pct <- fect::estimand(fit, "att", "overall",
                                  ci.method = "percentile")
  expect_s3_class(est_boot_pct, "data.frame")
  expect_equal(est_boot_pct$vartype, "bootstrap")
  expect_true(is.finite(est_boot_pct$ci.lo))
  expect_true(is.finite(est_boot_pct$ci.hi))
})


## -- S-12  Anti-regression: parametric fit unchanged -------------------------
## Parametric bootstrap requires never-treated control units, so we use a DGP
## where only 12 of 40 units are treated (units 1..12, periods 13..20), leaving
## 28 never-treated controls. This matches the DGP convention from
## test-estimand-parametric-cifix.R (make_panel_A).

.make_param_data <- function(seed = 42) {
  set.seed(seed)
  N <- 40L; TT <- 20L; T0 <- 12L; Ntr <- 12L
  alpha_i <- rnorm(N, 0, 2)
  xi_t    <- rnorm(TT, 0, 1)
  D       <- matrix(0L, TT, N)
  D[(T0 + 1L):TT, 1L:Ntr] <- 1L
  eps     <- matrix(rnorm(N * TT, 0, 1), TT, N)
  Y       <- outer(xi_t, rep(1, N)) + outer(rep(1, TT), alpha_i) + 3.0 * D + eps
  data.frame(id   = rep(1:N, each = TT),
             time = rep(1:TT, N),
             Y    = as.vector(Y),
             D    = as.vector(D))
}

.fit_param_s12 <- local({
  cached <- NULL
  function() {
    if (!is.null(cached)) return(cached)
    skip_on_cran()
    d <- .make_param_data()
    set.seed(42)
    cached <<- suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D, data = d, index = c("id", "time"),
        method = "fe", force = "two-way",
        se = TRUE, vartype = "parametric", nboots = 200,
        time.component.from = "nevertreated",
        parallel = FALSE, keep.sims = TRUE, CV = FALSE
      )
    ))
    cached
  }
})

test_that("S-12: parametric fit estimand(att, event.time) returns 'parametric' vartype", {
  skip_on_cran()
  fit <- .fit_param_s12()
  est <- fect::estimand(fit, "att", "event.time")
  expect_s3_class(est, "data.frame")
  expect_true(all(est$vartype == "parametric"))
})

test_that("S-12: parametric fit estimand(att, overall) is finite and 'parametric'", {
  skip_on_cran()
  fit <- .fit_param_s12()
  est <- fect::estimand(fit, "att", "overall")
  expect_s3_class(est, "data.frame")
  expect_true(is.finite(est$estimate))
  expect_true(is.finite(est$se))
  expect_equal(est$vartype, "parametric")
})


## -- S-SIM  100-rep coverage simulation: jackknife + normal, DGP-A -----------
##
## Acceptance criterion: coverage_jack >= 0.85 (true ATT = 3.0).
## Uses fit$est.avg to avoid any slot-contract overhead inside the loop;
## after the fix this equals estimand(fit, "att", "overall")$ci.lo/ci.hi.

.run_coverage_jack <- function(seed, N = 40, TT = 20, tau = 3.0,
                                tr_start = 11) {
  set.seed(seed)
  n_treated  <- N / 2
  unit_idx   <- rep(1:N, times = TT)
  time_idx   <- rep(1:TT, each = N)
  is_treated <- unit_idx <= n_treated
  D          <- as.integer(is_treated & time_idx >= tr_start)
  alpha_i    <- rnorm(N, sd = 1)
  gamma_t    <- rnorm(TT, sd = 0.5)
  Y0  <- outer(gamma_t, alpha_i, "+")[cbind(time_idx, unit_idx)] +
          rnorm(N * TT, sd = 0.5)
  Y   <- Y0 + tau * D
  dat <- data.frame(Y = Y, D = D, id = unit_idx, time = time_idx)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(Y ~ D, data = dat, index = c("id", "time"),
               method = "fe", force = "two-way",
               se = TRUE, vartype = "jackknife",
               parallel = FALSE, keep.sims = FALSE, CV = FALSE)
  ))
  att_avg <- fit$est.avg[1, "ATT.avg"]
  ci_lo   <- fit$est.avg[1, "CI.lower"]
  ci_hi   <- fit$est.avg[1, "CI.upper"]
  list(att = att_avg, covered = tau >= ci_lo & tau <= ci_hi)
}

test_that("S-SIM: jackknife + normal CI coverage >= 0.85 on DGP-A (100 reps)", {
  skip_on_cran()
  R             <- 100L
  results       <- lapply(seq_len(R), .run_coverage_jack)
  coverage_jack <- mean(sapply(results, `[[`, "covered"))
  expect_gte(coverage_jack, 0.85,
             label = paste0("jackknife coverage = ", round(coverage_jack, 3),
                            " (must be >= 0.85)"))
})

## ---------------------------------------------------------------
## Tests for v2.4.2 ci.method extensions and per-type defaults.
##
## v2.4.2 adds two ci.method values --- "bc" (bias-corrected
## percentile) and "normal" (Wald) --- and switches the per-type
## default (NULL trigger): att -> normal, att.cumu -> percentile,
## aptt -> bc, log.att -> bc.
##
## See statsclaw-workspace/fect/ref/v242-vartype-cimethod-design.md
## ---------------------------------------------------------------

suppressWarnings(data("simdata", package = "fect"))


.fit_canonical <- function(nboots = 100, keep.sims = TRUE) {
  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = simdata, index = c("id", "time"),
      method = "fe", force = "two-way",
      se = TRUE, nboots = nboots, parallel = FALSE,
      keep.sims = keep.sims
    )
  ))
}

## Positive-Y panel for log.att tests (avoids cell-drop hard-error).
.fit_positive_Y <- function(nboots = 30) {
  set.seed(7)
  N <- 30; TT <- 20
  df <- expand.grid(id = 1:N, time = 1:TT)
  treat_start <- sample(c(NA, 8:15), N, replace = TRUE)
  df$D <- ifelse(is.na(treat_start[df$id]) | df$time < treat_start[df$id],
                 0, 1)
  df$Y <- exp(2.0 + 0.05 * df$time + 0.3 * df$D + rnorm(nrow(df), sd = 0.1))
  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(Y ~ D, data = df, index = c("id", "time"),
               method = "fe", force = "two-way",
               se = TRUE, nboots = nboots, parallel = FALSE,
               keep.sims = TRUE)
  ))
}


## -- CI.1 ci.method enum accepts the four values --------------------

test_that("CI.1: ci.method accepts basic / percentile / bc / normal", {

  skip_on_cran()

  fit <- .fit_canonical()
  for (m in c("basic", "percentile", "bc", "normal")) {
    res <- fect::estimand(fit, "att", "overall", window = c(1, 5),
                          ci.method = m)
    expect_s3_class(res, "data.frame")
    expect_true(!is.na(res$estimate))
    expect_true(!is.na(res$se))
  }
})


## -- CI.2 ci.method = NULL triggers per-type defaults ----------------

test_that("CI.2: ci.method = NULL gives type-specific defaults", {

  skip_on_cran()

  fit <- .fit_canonical()

  ## att -> normal: should match the Wald CI from fit$est.att passthrough
  res_att <- fect::estimand(fit, "att", "event.time")
  ## Compare to explicit normal: should be identical (fast path same numbers)
  res_att_normal <- fect::estimand(fit, "att", "event.time",
                                   ci.method = "normal")
  expect_equal(res_att$estimate, res_att_normal$estimate)
  expect_equal(res_att$se,       res_att_normal$se)
  expect_equal(res_att$ci.lo,    res_att_normal$ci.lo)
  expect_equal(res_att$ci.hi,    res_att_normal$ci.hi)
})


test_that("CI.2b: att overall default = normal differs from basic", {

  skip_on_cran()

  fit <- .fit_canonical()

  res_default <- fect::estimand(fit, "att", "overall", window = c(1, 5))
  res_basic   <- fect::estimand(fit, "att", "overall", window = c(1, 5),
                                ci.method = "basic")
  res_normal  <- fect::estimand(fit, "att", "overall", window = c(1, 5),
                                ci.method = "normal")

  ## Default for att is normal; should equal the explicit normal.
  expect_equal(res_default$ci.lo, res_normal$ci.lo)
  expect_equal(res_default$ci.hi, res_normal$ci.hi)
  ## Basic and normal will generally differ.
  expect_false(isTRUE(all.equal(res_basic$ci.lo, res_normal$ci.lo)))
})


## -- CI.3 normal CI: ci.lo = est - z*SE, ci.hi = est + z*SE ----------

test_that("CI.3: normal CI is symmetric around the point estimate", {

  skip_on_cran()

  fit <- .fit_canonical()
  res <- fect::estimand(fit, "att", "overall", window = c(1, 5),
                        ci.method = "normal")

  z <- stats::qnorm(0.975)
  expect_equal(res$ci.lo, res$estimate - z * res$se, tolerance = 1e-10)
  expect_equal(res$ci.hi, res$estimate + z * res$se, tolerance = 1e-10)
})


## -- CI.4 bc CI: when bootstrap is symmetric around point, bc ~ percentile

test_that("CI.4: bc CI ~ percentile when bootstrap median = point estimate", {

  skip_on_cran()

  ## Use the internal helper with a large symmetric bootstrap. bc reduces
  ## to percentile when z0 = 0 (i.e., bootstrap median = point). With
  ## finite Monte Carlo noise the median isn't exactly the mean, so tol
  ## is loosened to 0.05 (5% of one bootstrap SD).
  set.seed(1)
  boot <- rnorm(5000, mean = 0.5, sd = 0.1)
  ci_pct <- fect:::.compute_ci(estimate = 0.5, boot = boot,
                                 ci.method = "percentile",
                                 conf.level = 0.95)
  ci_bc  <- fect:::.compute_ci(estimate = 0.5, boot = boot,
                                 ci.method = "bc",
                                 conf.level = 0.95)
  expect_equal(ci_pct$ci.lo, ci_bc$ci.lo, tolerance = 0.05)
  expect_equal(ci_pct$ci.hi, ci_bc$ci.hi, tolerance = 0.05)
})


## -- CI.5 bc CI: shifts cutoffs when bootstrap is biased --------------

test_that("CI.5: bc shifts cutoffs when bootstrap median != point estimate", {

  skip_on_cran()

  ## Bootstrap centered at 0.5, but point is at 0.7 (above bootstrap median).
  ## bc should shift cutoffs UP from raw percentile.
  set.seed(2)
  boot <- rnorm(1000, mean = 0.5, sd = 0.1)
  ci_pct <- fect:::.compute_ci(estimate = 0.7, boot = boot,
                                 ci.method = "percentile",
                                 conf.level = 0.95)
  ci_bc  <- fect:::.compute_ci(estimate = 0.7, boot = boot,
                                 ci.method = "bc",
                                 conf.level = 0.95)
  expect_true(ci_bc$ci.lo > ci_pct$ci.lo)
  expect_true(ci_bc$ci.hi > ci_pct$ci.hi)
})


## -- CI.6 vartype column reports method actually used at fit time ----

test_that("CI.6: vartype column reports the fit-time vartype regardless of ci.method", {

  skip_on_cran()

  fit <- .fit_canonical()
  for (m in c("basic", "percentile", "bc", "normal")) {
    res <- fect::estimand(fit, "att", "overall", window = c(1, 5),
                          ci.method = m)
    expect_true(res$vartype %in%
                c("bootstrap", "jackknife", "parametric"))
  }
})


## -- CI.7 cell-drop hard-error fires for log.att on negative-Y data --

test_that("CI.7: log.att on simdata triggers point-level hard-error (v2.4.2+)", {

  skip_on_cran()

  ## simdata has cells with Y <= 0 in the treated post-treatment window;
  ## v2.4.2's point-estimate-level hard-error fires before any bootstrap
  ## work. The bootstrap-level "log-ATT bootstrap is unreliable" message
  ## remains in R/po-estimands.R for strictly-positive panels where
  ## Y0_b crosses zero only in some replicates.
  fit_neg <- .fit_canonical()
  expect_error(
    fect::estimand(fit_neg, "log.att", "event.time"),
    "log\\.att requires Y > 0"
  )
})


test_that("CI.7b: log.att works on a positive-Y panel (no cell-drop pathology)", {

  skip_on_cran()

  fit_pos <- .fit_positive_Y(nboots = 30)
  res <- fect::estimand(fit_pos, "log.att", "event.time")

  expect_s3_class(res, "data.frame")
  expect_true(any(!is.na(res$estimate)))
  expect_true(any(res$n_cells > 0))
})


## -- CI.8 nboots default is 200 (v2.4.2) -----------------------------

test_that("CI.8: fect()'s nboots default is 200", {
  ## Cheap arg-validation; doesn't need a fit. fect.formula and
  ## fect.default are S3 methods, accessed via getS3method().
  ## v2.4.2 keeps nboots = 200 as default (unchanged from v2.4.1).
  ## Sufficient for SE / normal CI; bump to 1000+ for tail-quantile
  ## CIs accessed via estimand(). See CI.9-CI.11 for the warning gate.
  expect_equal(as.character(formals(fect::fect)$nboots), "200")
  expect_equal(as.character(formals(getS3method("fect", "formula"))$nboots),
               "200")
  expect_equal(as.character(formals(getS3method("fect", "default"))$nboots),
               "200")
})


## -- CI.9 estimand() warns on tail-CI methods at nboots < 1000 -------
## by = "overall" used because by = "event.time" + non-normal ci.method
## falls through to a "not yet implemented" path at this commit.

test_that("CI.9: estimand() warns on tail-CI methods with nboots < 1000", {
  skip_on_cran()
  fit_small <- .fit_canonical(nboots = 100)
  ## The warning is gated on Sys.getenv("TESTTHAT") inside estimand() to
  ## keep the suite-wide summary clean. Temporarily unset the env var here
  ## so the warning fires and can be asserted.
  withr::with_envvar(c(TESTTHAT = "false"), {
    for (m in c("basic", "percentile", "bc", "bca")) {
      expect_warning(
        fect::estimand(fit_small, "att", "overall", window = c(1, 5),
                       ci.method = m),
        "tail-quantile-based CIs may have"
      )
    }
  })
})


## -- CI.10 estimand() does NOT warn on normal CI ---------------------

test_that("CI.10: normal CI does not trigger the under-replication warning", {
  skip_on_cran()
  fit_small <- .fit_canonical(nboots = 100)
  expect_warning(
    fect::estimand(fit_small, "att", "overall", window = c(1, 5),
                   ci.method = "normal"),
    NA  ## NA = no warning expected
  )
})


## -- CI.11 estimand() does NOT warn at nboots >= 1000 ----------------

test_that("CI.11: tail-CI methods at nboots >= 1000 do not warn", {
  skip_on_cran()
  fit_big <- .fit_canonical(nboots = 1000)
  expect_warning(
    fect::estimand(fit_big, "att", "overall", window = c(1, 5),
                   ci.method = "bca"),
    NA
  )
  expect_warning(
    fect::estimand(fit_big, "att", "overall", window = c(1, 5),
                   ci.method = "percentile"),
    NA
  )
})

## Tests for the v2.4.2 ci.method = c("normal", "basic") arg on fect(),
## and the soft-deprecated quantile.CI legacy arg.
##
## What we're verifying:
##   1. Default behavior is byte-equal to the pre-v2.4.2 default
##      (quantile.CI = FALSE -> normal Wald CI).
##   2. ci.method = "normal" is byte-equal to legacy quantile.CI = FALSE.
##   3. ci.method = "basic" is byte-equal to legacy quantile.CI = TRUE.
##   4. Supplying quantile.CI explicitly emits a deprecation warning.
##   5. NOT supplying quantile.CI (NULL sentinel) emits no warning under the
##      modern API.
##   6. ci.method = "basic" with nboots < 1000 emits a tail-CI replicate
##      warning at fit time (mirrors the estimand .check_tail_ci_replicates
##      gate).
##   7. ci.method = "bca" / "bc" / "percentile" hard-error with a message
##      pointing to estimand() for the full 5-method surface.
##   8. ci.method = "garbage" hard-errors with a clear message.

library(testthat)

data(simdata, package = "fect")

base_call <- function(nboots = 50, parallel = FALSE, ...) {
    fect(Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
         method = "fe", force = "two-way", se = TRUE, nboots = nboots,
         parallel = parallel, seed = 42L, ...)
}

# ---- 1. Default behavior is byte-equal to legacy quantile.CI = FALSE -------

test_that("default ci.method yields the legacy quantile.CI = FALSE result", {
    out_default <- base_call()
    out_legacy_false <- suppressWarnings(base_call(quantile.CI = FALSE))
    expect_equal(out_default$est.avg, out_legacy_false$est.avg)
    expect_equal(out_default$est.att, out_legacy_false$est.att)
})

# ---- 2. ci.method = "normal" byte-equal to legacy FALSE --------------------

test_that("ci.method = 'normal' is byte-equal to quantile.CI = FALSE", {
    out_normal <- base_call(ci.method = "normal")
    out_legacy <- suppressWarnings(base_call(quantile.CI = FALSE))
    expect_equal(out_normal$est.avg, out_legacy$est.avg)
    expect_equal(out_normal$est.att, out_legacy$est.att)
})

# ---- 3. ci.method = "basic" byte-equal to legacy TRUE ----------------------

test_that("ci.method = 'basic' is byte-equal to quantile.CI = TRUE", {
    out_basic <- suppressWarnings(base_call(ci.method = "basic", nboots = 1000))
    out_legacy <- suppressWarnings(base_call(quantile.CI = TRUE, nboots = 1000))
    expect_equal(out_basic$est.avg, out_legacy$est.avg)
    expect_equal(out_basic$est.att, out_legacy$est.att)
})

# ---- 4. Supplying quantile.CI explicitly fires deprecation warning ---------

test_that("quantile.CI = TRUE fires a deprecation warning that mentions ci.method", {
    expect_warning(
        base_call(quantile.CI = TRUE, nboots = 1000),
        regexp = "quantile\\.CI.*deprecated.*ci\\.method"
    )
})

test_that("quantile.CI = FALSE also fires the deprecation warning", {
    ## "Soft-deprecate" means any user-supplied use of the legacy arg warns,
    ## even when supplying the legacy default value.
    expect_warning(
        base_call(quantile.CI = FALSE),
        regexp = "quantile\\.CI.*deprecated"
    )
})

# ---- 5. NULL sentinel: no warning under modern API -------------------------

test_that("modern API (no quantile.CI supplied) emits no quantile.CI warning", {
    expect_silent_about <- function(expr, regexp) {
        warned <- FALSE
        withCallingHandlers(
            expr,
            warning = function(w) {
                if (grepl(regexp, conditionMessage(w))) warned <<- TRUE
                invokeRestart("muffleWarning")
            }
        )
        expect_false(warned)
    }
    expect_silent_about(base_call(), "quantile\\.CI")
    expect_silent_about(base_call(ci.method = "normal"), "quantile\\.CI")
})

# ---- 6. nboots < 1000 + ci.method = "basic" warns at fit time --------------

test_that("ci.method = 'basic' with nboots < 1000 fires the tail-CI warning", {
    ## The warning is gated on Sys.getenv("TESTTHAT") inside fect() to keep
    ## the suite-wide summary clean. Temporarily unset the env var here so
    ## the warning fires and can be asserted.
    withr::with_envvar(c(TESTTHAT = "false"), {
        expect_warning(
            base_call(ci.method = "basic", nboots = 50),
            regexp = "tail quantiles.*1000|Efron 1987"
        )
    })
})

expect_silent_about <- function(expr, regexp) {
    warned <- FALSE
    withCallingHandlers(
        expr,
        warning = function(w) {
            if (grepl(regexp, conditionMessage(w))) warned <<- TRUE
            invokeRestart("muffleWarning")
        }
    )
    expect_false(warned)
}

test_that("ci.method = 'basic' with nboots >= 1000 emits no tail-CI warning", {
    expect_silent_about(
        base_call(ci.method = "basic", nboots = 1000),
        "tail quantiles|Efron 1987"
    )
})

test_that("ci.method = 'normal' never fires the tail-CI warning", {
    expect_silent_about(
        base_call(ci.method = "normal", nboots = 50),
        "tail quantiles"
    )
})

# ---- 7. ci.method in {bca, bc, percentile} hard-errors ---------------------

test_that("ci.method = 'bca' errors with a message pointing to estimand()", {
    expect_error(
        base_call(ci.method = "bca"),
        regexp = "bca.*not supported in fect.*estimand"
    )
})

test_that("ci.method = 'bc' errors with a message pointing to estimand()", {
    expect_error(
        base_call(ci.method = "bc"),
        regexp = "bc.*not supported in fect.*estimand"
    )
})

test_that("ci.method = 'percentile' errors with a message pointing to estimand()", {
    expect_error(
        base_call(ci.method = "percentile"),
        regexp = "percentile.*not supported in fect.*estimand"
    )
})

# ---- 8. ci.method validation -----------------------------------------------

test_that("ci.method = 'garbage' fails fast", {
    expect_error(
        base_call(ci.method = "garbage"),
        regexp = "ci\\.method.*must be one of"
    )
})

# ---- 9. ci.method = "basic" + vartype = "jackknife" -> error ---------------

test_that("ci.method = 'basic' + vartype = 'jackknife' hard-errors", {
    expect_error(
        suppressWarnings(base_call(vartype = "jackknife", ci.method = "basic")),
        regexp = "ci\\.method = \"basic\".*not supported.*jackknife"
    )
})

# ---- 10. ci.method = "basic" + parametric -> location-shift fix ------------
#
# fect's built-in CI machinery applies a location-shift fix at the avg-level
# and per-event-time CI sites so that ci.method = "basic" works on parametric
# fits.  The fect-level result should match estimand(fit, "att", "basic")
# byte-equally on the avg-level CI.

test_that("ci.method = 'basic' + parametric matches estimand at avg level", {
    skip_if_not_installed("fect")
    # Build a simple no-reversal panel (parametric requires no reversal)
    set.seed(1001L); N <- 30; TT <- 20; T0 <- 12; Ntr <- 5
    alpha_i <- rnorm(N); xi_t <- rnorm(TT)
    e <- matrix(rnorm(TT * N), TT, N)
    D <- matrix(0L, TT, N); D[(T0 + 1):TT, 1:Ntr] <- 1L
    Y <- outer(xi_t, rep(1, N)) + outer(rep(1, TT), alpha_i) + 3 * D + e
    df <- data.frame(id = rep(1:N, each = TT), time = rep(1:TT, N),
                     Y = c(Y), D = c(D))
    fit <- suppressWarnings(fect(
        Y ~ D, data = df, index = c("id", "time"),
        method = "fe", force = "two-way", CV = FALSE,
        se = TRUE, vartype = "parametric", para.error = "auto",
        ci.method = "basic",
        nboots = 200, parallel = FALSE,
        time.component.from = "nevertreated",
        keep.sims = TRUE, seed = 42L
    ))
    e <- suppressWarnings(estimand(fit, type = "att", by = "overall",
                                    ci.method = "basic"))
    expect_equal(unname(fit$est.avg[1, "CI.lower"]), e$ci.lo, tolerance = 1e-10)
    expect_equal(unname(fit$est.avg[1, "CI.upper"]), e$ci.hi, tolerance = 1e-10)
})

# ---- 9. vartype = "jackknife" + cl emits an "ignored" warning --------------
#
# fect's jackknife is leave-one-unit-out and does not consult cl.  Combining
# vartype = "jackknife" with a non-NULL cl silently produced unit-level SEs.
# The user-facing warning routes such callers to vartype = "bootstrap" + cl
# for cluster-aware inference.

test_that("vartype = 'jackknife' with cl warns that cl is ignored", {
    d <- simdata
    d$cl <- (d$id - 1) %/% 2 + 1   # 2 units per cluster
    expect_warning(
        suppressMessages(fect(Y ~ D, data = d, index = c("id", "time"),
                               method = "fe", force = "two-way",
                               vartype = "jackknife", cl = "cl",
                               se = TRUE, parallel = FALSE, seed = 42L)),
        "cl argument is ignored"
    )
})

test_that("vartype = 'jackknife' without cl does not emit the cl warning", {
    seen <- character(0)
    withCallingHandlers(
        suppressMessages(fect(Y ~ D, data = simdata, index = c("id", "time"),
                               method = "fe", force = "two-way",
                               vartype = "jackknife",
                               se = TRUE, parallel = FALSE, seed = 42L)),
        warning = function(w) {
            seen <<- c(seen, conditionMessage(w))
            invokeRestart("muffleWarning")
        }
    )
    expect_false(any(grepl("cl argument is ignored", seen, fixed = TRUE)))
})

test_that("vartype = 'bootstrap' with cl does not emit the jackknife warning", {
    d <- simdata
    d$cl <- (d$id - 1) %/% 2 + 1
    seen <- character(0)
    withCallingHandlers(
        suppressMessages(fect(Y ~ D, data = d, index = c("id", "time"),
                               method = "fe", force = "two-way",
                               vartype = "bootstrap", cl = "cl",
                               nboots = 50, se = TRUE,
                               parallel = FALSE, seed = 42L)),
        warning = function(w) {
            seen <<- c(seen, conditionMessage(w))
            invokeRestart("muffleWarning")
        }
    )
    expect_false(any(grepl("cl argument is ignored", seen, fixed = TRUE)))
})

## ---------------------------------------------------------------
## Tests for imputed_outcomes() and the slot contract.
##
## Coverage: column schema, cells filter (NULL / logical / formula),
## replicates expansion, direction = "on" / "off", error paths.
## ---------------------------------------------------------------

suppressWarnings(data("simdata", package = "fect"))


.fit_canonical <- function(keep.sims = FALSE, nboots = 50) {
  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = simdata, index = c("id", "time"),
      method = "fe", force = "two-way",
      se = TRUE, nboots = nboots, parallel = FALSE,
      keep.sims = keep.sims
    )
  ))
}


## -- I.1  Column schema, point estimates ----------------------------

test_that("I.1: imputed_outcomes() returns the documented column schema", {

  skip_on_cran()

  fit <- .fit_canonical()
  po  <- fect::imputed_outcomes(fit)

  expect_s3_class(po, "data.frame")

  expected_cols <- c("id", "time", "event.time", "cohort", "treated",
                     "Y_obs", "Y0_hat", "eff", "eff_debias", "W.agg")
  expect_setequal(names(po), expected_cols)

  ## Every row is treated
  expect_true(all(po$treated))

  ## Y0_hat = Y_obs - eff (within numeric tolerance)
  expect_equal(po$Y0_hat, po$Y_obs - po$eff)

  ## eff_debias is zero for plain imputation estimators (FE)
  expect_true(all(po$eff_debias == 0))

  ## W.agg is 1 when fit has no W
  expect_true(all(po$W.agg == 1))

  ## event.time is populated at most treated cells; can be NA for
  ## units that begin in the treated state (no pre-treatment data to
  ## anchor the event-time grid). Document and accept.
  expect_true(mean(!is.na(po$event.time)) > 0.95)
})


## -- I.2  cells filter: NULL = unfiltered ---------------------------

test_that("I.2: cells = NULL returns all treated cells", {

  skip_on_cran()

  fit <- .fit_canonical()
  po_full  <- fect::imputed_outcomes(fit, cells = NULL)
  po_again <- fect::imputed_outcomes(fit)
  expect_identical(po_full, po_again)
})


## -- I.3  cells filter: logical vector ------------------------------

test_that("I.3: cells = logical vector subsets correctly", {

  skip_on_cran()

  fit <- .fit_canonical()
  po  <- fect::imputed_outcomes(fit)

  mask <- po$event.time %in% 1:5
  po_logical <- fect::imputed_outcomes(fit, cells = mask)

  expect_equal(nrow(po_logical), sum(mask))
  expect_true(all(po_logical$event.time %in% 1:5))
})


## -- I.4  cells filter: one-sided formula ---------------------------

test_that("I.4: cells = formula subsets correctly", {

  skip_on_cran()

  fit <- .fit_canonical()
  po_form  <- fect::imputed_outcomes(fit, cells = ~ event.time %in% 1:5)
  po_full  <- fect::imputed_outcomes(fit)

  expect_equal(nrow(po_form), sum(po_full$event.time %in% 1:5))
  expect_true(all(po_form$event.time %in% 1:5))
})


## -- I.5  replicates = TRUE expands by nboots -----------------------

test_that("I.5: replicates = TRUE expands by nboots and adds replicate column", {

  skip_on_cran()

  fit <- .fit_canonical(keep.sims = TRUE, nboots = 25)
  po  <- fect::imputed_outcomes(fit)
  po_b <- fect::imputed_outcomes(fit, replicates = TRUE)

  expect_equal(nrow(po_b), nrow(po) * 25)
  expect_true("replicate" %in% names(po_b))
  expect_setequal(unique(po_b$replicate), seq_len(25))

  ## Every group of nboots rows shares the same (id, time, Y_obs, W.agg)
  one_cell <- po_b[po_b$id == po_b$id[1] & po_b$time == po_b$time[1], ]
  expect_equal(nrow(one_cell), 25)
  expect_equal(length(unique(one_cell$Y_obs)), 1)
  expect_equal(length(unique(one_cell$W.agg)), 1)

  ## Y0_hat varies across replicates
  expect_gt(length(unique(one_cell$Y0_hat)), 1)
})


## -- I.6  replicates = TRUE without keep.sims errors with locked wording

test_that("I.6: replicates = TRUE without keep.sims errors with locked wording", {

  skip_on_cran()

  fit <- .fit_canonical(keep.sims = FALSE)
  expect_error(
    fect::imputed_outcomes(fit, replicates = TRUE),
    "keep\\.sims = TRUE",
    fixed = FALSE
  )
})


## -- I.7  direction = "off" works on reversal panels ----------------

test_that("I.7: direction = 'off' returns a sensible long form on reversal panels", {

  skip_on_cran()

  fit <- .fit_canonical()
  ## simdata has reversal; fit$T.off should be populated.
  expect_false(is.null(fit$T.off))

  po_off <- fect::imputed_outcomes(fit, direction = "off")
  expect_true(nrow(po_off) > 0)
  expect_true(all(!is.na(po_off$event.time)))

  ## Cohort under "off" is the calendar time of first exit.
  expect_true(all(!is.na(po_off$cohort) | po_off$event.time != 1))
})


## -- I.8  validate_po_contract: bad fit errors ----------------------

test_that("I.8: passing a non-fect object errors with a contract message", {

  expect_error(
    fect::imputed_outcomes("not a fit"),
    "fect"
  )
})

## ---------------------------------------------------------------
## Tests for estimand(fit, "aptt", "event.time").
##
## Reference: Chen & Roth (2024 QJE).
## APTT_t = mean_{(t,i): D=1} (Y - Y0_hat) / mean_{(t,i): D=1} Y0_hat
##
## Validates against a manual computation from raw fit slots — this
## is the same recipe currently in vignettes/02-fect.Rmd.
## ---------------------------------------------------------------

suppressWarnings(data("simdata", package = "fect"))


.fit_canonical <- function(nboots = 50, keep.sims = TRUE) {
  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = simdata, index = c("id", "time"),
      method = "fe", force = "two-way",
      se = TRUE, nboots = nboots, parallel = FALSE,
      keep.sims = keep.sims
    )
  ))
}


## -- AP.1  Schema --------------------------------------------------

test_that("AP.1: estimand(fit, 'aptt', 'event.time') has the documented schema", {

  skip_on_cran()

  fit <- .fit_canonical()
  est <- fect::estimand(fit, "aptt", "event.time")

  expect_s3_class(est, "data.frame")
  expected_cols <- c("event.time", "estimate", "se",
                     "ci.lo", "ci.hi", "n_cells", "vartype")
  expect_setequal(names(est), expected_cols)

  expect_true(is.numeric(est$estimate))
  expect_true(all(!is.na(est$event.time)))
})


## -- AP.2  Manual recipe match -----------------------------------------

test_that("AP.2: estimate matches the manual APTT recipe from the vignette", {

  skip_on_cran()

  fit <- .fit_canonical()
  est <- fect::estimand(fit, "aptt", "event.time")

  ## Manual APTT computation: per event time, mean(eff)/mean(Y - eff).
  Don <- !is.na(fit$D.dat) & fit$D.dat == 1 & !is.na(fit$T.on)
  ets <- sort(unique(fit$T.on[Don]))

  manual <- vapply(ets, function(et) {
    m <- Don & fit$T.on == et
    eff_t <- fit$eff[m]
    Y_t   <- fit$Y.dat[m]
    Y0_t  <- Y_t - eff_t
    mean(eff_t, na.rm = TRUE) / mean(Y0_t, na.rm = TRUE)
  }, numeric(1))

  est_sorted <- est[order(est$event.time), ]
  expect_equal(est_sorted$estimate, manual, tolerance = 1e-10)
})


## -- AP.3  keep.sims = FALSE: error path ------------------------------

test_that("AP.3: APTT without keep.sims errors with the locked wording", {

  skip_on_cran()

  fit_no_keep <- .fit_canonical(keep.sims = FALSE)
  expect_error(
    fect::estimand(fit_no_keep, "aptt", "event.time"),
    "keep\\.sims = TRUE",
    fixed = FALSE
  )
})


## -- AP.4  vartype = "none" works without keep.sims --------------------

test_that("AP.4: vartype = 'none' returns point estimate only without keep.sims", {

  skip_on_cran()

  fit_no_keep <- .fit_canonical(keep.sims = FALSE)
  est <- fect::estimand(fit_no_keep, "aptt", "event.time", vartype = "none")
  expect_true(nrow(est) > 0)
  expect_true(all(is.na(est$se)))
  expect_true(all(is.na(est$ci.lo)))
  expect_true(all(est$vartype == "none"))
})

#################################
## support function
#################################
get_term <- function(d,
                     ii,
                     type = "on") {
    dd <- d
    iii <- ii
    ## dd <- dd[which(iii == 1)]
    first.pos <- min(which(iii == 1))
    if (first.pos != 1) {
        dd <- dd[-(1:(first.pos - 1))]
        iii <- iii[-(1:(first.pos - 1))]
    }
    T <- length(dd)
    if (0 %in% iii) {
        if (T > 1) {
            for (i in 1:(T - 1)) {
                if (iii[i + 1] == 0) {
                    dd[i + 1] <- dd[i]
                }
            }
        }
    }

    if (type == "off") {
        dd <- abs(dd - 1)
    }
    d1 <- dd[1:(T - 1)]
    d2 <- dd[2:T]

    if (T == 1) {
        term <- rep(NA, 1)
    } else if (sum(d1 == d2) == (T - 1)) {
        term <- rep(NA, T)
    } else {
        change.pos <- which(d1 != d2) + 1
        change.length <- length(change.pos)
        term <- NULL
        if (dd[1] == 0) {
            for (i in 1:(change.length)) {
                if (i == 1) {
                    part.term <- (2 - change.pos[i]):0
                } else {
                    if (i %% 2 == 0) {
                        part.term <- 1:(change.pos[i] - change.pos[i - 1])
                    } else {
                        part.term <- (change.pos[i - 1] - change.pos[i] + 1):0
                    }
                }
                term <- c(term, part.term)
            }
        } else if (dd[1] == 1) {
            for (i in 1:(change.length)) {
                if (i == 1) {
                    # if (type == "on") {
                    #    part.term <- 1:(change.pos[i] - 1)
                    # } else if (type == "off") {
                    part.term <- rep(NA, change.pos[i] - 1)
                    # }
                } else {
                    if (i %% 2 == 0) {
                        part.term <- (change.pos[i - 1] - change.pos[i] + 1):0
                    } else {
                        part.term <- 1:(change.pos[i] - change.pos[i - 1])
                    }
                }
                term <- c(term, part.term)
            }
        }
        if (dd[change.pos[change.length]] == 0) {
            term <- c(term, rep(NA, (T - change.pos[change.length] + 1)))
        } else {
            term <- c(term, 1:(T - change.pos[change.length] + 1))
        }
    }
    ## term.all <- rep(NA, length(d))
    if (first.pos != 1) {
        term <- c(rep(NA, (first.pos - 1)), term)
    }
    return(term)
}

###################################
## regressions for initial values
###################################

align_beta0 <- function(beta_raw, p) {
    if (p <= 0) {
        return(matrix(0, 1, 0))
    }

    beta_raw <- as.numeric(beta_raw)
    beta_names <- names(beta_raw)
    beta_full <- rep(0, p)
    filled <- rep(FALSE, p)
    used_raw <- rep(FALSE, length(beta_raw))

    if (!is.null(beta_names)) {
        for (j in seq_along(beta_raw)) {
            nm <- beta_names[j]
            idx <- suppressWarnings(as.integer(sub(".*?(\\d+)$", "\\1", nm)))
            if (!is.na(idx) && idx >= 1 && idx <= p && !used_raw[j]) {
                beta_full[idx] <- beta_raw[j]
                filled[idx] <- TRUE
                used_raw[j] <- TRUE
            }
        }
    }

    remaining_pos <- which(!filled)
    remaining_raw <- which(!used_raw)
    if (length(remaining_pos) > 0 && length(remaining_raw) > 0) {
        nfill <- min(length(remaining_pos), length(remaining_raw))
        beta_full[remaining_pos[seq_len(nfill)]] <- beta_raw[remaining_raw[seq_len(nfill)]]
    }

    beta_full[which(is.na(beta_full))] <- 0
    return(as.matrix(beta_full))
}

initialFit <- function(data, ## long form data
                       force,
                       w = NULL,
                       oci) { ## indicator

    N <- length(unique(data[, 2]))
    T <- length(unique(data[, 3]))
    p <- dim(data)[2] - 3

    x <- x.sub <- NULL
    if (p > 0) {
        x <- as.matrix(data[, 4:(dim(data)[2])])
        x.sub <- as.matrix(x[oci, ])
    }
    y <- as.matrix(data[, 1])
    y.sub <- as.matrix(y[oci, ])
    beta0 <- matrix(0, 1, 1)

    if (!is.null(w)) {
        use_weight <- 1
        w <- as.matrix(w)
        w.sub <- as.matrix(w[oci, ])
    } else {
        use_weight <- 0
    }

    ind <- NULL
    if (force == 0) {
        ind <- NULL
    } else if (force == 1) {
        ind <- as.matrix(data[, 2])
    } else if (force == 2) {
        ind <- as.matrix(data[, 3])
    } else if (force == 3) {
        ind <- as.matrix(data[, c(2, 3)])
    }

    if (force == 0) {
        if (p == 0) {
            if (use_weight == 1) {
                mu <- sum(y.sub * w.sub) / sum(w.sub)
                Y0 <- matrix(mu, T, N)
            } else {
                mu <- mean(c(y)[oci])
                Y0 <- matrix(mu, T, N)
            }
            ## res <- as.matrix(c(y) - mu)
        } else {
            if (use_weight == 1) {
                lm.fit <- lm(as.matrix(c(y)[oci]) ~ x.sub, weights = w.sub)
            } else {
                lm.fit <- lm(as.matrix(c(y)[oci]) ~ x.sub)
            }
            coef <- lm.fit$coefficients
            mu <- coef[1]
            beta0 <- align_beta0(coef[2:length(coef)], p)
            y0 <- mu + x %*% beta0
            Y0 <- matrix(y0, T, N)
            ## res <- as.matrix(lm.fit$residuals)
        }
    } else {
        colnames(y) <- y.name <- "y"
        colnames(ind) <- ind.name <- paste0("id.", c(1:dim(ind)[2]))
        if (p > 0) {
            colnames(x) <- x.name <- paste0("x.", c(1:dim(x)[2]))
            data.reg <- cbind.data.frame(y, x, ind)
            formula.reg <- paste0("y~", paste(x.name, collapse = "+"), "|", paste(ind.name, collapse = "+"))
        } else {
            data.reg <- cbind.data.frame(y, ind)
            formula.reg <- paste0("y~1|", paste(ind.name, collapse = "+"))
        }
        formula.reg <- as.formula(formula.reg)

        lm.fit <- feols(fml = formula.reg, data = data.reg[oci, ], fixef.rm = "none")
        # if (use_weight == 1) {
        #     lm.fit <- suppressWarnings(invisible(feols(
        #         fml = formula.reg,
        #         data = data.reg[oci, ], weights = w.sub,
        #         fixef.rm = "none"
        #     )))
        # } else {
        #     lm.fit <- suppressWarnings(invisible(feols(
        #         fml = formula.reg,
        #         data = data.reg[oci, ],
        #         fixef.rm = "none"
        #     )))
        # }

        y0 <- predict(lm.fit, newdata = data.reg)

        # lm.fit <- suppressWarnings(invisible(fastplm(y = as.matrix(c(y)[oci]), x = x.sub,
        #                   ind = as.matrix(ind[oci,]),drop.singletons = FALSE)))
        # y0 <- suppressWarnings(predict(lm.fit, x = x, ind = ind))
        Y0 <- matrix(y0, T, N)
        if (p > 0) {
            beta0 <- align_beta0(lm.fit$coefficients, p)
        }
    }

    bad.Y0 <- !is.finite(Y0)
    if (sum(bad.Y0) > 0) {
        mu.fill <- mean(as.numeric(y.sub), na.rm = TRUE)
        if (!is.finite(mu.fill)) {
            mu.fill <- 0
        }
        Y0[bad.Y0] <- mu.fill
    }

    result <- list(Y0 = Y0, beta0 = beta0)
    return(result)
}

################################################
##  regressions for initial values, probit model  ##
################################################

## if we do QR : the fixed effects term doesn't contain mu, and we need Y0, FE0, xi0, factor0
## if we do SVD : the fixed effects term contains mu, and we need Y0, FE0

BiInitialFit <- function(data, ## long form data
                         QR = 0,
                         force,
                         r,
                         oci) { ## indicator

    N <- length(unique(data[, 2]))
    T <- length(unique(data[, 3]))
    r <- min(r, T, N)
    p <- dim(data)[2] - 3

    x <- x.sub <- NULL
    if (p > 0) {
        x <- as.matrix(data[, 4:(dim(data)[2])])
        ## regard as missing
        x[setdiff(1:(N * T), oci), ] <- 0
        x.sub <- as.matrix(x[oci, ])
    }
    y <- as.matrix(data[, 1])
    beta0 <- matrix(0, 1, 1)

    ind <- NULL
    if (force == 1) {
        ind <- as.matrix(data[, 2])
    } else if (force == 2) {
        ind <- as.matrix(data[, 3])
    } else if (force == 3) {
        ind <- as.matrix(data[, c(2, 3)])
    }

    xi <- matrix(0, T, 1)
    alpha <- matrix(0, N, 1)
    factor <- matrix(0, T, r)
    loadings <- matrix(0, N, r)
    Y0 <- FE <- matrix(0, T, N)

    if (force == 0) { ## no additive fixed effects
        if (p == 0) {
            mu <- mean(c(y)[oci])
            y0 <- as.matrix(rep(mu, T * N))
            Y0 <- matrix(mu, T, N)
        } else {
            lm.fit <- lm(as.matrix(c(y)[oci]) ~ x.sub)
            coef <- lm.fit$coefficients
            mu <- coef[1]
            beta0 <- as.matrix(coef[2:length(coef)])
            if (sum(is.na(beta0)) > 0) {
                beta0[which(is.na(beta0))] <- 0
            }
            y0 <- mu + x %*% beta0
            Y0 <- matrix(y0, T, N)
        }
        if (QR == 0) {
            FE <- mu
        }
    } else { ## with additive fixed effects
        # plm.fit <- suppressWarnings(invisible(fastplm(y = as.matrix(c(y)[oci]), x = x.sub,
        #                   ind = as.matrix(ind[oci,]),drop.singletons = FALSE)))
        plm.fit <- NULL # to delete
        y0 <- suppressWarnings(predict(plm.fit, x = x, ind = ind))
        Y0 <- matrix(y0, T, N)
        mu <- plm.fit$intercept
        if (p > 0) {
            beta0 <- plm.fit$coefficients
        }
        if (force == 1) {
            alpha <- plm.fit$sfe.coefs[[1]]
        }
        if (force == 2) {
            xi <- plm.fit$sfe.coefs[[1]]
        } else if (force == 3) {
            alpha <- plm.fit$sfe.coefs[[1]]
            xi <- plm.fit$sfe.coefs[[2]]
        }
    }

    ## pca
    if (r > 0) { ## factor analysis of residuals

        res <- y - y0
        res[setdiff(1:(N * T), oci), ] <- 0
        res <- matrix(res, T, N)
        ife_pca <- panel_factor(res, r)
        ife <- ife_pca$FE
        factor <- ife_pca$factor
        loadings <- ife_pca$lambda
        Y0 <- Y0 + ife

        if (QR == 1) {
            ## initial value: need time fixed effects xi and factor f
            if (force == 0) {
                ## factor <- qr_factor(factor, loadings)$factor
                FE <- ife_pca$FE
            } else if (force == 1) {
                ## modify initial Y0
                Y0[setdiff(1:(N * T), oci)] <- Y0[setdiff(1:(N * T), oci)] - (sum(factor[1, ] * loadings[1, ]) + alpha[1])
                ## restrctions: alpha_1 = 0, f_1 = 0
                alpha <- alpha - alpha[1] + (loadings - matrix(rep(loadings[1, ], N), N, r, byrow = TRUE)) %*% as.matrix(factor[1, ])
                factor <- factor - matrix(rep(factor[1, ], T), T, r, byrow = TRUE)
                ife_qr <- qr_factor(factor, loadings)
                factor <- ife_qr$factor
                FE <- ife_qr$FE + matrix(rep(alpha, each = T), T, N)
            } else if (force == 2) {
                ## modify initial Y0
                Y0[setdiff(1:(N * T), oci)] <- Y0[setdiff(1:(N * T), oci)] - (sum(factor[1, ] * loadings[1, ]) + xi[1])
                ## restrctions: xi_1 = 0
                xi <- xi - xi[1] + (factor - matrix(rep(factor[1, ], T), T, r, byrow = TRUE)) %*% as.matrix(loadings[1, ])
                loadings <- loadings - matrix(rep(loadings[1, ], N), N, r, byrow = TRUE)
                ife_qr <- qr_factor(factor, loadings)
                factor <- ife_qr$factor
                FE <- ife_qr$FE + matrix(rep(xi, N), T, N)
            } else if (force == 3) {
                ## modify initial Y0
                Y0[setdiff(1:(N * T), oci)] <- Y0[setdiff(1:(N * T), oci)] - (sum(factor[1, ] * loadings[1, ]) + xi[1] + alpha[1])
                ## restrctions: xi_1 = 0 , f_1 = 0
                alpha <- alpha - alpha[1] + (loadings - matrix(rep(loadings[1, ], N), N, r, byrow = TRUE)) %*% as.matrix(factor[1, ])
                xi <- xi - xi[1] + (factor - matrix(rep(factor[1, ], T), T, r, byrow = TRUE)) %*% as.matrix(loadings[1, ])
                factor <- factor - matrix(rep(factor[1, ], T), T, r, byrow = TRUE)
                loadings <- loadings - matrix(rep(loadings[1, ], N), N, r, byrow = TRUE)
                ife_qr <- qr_factor(factor, loadings)
                factor <- ife_qr$factor
                FE <- ife_qr$FE + matrix(rep(alpha, each = T), T, N) + matrix(rep(xi, N), T, N)
            }
        } else {
            FE <- mu + matrix(rep(alpha, each = T), T, N) + matrix(rep(xi, N), T, N) + ife
        }
    } else { ## only adjust additive fixed effects
        if (QR == 1) {
            if (force == 1) {
                ## modify initial Y0
                Y0[setdiff(1:(N * T), oci)] <- Y0[setdiff(1:(N * T), oci)] - alpha[1]
                ## restrctions: alpha_1 = 0
                FE <- matrix(rep(alpha, each = T), T, N)
            } else if (force == 2) {
                ## modify initial Y0
                Y0[setdiff(1:(N * T), oci)] <- Y0[setdiff(1:(N * T), oci)] - xi[1]
                ## restrctions: xi_1 = 0
                xi <- xi - xi[1]
                FE <- matrix(rep(xi, N), T, N)
            } else if (force == 3) {
                ## modify initial Y0
                Y0[setdiff(1:(N * T), oci)] <- Y0[setdiff(1:(N * T), oci)] - (xi[1] + alpha[1])
                ## restrctions: xi_1 = 0 , alpha_1 = 0
                xi <- xi - xi[1]
                FE <- matrix(rep(alpha, each = T), T, N) + matrix(rep(xi, N), T, N)
            }
        } else {
            FE <- mu + matrix(rep(alpha, each = T), T, N) + matrix(rep(xi, N), T, N)
        }
    }

    result <- list(Y0 = Y0, FE0 = FE, xi0 = xi, factor0 = factor, loadings = loadings)
    return(result)
}

## cross validation sampling
cv.sample2 <- function(I, count) {
    N <- dim(I)[2]
    TT <- dim(I)[1]
    cv.id <- NULL
    oci <- which(c(I) == 1)
    if (count <= 3) {
        cv.id <- sample(oci, count, replace = FALSE)
    } else {
        ## remove boundary observation
        oci2 <- setdiff(oci, c((1:N) * TT, (TT * (0:(N - 1)) + 1)))
        ## randomly select 1/3
        subcount <- floor(count / 3)
        rm.id <- sample(oci2, subcount, replace = FALSE)
        rm.id.upper <- rm.id - 1
        rm.id.lower <- rm.id + 1

        cv.id1 <- unique(c(rm.id, rm.id.upper, rm.id.lower))
        pos <- sapply(1:length(cv.id1), function(i) cv.id1[i] %in% oci)
        cv.id1 <- cv.id1[pos]
        if (length(cv.id1) >= count) {
            cv.id1 <- cv.id1[1:count]
            cv.id2 <- NULL
        } else {
            cv.id2 <- sample(setdiff(oci, cv.id1), (count - length(cv.id1)), replace = FALSE)
        }
        cv.id <- sort(c(cv.id1, cv.id2))
    }
    return(cv.id)
}



## cross validation sampling
# cv.sample <- function(I, D, count,
#                      cv.count = 3,
#                      cv.treat = FALSE) {

#    N <- dim(I)[2]
#    TT <- dim(I)[1]
#    tr.pos <- which(apply(D, 2, sum) >= 1) ## treated units
#    D.fake <- matrix(0, TT, N)

#    cv.id <- NULL
#    if (cv.treat == FALSE) {
#        oci <- which(c(I) == 1)
#    } else {
#        D.fake[, tr.pos] <- 1
#        oci <- which(c(I) == 1 & c(D.fake) == 1)
#    }

#    if (length(oci) <= count) {
#        stop("Too few observations are valid for cross-validation. Try setting the option cv.treat to FALSE.\n")
#    }

#    if (cv.count == 1 || count <= 2) {  ## randomly missing
#        cv.id <- sample(oci, count, replace = FALSE)
#    } else {
## remove boundary observation
#        if (cv.treat == FALSE) {
#            rm.pos <- c()
#            for (i in 1:(cv.count - 1)) {
#                rm.pos <- c(rm.pos, (TT * (0:(N-1)) + i))
#            }
#            oci2 <- setdiff(oci, rm.pos)
#        } else {
#            rm.pos <- c()
#            for (i in 1:(cv.count - 1)) {
#                rm.pos <- c(rm.pos, (TT * (tr.pos - 1) + i))
#            }
#            oci2 <- setdiff(oci, rm.pos)
#        }

## randomly select 1/cv.count
#        subcount <- floor(count/cv.count)
#        if (subcount == 0) {
#            subcount <- 1
#        }
#        rm.id <- sample(oci2, subcount, replace = FALSE)
#        rm.id.all <- c()
#        for (i in 1:(cv.count - 1)) {
#            rm.id.all <- c(rm.id.all, rm.id - i)
#        }
#        rm.id.all <- c(rm.id.all, rm.id)

#        cv.id1 <- unique(rm.id.all)
#        pos <- unlist(sapply(1:length(cv.id1), function(i) cv.id1[i] %in% oci))
#        cv.id1 <- cv.id1[pos]

#        if (length(cv.id1) >= count) {
#            cv.id1 <- cv.id1[1:count]
#            cv.id2 <- NULL
#        } else {
#            cv.id2 <- sample(setdiff(oci, cv.id1), (count - length(cv.id1)), replace = FALSE)
#        }
#
#        cv.id <- sort(c(cv.id1, cv.id2))
#    }

#    return(cv.id)
# }

## cross validation sampling
cv.sample <- function(I, D, count,
                      cv.count = 3,
                      cv.treat = FALSE,
                      cv.donut = 1) {
    ## prop <- sum(c(I))
    N <- dim(I)[2]
    TT <- dim(I)[1]
    tr.pos <- which(apply(D, 2, sum) >= 1) ## treated units
    D.fake <- matrix(0, TT, N)

    prop <- (TT * N) / sum(c(I))

    cv.id <- NULL
    if (cv.treat == FALSE) {
        oci <- which(c(I) == 1)
    } else {
        D.fake[, tr.pos] <- 1
        oci <- which(c(I) == 1 & c(D.fake) == 1)
        if (length(oci) <= count) {
            stop("Too few observations are valid for cross-validation. Try to set the option cv.treat to FALSE or set a smaller cv.prop.\n")
        }
    }

    if (cv.count == 1 || count <= 2) { ## randomly missing
        cv.id <- sample(oci, count, replace = FALSE)
        rm.id.use <- cv.id
    } else {
        res <- TT %% cv.count
        int <- floor(TT / cv.count)
        rm.pos <- c()

        ## randomly select 1/cv.count
        subcount <- floor(count / cv.count * prop)

        if (subcount == 0) {
            subcount <- 1
        }

        if (cv.treat == FALSE) {
            for (i in 1:N) {
                rm.pos <- c(rm.pos, (TT * (i - 1) + res) + seq(from = 1, by = cv.count, length.out = int))
            }
        } else {
            for (i in tr.pos) {
                rm.pos <- c(rm.pos, (TT * (i - 1) + res) + seq(from = 1, by = cv.count, length.out = int))
            }
        }

        rm.id <- sample(rm.pos, subcount, replace = FALSE)
        rm.id.all <- rm.id
        rm.id.use <- NULL

        if (cv.count == 2) {
            rm.id.all <- c(rm.id, rm.id + 1)
            rm.id.use <- rm.id.all
        } else {
            for (i in 0:(cv.count - 1)) {
                rm.id.all <- c(rm.id.all, rm.id + i)
                if (i >= cv.donut && i <= (cv.count - cv.donut - 1)) {
                    rm.id.use <- c(rm.id.use, rm.id + i)
                }
            }
        }

        ## rm.id.all <- c(rm.id, rm.id + 1, rm.id + 2)

        rm.id.all <- intersect(rm.id.all, oci) ## remove missing values
        rm.id.all <- sort(rm.id.all)

        rm.id.use <- intersect(rm.id.use, oci)
        rm.id.use <- sort(rm.id.use)

        cv.id2 <- NULL
        if (length(rm.id.all) >= count) {
            cv.id <- rm.id.all[1:count]
            pos.cv <- which(rm.id.use >= min(cv.id) & rm.id.use <= max(cv.id))
            rm.id.use <- rm.id.use[pos.cv]
            # if (length(rm.id.use) >= count) {
            #    rm.id.use <- rm.id.use[1:count]
            # }
        } else {
            cv.id2 <- sample(setdiff(oci, rm.id.all), (count - length(rm.id.all)), replace = FALSE)
            cv.id <- sort(c(rm.id.all, cv.id2))
            ## rm.id.use <- sort(c(rm.id.use, cv.id2))
        }

        ## remove boundary observation
        # if (cv.treat == FALSE) {
        #    rm.pos <- c()
        #    for (i in 1:(cv.count - 1)) {
        #        rm.pos <- c(rm.pos, (TT * (0:(N-1)) + i))
        #    }
        #    oci2 <- setdiff(oci, rm.pos)
        # } else {
        #    rm.pos <- c()
        #    for (i in 1:(cv.count - 1)) {
        #        rm.pos <- c(rm.pos, (TT * (tr.pos - 1) + i))
        #    }
        #    oci2 <- setdiff(oci, rm.pos)
        # }

        ## randomly select 1/cv.count
        # subcount <- floor(count/cv.count)
        # if (subcount == 0) {
        #    subcount <- 1
        # }
        # rm.id <- sample(oci2, subcount, replace = FALSE)
        # rm.id.all <- c()
        # for (i in 1:(cv.count - 1)) {
        #    rm.id.all <- c(rm.id.all, rm.id - i)
        # }
        # rm.id.all <- c(rm.id.all, rm.id)

        # cv.id1 <- unique(rm.id.all)
        # pos <- unlist(sapply(1:length(cv.id1), function(i) cv.id1[i] %in% oci))
        # cv.id1 <- cv.id1[pos]

        # if (length(cv.id1) >= count) {
        #    cv.id1 <- cv.id1[1:count]
        #    cv.id2 <- NULL
        # } else {
        #    cv.id2 <- sample(setdiff(oci, cv.id1), (count - length(cv.id1)), replace = FALSE)
        # }

        # cv.id <- sort(c(cv.id1, cv.id2))
    }

    return(list(
        cv.id = cv.id, ## marked id
        est.id = rm.id.use
    )) ## id used for mspe
}

res.vcov <- function(res, ## TT*Nboots
                     cov.ar = 1) {
    T <- dim(res)[1]
    I <- is.na(res)
    count <- matrix(NA, T, T)

    res[is.na(res)] <- 0
    vcov <- res %*% t(res)


    for (i in 1:T) {
        for (j in 1:T) {
            if (i > j) {
                count[i, j] <- count[j, i]
            } else {
                if ((j - i) <= cov.ar) {
                    II <- I[i, ] + I[j, ]
                    count[i, j] <- min(1 / sum(II == 0), 1)
                } else {
                    count[i, j] <- 0
                }
            }
        }
    }
    vcov <- vcov * count
    return(vcov)
}


## ---------------------------------------------------------------------------
## Rolling (forward-only) cross-validation fold construction.
##
## For each unit (column of II), masks the LAST `cv.count` observed positions
## (in time order). Training uses everything else: the unit's earlier
## observations + all other units' full data. This closes the forward-leakage
## channel that the random-anchor cv.sample() leaves open at cv.donut = 0/1
## under serially correlated residuals.
##
## Arguments
##   II         TT x N integer matrix; II[t, i] = 1 if unit i is observed at t.
##   D          TT x N treatment indicator (0/1). When cv.treat = FALSE
##              (default for rolling on controls), D is ignored.
##   cv.count   integer; number of observations to mask per unit (default 3).
##   cv.treat   logical; if TRUE restrict masking to treated units only.
##              Default FALSE (mask control units).
##
## Returns a list with the same shape as cv.sample():
##   cv.id      integer vector of masked positions (1-indexed into vec(II))
##   rm.id.use  identical to cv.id (no donut shaving — every masked position
##              is scored, since rolling already prevents forward leakage).
##
## Deterministic: no random sampling. The same II + cv.count always produces
## the same cv.id. Callers therefore typically use k = 1 fold.
## ---------------------------------------------------------------------------
cv.sample.rolling <- function(II, D = NULL,
                              cv.count = 3,
                              cv.treat = FALSE,
                              min.T0 = 5L) {
    TT <- dim(II)[1]
    N  <- dim(II)[2]
    cv.id <- integer(0)

    ## Identify units to mask.
    if (isTRUE(cv.treat) && !is.null(D)) {
        ever_treat <- apply(D, 2, function(v) any(!is.na(v) & v >= 1))
        unit_idx   <- which(ever_treat)
    } else {
        unit_idx   <- seq_len(N)
    }

    ## Only mask units that retain >= min.T0 observations after masking.
    ## (Otherwise the downstream con2 check restores them and the rolling
    ## fold collapses to a tiny set of effective masks.)
    threshold <- min.T0 + cv.count
    for (j in unit_idx) {
        obs_t <- which(II[, j] == 1L)
        if (length(obs_t) >= threshold) {
            mask_t <- utils::tail(obs_t, cv.count)
            ## Convert (t, j) to a 1-indexed position into vec(II).
            cv.id <- c(cv.id, (j - 1L) * TT + mask_t)
        }
    }
    list(cv.id = cv.id, rm.id.use = cv.id)
}


v_replace <- function(needle, haystack) {
    sieved <- which(haystack == needle[1L])
    for (i in seq.int(1L, length(needle) - 1L)) {
        sieved <- sieved[haystack[sieved + i] == needle[i + 1L]]
    }
    out <- rep(NA, length(haystack))
    if (length(sieved) == 0) {
        return(out)
    }
    for (index in sieved) {
        if (!is.na(index)) {
            out[c(index:(index + length(needle) - 1))] <- needle
        }
    }
    return(out)
}

## Internal modern-theme helpers for plot.fect() and esplot().
## Not exported. Loaded by esplot() / plot.R when theme.bw = TRUE.

# Color constants for the modern recipe
.MODERN_EST_COLOR    <- "grey20"   # connected line / point default
.MODERN_REF_COLOR    <- "grey75"   # y=0 hline / treatment-onset vline
.MODERN_PLACEBO_PT   <- "#E07A2B"  # orange — placebo highlight point
.MODERN_PLACEBO_FILL <- "#F5C8A8"  # peach — placebo window background
.MODERN_CARRYOVER_PT   <- "#2B6CB0" # blue — carryover highlight point
.MODERN_CARRYOVER_FILL <- "#C7DEF0" # light-blue — carryover window background

.modern_theme <- function(base_size = 11) {
  ggplot2::theme_bw(base_size = base_size) +
    ggplot2::theme(
      plot.title         = ggplot2::element_text(size = base_size, face = "plain",
                                                 hjust = 0,
                                                 margin = ggplot2::margin(b = 4)),
      plot.subtitle      = ggplot2::element_text(size = base_size - 2, color = "grey40",
                                                 margin = ggplot2::margin(b = 8)),
      panel.grid.minor   = ggplot2::element_blank(),
      panel.grid.major.x = ggplot2::element_blank(),
      plot.margin        = ggplot2::margin(8, 12, 8, 8),
      ## Compact legend so it doesn't dominate the panel in subfigure use:
      ## tight top margin against the panel, small text, small keys.
      legend.position    = "bottom",
      legend.margin      = ggplot2::margin(t = 0, r = 0, b = 0, l = 0),
      legend.box.spacing = grid::unit(2, "pt"),
      legend.text        = ggplot2::element_text(size = base_size - 3),
      legend.title       = ggplot2::element_text(size = base_size - 3),
      legend.key.size    = grid::unit(10, "pt"),
      legend.spacing.x   = grid::unit(4, "pt"),
      legend.spacing.y   = grid::unit(0, "pt")
    )
}

## Lighten a color toward white. Used to auto-derive rect-fill from a point color
## (e.g., orange #E07A2B -> peach #F5C8A8).
.lighten_color <- function(col, amount = 0.75) {
  if (is.null(col) || length(col) == 0) return(NA_character_)
  pal <- grDevices::colorRampPalette(c(col, "white"))(100)
  pal[max(1L, min(100L, round(amount * 100)))]
}

###################################################################
## permutation
###################################################################
fect_permu <- function(Y,
                       X,
                       D,
                       I,
                       permu.dimension = "time",
                       r.cv = NULL,
                       lambda.cv = NULL,
                       m = 2,
                       method = "ife",
                       degree = 2,
                       knots = NULL,
                       force,
                       tol,
                       norm.para,
                       nboots,
                       parallel = TRUE,
                       cores = NULL) {
    TT <- dim(Y)[1]
    n.shuffle <- TT %/% m
    t.pos <- rep(1:n.shuffle, each = m)
    if (TT > n.shuffle * m) {
        t.pos <- c(t.pos, rep(n.shuffle + 1, TT - n.shuffle * m))
    }
    tt <- 1:TT
    ## list
    l.tt <- split(tt, t.pos)

    if (permu.dimension == "time") {
        sub.permu <- function() {
            sub.pos <- c()
            tt.length <- length(l.tt)
            one.rank <- sample(1:tt.length, tt.length, replace = FALSE)
            for (i in 1:tt.length) {
                sub.pos <- c(sub.pos, unlist(l.tt[[one.rank[i]]]))
            }
            Y.permu <- as.matrix(Y[sub.pos, ])
            I.permu <- as.matrix(I[sub.pos, ])
            # D.permu <- as.matrix(D[sub.pos, ])
            if (!is.null(X)) {
                X.permu <- X[sub.pos, , , drop = FALSE]
            }
            result <- try(one.permu(
                Y.permu,
                X.permu,
                D,
                I.permu,
                r.cv,
                lambda.cv,
                method,
                degree,
                knots,
                force,
                tol,
                norm.para
            ), silent = TRUE)


            if ("try-error" %in% class(result)) {
                return(NA)
            } else {
                return(result)
            }
        }
    }

    if (permu.dimension == "unit") {
        NN <- dim(Y)[2]
        index.NN <- c(1:NN)

        sub.permu <- function() {
            sample.unit <- sample(index.NN,
                size = length(index.NN),
                replace = FALSE
            )

            Y.permu <- as.matrix(Y[, sample.unit])
            I.permu <- as.matrix(I[, sample.unit])
            if (!is.null(X)) {
                X.permu <- X[, sample.unit, , drop = FALSE]
            }
            result <- try(one.permu(
                Y.permu,
                X.permu,
                D,
                I.permu,
                r.cv,
                lambda.cv,
                method,
                degree,
                knots,
                force,
                tol,
                norm.para
            ), silent = TRUE)


            if ("try-error" %in% class(result)) {
                return(NA)
            } else {
                return(result)
            }
        }
    }


    att.avg.permu <- rep(NA, nboots)

    if (!identical(parallel, FALSE)) {
        permu.out <- foreach(
            j = 1:nboots,
            .inorder = FALSE,
            .export = c("one.permu"),
            .packages = c("fect")
        ) %dopar% {
            return(sub.permu())
        }

        for (j in 1:nboots) {
            att.avg.permu[j] <- permu.out[[j]]
        }
    } else {
        for (j in 1:nboots) {
            permu <- sub.permu()
            att.avg.permu[j] <- permu
            ## report progress
            if (j %% 100 == 0) {
                message(".")
            }
        }
    }

    if (sum(is.na(c(att.avg.permu))) > 0) {
        permu.rm <- which(is.na(c(att.avg.permu)))
        att.avg.permu <- att.avg.permu[-permu.rm]
    }

    message(length(att.avg.permu), " permutes\n", sep = "")

    return(att.avg.permu)
}


## sub-function

one.permu <- function(Y, # Outcome variable, (T*N) matrix
                      X, # Explanatory variables:  (T*N*p) array
                      D, #  Indicator for treated unit (tr==1)
                      I,
                      r.cv = 0, # initial number of factors considered if CV==1
                      lambda.cv = 1,
                      method = "fe",
                      degree = 2,
                      knots = NULL,
                      force,
                      tol, # tolerance level
                      norm.para = NULL) {
    ## -------------------------------##
    ## Parsing data
    ## -------------------------------##
    placebo.pos <- na.pos <- NULL
    res.sd1 <- res.sd2 <- NULL

    ## unit id and time
    TT <- dim(Y)[1]
    N <- dim(Y)[2]
    if (is.null(X) == FALSE) {
        p <- dim(X)[3]
    } else {
        p <- 0
        X <- array(0, dim = c(1, 1, 0))
    }

    ## observed indicators:
    II <- I
    II[which(D == 1)] <- 0

    ## remove period when no obserbations are under control
    T0 <- apply(II, 1, sum)
    XX <- NULL
    if (min(T0) == 0) {
        t0.pos <- which(T0 == 0)
        TT <- TT - length(t0.pos)
        if (TT <= 2) {
            stop("No enough observations under control.\n")
        }
        Y <- as.matrix(Y[-t0.pos, ])
        D <- as.matrix(D[-t0.pos, ])
        I <- as.matrix(I[-t0.pos, ])
        II <- as.matrix(II[-t0.pos, ])
        if (is.null(X) == FALSE) {
            XX <- array(NA, dim = c(TT, N, p))
            for (i in 1:p) {
                subX <- X[, , i]
                XX[, , i] <- as.matrix(subX[-t0.pos, ])
            }
            X <- XX
        }
    }

    ## replicate data
    YY <- Y
    YY[which(II == 0)] <- 0 ## reset to 0
    D[which(I == 0)] <- 0
    oci <- which(c(II) == 1)

    if (sum(D) == 0) {
        stop("No valid observations under treatment.\n")
    }

    if (method %in% c("polynomial", "bspline")) {
        stop("Doesn't support in this version.")
    } else {
        ## initial fit using fastplm
        data.ini <- matrix(NA, (TT * N), (2 + 1 + p))
        data.ini[, 2] <- rep(1:N, each = TT) ## unit fe
        data.ini[, 3] <- rep(1:TT, N) ## time fe
        data.ini[, 1] <- c(Y) ## outcome
        if (p > 0) { ## covar
            for (i in 1:p) {
                data.ini[, (3 + i)] <- c(X[, , i])
            }
        }
        ## observed Y0 indicator:
        initialOut <- Y0 <- beta0 <- FE0 <- xi0 <- factor0 <- NULL

        initialOut <- initialFit(data = data.ini, force = force, oci = oci)
        Y0 <- initialOut$Y0
        beta0 <- initialOut$beta0
        if (p > 0 && sum(is.na(beta0)) > 0) {
            beta0[which(is.na(beta0))] <- 0
        }

        ## -------------------------------##
        ## ----------- Main Algorithm ----------- ##
        ## -------------------------------##

        est <- NULL
        if (method == "fe") {
            est <- inter_fe_ub(YY, Y0, X, II, beta0, 0, force = force, tol)
        } else if (method == "ife") {
            est <- inter_fe_ub(YY, Y0, X, II, beta0, r.cv, force = force, tol)
        } else if (method == "mc") {
            est <- inter_fe_mc(YY, Y0, X, II, beta0, 1, lambda.cv, force, tol)
        }

        if (!is.null(norm.para)) {
            est$fit <- est$fit * norm.para[1]
        }
        Y.ct <- est$fit
    }

    if (!is.null(norm.para)) {
        Y <- Y * norm.para[1]
    }

    eff <- Y - Y.ct

    att.avg <- sum(eff * D) / (sum(D))

    return(abs(att.avg))
}

## valid_controls() — control screening helper for parametric bootstrap
##
## Returns the subset of id.co indices that pass the pre/post-period
## observation threshold.  Used by fect_boot() before draw.error() runs.
##
## Arguments:
##   out        – original-fit result list (contains $D, $I, $r.cv)
##   method     – character scalar: "gsynth", "ife", "cfe"  (mc is blocked upstream)
##   predictive – character scalar: "notyettreated" or "nevertreated"
##   force      – integer scalar: 0, 1, 2, or 3

valid_controls <- function(out, method, predictive, force) {

  if (method == "mc") {
    stop("valid_controls: MC not supported in this release.")
  }

  id.co  <- which(colSums(out$D) == 0)
  id.tr  <- which(colSums(out$D) > 0)
  I.co   <- out$I[, id.co, drop = FALSE]
  TT     <- nrow(out$D)

  T0.ub      <- colSums(out$D[, id.tr, drop = FALSE] == 0)
  T0.ub.min  <- min(T0.ub)
  max.T0.ub  <- max(T0.ub)

  co.pre  <- colSums(I.co[seq_len(T0.ub.min), , drop = FALSE])
  co.post <- colSums(I.co[seq.int(max.T0.ub + 1L, TT), , drop = FALSE])

  threshold <- out$r.cv + as.integer(force %in% c(1, 3))

  valid.co <- id.co[(co.pre >= threshold) & (co.post >= 1L)]
  return(valid.co)
}

## Unit tests for impute_Y0() dispatcher and valid_controls() helper
## (Phase 1 + Phase 2 of the parametric-bootstrap refactor)
##
## These tests verify:
##   - impute_Y0 dispatches to the correct callee for each (method, predictive) combination
##   - impute_Y0 errors on unsupported combinations (mc)
##   - valid_controls returns the correct subset of control columns
##   - hasRevs+parametric gate fires correctly in fect()

test_that("impute_Y0 errors on mc method", {
  skip_on_cran()

  # Build a minimal data stub so impute_Y0 would reach the else branch
  # We don't need real data — just check the error fires before any callee call
  expect_error(
    impute_Y0(
      method     = "mc",
      predictive = "nevertreated",
      Y          = matrix(0, 5, 4),
      X          = NULL,
      D          = matrix(0, 5, 4),
      W          = NULL,
      I          = matrix(1, 5, 4),
      II         = matrix(1, 5, 4),
      T.on       = matrix(0L, 5, 4),
      tuning     = 1L,
      force      = 0,
      tol        = 1e-3,
      max.iteration = 10
    ),
    regexp = "MC support is planned for Phase 4"
  )
})

test_that("impute_Y0 errors on unknown method+predictive combination", {
  skip_on_cran()

  expect_error(
    impute_Y0(
      method     = "zzz",
      predictive = "nevertreated",
      Y          = matrix(0, 5, 4),
      X          = NULL,
      D          = matrix(0, 5, 4),
      W          = NULL,
      I          = matrix(1, 5, 4),
      II         = matrix(1, 5, 4),
      T.on       = matrix(0L, 5, 4),
      tuning     = 1L,
      force      = 0,
      tol        = 1e-3,
      max.iteration = 10
    ),
    regexp = "unsupported combination"
  )
})

test_that("valid_controls errors on mc method", {
  skip_on_cran()

  # Minimal out stub
  out_stub <- list(
    D   = matrix(c(0, 0, 1, 1, 0, 0, 0, 0, 0, 0), nrow = 5, ncol = 2),
    I   = matrix(1, 5, 2),
    r.cv = 1L
  )
  expect_error(
    valid_controls(out_stub, method = "mc", predictive = "nevertreated", force = 0),
    regexp = "MC not supported"
  )
})

test_that("valid_controls returns correct subset without FE adjustment", {
  skip_on_cran()

  # 5 time periods, 3 units: unit 1 = treated (onset t=3), units 2-3 = control
  # D:
  #   unit 1: c(0,0,1,1,1)
  #   unit 2: c(0,0,0,0,0)
  #   unit 3: c(0,0,0,0,0)
  TT <- 5
  N <- 3
  D <- matrix(0, TT, N)
  D[3:5, 1] <- 1
  I <- matrix(1, TT, N)
  r.cv <- 1L

  out_stub <- list(D = D, I = I, r.cv = r.cv)

  # T0.ub for treated unit 1 = 2 (periods 1-2 are pre-treatment)
  # max.T0.ub = 2, T0.ub.min = 2
  # co.pre[j] = sum(I.co[1:2, j]) for each control
  # co.post[j] = sum(I.co[3:5, j]) for each control
  # With full I, co.pre = 2, co.post = 3 for all controls
  # threshold (force=0): r.cv = 1
  # 2 >= 1 AND 3 >= 1 -> both controls should pass
  vc <- valid_controls(out_stub, method = "ife", predictive = "notyettreated", force = 0)
  expect_equal(vc, c(2L, 3L))
})

test_that("valid_controls applies +1 threshold when force has unit FE", {
  skip_on_cran()

  TT <- 5
  N <- 3
  D <- matrix(0, TT, N)
  D[3:5, 1] <- 1
  I <- matrix(1, TT, N)
  r.cv <- 2L  # threshold without FE = 2, with unit FE = 3

  out_stub <- list(D = D, I = I, r.cv = r.cv)

  # co.pre = 2, threshold (force=1) = 2 + 1 = 3
  # 2 < 3, so co.pre threshold fails -> both controls excluded
  vc_unit_fe <- valid_controls(out_stub, method = "ife", predictive = "notyettreated", force = 1)
  expect_equal(length(vc_unit_fe), 0L)

  # Without FE: threshold = 2, co.pre = 2 >= 2 -> both pass
  vc_no_fe <- valid_controls(out_stub, method = "ife", predictive = "notyettreated", force = 0)
  expect_equal(vc_no_fe, c(2L, 3L))
})

test_that("hasRevs+parametric gate fires in fect()", {
  skip_on_cran()

  # Load a dataset with treatment reversals
  data(simdata)  # standard simdata in the package
  df <- simdata

  # Manually introduce a reversal: flip treatment off for one unit in the middle
  # Find a treated unit and flip one period from 1 -> 0 in the middle
  treated_units <- unique(df$id[df$D == 1])
  if (length(treated_units) < 1) {
    skip("No treated units in simdata")
  }
  target_unit <- treated_units[1]
  unit_rows <- which(df$id == target_unit & df$D == 1)
  if (length(unit_rows) < 2) {
    skip("Insufficient treated periods to introduce reversal")
  }
  # Introduce a reversal: set one treated period back to 0
  mid_row <- unit_rows[ceiling(length(unit_rows) / 2)]
  df$D[mid_row] <- 0

  expect_error(
    fect(Y ~ D, data = df, index = c("id", "time"),
         method = "ife", vartype = "parametric", se = TRUE, nboots = 5),
    regexp = "parametric.*reversal|reversal.*parametric",
    ignore.case = TRUE
  )
})

## Paraboot parity and bug-fix validation tests
## Generated by tester Phase 1 (StatsClaw run: 2026-04-13-para-boot-refactor)
## Revised §3 tests: 2026-04-14 (revised test-spec.md §3 after planner fix)
##
## Covers: PAR-1 through PAR-7, BUG-1a/2a/3a, BUG-2b, BUG-1c/2c/3c,
##         BUG-4 through BUG-6, GATE-1 through GATE-3,
##         UNIT-1 through UNIT-4, DEAD-1
##
## NOTE on out$D vs out$D.dat:
## The fect() result list stores the treatment matrix as `D.dat` (TT x N matrix).
## `out$D` is a character string (the formula RHS variable name), not the matrix.
## All UNIT-* and valid_controls tests use out$D.dat for colSums() calls.
## valid_controls() accesses out$D internally; the caller passes a stub list
## with D set to the matrix (D.dat).

## ---- helpers ----------------------------------------------------------------

load_baseline <- function() {
  path <- testthat::test_path("fixtures", "paraboot-baseline.rds")
  if (!file.exists(path)) {
    skip("Pre-refactor fixture not found — run fixture capture first")
  }
  readRDS(path)
}

make_fixture_data <- function(seed = 42) {
  set.seed(seed)
  N <- 50; TT <- 20; T0 <- 12; Ntr <- 10
  id   <- rep(1:N, each = TT)
  time <- rep(1:TT, N)
  D    <- as.integer(id <= Ntr & time >= T0)
  alpha_i <- rep(rnorm(N, 0, 1), each = TT)
  xi_t    <- rep(rnorm(TT, 0, 0.5), N)
  Y <- alpha_i + xi_t + 2 * D + rnorm(N * TT, 0, 0.5)
  data.frame(id = id, time = time, Y = Y, D = D)
}

## ---- PAR-1: gsynth+parametric parity ----------------------------------------

test_that("PAR-1: gsynth+parametric parity (byte-identical pre/post refactor)", {
  skip_on_cran()
  baseline <- load_baseline()
  d <- make_fixture_data()

  set.seed(2026)
  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "gsynth", r = 1, se = TRUE,
    vartype = "parametric", nboots = 50,
    CV = FALSE, parallel = FALSE
  )))

  expect_true(identical(out$att.avg, baseline$gsynth_para$att.avg),
    info = "att.avg must be byte-identical")
  expect_true(identical(out$att, baseline$gsynth_para$att),
    info = "att vector must be byte-identical")
  expect_true(identical(out$count, baseline$gsynth_para$count),
    info = "count vector must be byte-identical")
  expect_true(identical(out$r.cv, baseline$gsynth_para$r.cv),
    info = "r.cv must be byte-identical")
  expect_true(identical(out$est.att, baseline$gsynth_para$est.att),
    info = "est.att (SE/CI table) must be byte-identical")
})

## ---- PAR-2: ife+nevertreated+parametric parity ------------------------------

test_that("PAR-2: ife+nevertreated+parametric parity (byte-identical; must equal baseline)", {
  skip_on_cran()
  baseline <- load_baseline()
  d <- make_fixture_data()

  ## Pin pre-2.4.3 EM defaults: baseline rds was captured at tol=1e-3,
  ## max.iteration=1000 (v2.4.2 defaults). v2.4.3 tightened defaults to
  ## tol=1e-5/max.iteration=5000. Test preserves byte-equality contract
  ## against the existing baseline by passing legacy values explicitly.
  set.seed(2026)
  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "ife", time.component.from = "nevertreated",
    r = 1, se = TRUE, vartype = "parametric", nboots = 50,
    CV = FALSE, parallel = FALSE,
    tol = 1e-3, max.iteration = 1000
  )))

  expect_true(identical(out$att.avg, baseline$ife_nev_para$att.avg))
  expect_true(identical(out$att,     baseline$ife_nev_para$att))
  expect_true(identical(out$count,   baseline$ife_nev_para$count))
  expect_true(identical(out$est.att, baseline$ife_nev_para$est.att))
})

## ---- PAR-3: bootstrap parity — ife+notyettreated ----------------------------

test_that("PAR-3: bootstrap vartype parity — ife+notyettreated", {
  skip_on_cran()
  baseline <- load_baseline()
  d <- make_fixture_data()

  ## Pin pre-2.4.3 EM defaults; see PAR-2.
  set.seed(2026)
  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "ife", r = 1, se = TRUE,
    vartype = "bootstrap", nboots = 50, CV = FALSE, parallel = FALSE,
    tol = 1e-3, max.iteration = 1000
  )))

  expect_true(identical(out$att.avg, baseline$ife_nt_boot$att.avg))
  expect_true(identical(out$att,     baseline$ife_nt_boot$att))
  expect_true(identical(out$est.att, baseline$ife_nt_boot$est.att))
})

## ---- PAR-4: bootstrap parity — ife+nevertreated -----------------------------

test_that("PAR-4: bootstrap vartype parity — ife+nevertreated", {
  skip_on_cran()
  baseline <- load_baseline()
  d <- make_fixture_data()

  ## Pin pre-2.4.3 EM defaults; see PAR-2.
  set.seed(2026)
  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "ife", time.component.from = "nevertreated",
    r = 1, se = TRUE, vartype = "bootstrap", nboots = 50, CV = FALSE, parallel = FALSE,
    tol = 1e-3, max.iteration = 1000
  )))

  expect_true(identical(out$att.avg, baseline$ife_nev_boot$att.avg))
  expect_true(identical(out$att,     baseline$ife_nev_boot$att))
  expect_true(identical(out$est.att, baseline$ife_nev_boot$est.att))
})

## ---- PAR-5: bootstrap parity — gsynth ---------------------------------------

test_that("PAR-5: bootstrap vartype parity — gsynth", {
  skip_on_cran()
  baseline <- load_baseline()
  d <- make_fixture_data()

  set.seed(2026)
  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "gsynth", r = 1, se = TRUE,
    vartype = "bootstrap", nboots = 50, CV = FALSE, parallel = FALSE
  )))

  expect_true(identical(out$att.avg, baseline$gsynth_boot$att.avg))
  expect_true(identical(out$att,     baseline$gsynth_boot$att))
  expect_true(identical(out$est.att, baseline$gsynth_boot$est.att))
})

## ---- PAR-6: jackknife parity — ife+notyettreated ----------------------------

test_that("PAR-6: jackknife parity — ife+notyettreated", {
  skip_on_cran()
  baseline <- load_baseline()
  d <- make_fixture_data()

  ## Pin pre-2.4.3 EM defaults; see PAR-2.
  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "ife", r = 1, se = TRUE,
    vartype = "jackknife", CV = FALSE, parallel = FALSE,
    tol = 1e-3, max.iteration = 1000
  )))

  expect_true(identical(out$att.avg, baseline$ife_jk$att.avg))
  expect_true(identical(out$att,     baseline$ife_jk$att))
  expect_true(identical(out$est.att, baseline$ife_jk$est.att))
})

## ---- PAR-7: jackknife parity — gsynth ----------------------------------------

test_that("PAR-7: jackknife parity — gsynth", {
  skip_on_cran()
  baseline <- load_baseline()
  d <- make_fixture_data()

  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "gsynth", r = 1, se = TRUE,
    vartype = "jackknife", CV = FALSE, parallel = FALSE
  )))

  expect_true(identical(out$att.avg, baseline$gsynth_jk$att.avg))
  expect_true(identical(out$att,     baseline$gsynth_jk$att))
  expect_true(identical(out$est.att, baseline$gsynth_jk$est.att))
})

## ---- BUG-1a: ife+notyettreated+parametric est.att changed from pre-fix ------

test_that("BUG-1a: ife+notyettreated+parametric now errors via notyettreated gate", {
  skip_on_cran()
  d <- make_fixture_data()
  expect_error(
    suppressWarnings(suppressMessages(fect(
      Y ~ D, data = d, index = c("id", "time"),
      method = "ife", r = 1, se = TRUE,
      vartype = "parametric", nboots = 50,
      CV = FALSE, parallel = FALSE
    ))),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

## ---- BUG-2a: cfe+nevertreated+parametric est.att changed from pre-fix -------

test_that("BUG-2a: cfe+nevertreated+parametric: est.att changed from pre-fix baseline", {
  skip_on_cran()
  baseline <- load_baseline()
  d <- make_fixture_data()

  set.seed(2026)
  out_post <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "cfe", time.component.from = "nevertreated",
    r = 1, se = TRUE, vartype = "parametric", nboots = 50,
    CV = FALSE, parallel = FALSE
  )))

  expect_true(
    isTRUE(all.equal(out_post$att.avg, baseline$cfe_nev_para$att.avg, tolerance = 1e-10)),
    info = "att.avg must be unchanged by Loop 2 fix"
  )
  expect_false(
    identical(out_post$est.att, baseline$cfe_nev_para$est.att),
    info = "est.att must differ from pre-fix baseline: Loop 2 was calling fect_nevertreated(ife) instead of fect_nevertreated(cfe)"
  )
})

## ---- BUG-3a: cfe+notyettreated+parametric est.att changed from pre-fix ------

test_that("BUG-3a: cfe+notyettreated+parametric now errors via notyettreated gate", {
  skip_on_cran()
  d <- make_fixture_data()
  expect_error(
    suppressWarnings(suppressMessages(fect(
      Y ~ D, data = d, index = c("id", "time"),
      method = "cfe", r = 0, se = TRUE,
      vartype = "parametric", nboots = 50,
      CV = FALSE, parallel = FALSE
    ))),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

## ---- BUG-2b: cfe+nevertreated no longer byte-identical to gsynth ------------

test_that("BUG-2b: cfe+nevertreated+parametric no longer byte-identical to gsynth+parametric", {
  skip_on_cran()
  d <- make_fixture_data()

  set.seed(2026)
  out_cfe_nev <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "cfe", time.component.from = "nevertreated",
    r = 1, se = TRUE, vartype = "parametric", nboots = 50,
    CV = FALSE, parallel = FALSE
  )))

  set.seed(2026)
  out_gsynth <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "gsynth", r = 1, se = TRUE,
    vartype = "parametric", nboots = 50,
    CV = FALSE, parallel = FALSE
  )))

  expect_false(
    identical(out_cfe_nev$est.att, out_gsynth$est.att),
    info = paste(
      "Pre-fix: cfe+nevertreated and gsynth had identical est.att (both used fect_nevertreated(ife) in Loop 2).",
      "Post-fix: cfe+nevertreated uses fect_nevertreated(cfe) in Loop 2; the two bootstrap distributions must differ."
    )
  )
})

## ---- BUG-1c: ife+notyettreated att.avg stability ----------------------------

test_that("BUG-1c: ife+notyettreated+parametric (se=TRUE) now errors via notyettreated gate", {
  skip_on_cran()
  d <- make_fixture_data()
  ## se=FALSE still works
  out_no_se <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "ife", r = 1, se = FALSE, CV = FALSE, parallel = FALSE
  )))
  expect_s3_class(out_no_se, "fect")
  ## se=TRUE, vartype="parametric" + default factors.from now errors
  expect_error(
    suppressWarnings(suppressMessages(fect(
      Y ~ D, data = d, index = c("id", "time"),
      method = "ife", r = 1, se = TRUE,
      vartype = "parametric", nboots = 50,
      CV = FALSE, parallel = FALSE
    ))),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

## ---- BUG-2c: cfe+nevertreated att.avg stability -----------------------------

test_that("BUG-2c: cfe+nevertreated att.avg unchanged between se=FALSE and se=TRUE (parametric)", {
  skip_on_cran()
  d <- make_fixture_data()

  out_no_se <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "cfe", time.component.from = "nevertreated",
    r = 1, se = FALSE, CV = FALSE, parallel = FALSE
  )))

  set.seed(2026)
  out_with_se <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "cfe", time.component.from = "nevertreated",
    r = 1, se = TRUE, vartype = "parametric", nboots = 50,
    CV = FALSE, parallel = FALSE
  )))

  expect_true(
    isTRUE(all.equal(out_with_se$att.avg, out_no_se$att.avg, tolerance = 1e-10)),
    info = "att.avg must be identical whether se=FALSE or se=TRUE"
  )
})

## ---- BUG-3c: cfe+notyettreated att.avg stability ----------------------------

test_that("BUG-3c: cfe+notyettreated+parametric (se=TRUE) now errors via notyettreated gate", {
  skip_on_cran()
  d <- make_fixture_data()
  ## se=FALSE still works
  out_no_se <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "cfe", r = 0, se = FALSE, CV = FALSE, parallel = FALSE
  )))
  expect_s3_class(out_no_se, "fect")
  ## se=TRUE, vartype="parametric" + default factors.from now errors
  expect_error(
    suppressWarnings(suppressMessages(fect(
      Y ~ D, data = d, index = c("id", "time"),
      method = "cfe", r = 0, se = TRUE,
      vartype = "parametric", nboots = 50,
      CV = FALSE, parallel = FALSE
    ))),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

## ---- BUG-4: ife+notyettreated+parametric SE smoke test -----------------------

test_that("BUG-4: ife+notyettreated+parametric now errors via notyettreated gate", {
  skip_on_cran()
  d <- make_fixture_data()
  expect_error(
    suppressWarnings(suppressMessages(fect(
      Y ~ D, data = d, index = c("id", "time"),
      method = "ife", r = 1, se = TRUE,
      vartype = "parametric", nboots = 50,
      CV = FALSE, parallel = FALSE
    ))),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

## ---- BUG-5: cfe+nevertreated+parametric SE smoke test -----------------------

test_that("BUG-5: cfe+nevertreated+parametric produces non-degenerate SE post-fix", {
  skip_on_cran()
  d <- make_fixture_data()
  set.seed(2026)
  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "cfe", time.component.from = "nevertreated",
    r = 1, se = TRUE, vartype = "parametric", nboots = 50,
    CV = FALSE, parallel = FALSE
  )))
  expect_s3_class(out, "fect")
  expect_true(!is.null(out$est.att))
})

## ---- BUG-6: cfe+notyettreated+parametric SE smoke test ----------------------

test_that("BUG-6: cfe+notyettreated+parametric now errors via notyettreated gate", {
  skip_on_cran()
  d <- make_fixture_data()
  expect_error(
    suppressWarnings(suppressMessages(fect(
      Y ~ D, data = d, index = c("id", "time"),
      method = "cfe", r = 0, se = TRUE,
      vartype = "parametric", nboots = 50,
      CV = FALSE, parallel = FALSE
    ))),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

## ---- GATE-1: hasRevs+parametric errors --------------------------------------

test_that("GATE-1: vartype='parametric' with treatment reversals errors with clear message", {
  skip_on_cran()
  N <- 30; TT <- 20
  set.seed(99)
  id   <- rep(1:N, each = TT)
  time <- rep(1:TT, N)
  D <- ifelse(id <= 10 & time >= 10 & time <= 15, 1, 0)
  Y <- rnorm(N * TT)
  d_rev <- data.frame(id = id, time = time, Y = Y, D = D)

  expect_error(
    suppressMessages(fect(
      Y ~ D, data = d_rev, index = c("id", "time"),
      method = "ife", r = 0, se = TRUE,
      vartype = "parametric", nboots = 30,
      CV = FALSE, parallel = FALSE
    )),
    regexp = "reversal|Parametric bootstrap is not valid",
    info = "Error message must mention treatment reversal"
  )
})

## ---- GATE-2: no reversals does NOT error ------------------------------------

test_that("GATE-2: vartype='parametric' without reversals does NOT error", {
  skip_on_cran()
  d <- make_fixture_data()
  expect_no_error(suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "gsynth", r = 1, se = TRUE,
    vartype = "parametric", nboots = 20,
    CV = FALSE, parallel = FALSE
  ))))
})

## ---- GATE-3: mc+parametric still errors -------------------------------------

test_that("GATE-3: vartype='parametric' with method='mc' still errors (Phase 3 scope unchanged)", {
  skip_on_cran()
  d <- make_fixture_data()
  expect_error(
    fect(
      Y ~ D, data = d, index = c("id", "time"),
      method = "mc", se = TRUE,
      vartype = "parametric", nboots = 30,
      CV = FALSE, parallel = FALSE
    ),
    regexp = "parametric"
  )
})

## ---- UNIT-1: impute_Y0 dispatches to fect_nevertreated for gsynth -----------

test_that("UNIT-1: impute_Y0 dispatches to fect_nevertreated for gsynth", {
  skip_on_cran()
  d <- make_fixture_data()
  out0 <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "gsynth", r = 1, se = FALSE, CV = FALSE, parallel = FALSE
  )))
  D_mat   <- out0$D.dat
  id.tr  <- which(colSums(D_mat) > 0)
  id.co  <- which(colSums(D_mat) == 0)
  id.boot <- c(id.tr, id.co)
  Y.boot  <- out0$Y.ct[, id.boot]; Y.boot[out0$I[, id.boot] == 0] <- 0
  result <- fect:::impute_Y0(
    method = "gsynth", predictive = "nevertreated",
    Y = Y.boot, X = NULL, D = D_mat[, id.boot],
    W = NULL, I = out0$I[, id.boot], II = out0$II[, id.boot],
    T.on = out0$T.on[, id.boot],
    tuning = out0$r.cv, boot = 1,
    force = out0$force, hasRevs = out0$hasRevs,
    tol = 1e-3, max.iteration = 1000
  )
  expect_true(is.list(result))
  expect_true("eff.tr" %in% names(result))
  expect_true("att.avg" %in% names(result))
})

## ---- UNIT-2: impute_Y0 dispatches to fect_fe for ife+notyettreated ----------

test_that("UNIT-2: impute_Y0 dispatches to fect_fe for ife+notyettreated", {
  skip_on_cran()
  d <- make_fixture_data()
  out0 <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "ife", r = 1, se = FALSE, CV = FALSE, parallel = FALSE
  )))
  D_mat   <- out0$D.dat
  id.tr  <- which(colSums(D_mat) > 0)
  id.co  <- which(colSums(D_mat) == 0)
  id.boot <- c(id.tr, id.co)
  Y.boot  <- out0$Y.ct[, id.boot]; Y.boot[out0$I[, id.boot] == 0] <- 0
  result <- fect:::impute_Y0(
    method = "ife", predictive = "notyettreated",
    Y = Y.boot, X = NULL, D = D_mat[, id.boot],
    W = NULL, I = out0$I[, id.boot], II = out0$II[, id.boot],
    T.on = out0$T.on[, id.boot],
    tuning = out0$r.cv, boot = 1,
    force = out0$force, hasRevs = out0$hasRevs,
    tol = 1e-3, max.iteration = 1000
  )
  expect_true(is.list(result))
  expect_true("eff.tr" %in% names(result))
  expect_true("att.avg" %in% names(result))
})

## ---- UNIT-3: impute_Y0 errors for mc+any ------------------------------------

test_that("UNIT-3: impute_Y0 errors for mc+any", {
  skip_on_cran()
  d <- make_fixture_data()
  out0 <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "ife", r = 1, se = FALSE, CV = FALSE, parallel = FALSE
  )))
  D_mat    <- out0$D.dat
  Y_mat    <- out0$Y.dat
  I_mat    <- out0$I
  II_mat   <- out0$II
  T_on_mat <- out0$T.on
  id.boot  <- 1:ncol(D_mat)
  expect_error(
    fect:::impute_Y0(
      method = "mc", predictive = "notyettreated",
      Y = Y_mat[, id.boot], X = NULL, D = D_mat[, id.boot],
      W = NULL, I = I_mat[, id.boot], II = II_mat[, id.boot],
      T.on = T_on_mat[, id.boot],
      tuning = 0, boot = 1,
      force = out0$force, hasRevs = out0$hasRevs,
      tol = 1e-3, max.iteration = 1000
    ),
    info = "impute_Y0 must error for mc method (Phase 4 not implemented)"
  )
})

## ---- UNIT-4: valid_controls returns non-empty subset for valid dataset -------

test_that("UNIT-4: valid_controls returns non-empty subset for valid dataset", {
  skip_on_cran()
  d <- make_fixture_data()
  out0 <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "ife", r = 1, se = FALSE, CV = FALSE, parallel = FALSE
  )))
  ## valid_controls uses out$D internally as the treatment matrix.
  ## Build a stub with D = out0$D.dat (the TT x N matrix).
  out_stub <- list(
    D    = out0$D.dat,
    I    = out0$I,
    r.cv = out0$r.cv
  )
  vc <- fect:::valid_controls(out_stub, method = "ife", predictive = "notyettreated", force = 0)
  expect_true(is.integer(vc) || is.numeric(vc))
  expect_true(length(vc) > 0)
  id.co <- which(colSums(out0$D.dat) == 0)
  expect_true(all(vc %in% id.co))
})

## ---- DEAD-1: Branch B is absent from boot.R source -------------------------

test_that("DEAD-1: Branch B parametric block is absent from boot.R source", {
  ## Prefer system.file; fall back to source repo path
  boot_path <- system.file("R", "boot.R", package = "fect")
  if (!nzchar(boot_path)) {
    ## Installed package uses compiled .rdb — read source directly
    src_path <- file.path(find.package("fect"), "..", "..", "..", "fect", "R", "boot.R")
    if (!file.exists(src_path)) {
      ## Try the known source repo path
      src_path <- "/Users/xyq/GitHub/statsclaw/.repos/fect/R/boot.R"
    }
    if (!file.exists(src_path)) {
      skip("Could not locate boot.R source")
    }
    boot_path <- src_path
  }
  boot_src <- readLines(boot_path)
  branch_b_lines <- grep(
    'method.*%in%.*c.*"ife".*"mc".*vartype.*==.*"parametric"',
    boot_src
  )
  expect_equal(
    length(branch_b_lines), 0L,
    info = "Branch B gate (ife/mc + parametric) must be absent from boot.R post-refactor"
  )
})

## ---- DEAD-2: package loads without error ------------------------------------

test_that("DEAD-2: R CMD check meta — package loads without error", {
  skip_on_cran()
  expect_true(requireNamespace("fect", quietly = TRUE))
})

fect_mspe <- function(
    out.fect,
    seed        = NULL,
    ## ----- masking strategy ----- ##
    cv.method   = "rolling",         # one of: "rolling" (default), "block", "all_units" (deprecated alias for "block"), "treated_units" (pending deprecation, see ?fect)
    cv.nobs     = 3,
    cv.donut    = 1,
    cv.buffer   = 1,
    cv.prop = 0.1,
    min.T0      = 5,
    k           = 20,
    ## ----- scoring ----- ##
    criterion   = "mspe",         # "mspe","wmspe","gmspe","wgmspe","mad","moment","gmoment"
    W           = NULL,           # TT x N observation weight matrix, or NULL
    norm.para   = NULL,           # normalization vector, or NULL
    proportion  = 0               # proportion cutoff for count.T.cv (same as fect_cv)
) {
    if (!is.null(seed)) set.seed(seed)
    caller_env <- parent.frame()

    ## ---- validate criterion ---- ##
    valid_criteria <- c("mspe", "wmspe", "gmspe", "wgmspe", "mad", "moment", "gmoment")
    criterion <- match.arg(criterion, valid_criteria)

    ## ---- validate cv.method ---- ##
    cv.method <- .fect_normalize_cv_method(
        cv.method,
        allowed = c("rolling", "block", "all_units", "treated_units")
    )
    cv.treat <- (cv.method == "treated_units")
    use_rolling <- (cv.method == "rolling")

    ## ---- helper functions (unchanged) ---- ##
    .build_rerun_args <- function(out_obj, formula_obj, data_obj, index_obj, caller_env) {
        formula_env <- environment(out_obj$call$formula)
        if (is.null(formula_env)) formula_env <- caller_env
        call_args <- as.list(out_obj$call)[-1]
        rerun_args <- list(
            formula = formula_obj,
            data = data_obj,
            index = index_obj
        )
        arg_names <- setdiff(names(call_args), c("", "formula", "data", "index"))
        for (nm in arg_names) {
            rerun_args[[nm]] <- eval(call_args[[nm]], envir = formula_env, enclos = caller_env)
        }
        rerun_args
    }
    .is_fect_output <- function(obj) {
        is.list(obj) && !is.null(obj$call) && !is.null(obj$Y.dat)
    }

    .get_last_matrix <- function(out_obj, nm, TT, N) {
        idx <- which(names(out_obj) == nm)
        if (length(idx) == 0) return(NULL)
        for (kk in rev(idx)) {
            obj <- out_obj[[kk]]
            if (is.matrix(obj) && all(dim(obj) == c(TT, N))) return(obj)
        }
        NULL
    }

    ## ---- validate out.fect ---- ##
    out_list <- if (.is_fect_output(out.fect)) list(out.fect) else out.fect
    if (!is.list(out_list) || length(out_list) == 0 ||
        !all(vapply(out_list, .is_fect_output, logical(1)))) {
        stop("`out.fect` must be a fect output or list of fect outputs.")
    }
    if (is.null(names(out_list)) || any(names(out_list) == "")) {
        names(out_list) <- paste0("out.fect_", seq_along(out_list))
    }

    ref <- out_list[[1]]
    ## Y.dat is always present in fect output; Y.ct.full may not be
    TT <- nrow(ref$Y.dat)
    N <- ncol(ref$Y.dat)
    y_ref <- .get_last_matrix(ref, "Y", TT, N)
    if (is.null(y_ref) && !is.null(ref$Y.dat)) y_ref <- ref$Y.dat
    d_ref <- .get_last_matrix(ref, "D", TT, N)
    if (is.null(d_ref) && !is.null(ref$D.dat)) d_ref <- ref$D.dat
    if (is.null(y_ref) || is.null(d_ref)) stop("`out.fect` must contain matrix Y and D.")

    ## validate W dimensions
    if (!is.null(W)) {
        if (!is.matrix(W) || nrow(W) != TT || ncol(W) != N) {
            stop("`W` must be a TT x N matrix.")
        }
    }

    ## ---- count.T.cv construction ---- ##
    ## Build period-level weights when criterion needs them
    count.T.cv <- NULL
    if (!is.null(ref$T.on)) {
        T_on <- ref$T.on
        count.T.cv <- table(T_on)
        count.T.cv <- count.T.cv[which(as.numeric(names(count.T.cv)) <= 0)]
        cv.prop.cut <- max(count.T.cv) * proportion
        cv.drop.index <- which(count.T.cv <= cv.prop.cut)
        count.T.cv <- count.T.cv / mean(count.T.cv)
        nm <- names(count.T.cv)
        count.T.cv <- c(count.T.cv, median(count.T.cv))
        names(count.T.cv) <- c(nm, "Control")
        count.T.cv[cv.drop.index] <- 0
    }

    ## ---- cv.sample masking setup ---- ##
    rmCV <- estCV <- NULL
    {
        I_mat <- if (!is.null(ref$I.dat)) ref$I.dat else {
            tmp <- .get_last_matrix(ref, "I", TT, N)
            if (is.null(tmp)) matrix(1L, TT, N)
            else tmp
        }
        D_mat <- d_ref
        II_mat <- I_mat
        II_mat[D_mat == 1] <- 0
        if (!is.null(ref$time.component.from) && ref$time.component.from == "nevertreated") {
            D_cum <- apply(D_mat, 2, function(vec) cummax(vec))
            ever_treated <- which(colSums(D_cum) > 0)
            II_mat[, ever_treated] <- 0
        }
        rm_count <- floor(sum(II_mat) * cv.prop)

        ## For nevertreated, only never-treated columns have observations in
        ## II_mat.  The feasibility checks (con1/con2) must count only
        ## rows/columns that carry observations in the *original* II_mat;
        ## otherwise every fold is deemed infeasible and the 200-iteration
        ## fallback strips all held-out positions, producing zero residuals.
        active_rows <- which(rowSums(II_mat) >= 1)
        active_cols <- which(colSums(II_mat) >= 1)
        n_active_rows <- length(active_rows)
        n_active_cols <- length(active_cols)

        rmCV <- list()
        estCV <- list()

        ## ---- rolling-window pre-computation (cv.method = "rolling") ---- ##
        rolling_folds <- NULL
        if (use_rolling) {
            rolling_folds <- .build_cv_mask_rolling(
                II = II_mat, D = D_mat, k = k,
                cv.nobs = cv.nobs, cv.buffer = cv.buffer,
                cv.prop = cv.prop, min.T0 = min.T0, seed = seed
            )
        }

        for (ii in 1:k) {
            if (use_rolling) {
                rmCV[[ii]]  <- rolling_folds[[ii]]$cv.id
                estCV[[ii]] <- rolling_folds[[ii]]$est.id
                next
            }
            cv.n <- 0
            repeat {
                cv.n <- cv.n + 1
                get.cv <- cv.sample(II_mat, D_mat,
                    count = rm_count,
                    cv.count = cv.nobs,
                    cv.treat = cv.treat,
                    cv.donut = cv.donut
                )
                cv.id <- get.cv$cv.id
                II.cv <- II_mat
                II.cv[cv.id] <- 0

                con1 <- sum(rowSums(II.cv[active_rows, , drop = FALSE]) >= 1) == n_active_rows
                con2 <- sum(colSums(II.cv[, active_cols, drop = FALSE]) >= min.T0) == n_active_cols

                if (con1 && con2) break
                if (cv.n >= 200) {
                    keep.1 <- which(apply(II.cv, 1, sum) < 1)
                    keep.2 <- which(apply(II.cv, 2, sum) < min.T0)
                    II.cv[keep.1, ] <- II_mat[keep.1, ]
                    II.cv[, keep.2] <- II_mat[, keep.2]
                    II.cv.valid <- II_mat
                    II.cv.valid[cv.id] <- -1
                    II.cv.valid[keep.1, ] <- II_mat[keep.1, ]
                    II.cv.valid[, keep.2] <- II_mat[, keep.2]
                    cv.id <- which(II.cv.valid != II_mat)
                    break
                }
            }
            rmCV[[ii]] <- cv.id
            if (cv.n < 200) {
                estCV[[ii]] <- get.cv$est.id
            } else {
                cv.id.old <- get.cv$cv.id
                cv.diff <- setdiff(cv.id.old, cv.id)
                estCV[[ii]] <- setdiff(get.cv$est.id, cv.diff)
            }
        }
    }

    ## ---- records data.frame ---- ##
    records <- data.frame(
        Rep = integer(0),
        Model = character(0),
        Hidden_N = integer(0),
        MSPE = numeric(0), WMSPE = numeric(0), GMSPE = numeric(0),
        WGMSPE = numeric(0), MAD = numeric(0),
        Moment = numeric(0), GMoment = numeric(0),
        RMSE = numeric(0), Bias = numeric(0),
        stringsAsFactors = FALSE
    )
    fits_last <- NULL
    scores_last <- NULL

    ## ---- cv.sample evaluation (single pass) ---- ##
    {
        fits <- vector("list", length(out_list))
        names(fits) <- names(out_list)

        for (i in seq_along(out_list)) {
            out_i <- out_list[[i]]
            y_true_i <- .get_last_matrix(out_i, "Y", TT, N)
            if (is.null(y_true_i) && !is.null(out_i$Y.dat)) y_true_i <- out_i$Y.dat
            d_mat_i <- .get_last_matrix(out_i, "D", TT, N)
            if (is.null(d_mat_i) && !is.null(out_i$D.dat)) d_mat_i <- out_i$D.dat
            if (is.null(y_true_i) || is.null(d_mat_i)) {
                stop("Each out.fect must provide Y and D matrices.")
            }

            formula_env_i <- environment(out_i$call$formula)
            if (is.null(formula_env_i)) formula_env_i <- caller_env
            data_i <- eval(out_i$call$data, envir = formula_env_i, enclos = caller_env)
            idx_i <- eval(out_i$call$index, envir = formula_env_i, enclos = caller_env)
            formula_obj_i <- tryCatch(
                eval(out_i$call$formula, envir = formula_env_i, enclos = caller_env),
                error = function(e) NULL
            )
            if (!inherits(formula_obj_i, "formula")) {
                formula_obj_i <- stats::reformulate(c(out_i$D, out_i$X), response = out_i$Y)
            }
            f_vars <- all.vars(formula_obj_i)
            y_col_i <- f_vars[1]

            rr_i <- match(data_i[[idx_i[2]]], out_i$rawtime)
            cc_i <- match(data_i[[idx_i[1]]], out_i$id)

            ## ---- cv.sample: accumulate residuals across k folds ---- ##
                all_resid <- c()
                all_obs_w <- c()
                all_time_idx <- c()

                for (ii in 1:k) {
                    ## hide fold ii positions in long-form data
                    long_rm <- rep(FALSE, nrow(data_i))
                    valid_rm <- !is.na(rr_i) & !is.na(cc_i)
                    mat_idx_rm <- cbind(rr_i[valid_rm], cc_i[valid_rm])
                    in_rm <- rep(FALSE, nrow(data_i))
                    in_rm[valid_rm] <- (mat_idx_rm[, 1] + (mat_idx_rm[, 2] - 1) * TT) %in% rmCV[[ii]]
                    hide_rows_ii <- which(in_rm)

                    if (length(hide_rows_ii) == 0) next

                    data_hidden <- data_i
                    data_hidden[hide_rows_ii, y_col_i] <- NA_real_

                    rerun_args <- .build_rerun_args(
                        out_obj = out_i,
                        formula_obj = formula_obj_i,
                        data_obj = data_hidden,
                        index_obj = idx_i,
                        caller_env = caller_env
                    )
                    ## Fix: always re-fit with fixed r (no CV) and no SE
                    rerun_args$CV <- FALSE
                    rerun_args$se <- FALSE
                    if (!is.null(out_i$r.cv)) {
                        rerun_args$r <- out_i$r.cv
                    }
                    ## Suppress messages from inner refits --- the
                    ## "units dropped" message and similar warnings
                    ## already fired when the user fit the original
                    ## model; emitting them once per model x fold
                    ## (= up to N_models * k times) is noise.
                    out_new <- suppressMessages(do.call(fect, rerun_args))

                    ## collect residuals at estCV positions
                    ## est_pos are flat column-major indices into the
                    ## *original* TT x N matrix (II_mat).  out_new from
                    ## the rerun may have different dimensions (dropped
                    ## rows/cols when all Y are NA), so we map through
                    ## rawtime/id labels instead of using flat indices
                    ## directly on out_new$Y.ct.full.
                    est_pos <- estCV[[ii]]
                    est_row <- ((est_pos - 1) %% TT) + 1
                    est_col <- ((est_pos - 1) %/% TT) + 1

                    ## map original (row, col) to rawtime/id labels,
                    ## then match into the rerun's coordinate system
                    est_time_labels <- out_i$rawtime[est_row]
                    est_id_labels   <- out_i$id[est_col]
                    rr_new <- match(est_time_labels, out_new$rawtime)
                    cc_new <- match(est_id_labels,   out_new$id)
                    valid_map <- !is.na(rr_new) & !is.na(cc_new)

                    pred_vals <- rep(NA_real_, length(est_pos))
                    ## Use Y.ct.full if available, fall back to Y.ct
                    yct_new <- if (!is.null(out_new$Y.ct.full)) out_new$Y.ct.full else out_new$Y.ct
                    pred_vals[valid_map] <- yct_new[cbind(rr_new[valid_map], cc_new[valid_map])]
                    ## y_true_i uses the original coordinate system (always TT x N)
                    actual_vals <- y_true_i[cbind(est_row, est_col)]
                    valid_score <- valid_map & !is.na(pred_vals) & !is.na(actual_vals)

                    resid_ii <- pred_vals[valid_score] - actual_vals[valid_score]
                    all_resid <- c(all_resid, resid_ii)

                    if (!is.null(W)) {
                        all_obs_w <- c(all_obs_w, W[cbind(est_row[valid_score], est_col[valid_score])])
                    }

                    ## time index for moment/gmoment
                    if (!is.null(out_i$T.on)) {
                        tidx <- as.character(out_i$T.on[cbind(est_row[valid_score], est_col[valid_score])])
                        tidx[which(is.na(tidx))] <- "Control"
                        all_time_idx <- c(all_time_idx, tidx)
                    }
                }

                if (length(all_resid) == 0) {
                    stop("No valid residuals collected from cv.sample folds.")
                }
                fits[[i]] <- NULL  ## cv.sample doesn't retain a single refit

                scores <- .score_residuals(
                    resid = all_resid,
                    obs_weights = if (length(all_obs_w) > 0) all_obs_w else NULL,
                    time_index = if (length(all_time_idx) > 0) all_time_idx else NULL,
                    count_weights = count.T.cv,
                    norm.para = norm.para
                )

            scores_last <- scores

            records <- rbind(
                records,
                data.frame(
                    Rep = 1L,
                    Model = names(out_list)[i],
                    Hidden_N = length(all_resid),
                    MSPE = scores["MSPE"],
                    WMSPE = scores["WMSPE"],
                    GMSPE = scores["GMSPE"],
                    WGMSPE = scores["WGMSPE"],
                    MAD = scores["MAD"],
                    Moment = scores["Moment"],
                    GMoment = scores["GMoment"],
                    RMSE = scores["RMSE"],
                    Bias = scores["Bias"],
                    stringsAsFactors = FALSE,
                    row.names = NULL
                )
            )
        }
        fits_last <- fits
    }

    summary <- aggregate(
        cbind(Hidden_N, MSPE, WMSPE, GMSPE, WGMSPE, MAD, Moment, GMoment, RMSE, Bias) ~ Model,
        data = records, FUN = mean
    )
    list(
        summary = summary,
        records = records,
        fits = fits_last,
        criterion = criterion,
        scores = scores_last
    )
}

## ---------------------------------------------------------------
## Tests for fect_nevertreated CV: cv.method dispatch (Section C),
## 1% selection rule verification (Section E), W + count.T.cv weights
## (Section F), end-to-end integration (Section G "Integration"), and
## cv.sample k-fold CV (Section H).
##
## Originally part of test-score-unify.R; split out 2026-05-03 for
## progress visibility. Shared fixtures live in helper-score-unify.R.
## ---------------------------------------------------------------

## =================================================================
## Section C: cv.method in fect_nevertreated (NEW for Phase 2)
## =================================================================

test_that("NT1: fect_nevertreated cv.method='loo' selects r.cv (IFE)", {

  skip_on_cran()
  nt_out <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "loo",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  expect_true(nt_out$r.cv >= 0 && nt_out$r.cv <= 3)
  expect_true(!is.null(nt_out$CV.out))
  mspe_col <- nt_out$CV.out[, "MSPE"]
  computed <- mspe_col[mspe_col < 1e19]
  if (length(computed) > 0) {
    expect_true(all(is.finite(computed)))
  }
  expect_true(!is.null(nt_out$Y.ct))
})

test_that("NT2: fect_nevertreated cv.method='treated_units' selects r.cv (IFE)", {

  skip_on_cran()
  nt_out <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "treated_units",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  expect_true(nt_out$r.cv >= 0 && nt_out$r.cv <= 3)
  expect_true(!is.null(nt_out$CV.out))
})

test_that("NT3: fect_nevertreated cv.method='all_units' selects r.cv (IFE)", {

  skip_on_cran()
  nt_out <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "all_units",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  expect_true(nt_out$r.cv >= 0 && nt_out$r.cv <= 3)
  expect_true(!is.null(nt_out$CV.out))
})

test_that("NT4: fect_nevertreated cv.method='loo' selects r.cv (CFE)", {

  skip_on_cran()
  nt_cfe <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "cfe",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "loo",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  expect_true(nt_cfe$r.cv >= 0 && nt_cfe$r.cv <= 3)
  expect_true(!is.null(nt_cfe$CV.out))
})

test_that("NT5: fect_nevertreated cv.method='treated_units' selects r.cv (CFE)", {

  skip_on_cran()
  nt_cfe <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "cfe",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "treated_units",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  expect_true(nt_cfe$r.cv >= 0 && nt_cfe$r.cv <= 3)
  expect_true(!is.null(nt_cfe$CV.out))
})

test_that("NT6: fect_nevertreated default cv.method is treated_units", {

  skip_on_cran()
  nt_default <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      se              = FALSE,
      parallel        = FALSE
    )
  ))

  nt_explicit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "treated_units",
      se              = FALSE,
      parallel        = FALSE
    )
  ))

  expect_equal(nt_default$r.cv, nt_explicit$r.cv)
})

test_that("NT7: Invalid cv.method rejected for nevertreated", {

  skip_on_cran()
  expect_error(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      cv.method       = "invalid"
    ),
    "cv.method|arg"
  )
})


## =================================================================
## Section E: 1% Selection Rule Verification (NEW for Phase 2)
## =================================================================

test_that("SEL1: 1% selection rule in IFE nevertreated", {

  skip_on_cran()
  # Large tol should NOT affect the 1% rule
  nt_bigtol <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      se              = FALSE,
      parallel        = FALSE,
      tol             = 0.5
    )
  ))
  nt_smalltol <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      se              = FALSE,
      parallel        = FALSE,
      tol             = 1e-3
    )
  ))
  expect_equal(nt_bigtol$r.cv, nt_smalltol$r.cv)
})

test_that("SEL2: 1% selection rule in CFE nevertreated", {

  skip_on_cran()
  nt_bigtol <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "cfe",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      se              = FALSE,
      parallel        = FALSE,
      tol             = 0.5
    )
  ))
  nt_smalltol <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "cfe",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      se              = FALSE,
      parallel        = FALSE,
      tol             = 1e-3
    )
  ))
  expect_equal(nt_bigtol$r.cv, nt_smalltol$r.cv)
})


## =================================================================
## Section F: W and count.T.cv in fect_nevertreated (NEW for Phase 2)
## =================================================================

test_that("WT1: W weights flow through nevertreated LOO scoring", {

  skip_on_cran()
  # W in fect() is a column name, not a matrix. Add a weight column to ntdata.
  ntdata_w <- ntdata
  ntdata_w$wt <- 1.0  # uniform weights

  nt_w <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata_w,
      index           = c("id", "time"),
      method          = "ife",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      W               = "wt",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  expect_true(nt_w$r.cv >= 0 && nt_w$r.cv <= 3)

  # With uniform weights, r.cv should match unweighted result
  nt_nw <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  expect_equal(nt_w$r.cv, nt_nw$r.cv)
})

test_that("WT2: Non-uniform W may change r selection", {

  skip_on_cran()
  # W in fect() is a column name. Add non-uniform weight column.
  ntdata_w <- ntdata
  set.seed(99)
  ntdata_w$wt <- runif(nrow(ntdata), 0.5, 2.0)

  nt_w <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata_w,
      index           = c("id", "time"),
      method          = "ife",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      W               = "wt",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  # Should not crash, r.cv should be valid
  expect_true(nt_w$r.cv >= 0 && nt_w$r.cv <= 3)
})


## =================================================================
## Section G: Integration Tests (NEW for Phase 2)
## =================================================================

test_that("INT1: End-to-end cv.method pipeline", {

  skip_on_cran()
  ## Use CV=FALSE for the fit since fect_mspe performs its own

  ## cross-validation masking; what matters is that the fit object
  ## has valid Y.dat/Y.ct for fect_mspe to score.
  ## NOTE: fect_mspe errors on CV=TRUE fits ("No valid residuals") —
  ## that is a separate source-code issue tracked for builder.
  fit_result <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = ntdata,
      index     = c("id", "time"),
      method    = "ife",
      CV        = FALSE,
      r         = 2,
      se        = FALSE,
      parallel  = FALSE
    )
  ))
  mspe_result <- suppressWarnings(suppressMessages(
    fect_mspe(fit_result, seed = 42, cv.method = "all_units")
  ))
  expect_true("summary" %in% names(mspe_result))
  expect_true(mspe_result$summary$RMSE > 0)
  if ("MSPE" %in% names(mspe_result$summary)) {
    expect_true(mspe_result$summary$MSPE > 0)
  }
})

test_that("INT2: IFE CV respects cv.method='treated_units'", {

  skip_on_cran()
  ## Original test used method="gsynth" + cv.method="loo", but simdata has
  ## treatment reversals which gsynth rejects. We test that cv.method is
  ## respected by using cv.method="treated_units" (non-default) with method="ife".
  ife_out <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      CV        = TRUE,
      r         = c(0, 3),
      cv.method = "treated_units",
      se        = FALSE,
      parallel  = FALSE
    )
  ))
  expect_true(ife_out$r.cv >= 0 && ife_out$r.cv <= 3)
})


## =================================================================
## Section H: cv.sample k-fold CV in fect_nevertreated (NEW)
## Tests for actual cv.sample-based cross-validation branches
## when cv.method="all_units" or "treated_units" in nevertreated.
## =================================================================

## ---- H.1: IFE smoke tests ---- ##

test_that("NTCV1: cv.method='all_units' IFE produces valid output", {

  skip_on_cran()
  set.seed(42)
  out_au <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      force           = "two-way",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "all_units",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  # r.cv is integer in [0, 3]
  expect_true(out_au$r.cv >= 0 && out_au$r.cv <= 3)
  # att.avg is finite
  expect_true(is.finite(out_au$att.avg))
  # est.att has no all-NA columns
  if (!is.null(out_au$est.att)) {
    na_cols <- apply(out_au$est.att, 2, function(x) all(is.na(x)))
    expect_false(all(na_cols),
                 info = "est.att should not have all columns be NA")
  }
  # CV.out exists with proper structure
  expect_true(!is.null(out_au$CV.out))
  mspe_col <- out_au$CV.out[, "MSPE"]
  computed <- mspe_col[mspe_col < 1e19]
  if (length(computed) > 0) {
    expect_true(all(is.finite(computed)))
    expect_true(all(computed > 0))
  }
})

test_that("NTCV2: cv.method='treated_units' IFE produces valid output", {

  skip_on_cran()
  set.seed(42)
  out_tu <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      force           = "two-way",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "treated_units",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  expect_true(out_tu$r.cv >= 0 && out_tu$r.cv <= 3)
  expect_true(is.finite(out_tu$att.avg))
  if (!is.null(out_tu$est.att)) {
    na_cols <- apply(out_tu$est.att, 2, function(x) all(is.na(x)))
    expect_false(all(na_cols))
  }
  expect_true(!is.null(out_tu$CV.out))
})

## ---- H.2: CFE smoke tests ---- ##

test_that("NTCV3: cv.method='all_units' CFE produces valid output", {

  skip_on_cran()
  set.seed(42)
  out_au_cfe <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "cfe",
      force           = "two-way",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "all_units",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  expect_true(out_au_cfe$r.cv >= 0 && out_au_cfe$r.cv <= 3)
  expect_true(is.finite(out_au_cfe$att.avg))
  expect_true(!is.null(out_au_cfe$CV.out))
})

test_that("NTCV4: cv.method='treated_units' CFE produces valid output", {

  skip_on_cran()
  set.seed(42)
  out_tu_cfe <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "cfe",
      force           = "two-way",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "treated_units",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  expect_true(out_tu_cfe$r.cv >= 0 && out_tu_cfe$r.cv <= 3)
  expect_true(is.finite(out_tu_cfe$att.avg))
  expect_true(!is.null(out_tu_cfe$CV.out))
})

## ---- H.3: LOO backward compatibility ---- ##

test_that("NTCV5: cv.method='loo' IFE backward compatibility", {

  skip_on_cran()
  set.seed(1234)
  out_loo <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      force           = "two-way",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "loo",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  expect_true(out_loo$r.cv >= 0 && out_loo$r.cv <= 3)
  expect_true(is.finite(out_loo$att.avg))
  expect_true(!is.null(out_loo$CV.out))

  # LOO is deterministic: re-running should produce identical r.cv
  set.seed(1234)
  out_loo2 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      force           = "two-way",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "loo",
      se              = FALSE,
      parallel        = FALSE
    )
  ))
  expect_equal(out_loo$r.cv, out_loo2$r.cv)
})

## ---- H.4: r-selection and ATT checks ---- ##

test_that("NTCV6: r-selection validity across all cv.methods (IFE)", {

  skip_on_cran()
  r_start <- 0
  r_end <- 3

  methods_list <- c("loo", "all_units", "treated_units")
  for (cm in methods_list) {
    set.seed(42)
    out <- suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data            = ntdata,
        index           = c("id", "time"),
        method          = "ife",
        force           = "two-way",
        time.component.from    = "nevertreated",
        CV              = TRUE,
        r               = c(r_start, r_end),
        cv.method       = cm,
        se              = FALSE,
        parallel        = FALSE
      )
    ))

    # r.cv in valid range
    expect_true(out$r.cv >= r_start && out$r.cv <= r_end,
                info = paste("cv.method =", cm, ": r.cv out of range"))

    # CV.out has correct number of rows
    expect_equal(nrow(out$CV.out), r_end - r_start + 1,
                 info = paste("cv.method =", cm, ": CV.out row count wrong"))

    # Score columns in CV.out: check MSPE for evaluated rows
    mspe_col <- out$CV.out[, "MSPE"]
    computed <- mspe_col[mspe_col < 1e19]
    if (length(computed) > 0) {
      expect_true(all(is.finite(computed)),
                  info = paste("cv.method =", cm, ": non-finite MSPE"))
      expect_true(all(computed >= 0),
                  info = paste("cv.method =", cm, ": negative MSPE"))
    }
  }
})

test_that("NTCV7: ATT consistency across cv.methods (IFE)", {

  skip_on_cran()
  att_vals <- numeric(3)
  methods_list <- c("loo", "all_units", "treated_units")
  for (i in seq_along(methods_list)) {
    set.seed(42)
    out <- suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data            = ntdata,
        index           = c("id", "time"),
        method          = "ife",
        force           = "two-way",
        time.component.from    = "nevertreated",
        CV              = TRUE,
        r               = c(0, 3),
        cv.method       = methods_list[i],
        se              = FALSE,
        parallel        = FALSE
      )
    ))
    att_vals[i] <- out$att.avg
    expect_true(is.finite(out$att.avg),
                info = paste("cv.method =", methods_list[i], ": att.avg not finite"))
  }

  # Sanity check: all ATT values should be in the same ballpark.
  # The true ATT is ~3.0 for this DGP.
  # Allow wide tolerance since different r.cv selections produce different ATTs.
  att_range <- max(att_vals) - min(att_vals)
  expect_true(att_range < 5.0,
              info = paste("ATT values too spread:",
                           paste(round(att_vals, 3), collapse = ", ")))
})

## ---- H.5: Edge cases ---- ##

test_that("NTCV-Edge1: r=c(0,0) with all cv.methods", {

  skip_on_cran()
  for (cm in c("loo", "all_units", "treated_units")) {
    set.seed(42)
    out <- suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data            = ntdata,
        index           = c("id", "time"),
        method          = "ife",
        force           = "two-way",
        time.component.from    = "nevertreated",
        CV              = TRUE,
        r               = c(0, 0),
        cv.method       = cm,
        se              = FALSE,
        parallel        = FALSE
      )
    ))
    expect_equal(out$r.cv, 0,
                 info = paste("cv.method =", cm, ": r.cv should be 0 when r=c(0,0)"))
    # Scores should still be finite
    mspe_col <- out$CV.out[, "MSPE"]
    computed <- mspe_col[mspe_col < 1e19]
    if (length(computed) > 0) {
      expect_true(all(is.finite(computed)),
                  info = paste("cv.method =", cm, ": non-finite MSPE with r=0"))
    }
  }
})

test_that("NTCV-Edge2: Small panel with all cv.methods", {

  skip_on_cran()
  # Create data with few pre-treatment periods for treated units
  small_data <- make_factor_data(N = 30, TT = 10, Ntr = 8, r = 1, seed = 99)

  for (cm in c("loo", "all_units", "treated_units")) {
    set.seed(42)
    out <- tryCatch(
      suppressWarnings(suppressMessages(
        fect::fect(
          Y ~ D,
          data            = small_data,
          index           = c("id", "time"),
          method          = "ife",
          force           = "two-way",
          time.component.from    = "nevertreated",
          CV              = TRUE,
          r               = c(0, 2),
          cv.method       = cm,
          se              = FALSE,
          parallel        = FALSE
        )
      )),
      error = function(e) e
    )
    # Should either succeed with valid r.cv or produce informative error
    if (!inherits(out, "error")) {
      expect_true(out$r.cv >= 0 && out$r.cv <= 2,
                  info = paste("cv.method =", cm, ": r.cv out of range (small panel)"))
      expect_true(is.finite(out$att.avg),
                  info = paste("cv.method =", cm, ": att.avg not finite (small panel)"))
    } else {
      # If it errors, the message should be informative (not a cryptic crash)
      expect_true(nchar(conditionMessage(out)) > 0,
                  info = paste("cv.method =", cm, ": error should be informative"))
    }
  }
})

test_that("NTCV-Edge3: Single treated unit", {

  skip_on_cran()
  single_tr_data <- make_factor_data(N = 30, TT = 15, Ntr = 1, r = 1, seed = 77)

  for (cm in c("loo", "all_units", "treated_units")) {
    set.seed(42)
    out <- tryCatch(
      suppressWarnings(suppressMessages(
        fect::fect(
          Y ~ D,
          data            = single_tr_data,
          index           = c("id", "time"),
          method          = "ife",
          force           = "two-way",
          time.component.from    = "nevertreated",
          CV              = TRUE,
          r               = c(0, 2),
          cv.method       = cm,
          se              = FALSE,
          parallel        = FALSE
        )
      )),
      error = function(e) e
    )
    if (!inherits(out, "error")) {
      expect_true(out$r.cv >= 0 && out$r.cv <= 2,
                  info = paste("cv.method =", cm, ": r.cv invalid (single treated)"))
    }
    # If it errors, that's acceptable for single treated unit edge case
  }
})

## ---- H.6: Property-based invariants for cv.sample ---- ##

test_that("NTCV-P1: Score non-negativity in CV.out", {

  skip_on_cran()
  for (cm in c("all_units", "treated_units")) {
    set.seed(42)
    out <- suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data            = ntdata,
        index           = c("id", "time"),
        method          = "ife",
        force           = "two-way",
        time.component.from    = "nevertreated",
        CV              = TRUE,
        r               = c(0, 3),
        cv.method       = cm,
        se              = FALSE,
        parallel        = FALSE
      )
    ))

    # All score columns should be non-negative
    score_cols <- c("MSPE", "WMSPE", "GMSPE", "WGMSPE", "MAD")
    for (sc in score_cols) {
      if (sc %in% colnames(out$CV.out)) {
        vals <- out$CV.out[, sc]
        computed <- vals[vals < 1e19]
        if (length(computed) > 0) {
          expect_true(all(computed >= 0),
                      info = paste("cv.method =", cm, ", score =", sc, ": negative value"))
        }
      }
    }
  }
})

test_that("NTCV-P4: LOO determinism", {

  skip_on_cran()
  set.seed(42)
  out1 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      force           = "two-way",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "loo",
      se              = FALSE,
      parallel        = FALSE
    )
  ))

  set.seed(999)  # different seed should not affect LOO
  out2 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data            = ntdata,
      index           = c("id", "time"),
      method          = "ife",
      force           = "two-way",
      time.component.from    = "nevertreated",
      CV              = TRUE,
      r               = c(0, 3),
      cv.method       = "loo",
      se              = FALSE,
      parallel        = FALSE
    )
  ))

  expect_equal(out1$r.cv, out2$r.cv)
  # CV.out scores should be identical
  expect_equal(out1$CV.out, out2$CV.out, tolerance = 1e-10)
})

test_that("NTCV-P5: cv.sample reproducibility with set.seed", {

  skip_on_cran()
  for (cm in c("all_units", "treated_units")) {
    set.seed(42)
    out1 <- suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data            = ntdata,
        index           = c("id", "time"),
        method          = "ife",
        force           = "two-way",
        time.component.from    = "nevertreated",
        CV              = TRUE,
        r               = c(0, 3),
        cv.method       = cm,
        se              = FALSE,
        parallel        = FALSE
      )
    ))

    set.seed(42)
    out2 <- suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data            = ntdata,
        index           = c("id", "time"),
        method          = "ife",
        force           = "two-way",
        time.component.from    = "nevertreated",
        CV              = TRUE,
        r               = c(0, 3),
        cv.method       = cm,
        se              = FALSE,
        parallel        = FALSE
      )
    ))

    expect_equal(out1$r.cv, out2$r.cv,
                 info = paste("cv.method =", cm, ": r.cv not reproducible"))
    expect_equal(out1$CV.out, out2$CV.out, tolerance = 1e-10,
                 info = paste("cv.method =", cm, ": CV.out not reproducible"))
  }
})

## ---------------------------------------------------------------
## Integration tests: cv.method = "rolling" wired into the main
## fect() CV dispatcher (additive change; block CV defaults
## untouched).
##
## Coverage:
##   1. fect(method = "ife", CV = TRUE, cv.method = "rolling", ...)
##   2. fect(method = "gsynth", CV = TRUE, cv.method = "rolling", ...)
##      (routes through fect_nevertreated)
##   3. fect(method = "cfe", CV = TRUE, cv.method = "rolling", ...)
##      with extra-FE column in index
##   4. fect_mspe(out, cv.method = "rolling", cv.buffer = 1, ...)
##   5. Regression: existing block CV defaults still work and produce
##      a finite r.cv.
## ---------------------------------------------------------------

.make_rolling_panel <- function(seed = 17) {
  set.seed(seed)
  N  <- 30L; T <- 25L
  ids   <- rep(seq_len(N), each = T)
  times <- rep(seq_len(T), N)
  F1 <- rnorm(T); L1 <- rnorm(N)
  err <- rnorm(N * T, sd = 0.3)
  mu  <- rep(L1, each = T) * rep(F1, N)
  Y   <- mu + err
  treated      <- 26L:30L
  treat_starts <- c(18L, 19L, 20L, 21L, 22L)
  D <- rep(0L, N * T)
  for (k in seq_along(treated)) {
    u <- treated[k]
    D[ids == u & times >= treat_starts[k]] <- 1L
  }
  Y <- Y + D * 1.5
  df <- data.frame(id = ids, time = times, Y = Y, D = D)
  ## Make panel slightly unbalanced (so masking some control units'
  ## tail observations doesn't drop entire time points).
  drop_keys <- c(
    paste(27L, c(24L, 25L), sep = "_"),
    paste(28L, c(22L, 23L, 24L, 25L), sep = "_"),
    paste(1L,  c(23L, 24L, 25L), sep = "_"),
    paste(2L,  c(24L, 25L), sep = "_")
  )
  df_keys <- paste(df$id, df$time, sep = "_")
  df[!df_keys %in% drop_keys, ]
}

test_that("fect dispatcher: cv.method = 'rolling' works for method = 'ife'", {

  skip_on_cran()

  df <- .make_rolling_panel(seed = 17)

  set.seed(101)
  out <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", force = "two-way",
      CV = TRUE, cv.method = "rolling",
      cv.buffer = 1, cv.nobs = 2, k = 2, cv.prop = 0.2,
      r = c(0, 2), min.T0 = 5,
      se = FALSE
    )
  ))

  expect_true(is.list(out))
  expect_true(!is.null(out$r.cv))
  expect_true(is.finite(as.numeric(out$r.cv)))
})

test_that("fect dispatcher: cv.method = 'rolling' works for method = 'gsynth'", {

  skip_on_cran()

  df <- .make_rolling_panel(seed = 19)

  set.seed(202)
  out <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "gsynth", force = "two-way",
      CV = TRUE, cv.method = "rolling",
      cv.buffer = 1, cv.nobs = 2, k = 2, cv.prop = 0.2,
      r = c(0, 2), min.T0 = 5,
      se = FALSE
    )
  ))

  expect_true(is.list(out))
  expect_true(!is.null(out$r.cv))
  expect_true(is.finite(as.numeric(out$r.cv)))
})

test_that("fect dispatcher: cv.method = 'rolling' works for method = 'cfe' with extra-FE", {

  skip_on_cran()

  if (!exists("sim_region", inherits = TRUE)) {
    suppressWarnings(try(utils::data("sim_region", package = "fect"),
                         silent = TRUE))
  }
  skip_if_not(exists("sim_region", inherits = TRUE),
              "sim_region dataset not available")

  set.seed(303)
  out <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = sim_region,
      index = c("id", "time", "region_time"),
      method = "cfe", force = "two-way",
      CV = TRUE, cv.method = "rolling",
      cv.buffer = 1, cv.nobs = 2, k = 2, cv.prop = 0.2,
      r = c(0, 1), min.T0 = 5,
      se = FALSE
    )
  ))

  expect_true(is.list(out))
  expect_true(!is.null(out$r.cv))
  expect_true(is.finite(as.numeric(out$r.cv)))
})

test_that("fect_mspe: cv.method = 'rolling' produces a finite MSPE", {

  skip_on_cran()

  df <- .make_rolling_panel(seed = 23)

  set.seed(404)
  ## A single fect output is the input.
  out <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", force = "two-way",
      CV = FALSE, r = 1, se = FALSE
    )
  ))

  res <- suppressWarnings(suppressMessages(
    fect::fect_mspe(
      out, cv.method = "rolling",
      cv.buffer = 1, cv.nobs = 2, k = 2, cv.prop = 0.2,
      min.T0 = 5, seed = 7
    )
  ))

  expect_true(is.list(res))
  expect_true(is.data.frame(res$summary))
  expect_true(any(is.finite(res$summary$MSPE)))
})

test_that("fect dispatcher: block-CV defaults still work (regression)", {

  skip_on_cran()

  df <- .make_rolling_panel(seed = 31)

  set.seed(505)
  ## Use the existing block CV path with all defaults.
  out_block <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", force = "two-way",
      CV = TRUE, cv.method = "all_units",
      cv.nobs = 2, k = 2, cv.prop = 0.2,
      r = c(0, 2), min.T0 = 5,
      se = FALSE
    )
  ))

  expect_true(is.list(out_block))
  expect_true(!is.null(out_block$r.cv))
  expect_true(is.finite(as.numeric(out_block$r.cv)))
})

## make-paraboot-baseline.R
## Pre-refactor baseline fixture capture helper (test-spec.md §1)
##
## Run this script from the fect project root BEFORE applying the refactor:
##   Rscript tests/testthat/make-paraboot-baseline.R
##
## Saves: tests/testthat/fixtures/paraboot-baseline.rds
##
## Seed convention: set.seed(42) for dataset; set.seed(2026) before each fect() call.
## parallel=FALSE is MANDATORY for reproducibility.

library(fect)

make_fixture_data <- function(seed = 42) {
  set.seed(seed)
  N   <- 50
  TT  <- 20
  T0  <- 12
  Ntr <- 10
  id   <- rep(1:N, each = TT)
  time <- rep(1:TT, N)
  D    <- as.integer(id <= Ntr & time >= T0)
  alpha_i <- rep(rnorm(N, 0, 1), each = TT)
  xi_t    <- rep(rnorm(TT, 0, 0.5), N)
  Y <- alpha_i + xi_t + 2 * D + rnorm(N * TT, 0, 0.5)
  data.frame(id = id, time = time, Y = Y, D = D)
}

d <- make_fixture_data(seed = 42)

## Fixture calls — test-spec.md §1.2
set.seed(2026)
gsynth_para_result <- suppressWarnings(suppressMessages(fect(
  Y ~ D, data=d, index=c("id","time"),
  method="gsynth", r=1, se=TRUE, vartype="parametric", nboots=50, CV=FALSE, parallel=FALSE
)))

set.seed(2026)
ife_nt_para_result <- suppressWarnings(suppressMessages(fect(
  Y ~ D, data=d, index=c("id","time"),
  method="ife", r=1, se=TRUE, vartype="parametric", nboots=50, CV=FALSE, parallel=FALSE
)))

set.seed(2026)
ife_nev_para_result <- suppressWarnings(suppressMessages(fect(
  Y ~ D, data=d, index=c("id","time"),
  method="ife", time.component.from="nevertreated", r=1, se=TRUE,
  vartype="parametric", nboots=50, CV=FALSE, parallel=FALSE
)))

set.seed(2026)
cfe_nev_para_result <- suppressWarnings(suppressMessages(fect(
  Y ~ D, data=d, index=c("id","time"),
  method="cfe", time.component.from="nevertreated", r=1, se=TRUE,
  vartype="parametric", nboots=50, CV=FALSE, parallel=FALSE
)))

set.seed(2026)
cfe_nt_para_result <- suppressWarnings(suppressMessages(fect(
  Y ~ D, data=d, index=c("id","time"),
  method="cfe", r=0, se=TRUE, vartype="parametric", nboots=50, CV=FALSE, parallel=FALSE
)))

set.seed(2026)
ife_nt_boot_result <- suppressWarnings(suppressMessages(fect(
  Y ~ D, data=d, index=c("id","time"),
  method="ife", r=1, se=TRUE, vartype="bootstrap", nboots=50, CV=FALSE, parallel=FALSE
)))

set.seed(2026)
ife_nev_boot_result <- suppressWarnings(suppressMessages(fect(
  Y ~ D, data=d, index=c("id","time"),
  method="ife", time.component.from="nevertreated", r=1, se=TRUE,
  vartype="bootstrap", nboots=50, CV=FALSE, parallel=FALSE
)))

set.seed(2026)
gsynth_boot_result <- suppressWarnings(suppressMessages(fect(
  Y ~ D, data=d, index=c("id","time"),
  method="gsynth", r=1, se=TRUE, vartype="bootstrap", nboots=50, CV=FALSE, parallel=FALSE
)))

set.seed(2026)
ife_jk_result <- suppressWarnings(suppressMessages(fect(
  Y ~ D, data=d, index=c("id","time"),
  method="ife", r=1, se=TRUE, vartype="jackknife", CV=FALSE, parallel=FALSE
)))

set.seed(2026)
gsynth_jk_result <- suppressWarnings(suppressMessages(fect(
  Y ~ D, data=d, index=c("id","time"),
  method="gsynth", r=1, se=TRUE, vartype="jackknife", CV=FALSE, parallel=FALSE
)))

## Save
baseline <- list(
  gsynth_para  = gsynth_para_result,
  ife_nt_para  = ife_nt_para_result,
  ife_nev_para = ife_nev_para_result,
  cfe_nev_para = cfe_nev_para_result,
  cfe_nt_para  = cfe_nt_para_result,
  ife_nt_boot  = ife_nt_boot_result,
  ife_nev_boot = ife_nev_boot_result,
  gsynth_boot  = gsynth_boot_result,
  ife_jk       = ife_jk_result,
  gsynth_jk    = gsynth_jk_result
)

dir.create("tests/testthat/fixtures", showWarnings = FALSE, recursive = TRUE)
saveRDS(baseline, "tests/testthat/fixtures/paraboot-baseline.rds")
message("Fixture saved to tests/testthat/fixtures/paraboot-baseline.rds")

## test-parametric-notyet-gate.R
## Gate tests for parametric + notyettreated hard gate
## StatsClaw run: 2026-04-14-notyet-parametric-gate
##
## Positive-fire tests (NGP-1 through NGP-4): gate fires when it should.
## Negative-fire tests (NGN-1 through NGN-3): gate does NOT fire for valid paths.

## ---- helpers ----------------------------------------------------------------

make_gate_fixture <- function(seed = 42) {
  ## Staggered adoption panel with 50 units (10 treated, 40 never-treated)
  ## Uses same structure as make_fixture_data() in test-paraboot-parity.R
  set.seed(seed)
  N <- 50; TT <- 20; T0 <- 12; Ntr <- 10
  id   <- rep(1:N, each = TT)
  time <- rep(1:TT, N)
  D    <- as.integer(id <= Ntr & time >= T0)
  alpha_i <- rep(rnorm(N, 0, 1), each = TT)
  xi_t    <- rep(rnorm(TT, 0, 0.5), N)
  Y <- alpha_i + xi_t + 2 * D + rnorm(N * TT, 0, 0.5)
  data.frame(id = id, time = time, Y = Y, D = D)
}

## Verify fixture has never-treated units (required for nevertreated tests)
.check_nev <- function(d) {
  D_mat <- tapply(d$D, list(d$time, d$id), identity)
  id_co <- which(colSums(D_mat) == 0)
  stopifnot("Fixture must have never-treated units" = length(id_co) > 0)
  invisible(TRUE)
}

## ============================================================
## POSITIVE-FIRE TESTS: gate fires for invalid combinations
## ============================================================

test_that("NGP-1: ife + vartype='parametric' + default factors.from errors", {
  skip_on_cran()
  d <- make_gate_fixture()
  expect_error(
    fect(
      Y ~ D, data = d, index = c("id", "time"),
      method = "ife", r = 1, se = TRUE,
      vartype = "parametric", nboots = 10,
      CV = FALSE, parallel = FALSE
    ),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

test_that("NGP-2: cfe + vartype='parametric' + default factors.from errors", {
  skip_on_cran()
  d <- make_gate_fixture()
  expect_error(
    fect(
      Y ~ D, data = d, index = c("id", "time"),
      method = "cfe", r = 1, se = TRUE,
      vartype = "parametric", nboots = 10,
      CV = FALSE, parallel = FALSE
    ),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

test_that("NGP-3: ife + vartype='parametric' + explicit factors.from='notyettreated' errors", {
  skip_on_cran()
  d <- make_gate_fixture()
  expect_error(
    fect(
      Y ~ D, data = d, index = c("id", "time"),
      method = "ife", time.component.from = "notyettreated",
      r = 1, se = TRUE, vartype = "parametric", nboots = 10,
      CV = FALSE, parallel = FALSE
    ),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

test_that("NGP-4: error message mentions both an alternative (nevertreated or bootstrap or jackknife)", {
  skip_on_cran()
  d <- make_gate_fixture()
  expect_error(
    fect(
      Y ~ D, data = d, index = c("id", "time"),
      method = "ife", r = 1, se = TRUE,
      vartype = "parametric", nboots = 10,
      CV = FALSE, parallel = FALSE
    ),
    regexp = "nevertreated|bootstrap|jackknife",
    ignore.case = TRUE
  )
})

## ============================================================
## NEGATIVE-FIRE TESTS: gate does NOT fire for valid paths
## ============================================================

test_that("NGN-1: gsynth + parametric does NOT error", {
  skip_on_cran()
  d <- make_gate_fixture()
  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "gsynth", r = 1, se = TRUE,
    vartype = "parametric", nboots = 10,
    CV = FALSE, parallel = FALSE
  )))
  expect_s3_class(out, "fect")
})

test_that("NGN-2: ife + nevertreated + parametric does NOT error", {
  skip_on_cran()
  d <- make_gate_fixture()
  .check_nev(d)
  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "ife", time.component.from = "nevertreated",
    r = 1, se = TRUE, vartype = "parametric", nboots = 10,
    CV = FALSE, parallel = FALSE
  )))
  expect_s3_class(out, "fect")
})

test_that("NGN-3: cfe + nevertreated + parametric does NOT error", {
  skip_on_cran()
  d <- make_gate_fixture()
  .check_nev(d)
  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = d, index = c("id", "time"),
    method = "cfe", time.component.from = "nevertreated",
    r = 1, se = TRUE, vartype = "parametric", nboots = 10,
    CV = FALSE, parallel = FALSE
  )))
  expect_s3_class(out, "fect")
})

###################################################################
## Cross-validation
###################################################################
fect_binary_cv <- function(Y, # Outcome variable, (T*N) matrix
                           X, # Explanatory variables:  (T*N*p) array
                           D, #  Indicator for treated unit (tr==1) 
                           I,
                           II, 
                           T.on, 
                           T.off = NULL, 
                           k = 20, # CV time
                           cv.prop = 0.1,
                           cv.method = "rolling",
                           cv.nobs = 3,
                           cv.buffer = 1,
                           min.T0 = 5,
                           r = 0, # initial number of factors considered if CV==1
                           r.end,
                           QR = FALSE, 
                           force, 
                           hasRevs = 1,
                           tol, # tolerance level
                           group.level = NULL,
                           group = NULL
                           ) {  
    
    ##-------------------------------##
    ## Parsing data
    ##-------------------------------##

    ## ---- cv.method → cv.treat mapping ---- ##
    cv.method <- .fect_normalize_cv_method(
        cv.method,
        allowed = c("rolling", "block", "all_units", "treated_units")
    )
    cv.treat <- (cv.method == "treated_units")
    use_rolling <- (cv.method == "rolling")

    ## unit id and time
    TT <- dim(Y)[1]
    N <- dim(Y)[2]
    if (is.null(X) == FALSE) {
        p <- dim(X)[3]
    } else {
        p <- 0 
        X <- array(0, dim = c(1, 1, 0))
    }

    ## replicate data
    YY <- Y
    YY[which(II == 0)] <- 0 ## reset to 0 
    t.on <- c(T.on)
    T0.min <- min(apply(II, 2, sum))

    ## ----------------- candidate r ------------ ##
    r.max <- min(TT, r.end)
    r.cv <- 0 ## initial value

    ##  --------- initial fit using fastplm --------- ##
    data.ini <- matrix(NA, (TT*N), (2 + 1 + p))
    data.ini[, 2] <- rep(1:N, each = TT)         ## unit fe
    data.ini[, 3] <- rep(1:TT, N)                ## time fe
    data.ini[, 1] <- c(Y)                       ## outcome
    if (p > 0) {                                ## covar
        for (i in 1:p) {
            data.ini[, (3 + i)] <- c(X[, , i])
        }
    }
    ## observed Y0 indicator:
    initialOut <- Y0 <- beta0 <- FE0 <- xi0 <- factor0 <- NULL
    oci <- which(c(II) == 1)


    if (r.max == 0) {
        r.cv <- 0
        message("Cross validation cannot be performed since available pre-treatment records of treated units are too few. So set r.cv = 0.\n ")
        initialOut <- BiInitialFit(data = data.ini, QR = QR, r = 0, force = force, oci = oci)
        Y0 <- initialOut$Y0
        FE0 <- initialOut$FE0
        if (QR == 1) {
            xi0 <- initialOut$xi0
            factor0 <- initialOut$factor0
        }
        if (QR == FALSE) {
            est.best <- inter_fe_d_ub(YY, Y0, FE0, X, II, r.cv, force, tol = tol)
        } else {
            est.best <- inter_fe_d_qr_ub(YY, Y0, FE0, factor0, xi0, X, II, r.cv, force, tol = tol)
        }
    }
    

    ## ------------- restrictions on candidate hyper parameters ---------- ##
    obs.con <- (sum(II) - r.end * (N + TT) + r.end^2 - p) <= 0
    if (obs.con) {
        while((sum(II) - r.end * (N + TT) + r.end^2 - p) <= 0) {
            r.end <- r.end - 1
        }
    }
    if (r.end >= T0.min) {
        message("Facotr number should not be greater than ", T0.min - 1, "\n", sep = "")
        r.end <- T0.min-1
    } else {
        if (obs.con) {
            message("Facotr number should not be greater than ", r.end, "\n", sep = "")
        }
    }

        ##-------------------------------##
    ## ----------- Main Algorithm ----------- ##
        ##-------------------------------##
    
    validX <- 1 ## no multi-colinearity 
    
    ##----------------------------------------------------##
    ##                 Cross-validation of r              ##
    ##----------------------------------------------------##
    
    

    if (r.max > 0) {

        r.old <- r ## save the minimal number of factors 
        message("Cross-validating ...","\n") 

                         ## ----- ##
        ## ------------- initialize ------------ ##
                         ## ----- ##
        
        cv.pos <- which(t.on<=0)
        t.on.cv <- t.on[cv.pos]
        count.on.cv <- as.numeric(table(t.on.cv))
        ## tot.id <- which(c(II)==1) ## observed control data
        ## cv.count <- ceiling((sum(II)*sum(II))/sum(I))
        rm.count <- floor(sum(II)*cv.prop)
        cv.count <- sum(II) - rm.count

        #ociCV <- matrix(NA, cv.count, k) ## store indicator
        #rmCV <- matrix(NA, (length(oci) - cv.count), k) ## removed indicator
        ociCV <- list()
        rmCV <- list()
        Y0CV <- array(NA, dim = c(TT, N, k)) ## store initial Y0
        if (p > 0) {
           beta0CV <- array(NA, dim = c(p, 1, k)) 
        } else {
            beta0CV <- array(0, dim = c(1, 0, k)) ## store initial beta0
        }
        
        ## ---- rolling-window pre-computation (cv.method = "rolling") ---- ##
        rolling_folds <- NULL
        if (use_rolling) {
            rolling_folds <- .build_cv_mask_rolling(
                II = II, D = D, k = k,
                cv.nobs = cv.nobs, cv.buffer = cv.buffer,
                cv.prop = cv.prop, min.T0 = min.T0, seed = NULL
            )
        }

        for (i in 1:k) {
            if (use_rolling) {
                cv.id <- rolling_folds[[i]]$cv.id
            } else {
                cv.n <- 0
                repeat{
                    cv.n <- cv.n + 1
                    cv.id <- cv.sample(II, D, rm.count, cv.nobs, cv.treat)
                    ## cv.id <- cv.sample(II, as.integer(sum(II) - cv.count))
                    ## cv.id <- sample(oci, as.integer(sum(II) - cv.count), replace = FALSE)
                    #II.cv <- II
                    #II.cv[cv.id] <- 0
                    II.cv.valid <- II.cv <- II
                    II.cv[cv.id] <- 0
                    II.cv.valid[cv.id] <- -1
                    con1 <- sum(apply(II.cv, 1, sum) >= 1) == TT
                    con2 <- sum(apply(II.cv, 2, sum) >= 1) == N
                    if (con1 & con2) {
                        break
                    }
                    if (cv.n>=100) {
                        message("Some units have too few pre-treatment observations. Remove them automatically.")
                        keep.1 <- which(apply(II.cv, 1, sum) < 1)
                        keep.2 <- which(apply(II.cv, 2, sum) < 1)
                        II.cv[keep.1,] <- II[keep.1,]
                        II.cv[,keep.2] <- II[,keep.2]
                        II.cv.valid[keep.1,] <- II[keep.1,]
                        II.cv.valid[,keep.2] <- II[,keep.2]
                        cv.id <- which(II.cv.valid!=II)
                        break
                    }
                }
            }
            rmCV[[i]] <- cv.id
            ociCV[[i]] <- setdiff(oci, cv.id)

        }
    
        ##  --------------------------------------------- ##
##  ---------------- cross validation for ife model ------------------  ##
        ##  --------------------------------------------- ##
        
            
        message("Probit model with interactive fixed effects ...\n")
        
        MSPE.best <- NULL
        CV.out <- matrix(NA, (r.max - r.old + 1), 4)
        colnames(CV.out) <- c("r", "IC", "Log-likelihood", "MSPE")
        CV.out[,"r"] <- c(r.old:r.max)
        CV.out[,"MSPE"] <- 1e20

        CVinitialOut <- fit.cv <- Y0.cv <- FE0.cv <- xi0.cv <- factor0.cv <- NULL
                    
        for (i in 1:dim(CV.out)[1]) { ## cross-validation loop starts 

            ## inter FE based on control, before & after 
            r <- CV.out[i, "r"]  
            ## k <- 5
            SSE <- 0
            for (ii in 1:k) {
                II.cv <- II
                II.cv[rmCV[[ii]]] <- 0
                YY.cv <- YY
                YY.cv[rmCV[[ii]]] <- 0 

                CVinitialOut <- BiInitialFit(data = data.ini, QR = QR, r = r, force = force, oci = ociCV[[ii]])
                Y0.cv <- CVinitialOut$Y0
                FE0.cv <- CVinitialOut$FE0
                if (QR == 1) {
                    xi0.cv <- CVinitialOut$xi0
                    factor0.cv <- CVinitialOut$factor0
                }
                if (QR == FALSE) {
                    est.cv.fit <- inter_fe_d_ub(YY.cv, Y0.cv, FE0.cv, X, II.cv, r = r, force, tol = tol)
                } else {
                    est.cv.fit <- inter_fe_d_qr_ub(YY.cv, Y0.cv, FE0.cv, factor0.cv, xi0.cv, X, II.cv, r = r, force, tol = tol)
                }
                fit.cv <- ifelse(est.cv.fit$fit >= 0, 1, 0)

                SSE <- SSE + sum((YY[rmCV[[ii]]]-fit.cv[rmCV[[ii]]])^2)
            }
            MSPE <- SSE/(k*(sum(II) - cv.count))

            ## over-all 
            initialOut <- BiInitialFit(data = data.ini, QR = QR, r = r, force = force, oci = oci)
            Y0 <- initialOut$Y0
            FE0 <- initialOut$FE0
            if (QR == 1) {
                xi0 <- initialOut$xi0
                factor0 <- initialOut$factor0
            }
            if (QR == FALSE) {
                est.cv <- inter_fe_d_ub(YY, Y0, FE0, X, II, r, force, tol = tol)
            } else {
                est.cv <- inter_fe_d_qr_ub(YY, Y0, FE0, factor0, xi0, X, II, r, force, tol = tol)
            }

            IC <- est.cv$IC
            Loglikelihood <- est.cv$loglikelihood

            if (min(CV.out[,"MSPE"]) > MSPE) {
                ## at least 10% improvement for MPSE
                MSPE.best <- MSPE
                est.best <- est.cv  
                r.cv <- r
            } else {
                if (r == r.cv + 1) message("*")
            }

            CV.out[i, 2:4] <- c(IC, Loglikelihood, MSPE)

            message("\n r = ",r, "; IC = ",
                sprintf("%.5f",IC), "; Log-likelihood = ",
                sprintf("%.5f",Loglikelihood), "; MSPE = ",
                sprintf("%.5f",MSPE), sep="")
        }
        if (r > (TT-1)) {message(" (r hits maximum)")}
        message("\n\n r* = ",r.cv, sep="")
        message("\n\n") 

    }    ## End of Cross-Validation 
    validF <- ifelse(r.cv > 0, 1, 0)
    validX <- est.best$validX 

        ##------------------------------##
    ## ----------- Summarize -------------- ##
        ##------------------------------##    

    ##-------------------------------##
    ##   ATT and Counterfactuals     ##
    ##-------------------------------##

    ## 0. relevant parameters
    IC <- est.best$IC
    loglikelihood <- est.best$loglikelihood

    if (p>0) {
        na.pos <- is.nan(est.best$beta)
        beta <- est.best$beta
        if( sum(na.pos) > 0 ) {
            beta[na.pos] <- NA
        }
    } else {
        beta <- NA
    }
   
    ## 1. estimated att and counterfactuals
    Y.ct <- pnorm(est.best$fit)
    eff <- Y - Y.ct    
    att.avg <- sum(eff * D)/(sum(D))

    ## att.avg.unit
    tr.pos <- which(apply(D, 2, sum) > 0)
    att.unit <- sapply(1:length(tr.pos), function(vec){return(sum(eff[, tr.pos[vec]] * D[, tr.pos[vec]]) / sum(D[, tr.pos[vec]]))})
    att.avg.unit <- mean(att.unit)

    ## 2. rmse for treated units' observations under control: useless for binary
    ## average marginal effect
    marginal <- NULL
    #if (binary == TRUE) {
        if (p > 0) {
            dense <- dnorm(c(est.best$fit[which(II == 1)]))
            marginal <- as.matrix(sapply(1:p, function(vec){ mean(beta[vec] * dense)} ))
        }
    #}

    ## 3. unbalanced output
    if (0%in%I) {
        eff[which(I == 0)] <- NA
        Y.ct[which(I == 0)] <- NA
        est.best$fit[which(I == 0)] <- NA
    } 

    ## 4. dynamic effects
    t.on <- c(T.on)
    eff.v <- c(eff) ## a vector
    rm.pos1 <- which(is.na(eff.v))
    rm.pos2 <- which(is.na(t.on))
    eff.v.use1 <- eff.v
    t.on.use <- t.on
    n.on.use <- rep(1:N, each = TT)

    if (NA %in% eff.v | NA %in% t.on) {
        eff.v.use1 <- eff.v[-c(rm.pos1, rm.pos2)]
        t.on.use <- t.on[-c(rm.pos1, rm.pos2)]
        n.on.use <- n.on.use[-c(rm.pos1, rm.pos2)]
    }

    pre.pos <- which(t.on.use <= 0)
    eff.pre <- cbind(eff.v.use1[pre.pos], t.on.use[pre.pos], n.on.use[pre.pos])
    colnames(eff.pre) <- c("eff", "period", "unit")

    time.on <- sort(unique(t.on.use))
    att.on <- as.numeric(tapply(eff.v.use1, t.on.use, mean)) ## NA already removed
    count.on <- as.numeric(table(t.on.use))

    ## 5. placebo effect, if placeboTest == 1 , no placebo for CV


    ## 6. switch-off effects
    if (hasRevs == 1) {    
        t.off <- c(T.off)
        rm.pos3 <- which(is.na(t.off))
        eff.v.use2 <- eff.v
        t.off.use <- t.off

        if (NA %in% eff.v | NA %in% t.off) {
            eff.v.use2 <- eff.v[-c(rm.pos1, rm.pos3)]
            t.off.use <- t.off[-c(rm.pos1, rm.pos3)]
        }

        time.off <- sort(unique(t.off.use))
        att.off <- as.numeric(tapply(eff.v.use2, t.off.use, mean)) ## NA already removed
        count.off <- as.numeric(table(t.off.use))
    }

    ## 7. cohort effects
    if (!is.null(group)) {
        cohort <- cbind(c(group), c(D), c(eff.v))
        rm.pos <- unique(c(rm.pos1, which(cohort[, 2] == 0)))
        cohort <- cohort[-rm.pos, ]

        g.level <- sort(unique(cohort[, 1]))
        raw.group.att <- as.numeric(tapply(cohort[, 3], cohort[, 1], mean))

        group.att <- rep(NA, length(group.level))
        group.att[which(group.level %in% g.level)] <- raw.group.att 
    }

    ##-------------------------------##
    ##           Storage 
    ##-------------------------------##  

    ##control group residuals
    out<-list(
        ## main results 
        T.on = T.on,
        Y.ct = Y.ct,
        eff = eff,
        att.avg = att.avg,
        att.avg.unit = att.avg.unit,
        ## supporting
        force = force,
        T = TT,
        N = N,
        p = p, 
        r.cv = r.cv,
        est = est.best,
        mu = est.best$mu,
        beta = beta, 
        validX = validX,
        validF = validF,
        niter = est.best$niter, 
        time = time.on,
        att = att.on,
        count = count.on,
        eff.pre = eff.pre,
        CV.out = CV.out
    )

    if (hasRevs == 1) {
        out <- c(out, list(time.off = time.off, 
                           att.off = att.off,
                           count.off = count.off))
    }
    if (r.cv > 0) {
        out<-c(out,list(factor = as.matrix(est.best$factor),
                        lambda = as.matrix(est.best$lambda))) 
    }
    if (force == 1) {
        out<-c(out, list(alpha = est.best$alpha))
    } else if (force == 2) {
        out<-c(out,list(xi = est.best$xi))
    } else if (force == 3) {
        out<-c(out,list(alpha = est.best$alpha, xi = est.best$xi))
    }

    if (!is.null(group)) {
        out <- c(out, list(group.att = group.att))
    }
 
    return(out)
} ## binary cross-validation function ends

.has_honest <- function() {
  requireNamespace("HonestDiDFEct", quietly = TRUE)
}

.honest <- function(fun) {
  if (!.has_honest()) {
    stop("Optional dependency ``HonestDiDFEct'' not available. Install from https://github.com/lzy318/HonestDiDFEct .", call. = FALSE)
  }
  getExportedValue("HonestDiDFEct", fun)
}


fect_sens <- function(
    fect.out,
    post.periods = NA, # Post-treatment periods
    l_vec = NA, # Optional custom weighting vector
    Mbarvec = seq(0, 1, by = 0.1), # Vector of Mbar for RM analysis
    Mvec = seq(0, 0.25, 0.05), # Vector of M for Smoothness analysis
    periodMbarvec = c(0, 0.5), # Vector of Mbar for period-by-period RM analysis
    periodMvec = c(0, 0.1), # Vector of M for Smoothness analysis
    parallel = FALSE,
    cores = NULL) {
  if (!identical(parallel, FALSE) && is.null(cores)) {
    cores <- parallelly::availableCores(omit = 2, max = 8) # use at most 8 cores
    cl <- parallel::makeCluster(cores)
    doParallel::registerDoParallel(cl)
  }

  # -------------------------------------------------------------------
  # 1) Identify relevant periods and extract fect estimates + vcov
  # -------------------------------------------------------------------
  # We want event-time strings for pre and post
  pre.periods <- fect.out$placebo.period[1]:fect.out$placebo.period[2]
  if (any(is.na(post.periods))) {
    post.periods <- 1:length(fect.out$time) - length(fect.out$pre.periods) - (fect.out$placebo.period[2] - fect.out$placebo.period[1] + 1)
  }
  all.periods.char <- as.character(c(pre.periods, post.periods))

  # Make sure that:
  #   - rownames(fect.out$est.att) match these event times (e.g. "-2","-1","0","1", etc.)
  #   - fect.out$att.vcov has dimension names that match rownames(fect.out$est.att)
  # If 'att.vcov' lacks dimnames, you can do:
  #   dimnames(fect.out$att.vcov) <- list(rownames(fect.out$est.att),
  #                                       rownames(fect.out$est.att))

  # Numeric indices corresponding to those event-time strings
  idx <- match(all.periods.char, rownames(fect.out$est.att))
  idx <- idx[!is.na(idx)] # remove anything not found

  # Extract DTE estimates (beta.hat) and var-cov (vcov.hat)
  beta.hat <- fect.out$est.att[idx, 1]

  if (is.matrix(fect.out$att.vcov)) {
    vcov.hat <- fect.out$att.vcov[idx, idx]
  } else if (is.matrix(fect.out$att.boot)) {
    # Fallback: compute vcov from bootstrap samples if att.vcov is unavailable
    vcov.hat <- cov(t(fect.out$att.boot[idx, , drop = FALSE]),
                    use = "pairwise.complete.obs")
    if (!is.matrix(vcov.hat)) {
      stop("fect_sens requires a valid variance-covariance matrix. ",
           "Could not compute one from bootstrap samples. ",
           "Please re-run fect() with se = TRUE and sufficient nboots.",
           call. = FALSE)
    }
  } else {
    stop("fect_sens requires a valid variance-covariance matrix (att.vcov) from fect(). ",
         "Please re-run fect() with se = TRUE and ensure sufficient bootstrap iterations (nboots).",
         call. = FALSE)
  }

  # Counts of pre and post periods
  numPrePeriods <- length(pre.periods)
  numPostPeriods <- length(post.periods)

  # -------------------------------------------------------------------
  # 2) Construct weights for an "overall" post-treatment ATT
  # -------------------------------------------------------------------
  # If the user did NOT provide a custom weighting vector l_vec,
  # default to count-based weighting (counts from each post period).
  if (any(is.na(l_vec))) {
    # Indices for post-periods in the rownames
    post.idx <- match(as.character(post.periods), rownames(fect.out$est.att))
    post.idx <- post.idx[!is.na(post.idx)]

    count <- fect.out$count[post.idx]
    w.att <- count / sum(count)
  } else {
    w.att <- l_vec
  }

  # We will attach everything to the original fect.out in new sub-lists.
  # For each approach (RM or Smoothness), we will store both:
  #   (a) Weighted-average results
  #   (b) Period-by-period results

  # Initialize empty placeholders
  fect.out$RM_Sensitivity <- NULL
  fect.out$Smooth_Sensitivity <- NULL

  # -------------------------------------------------------------------
  # 3) Relative Magnitude Analysis (RM), if Mbarvec is non-empty
  # -------------------------------------------------------------------
  if (!is.null(Mbarvec) && length(Mbarvec) > 0) {
    # 3a) Weighted-average, across the entire post-treatment window
    rm_sens_results <- suppressWarnings(.honest("createSensitivityResults_relativeMagnitudes")(
      betahat = beta.hat,
      sigma = vcov.hat,
      numPrePeriods = numPrePeriods,
      numPostPeriods = numPostPeriods,
      l_vec = w.att,
      Mbarvec = Mbarvec,
      parallel = parallel
    ))


    rm_original_cs <- suppressWarnings(.honest("constructOriginalCS")(
      betahat        = beta.hat,
      sigma          = vcov.hat,
      numPrePeriods  = numPrePeriods,
      numPostPeriods = numPostPeriods,
      l_vec          = w.att
    ))
  }
  if (!is.null(periodMbarvec) && length(periodMbarvec) > 0) {
    # 3b) Period-by-period robust confidence sets
    #     We'll loop over each post-treatment period and set a vector of 0s except 1 for that period
    rm_period_output <- cbind.data.frame() # Will accumulate results

    for (t_i in seq_len(numPostPeriods)) {
      # l_vec for the "t_i-th" post period (in 1..numPostPeriods)
      # We need a vector of length numPrePeriods + numPostPeriods
      # Set that post period's position to 1
      dte_l <- rep(0, numPostPeriods)
      dte_l[t_i] <- 1

      # For each t_i, we run createSensitivityResults_relativeMagnitudes
      # across all Mbar in Mbarvec
      honest.dte <- suppressWarnings(.honest("createSensitivityResults_relativeMagnitudes")(
        betahat        = beta.hat,
        sigma          = vcov.hat,
        numPrePeriods  = numPrePeriods,
        numPostPeriods = numPostPeriods,
        l_vec          = dte_l,
        Mbarvec        = periodMbarvec,
        parallel       = parallel
      ))

      # Convert to data.frame
      # The returned object typically has columns lb, ub, Mbar, etc.
      # We add a column for the actual "post period"
      honest.dte <- as.data.frame(honest.dte)
      # Actual event time is post.periods[t_i]
      honest.dte$postPeriod <- post.periods[t_i]

      # Append
      rm_period_output <- rbind(rm_period_output, honest.dte)
    }

    # Store results in fect.out$sensitivity.rm
    fect.out$sensitivity.rm <- list(
      results = rm_sens_results,
      original = rm_original_cs,
      periods = rm_period_output
    )
  }

  # -------------------------------------------------------------------
  # 4) Smoothness Analysis (C-LF), if Mvec is non-empty
  # -------------------------------------------------------------------
  if (!is.null(Mvec) && length(Mvec) > 0) {
    # 4a) Weighted-average analysis

    smooth_sens_results <- suppressWarnings(.honest("createSensitivityResults")(
      betahat = beta.hat,
      sigma = vcov.hat,
      numPrePeriods = numPrePeriods,
      numPostPeriods = numPostPeriods,
      method = "C-LF",
      l_vec = w.att,
      Mvec = Mvec,
      parallel = parallel
    ))

    sm_original_cs <- suppressWarnings(.honest("constructOriginalCS")(
      betahat        = beta.hat,
      sigma          = vcov.hat,
      numPrePeriods  = numPrePeriods,
      numPostPeriods = numPostPeriods,
      l_vec          = w.att
    ))
  }
  if (!is.null(periodMvec) && length(periodMvec) > 0) {
    # 4b) Period-by-period robust confidence sets
    smooth_period_output <- cbind.data.frame()

    for (t_i in seq_len(numPostPeriods)) {
      # l_vec for the "t_i-th" post period
      dte_l <- rep(0, numPostPeriods)
      dte_l[t_i] <- 1

      honest.dte <- suppressWarnings(.honest("createSensitivityResults")(
        betahat = beta.hat,
        sigma = vcov.hat,
        numPrePeriods = numPrePeriods,
        numPostPeriods = numPostPeriods,
        method = "C-LF",
        l_vec = dte_l,
        Mvec = periodMvec,
        parallel = parallel
      ))

      honest.dte <- as.data.frame(honest.dte)
      honest.dte$postPeriod <- post.periods[t_i]

      smooth_period_output <- rbind(smooth_period_output, honest.dte)
    }

    # Store results in fect.out$sensitivity.smooth
    fect.out$sensitivity.smooth <- list(
      results = smooth_sens_results,
      original = sm_original_cs,
      periods = smooth_period_output
    )
  }
  if (!identical(parallel, FALSE)) {
    stopCluster(cl)
  }
  # -------------------------------------------------------------------
  # 5) Return the updated fect.out
  # -------------------------------------------------------------------
  return(fect.out)
}

test_that("fect_sens attaches sensitivity results when inputs present", {
  skip_on_cran()
  skip_if_not_installed("HonestDiDFEct")
  suppressWarnings(try(data("simdata", package = "fect"), silent = TRUE))
  expect_true(exists("simdata"))

  set.seed(3)
  out <- fect::fect(
    Y ~ D + X1 + X2,
    data = simdata,
    index = c("id", "time"),
    method = "ife",
    r = 1,
    CV = FALSE,
    se = TRUE,
    nboots = 20,
    parallel = FALSE
  )

  # ensure names and shapes needed by fect_sens exist
  expect_true(!is.null(out$est.att))
  expect_true(!is.null(out$att.vcov))

  # Ensure placebo.period exists to avoid length-0 error inside fect_sens
  if (is.null(out$placebo.period)) {
    out$placebo.period <- c(-3, -1)
  }
  # restrict periods so numPre+numPost matches available row count
  out$placebo.period <- c(-3, -1)
  out2 <- fect_sens(out, post.periods = 1:10, parallel = FALSE)
  expect_true(!is.null(out2$sensitivity.rm) || !is.null(out2$sensitivity.smooth))
})

## -----------------------------------------------------------------------------
## loading_bound.R
##
## Entropy-regularized simplex projection of treated factor loadings for GSC.
##
## Solves, per treated unit i:
##
##   minimize   (1/gamma) * sum_j w_j * log(w_j / q_j)
##            + || u_pre - F_pre %*% t(Lambda_co) %*% w ||_2^2
##   over       w in Delta_{Nco}  (simplex)
##
## The resulting loading is lambda_hat = t(Lambda_co) %*% w, which lies in the
## convex hull of control loadings by construction. Under the factor model this
## yields Y_hat(0) = F %*% lambda_hat = F %*% t(Lambda_co) %*% w, a convex
## combination of factor-implied control outcomes.
##
## See statsclaw-workspace/fect/runs/REQ-bounded-loadings/spec.md for the full
## design and the theoretical justification.
## -----------------------------------------------------------------------------

.bounded_loading_obj <- function(theta, u_pre, F_pre, Lambda_co, gamma, q) {
    ## Softmax reparameterization: w_j = exp(theta_j) / sum_k exp(theta_k).
    th   <- theta - max(theta)            # numerical stability
    ew   <- exp(th)
    w    <- ew / sum(ew)
    w_safe <- pmax(w, 1e-300)
    lam  <- crossprod(Lambda_co, w)       # r-vector
    res  <- u_pre - F_pre %*% lam
    ent  <- sum(w * (log(w_safe) - log(q)))
    val  <- (1 / gamma) * ent + sum(res^2)
    if (!is.finite(val)) val <- 1e30      # L-BFGS-B needs finite values
    val
}

.bounded_loading_grad <- function(theta, u_pre, F_pre, Lambda_co, gamma, q) {
    th   <- theta - max(theta)
    ew   <- exp(th)
    w    <- ew / sum(ew)
    w_safe <- pmax(w, 1e-300)
    lam  <- crossprod(Lambda_co, w)
    res  <- u_pre - F_pre %*% lam

    ## d/dw of fit term: -2 * Lambda_co %*% t(F_pre) %*% res   (Nco-vector)
    g_fit_w <- -2 * as.numeric(Lambda_co %*% crossprod(F_pre, res))

    ## d/dw of entropy term: (1/gamma) * (log(w/q) + 1)
    g_ent_w <- (1 / gamma) * (log(w_safe / q) + 1)

    g_w <- g_fit_w + g_ent_w

    ## Chain rule for softmax: g_theta = w * (g_w - sum(w * g_w))
    g <- w * (g_w - sum(w * g_w))
    g[!is.finite(g)] <- 0
    g
}

.mirror_descent_bounded_loading <- function(
    u_pre, F_pre, Lambda_co, gamma, q,
    tol = 1e-8, max_iter = 500L
) {
    Nco <- nrow(Lambda_co)
    w   <- rep(1 / Nco, Nco)
    obj_new <- Inf
    for (iter in seq_len(max_iter)) {
        lam  <- crossprod(Lambda_co, w)
        res  <- u_pre - F_pre %*% lam
        g_w  <- -2 * as.numeric(Lambda_co %*% crossprod(F_pre, res)) +
                 (1 / gamma) * (log(pmax(w, 1e-300) / q) + 1)
        step <- 1 / (iter + 1)             # diminishing step
        ## Clip gradient to prevent exp() overflow on the log-simplex step
        g_w <- pmax(pmin(g_w, 1e6), -1e6)
        lw   <- log(pmax(w, 1e-300)) - step * g_w
        lw   <- lw - max(lw)
        ew   <- exp(lw)
        ew_sum <- sum(ew)
        if (!is.finite(ew_sum) || ew_sum <= 0) {
            ## Numerical failure; fall back to uniform and terminate
            w <- rep(1 / Nco, Nco)
            break
        }
        w_new <- ew / ew_sum
        if (any(!is.finite(w_new))) {
            w <- rep(1 / Nco, Nco)
            break
        }

        obj_new <- sum(res^2) +
            (1 / gamma) * sum(w_new * (log(pmax(w_new, 1e-300)) - log(q)))
        delta <- max(abs(w_new - w))
        if (isTRUE(is.finite(delta) && delta < tol)) {
            w <- w_new
            break
        }
        w <- w_new
    }
    list(w = w, n_iter = iter, obj = obj_new)
}

.solve_bounded_loading <- function(
    u_pre,
    F_pre,
    Lambda_co,
    gamma,
    q        = NULL,
    tol      = 1e-8,
    max_iter = 200L,
    fallback = TRUE
) {
    u_pre     <- as.numeric(u_pre)
    F_pre     <- as.matrix(F_pre)
    Lambda_co <- as.matrix(Lambda_co)
    Nco       <- nrow(Lambda_co)
    stopifnot(length(u_pre) == nrow(F_pre))
    stopifnot(ncol(F_pre) == ncol(Lambda_co))

    if (Nco < 2L) {
        stop("bounded-loading projection requires at least 2 control units; got Nco = ",
             Nco, ".", call. = FALSE)
    }
    if (!is.numeric(gamma) || length(gamma) != 1L || !is.finite(gamma) || gamma <= 0) {
        stop("gamma must be a single finite positive numeric; got ", gamma, ".",
             call. = FALSE)
    }
    if (is.null(q)) {
        q <- rep(1 / Nco, Nco)
    } else {
        q <- as.numeric(q)
        if (length(q) != Nco || any(q <= 0) || abs(sum(q) - 1) > 1e-8) {
            stop("q must be a positive Nco-vector summing to 1.", call. = FALSE)
        }
    }

    ## Degenerate: all-zero Lambda_co -- return uniform w
    if (max(abs(Lambda_co)) < 1e-12) {
        w <- q
        return(list(
            w          = w,
            lambda_hat = as.numeric(crossprod(Lambda_co, w)),
            obj        = sum(u_pre^2),
            converged  = TRUE,
            n_iter     = 0L,
            method     = "degenerate-return-uniform"
        ))
    }

    if (Nco <= ncol(Lambda_co)) {
        warning(
            "Nco (", Nco, ") <= r (", ncol(Lambda_co),
            "); conv(Lambda_co) has empty interior in R^r and the projection may ",
            "be highly constrained.",
            call. = FALSE
        )
    }

    theta0 <- rep(0, Nco)
    opt <- tryCatch(
        stats::optim(
            par     = theta0,
            fn      = .bounded_loading_obj,
            gr      = .bounded_loading_grad,
            method  = "L-BFGS-B",
            u_pre   = u_pre,
            F_pre   = F_pre,
            Lambda_co = Lambda_co,
            gamma   = gamma,
            q       = q,
            control = list(maxit = max_iter, factr = max(tol / .Machine$double.eps, 1))
        ),
        error = function(e) list(convergence = 99L, message = conditionMessage(e))
    )

    converged_lbfgs <- isTRUE(opt$convergence == 0L)

    if (converged_lbfgs) {
        th   <- opt$par - max(opt$par)
        ew   <- exp(th)
        w    <- ew / sum(ew)
        lam  <- as.numeric(crossprod(Lambda_co, w))
        return(list(
            w          = w,
            lambda_hat = lam,
            obj        = opt$value,
            converged  = TRUE,
            n_iter     = if (!is.null(opt$counts)) unname(opt$counts[1]) else NA_integer_,
            method     = "lbfgs"
        ))
    }

    if (!fallback) {
        return(list(
            w          = rep(1 / Nco, Nco),
            lambda_hat = as.numeric(crossprod(Lambda_co, rep(1 / Nco, Nco))),
            obj        = NA_real_,
            converged  = FALSE,
            n_iter     = NA_integer_,
            method     = "lbfgs-failed"
        ))
    }

    md <- .mirror_descent_bounded_loading(
        u_pre     = u_pre,
        F_pre     = F_pre,
        Lambda_co = Lambda_co,
        gamma     = gamma,
        q         = q,
        tol       = tol,
        max_iter  = 500L
    )
    list(
        w          = md$w,
        lambda_hat = as.numeric(crossprod(Lambda_co, md$w)),
        obj        = md$obj,
        converged  = TRUE,
        n_iter     = md$n_iter,
        method     = "mirror-descent"
    )
}

## -----------------------------------------------------------------------------
## .cv_gamma_loading
##
## Selects gamma by 5-fold CV over pre-treatment periods. Folds are constructed
## across pre-treatment time indices (shared across treated units); per-fold
## score is the mean squared prediction error of the bounded projection on the
## held-out pre-treatment observations.
## -----------------------------------------------------------------------------

.cv_gamma_loading <- function(
    U_tr_pre,        # T0 x Ntr matrix OR list of length Ntr of unit-specific vectors
    F_hat_pre,       # T0 x r matrix (balanced case) OR list per unit (unbalanced)
    Lambda_co,       # Nco x r matrix
    gamma_grid,      # positive numeric vector
    cv_k        = 5L,
    tol         = 1e-8,
    max_iter    = 200L
) {
    Lambda_co <- as.matrix(Lambda_co)
    stopifnot(all(gamma_grid > 0))

    is_balanced <- is.matrix(U_tr_pre) && is.matrix(F_hat_pre)

    if (is_balanced) {
        T0  <- nrow(U_tr_pre)
        Ntr <- ncol(U_tr_pre)
        fold_id <- sample(rep_len(seq_len(cv_k), T0))
    } else {
        Ntr <- length(U_tr_pre)
        fold_id <- NULL   # per-unit folding for unbalanced
    }

    cv_mspe <- numeric(length(gamma_grid))
    for (gi in seq_along(gamma_grid)) {
        gamma_i <- gamma_grid[gi]
        sq_err  <- 0
        n_obs   <- 0L

        if (is_balanced) {
            for (ii in seq_len(cv_k)) {
                tr_idx <- which(fold_id != ii)
                te_idx <- which(fold_id == ii)
                if (length(tr_idx) < ncol(F_hat_pre) || length(te_idx) == 0L) next
                F_tr <- F_hat_pre[tr_idx, , drop = FALSE]
                F_te <- F_hat_pre[te_idx, , drop = FALSE]
                for (i in seq_len(Ntr)) {
                    u_tr <- U_tr_pre[tr_idx, i]
                    u_te <- U_tr_pre[te_idx, i]
                    sol <- .solve_bounded_loading(
                        u_pre     = u_tr,
                        F_pre     = F_tr,
                        Lambda_co = Lambda_co,
                        gamma     = gamma_i,
                        tol       = tol,
                        max_iter  = max_iter
                    )
                    pred <- as.numeric(F_te %*% sol$lambda_hat)
                    sq_err <- sq_err + sum((u_te - pred)^2)
                    n_obs  <- n_obs + length(u_te)
                }
            }
        } else {
            for (i in seq_len(Ntr)) {
                u_i <- as.numeric(U_tr_pre[[i]])
                F_i <- as.matrix(F_hat_pre[[i]])
                T0i <- length(u_i)
                if (T0i < cv_k + ncol(F_i)) next
                fold_i <- sample(rep_len(seq_len(cv_k), T0i))
                for (ii in seq_len(cv_k)) {
                    tr_idx <- which(fold_i != ii)
                    te_idx <- which(fold_i == ii)
                    if (length(tr_idx) < ncol(F_i) || length(te_idx) == 0L) next
                    sol <- .solve_bounded_loading(
                        u_pre     = u_i[tr_idx],
                        F_pre     = F_i[tr_idx, , drop = FALSE],
                        Lambda_co = Lambda_co,
                        gamma     = gamma_i,
                        tol       = tol,
                        max_iter  = max_iter
                    )
                    pred <- as.numeric(F_i[te_idx, , drop = FALSE] %*% sol$lambda_hat)
                    sq_err <- sq_err + sum((u_i[te_idx] - pred)^2)
                    n_obs  <- n_obs + length(te_idx)
                }
            }
        }

        cv_mspe[gi] <- if (n_obs > 0L) sq_err / n_obs else Inf
    }

    best <- which.min(cv_mspe)
    list(
        gamma_cv   = gamma_grid[best],
        cv_mspe    = cv_mspe,
        gamma_grid = gamma_grid
    )
}

## Default gamma grid: 9 log-spaced points over [0.01, 100]
.default_gamma_grid <- function() 10 ^ seq(-2, 2, length.out = 9L)

## Unit tests for .solve_bounded_loading and .cv_gamma_loading internals.
## See statsclaw-workspace/fect/runs/REQ-bounded-loadings/test-spec.md section S.

test_that("S.1: solver output lies on the simplex", {
    set.seed(1)
    T0 <- 20; r <- 2; Nco <- 15
    F_pre     <- matrix(rnorm(T0 * r), T0, r)
    Lambda_co <- matrix(rnorm(Nco * r), Nco, r)
    u_pre     <- F_pre %*% c(0.3, 0.5) + rnorm(T0, sd = 0.1)
    sol       <- fect:::.solve_bounded_loading(u_pre, F_pre, Lambda_co, gamma = 1)
    expect_equal(sum(sol$w), 1, tolerance = 1e-10)
    expect_true(all(sol$w >= -1e-12))
    expect_true(sol$converged)
    expect_equal(length(sol$lambda_hat), r)
})

test_that("S.2: lambda_hat = t(Lambda_co) %*% w", {
    set.seed(2)
    T0 <- 20; r <- 3; Nco <- 20
    F_pre     <- matrix(rnorm(T0 * r), T0, r)
    Lambda_co <- matrix(rnorm(Nco * r), Nco, r)
    u_pre     <- F_pre %*% c(0.2, 0.4, -0.1) + rnorm(T0, sd = 0.1)
    sol       <- fect:::.solve_bounded_loading(u_pre, F_pre, Lambda_co, gamma = 1)
    expect_equal(as.numeric(sol$lambda_hat),
                 as.numeric(crossprod(Lambda_co, sol$w)),
                 tolerance = 1e-8)
})

test_that("S.3: small gamma -> weights shrink toward uniform; large gamma -> near-unregularized", {
    set.seed(3)
    T0 <- 30; r <- 2; Nco <- 12
    F_pre     <- matrix(rnorm(T0 * r), T0, r)
    Lambda_co <- matrix(rnorm(Nco * r), Nco, r)
    u_pre     <- as.numeric(F_pre %*% c(0.4, 0.2))   # no noise; OLS target inside hull
    sol_small <- fect:::.solve_bounded_loading(u_pre, F_pre, Lambda_co, gamma = 1e-4)
    sol_large <- fect:::.solve_bounded_loading(u_pre, F_pre, Lambda_co, gamma = 1e6)
    ## Small gamma: entropy dominates, weights pulled very close to uniform
    expect_lt(max(abs(sol_small$w - 1 / Nco)), 0.05)
    ## Large gamma: entropy negligible, projection fits the noise-free target well
    resid_large <- u_pre - as.numeric(F_pre %*% sol_large$lambda_hat)
    expect_lt(sum(resid_large^2), 1e-4)
})

test_that("S.4: exterior OLS target -> projection on hull, nontrivial residual", {
    set.seed(4)
    T0 <- 20; r <- 2; Nco <- 10
    Lambda_co <- matrix(rnorm(Nco * r, sd = 0.5), Nco, r)
    F_pre     <- matrix(rnorm(T0 * r), T0, r)
    lambda_tr_true <- c(3, 3)              # far outside hull
    u_pre <- as.numeric(F_pre %*% lambda_tr_true) + rnorm(T0, sd = 0.1)
    sol   <- fect:::.solve_bounded_loading(u_pre, F_pre, Lambda_co, gamma = 1e6)
    ## Projected loading should lie within the bounding box of Lambda_co
    expect_lte(max(sol$lambda_hat), max(Lambda_co) + 1e-6)
    expect_gte(min(sol$lambda_hat), min(Lambda_co) - 1e-6)
    ## Residual should be substantially nonzero (cannot fit (3,3) with hull)
    resid <- u_pre - as.numeric(F_pre %*% sol$lambda_hat)
    expect_gt(sum(resid^2), 0.1)
})

test_that("S.5: Nco < 2 raises informative error", {
    expect_error(
        fect:::.solve_bounded_loading(
            u_pre = rnorm(10), F_pre = matrix(rnorm(10), 10, 1),
            Lambda_co = matrix(0.5, 1, 1), gamma = 1
        ),
        regexp = "at least 2 control units"
    )
})

test_that("S.6: all-zero Lambda_co returns uniform w, zero lambda_hat", {
    sol <- fect:::.solve_bounded_loading(
        u_pre     = rnorm(10),
        F_pre     = matrix(rnorm(20), 10, 2),
        Lambda_co = matrix(0, 15, 2),
        gamma     = 1
    )
    expect_equal(sol$w, rep(1 / 15, 15), tolerance = 1e-10)
    expect_equal(as.numeric(sol$lambda_hat), c(0, 0), tolerance = 1e-12)
    expect_true(sol$converged)
    expect_equal(sol$method, "degenerate-return-uniform")
})

test_that("S.7: mirror-descent fallback can run on ill-posed inputs", {
    set.seed(7)
    T0 <- 5; r <- 4; Nco <- 30       # r > T0 -- ill-posed regression
    F_pre     <- matrix(rnorm(T0 * r), T0, r)
    Lambda_co <- matrix(rnorm(Nco * r), Nco, r)
    u_pre     <- rnorm(T0)
    ## fallback = TRUE (default) ensures we don't silently fail
    sol <- fect:::.solve_bounded_loading(u_pre, F_pre, Lambda_co,
                                         gamma = 1, fallback = TRUE)
    expect_true(sol$converged)
    expect_true(sol$method %in% c("lbfgs", "mirror-descent"))
    expect_equal(sum(sol$w), 1, tolerance = 1e-6)
    expect_true(all(sol$w >= -1e-8))
})

test_that("S.8: .cv_gamma_loading picks from grid, returns valid selection", {
    set.seed(8)
    T0 <- 30; r <- 2; Nco <- 20; Ntr <- 5
    F_pre     <- matrix(rnorm(T0 * r), T0, r)
    Lambda_co <- matrix(rnorm(Nco * r), Nco, r)
    ## Treated loadings sampled from same dist; interior case
    Lambda_tr <- matrix(rnorm(Ntr * r), Ntr, r)
    U_tr_pre  <- F_pre %*% t(Lambda_tr) + matrix(rnorm(T0 * Ntr, sd = 0.3), T0, Ntr)
    grid      <- c(0.1, 1.0, 10.0)
    res <- fect:::.cv_gamma_loading(
        U_tr_pre   = U_tr_pre,
        F_hat_pre  = F_pre,
        Lambda_co  = Lambda_co,
        gamma_grid = grid,
        cv_k       = 3L
    )
    expect_true(res$gamma_cv %in% grid)
    expect_length(res$cv_mspe, length(grid))
    expect_true(all(is.finite(res$cv_mspe)))
})

###################################################################
## Cross-validation
###################################################################
fect_cv <- function(Y, # Outcome variable, (T*N) matrix
                    X, # Explanatory variables:  (T*N*p) array
                    D, #  Indicator for treated unit (tr==1)
                    W,
                    I,
                    II,
                    T.on,
                    T.off = NULL,
                    T.on.carry = NULL,
                    T.on.balance = NULL,
                    balance.period = NULL,
                    method = "ife",
                    criterion = "mspe",
                    k = 20, # CV time
                    cv.prop = 0.1,
                    cv.method = "rolling",
                    cv.nobs = 3,
                    cv.donut = 1,
                    cv.buffer = 1,
                    min.T0 = 5,
                    r = 0, # initial number of factors considered if CV==1
                    r.end,
                    proportion = 0,
                    nlambda = 10,
                    lambda = NULL,
                    force,
                    hasRevs = 1,
                    tol, # tolerance level
                    max.iteration = 1000,
                    norm.para = NULL,
                    group.level = NULL,
                    group = NULL,
                    time.component.from = "notyettreated",
                    X.extra.FE = NULL,
                    X.Z = NULL,
                    X.Q = NULL,
                    X.gamma = NULL,
                    X.kappa = NULL,
                    Zgamma.id = NULL,
                    kappaQ.id = NULL,
                    parallel = FALSE,
                    cores = NULL,
                    do_parallel_cv   = FALSE,   ## pre-computed flag from default.R
                    do_parallel_boot = FALSE,    ## threaded through; not used in cv.R
                    cv.rule = "1se",             ## "1se" (default), "min", or "1pct" (legacy)
                    W.in.fit = TRUE              ## whether W enters the outcome-model fit
                    ) {
    cv.rule <- .fect_validate_cv_rule(cv.rule)
    ## -------------------------------##
    ## Parsing data
    ## -------------------------------##

    ## Two-tier tolerance: CV uses looser tol for r/lambda selection speed,
    ## final estimation uses the user's tol for precision.
    cv_tol <- max(tol, 1e-3)

    placebo.pos <- na.pos <- NULL
    ## unit id and time
    TT <- dim(Y)[1]
    N <- dim(Y)[2]
    if (is.null(X) == FALSE) {
        p <- dim(X)[3]
    } else {
        p <- 0
        X <- array(0, dim = c(1, 1, 0))
    }

    if (is.null(W) || !W.in.fit) {
        W.use <- as.matrix(0)
        use_weight <- 0
    } else {
        use_weight <- 1
        W.use <- W
        W.use[which(II == 0)] <- 0
    }

    ## replicate data
    YY <- Y
    YY[which(II == 0)] <- 0 ## reset to 0
    if (use_weight) {
        WW <- W
        WW[which(II == 0)] <- 0 ## reset to 0
    }
    t.on <- c(T.on)
    II.colsums <- apply(II, 2, sum)
    T0.min <- min(II.colsums[II.colsums > 0])

    D.c <- apply(D, 2, function(vec) {
        cumsum(vec)
    })
    D.c <- ifelse(D.c > 0, 1, 0)
    D.sum <- colSums(D.c)
    tr <- which(D.sum >= 1)
    Ntr <- length(tr)
    co <- which(D.sum == 0)
    Nco <- length(co)

    ##  --------- initial fit using fastplm --------- ##
    data.ini <- matrix(NA, (TT * N), (2 + 1 + p))
    data.ini[, 2] <- rep(1:N, each = TT) ## unit fe
    data.ini[, 3] <- rep(1:TT, N) ## time fe
    data.ini[, 1] <- c(Y) ## outcome
    if (p > 0) { ## covar
        for (i in 1:p) {
            data.ini[, (3 + i)] <- c(X[, , i])
        }
    }
    ## observed Y0 indicator:
    oci <- which(c(II) == 1)
    if (use_weight == 1) {
        initialOut <- initialFit(
            data = data.ini,
            force = force,
            w = c(W),
            oci = oci
        )
    } else {
        initialOut <- initialFit(
            data = data.ini,
            force = force,
            oci = oci
        )
    }

    Y0 <- initialOut$Y0
    beta0 <- initialOut$beta0
    if (p > 0 && sum(is.na(beta0)) > 0) {
        beta0[which(is.na(beta0))] <- 0
    }

    ## ------------- restrictions on candidate hyper parameters ---------- ##
    obs.con <- (sum(II) - r.end * (N + TT) + r.end^2 - p) <= 0
    if (obs.con) {
        while ((sum(II) - r.end * (N + TT) + r.end^2 - p) <= 0) {
            r.end <- r.end - 1
        }
    }
    if (r.end >= T0.min) {
        if (method %in% c("both", "ife", "gsynth")) {
            message("Factor number should not be greater than ", T0.min - 1, "\n", sep = "")
        }
        r.end <- T0.min - 1
    } else {
        if (obs.con) {
            if (method %in% c("both", "ife", "gsynth")) {
                message("Factor number should not be greater than ", r.end, "\n", sep = "")
            }
        }
    }

    ## -------------------------------##
    ## ----------- Main Algorithm ----------- ##
    ## -------------------------------##

    validX <- 1 ## no multi-colinearity
    CV.out.ife <- CV.out.mc <- NULL

    ## ----------------------------------------------------##
    ##         Cross-validation of r and lambda           ##
    ## ----------------------------------------------------##

    r.max <- min(TT, r.end)
    r.cv <- 0 ## initial value

    if (method %in% c("ife", "both", "gsynth") && FALSE) {
        r.cv <- 0
        est.best <- inter_fe_ub(YY, Y0,
            X, II, W.use, beta0,
            0,
            force = force,
            tol, max.iteration
        )
        message("Cross validation cannot be performed since available pre-treatment records of treated units are too few. So set r.cv = 0.\n ")
    } else {
        r.old <- r ## save the minimal number of factors
        message("Cross-validating ...", "\n")
        if (criterion == "mspe") {
            message("Criterion: Mean Squared Prediction Error\n")
        } else if (criterion == "wmspe") {
            message("Criterion: Weighted Mean Squared Prediction Error\n")
        } else if (criterion == "gmspe") {
            message("Criterion: Geometric Mean Squared Prediction Error\n")
        } else if (criterion == "wgmspe") {
            message("Criterion: Weighted Geometric Mean Squared Prediction Error\n")
        } else if (criterion == "mad") {
            message("Criterion: Median Absolute Deviation\n")
        } else if (criterion == "moment") {
            message("Criterion: Moment Conditions\n")
        } else if (criterion == "gmoment") {
            message("Criterion: Geometric Moment Conditions\n")
        } else if (criterion == "pc") {
            message("Criterion: PC\n")
        }

        ## for gsynth, use the cross-validation function in fect_nevertreated
        if (method == "gsynth") {
            message("Interactive fixed effects model...\n")
            out <- fect_nevertreated(
                Y = Y, D = D, X = X, W = W, I = I, II = II,
                T.on = T.on, T.off = T.off,
                T.on.balance = T.on.balance,
                balance.period = balance.period,
                r = r, r.end = r.end, CV = TRUE,
                force = force, hasRevs = hasRevs,
                tol = tol, boot = 0,
                norm.para = norm.para,
                group.level = group.level, group = group,
                cv.method = cv.method, cv.nobs = cv.nobs,
                cv.prop = cv.prop, cv.donut = cv.donut, cv.buffer = cv.buffer,
                min.T0 = min.T0, k = k, criterion = criterion,
                parallel = parallel, cores = cores,
                do_parallel_cv = do_parallel_cv
            )
            return(out)
        }

        ## for ife with nevertreated, use the cross-validation function in fect_nevertreated
        if (method == "ife" && time.component.from == "nevertreated") {
            message("IFE model with nevertreated factors...\n")
            out <- fect_nevertreated(
                Y = Y, D = D, X = X, W = W, I = I, II = II,
                T.on = T.on, T.off = T.off,
                T.on.balance = T.on.balance,
                balance.period = balance.period,
                r = r, r.end = r.end, CV = TRUE,
                force = force, hasRevs = hasRevs,
                tol = tol, boot = 0,
                norm.para = norm.para,
                group.level = group.level, group = group,
                cv.method = cv.method, cv.nobs = cv.nobs,
                cv.prop = cv.prop, cv.donut = cv.donut, cv.buffer = cv.buffer,
                min.T0 = min.T0, k = k, criterion = criterion,
                parallel = parallel, cores = cores,
                do_parallel_cv = do_parallel_cv
            )
            return(out)
        }

        ## for cfe with nevertreated, use the cross-validation function in fect_nevertreated
        if (method == "cfe" && time.component.from == "nevertreated") {
            message("CFE model with nevertreated factors...\n")
            out <- fect_nevertreated(
                Y = Y, D = D, X = X, W = W, I = I, II = II,
                T.on = T.on, T.off = T.off,
                T.on.balance = T.on.balance,
                balance.period = balance.period,
                r = r, r.end = r.end, CV = TRUE,
                force = force, hasRevs = hasRevs,
                tol = tol, boot = 0,
                norm.para = norm.para,
                group.level = group.level, group = group,
                method = "cfe",
                X.extra.FE = X.extra.FE, X.Z = X.Z, X.Q = X.Q,
                X.gamma = X.gamma, X.kappa = X.kappa,
                Zgamma.id = Zgamma.id, kappaQ.id = kappaQ.id,
                cv.method = cv.method, cv.nobs = cv.nobs,
                cv.prop = cv.prop, cv.donut = cv.donut, cv.buffer = cv.buffer,
                min.T0 = min.T0, k = k, criterion = criterion,
                parallel = parallel, cores = cores,
                do_parallel_cv = do_parallel_cv
            )
            return(out)
        }

        ## ---- cv.method → cv.treat mapping (after nevertreated delegation) ---- ##
        cv.method <- .fect_normalize_cv_method(
            cv.method,
            allowed = c("rolling", "block", "all_units", "treated_units")
        )
        cv.treat <- (cv.method == "treated_units")
        use_rolling <- (cv.method == "rolling")

        ## ----- ##
        ## ------------- initialize ------------ ##
        ## ----- ##

        cv.pos <- which(t.on <= 0)
        t.on.cv <- t.on[cv.pos]
        count.on.cv <- as.numeric(table(t.on.cv))
        ## tot.id <- which(c(II)==1) ## observed control data
        ## cv.count <- ceiling((sum(II)*sum(II))/sum(I))
        rm.count <- floor(sum(II) * cv.prop)
        cv.count <- sum(II) - rm.count

        # ociCV <- matrix(NA, cv.count, k) ## store indicator
        # rmCV <- matrix(NA, rm.count, k) ## removed indicator
        ociCV <- list()
        rmCV <- list()
        estCV <- NULL ## used for mspe
        if (use_weight == 1) {
            W.rmCV <- list()
        }

        Y0CV <- array(NA, dim = c(TT, N, k)) ## store initial Y0
        if (p > 0) {
            beta0CV <- array(NA, dim = c(p, 1, k))
        } else {
            beta0CV <- array(0, dim = c(1, 0, k)) ## store initial beta0
        }

        ## ---- rolling-window pre-computation (cv.method = "rolling") ---- ##
        ## When rolling, masks are constructed via .build_cv_mask_rolling
        ## (per-fold sampling of cv.prop of eligible units; per-unit random
        ## anchor; cv.nobs scored holdout, cv.buffer past-side buffer,
        ## drop-future as the rolling-window step). The con1/con2
        ## block-CV feasibility checks are skipped --- rolling preserves
        ## per-time donor coverage by construction via per-fold unit
        ## sampling (unsampled units stay fully observed).
        rolling_folds <- NULL
        if (use_rolling) {
            rolling_folds <- .build_cv_mask_rolling(
                II = II, D = D, k = k,
                cv.nobs = cv.nobs, cv.buffer = cv.buffer,
                cv.prop = cv.prop, min.T0 = min.T0, seed = NULL
            )
        }

        ## cv.id.all <- c()
        flag <- 0
        for (i in 1:k) {
            if (use_rolling) {
                cv.n <- 0
                cv.id <- rolling_folds[[i]]$cv.id
                est.id <- rolling_folds[[i]]$est.id
            } else {
                cv.n <- 0
                repeat{
                    cv.n <- cv.n + 1
                    # cv.id <- cv.sample(II, as.integer(sum(II) - cv.count))
                    get.cv <- cv.sample(II, D,
                        count = rm.count,
                        cv.count = cv.nobs,
                        cv.treat = cv.treat,
                        cv.donut = cv.donut
                    )
                    cv.id <- get.cv$cv.id
                    ## cv.id <- sample(oci, as.integer(sum(II) - cv.count), replace = FALSE)
                    II.cv.valid <- II.cv <- II
                    II.cv[cv.id] <- 0
                    II.cv.valid[cv.id] <- -1
                    ## ziyi: if certain rows or columns doesn't satisfy con1 or con2,
                    ## replace the row or column of II.cv using the corresponding rows or columns in II

                    con1 <- sum(apply(II.cv, 1, sum) >= 1) == TT
                    con2 <- sum(apply(II.cv, 2, sum) >= min.T0) == N

                    if (con1 == TRUE & con2 == TRUE) {
                        break
                    }

                    if (cv.n >= 200) {
                        flag <- 1
                        # message("Some units have too few pre-treatment observations. Remove them automatically in Cross-Validation.")
                        keep.1 <- which(apply(II.cv, 1, sum) < 1)
                        keep.2 <- which(apply(II.cv, 2, sum) < min.T0)
                        II.cv[keep.1, ] <- II[keep.1, ]
                        II.cv[, keep.2] <- II[, keep.2]
                        II.cv.valid[keep.1, ] <- II[keep.1, ]
                        II.cv.valid[, keep.2] <- II[, keep.2]
                        cv.id <- which(II.cv.valid != II)
                        break
                    }
                }
            }

            if (length(cv.id) == 0) {
                stop("Some units have too few pre-treatment observations. Set a larger \"cv.prop\" or set cv.method to \"all_units\".")
            }

            rmCV[[i]] <- cv.id
            ocicv <- setdiff(oci, cv.id)
            ociCV[[i]] <- ocicv
            if (use_weight) {
                W.estCV <- list()
            }

            if (use_rolling) {
                estCV <- c(estCV, list(est.id))
            } else if (cv.n < 200) {
                estCV <- c(estCV, list(get.cv$est.id))
            } else {
                cv.id.old <- get.cv$cv.id
                cv.diff <- setdiff(cv.id.old, cv.id)
                estCV <- c(estCV, list(setdiff(get.cv$est.id, cv.diff)))
            }

            if (use_weight == 0) {
                initialOutCv <- initialFit(
                    data = data.ini,
                    force = force,
                    oci = ocicv
                )
            } else {
                initialOutCv <- initialFit(
                    data = data.ini,
                    force = force,
                    w = c(W),
                    oci = ocicv
                )
            }

            Y0CV[, , i] <- initialOutCv$Y0

            if (p > 0) {
                beta0cv <- initialOutCv$beta0
                if (sum(is.na(beta0cv)) > 0) {
                    beta0cv[which(is.na(beta0cv))] <- 0
                }
                beta0CV[, , i] <- beta0cv
            }
        }

        if (flag == 1) {
            message("Some units have too few pre-treatment observations. Remove them automatically in Cross-Validation.\n")
        }

        ## get count matrix
        if (use_weight == 0) {
            count.T.cv <- count.T.cv.old <- table(T.on)
            count.T.cv.old <- count.T.cv <- count.T.cv[which(as.numeric(names(count.T.cv)) <= 0)]
            cv.prop.cut <- max(count.T.cv.old) * proportion
            cv.drop.index <- which(count.T.cv.old <= cv.prop.cut)

            count.T.cv <- count.T.cv / mean(count.T.cv)
            name.count.T.cv <- names(count.T.cv)
            count.T.cv <- c(count.T.cv, median(count.T.cv))
            names(count.T.cv) <- c(name.count.T.cv, "Control")
            count.T.cv[cv.drop.index] <- 0 # set weights to 0 for the periods when the number of treated observations is less than proportion
        }
        if (use_weight == 1) {
            count.T.cv <- count.T.cv.old <- aggregate(c(W), by = list(relative = c(T.on)), FUN = sum)
            count.T.cv.old <- count.T.cv <- count.T.cv[which(count.T.cv[, "relative"] <= 0), ]
            cv.prop.cut <- max(count.T.cv.old[, 2]) * proportion
            cv.drop.index <- which(count.T.cv.old[, 2] <= cv.prop.cut)

            name.count.T.cv <- count.T.cv[, 1]
            count.T.cv <- count.T.cv[, 2] / mean(count.T.cv[, 2])
            count.T.cv <- c(count.T.cv, median(count.T.cv))
            names(count.T.cv) <- c(name.count.T.cv, "Control")
            count.T.cv[cv.drop.index] <- 0
        }


        ##  --------------------------------------------- ##
        ##  ---------------- cross validation for ife model ------------------  ##
        ##  --------------------------------------------- ##

        if (method %in% c("ife", "both")) {
            message("Interactive fixed effects model...\n")

            r.pc <- est.pc.best <- MSPE.best <- WMSPE.best <- MSPE.pc.best <- NULL
            gmoment.best <- moment.best <- MAD.best <- GMSPE.best <- WGMSPE.best <- NULL

            if (criterion == "PC") {
                CV.out.ife <- matrix(NA, (r.max - r.old + 1), 6)
                colnames(CV.out.ife) <- c("r", "sigma2", "IC", "PC", "MSPTATT", "MSE")
            } else {
                CV.out.ife <- matrix(NA, (r.max - r.old + 1), 13)
                colnames(CV.out.ife) <- c(
                    "r", "sigma2", "IC", "PC",
                    "MSPE", "WMSPE", "GMSPE", "WGMSPE", "MAD", "Moment", "GMoment", "MSPTATT", "MSE"
                )
            }

            CV.out.ife[, "r"] <- c(r.old:r.max)
            CV.out.ife[, "PC"] <- CV.out.ife[, "GMoment"] <- CV.out.ife[, "Moment"] <- CV.out.ife[, "MAD"] <- CV.out.ife[, "MSPE"] <- CV.out.ife[, "WMSPE"] <- CV.out.ife[, "GMSPE"] <- CV.out.ife[, "WGMSPE"] <- 1e20

            ## Per-fold SE matrix parallel to CV.out.ife. Populated below in
            ## both parallel and serial branches; consumed at the end of the
            ## IFE CV block to apply `cv.rule` (see `.fect_apply_cv_rule`).
            ## Rows correspond to r values (same indexing as CV.out.ife);
            ## columns are criterion-specific SEs across folds. NA = not
            ## computable (e.g., k = 1 or only one finite fold score).
            CV.out.ife.se <- matrix(NA_real_, nrow(CV.out.ife), ncol(CV.out.ife))
            colnames(CV.out.ife.se) <- colnames(CV.out.ife)
            CV.out.ife.se[, "r"] <- CV.out.ife[, "r"]

            ## ---- Parallel CV setup (IFE, notyettreated) ---- ##
            ## do_parallel_cv is TRUE when user said parallel=TRUE (auto mode) or parallel="cv" (explicit override).
            ## When parallel=TRUE (auto), threshold gates; when parallel="cv" (explicit), threshold is bypassed.
            ife_size <- Nco * TT
            ife_threshold_met <- ife_size > .CV_PARALLEL_THRESH$ife
            ## "cv" in as.character(parallel) but NOT isTRUE(parallel) → explicit override
            use_explicit_cv <- "cv" %in% as.character(parallel) && !isTRUE(parallel)
            cv_ife_parallel <- do_parallel_cv && (ife_threshold_met || use_explicit_cv)

            if (cv_ife_parallel && k > 1 && criterion != "pc") {
                if (is.null(cores)) {
                    cores <- max(1L, min(parallelly::availableCores(omit = 2L), 8L))
                }
                old.future.plan.cv <- future::plan()
                on.exit(future::plan(old.future.plan.cv), add = TRUE)
                future::plan(future::cluster, workers = .fect_make_future_cluster(cores))
                avail <- parallelly::availableCores()
                msg_line <- sprintf("Parallel CV: using %d of %d available cores.", cores, avail)
                pad <- strrep(" ", max(0, 56 - nchar(msg_line)))
                message("\n",
                    " +----------------------------------------------------------+\n",
                    " | ", msg_line, pad, " |\n",
                    " |                                                          |\n",
                    " | To change: set cores = <n> in fect().                    |\n",
                    " | Default: min(available - 2, 8).                          |\n",
                    " +----------------------------------------------------------+\n")

                ## Build flat r×k task list
                r_seq <- CV.out.ife[, "r"]
                tasks <- vector("list", length(r_seq) * k)
                idx <- 1L
                for (ri in seq_along(r_seq)) {
                    for (ii in 1:k) {
                        tasks[[idx]] <- list(r = r_seq[ri], ii = ii, ri = ri)
                        idx <- idx + 1L
                    }
                }

                ## Capture internal helper in closure so future workers can see it.
                ## Internal (dot-prefix) functions are not exported and may not be found
                ## by workers via package namespace lookup. Explicit capture ensures
                ## the function body is serialized with the task closure.
                .score_fn_ife <- .fect_cv_score_one_ife_all

                ## --- Flat r×k parallel dispatch ---
                fold_scores <- future.apply::future_lapply(
                    tasks,
                    FUN = function(task) {
                        .score_fn_ife(
                            ii       = task$ii,
                            r        = task$r,
                            YY       = YY,
                            Y0CV     = Y0CV,
                            X        = X,
                            II       = II,
                            W.use    = W.use,
                            WW       = if (use_weight == 1) WW else NULL,
                            beta0CV  = beta0CV,
                            rmCV     = rmCV,
                            estCV    = estCV,
                            T.on     = T.on,
                            force    = force,
                            cv_tol   = cv_tol,
                            max.iteration = max.iteration,
                            use_weight    = use_weight
                        )
                    },
                    future.seed = TRUE,
                    future.packages = "fect"
                )

                ## --- Aggregate per-r MSPE sequentially; apply 1% rule in master ---
                ## (1-SE rule applied at the end via .fect_apply_cv_rule.)
                n_r <- length(r_seq)
                for (i in seq_len(n_r)) {
                    r <- r_seq[i]
                    task_idx <- which(vapply(tasks, function(t) t$ri == i, logical(1)))

                    ## Both pooled scores (back-compat MSPE values) and per-fold
                    ## scores (for SE → 1-SE rule) are computed in one pass.
                    agg <- .fect_cv_aggregate_folds(
                        fold_list   = fold_scores[task_idx],
                        count.T.cv  = count.T.cv,
                        use_weight  = use_weight,
                        norm.para   = NULL
                    )
                    scores <- agg$pooled
                    MSPE    <- scores["MSPE"]
                    WMSPE   <- scores["WMSPE"]
                    GMSPE   <- scores["GMSPE"]
                    WGMSPE  <- scores["WGMSPE"]
                    MAD     <- scores["MAD"]
                    moment  <- scores["Moment"]
                    gmoment <- scores["GMoment"]

                    ## Store per-fold SEs in CV.out.ife.se
                    se_v <- agg$se
                    if (!is.null(norm.para)) {
                        se_v[c("MSPE","WMSPE","GMSPE","WGMSPE","MAD","Moment","GMoment")] <-
                            se_v[c("MSPE","WMSPE","GMSPE","WGMSPE","MAD","Moment","GMoment")] *
                            (norm.para[1]^2)
                    }
                    for (cn in c("MSPE","WMSPE","GMSPE","WGMSPE","MAD","Moment","GMoment")) {
                        if (cn %in% colnames(CV.out.ife.se)) {
                            CV.out.ife.se[i, cn] <- se_v[cn]
                        }
                    }

                    ## Full-data fit for IC/PC/sigma2 — sequential in master
                    est.cv <- inter_fe_ub(YY, Y0, X, II, W.use, beta0, r, force, cv_tol, max.iteration)
                    sigma2 <- est.cv$sigma2
                    IC <- est.cv$IC
                    PC <- est.cv$PC

                    eff.v.cv <- c(Y - est.cv$fit)[cv.pos]
                    meff <- as.numeric(tapply(eff.v.cv, t.on.cv, mean))
                    MSPTATT <- sum(meff^2 * count.on.cv) / sum(count.on.cv)
                    MSE <- sum(eff.v.cv^2) / length(eff.v.cv)

                    if (!is.null(norm.para)) {
                        MSPE    <- MSPE    * (norm.para[1]^2)
                        WMSPE   <- WMSPE   * (norm.para[1]^2)
                        GMSPE   <- GMSPE   * (norm.para[1]^2)
                        WGMSPE  <- WGMSPE  * (norm.para[1]^2)
                        MAD     <- MAD     * (norm.para[1]^2)
                        moment  <- moment  * (norm.para[1]^2)
                        gmoment <- gmoment * (norm.para[1]^2)
                        sigma2 <- sigma2 * (norm.para[1]^2)
                        IC <- est.cv$IC - log(est.cv$sigma2) + log(sigma2)
                        PC <- PC * (norm.para[1]^2)
                    }

                    if (criterion == "mspe") {
                        if ((min(CV.out.ife[, "MSPE"]) - MSPE) > 0.01 * min(CV.out.ife[, "MSPE"])) {
                            MSPE.best <- MSPE; est.best <- est.cv; r.cv <- r
                        } else {
                            if (r == r.cv + 1) message("*")
                        }
                    } else if (criterion == "wmspe") {
                        if ((min(CV.out.ife[, "WMSPE"]) - WMSPE) > 0.01 * min(CV.out.ife[, "WMSPE"])) {
                            WMSPE.best <- WMSPE; est.best <- est.cv; r.cv <- r
                        } else {
                            if (r == r.cv + 1) message("*")
                        }
                    } else if (criterion == "gmspe") {
                        if ((min(CV.out.ife[, "GMSPE"]) - GMSPE) > 0.01 * min(CV.out.ife[, "GMSPE"])) {
                            GMSPE.best <- GMSPE; est.best <- est.cv; r.cv <- r
                        } else {
                            if (r == r.cv + 1) message("*")
                        }
                    } else if (criterion == "wgmspe") {
                        if ((min(CV.out.ife[, "WGMSPE"]) - WGMSPE) > 0.01 * min(CV.out.ife[, "WGMSPE"])) {
                            WGMSPE.best <- WGMSPE; est.best <- est.cv; r.cv <- r
                        } else {
                            if (r == r.cv + 1) message("*")
                        }
                    } else if (criterion == "mad") {
                        if ((min(CV.out.ife[, "MAD"]) - MAD) > 0.01 * min(CV.out.ife[, "MAD"])) {
                            MAD.best <- MAD; est.best <- est.cv; r.cv <- r
                        } else {
                            if (r == r.cv + 1) message("*")
                        }
                    } else if (criterion == "moment") {
                        if ((min(CV.out.ife[, "Moment"]) - moment) > 0.01 * min(CV.out.ife[, "Moment"])) {
                            moment.best <- moment; est.best <- est.cv; r.cv <- r
                        } else {
                            if (r == r.cv + 1) message("*")
                        }
                    } else if (criterion == "gmoment") {
                        if ((min(CV.out.ife[, "GMoment"]) - gmoment) > 0.01 * min(CV.out.ife[, "GMoment"])) {
                            gmoment.best <- gmoment; est.best <- est.cv; r.cv <- r
                        } else {
                            if (r == r.cv + 1) message("*")
                        }
                    }

                    CV.out.ife[i, 2:13] <- c(sigma2, IC, PC, MSPE, WMSPE, GMSPE, WGMSPE, MAD, moment, gmoment, MSPTATT, MSE)
                    message(
                        "r = ", r, "; sigma2 = ",
                        sprintf("%.5f", sigma2), "; IC = ",
                        sprintf("%.5f", IC), "; PC = ",
                        sprintf("%.5f", PC), "; MSPE = ",
                        sprintf("%.5f", MSPE)
                    )
                } ## end parallel aggregate loop
                ## Ensure r.cv carries the name "r" to match the serial path's
                ## CV.out.ife[i, "r"] extraction which preserves the column name.
                names(r.cv) <- "r"

            } else {
            ## ---- Serial path (original code) ---- ##
            for (i in 1:dim(CV.out.ife)[1]) { ## cross-validation loop starts
                ## inter FE based on control, before & after
                r <- CV.out.ife[i, "r"]
                ## k <- 5
                if (criterion %in% c("mspe", "wmspe", "gmspe", "wgmspe", "mad", "moment", "gmoment")) {
                    fold_list <- vector("list", k)
                    for (ii in 1:k) {
                        fold_list[[ii]] <- .fect_cv_score_one_ife_all(
                            ii       = ii,
                            r        = r,
                            YY       = YY,
                            Y0CV     = Y0CV,
                            X        = X,
                            II       = II,
                            W.use    = W.use,
                            WW       = if (use_weight == 1) WW else NULL,
                            beta0CV  = beta0CV,
                            rmCV     = rmCV,
                            estCV    = estCV,
                            T.on     = T.on,
                            force    = force,
                            cv_tol   = cv_tol,
                            max.iteration = max.iteration,
                            use_weight    = use_weight
                        )
                    }
                    agg <- .fect_cv_aggregate_folds(
                        fold_list   = fold_list,
                        count.T.cv  = count.T.cv,
                        use_weight  = use_weight,
                        norm.para   = NULL
                    )
                    scores <- agg$pooled
                    MSPE    <- scores["MSPE"]
                    WMSPE   <- scores["WMSPE"]
                    GMSPE   <- scores["GMSPE"]
                    WGMSPE  <- scores["WGMSPE"]
                    MAD     <- scores["MAD"]
                    moment  <- scores["Moment"]
                    gmoment <- scores["GMoment"]

                    ## Store per-fold SEs in CV.out.ife.se for end-of-loop rule
                    se_v <- agg$se
                    if (!is.null(norm.para)) {
                        se_v[c("MSPE","WMSPE","GMSPE","WGMSPE","MAD","Moment","GMoment")] <-
                            se_v[c("MSPE","WMSPE","GMSPE","WGMSPE","MAD","Moment","GMoment")] *
                            (norm.para[1]^2)
                    }
                    for (cn in c("MSPE","WMSPE","GMSPE","WGMSPE","MAD","Moment","GMoment")) {
                        if (cn %in% colnames(CV.out.ife.se)) {
                            CV.out.ife.se[i, cn] <- se_v[cn]
                        }
                    }
                }

                est.cv <- inter_fe_ub(
                    YY,
                    Y0,
                    X,
                    II,
                    W.use,
                    beta0,
                    r,
                    force,
                    cv_tol,
                    max.iteration
                ) ## overall
                sigma2 <- est.cv$sigma2
                IC <- est.cv$IC
                PC <- est.cv$PC

                eff.v.cv <- c(Y - est.cv$fit)[cv.pos]
                meff <- as.numeric(tapply(eff.v.cv, t.on.cv, mean))
                MSPTATT <- sum(meff^2 * count.on.cv) / sum(count.on.cv)
                MSE <- sum(eff.v.cv^2) / length(eff.v.cv)

                if (!is.null(norm.para)) {
                    if (criterion %in% c("mspe", "wmspe", "gmspe", "wgmspe", "mad", "moment", "gmoment")) {
                        MSPE <- MSPE * (norm.para[1]^2)
                        WMSPE <- WMSPE * (norm.para[1]^2)
                        GMSPE <- GMSPE * (norm.para[1]^2)
                        WGMSPE <- WGMSPE * (norm.para[1]^2)
                        MAD <- MAD * (norm.para[1]^2)
                        moment <- moment * (norm.para[1]^2)
                        gmoment <- gmoment * (norm.para[1]^2)
                    }
                    sigma2 <- sigma2 * (norm.para[1]^2)
                    IC <- est.cv$IC - log(est.cv$sigma2) + log(sigma2)
                    PC <- PC * (norm.para[1]^2)
                }

                if (criterion == "mspe") {
                    if ((min(CV.out.ife[, "MSPE"]) - MSPE) > 0.01 * min(CV.out.ife[, "MSPE"])) {
                        ## at least 1% improvement for MPSE
                        MSPE.best <- MSPE
                        est.best <- est.cv
                        r.cv <- r
                    } else {
                        if (r == r.cv + 1) message("*")
                    }
                } else if (criterion == "wmspe") {
                    if ((min(CV.out.ife[, "WMSPE"]) - WMSPE) > 0.01 * min(CV.out.ife[, "WMSPE"])) {
                        ## at least 1% improvement for MPSE
                        WMSPE.best <- WMSPE
                        est.best <- est.cv
                        r.cv <- r
                    } else {
                        if (r == r.cv + 1) message("*")
                    }
                } else if (criterion == "gmspe") {
                    if ((min(CV.out.ife[, "GMSPE"]) - GMSPE) > 0.01 * min(CV.out.ife[, "GMSPE"])) {
                        ## at least 1% improvement for MPSE
                        GMSPE.best <- GMSPE
                        est.best <- est.cv
                        r.cv <- r
                    } else {
                        if (r == r.cv + 1) message("*")
                    }
                } else if (criterion == "wgmspe") {
                    if ((min(CV.out.ife[, "WGMSPE"]) - WGMSPE) > 0.01 * min(CV.out.ife[, "WGMSPE"])) {
                        ## at least 1% improvement for MPSE
                        WGMSPE.best <- WGMSPE
                        est.best <- est.cv
                        r.cv <- r
                    } else {
                        if (r == r.cv + 1) message("*")
                    }
                } else if (criterion == "mad") {
                    if ((min(CV.out.ife[, "MAD"]) - MAD) > 0.01 * min(CV.out.ife[, "MAD"])) {
                        MAD.best <- MAD
                        est.best <- est.cv
                        r.cv <- r
                    } else {
                        if (r == r.cv + 1) message("*")
                    }
                } else if (criterion == "moment") {
                    if ((min(CV.out.ife[, "Moment"]) - moment) > 0.01 * min(CV.out.ife[, "Moment"])) {
                        moment.best <- moment
                        est.best <- est.cv
                        r.cv <- r
                    } else {
                        if (r == r.cv + 1) message("*")
                    }
                } else if (criterion == "gmoment") {
                    if ((min(CV.out.ife[, "GMoment"]) - gmoment) > 0.01 * min(CV.out.ife[, "GMoment"])) {
                        gmoment.best <- gmoment
                        est.best <- est.cv
                        r.cv <- r
                    } else {
                        if (r == r.cv + 1) message("*")
                    }
                } else if (criterion == "pc") {
                    if (PC < min(CV.out.ife[, "PC"])) {
                        est.pc.best <- est.cv
                        r.pc <- r
                    }
                }

                if (criterion != "pc") {
                    CV.out.ife[i, 2:13] <- c(sigma2, IC, PC, MSPE, WMSPE, GMSPE, WGMSPE, MAD, moment, gmoment, MSPTATT, MSE)
                } else {
                    CV.out.ife[i, 2:6] <- c(sigma2, IC, PC, MSPTATT, MSE)
                }

                if (criterion == "pc") {
                    message("r = ", r, "; sigma2 = ",
                        sprintf("%.5f", sigma2), "; IC = ",
                        sprintf("%.5f", IC), "; PC = ",
                        sprintf("%.5f", PC), "; MSPTATT = ",
                        sprintf("%.5f", MSPTATT), "; MSE = ",
                        sprintf("%.5f", MSE),
                        sep = ""
                    )
                } else {
                    # message("r = ",r, "; sigma2 = ",
                    #    sprintf("%.5f",sigma2), "; IC = ",
                    #    sprintf("%.5f",IC), "; PC = ",
                    #    sprintf("%.5f",PC), "; MSPE = ",
                    #    sprintf("%.5f",MSPE), "; GMSPE = ",
                    #    sprintf("%.5f",GMSPE), "; Moment = ",
                    #    sprintf("%.5f",moment), "; MSPTATT = ",
                    #    sprintf("%.5f",MSPTATT), "; MSE = ",
                    #    sprintf("%.5f",MSE), sep="")
                    message(
                        "r = ", r, "; sigma2 = ",
                        sprintf("%.5f", sigma2), "; IC = ",
                        sprintf("%.5f", IC), "; PC = ",
                        sprintf("%.5f", PC), "; MSPE = ",
                        sprintf("%.5f", MSPE)
                    )
                }
            } ## end of while: search for r_star over
            } ## end of serial/parallel if-else block

            # MSPE.best <- min(CV.out[,"MSPE"])
            # PC.best <- min(CV.out[,"PC"])

            ## --- Apply cv.rule (added v2.3.0) -----------------------------
            ## The in-loop assignments above use the legacy 1% rule. Override
            ## r.cv here based on the user-selected rule, using full CV.out.ife
            ## and CV.out.ife.se. Only applies to MSPE-family criteria; the
            ## "pc" branch keeps its own argmin.
            crit_col_map <- c(mspe = "MSPE", wmspe = "WMSPE", gmspe = "GMSPE",
                              wgmspe = "WGMSPE", mad = "MAD",
                              moment = "Moment", gmoment = "GMoment")
            if (criterion %in% names(crit_col_map)) {
                crit_col <- crit_col_map[[criterion]]
                means <- CV.out.ife[, crit_col]
                ses   <- CV.out.ife.se[, crit_col]
                ## Treat the loop sentinel value (1e20) as missing for selection.
                means[!is.finite(means) | means >= 1e19] <- NA_real_
                i_pick <- .fect_apply_cv_rule(means, ses, rule = cv.rule)
                if (!is.na(i_pick) && i_pick >= 1L && i_pick <= nrow(CV.out.ife)) {
                    new_r_cv <- unname(CV.out.ife[i_pick, "r"])
                    if (!is.null(new_r_cv) && is.finite(new_r_cv)) {
                        if (new_r_cv != as.integer(unname(r.cv))) {
                            message(sprintf(
                                "  [cv.rule = %s] r.cv adjusted from %d to %d (1-SE band)\n",
                                cv.rule,
                                as.integer(unname(r.cv)),
                                as.integer(new_r_cv)
                            ))
                        }
                        ## Preserve the prior branch's names convention: the
                        ## parallel branch sets names(r.cv) <- "r"; the serial
                        ## branch leaves it unnamed. Carry forward whichever
                        ## the loop produced rather than overwriting.
                        had_name <- !is.null(names(r.cv))
                        r.cv <- new_r_cv
                        if (had_name) names(r.cv) <- "r"
                        ## Refit at the chosen r so est.best matches r.cv.
                        est.best <- inter_fe_ub(
                            YY, Y0, X, II, W.use, beta0,
                            as.integer(unname(r.cv)), force, cv_tol, max.iteration
                        )
                        if (!is.null(scores <- CV.out.ife[i_pick, ])) {
                            MSPE.best   <- scores["MSPE"]
                            WMSPE.best  <- scores["WMSPE"]
                            GMSPE.best  <- scores["GMSPE"]
                            WGMSPE.best <- scores["WGMSPE"]
                            MAD.best    <- scores["MAD"]
                            moment.best <- scores["Moment"]
                            gmoment.best <- scores["GMoment"]
                        }
                    }
                }
            }
            ## --------------------------------------------------------------

            ## compare
            if (criterion == "both") {
                if (r.cv > r.pc) {
                    message("\n Factor number selected via cross validation may be larger than the true number. Using the PC criterion.\n\n ")
                    r.cv <- r.pc
                    est.best <- est.pc.best
                    MSPE.best <- MSPE.pc.best
                }
                est.best.ife <- est.best
                MSPE.best.ife <- MSPE.best
            } else if (criterion == "pc") {
                est.best.ife <- est.pc.best
                r.cv <- r.pc
            } else {
                est.best.ife <- est.best
                MSPE.best.ife <- MSPE.best
                WMSPE.best.ife <- WMSPE.best
                GMSPE.best.ife <- GMSPE.best
                WGMSPE.best.ife <- WGMSPE.best
                MAD.best.ife <- MAD.best
                moment.best.ife <- moment.best
                gmoment.best.ife <- gmoment.best
            }

            if (r > (TT - 1)) {
                message(" (r hits maximum)")
            }
            message("\n r* = ", r.cv, sep = "")
        }

        ##  ------------------------------------- ##
        ##  ---------------- cross validation for mc ---------------  ##
        ##  ------------------------------------- ##
        if (method %in% c("mc", "both")) {
            message("Matrix completion method...\n")
            eigen.all <- NULL
            if (is.null(lambda) || length(lambda) == 1) {
                ## create the hyper-parameter sequence
                ## biggest candidate lambda
                ## Y.lambda <- YY
                Y.lambda <- YY - Y0
                ## Y.lambda[which(II == 0)] <- Y0[which(II == 0)]
                Y.lambda[which(II == 0)] <- 0
                if (use_weight) {
                    Y.lambda <- Y.lambda * W
                }
                eigen.all <- svd(Y.lambda / (TT * N))$d
                lambda.max <- log10(max(eigen.all))
                lambda <- rep(NA, nlambda)
                lambda.by <- 3 / (nlambda - 2)
                for (i in 1:(nlambda - 1)) {
                    lambda[i] <- 10^(lambda.max - (i - 1) * lambda.by)
                }
                lambda[nlambda] <- 0
            } else {
                Y.lambda <- YY - Y0
                Y.lambda[which(II == 0)] <- 0
                if (use_weight) {
                    Y.lambda <- Y.lambda * W
                }
                eigen.all <- svd(Y.lambda / (TT * N))$d
            }

            ## store all MSPE
            MSPE.best <- WMSPE.best <- GMSPE.best <- WGMSPE.best <- MAD.best <- moment.best <- gmoment.best <- NULL
            CV.out.mc <- matrix(NA, length(lambda), 10)
            colnames(CV.out.mc) <- c("lambda.norm", "MSPE", "WMSPE", "GMSPE", "WGMSPE", "MAD", "Moment", "GMoment", "MSPTATT", "MSE")
            CV.out.mc[, "lambda.norm"] <- c(lambda / max(eigen.all))

            ## Per-fold SE matrix parallel to CV.out.mc (added v2.3.0).
            ## Same role as CV.out.ife.se: store per-fold SE per criterion so
            ## the end-of-MC-block `cv.rule` override can apply 1-SE selection.
            CV.out.mc.se <- matrix(NA_real_, length(lambda), 10)
            colnames(CV.out.mc.se) <- colnames(CV.out.mc)
            CV.out.mc.se[, "lambda.norm"] <- CV.out.mc[, "lambda.norm"]
            CV.out.mc[, "GMoment"] <- CV.out.mc[, "Moment"] <- CV.out.mc[, "MAD"] <- CV.out.mc[, "WGMSPE"] <- CV.out.mc[, "WGMSPE"] <- CV.out.mc[, "GMSPE"] <- CV.out.mc[, "WMSPE"] <- CV.out.mc[, "MSPE"] <- 1e20

            ## ---- Parallel CV setup (MC, notyettreated) ---- ##
            ## do_parallel_cv is TRUE when user said parallel=TRUE (auto mode) or parallel="cv" (explicit override).
            ## Threshold only governs auto mode: explicit "cv" bypasses it.
            mc_size <- Nco * TT
            mc_threshold_met <- mc_size > .CV_PARALLEL_THRESH$mc
            ## Re-derive use_explicit_cv for the MC block. When method == "mc" only, the IFE block
            ## did not run and use_explicit_cv (from the IFE block) is not in scope. Always use
            ## use_explicit_cv_mc here — a locally-computed copy — to avoid R scoping issues.
            use_explicit_cv_mc <- "cv" %in% as.character(parallel) && !isTRUE(parallel)
            cv_mc_parallel <- do_parallel_cv && (mc_threshold_met || use_explicit_cv_mc)

            if (cv_mc_parallel && k > 1) {
                if (is.null(cores)) {
                    cores <- max(1L, min(parallelly::availableCores(omit = 2L), 8L))
                }
                old.future.plan.mc <- future::plan()
                ## Use after = FALSE to prepend this restore so it fires BEFORE the IFE
                ## restore when method = "both" (both blocks register on.exit in the same
                ## fect_cv call; the IFE block registers first so it fires last by default).
                ## With after = FALSE, MC restore fires first (restoring to the IFE-activated
                ## multisession plan), then IFE restore fires (restoring to the caller's
                ## original plan). Net: caller's plan correctly preserved.
                on.exit(future::plan(old.future.plan.mc), add = TRUE, after = FALSE)
                future::plan(future::cluster, workers = .fect_make_future_cluster(cores))
                avail <- parallelly::availableCores()
                msg_line <- sprintf("Parallel CV (MC): using %d of %d available cores.", cores, avail)
                pad <- strrep(" ", max(0, 56 - nchar(msg_line)))
                message("\n",
                    " +----------------------------------------------------------+\n",
                    " | ", msg_line, pad, " |\n",
                    " |                                                          |\n",
                    " | To change: set cores = <n> in fect().                    |\n",
                    " | Default: min(available - 2, 8).                          |\n",
                    " +----------------------------------------------------------+\n")

                ## Build flat lambda×k task list
                tasks <- vector("list", length(lambda) * k)
                idx <- 1L
                for (li in seq_along(lambda)) {
                    for (ii in 1:k) {
                        tasks[[idx]] <- list(lambda_i = lambda[li], ii = ii, li = li)
                        idx <- idx + 1L
                    }
                }

                ## Capture MC helper in closure for worker serialization.
                ## Internal (dot-prefix) helpers are not exported; closure capture ensures
                ## the function body is serialized with the task, same pattern as Phase 1 IFE.
                .score_fn_mc <- .fect_cv_score_one_mc_all

                ## --- Flat lambda×k parallel dispatch ---
                fold_scores <- future.apply::future_lapply(
                    tasks,
                    FUN = function(task) {
                        .score_fn_mc(
                            ii         = task$ii,
                            lambda_i   = task$lambda_i,
                            YY         = YY,
                            Y0CV       = Y0CV,
                            X          = X,
                            II         = II,
                            W.use      = W.use,
                            WW         = if (use_weight == 1) WW else NULL,
                            beta0CV    = beta0CV,
                            rmCV       = rmCV,
                            estCV      = estCV,
                            T.on       = T.on,
                            force      = force,
                            cv_tol     = cv_tol,
                            max.iteration = max.iteration,
                            use_weight    = use_weight
                        )
                    },
                    future.seed = TRUE,
                    future.packages = "fect"
                )

                ## --- Aggregate per-lambda MSPE sequentially; apply 1% rule in master ---
                ## (1-SE rule applied at the end of the MC block via .fect_apply_cv_rule.)
                ## break_check is NOT applied in parallel mode — all lambdas are computed.
                for (i in seq_along(lambda)) {
                    task_idx <- which(vapply(tasks, function(t) t$li == i, logical(1)))

                    ## Per-fold + pooled aggregation (added v2.3.0).
                    agg <- .fect_cv_aggregate_folds(
                        fold_list   = fold_scores[task_idx],
                        count.T.cv  = count.T.cv,
                        use_weight  = use_weight,
                        norm.para   = NULL
                    )
                    scores <- agg$pooled
                    MSPE    <- scores["MSPE"]
                    WMSPE   <- scores["WMSPE"]
                    GMSPE   <- scores["GMSPE"]
                    WGMSPE  <- scores["WGMSPE"]
                    MAD     <- scores["MAD"]
                    moment  <- scores["Moment"]
                    gmoment <- scores["GMoment"]

                    ## Full-data fit for MSPTATT/MSE — sequential in master (same as serial path)
                    est.cv <- inter_fe_mc(
                        YY, Y0, X, II, W.use, beta0,
                        1, lambda[i],
                        force, cv_tol, max.iteration
                    )

                    eff.v.cv <- c(Y - est.cv$fit)[cv.pos]
                    meff <- as.numeric(tapply(eff.v.cv, t.on.cv, mean))
                    MSPTATT <- sum(meff^2 * count.on.cv) / sum(count.on.cv)
                    MSE <- sum(eff.v.cv^2) / length(eff.v.cv)

                    if (!is.null(norm.para)) {
                        MSPE   <- MSPE   * (norm.para[1]^2)
                        WMSPE  <- WMSPE  * (norm.para[1]^2)
                        GMSPE  <- GMSPE  * (norm.para[1]^2)
                        WGMSPE <- WGMSPE * (norm.para[1]^2)
                        MAD    <- MAD    * (norm.para[1]^2)
                        moment <- moment * (norm.para[1]^2)
                        gmoment <- gmoment * (norm.para[1]^2)
                    }

                    ## Persist per-fold SEs in CV.out.mc.se for end-of-block rule
                    se_v <- agg$se
                    if (!is.null(norm.para)) {
                        se_v[c("MSPE","WMSPE","GMSPE","WGMSPE","MAD","Moment","GMoment")] <-
                            se_v[c("MSPE","WMSPE","GMSPE","WGMSPE","MAD","Moment","GMoment")] *
                            (norm.para[1]^2)
                    }
                    for (cn in c("MSPE","WMSPE","GMSPE","WGMSPE","MAD","Moment","GMoment")) {
                        if (cn %in% colnames(CV.out.mc.se)) {
                            CV.out.mc.se[i, cn] <- se_v[cn]
                        }
                    }

                    ## 1% rule — identical logic to serial path, but no break_check
                    if (criterion == "mspe") {
                        if ((min(CV.out.mc[, "MSPE"]) - MSPE) > 0.01 * min(CV.out.mc[, "MSPE"])) {
                            MSPE.best <- MSPE; est.best <- est.cv; lambda.cv <- lambda[i]
                        }
                    } else if (criterion == "wmspe") {
                        if ((min(CV.out.mc[, "WMSPE"]) - WMSPE) > 0.01 * min(CV.out.mc[, "WMSPE"])) {
                            WMSPE.best <- WMSPE; est.best <- est.cv; lambda.cv <- lambda[i]
                        }
                    } else if (criterion == "gmspe") {
                        if ((min(CV.out.mc[, "GMSPE"]) - GMSPE) > 0.01 * min(CV.out.mc[, "GMSPE"])) {
                            GMSPE.best <- GMSPE; est.best <- est.cv; lambda.cv <- lambda[i]
                        }
                    } else if (criterion == "wgmspe") {
                        if ((min(CV.out.mc[, "WGMSPE"]) - WGMSPE) > 0.01 * min(CV.out.mc[, "WGMSPE"])) {
                            WGMSPE.best <- WGMSPE; est.best <- est.cv; lambda.cv <- lambda[i]
                        }
                    } else if (criterion == "mad") {
                        if ((min(CV.out.mc[, "MAD"]) - MAD) > 0.01 * min(CV.out.mc[, "MAD"])) {
                            MAD.best <- MAD; est.best <- est.cv; lambda.cv <- lambda[i]
                        }
                    } else if (criterion == "moment") {
                        if ((min(CV.out.mc[, "Moment"]) - moment) > 0.01 * min(CV.out.mc[, "Moment"])) {
                            moment.best <- moment; est.best <- est.cv; lambda.cv <- lambda[i]
                        }
                    } else if (criterion == "gmoment") {
                        if ((min(CV.out.mc[, "GMoment"]) - gmoment) > 0.01 * min(CV.out.mc[, "GMoment"])) {
                            gmoment.best <- gmoment; est.best <- est.cv; lambda.cv <- lambda[i]
                        }
                    }

                    CV.out.mc[i, 2:10] <- c(MSPE, WMSPE, GMSPE, WGMSPE, MAD, moment, gmoment, MSPTATT, MSE)
                    message("lambda.norm = ",
                        sprintf("%.5f", lambda[i] / max(eigen.all)), "; MSPE = ",
                        sprintf("%.5f", MSPE), "; MSPTATT = ",
                        sprintf("%.5f", MSPTATT), "; MSE = ",
                        sprintf("%.5f", MSE),
                        sep = ""
                    )
                }
                ## end parallel MC CV

            } else {
            ## ---- Serial path (existing code) ---- ##
            break_count <- 0
            break_check <- 0
            for (i in 1:length(lambda)) {
                ## k <- 5
                fold_list <- vector("list", k)
                for (ii in 1:k) {
                    fold_list[[ii]] <- .fect_cv_score_one_mc_all(
                        ii        = ii,
                        lambda_i  = lambda[i],
                        YY        = YY,
                        Y0CV      = Y0CV,
                        X         = X,
                        II        = II,
                        W.use     = W.use,
                        WW        = if (use_weight == 1) WW else NULL,
                        beta0CV   = beta0CV,
                        rmCV      = rmCV,
                        estCV     = estCV,
                        T.on      = T.on,
                        force     = force,
                        cv_tol    = cv_tol,
                        max.iteration = max.iteration,
                        use_weight    = use_weight
                    )
                }
                ## Per-fold + pooled aggregation (added v2.3.0)
                agg_mc <- .fect_cv_aggregate_folds(
                    fold_list   = fold_list,
                    count.T.cv  = count.T.cv,
                    use_weight  = use_weight,
                    norm.para   = NULL
                )
                scores <- agg_mc$pooled
                MSPE    <- scores["MSPE"]
                WMSPE   <- scores["WMSPE"]
                GMSPE   <- scores["GMSPE"]
                WGMSPE  <- scores["WGMSPE"]
                MAD     <- scores["MAD"]
                moment  <- scores["Moment"]
                gmoment <- scores["GMoment"]

                est.cv <- inter_fe_mc(
                    YY, Y0, X, II, W.use, beta0,
                    1, lambda[i],
                    force, cv_tol, max.iteration
                ) ## overall

                eff.v.cv <- c(Y - est.cv$fit)[cv.pos]
                meff <- as.numeric(tapply(eff.v.cv, t.on.cv, mean))
                MSPTATT <- sum(meff^2 * count.on.cv) / sum(count.on.cv)
                MSE <- sum(eff.v.cv^2) / length(eff.v.cv)

                if (!is.null(norm.para)) {
                    MSPE <- MSPE * (norm.para[1]^2)
                    WMSPE <- WMSPE * (norm.para[1]^2)
                    GMSPE <- GMSPE * (norm.para[1]^2)
                    WGMSPE <- WGMSPE * (norm.para[1]^2)
                    MAD <- MAD * (norm.para[1]^2)
                    moment <- moment * (norm.para[1]^2)
                    gmoment <- gmoment * (norm.para[1]^2)
                }

                ## Persist per-fold SEs in CV.out.mc.se (added v2.3.0)
                {
                    se_v <- agg_mc$se
                    if (!is.null(norm.para)) {
                        se_v[c("MSPE","WMSPE","GMSPE","WGMSPE","MAD","Moment","GMoment")] <-
                            se_v[c("MSPE","WMSPE","GMSPE","WGMSPE","MAD","Moment","GMoment")] *
                            (norm.para[1]^2)
                    }
                    for (cn in c("MSPE","WMSPE","GMSPE","WGMSPE","MAD","Moment","GMoment")) {
                        if (cn %in% colnames(CV.out.mc.se)) {
                            CV.out.mc.se[i, cn] <- se_v[cn]
                        }
                    }
                }

                if (criterion == "mspe") {
                    if ((min(CV.out.mc[, "MSPE"]) - MSPE) > 0.01 * min(CV.out.mc[, "MSPE"])) {
                        ## at least 1% improvement for MPSE
                        MSPE.best <- MSPE
                        est.best <- est.cv
                        lambda.cv <- lambda[i]
                        break_count <- 0
                        break_check <- 0
                    } else {
                        if (i > 1) {
                            if (lambda.cv == lambda[i - 1]) {
                                message("*")
                                break_check <- 1
                                break_count <- 0
                            }
                        }
                    }
                } else if (criterion == "wmspe") {
                    if ((min(CV.out.mc[, "WMSPE"]) - WMSPE) > 0.01 * min(CV.out.mc[, "WMSPE"])) {
                        ## at least 1% improvement for MPSE
                        WMSPE.best <- WMSPE
                        est.best <- est.cv
                        lambda.cv <- lambda[i]
                        break_check <- 0
                        break_count <- 0
                    } else {
                        if (i > 1) {
                            if (lambda.cv == lambda[i - 1]) {
                                message("*")
                                break_check <- 1
                                break_count <- 0
                            }
                        }
                    }
                } else if (criterion == "gmspe") {
                    if ((min(CV.out.mc[, "GMSPE"]) - GMSPE) > 0.01 * min(CV.out.mc[, "GMSPE"])) {
                        ## at least 1% improvement for MPSE
                        GMSPE.best <- GMSPE
                        est.best <- est.cv
                        lambda.cv <- lambda[i]
                        break_check <- 0
                        break_count <- 0
                    } else {
                        if (i > 1) {
                            if (lambda.cv == lambda[i - 1]) {
                                message("*")
                                break_check <- 1
                                break_count <- 0
                            }
                        }
                    }
                } else if (criterion == "wgmspe") {
                    if ((min(CV.out.mc[, "WGMSPE"]) - WGMSPE) > 0.01 * min(CV.out.mc[, "WGMSPE"])) {
                        ## at least 1% improvement for MPSE
                        WGMSPE.best <- WGMSPE
                        est.best <- est.cv
                        lambda.cv <- lambda[i]
                        break_check <- 0
                        break_count <- 0
                    } else {
                        if (i > 1) {
                            if (lambda.cv == lambda[i - 1]) {
                                message("*")
                                break_check <- 1
                                break_count <- 0
                            }
                        }
                    }
                } else if (criterion == "mad") {
                    if ((min(CV.out.mc[, "MAD"]) - MAD) > 0.01 * min(CV.out.mc[, "MAD"])) {
                        ## at least 1% improvement for MPSE
                        MAD.best <- MAD
                        est.best <- est.cv
                        lambda.cv <- lambda[i]
                        break_check <- 0
                        break_count <- 0
                    } else {
                        if (i > 1) {
                            if (lambda.cv == lambda[i - 1]) {
                                message("*")
                                break_check <- 1
                                break_count <- 0
                            }
                        }
                    }
                } else if (criterion == "moment") {
                    if ((min(CV.out.mc[, "Moment"]) - moment) > 0.01 * min(CV.out.mc[, "Moment"])) {
                        ## at least 1% improvement for MPSE
                        moment.best <- moment
                        est.best <- est.cv
                        lambda.cv <- lambda[i]
                        break_check <- 0
                        break_count <- 0
                    } else {
                        if (i > 1) {
                            if (lambda.cv == lambda[i - 1]) {
                                message("*")
                                break_check <- 1
                                break_count <- 0
                            }
                        }
                    }
                } else if (criterion == "gmoment") {
                    if ((min(CV.out.mc[, "GMoment"]) - gmoment) > 0.01 * min(CV.out.mc[, "GMoment"])) {
                        ## at least 1% improvement for MPSE
                        gmoment.best <- gmoment
                        est.best <- est.cv
                        lambda.cv <- lambda[i]
                        break_check <- 0
                        break_count <- 0
                    } else {
                        if (i > 1) {
                            if (lambda.cv == lambda[i - 1]) {
                                message("*")
                                break_check <- 1
                                break_count <- 0
                            }
                        }
                    }
                }

                if (break_check == 1) {
                    break_count <- break_count + 1
                }

                CV.out.mc[i, 2:10] <- c(MSPE, WMSPE, GMSPE, WGMSPE, MAD, moment, gmoment, MSPTATT, MSE)
                message("lambda.norm = ",
                    sprintf("%.5f", lambda[i] / max(eigen.all)), "; MSPE = ",
                    sprintf("%.5f", MSPE), "; MSPTATT = ",
                    sprintf("%.5f", MSPTATT), "; MSE = ",
                    sprintf("%.5f", MSE),
                    sep = ""
                )
                if (break_count == 3) {
                    break
                }
            }
            } ## end else (serial MC CV)

            ## --- Apply cv.rule to lambda selection (added v2.3.0) ----------
            ## Mirrors the IFE end-of-loop override. Applies only to
            ## MSPE-family criteria.
            crit_col_map_mc <- c(mspe = "MSPE", wmspe = "WMSPE", gmspe = "GMSPE",
                                 wgmspe = "WGMSPE", mad = "MAD",
                                 moment = "Moment", gmoment = "GMoment")
            if (criterion %in% names(crit_col_map_mc)) {
                crit_col_mc <- crit_col_map_mc[[criterion]]
                means_mc <- CV.out.mc[, crit_col_mc]
                ses_mc   <- CV.out.mc.se[, crit_col_mc]
                ## Treat NAs (lambdas where break_check skipped) as missing
                means_mc[!is.finite(means_mc)] <- NA_real_
                i_pick_mc <- .fect_apply_cv_rule(means_mc, ses_mc, rule = cv.rule)
                if (!is.na(i_pick_mc) && i_pick_mc >= 1L && i_pick_mc <= length(lambda)) {
                    new_lambda_cv <- lambda[i_pick_mc]
                    if (!identical(new_lambda_cv, lambda.cv)) {
                        message(sprintf(
                            "  [cv.rule = %s] lambda.cv adjusted (1-SE band)",
                            cv.rule
                        ))
                        est.best <- inter_fe_mc(
                            YY, Y0, X, II, W.use, beta0,
                            1, new_lambda_cv,
                            force, cv_tol, max.iteration
                        )
                        lambda.cv <- new_lambda_cv
                    }
                }
            }
            ## --------------------------------------------------------------

            est.best.mc <- est.best
            MSPE.best.mc <- MSPE.best
            WMSPE.best.mc <- WMSPE.best
            GMSPE.best.mc <- GMSPE.best
            WGMSPE.best.mc <- WGMSPE.best
            MAD.best.mc <- MAD.best
            moment.best.mc <- moment.best
            gmoment.best.mc <- gmoment.best
            message("\n lambda.norm* = ", lambda.cv / max(eigen.all), sep = "")
            message("\n")
        }
    } ## End of Cross-Validation


    if (method == "ife") {
        est.best <- est.best.ife
        validF <- ifelse(r.cv > 0, 1, 0)
    } else if (method == "mc") {
        est.best <- est.best.mc
        validF <- est.best$validF
    } else {
        if (criterion == "mspe") {
            if (MSPE.best.ife <= MSPE.best.mc) {
                est.best <- est.best.ife
                validF <- ifelse(r.cv > 0, 1, 0)
                method <- "ife"
            } else {
                est.best <- est.best.mc
                validF <- est.best$validF
                method <- "mc"
            }
        }
        if (criterion == "wmspe") {
            if (WMSPE.best.ife <= WMSPE.best.mc) {
                est.best <- est.best.ife
                validF <- ifelse(r.cv > 0, 1, 0)
                method <- "ife"
            } else {
                est.best <- est.best.mc
                validF <- est.best$validF
                method <- "mc"
            }
        }
        if (criterion == "gmspe") {
            if (GMSPE.best.ife <= GMSPE.best.mc) {
                est.best <- est.best.ife
                validF <- ifelse(r.cv > 0, 1, 0)
                method <- "ife"
            } else {
                est.best <- est.best.mc
                validF <- est.best$validF
                method <- "mc"
            }
        }
        if (criterion == "wgmspe") {
            if (WGMSPE.best.ife <= WGMSPE.best.mc) {
                est.best <- est.best.ife
                validF <- ifelse(r.cv > 0, 1, 0)
                method <- "ife"
            } else {
                est.best <- est.best.mc
                validF <- est.best$validF
                method <- "mc"
            }
        }
        if (criterion == "mad") {
            if (MAD.best.ife <= MAD.best.mc) {
                est.best <- est.best.ife
                validF <- ifelse(r.cv > 0, 1, 0)
                method <- "ife"
            } else {
                est.best <- est.best.mc
                validF <- est.best$validF
                method <- "mc"
            }
        }
        if (criterion == "moment") {
            if (moment.best.ife <= moment.best.mc) {
                est.best <- est.best.ife
                validF <- ifelse(r.cv > 0, 1, 0)
                method <- "ife"
            } else {
                est.best <- est.best.mc
                validF <- est.best$validF
                method <- "mc"
            }
        }
        if (criterion == "gmoment") {
            if (gmoment.best.ife <= gmoment.best.mc) {
                est.best <- est.best.ife
                validF <- ifelse(r.cv > 0, 1, 0)
                method <- "ife"
            } else {
                est.best <- est.best.mc
                validF <- est.best$validF
                method <- "mc"
            }
        }
        message("\n Recommended method through cross-validation: ", method, sep = "")
        message("\n")
    }
    validX <- est.best$validX

    ## ------------------------------##
    ## ----------- Summarize -------------- ##
    ## ------------------------------##

    ## 00. run a fect to obtain residuals
    if (method == "ife") {
        if (r.cv == 0) {
            est.fect <- est.best
        } else {
            est.fect <- inter_fe_ub(YY, Y0, X, II, W.use, beta0, 0, force = force, tol, max.iteration)
        }
    } else {
        est.fect <- inter_fe_ub(YY, Y0, X, II, W.use, beta0, 0, force = force, tol, max.iteration)
    }


    ## -------------------------------##
    ##   ATT and Counterfactuals     ##
    ## -------------------------------##

    ## we first adjustment for normalization
    if (!is.null(norm.para)) {
        Y <- Y * norm.para[1]
        if (method == "ife") {
            ## variance of the error term
            sigma2 <- est.best$sigma2 * (norm.para[1]^2)
            IC <- est.best$IC - log(est.best$sigma2) + log(sigma2)
            PC <- est.best$PC * (norm.para[1]^2)
            est.best$sigma2 <- sigma2
            est.best$IC <- IC
            est.best$PC <- PC
        }

        ## output of estimates
        est.best$mu <- est.best$mu * norm.para[1]
        if (method == "ife" && r.cv > 0) {
            est.best$lambda <- est.best$lambda * norm.para[1]
            est.best$VNT <- est.best$VNT * norm.para[1]
        }
        if (force %in% c(1, 3)) {
            est.best$alpha <- est.best$alpha * norm.para[1]
        }
        if (force %in% c(2, 3)) {
            est.best$xi <- est.best$xi * norm.para[1]
        }
        # if (p>0) {
        #    est.best$beta <- est.best$beta * norm.para[1]
        # }
        est.best$residuals <- est.best$residuals * norm.para[1]
        est.best$fit <- est.best$fit * norm.para[1]
        est.fect$fit <- est.fect$fit * norm.para[1]
        est.fect$sigma2 <- est.fect$sigma2 * norm.para[1]
    }

    ## 0. revelant parameters
    sigma2 <- IC <- PC <- NULL
    if (method == "ife") {
        sigma2 <- est.best$sigma2
        IC <- est.best$IC
        PC <- est.best$PC
    }

    if (p > 0) {
        na.pos <- is.nan(est.best$beta)
        beta <- est.best$beta
        if (sum(na.pos) > 0) {
            beta[na.pos] <- NA
        }
    } else {
        beta <- NA
    }

    ## 1. estimated att and counterfactuals
    Y.ct.equiv <- Y.ct <- NULL
    Y.ct <- est.best$fit
    eff <- Y - Y.ct
    missing.index <- which(is.na(eff))
    if (length(missing.index) > 0) {
        I[missing.index] <- 0
        II[missing.index] <- 0
    }
    if (0 %in% I) {
        eff[which(I == 0)] <- NA
    }
    complete.index <- which(!is.na(eff))
    att.avg <- sum(eff[complete.index] * D[complete.index]) / (sum(D[complete.index]))

    att.avg.balance <- NA
    if (!is.null(balance.period)) {
        complete.index2 <- which(!is.na(T.on.balance))
        att.avg.balance <- sum(eff[complete.index2] * D[complete.index2]) / (sum(D[complete.index2]))
    }

    # weighted effect
    att.avg.W <- NA
    if (!is.null(W)) {
        att.avg.W <- sum(eff[complete.index] * D[complete.index] * W[complete.index]) / (sum(D[complete.index] * W[complete.index]))
    }


    ## att.avg.unit
    tr.pos <- which(apply(D, 2, sum) > 0)
    att.unit <- sapply(1:length(tr.pos), function(vec) {
        return(sum(eff[, tr.pos[vec]] * D[, tr.pos[vec]]) / sum(D[, tr.pos[vec]]))
    })
    att.avg.unit <- mean(att.unit)

    eff.equiv <- Y - est.fect$fit
    equiv.att.avg <- sum(eff.equiv * D) / (sum(D))

    ## 2. rmse for treated units' observations under control
    tr <- which(apply(D, 2, sum) > 0)
    tr.co <- which((as.matrix(1 - D[, tr]) * as.matrix(II[, tr])) == 1)
    eff.tr <- as.matrix(eff[, tr])
    v.eff.tr <- eff.tr[tr.co]
    rmse <- sqrt(mean(v.eff.tr^2))

    ## 3. unbalanced output
    if (0 %in% I) {
        eff[which(I == 0)] <- NA
        est.best$fit[which(I == 0)] <- NA
        ## eff.equiv[which(I == 0)] <- NA
    }
    est.best$residuals[which(II == 0)] <- NA
    if (method == "mc") {
        est.best$sigma2 <- mean(c(est.best$residuals[which(II == 1)])^2) ## mean squared error of residuals
    }

    ## 4. dynamic effects
    t.on <- c(T.on)
    eff.v <- c(eff) ## a vector
    rm.pos1 <- which(is.na(eff.v))
    rm.pos2 <- which(is.na(t.on))
    eff.v.use1 <- eff.v
    t.on.use <- t.on
    n.on.use <- rep(1:N, each = TT)

    eff.equiv.v <- c(eff.equiv)

    if (NA %in% eff.v | NA %in% t.on) {
        eff.v.use1 <- eff.v[-c(rm.pos1, rm.pos2)]
        t.on.use <- t.on[-c(rm.pos1, rm.pos2)]
        n.on.use <- n.on.use[-c(rm.pos1, rm.pos2)]
        eff.equiv.v <- eff.equiv.v[-c(rm.pos1, rm.pos2)]
    }

    pre.pos <- which(t.on.use <= 0)
    eff.pre <- cbind(eff.v.use1[pre.pos], t.on.use[pre.pos], n.on.use[pre.pos])
    colnames(eff.pre) <- c("eff", "period", "unit")

    eff.pre.equiv <- cbind(eff.equiv.v[pre.pos], t.on.use[pre.pos], n.on.use[pre.pos])
    colnames(eff.pre.equiv) <- c("eff.equiv", "period", "unit")

    pre.sd <- tapply(eff.pre.equiv[, 1], eff.pre.equiv[, 2], sd)
    pre.sd <- cbind(pre.sd, sort(unique(eff.pre.equiv[, 2])), table(eff.pre.equiv[, 2]))
    colnames(pre.sd) <- c("sd", "period", "count")

    time.on <- sort(unique(t.on.use))
    att.on <- as.numeric(tapply(eff.v.use1, t.on.use, mean)) ## NA already removed
    count.on <- as.numeric(table(t.on.use))

    if (!is.null(W)) {
        W.v <- c(W)
        rm.pos.W <- which(is.na(W))
        if (NA %in% eff.v | NA %in% t.on | NA %in% W.v) {
            eff.v.use.W <- eff.v[-c(rm.pos1, rm.pos2, rm.pos.W)]
            W.v.use <- W.v[-c(rm.pos1, rm.pos2, rm.pos.W)]
            t.on.use.W <- t.on[-c(rm.pos1, rm.pos2, rm.pos.W)]
            n.on.use.W <- n.on.use[-c(rm.pos1, rm.pos2, rm.pos.W)]
        } else {
            eff.v.use.W <- eff.v.use1
            t.on.use.W <- t.on.use
            n.on.use.W <- n.on.use
            W.v.use <- W.v
        }
        time.on.W <- sort(unique(t.on.use.W))
        att.on.sum.W <- as.numeric(tapply(eff.v.use.W * W.v.use, t.on.use.W, sum)) ## NA already removed
        W.on.sum <- as.numeric(tapply(W.v.use, t.on.use.W, sum))
        att.on.W <- att.on.sum.W / W.on.sum
        count.on.W <- as.numeric(table(t.on.use.W))
    } else {
        att.on.sum.W <- att.on.W <- count.on.W <- time.on.W <- W.on.sum <- NULL
    }

    ## 4.2 balance effect
    balance.att <- NULL
    if (!is.null(balance.period)) {
        t.on.balance <- c(T.on.balance)
        rm.pos4 <- which(is.na(t.on.balance))
        t.on.balance.use <- t.on.balance

        if (NA %in% eff.v | NA %in% t.on.balance) {
            eff.v.use3 <- eff.v[-c(rm.pos1, rm.pos4)]
            t.on.balance.use <- t.on.balance[-c(rm.pos1, rm.pos4)]
        }

        balance.time <- sort(unique(t.on.balance.use))
        balance.att <- as.numeric(tapply(eff.v.use3, t.on.balance.use, mean)) ## NA already removed
        balance.count <- as.numeric(table(t.on.balance.use))
    }



    ## 5 carryover effect
    carry.att <- NULL
    if (!is.null(T.on.carry)) {
        t.on.carry <- c(T.on.carry)
        rm.pos4 <- which(is.na(t.on.carry))
        t.on.carry.use <- t.on.carry

        if (NA %in% eff.v | NA %in% t.on.carry) {
            eff.v.use3 <- eff.v[-c(rm.pos1, rm.pos4)]
            t.on.carry.use <- t.on.carry[-c(rm.pos1, rm.pos4)]
        }

        carry.time <- sort(unique(t.on.carry.use))
        carry.att <- as.numeric(tapply(eff.v.use3, t.on.carry.use, mean)) ## NA already removed

        # if (!is.null(time.on.carry.seq)) {
        #    carry.att.med <- rep(NA, length(time.on.carry.seq))
        #    carry.att.med[which(time.on.carry.seq %in% carry.time)] <- carry.att
        #    carry.att <- carry.att.med

        # }
    }



    ## 6. switch-off effects
    eff.off <- eff.equiv <- off.sd <- NULL
    if (hasRevs == 1) {
        t.off <- c(T.off)
        rm.pos3 <- which(is.na(t.off))
        eff.v.use2 <- eff.v
        t.off.use <- t.off

        if (NA %in% eff.v | NA %in% t.off) {
            eff.v.use2 <- eff.v[-c(rm.pos1, rm.pos3)]
            t.off.use <- t.off[-c(rm.pos1, rm.pos3)]
        }

        off.pos <- which(t.off.use > 0)
        eff.off <- cbind(eff.v.use2[off.pos], t.off.use[off.pos], n.on.use[off.pos])
        colnames(eff.off) <- c("eff", "period", "unit")

        eff.off.equiv <- cbind(eff.equiv.v[off.pos], t.off.use[off.pos], n.on.use[off.pos])
        colnames(eff.off.equiv) <- c("off.equiv", "period", "unit")

        off.sd <- tapply(eff.off.equiv[, 1], eff.off.equiv[, 2], sd)
        off.sd <- cbind(off.sd, sort(unique(eff.off.equiv[, 2])), table(eff.off.equiv[, 2]))
        colnames(off.sd) <- c("sd", "period", "count")

        time.off <- sort(unique(t.off.use))
        att.off <- as.numeric(tapply(eff.v.use2, t.off.use, mean)) ## NA already removed
        count.off <- as.numeric(table(t.off.use))

        if (!is.null(W)) {
            if (NA %in% eff.v | NA %in% t.off | NA %in% W.v) {
                eff.v.use2.W <- eff.v[-c(rm.pos1, rm.pos3, rm.pos.W)]
                W.v.use2 <- W.v[-c(rm.pos1, rm.pos3, rm.pos.W)]
                t.off.use.W <- t.off[-c(rm.pos1, rm.pos3, rm.pos.W)]
            } else {
                eff.v.use2.W <- eff.v.use2
                t.off.use.W <- t.off.use
                W.v.use2 <- W.v
            }

            time.off.W <- sort(unique(t.off.use.W))
            att.off.sum.W <- as.numeric(tapply(eff.v.use2.W * W.v.use2, t.off.use.W, sum))
            W.off.sum <- as.numeric(tapply(W.v.use2, t.off.use.W, sum))
            att.off.W <- att.off.sum.W / W.off.sum ## NA already removed
            count.off.W <- as.numeric(table(t.off.use.W))
        } else {
            W.off.sum <- att.off.sum.W <- att.off.W <- count.off.W <- time.off.W <- NULL
        }
    }

    ## 8. loess HTE by time
    D.missing <- D
    D.missing[which(D == 0)] <- NA
    eff.calendar <- apply(eff * D.missing, 1, mean, na.rm = TRUE)
    N.calendar <- apply(!is.na(eff * D.missing), 1, sum)
    T.calendar <- c(1:TT)
    if (sum(!is.na(eff.calendar)) > 1) {
        # loess fit
        loess.fit <- suppressWarnings(try(loess(eff.calendar ~ T.calendar, weights = N.calendar), silent = TRUE))
        if ("try-error" %in% class(loess.fit)) {
            eff.calendar.fit <- eff.calendar
            calendar.enp <- NULL
        } else {
            eff.calendar.fit <- eff.calendar
            eff.calendar.fit[which(!is.na(eff.calendar))] <- loess.fit$fit
            calendar.enp <- loess.fit$enp
        }
    } else {
        eff.calendar.fit <- eff.calendar
        calendar.enp <- NULL
    }

    ## 7. cohort effects
    if (!is.null(group)) {
        cohort <- cbind(c(group), c(D), c(eff.v))
        rm.pos <- unique(c(rm.pos1, which(cohort[, 2] == 0)))
        cohort <- cohort[-rm.pos, ]

        g.level <- sort(unique(cohort[, 1]))
        raw.group.att <- as.numeric(tapply(cohort[, 3], cohort[, 1], mean))

        group.att <- rep(NA, length(group.level))
        group.att[which(group.level %in% g.level)] <- raw.group.att

        # by-group dynamic effects
        group.level.name <- names(group.level)

        group.output <- list()
        for (i in c(1:length(group.level))) {
            sub.group <- group.level[i]
            sub.group.name <- group.level.name[i]

            ## by-group dynamic effects
            t.on.sub <- c(T.on[which(group == sub.group)])
            eff.v.sub <- c(eff[which(group == sub.group)]) ## a vector
            rm.pos1.sub <- which(is.na(eff.v.sub))
            rm.pos2.sub <- which(is.na(t.on.sub))
            eff.v.use1.sub <- eff.v.sub
            t.on.use.sub <- t.on.sub
            if (NA %in% eff.v.sub | NA %in% t.on.sub) {
                eff.v.use1.sub <- eff.v.sub[-c(rm.pos1.sub, rm.pos2.sub)]
                t.on.use.sub <- t.on.sub[-c(rm.pos1.sub, rm.pos2.sub)]
            }
            if (length(t.on.use.sub) > 0) {
                time.on.sub <- sort(unique(t.on.use.sub))
                att.on.sub <- as.numeric(tapply(
                    eff.v.use1.sub,
                    t.on.use.sub,
                    mean
                )) ## NA already removed
                count.on.sub <- as.numeric(table(t.on.use.sub))
            } else {
                time.on.sub <- att.on.sub <- count.on.sub <- NULL
            }

            suboutput <- list(
                att.on = att.on.sub,
                time.on = time.on.sub,
                count.on = count.on.sub
            )


            ## T.off
            if (hasRevs == 1) {
                t.off.sub <- c(T.off[which(group == sub.group)])
                rm.pos3.sub <- which(is.na(t.off.sub))
                eff.v.use2.sub <- eff.v.sub
                t.off.use.sub <- t.off.sub
                if (NA %in% eff.v.sub | NA %in% t.off.sub) {
                    eff.v.use2.sub <- eff.v.sub[-c(rm.pos1.sub, rm.pos3.sub)]
                    t.off.use.sub <- t.off.sub[-c(rm.pos1.sub, rm.pos3.sub)]
                }
                if (length(t.off.use.sub) > 0) {
                    time.off.sub <- sort(unique(t.off.use.sub))
                    att.off.sub <- as.numeric(tapply(eff.v.use2.sub, t.off.use.sub, mean)) ## NA already removed
                    count.off.sub <- as.numeric(table(t.off.use.sub))
                } else {
                    time.off.sub <- att.off.sub <- count.off.sub <- NULL
                }

                suboutput <- c(suboutput, list(
                    att.off = att.off.sub,
                    count.off = count.off.sub,
                    time.off = time.off.sub
                ))
            }
            group.output[[sub.group.name]] <- suboutput
        }
    }



    ## -------------------------------##
    ##           Storage
    ## -------------------------------##

    ## control group residuals
    out <- list(
        ## main results
        sigma2 = est.best$sigma2,
        sigma2.fect = est.fect$sigma2,
        T.on = T.on,
        Y.ct = est.best$fit,
        eff = eff,
        att.avg = att.avg,
        att.avg.unit = att.avg.unit,
        ## supporting
        force = force,
        T = TT,
        N = N,
        Ntr = Ntr,
        Nco = Nco,
        p = p,
        D = D,
        Y = Y,
        X = X,
        I = I,
        II = II,
        tr = tr,
        co = co,
        est = est.best,
        method = method,
        mu = est.best$mu,
        beta = beta,
        validX = validX,
        validF = validF,
        niter = est.best$niter,
        time = time.on,
        att = att.on,
        count = count.on,
        eff.calendar = eff.calendar,
        N.calendar = N.calendar,
        eff.calendar.fit = eff.calendar.fit,
        calendar.enp = calendar.enp,
        eff.pre = eff.pre,
        eff.pre.equiv = eff.pre.equiv,
        pre.sd = pre.sd,
        rmse = rmse,
        rmCV = rmCV,
        estCV = estCV,
        res = est.best$res
    )

    if (hasRevs == 1) {
        out <- c(out, list(
            time.off = time.off,
            att.off = att.off,
            count.off = count.off,
            eff.off = eff.off,
            eff.off.equiv = eff.off.equiv,
            off.sd = off.sd
        ))
    }

    if (!is.null(W)) {
        out <- c(out, list(
            W = W,
            att.avg.W = att.avg.W,
            att.on.sum.W = att.on.sum.W,
            att.on.W = att.on.W,
            count.on.W = count.on.W,
            time.on.W = time.on.W,
            W.on.sum = W.on.sum
        ))
        if (hasRevs == 1) {
            out <- c(out, list(
                att.off.sum.W = att.off.sum.W,
                att.off.W = att.off.W,
                count.off.W = count.off.W,
                time.off.W = time.off.W,
                W.off.sum = W.off.sum
            ))
        }
    }

    if (!is.null(T.on.carry)) {
        out <- c(out, list(carry.att = carry.att, carry.time = carry.time))
    }

    if (!is.null(balance.period)) {
        out <- c(out, list(balance.att = balance.att, balance.time = balance.time, balance.count = balance.count, balance.avg.att = att.avg.balance))
    }

    if (force == 1) {
        out <- c(out, list(
            alpha = est.best$alpha,
            alpha.tr = as.matrix(est.best$alpha[tr, ]),
            alpha.co = as.matrix(est.best$alpha[co, ])
        ))
    } else if (force == 2) {
        out <- c(out, list(xi = est.best$xi))
    } else if (force == 3) {
        out <- c(out, list(
            alpha = est.best$alpha, xi = est.best$xi,
            alpha.tr = as.matrix(est.best$alpha[tr, ]),
            alpha.co = as.matrix(est.best$alpha[co, ])
        ))
    }

    if (method == "ife") {
        out <- c(out, list(r.cv = r.cv, IC = IC, PC = PC))
        if (r.cv > 0) {
            out <- c(out, list(
                factor = as.matrix(est.best$factor),
                lambda = as.matrix(est.best$lambda),
                lambda.tr = as.matrix(est.best$lambda[tr, ]),
                lambda.co = as.matrix(est.best$lambda[co, ])
            ))
        }
    }

    if (method == "mc") {
        out <- c(out, list(
            lambda.cv = lambda.cv, lambda.seq = lambda,
            lambda.norm = lambda.cv / max(eigen.all),
            eigen.all = eigen.all
        ))
    }

    ## CV results
    if (!is.null(CV.out.ife)) {
        out <- c(out, list(CV.out.ife = CV.out.ife))
    }

    if (!is.null(CV.out.mc)) {
        out <- c(out, list(CV.out.mc = CV.out.mc))
    }

    if (!is.null(group)) {
        out <- c(out, list(
            group.att = group.att,
            group.output = group.output
        ))
    }

    # Keep the selected concrete method for downstream bootstrap/inference.
    out <- c(out, list(method = method))

    return(out)
} ## cross-validation function ends

## ---------------------------------------------------------------
## Tests for fect_cv: regression coverage (S2) + cv.method extension
## in fect_cv (Section B, NEW for Phase 2).
##
## Originally part of test-score-unify.R; split out 2026-05-03 for
## progress visibility. Shared fixtures live in helper-score-unify.R.
## ---------------------------------------------------------------

## =================================================================
## S2: fect_cv regression test (before vs after refactor)
## =================================================================

test_that("S2.1: IFE method CV - r.cv and CV.out snapshot", {

  skip_on_cran()
  cv_out <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      CV        = TRUE,
      r         = c(0, 3),
      criterion = "mspe",
      se        = FALSE,
      parallel  = FALSE
    )
  ))
  # r.cv should be a non-negative integer in [0, 3]
  expect_true(cv_out$r.cv >= 0 && cv_out$r.cv <= 3)

  # CV.out should exist and have MSPE column with finite positive values
  expect_true(!is.null(cv_out$CV.out))
  mspe_col <- cv_out$CV.out[, "MSPE"]
  # At least some entries should be less than 1e20 (the init value)
  expect_true(any(mspe_col < 1e19))
  expect_true(all(is.finite(mspe_col[mspe_col < 1e19])))
  expect_true(all(mspe_col[mspe_col < 1e19] > 0))
})

test_that("S2.2: MC method CV - lambda.cv selection", {

  skip_on_cran()
  cv_mc <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2,
      data      = simdata,
      index     = c("id", "time"),
      method    = "mc",
      CV        = TRUE,
      criterion = "mspe",
      se        = FALSE,
      parallel  = FALSE
    )
  ))
  # lambda.cv should be selected
  expect_true(!is.null(cv_mc$lambda.cv) || !is.null(cv_mc$r.cv))
  expect_true(!is.null(cv_mc$CV.out))
})

test_that("S2.3: GMoment column correctly populated (IFE)", {

  skip_on_cran()
  cv_out <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      CV        = TRUE,
      r         = c(0, 3),
      criterion = "mspe",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  # GMoment values should be finite and positive where computed
  gm_col <- cv_out$CV.out[, "GMoment"]
  computed <- gm_col[gm_col < 1e19]
  if (length(computed) > 0) {
    expect_true(all(is.finite(computed)))
    expect_true(all(computed > 0))

    # GMoment should generally differ from MSPTATT
    msptatt_col <- cv_out$CV.out[, "MSPTATT"]
    msptatt_computed <- msptatt_col[gm_col < 1e19]
    if (length(msptatt_computed) > 0) {
      if (length(computed) > 1) {
        expect_false(
          all(abs(computed - msptatt_computed) < 1e-15),
          info = "GMoment should not be identical to MSPTATT for all r values"
        )
      }
    }
  }
})


## =================================================================
## Section B: cv.method in fect_cv (NEW for Phase 2)
## =================================================================

test_that("CV1: cv.method='all_units' selects r.cv", {

  skip_on_cran()
  cv_out <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      CV        = TRUE,
      r         = c(0, 3),
      cv.method = "all_units",
      se        = FALSE,
      parallel  = FALSE
    )
  ))
  # r.cv is integer in [0, 3]
  expect_true(cv_out$r.cv >= 0 && cv_out$r.cv <= 3)
  # CV.out exists with MSPE column
  expect_true(!is.null(cv_out$CV.out))
  mspe_col <- cv_out$CV.out[, "MSPE"]
  computed <- mspe_col[mspe_col < 1e19]
  expect_true(all(is.finite(computed)))
  expect_true(all(computed > 0))
})

test_that("CV2: cv.method='treated_units' selects r.cv", {

  skip_on_cran()
  cv_out <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      CV        = TRUE,
      r         = c(0, 3),
      cv.method = "treated_units",
      se        = FALSE,
      parallel  = FALSE
    )
  ))
  # r.cv is integer in [0, 3]
  expect_true(cv_out$r.cv >= 0 && cv_out$r.cv <= 3)
  # CV.out exists
  expect_true(!is.null(cv_out$CV.out))
  mspe_col <- cv_out$CV.out[, "MSPE"]
  computed <- mspe_col[mspe_col < 1e19]
  expect_true(all(is.finite(computed)))
  expect_true(all(computed > 0))
})

test_that("CV3: Invalid cv.method rejected", {

  skip_on_cran()
  expect_error(
    fect::fect(
      Y ~ D + X1 + X2,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      CV        = TRUE,
      cv.method = "invalid"
    ),
    "cv.method|arg"
  )
})

###################################################################
## CV rule selection helpers (added v2.3.0)
##
## The classic fect rule is a 1% relative-tolerance check applied
## sequentially during the (r) loop:
##     update r.cv if (best - current) > 0.01 * best
##
## v2.3.0 introduces the 1-SE rule (Breiman, Friedman, Olshen & Stone 1984;
## Hastie, Tibshirani & Friedman 2009 §7.10): pick the smallest r whose mean
## CV criterion is within one fold-SE of the best. Default `cv.rule = "1se"`.
## The 1% rule is preserved as `cv.rule = "1pct"` for byte-identical
## reproducibility of pre-2.3.0 fits. A pure-min rule is also available as
## `cv.rule = "min"`.
##
## The fold SE is computed from per-fold criterion values:
##     SE_r = sd(score_per_fold_r) / sqrt(K)
##
## Per-fold scores are obtained by calling .score_residuals() on each fold's
## residuals separately (instead of pooling residuals first), then taking
## column-wise sd / sqrt(K).
###################################################################


## Compute per-fold criterion vectors and pooled criterion vector for a list
## of fold-result lists (each list item containing $resid, $time_idx, optional
## $obs_w). Returns:
##   $pooled : named vector of criteria computed from pooled residuals (legacy)
##   $folds  : K x ncriteria matrix of per-fold criterion values
##   $mean   : named vector of mean-across-folds criterion values
##   $se     : named vector of (sd / sqrt(K)) per criterion
##
## When K == 1, $se is set to 0 for all criteria (no fold dispersion).
.fect_cv_aggregate_folds <- function(fold_list,
                                     count.T.cv,
                                     use_weight = 0L,
                                     norm.para = NULL) {
    ## Pooled (legacy) score
    all_resid    <- unlist(lapply(fold_list, `[[`, "resid"))
    all_time_idx <- unlist(lapply(fold_list, `[[`, "time_idx"))
    all_obs_w    <- if (use_weight == 1L)
                        unlist(lapply(fold_list, `[[`, "obs_w"))
                    else NULL

    pooled <- .score_residuals(
        resid         = all_resid,
        obs_weights   = all_obs_w,
        time_index    = all_time_idx,
        count_weights = count.T.cv,
        norm.para     = norm.para
    )

    ## Per-fold scores
    per_fold <- lapply(fold_list, function(fl) {
        .score_residuals(
            resid         = fl$resid,
            obs_weights   = if (use_weight == 1L) fl$obs_w else NULL,
            time_index    = fl$time_idx,
            count_weights = count.T.cv,
            norm.para     = norm.para
        )
    })
    score_mat <- do.call(rbind, per_fold)
    K <- nrow(score_mat)

    means <- if (K >= 1L) colMeans(score_mat, na.rm = TRUE) else pooled
    ses <- if (K > 1L) {
        apply(score_mat, 2L, function(x) {
            x_finite <- x[is.finite(x)]
            if (length(x_finite) <= 1L) return(0.0)
            stats::sd(x_finite) / sqrt(length(x_finite))
        })
    } else {
        rep(0.0, length(pooled))
    }
    names(ses) <- names(pooled)

    list(pooled = pooled, folds = score_mat, mean = means, se = ses)
}


## Apply a CV rule to select the row index of CV.out matching the chosen
## hyper-parameter (r or lambda).
##
## means: numeric vector indexed by row (one per candidate hyper-param value).
##        NAs and Inf are treated as missing (excluded from selection).
## ses:   numeric vector of SE per row, OR NULL (used only when rule == "1se").
## rule:  one of "1se" (default), "min", "1pct".
##
## Returns the row index (integer). Uses the smallest index among ties
## consistent with bias-toward-parsimony.
.fect_apply_cv_rule <- function(means, ses = NULL, rule = c("1se", "min", "1pct")) {
    rule <- match.arg(rule)
    valid <- is.finite(means)
    if (!any(valid)) return(NA_integer_)

    if (rule == "min") {
        m_min <- min(means[valid])
        return(min(which(valid & means == m_min)))
    }

    if (rule == "1pct") {
        m_min <- min(means[valid])
        threshold <- m_min * 1.01
        return(min(which(valid & means <= threshold)))
    }

    ## "1se"
    i_min <- min(which(valid & means == min(means[valid])))
    se_min <- if (!is.null(ses) && is.finite(ses[i_min])) ses[i_min] else 0
    threshold <- means[i_min] + se_min
    min(which(valid & means <= threshold))
}


## Validate user-supplied cv.rule argument and return a single canonical string.
.fect_validate_cv_rule <- function(cv.rule) {
    if (is.null(cv.rule)) return("1se")
    if (!is.character(cv.rule) || length(cv.rule) != 1L) {
        stop("'cv.rule' must be a single character string: ",
             "'1se' (default), 'min', or '1pct'.")
    }
    if (!cv.rule %in% c("1se", "min", "1pct")) {
        stop("'cv.rule' must be one of '1se', 'min', '1pct'. Got: '",
             cv.rule, "'.")
    }
    cv.rule
}

#' Over-identification test for causal moderation (cm)
#'
#' Implements the over-identification test described in the slide:
#' run two auxiliary regressions of residualized outcomes on the moderator
#' and covariates (including quadratic and interaction terms) with unit and
#' time fixed effects, then test using \eqn{n \times R^2 \sim \chi^2(df)},
#' where df is the number of *nonlinear* (quadratic + interaction) terms that
#' remain in the fitted model.
#'
#' @param x A `fect` object. Must be estimated with `cm=TRUE` so that `x$est.cm` exists.
#' @param moderator Character scalar. Name of the moderator \eqn{M_{it}} (must be in `x$X`).
#' @param covariates Optional character vector. Names of other covariates \eqn{X_{it}}.
#'   Default is all covariates in `x$X` except `moderator`.
#' @param quadratic Logical. Include quadratic terms \eqn{M_{it}^2} and \eqn{X_{it}^2}. Default TRUE.
#' @param interaction Logical. Include interaction terms \eqn{M_{it} \times X_{it}}. Default TRUE.
#'
#' @return A list with elements `e1` and `e0`, each containing:
#'   - `n`: sample size used in the auxiliary regression
#'   - `r2`: R-squared of the auxiliary regression
#'   - `stat`: \eqn{n \times R^2}
#'   - `df`: degrees of freedom (# nonlinear terms kept)
#'   - `p`: p-value from chi-square test
#'   - `model`: the fitted `fixest` model
#'
#' @export
fect_iden <- function(x,
                      moderator,
                      covariates = NULL,
                      quadratic = TRUE,
                      interaction = TRUE) {
  if (is.null(x$est) || is.null(x$est$fit)) {
    stop("`x$est$fit` is required to compute g0.\n")
  }
  if (is.null(x$est.cm) || is.null(x$est.cm$fit)) {
    stop("This test requires `cm=TRUE` so that `x$est.cm$fit` exists.\n")
  }
  if (is.null(x$Y.dat) || is.null(x$D.dat)) {
    stop("`x$Y.dat` and `x$D.dat` are required.\n")
  }
  if (is.null(x$id) || is.null(x$rawtime)) {
    stop("`x$id` and `x$rawtime` are required.\n")
  }
  if (is.null(x$X) || length(x$X) == 0) {
    stop("`x$X` (covariate names) is required.\n")
  }

  if (!is.character(moderator) || length(moderator) != 1) {
    stop("`moderator` must be a single character string.\n")
  }
  if (!moderator %in% x$X) {
    stop("`moderator` must be one of `x$X`.\n")
  }

  if (is.null(covariates)) {
    covariates <- setdiff(x$X, moderator)
  }
  if (!is.character(covariates)) {
    stop("`covariates` must be a character vector.\n")
  }
  if (length(covariates) > 0 && any(!covariates %in% x$X)) {
    stop("All `covariates` must be in `x$X`.\n")
  }

  # Extract covariate array (T x N x p) from the fect object.
  # In `fect` objects there are typically two entries named "X":
  # - the first is `x$X` (character vector of covariate names)
  # - the second is the actual covariate array (T x N x p)
  X.arr <- NULL
  X.cands <- x[names(x) == "X"]
  if (length(X.cands) >= 2) {
    # Using `[2]$X` (not `[[2]]$X`) works whether the second element is a list or an array.
    X.arr <- X.cands[2]$X
  } else if (is.array(x$X)) {
    # Fallback: some internal objects may store the array directly at x$X
    X.arr <- x$X
  }
  if (is.null(X.arr) || length(dim(X.arr)) != 3) {
    stop("Cannot locate covariate array in `x`.\n")
  }

  TT <- nrow(x$Y.dat)
  N <- ncol(x$Y.dat)
  if (nrow(x$D.dat) != TT || ncol(x$D.dat) != N) {
    stop("Dimensions of `x$Y.dat` and `x$D.dat` do not match.\n")
  }

  g0 <- x$est$fit
  g1 <- x$est.cm$fit
  if (!is.matrix(g0) || !is.matrix(g1) || any(dim(g0) != c(TT, N)) || any(dim(g1) != c(TT, N))) {
    stop("`x$est$fit` and `x$est.cm$fit` must be T-by-N matrices matching `x$Y.dat`.\n")
  }

  Y <- x$Y.dat
  D <- x$D.dat
  I <- if (!is.null(x$I.dat)) x$I.dat else matrix(1, TT, N)

  # Residual-like objects defined in the slide:
  # e1_it = D_it (Y_it - g1_it), on treated cells
  # e0_it = (1-D_it) (Y_it - g0_it), on control cells
  e1 <- D * (Y - g1)
  e0 <- (1 - D) * (Y - g0)

  # Long-form indexing consistent with c(matrix): time varies fastest within unit.
  unit <- rep(x$id, each = TT)
  time <- rep(x$rawtime, times = N)

  # Extract moderator and covariates into vectors aligned with c(Y).
  m_idx <- which(x$X == moderator)[1]
  M.vec <- c(X.arr[, , m_idx])

  X.dat <- list()
  if (length(covariates) > 0) {
    for (nm in covariates) {
      idx <- which(x$X == nm)[1]
      X.dat[[nm]] <- c(X.arr[, , idx])
    }
  }

  Y.vec <- c(Y)
  D.vec <- c(D)
  I.vec <- c(I)
  g0.vec <- c(g0)
  g1.vec <- c(g1)
  e1.vec <- c(e1)
  e0.vec <- c(e0)

  base_df <- data.frame(
    unit = as.factor(unit),
    time = as.factor(time),
    M = M.vec
  )
  names(base_df)[names(base_df) == "M"] <- moderator
  if (length(X.dat) > 0) {
    for (nm in names(X.dat)) {
      base_df[[nm]] <- X.dat[[nm]]
    }
  }

  build_formula <- function(lhs) {
    linear_terms <- c(moderator, covariates)
    nonlinear_terms <- c()

    if (isTRUE(quadratic)) {
      nonlinear_terms <- c(nonlinear_terms, paste0("I(", moderator, "^2)"))
      if (length(covariates) > 0) {
        nonlinear_terms <- c(nonlinear_terms, paste0("I(", covariates, "^2)"))
      }
    }
    if (isTRUE(interaction) && length(covariates) > 0) {
      nonlinear_terms <- c(nonlinear_terms, paste0(moderator, ":", covariates))
    }

    rhs_terms <- c(linear_terms, nonlinear_terms)
    rhs_terms <- rhs_terms[nzchar(rhs_terms)]
    rhs <- if (length(rhs_terms) > 0) paste(rhs_terms, collapse = " + ") else "1"
    stats::as.formula(paste0(lhs, " ~ ", rhs, " | unit + time"))
  }

  r2_of <- function(mod) {
    # Prefer fixest fitstat if available; fall back to r2().
    r2v <- tryCatch({
      fs <- fixest::fitstat(mod, "r2")
      as.numeric(fs$r2)
    }, error = function(e) NA_real_)
    if (is.na(r2v)) {
      r2v <- tryCatch(as.numeric(fixest::r2(mod, type = "r2")), error = function(e) NA_real_)
    }
    r2v
  }

  df_nonlinear <- function(mod) {
    cn <- names(stats::coef(mod))
    if (is.null(cn) || length(cn) == 0) return(0L)

    keep <- rep(FALSE, length(cn))
    # quadratic terms
    if (isTRUE(quadratic)) {
      keep <- keep | grepl(paste0("^I\\(", moderator, "\\^2\\)$"), cn)
      if (length(covariates) > 0) {
        for (nm in covariates) {
          keep <- keep | grepl(paste0("^I\\(", nm, "\\^2\\)$"), cn)
        }
      }
    }
    # interaction terms
    if (isTRUE(interaction) && length(covariates) > 0) {
      for (nm in covariates) {
        keep <- keep | (cn == paste0(moderator, ":", nm)) | (cn == paste0(nm, ":", moderator))
      }
    }
    sum(keep)
  }

  run_one <- function(resp_vec, keep_idx, lhs_name) {
    dat <- base_df[keep_idx, , drop = FALSE]
    dat[[lhs_name]] <- resp_vec[keep_idx]

    fml <- build_formula(lhs_name)
    mod <- fixest::feols(fml, data = dat, warn = FALSE)

    n <- stats::nobs(mod)
    r2 <- r2_of(mod)
    df <- df_nonlinear(mod)
    stat <- n * r2
    p <- if (is.finite(stat) && df > 0) (1 - stats::pchisq(stat, df = df)) else NA_real_

    list(n = n, r2 = r2, stat = stat, df = df, p = p, model = mod)
  }

  # Restrict to observed cells (I==1) and appropriate treatment status.
  keep_common <- which(I.vec == 1 & !is.na(Y.vec))
  keep_e1 <- intersect(keep_common, which(D.vec == 1 & !is.na(g1.vec) & !is.na(e1.vec)))
  keep_e0 <- intersect(keep_common, which(D.vec == 0 & !is.na(g0.vec) & !is.na(e0.vec)))

  out <- list(
    e1 = run_one(e1.vec, keep_e1, "e1"),
    e0 = run_one(e0.vec, keep_e0, "e0"),
    call = match.call()
  )
  class(out) <- "fect_iden"
  out
}

# Interactive Fixed Effect Model
# Version 1.02
# LLC WY XYQ, 2019.2.2


## generic function
interFE <- function(formula = NULL,
                    data, # a data frame
                    Y, # outcome variable
                    X = NULL, # covariates
                    W = NULL, # weights
                    index, # id and time indicators
                    r = 0, # number of factors
                    force = "two-way", # additived fixed effects
                    se = FALSE, # standard error
                    nboots = 500, # number of bootstrap runs
                    seed = NULL,
                    tol = 1e-3,
                    max_iteration = 500,
                    binary = FALSE,
                    QR = FALSE,
                    normalize = FALSE) {
    UseMethod("interFE")
}

## formula method
interFE.formula <- function(formula = NULL, data, # a data frame
                            Y, # outcome variable
                            X = NULL, # covariates
                            W = NULL, # weights
                            index, # id and time indicators
                            r = 0, # number of factors
                            force = "two-way", # additived fixed effects
                            se = 0, # standard error
                            nboots = 500, # number of bootstrap runs
                            seed = NULL,
                            tol = 1e-3,
                            max_iteration = 500,
                            binary = FALSE,
                            QR = FALSE,
                            normalize = FALSE) {
    ## parsing
    varnames <- all.vars(formula)
    Yname <- varnames[1]
    Xname <- varnames[2:length(varnames)]


    ## check binary outcome
    if (binary == TRUE) {
        unique_y <- sort(unique(data[, Yname]))
        if (length(unique_y) != 2) {
            stop("Outcome should only contain 0 and 1.")
        } else {
            if (sum(unique_y == c(0, 1)) != 2) {
                stop("Outcome should only contain 0 and 1.")
            }
        }
    }


    ## run the model
    out <- interFE.default(
        formula = NULL, data = data, Y = Yname, X = Xname, W,
        index, # id and time indicators
        r, # number of factors
        force, # additived fixed effects
        se, # standard error
        nboots, # number of bootstrap runs
        seed,
        tol,
        max_iteration,
        binary,
        QR,
        normalize
    )
    out$call <- match.call()
    out$formula <- formula
    return(out)
}


print.interFE <- function(x,
                          ...) {
    message("Call:\n")
    print(x$call, digits = 4)
    message("\nEstimated Coefficients:\n")
    print(x$est.table, digits = 4)
}


###################################
# panel interactive fixed effects
###################################

interFE.default <- function(formula = NULL, data, # a data frame
                            Y, # outcome variable
                            X = NULL, # covariates
                            W = NULL,
                            index, # id and time indicators
                            r = 0, # number of factors
                            force = "two-way", # additived fixed effects
                            se = 0, # standard error
                            nboots = 500, # number of bootstrap runs
                            seed = NULL,
                            tol = 1e-5,
                            max_iteration = 500,
                            binary = FALSE,
                            QR = FALSE,
                            normalize = FALSE) {
    ## -------------------------------#
    ## Parameters
    ## -------------------------------#

    ## index
    if (length(index) != 2 | sum(index %in% colnames(data)) != 2) {
        stop("\"index\" option misspecified. Try, for example, index = c(\"unit.id\", \"time\").")
    }
    if (force == "none") { # no additive fixed effects imposed
        force <- 0
    } else if (force == "unit") { # unit fixed-effect
        force <- 1
    } else if (force == "time") { # time fixed-effect
        force <- 2
    } else if (force == "two-way") { # two-way fixed-effect
        force <- 3
    }
    if (!force %in% c(0, 1, 2, 3)) {
        stop("\"force\" option misspecified; choose from c(\"none\", \"unit\", \"time\", \"two-way\").")
    }

    ## -------------------------------#
    ## Parsing raw data
    ## -------------------------------#

    ## store variable names
    Yname <- Y
    Xname <- X
    Wname <- W


    if (!is.null(W)) {
        data <- data[which(data[, Wname] > 0), ]
    }

    id <- index[1]
    time <- index[2]

    id.series <- unique(data[, id])
    ## sort data
    data <- data[order(data[, id], data[, time]), ]

    ## dimensions
    T <- length(unique(data[, time]))
    N <- length(unique(data[, id]))
    p <- length(Xname)

    ## normalize
    if (binary == TRUE) {
        normalize <- FALSE
    }
    norm.para <- NULL
    if (normalize == TRUE) {
        sd.Y <- sd(as.matrix(data[, Yname]))
        data[, c(Yname, Xname)] <- data[, c(Yname, Xname)] / sd.Y
        ## if (length(Xname)>0) {
        ##     sd.X <- apply(as.matrix(data[,Xname]),2,sd)
        ##     data[,Xname] <- as.matrix(data[,Xname])/sd.X
        ##     norm.para <- c(sd.Y,sd.X)
        ## } else {
        norm.para <- sd.Y
        ## }
    }

    if (p > 0) {
        for (i in 1:p) {
            if (sum(is.na(data[, Xname[i]])) > 0) {
                stop(paste("Missing values in variable \"", Xname[i], "\".", sep = ""))
            }

            if (sum(tapply(data[, Xname[i]], data[, id], var), na.rm = TRUE) == 0) {
                stop(paste("Variable \"", Xname[i], "\" is unit-invariant. Try to remove it.", sep = ""))
            }
            if (sum(tapply(data[, Xname[i]], data[, time], var), na.rm = TRUE) == 0) {
                stop(paste("Variable \"", Xname[i], "\" is time-invariant. Try to remove it.", sep = ""))
            }
        }
    }

    ## check balanced panel
    if (dim(data)[1] < T * N) {
        data[, time] <- as.numeric(as.factor(data[, time]))
        ## ob <- "time_ob_ls"

        ## while (ob %in% colnames(data)) {
        ##     ob <- paste(ob, ob, sep = "_")
        ## }

        ## data[, ob] <- data[, time]
        ## for (i in 1:N) {
        ##     data[data[,id] == id.series[i], ob] <- data[data[,id] == id.series[i],time] + (i - 1) * TT
        ## }

        ob.indicator <- data[, time]
        id.indicator <- table(data[, id])
        sub.start <- 1
        for (i in 1:(N - 1)) {
            sub.start <- sub.start + id.indicator[i]
            sub.end <- sub.start + id.indicator[i + 1] - 1
            ob.indicator[sub.start:sub.end] <- ob.indicator[sub.start:sub.end] + i * T
        }

        variable <- c(Yname, Xname, Wname)

        data_I <- matrix(0, N * T, 1)
        data_I[ob.indicator, 1] <- 1
        data_ub <- as.matrix(data[, variable])
        data <- data_ub_adj(data_I, data_ub)
        colnames(data) <- variable
    }

    I <- matrix(1, T, N)
    Y.ind <- matrix(data[, Yname], T, N)
    I[is.nan(Y.ind)] <- 0

    if (0 %in% I) {
        data[is.nan(data)] <- 0
    }

    ## parse data
    Y <- matrix(data[, Yname], T, N)
    if (!is.null(Wname)) {
        W <- matrix(data[, Wname], T, N)
        W[which(I == 0)] <- 0
        use_weight <- 1
    } else {
        use_weight <- 0
    }

    ## time-varying covariates
    X <- array(0, dim = c(T, N, p))
    if (p > 0) {
        for (i in 1:p) {
            X[, , i] <- matrix(data[, Xname[i]], T, N)
        }
    }

    ## -------------------------------#
    ## Estimation
    ## -------------------------------#
    initialOut <- Y0 <- NULL
    FE0 <- xi0 <- factor0 <- NULL
    beta0 <- matrix(0, 1, 1)
    if (p > 0) {
        beta0 <- as.matrix(rep(0, p))
    }

    if (0 %in% I || use_weight) {
        data.ini <- matrix(NA, (T * N), (2 + 1 + p))
        data.ini[, 2] <- rep(1:N, each = T) ## unit fe
        data.ini[, 3] <- rep(1:T, N) ## time fe
        data.ini[, 1] <- c(Y) ## outcome
        if (p > 0) { ## covar
            for (i in 1:p) {
                data.ini[, (3 + i)] <- c(X[, , i])
            }
        }

        if (binary == FALSE) {
            if (use_weight) {
                initialOut <- initialFit(data.ini, force, w = c(W), which(c(I) == 1))
            } else {
                initialOut <- initialFit(data.ini, force, NULL, which(c(I) == 1))
            }

            Y0 <- initialOut$Y0
            if (p > 0) {
                beta0 <- initialOut$beta0
            }
        } else {
            initialOut <- BiInitialFit(data = data.ini, QR = QR, r = r, force = force, oci = which(c(I) == 1))
            Y0 <- initialOut$Y0
            FE0 <- initialOut$FE0
            if (QR == 1) {
                xi0 <- initialOut$xi0
                factor0 <- initialOut$factor0
            }
        }
    }

    ## estimates
    if (binary == FALSE) {
        if ((!0 %in% I) & (!use_weight)) {
            out <- inter_fe(Y = Y, X = X, r = r, beta0_in = beta0, force = force, max_iter = max_iteration, tol = tol)
        } else {
            if (!use_weight) {
                W <- as.matrix(0)
            }
            out <- inter_fe_ub(Y = Y, Y0 = Y0, X = X, I = I, W_in = W, beta0 = beta0, r = r, force = force, max_iter = max_iteration, tol = tol)
        }
    } else {
        if (QR == FALSE) {
            if (!0 %in% I) {
                out <- inter_fe_d(Y, Y0, FE0, X, r = r, force, tol = tol)
            } else {
                out <- inter_fe_d_ub(Y, Y0, FE0, X, I, r = r, force, tol = tol)
            }
        } else {
            if (!0 %in% I) {
                out <- inter_fe_d_qr(Y, Y0, FE0, factor0, xi0, X, r = r, force, tol = tol)
            } else {
                out <- inter_fe_d_qr_ub(Y, Y0, FE0, factor0, xi0, X, I, r = r, force, tol = tol)
            }
        }
    }


    if (is.null(norm.para)) {
        if (p > 0) {
            beta <- as.matrix(out$beta)
            if (!0 %in% I) {
                beta0 <- beta
                beta0[is.nan(beta0)] <- 0
            }
        }
        mu <- out$mu
    } else {
        mu <- out$mu * norm.para[1]
        if (p > 0) {
            beta <- as.matrix(out$beta)
            if (!0 %in% I) {
                beta0 <- beta
                beta0[is.nan(beta0)] <- 0
            }
            ## beta<-as.matrix(out$beta)*norm.para[1]/norm.para[2:length(norm.para)]
        }
    }


    ## -------------------------------#
    ## Standard Errors
    ## -------------------------------#

    ## function to get two-sided p-values
    get.pvalue <- function(vec) {
        if (NaN %in% vec | NA %in% vec) {
            nan.pos <- is.nan(vec)
            na.pos <- is.na(vec)
            pos <- c(which(nan.pos), which(na.pos))
            vec.a <- vec[-pos]
            a <- sum(vec.a >= 0) / (length(vec) - sum(nan.pos | na.pos)) * 2
            b <- sum(vec.a <= 0) / (length(vec) - sum(nan.pos | na.pos)) * 2
        } else {
            a <- sum(vec >= 0) / length(vec) * 2
            b <- sum(vec <= 0) / length(vec) * 2
        }
        return(min(as.numeric(min(a, b)), 1))
    }

    if (se == TRUE) {
        if (is.null(seed) == FALSE) {
            set.seed(seed)
        }
        ## to store results
        est.boot <- matrix(NA, nboots, (p + 1))
        message("Bootstraping...\n")
        pb <- txtProgressBar(min = 0, max = nboots, initial = 0)
        for (i in 1:nboots) {
            smp <- sample(1:N, N, replace = TRUE)
            Y.boot <- Y[, smp]
            X.boot <- X[, smp, , drop = FALSE]
            I.boot <- I[, smp]
            W.boot <- NULL
            if (use_weight) {
                W.boot <- W[, smp]
            } else {
                W.boot <- as.matrix(0)
            }

            if (binary == FALSE) {
                if ((!0 %in% I) & (!use_weight)) {
                    inter.out <- try(inter_fe(
                        Y = Y.boot, X = X.boot, r = r, tol = tol,
                        force = force, beta0_in = beta0, max_iter = max_iteration
                    ), silent = TRUE)
                } else {
                    if (!use_weight) {
                        W.boot <- as.matrix(0)
                    }
                    Y0.boot <- Y0[, smp]
                    inter.out <- try(inter_fe_ub(
                        Y = Y.boot, Y0 = Y0.boot, X = X.boot, I = I.boot, W_in = W.boot, tol = tol,
                        beta0 = beta0, r = r, force = force, max_iter = max_iteration
                    ), silent = TRUE)
                }
            } else {
                data.ini.boot <- matrix(NA, (T * N), (2 + 1 + p))
                data.ini.boot[, 2] <- rep(1:N, each = T) ## unit fe
                data.ini.boot[, 3] <- rep(1:T, N) ## time fe
                data.ini.boot[, 1] <- c(Y.boot) ## outcome
                if (p > 0) { ## covar
                    for (j in 1:p) {
                        data.ini.boot[, (3 + j)] <- c(X.boot[, , j])
                    }
                }
                initialOut.boot <- BiInitialFit(data = data.ini.boot, QR = QR, r = r, force = force, oci = which(c(I.boot) == 1))
                Y0.boot <- initialOut.boot$Y0
                FE0.boot <- initialOut.boot$FE0
                if (QR == 1) {
                    xi0.boot <- initialOut.boot$xi0
                    factor0.boot <- initialOut.boot$factor0
                }

                if (QR == FALSE) {
                    if (!0 %in% I) {
                        inter.out <- try(inter_fe_d(Y.boot, Y0.boot, FE0.boot, X.boot, r = r, force, tol = tol), silent = TRUE)
                    } else {
                        inter.out <- try(inter_fe_d_ub(Y.boot, Y0.boot, FE0.boot, X.boot, I.boot, r = r, force, tol = tol), silent = TRUE)
                    }
                } else {
                    if (!0 %in% I) {
                        inter.out <- try(inter_fe_d_qr(Y.boot, Y0.boot, FE0.boot, factor0.boot, xi0.boot, X.boot, r = r, force, tol = tol), silent = TRUE)
                    } else {
                        inter.out <- try(inter_fe_d_qr_ub(Y.boot, Y0.boot, FE0.boot, factor0.boot, xi0.boot, X.boot, I.boot, r = r, force, tol = tol), silent = TRUE)
                    }
                }
            }

            if ("try-error" %in% class(inter.out)) {
                inter.out <- list(beta = NA, mu = NA)
            } else {
                if (is.null(norm.para)) {
                    est.boot[i, ] <- c(c(inter.out$beta), inter.out$mu)
                } else {
                    if (p > 0) {
                        est.boot[i, ] <-
                            c(
                                c(inter.out$beta),
                                inter.out$mu * norm.para[1]
                            )
                    } else {
                        est.boot[i, ] <- c(c(inter.out$beta), inter.out$mu * norm.para[1])
                    }
                }
            }

            setTxtProgressBar(pb, i)
        }
        close(pb)
        ## T*2: lower,upper
        CI <- t(apply(est.boot, 2, function(vec) {
            quantile(vec, c(0.025, 0.975), na.rm = TRUE)
        }))
        SE <- apply(est.boot, 2, sd, na.rm = TRUE)
        pvalue <- apply(est.boot, 2, get.pvalue)

        ## estimate table
        if (p > 0) {
            est.table <- cbind(c(beta, mu), SE, CI, pvalue)
        } else {
            est.table <- cbind(c(mu), SE, CI, pvalue)
        }

        colnames(est.table) <- c("Coef", "S.E.", "CI.lower", "CI.upper", "p.value")
    } else {
        if (p > 0) {
            est.table <- as.matrix(c(beta, mu))
        } else {
            est.table <- as.matrix(c(mu))
        }
    }
    rownames(est.table) <- c(Xname, "_const")

    ## -------------------------------#
    ## Storage
    ## -------------------------------#
    if (!is.null(norm.para)) {
        out$mu <- out$mu * norm.para[1]
        if (p > 0) {
            out$beta <- out$beta
        }
        if (r > 0) {
            out$lambda <- out$lambda * norm.para[1]
            out$VNT <- out$VNT * norm.para[1]
        }
        if (force %in% c(1, 3)) {
            out$alpha <- out$alpha * norm.para[1]
        }
        if (force %in% c(2, 3)) {
            out$xi <- out$xi * norm.para[1]
        }
        out$IC <- out$IC - log(out$sigma2) + log(out$sigma2 * (norm.para[1]^2))
        out$PC <- out$PC * (norm.para[1]^2)
        out$sigma2 <- out$sigma2 * (norm.para[1]^2)
        out$residuals <- out$residuals * norm.para[1]
        out$fit <- out$fit * norm.para[1]
    }

    out <- c(out, list(
        dat.Y = Y,
        dat.X = X,
        Y = Yname,
        X = Xname,
        index = c(id, time)
    ))
    if (se == TRUE) {
        out <- c(out, list(
            est.table = est.table,
            est.boot = est.boot # bootstrapped coef.
        ))
    } else {
        out <- c(out, list(est.table = est.table))
    }
    class(out) <- "interFE"
    return(out)
}

###############################################
## Inference
###############################################
basic_ci_alpha <- function(theta, boots, alpha) {
  n_p <- length(theta)
  probs <- c(alpha / 2, 1 - alpha / 2)
  ci_mat <- t(sapply(seq_len(n_p), function(i) {
    vals <- boots[i, ]
    vals <- vals[!is.na(vals)]
    if (length(vals) > 1) {
      qs <- tryCatch(
        quantile(vals, probs = probs, na.rm = TRUE),
        error = function(e) c(NA_real_, NA_real_)
      )
      if (all(!is.na(qs))) {
        # reflect about theta[i]
        lower <- 2 * theta[i] - qs[2]
        upper <- 2 * theta[i] - qs[1]
        return(c(lower, upper))
      }
    }
    c(NA_real_, NA_real_)
  }))

  colnames(ci_mat) <- c("lower", "upper")
  ci_mat
}

## Reflected pivot CI on a (possibly H0-centered) bootstrap distribution.
## Mirrors the location-shift fix in R/po-estimands.R (commit b4e9fbf):
## when `shift = TRUE` (i.e., vartype == "parametric"), shift the boot
## distribution row-wise from mean ~ 0 (H0) to mean ~ theta (H1) before
## computing the basic interval [2*theta - q_high, 2*theta - q_low]. The
## shift is variance-preserving, so the SE computed elsewhere on the
## unshifted boot is unaffected; the H0-based p-value (computed via
## get.pvalue() on the unshifted boot) is also unchanged.
##
## - theta: vector of point estimates (length p)
## - boots: p x B matrix of bootstrap draws
## - alpha: significance level (e.g., 0.05 for 95% CI; 0.10 for 90% CI bound)
## - shift: if TRUE, apply the parametric H0 -> H1 location shift
.basic_ci_shifted <- function(theta, boots, alpha, shift = FALSE) {
  if (isTRUE(shift) && is.matrix(boots) && nrow(boots) > 0L) {
    rm <- rowMeans(boots, na.rm = TRUE)
    boots <- boots - rm + theta
  }
  ci_mat <- basic_ci_alpha(theta, boots, alpha)
  colnames(ci_mat) <- c("CI.lower", "CI.upper")
  ci_mat
}

## Single-CI version: theta scalar, boots vector. Same shift semantics.
.basic_ci_shifted_one <- function(theta, boots, alpha, shift = FALSE) {
  if (isTRUE(shift) && length(boots) > 0L) {
    boots <- boots - mean(boots, na.rm = TRUE) + theta
  }
  qs <- quantile(boots, c(1 - alpha / 2, alpha / 2), na.rm = TRUE)
  c(2 * theta - unname(qs[1]),
    2 * theta - unname(qs[2]))
}

# Reduce closure payload before parallel export by keeping only symbols
# that the function body actually references from its local frame.
trim_closure_env <- function(fun) {
  if (!is.function(fun)) {
    return(fun)
  }

  old_env <- environment(fun)
  if (is.null(old_env) || identical(old_env, emptyenv())) {
    return(fun)
  }

  globals <- codetools::findGlobals(fun, merge = FALSE)
  needed <- unique(c(globals$variables, globals$functions))
  if (length(needed) == 0) {
    return(fun)
  }

  env_names <- ls(old_env, all.names = TRUE)
  keep <- intersect(needed, env_names)
  if (length(keep) == 0) {
    return(fun)
  }

  new_env <- new.env(parent = parent.env(old_env))
  for (nm in keep) {
    assign(nm, get(nm, envir = old_env, inherits = FALSE), envir = new_env)
  }
  environment(fun) <- new_env
  fun
}

fect_boot <- function(
  Y,
  X,
  D,
  W,
  cl = NULL,
  I,
  II,
  cm=FALSE,
  II.cm=NULL,
  T.on,
  T.off = NULL,
  T.on.carry = NULL,
  T.on.balance = NULL,
  balance.period = NULL,
  method = "ife",
  degree = 2,
  cfe = NULL,
  X.extra.FE = NULL,
  X.Z = NULL,
  X.Q = NULL,
  X.gamma = NULL,
  X.kappa = NULL,
  Zgamma.id = NULL,
  kappaQ.id = NULL,
  ind.matrix = NULL,
  knots = NULL,
  criterion = "mspe",
  CV = 0,
  k = 20,
  cv.prop = 0.1,
  cv.method = "rolling",
  cv.nobs = 3,
  r = 0,
  r.end = 3,
  lambda = NULL,
  nlambda = 10,
  alpha = 0.05,
  binary = 0,
  QR = 0,
  force = 0,
  hasRevs = 1,
  tol = 1e-3,
  max.iteration = 1000,
  norm.para = NULL,
  placebo.period = NULL,
  placeboTest = 0,
  carryoverTest = 0,
  carryover.period = NULL,
  vartype = "bootstrap",
  para.error = "auto",
  quantile.CI = FALSE,
  nboots = 200,
  parallel = TRUE,
  cores = NULL,
  do_parallel_cv = FALSE,
  group.level = NULL,
  group = NULL,
  dis = 0,
  keep.sims = FALSE,
  time.component.from = "notyettreated",
  loading.bound = "none",
  gamma.loading = NULL,
  gamma.loading.grid = NULL,
  W.in.fit = TRUE,
  W.in.agg = TRUE
) {
  do_parallel_boot <- isTRUE(parallel) || "boot" %in% as.character(parallel)
  na.pos <- NULL
  TT <- dim(Y)[1]
  N <- dim(Y)[2]
  if (is.null(X) == FALSE) {
    p <- dim(X)[3]
  } else {
    p <- 0
  }
  if (is.null(W)) {
    W.use <- as.matrix(0)
    use_weight <- 0
  } else {
    W.use <- W
    W.use[which(II == 0)] <- 0
    use_weight <- 1
  }

  if (hasRevs == 1) {
    ## D.fake : check reversals
    D.fake <- apply(D, 2, function(vec) {
      cumsum(vec)
    })
    D.fake <- ifelse(D.fake > 0, 1, 0)
    D.fake[which(I == 0)] <- 0

    rev <- which(apply(D.fake == D, 2, sum) != TT)
    co <- which(apply(D, 2, sum) == 0)
    tr.all <- which(apply(D, 2, sum) > 0)
    tr <- tr.all[which(!tr.all %in% rev)]

    Nrev <- length(rev)
    Ntr <- length(tr)
    Nco <- length(co)
  } else {
    ## treatement indicator
    tr <- which(apply(D, 2, sum) > 0)
    co <- which(apply(D, 2, sum) == 0)
    Ntr <- length(tr)
    Nco <- length(co)
  }

  ## Internal: when `loading.bound != "none"`, normalize
  ## `method = "ife"` + `time.component.from = "nevertreated"` to
  ## `method = "gsynth"` so the bootstrap loop routes to fect_nevertreated
  ## (which threads loading.bound) instead of fect_fe (which would silently
  ## drop it). The default `loading.bound = "none"` path is left untouched
  ## to preserve byte-identical RNG sequencing with pre-feature behavior.
  if (!identical(loading.bound, "none") &&
      identical(method, "ife") &&
      identical(time.component.from, "nevertreated")) {
      method <- "gsynth"
  }

  ## estimation
  if (CV == 0) {
    if (method == "gsynth") {
      out <- fect_nevertreated(
        Y = Y,
        X = X,
        D = D,
        W = W,
        W.in.fit = W.in.fit,
        I = I,
        II = II,
        T.on = T.on,
        T.off = T.off,
        CV = 0,
        T.on.balance = T.on.balance,
        balance.period = balance.period,
        r = r,
        binary = binary,
        QR = QR,
        force = force,
        hasRevs = hasRevs,
        tol = tol,
        max.iteration = max.iteration,
        boot = 0,
        norm.para = norm.para,
        placebo.period = placebo.period,
        placeboTest = placeboTest,
        carryover.period = carryover.period,
        carryoverTest = carryoverTest,
        group.level = group.level,
        group = group,
        loading.bound      = loading.bound,
        gamma.loading      = gamma.loading,
        gamma.loading.grid = gamma.loading.grid
      )
    } else if (method == "ife") {
      out <- fect_fe(
        Y = Y,
        X = X,
        D = D,
        W = W,
        W.in.fit = W.in.fit,
        I = I,
        II = II,
        cm = cm,
        II.cm = II.cm,
        T.on = T.on,
        T.off = T.off,
        T.on.carry = T.on.carry,
        T.on.balance = T.on.balance,
        balance.period = balance.period,
        r.cv = r,
        binary = binary,
        QR = QR,
        force = force,
        hasRevs = hasRevs,
        tol = tol,
        max.iteration = max.iteration,
        boot = 0,
        norm.para = norm.para,
        placebo.period = placebo.period,
        placeboTest = placeboTest,
        carryover.period = carryover.period,
        carryoverTest = carryoverTest,
        group.level = group.level,
        group = group
      )
    } else if (method == "mc") {
      out <- try(
        fect_mc(
          Y = Y,
          X = X,
          D = D,
          W = W,
          W.in.fit = W.in.fit,
          I = I,
          II = II,
          T.on = T.on,
          T.off = T.off,
          T.on.carry = T.on.carry,
          T.on.balance = T.on.balance,
          balance.period = balance.period,
          lambda.cv = lambda,
          force = force,
          hasRevs = hasRevs,
          tol = tol,
          max.iteration = max.iteration,
          boot = 0,
          norm.para = norm.para,
          placebo.period = placebo.period,
          placeboTest = placeboTest,
          carryover.period = carryover.period,
          carryoverTest = carryoverTest,
          group.level = group.level,
          group = group
        ),
        silent = TRUE
      )
      if ("try-error" %in% class(out)) {
        stop("\nCannot estimate using full data with MC algorithm.\n")
      }
    } else if (method == "cfe" && time.component.from == "nevertreated") {
      out <- try(
        fect_nevertreated(
          Y = Y,
          X = X,
          D = D,
          W = W,
          W.in.fit = W.in.fit,
          I = I,
          II = II,
          T.on = T.on,
          T.off = T.off,
          CV = 0,
          T.on.balance = T.on.balance,
          balance.period = balance.period,
          r = r,
          binary = binary,
          QR = QR,
          force = force,
          hasRevs = hasRevs,
          tol = tol,
          max.iteration = max.iteration,
          boot = 0,
          norm.para = norm.para,
          placebo.period = placebo.period,
          placeboTest = placeboTest,
          carryover.period = carryover.period,
          carryoverTest = carryoverTest,
          group.level = group.level,
          group = group,
          method = "cfe",
          X.extra.FE = X.extra.FE,
          X.Z = X.Z,
          X.Q = X.Q,
          X.gamma = X.gamma,
          X.kappa = X.kappa,
          Zgamma.id = Zgamma.id,
          kappaQ.id = kappaQ.id
        ),
        silent = TRUE
      )
      if ("try-error" %in% class(out)) {
        print(out)
        stop("\nCannot estimate.\n")
      }
    } else if (method %in% c("cfe")) {
      out <- try(
        fect_cfe(
          Y = Y,
          X = X,
          D = D,
          W = W,
          W.in.fit = W.in.fit,
          X.extra.FE = X.extra.FE,
          X.Z = X.Z,
          X.Q = X.Q,
          X.gamma = X.gamma,
          X.kappa = X.kappa,
          Zgamma.id = Zgamma.id,
          kappaQ.id = kappaQ.id,
          I = I,
          II = II,
          cm = cm,
          II.cm = II.cm,
          T.on = T.on,
          T.off = T.off,
          T.on.carry = T.on.carry,
          T.on.balance = T.on.balance,
          balance.period = balance.period,
          r.cv = r,
          binary = binary,
          QR = QR,
          force = force,
          hasRevs = hasRevs,
          tol = tol,
          max.iteration = max.iteration,
          boot = 0,
          norm.para = norm.para,
          placebo.period = placebo.period,
          placeboTest = placeboTest,
          carryover.period = carryover.period,
          carryoverTest = carryoverTest,
          group.level = group.level,
          group = group
        ),
        silent = TRUE
      )
      # I.report <- out$I
      # II.report <- out$II
      if ("try-error" %in% class(out)) {
        print(out)
        stop("\nCannot estimate.\n")
      }
    }
  } else {
    ## cross-valiadtion
    if (binary == 0) {
      out <- fect_cv(
        Y = Y,
        X = X,
        D = D,
        W = W,
        W.in.fit = W.in.fit,
        I = I,
        II = II,
        T.on = T.on,
        T.off = T.off,
        T.on.carry = T.on.carry,
        T.on.balance = T.on.balance,
        balance.period = balance.period,
        method = method,
        criterion = criterion,
        k = k,
        r = r,
        r.end = r.end,
        nlambda = nlambda,
        lambda = lambda,
        force = force,
        hasRevs = hasRevs,
        tol = tol,
        max.iteration = max.iteration,
        norm.para = norm.para,
        group.level = group.level,
        group = group,
        cv.prop = cv.prop,
        cv.method = cv.method,
        cv.nobs = cv.nobs,
        time.component.from = time.component.from,
        X.extra.FE = X.extra.FE,
        X.Z = X.Z,
        X.Q = X.Q,
        X.gamma = X.gamma,
        X.kappa = X.kappa,
        Zgamma.id = Zgamma.id,
        kappaQ.id = kappaQ.id,
        parallel = parallel,
        cores = cores,
        do_parallel_cv = do_parallel_cv
      )

      if (!is.null(out$method)) {
        method <- out$method
      }
    } else {
      out <- fect_binary_cv(
        Y = Y,
        X = X,
        D = D,
        I = I,
        II = II,
        T.on = T.on,
        T.off = T.off,
        k = k,
        r = r,
        r.end = r.end,
        QR = QR,
        force = force,
        hasRevs = hasRevs,
        tol = tol,
        group.level = group.level,
        group = group
      )
      if (!is.null(out$method)) {
        method <- out$method
      }
    }
  }

  ## output
  validX <- out$validX
  eff <- out$eff
  att.avg <- out$att.avg
  att.avg.unit <- out$att.avg.unit
  calendar.eff <- out$eff.calendar
  calendar.eff.fit <- out$eff.calendar.fit
  calendar.N <- out$N.calendar

  group.att <- out$group.att

  att <- out$att
  time.on <- out$time
  target.enp <- out$calendar.enp

  time.off <- NULL
  if (hasRevs == 1) {
    att.off <- out$att.off
    time.off <- out$time.off
  }
  carry.att <- carry.time <- NULL
  if (!is.null(T.on.carry)) {
    carry.att <- out$carry.att
    carry.time <- out$carry.time
  }

  if (!is.null(balance.period)) {
    balance.att <- out$balance.att
    balance.time <- out$balance.time
    balance.count <- out$balance.count
    balance.avg.att <- out$balance.avg.att
    if (!is.null(placebo.period) & placeboTest == TRUE) {
      balance.att.placebo <- out$balance.att.placebo
    }
  }

  if (!is.null(W)) {
    att.avg.W <- out$att.avg.W
    att.on.sum.W <- out$att.on.sum.W
    att.on.W <- out$att.on.W
    count.on.W <- out$count.on.W
    time.on.W <- out$time.on.W
    W.on.sum <- out$W.on.sum
    if (!is.null(placebo.period) & placeboTest == TRUE) {
      att.placebo.W <- out$att.placebo.W
    }
    if (hasRevs == 1) {
      att.off.sum.W <- out$att.off.sum.W
      att.off.W <- out$att.off.W
      count.off.W <- out$count.off.W
      time.off.W <- out$time.off.W
      W.off.sum <- out$W.off.sum
    }
    if (!is.null(carryover.period) & carryoverTest == TRUE) {
      att.carryover.W <- out$att.carryover.W
    }
  }

  eff.out <- out$eff
  fit.out <- out$Y.ct
  N_unit <- dim(out$res)[2]

  if (!is.null(group)) {
    group.output.origin <- out$group.output
    group.output.name <- names(out$group.output)
    group.time.on <- list()
    group.time.off <- list()
    for (sub.name in group.output.name) {
      group.time.on[[sub.name]] <- out$group.output[[sub.name]]$time.on
      if (hasRevs == 1) {
        group.time.off[[sub.name]] <- out$group.output[[sub.name]]$time.off
      }
    }
  }

  if (p > 0) {
    beta <- out$beta
  } else {
    beta <- matrix(0, 1, 0)
  }

  ## Cache the CV-selected gamma from the main fit, so bootstrap replications
  ## reuse it instead of re-running CV each replication. Re-running CV per
  ## bootstrap is both slow and numerically unstable (each resample constructs
  ## its own fold set on perturbed data; at T0 >= 40 the inner solver
  ## occasionally hits ill-conditioned cases). Using a fixed, main-fit-CV
  ## gamma across bootstrap reps is the standard practice for nested
  ## CV + bootstrap and matches how r is handled in IFE CV.
  if (identical(loading.bound, "simplex") &&
      is.null(gamma.loading) &&
      !is.null(out$gamma.loading) &&
      is.finite(out$gamma.loading)) {
    gamma.loading <- out$gamma.loading
  }

  if (is.null(cl)) {
    cl.unique <- NULL
  } else {
    cl.unique <- unique(c(cl))
  }

  if (vartype == "jackknife") {
    nboots <- N
    if (N > 1000) {
      warning(
        "vartype = \"jackknife\" with N = ", N, " requires ", N,
        " leave-one-out refits and will be slow at the v2.4.2 EM ",
        "convergence defaults (tol = 1e-5, max.iteration = 5000). ",
        "Consider vartype = \"bootstrap\" (B = 1000 typically faster ",
        "than full leave-one-out at N > 500) for tractability.",
        call. = FALSE
      )
    }
  }

  ## bootstrapped estimates
  if (keep.sims) {
    if (vartype == "jackknife") {
      D.boot <- array(NA, dim = c(TT, N - 1, nboots))
      I.boot <- array(NA, dim = c(TT, N - 1, nboots))
      eff.boot <- array(0, dim = c(TT, N - 1, nboots)) ## to store results
    } else {
      D.boot <- array(NA, dim = c(TT, N, nboots))
      I.boot <- array(NA, dim = c(TT, N, nboots))
      eff.boot <- array(0, dim = c(TT, N, nboots)) ## to store results
    }
    colnames.boot <- c()
  }

  att.avg.boot <- matrix(0, 1, nboots)
  att.avg.unit.boot <- matrix(0, 1, nboots)
  att.boot <- matrix(0, length(time.on), nboots)
  att.count.boot <- matrix(0, length(time.on), nboots)
  beta.boot <- marginal.boot <- att.off.boot <- att.off.count.boot <- NULL
  calendar.eff.boot <- matrix(0, TT, nboots)
  calendar.eff.fit.boot <- matrix(0, TT, nboots)

  if (hasRevs == 1) {
    eff.off.boot <- c()
    att.off.boot <- matrix(0, length(time.off), nboots)
    att.off.count.boot <- matrix(0, length(time.off), nboots)
  }
  if (!is.null(T.on.carry)) {
    carry.att.boot <- matrix(0, length(carry.att), nboots)
  }
  if (!is.null(balance.period)) {
    balance.att.boot <- matrix(0, length(balance.att), nboots) # dynamic att
    balance.count.boot <- matrix(0, length(balance.att), nboots)
    balance.avg.att.boot <- matrix(0, 1, nboots)
    if (!is.null(placebo.period) & placeboTest == TRUE) {
      balance.att.placebo.boot <- matrix(0, 1, nboots)
    }
  }

  if (!is.null(W)) {
    att.avg.W.boot <- matrix(0, 1, nboots)
    att.on.W.boot <- matrix(0, length(time.on.W), nboots)
    att.on.count.W.boot <- matrix(0, length(time.on.W), nboots)
    if (!is.null(placebo.period) & placeboTest == TRUE) {
      att.placebo.W.boot <- matrix(0, 1, nboots)
    }
    if (hasRevs == 1) {
      att.off.W.boot <- matrix(0, length(time.off.W), nboots)
      att.off.count.W.boot <- matrix(0, length(time.off.W), nboots)
      if (!is.null(carryover.period) & carryoverTest == TRUE) {
        att.carryover.W.boot <- matrix(0, 1, nboots)
      }
    }
  }

  if (p > 0) {
    beta.boot <- matrix(0, p, nboots)
    if (binary == TRUE) {
      marginal.boot <- matrix(0, p, nboots)
    }
  }
  if (!is.null(placebo.period) & placeboTest == TRUE) {
    att.placebo.boot <- matrix(0, 1, nboots)
  }
  if (!is.null(carryover.period) & carryoverTest == TRUE) {
    att.carryover.boot <- matrix(0, 1, nboots)
  }

  group.att.boot <- NULL
  if (!is.null(group.att)) {
    group.att.boot <- matrix(0, length(group.att), nboots)
  }

  group.atts.boot <- NULL
  group.atts.off.boot <- NULL
  group.atts.placebo.boot <- NULL
  group.atts.carryover.boot <- NULL
  if (!is.null(group)) {
    group.atts.boot <- list()
    group.atts.off.boot <- list()
    group.att.placebo.boot <- list()
    group.att.carryover.boot <- list()
    for (sub.name in group.output.name) {
      subgroup.time.on <- group.time.on[[sub.name]]
      group.atts.boot[[sub.name]] <- matrix(0, length(subgroup.time.on), nboots)
      if (hasRevs == 1) {
        subgroup.time.off <- group.time.off[[sub.name]]
        group.atts.off.boot[[sub.name]] <- matrix(
          0,
          length(subgroup.time.off),
          nboots
        )
      }
      if (placeboTest) {
        group.att.placebo.boot[[sub.name]] <- matrix(0, 1, nboots)
      }
      if (carryoverTest) {
        group.att.carryover.boot[[sub.name]] <- matrix(0, 1, nboots)
      }
    }
  }

  if (dis) {
    if (vartype == "jackknife") {
      message("Jackknife estimates ... ")
    } else {
      message("Bootstrapping for uncertainties ... ")
    }
  }

  if (binary == TRUE & vartype == "parametric") {
    one.nonpara <- function(num = NULL) {
      Y.boot <- Y
      Y.fit <- out$Y.ct
      Y.fit[which(is.na(Y.fit))] <- 0
      Y.boot.new <- matrix(
        sapply(1:length(c(Y.fit)), function(i) {
          rbinom(1, 1, c(Y.fit)[i])
        }),
        TT,
        N
      )
      Y.boot[which(II == 1)] <- Y.boot.new[which(II == 1)]

      placebo.period.boot <- NULL
      carryover.period.boot <- NULL
      if (placeboTest == TRUE) {
        placebo.period.boot <- placebo.period
      }
      if (carryoverTest == TRUE) {
        carryover.period.boot <- carryover.period
      }

      boot <- try(
        fect_fe(
          Y = Y.boot,
          X = X,
          D = D,
          W = W,
          W.in.fit = W.in.fit,
          I = I,
          II = II,
          cm = cm,
          II.cm = II.cm,
          T.on = T.on,
          T.off = T.off,
          T.on.carry = T.on.carry,
          T.on.balance = T.on.balance,
          balance.period = balance.period,
          r.cv = out$r.cv,
          binary = binary,
          QR = QR,
          force = force,
          hasRevs = hasRevs,
          tol = tol,
          max.iteration = max.iteration,
          boot = 1,
          norm.para = norm.para,
          calendar.enp.seq = target.enp,
          time.on.seq = time.on,
          time.off.seq = time.off,
          time.on.carry.seq = carry.time,
          time.on.balance.seq = balance.time,
          time.on.seq.W = time.on.W,
          time.off.seq.W = time.off.W,
          placebo.period = placebo.period.boot,
          placeboTest = placeboTest,
          carryoverTest = carryoverTest,
          carryover.period = carryover.period.boot,
          group.level = group.level,
          group = group
        ),
        silent = TRUE
      )

      if ("try-error" %in% class(boot)) {
        boot0 <- list(
          att.avg = NA,
          att = NA,
          count = NA,
          beta = NA,
          att.off = NA,
          count.off = NA,
          eff.calendar = NA,
          eff.calendar.fit = NA,
          att.placebo = NA,
          att.avg.unit = NA,
          att.carryover = NA,
          group.att = NA,
          marginal = NA,
          carry.att = NA,
          balance.att = NA,
          balance.att.placebo = NA,
          balance.count = NA,
          att.avg.W = NA,
          att.on.W = NA,
          count.on.W = NA,
          time.on.W = NA,
          att.placebo.W = NA,
          att.off.W = NA,
          count.off.W = NA,
          time.off.W = NA,
          att.carryover.W = NA,
          balance.avg.att = NA,
          balance.time = NA,
          group.output = list(),
          eff = if (keep.sims) matrix(NA_real_, TT, N) else NULL,
          D = if (keep.sims) matrix(NA_real_, TT, N) else NULL,
          I = if (keep.sims) matrix(NA_real_, TT, N) else NULL,
          boot.id = NULL
        )
        return(boot0)
      } else {
        return(boot)
      }
    }
  } else if (
    binary == FALSE & method %in% c("gsynth", "ife", "cfe") & vartype == "parametric"
  ) {
    ## Resolve para.error = "auto" to the concrete mode for this dataset.
    para.error.resolved <- if (identical(para.error, "auto")) {
        if (0 %in% I) "ar" else "empirical"
    } else {
        para.error
    }
    message("Parametric Bootstrap (para.error = \"", para.error.resolved, "\") \n")
    sum.D <- colSums(out$D)
    id.tr <- which(sum.D > 0)
    I.tr <- as.matrix(out$I[, id.tr])
    id.co <- which(sum.D == 0)
    Nco <- length(id.co)
    Ntr <- length(id.tr)

    fit.out[which(out$I == 0)] <- 0

    error.co <- out$res.full[, id.co]
    I.co <- out$I[, id.co]
    valid.co <- valid_controls(out, method, time.component.from, force)

    draw.error <- function() {
      repeat {
        fake.tr <- sample(id.co, 1, replace = FALSE)
        if (fake.tr %in% valid.co) {
          break
        }
      }

      id.co.rest <- id.co[which(!id.co %in% fake.tr)]
      repeat {
        id.co.pseudo <- sample(id.co.rest, Nco, replace = TRUE)
        if (sum(apply(as.matrix(out$I[, id.co.pseudo]), 1, sum) >= 1) == TT) {
          break
        }
      }

      id.pseudo <- c(rep(fake.tr, Ntr), id.co.pseudo) ## Ntr + ...
      I.id.pseudo <- out$I[, id.pseudo]
      II.id.pseudo <- out$II[, id.pseudo]
      ## obtain the prediction eror
      D.pseudo <- out$D[, c(id.tr, id.co.pseudo)] ## fake.tr + control left
      Y.pseudo <- out$Y[, id.pseudo]
      T.on.pseudo <- T.on[, id.pseudo]

      X.pseudo <- NULL
      if (p > 0) {
        X.pseudo <- X[, id.pseudo, , drop = FALSE]
      }

      ## Subset CFE arrays to pseudo-sample indices
      X.extra.FE.pseudo <- if (!is.null(X.extra.FE) && length(dim(X.extra.FE)) == 3 && dim(X.extra.FE)[2] > 0) {
          X.extra.FE[, id.pseudo, , drop = FALSE]
      } else {
          X.extra.FE
      }
      X.Z.pseudo <- if (!is.null(X.Z) && length(dim(X.Z)) == 3 && dim(X.Z)[2] > 0) {
          X.Z[, id.pseudo, , drop = FALSE]
      } else {
          X.Z
      }
      X.Q.pseudo <- if (!is.null(X.Q) && length(dim(X.Q)) == 3 && dim(X.Q)[2] > 0) {
          X.Q[, id.pseudo, , drop = FALSE]
      } else {
          X.Q
      }
      X.gamma.pseudo <- if (!is.null(X.gamma) && length(dim(X.gamma)) == 3 && dim(X.gamma)[2] > 0) {
          X.gamma[, id.pseudo, , drop = FALSE]
      } else {
          X.gamma
      }
      X.kappa.pseudo <- if (!is.null(X.kappa) && length(dim(X.kappa)) == 3 && dim(X.kappa)[2] > 0) {
          X.kappa[, id.pseudo, , drop = FALSE]
      } else {
          X.kappa
      }

      ## output
      synth.out <- try(
        impute_Y0(
          method     = method,
          predictive = time.component.from,
          Y          = Y.pseudo,
          X          = X.pseudo,
          D          = D.pseudo,
          W          = NULL,
          I          = I.id.pseudo,
          II         = II.id.pseudo,
          T.on       = T.on.pseudo,
          X.extra.FE = X.extra.FE.pseudo,
          X.Z        = X.Z.pseudo,
          X.Q        = X.Q.pseudo,
          X.gamma    = X.gamma.pseudo,
          X.kappa    = X.kappa.pseudo,
          Zgamma.id  = Zgamma.id,
          kappaQ.id  = kappaQ.id,
          tuning     = out$r.cv,
          boot       = 1,
          force      = force,
          hasRevs    = hasRevs,
          tol        = tol,
          max.iteration = max.iteration,
          norm.para  = norm.para
        ),
        silent = TRUE
      )

      if ("try-error" %in% class(synth.out)) {
        return(matrix(NA, TT, Ntr))
      } else {
        if ("eff" %in% names(synth.out)) {
          if (is.null(norm.para)) {
            output <- synth.out$eff.tr
          } else {
            output <- synth.out$eff.tr / norm.para[1]
          }

          return(as.matrix(output)) ## TT * Ntr
        } else {
          return(matrix(NA, TT, Ntr))
        }
      }
    }

    message("\rSimulating errors ...")
    if (do_parallel_boot) {
      ## Phase A: future_lapply (was foreach %dopar%, which inherited whatever
      ## backend the global foreach registry held — see notes/ stage-1).
      error.list <- .fect_with_quiet_pkg_warnings(
        future.apply::future_lapply(
          seq_len(nboots),
          FUN = function(j) draw.error(),
          future.seed = TRUE,
          future.packages = c("fect", "mvtnorm", "fixest")
        )
      )
      error.tr <- abind(error.list, along = 3)
    } else {
      error.tr <- array(NA, dim = c(TT, Ntr, nboots))
      for (j in 1:nboots) {
        error.tr[,, j] <- draw.error()
        # if (j %% 100 == 0) {
        #     message(".")
        # }
      }
    }

    if (para.error.resolved == "ar") {
      ## calculate vcov of ep_tr
      na.sum <- sapply(1:nboots, function(vec) {
        sum(is.na(c(error.tr[,, vec])))
      })
      na.rm <- na.sum == TT * Ntr
      na.rm.count <- sum(na.rm)
      rm.pos <- which(na.rm == TRUE)

      if (na.rm.count > 0) {
        if (na.rm.count == nboots) {
          stop("fail to simulate errors.\n")
        }
        error.tr <- error.tr[,, -rm.pos, drop = FALSE]
      }

      error.tr.adj <- array(NA, dim = c(TT, nboots - na.rm.count, Ntr))
      for (i in 1:Ntr) {
        error.tr.adj[,, i] <- error.tr[, i, ]
      }
      vcov_tr <- array(NA, dim = c(TT, TT, Ntr))
      for (i in 1:Ntr) {
        vcov_tr[,, i] <- res.vcov(res = error.tr.adj[,, i], cov.ar = TT - 1L)
        vcov_tr[,, i][is.na(vcov_tr[,, i]) | is.nan(vcov_tr[,, i])] <- 0
      }
      ## calculate vcov of e_co
      vcov_co <- res.vcov(res = error.co, cov.ar = TT - 1L)
      vcov_co[is.na(vcov_co) | is.nan(vcov_co)] <- 0
    }

    one.nonpara <- function(num = NULL) {
      ## boostrap ID
      repeat {
        fake.co <- sample(id.co, Nco, replace = TRUE)
        if (sum(apply(as.matrix(II[, fake.co]), 1, sum) >= 1) == TT) {
          break
        }
      }
      id.boot <- c(id.tr, fake.co)
      r.boot <- out$r.cv
      if (!is.null(r.boot) && length(r.boot) == 1 && !is.na(r.boot)) {
        n.unit.boot <- length(id.boot)
        r.boot <- min(r.boot, TT, n.unit.boot)
      }

      ## get the error for the treated and control
      error.tr.boot <- matrix(NA, TT, Ntr)
      if (para.error.resolved == "ar") {
        ## Path AR: draw from MVN with AR-vcov estimated from Loop 1 pool.
        ## Works for fully-observed and missing-data panels.
        for (w in 1:Ntr) {
          error.tr.boot[, w] <- t(rmvnorm(
            n = 1,
            rep(0, TT),
            vcov_tr[,, w],
            method = "svd"
          ))
        }
        error.tr.boot[which(I.tr == 0)] <- 0
        error.co.boot <- t(rmvnorm(
          n = Nco,
          rep(0, TT),
          vcov_co,
          method = "svd"
        ))
        error.co.boot[which(as.matrix(I[, fake.co]) == 0)] <- 0

      } else if (para.error.resolved == "empirical") {
        ## Path empirical: column-resample from the Loop 1 pool.
        ## Requires fully-observed panel (validated at fit time).
        for (w in 1:Ntr) {
          error.tr.boot[, w] <- error.tr[,
            w,
            sample(1:nboots, 1, replace = TRUE)
          ]
        }
        error.co.boot <- error.co[, sample(1:Nco, Nco, replace = TRUE)]

      } else {
        ## Path wild: unit-level Rademacher sign-flip on Loop 1 pool draws.
        ## Preserves within-unit AR structure (sign applied to entire time series).
        ## Requires fully-observed panel (validated at fit time).
        ## This is variant-(i): treated cells receive error.tr.boot (from Loop 1 pool),
        ## NOT the observed treatment effect. The bootstrap distribution is H0-centered.
        ## The po-estimands.R location-shift (commit b4e9fbf) re-centers at theta-hat.
        signs    <- sample(c(-1, 1), Ntr, replace = TRUE)
        co_signs <- sample(c(-1, 1), Nco, replace = TRUE)

        for (w in 1:Ntr) {
          j <- sample(1:dim(error.tr)[3], 1, replace = TRUE)
          error.tr.boot[, w] <- signs[w] * error.tr[, w, j]
        }

        co_picks      <- sample(1:Nco, Nco, replace = TRUE)
        error.co.boot <- error.co[, co_picks, drop = FALSE]
        ## Apply per-unit sign to entire column (unit's full time series).
        ## t(t(M) * v) multiplies column k of M by v[k].
        error.co.boot <- t(t(error.co.boot) * co_signs)
      }

      Y.boot <- fit.out[, id.boot]
      Y.boot[, 1:Ntr] <- as.matrix(Y.boot[, 1:Ntr] + error.tr.boot)
      Y.boot[, (Ntr + 1):length(id.boot)] <- Y.boot[,
        (Ntr + 1):length(id.boot)
      ] +
        error.co.boot
      X.boot <- NULL
      if (p > 0) {
        X.boot <- X[, id.boot, , drop = FALSE]
      }
      D.boot <- out$D[, id.boot, drop = FALSE]
      I.boot <- out$I[, id.boot, drop = FALSE]
      II.boot <- out$II[, id.boot, drop = FALSE]
      II.cm.boot <- out$II.cm[, id.boot, drop = FALSE]
      W.boot <- NULL
      if (!is.null(W)) {
        W.boot <- NULL
      }
      boot.group <- NULL
      if (!is.null(group)) {
        if (is.matrix(group) || length(dim(group)) == 2) {
          boot.group <- group[, id.boot, drop = FALSE]
        } else {
          # Fallback: if group is a length-N vector, align it to boot.id
          boot.group <- group[id.boot]
        }
      }

      ## Extra-FE slicing for cfe branches (new for Phase 3b)
      X.extra.FE.boot <- if (!is.null(X.extra.FE) && length(dim(X.extra.FE)) == 3 && dim(X.extra.FE)[2] > 0) {
          X.extra.FE[, id.boot, , drop = FALSE]
      } else {
          X.extra.FE
      }
      X.Z.boot <- if (!is.null(X.Z) && length(dim(X.Z)) == 3 && dim(X.Z)[2] > 0) {
          X.Z[, id.boot, , drop = FALSE]
      } else {
          X.Z
      }
      X.Q.boot <- if (!is.null(X.Q) && length(dim(X.Q)) == 3 && dim(X.Q)[2] > 0) {
          X.Q[, id.boot, , drop = FALSE]
      } else {
          X.Q
      }
      X.gamma.boot <- if (!is.null(X.gamma) && length(dim(X.gamma)) == 3 && dim(X.gamma)[2] > 0) {
          X.gamma[, id.boot, , drop = FALSE]
      } else {
          X.gamma
      }
      X.kappa.boot <- if (!is.null(X.kappa) && length(dim(X.kappa)) == 3 && dim(X.kappa)[2] > 0) {
          X.kappa[, id.boot, , drop = FALSE]
      } else {
          X.kappa
      }

      synth.out <- try(
        impute_Y0(
          method     = method,
          predictive = time.component.from,
          Y          = Y.boot,
          X          = X.boot,
          D          = D.boot,
          W          = W.boot,
          I          = I.boot,
          II         = II.boot,
          T.on       = T.on[, id.boot],
          T.on.balance = T.on.balance[, id.boot],
          balance.period = balance.period,
          X.extra.FE = X.extra.FE.boot,
          X.Z        = X.Z.boot,
          X.Q        = X.Q.boot,
          X.gamma    = X.gamma.boot,
          X.kappa    = X.kappa.boot,
          Zgamma.id  = Zgamma.id,
          kappaQ.id  = kappaQ.id,
          tuning     = out$r.cv,
          boot       = 1,
          placeboTest = placeboTest,
          placebo.period = placebo.period,
          carryover.period = carryover.period,
          carryoverTest = carryoverTest,
          calendar.enp.seq = target.enp,
          time.on.seq      = time.on,
          time.off.seq     = time.off,
          time.on.seq.W    = time.on.W,
          time.off.seq.W   = time.off.W,
          time.on.seq.group    = group.time.on,
          time.off.seq.group   = group.time.off,
          time.on.balance.seq  = balance.time,
          norm.para    = norm.para,
          tol          = tol,
          max.iteration = max.iteration,
          group.level  = group.level,
          group        = boot.group,
          force        = force,
          hasRevs      = hasRevs
        ),
        silent = TRUE
      )

      if ("try-error" %in% class(synth.out)) {
        boot0 <- list(
          att.avg = NA,
          att = NA,
          count = NA,
          beta = NA,
          att.off = NA,
          count.off = NA,
          eff.calendar = NA,
          eff.calendar.fit = NA,
          att.placebo = NA,
          att.avg.unit = NA,
          att.carryover = NA,
          group.att = NA,
          marginal = NA,
          balance.att = NA,
          balance.att.placebo = NA,
          balance.count = NA,
          balance.avg.att = NA,
          balance.time = NA,
          att.avg.W = NA,
          att.on.W = NA,
          count.on.W = NA,
          time.on.W = NA,
          att.placebo.W = NA,
          att.off.W = NA,
          count.off.W = NA,
          time.off.W = NA,
          att.carryover.W = NA,
          group.output = list(),
          eff = if (keep.sims) matrix(NA_real_, TT, N) else NULL,
          D = if (keep.sims) matrix(NA_real_, TT, N) else NULL,
          I = if (keep.sims) matrix(NA_real_, TT, N) else NULL,
          boot.id = NULL
        )
        return(boot0)
      } else {
        synth.out$boot.id <- id.boot
        return(synth.out)
      }
    }
  } else {
    one.nonpara <- function(num = NULL) {
      ## Y.input is what gets passed to the per-method bootstrap refit.
      ## For case bootstrap and jackknife it is just the original Y.
      Y.input <- Y
      if (is.null(num)) {
        ## case bootstrap (resample units with replacement)
        if (is.null(cl)) {
          if (hasRevs == 0) {
            if (Nco > 0) {
              repeat {
                fake.co <- sample(co, Nco, replace = TRUE)
                fake.tr <- sample(tr, Ntr, replace = TRUE)
                boot.id <- c(fake.tr, fake.co)
                if (sum(apply(as.matrix(II[, boot.id]), 1, sum) >= 1) == TT) {
                  break
                }
              }
            } else {
              repeat {
                boot.id <- sample(tr, Ntr, replace = TRUE)
                if (sum(apply(as.matrix(II[, boot.id]), 1, sum) >= 1) == TT) {
                  break
                }
              }
            }
          } else {
            if (Ntr > 0) {
              if (Nco > 0) {
                repeat {
                  fake.co <- sample(co, Nco, replace = TRUE)
                  fake.tr <- sample(tr, Ntr, replace = TRUE)
                  fake.rev <- sample(rev, Nrev, replace = TRUE)
                  boot.id <- c(fake.rev, fake.tr, fake.co)
                  if (sum(apply(as.matrix(II[, boot.id]), 1, sum) >= 1) == TT) {
                    break
                  }
                }
              } else {
                repeat {
                  fake.tr <- sample(tr, Ntr, replace = TRUE)
                  fake.rev <- sample(rev, Nrev, replace = TRUE)
                  boot.id <- c(fake.rev, fake.tr)
                  if (sum(apply(as.matrix(II[, boot.id]), 1, sum) >= 1) == TT) {
                    break
                  }
                }
              }
            } else {
              if (Nco > 0) {
                repeat {
                  fake.co <- sample(co, Nco, replace = TRUE)
                  fake.rev <- sample(rev, Nrev, replace = TRUE)
                  boot.id <- c(fake.rev, fake.co)
                  if (sum(apply(as.matrix(II[, boot.id]), 1, sum) >= 1) == TT) {
                    break
                  }
                }
              } else {
                repeat {
                  boot.id <- sample(rev, Nrev, replace = TRUE)
                  if (sum(apply(as.matrix(II[, boot.id]), 1, sum) >= 1) == TT) {
                    break
                  }
                }
              }
            }
          }
        } else {
          cl.id <- c(apply(cl, 2, mean))
          cl.boot <- sample(cl.unique, length(cl.unique), replace = TRUE)
          cl.boot.uni <- unique(cl.boot)
          cl.boot.count <- as.numeric(table(cl.boot))
          boot.id <- c()
          for (kk in 1:length(cl.boot.uni)) {
            boot.id <- c(
              boot.id,
              rep(which(cl.id == cl.boot.uni[kk]), cl.boot.count[kk])
            )
          }
        }

        boot.group <- group[, boot.id]
      } else {
        ## jackknife
        boot.group <- group[, -num]
        boot.id <- 1:N
        boot.id <- boot.id[-num]
      }

      r.boot <- 0
      if (!is.null(out$r.cv) && length(out$r.cv) >= 1 && !is.na(out$r.cv[1])) {
        r.boot <- min(as.numeric(out$r.cv[1]), TT, length(boot.id))
      }

      X.boot <- X[, boot.id, , drop = FALSE]
      D.boot <- D[, boot.id]
      I.boot <- I[, boot.id]
      W.boot <- NULL
      if (!is.null(W)) {
        W.boot <- W[, boot.id]
      }

      if (
        sum(c(D.boot) == 0) == 0 |
          sum(c(D.boot) == 1) == 0 |
          sum(c(I.boot) == 1) == 0
      ) {
        boot0 <- list(
          att.avg = NA,
          att = NA,
          count = NA,
          beta = NA,
          att.off = NA,
          count.off = NA,
          eff.calendar = NA,
          eff.calendar.fit = NA,
          att.placebo = NA,
          att.avg.unit = NA,
          att.carryover = NA,
          group.att = list(),
          balance.att = NA,
          balance.att.placebo = NA,
          balance.count = NA,
          balance.avg.att = NA,
          balance.time = NA,
          att.avg.W = NA,
          att.on.W = NA,
          count.on.W = NA,
          time.on.W = NA,
          att.placebo.W = NA,
          att.off.W = NA,
          count.off.W = NA,
          time.off.W = NA,
          att.carryover.W = NA,
          group.out = list(),
          eff = matrix(NA_real_, TT, length(boot.id)),
          D = matrix(NA_real_, TT, length(boot.id)),
          I = matrix(NA_real_, TT, length(boot.id)),
          boot.id = boot.id
        )
        return(boot0)
      } else {
        T.off.boot <- NULL
        if (hasRevs == TRUE) {
          T.off.boot <- T.off[, boot.id]
        }
        placebo.period.boot <- NULL
        if (placeboTest == TRUE) {
          placebo.period.boot <- placebo.period
        }
        carryover.period.boot <- NULL
        if (carryoverTest == TRUE) {
          carryover.period.boot <- carryover.period
        }

        boot <- NULL
        if (method == "gsynth") {
          boot <- try(
            fect_nevertreated(
              Y = Y.input[, boot.id],
              X = X.boot,
              D = D.boot,
              W = W.boot,
              W.in.fit = W.in.fit,
              I = I.boot,
              II = II[, boot.id],
              T.on = T.on[, boot.id],
              T.off = T.off.boot,
              CV = 0,
              T.on.balance = T.on.balance[, boot.id],
              balance.period = balance.period,
              r = r.boot,
              binary = binary,
              QR = QR,
              force = force,
              hasRevs = hasRevs,
              tol = tol,
              max.iteration = max.iteration,
              boot = 1,
              norm.para = norm.para,
              calendar.enp.seq = target.enp,
              time.on.seq = time.on,
              time.off.seq = time.off,
              time.on.seq.W = time.on.W,
              time.off.seq.W = time.off.W,
              placebo.period = placebo.period.boot,
              placeboTest = placeboTest,
              time.on.balance.seq = balance.time,
              carryoverTest = carryoverTest,
              carryover.period = carryover.period.boot,
              group.level = group.level,
              group = boot.group,
              time.on.seq.group = group.time.on,
              time.off.seq.group = group.time.off,
              loading.bound      = loading.bound,
              gamma.loading      = gamma.loading,
              gamma.loading.grid = gamma.loading.grid
            ),
            silent = TRUE
          )
        } else if (method == "ife") {
          boot <- try(
            fect_fe(
              Y = Y.input[, boot.id],
              X = X.boot,
              D = D.boot,
              W = W.boot,
              W.in.fit = W.in.fit,
              I = I.boot,
              II = II[, boot.id],
              cm = cm,
              II.cm = II.cm[, boot.id],
              T.on = T.on[, boot.id],
              T.off = T.off.boot,
              T.on.carry = T.on.carry[, boot.id],
              T.on.balance = T.on.balance[, boot.id],
              balance.period = balance.period,
              r.cv = r.boot,
              binary = binary,
              QR = QR,
              force = force,
              hasRevs = hasRevs,
              tol = tol,
              max.iteration = max.iteration,
              boot = 1,
              norm.para = norm.para,
              calendar.enp.seq = target.enp,
              time.on.seq = time.on,
              time.off.seq = time.off,
              time.on.seq.W = time.on.W,
              time.off.seq.W = time.off.W,
              time.on.carry.seq = carry.time,
              time.on.balance.seq = balance.time,
              placebo.period = placebo.period.boot,
              placeboTest = placeboTest,
              carryoverTest = carryoverTest,
              carryover.period = carryover.period.boot,
              group.level = group.level,
              group = boot.group,
              time.on.seq.group = group.time.on,
              time.off.seq.group = group.time.off
            ),
            silent = TRUE
          )
        } else if (method == "mc") {
          boot <- try(
            fect_mc(
              Y = Y.input[, boot.id],
              X = X.boot,
              D = D[, boot.id],
              W = W.boot,
              W.in.fit = W.in.fit,
              I = I[, boot.id],
              II = II[, boot.id],
              T.on = T.on[, boot.id],
              T.off = T.off.boot,
              T.on.carry = T.on.carry[, boot.id],
              T.on.balance = T.on.balance[, boot.id],
              balance.period = balance.period,
              lambda.cv = out$lambda.cv,
              force = force,
              hasF = out$validF,
              hasRevs = hasRevs,
              tol = tol,
              max.iteration = max.iteration,
              boot = 1,
              norm.para = norm.para,
              calendar.enp.seq = target.enp,
              time.on.seq = time.on,
              time.off.seq = time.off,
              time.on.seq.W = time.on.W,
              time.off.seq.W = time.off.W,
              time.on.carry.seq = carry.time,
              time.on.balance.seq = balance.time,
              placebo.period = placebo.period.boot,
              placeboTest = placeboTest,
              carryoverTest = carryoverTest,
              carryover.period = carryover.period.boot,
              group.level = group.level,
              group = boot.group,
              time.on.seq.group = group.time.on,
              time.off.seq.group = group.time.off
            ),
            silent = TRUE
          )
        } else if (method == "cfe") {
          X.extra.FE.boot <- X.extra.FE[, boot.id, , drop = FALSE]
          X.Z.boot <- X.Z[, boot.id, , drop = FALSE]
          X.Q.boot <- X.Q[, boot.id, , drop = FALSE]
          X.gamma.boot <- X.gamma[, boot.id, , drop = FALSE]
          X.kappa.boot <- X.kappa[, boot.id, , drop = FALSE]
          boot <- try(
            fect_cfe(
              Y = Y.input[, boot.id],
              X = X.boot,
              D = D.boot,
              W = W.boot,
              W.in.fit = W.in.fit,
              X.extra.FE = X.extra.FE.boot,
              X.Z = X.Z.boot,
              X.Q = X.Q.boot,
              X.gamma = X.gamma.boot,
              X.kappa = X.kappa.boot,
              Zgamma.id = Zgamma.id,
              kappaQ.id = kappaQ.id,
              I = I.boot,
              II = II[, boot.id],
              T.on = T.on[, boot.id],
              T.off = T.off.boot,
              T.on.carry = T.on.carry[, boot.id],
              T.on.balance = T.on.balance[, boot.id],
              balance.period = balance.period,
              r.cv = r.boot,
              binary = binary,
              QR = QR,
              force = force,
              hasRevs = hasRevs,
              tol = tol,
              max.iteration = max.iteration,
              boot = 1,
              norm.para = norm.para,
              placebo.period = placebo.period.boot,
              placeboTest = placeboTest,
              carryover.period = carryover.period.boot,
              carryoverTest = carryoverTest,
              group.level = group.level,
              group = boot.group,
              time.on.seq.group = group.time.on,
              time.off.seq.group = group.time.off,
              time.on.seq = time.on,
              time.off.seq = time.off
            ),
            silent = TRUE
          )
        }

        if (is.null(boot)) {
          stop(paste0("Unsupported bootstrap method: ", method))
        }
        if ("try-error" %in% class(boot)) {
          boot0 <- list(
            att.avg = NA,
            att = NA,
            count = NA,
            beta = NA,
            att.off = NA,
            count.off = NA,
            eff.calendar = NA,
            eff.calendar.fit = NA,
            att.placebo = NA,
            att.avg.unit = NA,
            att.carryover = NA,
            group.att = NA,
            marginal = NA,
            carry.att = NA,
            group.output = list(),
            balance.att = NA,
            balance.att.placebo = NA,
            balance.count = NA,
            balance.avg.att = NA,
            balance.time = NA,
            att.avg.W = NA,
            att.on.W = NA,
            count.on.W = NA,
            time.on.W = NA,
            att.placebo.W = NA,
            att.off.W = NA,
            count.off.W = NA,
            time.off.W = NA,
            att.carryover.W = NA,
            eff = matrix(NA_real_, TT, length(boot.id)),
            D = matrix(NA_real_, TT, length(boot.id)),
            I = matrix(NA_real_, TT, length(boot.id)),
            boot.id = boot.id
          )
          return(boot0)
        } else {
          boot$boot.id <- boot.id
          return(boot)
        }
      }
    }
  }

  ## jack.seq <- sample(1:N, N, replace = FALSE)
  boot.seq <- NULL
  if (vartype == "jackknife") {
    ## nboots <- min(N, nboots)
    ## boot.seq <- jack.seq[1:nboots]
    boot.seq <- 1:N
  }

  ## computing
  one.nonpara <- trim_closure_env(one.nonpara)
  make_boot_na <- function() {
    list(
      att.avg = NA,
      att = rep(NA_real_, length(time.on)),
      count = rep(NA_real_, length(time.on)),
      beta = if (p > 0) rep(NA_real_, p) else NA,
      att.off = if (hasRevs == 1) rep(NA_real_, length(time.off)) else NA,
      count.off = if (hasRevs == 1) rep(NA_real_, length(time.off)) else NA,
      eff.calendar = rep(NA_real_, TT),
      eff.calendar.fit = rep(NA_real_, TT),
      att.placebo = NA,
      att.avg.unit = NA,
      att.carryover = NA,
      group.att = NA,
      marginal = NA,
      carry.att = if (!is.null(T.on.carry)) rep(NA_real_, length(carry.att)) else NA,
      group.output = list(),
      balance.att = if (!is.null(balance.period)) rep(NA_real_, length(balance.att)) else NA,
      balance.att.placebo = NA,
      balance.count = if (!is.null(balance.period)) rep(NA_real_, length(balance.att)) else NA,
      balance.avg.att = NA,
      balance.time = if (!is.null(balance.period)) balance.time else NA,
      att.avg.W = NA,
      att.on.W = if (!is.null(W)) rep(NA_real_, length(time.on.W)) else NA,
      count.on.W = if (!is.null(W)) rep(NA_real_, length(time.on.W)) else NA,
      time.on.W = if (!is.null(W)) time.on.W else NA,
      att.placebo.W = NA,
      att.off.W = if (!is.null(W) && hasRevs == 1) rep(NA_real_, length(time.off.W)) else NA,
      count.off.W = if (!is.null(W) && hasRevs == 1) rep(NA_real_, length(time.off.W)) else NA,
      time.off.W = if (!is.null(W) && hasRevs == 1) time.off.W else NA,
      att.carryover.W = NA,
      eff = if (keep.sims) matrix(NA_real_, TT, N) else NULL,
      D = if (keep.sims) matrix(NA_real_, TT, N) else NULL,
      I = if (keep.sims) matrix(NA_real_, TT, N) else NULL,
      boot.id = NULL
    )
  }
  if (do_parallel_boot) {
    old_rng_misuse <- getOption("doFuture.rng.onMisuse")
    options(doFuture.rng.onMisuse = "ignore")
    on.exit(options(doFuture.rng.onMisuse = old_rng_misuse), add = TRUE)

    ## v2.4.3: bump future.globals.maxSize locally to 2 GiB for the
    ## parallel block. Belt-and-braces guard for very large panels
    ## (high N*T, large factor rank, dense covariates) whose per-worker
    ## export can exceed the 500 MiB default even after the trim below.
    ## Honour any larger user-set cap via max().
    old_future_max <- getOption("future.globals.maxSize", 500 * 1024^2)
    options(future.globals.maxSize = max(old_future_max, 2 * 1024^3))
    on.exit(options(future.globals.maxSize = old_future_max), add = TRUE)

    quiet_nonpara <- function(j) {
      suppressMessages(suppressWarnings(one.nonpara(boot.seq[j])))
    }
    ## v2.4.3: trim the wrapper's closure env so foreach/future export
    ## ships only `one.nonpara` (already trimmed at L1600) and `boot.seq`,
    ## NOT the full fect_boot() frame (Y, D, X, W, out, ...). Prevents the
    ## quiet_nonpara=728MiB blowup reported on IFE + nboots=1000.
    quiet_nonpara <- trim_closure_env(quiet_nonpara)

    run_dopar_retry <- function(idx, workers) {
      ## Build the PSOCK cluster via the package helper, which bakes
      ## .libPaths() into worker startup (rscript_libs) — eliminating
      ## the post-construction `clusterCall(.libPaths)` race that
      ## occasionally killed a worker before iter 1 in mixed CV+boot
      ## test runs (E.6 in test-cv-parallel.R). Retry-with-backoff on
      ## socket-init failure; shrink workers between attempts.
      cl <- NULL
      init_err <- NULL
      try_workers <- max(1L, as.integer(workers))
      for (attempt in seq_len(3L)) {
        cl <- tryCatch(
          .fect_make_future_cluster(try_workers),
          error = function(e) {
            init_err <<- e
            NULL
          }
        )
        if (!is.null(cl)) break
        Sys.sleep(0.5 * attempt)
        try_workers <- max(1L, try_workers - 1L)
      }
      if (is.null(cl)) {
        stop("PSOCK cluster initialization failed: ",
             conditionMessage(init_err))
      }
      on.exit({
        try(parallel::stopCluster(cl), silent = TRUE)
      }, add = TRUE)
      doParallel::registerDoParallel(cl)
      suppressWarnings(foreach(
        j = idx,
        .inorder = TRUE,
        .errorhandling = "pass",
        .export = c(
          "fect_fe",
          "fect_mc",
          "fect_cfe",
          "get_term",
          "fect_nevertreated",
          "initialFit",
          ".reconstruct_gamma_fit_tr",
          ".reconstruct_kappa_fit",
          ".extract_and_apply_typeB_fe"
        ),
        .packages = c("fect", "mvtnorm", "fixest")
      ) %dopar%
        {
          return(quiet_nonpara(j))
        })
    }

    boot.out <- tryCatch(
      {
        suppressWarnings(foreach(
          j = 1:nboots,
          .inorder = FALSE,
          .errorhandling = "pass",
          .export = c(
            "fect_fe",
            "fect_mc",
            "fect_cfe",
            "get_term",
            "fect_nevertreated",
            "initialFit",
            ".reconstruct_gamma_fit_tr",
            ".reconstruct_kappa_fit",
            ".extract_and_apply_typeB_fe"
          ),
          .packages = c("fect", "mvtnorm", "fixest"),
          .options.future = list(seed = TRUE)
        ) %dopar%
          {
            return(quiet_nonpara(j))
          })
      },
      error = function(e) {
        if (!requireNamespace("doParallel", quietly = TRUE)) {
          stop(e)
        }
        warning(
          paste0(
            "Future backend failed (",
            conditionMessage(e),
            "). Retrying with doParallel backend."
          )
        )
        workers <- cores
        if (is.null(workers) || !is.numeric(workers) || length(workers) != 1 || is.na(workers)) {
          raw_cores <- parallelly::availableCores(omit = 2L)
          workers <- max(1L, min(raw_cores, 8L))
        }
        workers <- max(1L, as.integer(workers))
        ## If doParallel ALSO can't bring up a cluster (cluster-init
        ## storm after CV plan teardown, etc.), degrade to sequential
        ## rather than letting fect() die. Bootstrap completes with
        ## reduced throughput; downstream aggregation is unaffected.
        tryCatch(
          run_dopar_retry(1:nboots, workers),
          error = function(e2) {
            warning(
              paste0(
                "doParallel backend also failed (",
                conditionMessage(e2),
                "). Falling back to sequential bootstrap."
              )
            )
            lapply(seq_len(nboots), quiet_nonpara)
          }
        )
      }
    )

    failed.idx <- which(vapply(boot.out, inherits, logical(1), "error"))
    if (length(failed.idx) > 0 && requireNamespace("doParallel", quietly = TRUE)) {
      warning(
        paste0(
          "Retrying ",
          length(failed.idx),
          " failed bootstrap iterations with doParallel backend."
        )
      )
      retry.workers <- cores
      if (is.null(retry.workers) || !is.numeric(retry.workers) || length(retry.workers) != 1 || is.na(retry.workers)) {
        raw_cores <- parallelly::availableCores(omit = 2L)
        retry.workers <- max(1L, min(raw_cores, 8L))
      }
      retry.workers <- max(1L, min(as.integer(retry.workers), length(failed.idx)))

      # Retry with decreasing worker counts to avoid socket fragility in some IDE sessions.
      worker.plan <- unique(c(
        retry.workers,
        max(1L, retry.workers %/% 2L),
        2L,
        1L
      ))
      pending <- failed.idx
      for (w in worker.plan) {
        if (length(pending) == 0) {
          break
        }
        w.use <- max(1L, min(w, length(pending)))
        retry.out <- tryCatch(
          run_dopar_retry(pending, w.use),
          error = function(e) {
            warning(
              paste0(
                "doParallel retry failed with ",
                w.use,
                " worker(s): ",
                conditionMessage(e)
              )
            )
            vector("list", length(pending))
          }
        )
        for (k in seq_along(pending)) {
          if (!is.null(retry.out[[k]]) && !inherits(retry.out[[k]], "error")) {
            boot.out[[pending[k]]] <- retry.out[[k]]
          }
        }
        pending <- pending[vapply(boot.out[pending], inherits, logical(1), "error")]
      }

      # Last resort: fill any remaining failures directly to preserve finite SE.
      if (length(pending) > 0) {
        warning(
          paste0(
            "Final local retry for ",
            length(pending),
            " iteration(s) after parallel socket errors."
          )
        )
        for (j in pending) {
          boot.out[[j]] <- quiet_nonpara(j)
        }
      }
    }

    for (j in 1:nboots) {
      if (inherits(boot.out[[j]], "error")) {
        warning(
          paste0("Bootstrap iteration ", j, " failed in parallel worker: ", conditionMessage(boot.out[[j]]))
        )
        boot.out[[j]] <- make_boot_na()
      }
      att.avg.boot[, j] <- boot.out[[j]]$att.avg
      att.avg.unit.boot[, j] <- boot.out[[j]]$att.avg.unit
      att.boot[, j] <- boot.out[[j]]$att
      att.count.boot[, j] <- boot.out[[j]]$count
      if (keep.sims) {
        colnames(boot.out[[j]]$eff) <- boot.out[[j]]$boot.id
        eff.boot[,, j] <- boot.out[[j]]$eff
        D.boot[,, j] <- boot.out[[j]]$D
        I.boot[,, j] <- boot.out[[j]]$I
        if (is.null(boot.out[[j]]$boot.id)) {
          colnames.boot <- c(colnames.boot, list(1:N))
        } else {
          colnames.boot <- c(colnames.boot, list(boot.out[[j]]$boot.id))
        }
      }
      calendar.eff.boot[, j] <- boot.out[[j]]$eff.calendar
      calendar.eff.fit.boot[, j] <- boot.out[[j]]$eff.calendar.fit
      if (p > 0) {
        beta.boot[, j] <- boot.out[[j]]$beta
        if (binary == TRUE) {
          marginal.boot[, j] <- boot.out[[j]]$marginal
        }
      }
      if (hasRevs == 1) {
        att.off.boot[, j] <- boot.out[[j]]$att.off
        att.off.count.boot[, j] <- boot.out[[j]]$count.off
        if (keep.sims) {
          eff.off.boot <- c(eff.off.boot, list(boot.out[[j]]$eff.off))
        }
      }
      if (!is.null(T.on.carry)) {
        carry.att.boot[, j] <- boot.out[[j]]$carry.att
      }
      if (!is.null(balance.period)) {
        balance.att.boot[, j] <- boot.out[[j]]$balance.att
        balance.count.boot[, j] <- boot.out[[j]]$balance.count
        balance.avg.att.boot[, j] <- boot.out[[j]]$balance.avg.att
        if (!is.null(placebo.period) & placeboTest == TRUE) {
          balance.att.placebo.boot[, j] <- boot.out[[j]]$balance.att.placebo
        }
      }
      if (!is.null(W)) {
        att.avg.W.boot[, j] <- boot.out[[j]]$att.avg.W
        att.on.W.boot[, j] <- boot.out[[j]]$att.on.W
        att.on.count.W.boot[, j] <- boot.out[[j]]$count.on.W
        if (!is.null(placebo.period) & placeboTest == TRUE) {
          att.placebo.W.boot[, j] <- boot.out[[j]]$att.placebo.W
        }
        if (hasRevs == 1) {
          att.off.W.boot[, j] <- boot.out[[j]]$att.off.W
          att.off.count.W.boot[, j] <- boot.out[[j]]$count.off.W
          if (!is.null(carryover.period) & carryoverTest == TRUE) {
            att.carryover.W.boot[, j] <- boot.out[[j]]$att.carryover.W
          }
        }
      }

      if (!is.null(placebo.period) & placeboTest == TRUE) {
        att.placebo.boot[, j] <- boot.out[[j]]$att.placebo
      }
      if (!is.null(carryover.period) & carryoverTest == TRUE) {
        att.carryover.boot[, j] <- boot.out[[j]]$att.carryover
      }
      if (!is.null(group)) {
        group.att.boot[, j] <- boot.out[[j]]$group.att
        for (sub.name in group.output.name) {
          if (is.null(boot.out[[j]]$group.output[[sub.name]]$att.on)) {
            group.atts.boot[[sub.name]][, j] <- NA
          } else {
            group.atts.boot[[sub.name]][, j] <- boot.out[[j]]$group.output[[
              sub.name
            ]]$att.on
          }
          if (hasRevs == 1) {
            if (is.null(boot.out[[j]]$group.output[[sub.name]]$att.off)) {
              group.atts.off.boot[[sub.name]][, j] <- NA
            } else {
              group.atts.off.boot[[sub.name]][, j] <- boot.out[[
                j
              ]]$group.output[[sub.name]]$att.off
            }
          }
          if (placeboTest) {
            if (is.null(boot.out[[j]]$group.output[[sub.name]]$att.placebo)) {
              group.att.placebo.boot[[sub.name]][, j] <- NA
            } else {
              group.att.placebo.boot[[sub.name]][, j] <- boot.out[[
                j
              ]]$group.output[[sub.name]]$att.placebo
            }
          }
          if (carryoverTest) {
            if (is.null(boot.out[[j]]$group.output[[sub.name]]$att.carryover)) {
              group.att.carryover.boot[[sub.name]][, j] <- NA
            } else {
              group.att.carryover.boot[[sub.name]][, j] <- boot.out[[
                j
              ]]$group.output[[sub.name]]$att.carryover
            }
          }
        }
      }
    }
  } else {
    boot.out <- vector("list", nboots)
    # pb <- txtProgressBar(
    #     min = 0,
    #     max = nboots,
    #     style = 3,
    #     width = 50,
    #     char = "="
    # )
    for (j in 1:nboots) {
      boot <- one.nonpara(boot.seq[j])
      boot.out[[j]] <- boot
      att.avg.boot[, j] <- boot$att.avg
      att.avg.unit.boot[, j] <- boot$att.avg.unit
      att.boot[, j] <- boot$att
      att.count.boot[, j] <- boot$count
      if (keep.sims) {
        colnames(boot$eff) <- boot$boot.id
        # assign("boot", boot, .GlobalEnv)
        eff.boot[,, j] <- boot$eff
        D.boot[,, j] <- boot$D
        I.boot[,, j] <- boot$I
        if (is.null(boot$boot.id)) {
          colnames.boot <- c(colnames.boot, list(1:N)) # Parametric bootstrap
          # assign("boot", boot, .GlobalEnv)
        } else {
          colnames.boot <- c(colnames.boot, list(boot$boot.id)) # Raw bootstrap and jackknife
        }
      }
      calendar.eff.boot[, j] <- boot$eff.calendar
      calendar.eff.fit.boot[, j] <- boot$eff.calendar.fit
      if (p > 0) {
        beta.boot[, j] <- boot$beta
        if (binary == TRUE) {
          marginal.boot[, j] <- boot$marginal
        }
      }
      if (hasRevs == 1) {
        if (keep.sims) {
          eff.off.boot <- c(eff.off.boot, list(boot$eff.off))
        }
        att.off.boot[, j] <- boot$att.off
        att.off.count.boot[, j] <- boot$count.off
      }
      if (!is.null(T.on.carry)) {
        carry.att.boot[, j] <- boot$carry.att
      }
      if (!is.null(balance.period)) {
        balance.att.boot[, j] <- boot$balance.att
        balance.count.boot[, j] <- boot$balance.count
        balance.avg.att.boot[, j] <- boot$balance.avg.att
        if (!is.null(placebo.period) & placeboTest == TRUE) {
          balance.att.placebo.boot[, j] <- boot$balance.att.placebo
        }
      }
      if (!is.null(W)) {
        att.avg.W.boot[, j] <- boot$att.avg.W
        att.on.W.boot[, j] <- boot$att.on.W
        att.on.count.W.boot[, j] <- boot$count.on.W
        if (!is.null(placebo.period) & placeboTest == TRUE) {
          att.placebo.W.boot[, j] <- boot$att.placebo.W
        }
        if (hasRevs == 1) {
          att.off.W.boot[, j] <- boot$att.off.W
          att.off.count.W.boot[, j] <- boot$count.off.W
          if (!is.null(carryover.period) & carryoverTest == TRUE) {
            att.carryover.W.boot[, j] <- boot$att.carryover.W
          }
        }
      }
      if (!is.null(placebo.period) & placeboTest == TRUE) {
        att.placebo.boot[, j] <- boot$att.placebo
      }
      if (!is.null(carryover.period) & carryoverTest == TRUE) {
        att.carryover.boot[, j] <- boot$att.carryover
      }
      if (!is.null(group)) {
        group.att.boot[, j] <- boot$group.att
        for (sub.name in group.output.name) {
          if (is.null(boot$group.output[[sub.name]]$att.on)) {
            group.atts.boot[[sub.name]][, j] <- NA
          } else {
            group.atts.boot[[sub.name]][, j] <- boot$group.output[[
              sub.name
            ]]$att.on
          }
          if (hasRevs == 1) {
            if (is.null(boot$group.output[[sub.name]]$att.off)) {
              group.atts.off.boot[[sub.name]][, j] <- NA
            } else {
              group.atts.off.boot[[sub.name]][, j] <- boot$group.output[[
                sub.name
              ]]$att.off
            }
          }
          if (placeboTest) {
            if (is.null(boot$group.output[[sub.name]]$att.placebo)) {
              group.att.placebo.boot[[sub.name]][, j] <- NA
            } else {
              group.att.placebo.boot[[sub.name]][, j] <- boot$group.output[[
                sub.name
              ]]$att.placebo
            }
          }
          if (carryoverTest) {
            if (is.null(boot$group.output[[sub.name]]$att.carryover)) {
              group.att.carryover.boot[[sub.name]][, j] <- NA
            } else {
              group.att.carryover.boot[[sub.name]][, j] <- boot$group.output[[
                sub.name
              ]]$att.carryover
            }
          }
        }
      }
      # setTxtProgressBar(pb, j)
    }
    # close(pb)
  }
  ## end of bootstrapping
  ## remove failure bootstrap
  ## alternative condition? max(apply(is.na(att.boot),2,sum)) == dim(att.boot)[1]
  att.boot.original <- att.boot
  if (sum(is.na(c(att.avg.boot))) > 0) {
    boot.rm <- which(is.na(c(att.avg.boot)))
    att.avg.boot <- t(as.matrix(att.avg.boot[, -boot.rm]))
    att.avg.unit.boot <- t(as.matrix(att.avg.unit.boot[, -boot.rm]))
    att.boot <- as.matrix(att.boot[, -boot.rm])
    att.count.boot <- as.matrix(att.count.boot[, -boot.rm])
    calendar.eff.boot <- as.matrix(calendar.eff.boot[, -boot.rm])
    calendar.eff.fit.boot <- as.matrix(calendar.eff.fit.boot[, -boot.rm])
    if (p > 0) {
      beta.boot <- as.matrix(beta.boot[, -boot.rm])
      if (dim(beta.boot)[2] == 1) {
        beta.boot <- t(beta.boot)
      }
    }
    if (hasRevs == 1) {
      att.off.boot <- as.matrix(att.off.boot[, -boot.rm])
      att.off.count.boot <- as.matrix(att.off.count.boot[, -boot.rm])
    }
    if (!is.null(T.on.carry)) {
      carry.att.boot <- as.matrix(carry.att.boot[, -boot.rm])
    }
    if (!is.null(balance.period)) {
      balance.att.boot <- as.matrix(balance.att.boot[, -boot.rm])
      balance.count.boot <- as.matrix(balance.count.boot[, -boot.rm])
      balance.avg.att.boot <- t(as.matrix(balance.avg.att.boot[, -boot.rm]))
      if (!is.null(placebo.period) & placeboTest == TRUE) {
        balance.att.placebo.boot <- t(as.matrix(balance.att.placebo.boot[,
          -boot.rm
        ]))
      }
    }
    if (!is.null(W)) {
      att.avg.W.boot <- t(as.matrix(att.avg.W.boot[, -boot.rm]))
      att.on.W.boot <- as.matrix(att.on.W.boot[, -boot.rm])
      att.on.count.W.boot <- as.matrix(att.on.count.W.boot[, -boot.rm])
      if (!is.null(placebo.period) & placeboTest == TRUE) {
        att.placebo.W.boot <- t(as.matrix(att.placebo.W.boot[, -boot.rm]))
      }
      if (hasRevs == 1) {
        att.off.W.boot <- as.matrix(att.off.W.boot[, -boot.rm])
        att.off.count.W.boot <- as.matrix(att.off.count.W.boot[, -boot.rm])
        if (!is.null(carryover.period) & carryoverTest == TRUE) {
          att.carryover.W.boot <- t(as.matrix(att.carryover.W.boot[, -boot.rm]))
        }
      }
    }
    if (!is.null(placebo.period) & placeboTest == TRUE) {
      att.placebo.boot <- t(as.matrix(att.placebo.boot[, -boot.rm]))
    }
    if (!is.null(carryover.period) & carryoverTest == TRUE) {
      att.carryover.boot <- t(as.matrix(att.carryover.boot[, -boot.rm]))
    }
    if (!is.null(group)) {
      if (dim(group.att.boot)[1] == 1) {
        group.att.boot <- t(as.matrix(group.att.boot[, -boot.rm]))
      } else {
        group.att.boot <- as.matrix(group.att.boot[, -boot.rm])
      }

      for (sub.name in group.output.name) {
        group.atts.boot[[sub.name]] <- as.matrix(group.atts.boot[[sub.name]][,
          -boot.rm
        ])
        if (hasRevs == 1) {
          group.atts.off.boot[[sub.name]] <- as.matrix(group.atts.off.boot[[
            sub.name
          ]][, -boot.rm])
        }
        if (placeboTest) {
          group.att.placebo.boot[[
            sub.name
          ]] <- t(as.matrix(group.att.placebo.boot[[sub.name]][, -boot.rm]))
        }
        if (carryoverTest) {
          group.att.carryover.boot[[
            sub.name
          ]] <- t(as.matrix(group.att.carryover.boot[[sub.name]][, -boot.rm]))
        }
      }
    }
    if (keep.sims) {
      eff.boot <- eff.boot[, , -boot.rm, drop = FALSE]
      D.boot <- D.boot[, , -boot.rm, drop = FALSE]
      I.boot <- I.boot[, , -boot.rm, drop = FALSE]
      colnames.boot <- colnames.boot[-boot.rm]
    }
  }
  if (dis) {
    message(dim(att.boot)[2], " runs\n", sep = "")
  }
  if (vartype != "parametric") {
    ## Calculation of average outcomes
    # --- Input Checks ---
    if (!exists("boot.out")) {
      stop("List 'boot.out' required for non-parametric vartype.")
    }
    if (exists("boot.out") && !is.list(boot.out)) {
      stop("'boot.out' must be a list.")
    }
    if (!exists("T.on") || !exists("N")) {
      stop("'T.on' and 'N' must be defined.")
    }
    T_val <- tryCatch(
      if (is.matrix(T.on) || is.data.frame(T.on)) nrow(T.on) else length(T.on),
      error = function(e) 0
    )
    if (!is.numeric(T_val) || T_val < 0) {
      stop("'T.on' does not yield a valid number of time periods.")
    }
    if (!is.numeric(N) || N < 0) {
      stop("'N' must be a non-negative numeric value.")
    }
    if (!exists("D")) {
      stop("Reference treatment matrix 'D' must be defined.")
    }
    if (!is.matrix(D) && !is.data.frame(D)) {
      stop("'D' must be a matrix or data frame.")
    }
    if (!is.numeric(D)) {
      warning("Coercing reference matrix D to numeric.")
      mode(D) <- "numeric"
    }
    if (nrow(D) != T_val || (N > 0 && ncol(D) != N)) {
      stop("Dimensions of 'D' do not match T.on and N.")
    }
    # vartype check is general, but this whole block is for vartype != "parametric"
    if (!is.character(vartype) || length(vartype) != 1) {
      stop("'vartype' must be a single string.")
    }
    valid_vartypes <- c("bootstrap", "jackknife") # Only non-parametric types are valid here
    if (!(vartype %in% valid_vartypes)) {
      stop(paste(
        "'vartype' must be one of:",
        paste(valid_vartypes, collapse = ", ")
      ))
    }

    # --- Preliminaries ---
    D0 <- D # Use the provided reference matrix D as D0
    n_replicates <- if (exists("boot.out")) length(boot.out) else 0 # Simplified as boot.out must exist here

    if (n_replicates == 0 && N > 0) {
      stop("'boot.out' is empty for non-parametric vartype.")
    } # Note: Original code had warning, changed to stop to match other similar critical checks. If this should be a warning, it can be reverted. The prompt implies only specific changes. Let's assume this was an oversight and keep it as stop for consistency or revert if original intent was warning. Given "Do not make any other changes", I will revert this to warning if it was originally a warning. The provided code has: `if (n_replicates == 0 && N > 0) { stop("'boot.out' is empty for non-parametric vartype.") }` - so this is correct.

    if (vartype == "jackknife" && N <= 1) {
      warning("Jackknife requires N > 1 units. Returning empty Y.avg.")
      if (!exists("out")) {
        out <- list()
      }
      out$Y.avg <- data.frame(
        period = numeric(0),
        treated = numeric(0),
        counterfactual = numeric(0),
        lower.tr = numeric(0),
        upper.tr = numeric(0),
        lower90.tr = numeric(0),
        upper90.tr = numeric(0),
        lower.ct = numeric(0),
        upper.ct = numeric(0),
        lower90.ct = numeric(0),
        upper90.ct = numeric(0)
      )
      return(invisible(NULL))
    }
    if (T_val == 0 || N == 0) {
      warning(paste(
        "Input data has zero ",
        if (T_val == 0) "time periods" else "units",
        ". Returning empty Y.avg.",
        sep = ""
      ))
      if (!exists("out")) {
        out <- list()
      }
      out$Y.avg <- data.frame(
        period = numeric(0),
        treated = numeric(0),
        counterfactual = numeric(0),
        lower.tr = numeric(0),
        upper.tr = numeric(0),
        lower90.tr = numeric(0),
        upper90.tr = numeric(0),
        lower.ct = numeric(0),
        upper.ct = numeric(0),
        lower90.ct = numeric(0),
        upper90.ct = numeric(0)
      )
      return(invisible(NULL)) # Directly return for non-parametric case
    }

    # --- Helper Functions ---
    calculate_first_treatment <- function(D_mat, N_val) {
      if (N_val == 0) {
        return(integer(0))
      }
      if (!is.matrix(D_mat)) {
        D_mat <- as.matrix(D_mat)
      }
      if (!is.numeric(D_mat)) {
        mode(D_mat) <- "numeric"
      }
      # Vectorized: first index per column where D >= 1 (ignoring NAs),
      # identical to apply(D_mat, 2, find_first, min_val = 1)
      Tn <- nrow(D_mat)
      if (is.null(Tn) || Tn == 0L) {
        return(rep(NA_integer_, ncol(D_mat)))
      }
      cond <- !is.na(D_mat) & D_mat >= 1
      has_true <- colSums(cond) > 0L
      if (!any(has_true)) {
        return(rep(NA_integer_, ncol(D_mat)))
      }
      score_mat <- matrix(Tn + 1L - seq_len(Tn), nrow = Tn, ncol = ncol(D_mat))
      score_mat <- score_mat * as.integer(cond)
      first_idx <- max.col(t(score_mat), ties.method = "first")
      out_idx <- rep(NA_integer_, ncol(D_mat))
      out_idx[has_true] <- first_idx[has_true]
      out_idx
    }

    # --- Determine Treatment Timing and Type ---
    first_treat_period_D0 <- if (N > 0) {
      calculate_first_treatment(D0, N)
    } else {
      integer(0)
    }
    treated_units_idx_D0 <- which(!is.na(first_treat_period_D0))
    control_units_idx_D0 <- which(is.na(first_treat_period_D0))
    Ntr <- length(treated_units_idx_D0)
    Nco <- length(control_units_idx_D0)

    if (Ntr == 0) {
      warning("No treated units found based on reference D matrix (D0).")
      is_staggered <- FALSE
      common_G_D0 <- NA
    } else {
      unique_adoption_times_D0 <- unique(first_treat_period_D0[
        treated_units_idx_D0
      ])
      is_staggered <- length(unique_adoption_times_D0) > 1
      if (!is_staggered) {
        common_G_D0 <- unique_adoption_times_D0
      } else {
        common_G_D0 <- NA
      }
    }

    # --- Period Labels ---
    calendar_periods_labels <- seq.int(T_val) # Default labels (1:T)
    if (!is.null(rownames(D0))) {
      cal_labels_from_rows <- tryCatch(
        as.numeric(rownames(D0)),
        warning = function(w) NULL
      )
      if (
        !is.null(cal_labels_from_rows) &&
          !anyNA(cal_labels_from_rows) &&
          length(cal_labels_from_rows) == T_val
      ) {
        calendar_periods_labels <- cal_labels_from_rows
      } else {
        warning("Rownames of D not valid numeric labels. Using 1:T.")
      }
    }
    periods <- NULL # Will be set later based on staggered/non-staggered

    # --- Initialize Placeholders ---
    event_times_range <- NULL
    period_type_attr <- "calendar_time"
    repl_tr <- NULL
    repl_cf <- NULL
    theta_tr <- NULL
    theta_cf <- NULL

    # --- Calculate Averages and Determine Output Periods ---

    if (is_staggered) {
      period_type_attr <- "event_time"
      min_event_time <- -(T_val - 1) + 1
      max_event_time <- T_val
      event_times_range <- min_event_time:max_event_time
      n_event_times <- length(event_times_range)
      event_time_matrix_D0 <- outer(
        seq.int(T_val),
        first_treat_period_D0,
        FUN = "-"
      ) +
        1

      # Calculate Replicates if Bootstrap/Jackknife
      event_repl_cf <- matrix(
        NA_real_,
        nrow = n_event_times,
        ncol = n_replicates
      )
      event_repl_tr <- matrix(
        NA_real_,
        nrow = n_event_times,
        ncol = n_replicates
      )
      rownames(event_repl_cf) <- rownames(event_repl_tr) <- event_times_range
      if (n_replicates > 0) {
        for (b in seq_len(n_replicates)) {
          repl_data <- boot.out[[b]]
          if (
            is.null(repl_data) ||
              !is.list(repl_data) ||
              is.null(repl_data$D) ||
              is.null(repl_data$Y) ||
              is.null(repl_data$Y.ct)
          ) {
            next
          }
          D_b <- repl_data$D
          Y_b <- repl_data$Y
          Yct_b <- repl_data$Y.ct
          if (!is.matrix(D_b)) {
            D_b <- as.matrix(D_b)
          }
          if (!is.matrix(Y_b)) {
            Y_b <- as.matrix(Y_b)
          }
          if (!is.matrix(Yct_b)) {
            Yct_b <- as.matrix(Yct_b)
          }
          if (!is.numeric(D_b)) {
            mode(D_b) <- "numeric"
          }
          if (!is.numeric(Y_b)) {
            warning(paste("Coercing replicate", b, "Y to numeric."))
            mode(Y_b) <- "numeric"
          }
          if (!is.numeric(Yct_b)) {
            warning(paste("Coercing replicate", b, "Y.ct to numeric."))
            mode(Yct_b) <- "numeric"
          }

          N_b <- ncol(D_b)
          expected_N_b <- N
          if (vartype == "jackknife") {
            expected_N_b <- N - 1
          }
          if (N > 0 && N_b != expected_N_b) {
            warning(
              paste(
                "Replicate",
                b,
                ": For vartype '",
                vartype,
                "', expected ",
                expected_N_b,
                " columns in D_b, but got ",
                N_b,
                ". Skipping replicate."
              ),
              call. = FALSE
            )
            next
          }
          if (
            nrow(D_b) != T_val ||
              nrow(Y_b) != T_val ||
              nrow(Yct_b) != T_val ||
              (N_b > 0 && (ncol(Y_b) != N_b || ncol(Yct_b) != N_b))
          ) {
            warning(
              paste(
                "Replicate",
                b,
                ": Dimension mismatch with T_val. Skipping replicate."
              ),
              call. = FALSE
            )
            next
          }
          if (N_b == 0) {
            next
          } # If a replicate has 0 units, skip it (relevant if N > 0 originally)

          Y_b[is.na(Yct_b)] <- NA

          first_treat_period_b <- calculate_first_treatment(D_b, N_b)
          event_time_matrix_b <- outer(
            seq.int(T_val),
            first_treat_period_b,
            FUN = "-"
          ) +
            1

          # Vectorized aggregation across all event times while preserving
          # identical selection semantics and element order
          evt_vec <- as.vector(event_time_matrix_b)
          valid_evt <- !is.na(evt_vec)
          D_vec <- as.vector(D_b)
          Y_vec <- as.vector(Y_b)
          Yct_vec <- as.vector(Yct_b)

          lvl <- event_times_range
          lvl_chr <- as.character(lvl)

          # Indices for all observations at each event time (e can be <= 0 or > 0)
          idx_all_list <- split(
            which(valid_evt),
            factor(evt_vec[valid_evt], levels = lvl)
          )
          # Indices for treated-period observations when e > 0 (D == 1)
          idx_d1_list <- split(
            which(valid_evt & D_vec == 1),
            factor(evt_vec[valid_evt & D_vec == 1], levels = lvl)
          )

          tr_col <- rep(NA_real_, length(lvl))
          cf_col <- rep(NA_real_, length(lvl))
          for (e_i in seq_along(lvl)) {
            e_val <- lvl[e_i]
            if (e_val <= 0) {
              idx_use <- idx_all_list[[lvl_chr[e_i]]]
              if (!is.null(idx_use) && length(idx_use) > 0L) {
                tr_col[e_i] <- mean(Y_vec[idx_use], na.rm = TRUE)
                cf_col[e_i] <- mean(Yct_vec[idx_use], na.rm = TRUE)
              }
            } else {
              idx_use <- idx_d1_list[[lvl_chr[e_i]]]
              if (!is.null(idx_use) && length(idx_use) > 0L) {
                tr_col[e_i] <- mean(Y_vec[idx_use], na.rm = TRUE)
                cf_col[e_i] <- mean(Yct_vec[idx_use], na.rm = TRUE)
              }
            }
          }

          event_repl_tr[, b] <- tr_col
          event_repl_cf[, b] <- cf_col
        }
      }

      # MODIFICATION START: Keep all event times; rows with all NA replicates will result in NA averages.
      periods <- event_times_range
      repl_tr <- event_repl_tr
      repl_cf <- event_repl_cf

      # rowMeans with na.rm=TRUE produces NaN if all replicates for a period are NA.
      # These NaNs will propagate to NAs in the final data.frame.
      theta_tr <- rowMeans(repl_tr, na.rm = TRUE)
      theta_cf <- rowMeans(repl_cf, na.rm = TRUE)
      # MODIFICATION END
    } else {
      # Non-Staggered Case
      period_type_attr <- "calendar_time"
      # periods <- calendar_periods_labels # This was assigned earlier, will be used below.
      # num_periods_out <- T_val # This was used later, now length(periods) is the source of truth

      calendar_repl_cf <- matrix(NA_real_, nrow = T_val, ncol = n_replicates)
      calendar_repl_tr <- matrix(NA_real_, nrow = T_val, ncol = n_replicates)
      rownames(calendar_repl_cf) <- rownames(calendar_repl_tr) <- 1:T_val # Use 1:T_val for internal consistency
      if (n_replicates > 0) {
        for (b in seq_len(n_replicates)) {
          repl_data <- boot.out[[b]]
          if (
            is.null(repl_data) ||
              !is.list(repl_data) ||
              is.null(repl_data$D) ||
              is.null(repl_data$Y) ||
              is.null(repl_data$Y.ct)
          ) {
            next
          }
          D_b <- repl_data$D
          Y_b <- repl_data$Y
          Yct_b <- repl_data$Y.ct
          if (!is.matrix(D_b)) {
            D_b <- as.matrix(D_b)
          }
          if (!is.matrix(Y_b)) {
            Y_b <- as.matrix(Y_b)
          }
          if (!is.matrix(Yct_b)) {
            Yct_b <- as.matrix(Yct_b)
          }
          if (!is.numeric(D_b)) {
            mode(D_b) <- "numeric"
          }
          if (!is.numeric(Y_b)) {
            warning(paste("Coercing replicate", b, "Y to numeric."))
            mode(Y_b) <- "numeric"
          }
          if (!is.numeric(Yct_b)) {
            warning(paste("Coercing replicate", b, "Y.ct to numeric."))
            mode(Yct_b) <- "numeric"
          }

          N_b <- ncol(D_b)
          expected_N_b <- N

          present_original_indices <- if (N > 0) 1:N else integer(0)
          replicate_col_to_original_idx <- if (N_b > 0) 1:N_b else integer(0) # Default mapping for bootstrap

          if (vartype == "jackknife") {
            expected_N_b <- N - 1
            if (N_b == N - 1 && n_replicates == N && N > 0) {
              # N > 0 for jackknife to make sense
              omitted_original_idx <- b
              present_original_indices <- (1:N)[-omitted_original_idx]
              # Map columns in this jackknife replicate D_b back to their original unit indices
              # Assuming D_b's columns are the N-1 present units, in their original relative order.
              replicate_col_to_original_idx <- present_original_indices
            } else if (N > 0) {
              # Only warn if N > 0, jackknife on N=0 is already handled.
              warning(
                paste(
                  "Replicate",
                  b,
                  ": Jackknife replicate has",
                  N_b,
                  "columns (expected N-1=",
                  N - 1,
                  ") or n_replicates (",
                  n_replicates,
                  ") != N (",
                  N,
                  "). Skipping replicate."
                ),
                call. = FALSE
              )
              next
            }
            # if N=0, expected_N_b = -1. N_b is likely 0. This path is complex if N=0.
            # However, N=0 is caught by an early return. So N > 0 here.
          }

          if (N > 0 && N_b != expected_N_b) {
            # Check for N > 0 before N_b != expected_N_b
            warning(
              paste(
                "Replicate",
                b,
                ": For vartype '",
                vartype,
                "', expected ",
                expected_N_b,
                " columns in D_b, but got ",
                N_b,
                ". Skipping replicate."
              ),
              call. = FALSE
            )
            next
          }

          if (
            nrow(D_b) != T_val ||
              nrow(Y_b) != T_val ||
              nrow(Yct_b) != T_val ||
              (N_b > 0 && (ncol(Y_b) != N_b || ncol(Yct_b) != N_b))
          ) {
            warning(
              paste(
                "Replicate",
                b,
                ": Dimension mismatch with T_val. Skipping replicate."
              ),
              call. = FALSE
            )
            next
          }

          if (N_b == 0 && N > 0) {
            next
          } # If N > 0, but this replicate has 0 units, skip. If N=0, N_b=0 is expected.

          Y_b[is.na(Yct_b)] <- NA

          # Get the column indices *in the current replicate D_b* that correspond to these target units
          replicate_cols_for_target_units <- which(
            colSums(D_b == 1, na.rm = TRUE) > 0
          )
          if (length(replicate_cols_for_target_units) == 0 && Ntr > 0) {
            next
          }

          for (t in seq.int(T_val)) {
            cols_to_avg <- integer(0)
            if (Ntr == 0) {
              # No treated units in D0, so effectively averaging "control" outcomes for "placebo" treated group
            } else if (!is.na(common_G_D0) && t < common_G_D0) {
              # Pre-treatment period for non-staggered
              cols_to_avg <- replicate_cols_for_target_units
            } else if (!is.na(common_G_D0) && t >= common_G_D0) {
              # Post-treatment period for non-staggered
              # Only average over units that are actually treated (D_b[t,col]==1) in this replicate at this time
              cols_to_avg <- D_b[t, ] == 1
            } else if (is.na(common_G_D0) && Ntr > 0) {
              # Should not happen if is_staggered is FALSE and Ntr > 0
              warning(
                paste(
                  "Replicate",
                  b,
                  ": In non-staggered case with Ntr > 0, but common_G_D0 is NA. Skipping time",
                  t,
                  "for this replicate."
                ),
                call. = FALSE
              )
              next # Skip this time period for this replicate
            }
            # If Ntr > 0 but replicate_cols_for_target_units is empty, cols_to_avg will be empty.

            if (length(cols_to_avg) > 0) {
              calendar_repl_tr[t, b] <- mean(Y_b[t, cols_to_avg], na.rm = TRUE)
              calendar_repl_cf[t, b] <- mean(
                Yct_b[t, cols_to_avg],
                na.rm = TRUE
              )
            }
            # If cols_to_avg is empty, the NA_real_ initialized in calendar_repl_tr/cf remains.
          }
        }
      }

      # MODIFICATION START: Keep all calendar times; rows with all NA replicates will result in NA averages.
      periods <- calendar_periods_labels # Use the potentially custom labels
      repl_tr <- calendar_repl_tr
      repl_cf <- calendar_repl_cf

      # rowMeans with na.rm=TRUE produces NaN if all replicates for a period are NA.
      # These NaNs will propagate to NAs in the final data.frame.
      theta_tr <- rowMeans(repl_tr, na.rm = TRUE)
      theta_cf <- rowMeans(repl_cf, na.rm = TRUE)
      # MODIFICATION END
    } # End Staggered/Non-Staggered block

    # --- Confidence Interval Calculation (Non-Parametric) ---
    num_periods_out <- length(periods)

    ci_tr95 <- matrix(
      NA_real_,
      nrow = num_periods_out,
      ncol = 2,
      dimnames = list(NULL, c("lower", "upper"))
    )
    ci_cf95 <- matrix(
      NA_real_,
      nrow = num_periods_out,
      ncol = 2,
      dimnames = list(NULL, c("lower", "upper"))
    )
    ci_tr90 <- matrix(
      NA_real_,
      nrow = num_periods_out,
      ncol = 2,
      dimnames = list(NULL, c("lower", "upper"))
    )
    ci_cf90 <- matrix(
      NA_real_,
      nrow = num_periods_out,
      ncol = 2,
      dimnames = list(NULL, c("lower", "upper"))
    )

    if (vartype == "bootstrap") {
      if (n_replicates > 0 && num_periods_out > 0) {
        if (
          nrow(repl_tr) != num_periods_out || nrow(repl_cf) != num_periods_out
        ) {
          # This warning should ideally not be triggered if logic is correct upstream.
          warning(
            "Bootstrap replicate matrix dimensions mismatch with number of output periods. Skipping CI."
          )
        } else {
          ci_tr95 <- basic_ci_alpha(theta_tr, repl_tr, alpha = 0.05)
          ci_cf95 <- basic_ci_alpha(theta_cf, repl_cf, alpha = 0.05)
          ci_tr90 <- basic_ci_alpha(theta_tr, repl_tr, alpha = 0.10)
          ci_cf90 <- basic_ci_alpha(theta_cf, repl_cf, alpha = 0.10)
        }
      } else {
        if (num_periods_out > 0) {
          warning(
            "Cannot calculate bootstrap CIs: No replicates remain or available."
          )
        }
      } # num_periods_out > 0 condition added to warning
    } else if (vartype == "jackknife") {
      if (n_replicates > 1 && num_periods_out > 0) {
        # Jackknife needs n_replicates > 1 (i.e., N > 1)
        if (
          nrow(repl_tr) != num_periods_out || nrow(repl_cf) != num_periods_out
        ) {
          warning(
            "Jackknife replicate matrix dimensions mismatch with number of output periods. Skipping CI."
          )
        } else {
          jackknife_ci_alpha <- function(theta, replicates, alpha) {
            n_p <- length(theta) # number of periods
            n_repl <- ncol(replicates) # number of jackknife replicates (should be N)

            # Handle cases where theta might be NA (e.g. all replicates for that period were NA)
            # or where all replicates for a period are NA
            # dev_sq will propagate NAs if theta is NA or replicates are NA
            dev_sq <- (replicates - theta)^2 # (num_periods_out x n_repl) matrix

            # sum_dev_sq will be NA if theta was NA.
            # If theta is a number, but all replicates[p,] are NA, sum_dev_sq[p] will be 0 (due to na.rm=TRUE).
            sum_dev_sq <- rowSums(dev_sq, na.rm = TRUE)
            valid_counts <- rowSums(!is.na(replicates)) # Count non-NA replicates per period

            jk_variance <- ((n_repl - 1) / n_repl) * sum_dev_sq
            # For periods where all replicates were NA (so theta is NA), jk_variance is NA.
            # For periods where theta is a number but all replicates were NA (sum_dev_sq=0), jk_variance is 0.
            # For periods where valid_counts <=1, variance is not well-defined or unstable.
            jk_variance[valid_counts <= 1] <- NA_real_

            jk_se <- sqrt(jk_variance)
            jk_se[!is.finite(jk_se)] <- NA_real_ # Catches NaN from sqrt(negative) or Inf

            z_crit <- qnorm(1 - alpha / 2)

            # ok condition: theta must be non-NA, se must be non-NA, and se must be positive.
            # If se is 0 (e.g. from sum_dev_sq=0), CI would be [theta, theta], which is not informative.
            # Setting to NA is more appropriate for such cases.
            ok <- !is.na(theta) & !is.na(jk_se) & jk_se > 0

            lower <- rep(NA_real_, n_p)
            upper <- rep(NA_real_, n_p)
            lower[ok] <- theta[ok] - z_crit * jk_se[ok]
            upper[ok] <- theta[ok] + z_crit * jk_se[ok]

            ci <- cbind(lower, upper)
            colnames(ci) <- c("lower", "upper")
            ci
          }
          ci_tr95 <- jackknife_ci_alpha(theta_tr, repl_tr, alpha = 0.05)
          ci_cf95 <- jackknife_ci_alpha(theta_cf, repl_cf, alpha = 0.05)
          ci_tr90 <- jackknife_ci_alpha(theta_tr, repl_tr, alpha = 0.10)
          ci_cf90 <- jackknife_ci_alpha(theta_cf, repl_cf, alpha = 0.10)
        }
      } else {
        if (num_periods_out > 0) {
          warning(
            "Cannot calculate jackknife CIs: Need >1 replicate (N > 1) or no periods remain."
          )
        }
      } # num_periods_out > 0 condition added
    } else {
      stop(
        "Internal error: Invalid 'vartype' for non-parametric CI calculation."
      )
    }

    # --- Assemble into out$Y.avg ---
    if (!exists("out")) {
      out <- list()
    }

    # Check consistency of component lengths before creating data.frame
    # This was already robust, but good to be aware with the changes.
    # num_periods_out is now full length of event_times_range or calendar_periods_labels.
    # theta_tr/cf will have this length, possibly with NaNs.
    # CIs are initialized to this size with NAs.

    # Ensure CI matrices are correctly dimensioned and numeric, even if all NA
    ensure_numeric_matrix_cols <- function(mat, expected_rows) {
      if (
        is.null(mat) ||
          !is.matrix(mat) ||
          ncol(mat) != 2 ||
          nrow(mat) != expected_rows
      ) {
        # If mat is not as expected (e.g. NULL from a failed CI calculation, or wrong dims), return a compliant NA matrix
        return(matrix(
          NA_real_,
          nrow = expected_rows,
          ncol = 2,
          dimnames = list(NULL, c("lower", "upper"))
        ))
      }
      if (!is.numeric(mat)) {
        # Coerce if not numeric (e.g. character matrix)
        mat_num <- matrix(
          as.numeric(mat),
          nrow = nrow(mat),
          ncol = ncol(mat),
          dimnames = dimnames(mat)
        )
        mat <- mat_num
      }
      mat[is.nan(mat)] <- NA_real_ # Standardize NaN to NA
      mat
    }

    ci_tr95 <- ensure_numeric_matrix_cols(ci_tr95, num_periods_out)
    ci_cf95 <- ensure_numeric_matrix_cols(ci_cf95, num_periods_out)
    ci_tr90 <- ensure_numeric_matrix_cols(ci_tr90, num_periods_out)
    ci_cf90 <- ensure_numeric_matrix_cols(ci_cf90, num_periods_out)

    if (num_periods_out > 0) {
      # Convert theta_tr/cf (which might contain NaN from rowMeans) to numeric vectors.
      # data.frame conversion will turn NaNs into NAs.
      theta_tr_vec <- if (length(theta_tr) == num_periods_out) {
        as.numeric(theta_tr)
      } else {
        rep(NA_real_, num_periods_out)
      }
      theta_cf_vec <- if (length(theta_cf) == num_periods_out) {
        as.numeric(theta_cf)
      } else {
        rep(NA_real_, num_periods_out)
      }

      out$Y.avg <- data.frame(
        period = periods, # This now contains all original periods
        treated = theta_tr_vec, # Will be NA if all replicates were NA for this period
        counterfactual = theta_cf_vec, # Will be NA if all replicates were NA for this period
        lower.tr = ci_tr95[, "lower"],
        upper.tr = ci_tr95[, "upper"],
        lower90.tr = ci_tr90[, "lower"],
        upper90.tr = ci_tr90[, "upper"],
        lower.ct = ci_cf95[, "lower"],
        upper.ct = ci_cf95[, "upper"],
        lower90.ct = ci_cf90[, "lower"],
        upper90.ct = ci_cf90[, "upper"]
      )
    } else {
      # This case handles T_val=0 or N=0 (via early return) or if periods somehow becomes length 0
      out$Y.avg <- data.frame(
        period = numeric(0),
        treated = numeric(0),
        counterfactual = numeric(0),
        lower.tr = numeric(0),
        upper.tr = numeric(0),
        lower90.tr = numeric(0),
        upper90.tr = numeric(0),
        lower.ct = numeric(0),
        upper.ct = numeric(0),
        lower90.ct = numeric(0),
        upper90.ct = numeric(0)
      )
    }
    attr(out$Y.avg, "period_type") <- period_type_attr
  } # when vartype == "parametric" the processing is done in plot.R
  ####################################
  ## Variance and CIs
  ####################################

  ## function to get two-sided p-values
  get.pvalue <- function(vec) {
    if (NaN %in% vec | NA %in% vec) {
      nan.pos <- is.nan(vec)
      na.pos <- is.na(vec)
      pos <- c(which(nan.pos), which(na.pos))
      vec.a <- vec[-pos]
      a <- sum(vec.a >= 0) / (length(vec) - sum(nan.pos | na.pos)) * 2
      b <- sum(vec.a <= 0) / (length(vec) - sum(nan.pos | na.pos)) * 2
    } else {
      a <- sum(vec >= 0) / length(vec) * 2
      b <- sum(vec <= 0) / length(vec) * 2
    }
    return(min(as.numeric(min(a, b)), 1))
  }

  ## ATT estimates
  if (vartype == "jackknife") {
    att.j <- jackknifed(att, att.boot, alpha, quantile.CI = quantile.CI)
    att.bound <- cbind(
      att + qnorm(alpha) * att.j$se,
      att + qnorm(1 - alpha) * att.j$se
    )
    colnames(att.bound) <- c("CI.lower", "CI.upper")
    rownames(att.bound) <- out$time

    est.att <- cbind(att, att.j$se, att.j$CI.l, att.j$CI.u, att.j$P, out$count)
    est.att90 <- cbind(att, att.j$se, att.bound, att.j$P, out$count)
    colnames(est.att) <- c(
      "ATT",
      "S.E.",
      "CI.lower",
      "CI.upper",
      "p.value",
      "count"
    )
    colnames(est.att90) <- c(
      "ATT",
      "S.E.",
      "CI.lower",
      "CI.upper",
      "p.value",
      "count"
    )
    rownames(est.att) <- out$time
    rownames(est.att90) <- out$time
    vcov.att <- att.j$vcov

    eff.calendar.j <- jackknifed(
      calendar.eff,
      calendar.eff.boot,
      alpha,
      quantile.CI = quantile.CI
    )
    est.eff.calendar <- cbind(
      calendar.eff,
      eff.calendar.j$se,
      eff.calendar.j$CI.l,
      eff.calendar.j$CI.u,
      eff.calendar.j$P,
      calendar.N
    )
    colnames(est.eff.calendar) <- c(
      "ATT-calendar",
      "S.E.",
      "CI.lower",
      "CI.upper",
      "p.value",
      "count"
    )

    eff.calendar.fit.j <- jackknifed(
      calendar.eff.fit,
      calendar.eff.fit.boot,
      alpha,
      quantile.CI = quantile.CI
    )
    est.eff.calendar.fit <- cbind(
      calendar.eff.fit,
      eff.calendar.fit.j$se,
      eff.calendar.fit.j$CI.l,
      eff.calendar.fit.j$CI.u,
      eff.calendar.fit.j$P,
      calendar.N
    )
    colnames(est.eff.calendar.fit) <- c(
      "ATT-calendar Fitted",
      "S.E.",
      "CI.lower",
      "CI.upper",
      "p.value",
      "count"
    )

    if (hasRevs == 1) {
      att.off.j <- jackknifed(
        att.off,
        att.off.boot,
        alpha,
        quantile.CI = quantile.CI
      )
      est.att.off <- cbind(
        att.off,
        att.off.j$se,
        att.off.j$CI.l,
        att.off.j$CI.u,
        att.off.j$P,
        out$count.off
      )
      colnames(est.att.off) <- c(
        "ATT.OFF",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value",
        "count"
      )
      rownames(est.att.off) <- out$time.off
      vcov.att.off <- att.off.j$vcov

      att.off.bound <- cbind(
        att.off + qnorm(alpha) * att.off.j$se,
        att.off + qnorm(1 - alpha) * att.off.j$se
      )
      colnames(att.off.bound) <- c("CI.lower", "CI.upper")
      rownames(att.off.bound) <- out$time.off
    }

    if (!is.null(T.on.carry)) {
      carry.att.j <- jackknifed(
        carry.att,
        carry.att.boot,
        alpha,
        quantile.CI = quantile.CI
      )
      est.carry.att <- cbind(
        carry.att,
        carry.att.j$se,
        carry.att.j$CI.l,
        carry.att.j$CI.u,
        carry.att.j$P
      )

      colnames(est.carry.att) <- c(
        "ATT",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value"
      )
      rownames(est.carry.att) <- carry.time
    }

    if (!is.null(balance.period)) {
      balance.att.j <- jackknifed(
        balance.att,
        balance.att.boot,
        alpha,
        quantile.CI = quantile.CI
      )
      est.balance.att <- cbind(
        balance.att,
        balance.att.j$se,
        balance.att.j$CI.l,
        balance.att.j$CI.u,
        balance.att.j$P,
        out$balance.count
      )
      colnames(est.balance.att) <- c(
        "ATT",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value",
        "count"
      )
      rownames(est.balance.att) <- out$balance.time
      vcov.balance.att <- balance.att.j$vcov

      balance.avg.att.j <- jackknifed(
        balance.avg.att,
        balance.avg.att.boot,
        alpha,
        quantile.CI = quantile.CI
      )
      est.balance.avg <- t(as.matrix(c(
        balance.avg.att,
        balance.avg.att.j$se,
        balance.avg.att.j$CI.l,
        balance.avg.att.j$CI.u,
        balance.avg.att.j$P
      )))
      colnames(est.balance.avg) <- c(
        "ATT.avg",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value"
      )

      balance.att.bound <- cbind(
        balance.att + qnorm(alpha) * balance.att.j$se,
        balance.att + qnorm(1 - alpha) * balance.att.j$se
      )
      colnames(balance.att.bound) <- c("CI.lower", "CI.upper")
      rownames(balance.att.bound) <- out$balance.time

      if (!is.null(placebo.period) & placeboTest == TRUE) {
        balance.att.placebo.j <- jackknifed(
          balance.att.placebo,
          balance.att.placebo.boot,
          alpha,
          quantile.CI = quantile.CI
        )
        est.balance.placebo <- t(as.matrix(c(
          balance.att.placebo,
          balance.att.placebo.j$se,
          balance.att.placebo.j$CI.l,
          balance.att.placebo.j$CI.u,
          balance.att.placebo.j$P
        )))
        colnames(est.balance.placebo) <- c(
          "ATT.placebo",
          "S.E.",
          "CI.lower",
          "CI.upper",
          "p.value"
        )
      }
    }

    if (!is.null(W)) {
      att.avg.W.j <- jackknifed(
        att.avg.W,
        att.avg.W.boot,
        alpha,
        quantile.CI = quantile.CI
      )
      est.avg.W <- t(as.matrix(c(
        att.avg.W,
        att.avg.W.j$se,
        att.avg.W.j$CI.l,
        att.avg.W.j$CI.u,
        att.avg.W.j$P
      )))
      colnames(est.avg.W) <- c(
        "ATT.avg",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value"
      )

      att.on.W.j <- jackknifed(
        att.on.W,
        att.on.W.boot,
        alpha,
        quantile.CI = quantile.CI
      )
      est.att.W <- cbind(
        att.on.W,
        att.on.W.j$se,
        att.on.W.j$CI.l,
        att.on.W.j$CI.u,
        att.on.W.j$P,
        count.on.W
      )
      colnames(est.att.W) <- c(
        "ATT",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value",
        "count"
      )
      rownames(est.att.W) <- time.on.W

      vcov.att.W <- att.on.W.j$vcov

      att.W.bound <- cbind(
        att.on.W + qnorm(alpha) * att.on.W.j$se,
        att.on.W + qnorm(1 - alpha) * att.on.W.j$se
      )
      colnames(att.W.bound) <- c("CI.lower", "CI.upper")
      rownames(att.W.bound) <- time.on.W

      est.att90.W <- cbind(
        att.on.W,
        att.on.W.j$se,
        att.W.bound,
        att.on.W.j$P,
        count.on.W
      )
      colnames(est.att90.W) <- c(
        "ATT",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value",
        "count"
      )
      rownames(est.att90.W) <- time.on.W

      if (!is.null(placebo.period) & placeboTest == TRUE) {
        att.placebo.W.j <- jackknifed(
          att.placebo.W,
          att.placebo.W.boot,
          alpha,
          quantile.CI = quantile.CI
        )
        est.placebo.W <- t(as.matrix(c(
          att.placebo.W,
          att.placebo.W.j$se,
          att.placebo.W.j$CI.l,
          att.placebo.W.j$CI.u,
          att.placebo.W.j$P
        )))
        colnames(est.placebo.W) <- c(
          "ATT.placebo",
          "S.E.",
          "CI.lower",
          "CI.upper",
          "p.value"
        )
      }
      if (hasRevs == 1) {
        att.off.W.j <- jackknifed(
          att.off.W,
          att.off.W.boot,
          alpha,
          quantile.CI = quantile.CI
        )
        est.att.off.W <- cbind(
          att.off.W,
          att.off.W.j$se,
          att.off.W.j$CI.l,
          att.off.W.j$CI.u,
          att.off.W.j$P,
          count.off.W
        )
        colnames(est.att.off.W) <- c(
          "ATT",
          "S.E.",
          "CI.lower",
          "CI.upper",
          "p.value",
          "count"
        )
        rownames(est.att.off.W) <- time.off.W
        vcov.att.off.W <- att.off.W.j$vcov
        att.off.W.bound <- cbind(
          att.off.W + qnorm(alpha) * att.off.W.j$se,
          att.off.W + qnorm(1 - alpha) * att.off.W.j$se
        )
        colnames(att.off.W.bound) <- c("CI.lower", "CI.upper")
        rownames(att.off.W.bound) <- out$time.off

        if (!is.null(carryover.period) & carryoverTest == TRUE) {
          att.carryover.W.j <- jackknifed(
            att.carryover.W,
            att.carryover.W.boot,
            alpha,
            quantile.CI = quantile.CI
          )
          est.carryover.W <- t(as.matrix(c(
            att.carryover.W,
            att.carryover.W.j$se,
            att.carryover.W.j$CI.l,
            att.carryover.W.j$CI.u,
            att.carryover.W.j$P
          )))
          colnames(est.carryover.W) <- c(
            "ATT.carryover",
            "S.E.",
            "CI.lower",
            "CI.upper",
            "p.value"
          )
        }
      }
    }

    ## average (over time) ATT
    att.avg.j <- jackknifed(
      att.avg,
      att.avg.boot,
      alpha,
      quantile.CI = quantile.CI
    )
    est.avg <- t(as.matrix(c(
      att.avg,
      att.avg.j$se,
      att.avg.j$CI.l,
      att.avg.j$CI.u,
      att.avg.j$P
    )))
    colnames(est.avg) <- c("ATT.avg", "S.E.", "CI.lower", "CI.upper", "p.value")

    att.avg.unit.j <- jackknifed(
      att.avg.unit,
      att.avg.unit.boot,
      alpha,
      quantile.CI = quantile.CI
    )
    est.avg.unit <- t(as.matrix(c(
      att.avg.unit,
      att.avg.unit.j$se,
      att.avg.unit.j$CI.l,
      att.avg.unit.j$CI.u,
      att.avg.unit.j$P
    )))
    colnames(est.avg.unit) <- c(
      "ATT.avg.unit",
      "S.E.",
      "CI.lower",
      "CI.upper",
      "p.value"
    )

    ## regression coefficents
    if (p > 0) {
      beta.j <- jackknifed(beta, beta.boot, alpha, quantile.CI = quantile.CI)
      est.beta <- cbind(beta, beta.j$se, beta.j$CI.l, beta.j$CI.u, beta.j$P)
      colnames(est.beta) <- c("beta", "S.E.", "CI.lower", "CI.upper", "p.value")

      if (binary == TRUE) {
        marginal.j <- jackknifed(
          out$marginal,
          marginal.boot,
          alpha,
          quantile.CI = quantile.CI
        )
        est.marginal <- cbind(
          out$marginal,
          marginal.j$se,
          marginal.j$CI.l,
          marginal.j$CI.u,
          marginal.j$P
        )
        colnames(est.marginal) <- c(
          "marginal",
          "S.E.",
          "CI.lower",
          "CI.upper",
          "p.value"
        )
      }
    }

    ## placebo test
    if (!is.null(placebo.period) & placeboTest == TRUE) {
      att.placebo <- out$att.placebo
      att.placebo.j <- jackknifed(
        att.placebo,
        att.placebo.boot,
        alpha,
        quantile.CI = quantile.CI
      )
      att.placebo.bound <- c(
        att.placebo + qnorm(alpha) * att.placebo.j$se,
        att.placebo + qnorm(1 - alpha) * att.placebo.j$se
      )
      est.placebo <- t(as.matrix(c(
        att.placebo,
        att.placebo.j$se,
        att.placebo.j$CI.l,
        att.placebo.j$CI.u,
        att.placebo.j$P,
        att.placebo.bound
      )))
      colnames(est.placebo) <- c(
        "ATT.placebo",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value",
        "CI.lower(90%)",
        "CI.upper(90%)"
      )
    }

    ## carryover test
    if (!is.null(carryover.period) & carryoverTest == TRUE) {
      att.carryover <- out$att.carryover
      att.carryover.j <- jackknifed(
        att.carryover,
        att.carryover.boot,
        alpha,
        quantile.CI = quantile.CI
      )
      att.carryover.bound <- c(
        att.carryover + qnorm(alpha) * att.carryover.j$se,
        att.carryover + qnorm(1 - alpha) * att.carryover.j$se
      )

      est.carryover <- t(as.matrix(c(
        att.carryover,
        att.carryover.j$se,
        att.carryover.j$CI.l,
        att.carryover.j$CI.u,
        att.carryover.j$P,
        att.carryover.bound
      )))
      colnames(est.carryover) <- c(
        "ATT.carryover",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value",
        "CI.lower(90%)",
        "CI.upper(90%)"
      )
    }

    ## cohort effect
    est.group.out <- NULL
    if (!is.null(group)) {
      group.att.j <- jackknifed(
        group.att,
        group.att.boot,
        alpha,
        quantile.CI = quantile.CI
      )
      est.group.att <- cbind(
        group.att,
        group.att.j$se,
        group.att.j$CI.l,
        group.att.j$CI.u,
        group.att.j$P
      )
      colnames(est.group.att) <- c(
        "ATT",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value"
      )

      est.group.out <- list()
      for (sub.name in group.output.name) {
        subgroup.atts <- group.output.origin[[sub.name]]$att.on
        subgroup.atts.boot <- group.atts.boot[[sub.name]]
        subgroup.est.att <- NULL
        subgroup.att.bound <- NULL

        if (dim(subgroup.atts.boot)[1] > 0) {
          subgroup.att.j <- jackknifed(
            subgroup.atts,
            subgroup.atts.boot,
            alpha,
            quantile.CI = quantile.CI
          )
          subgroup.est.att <- cbind(
            subgroup.atts,
            subgroup.att.j$se,
            subgroup.att.j$CI.l,
            subgroup.att.j$CI.u,
            subgroup.att.j$P,
            group.output.origin[[sub.name]]$count.on
          )
          colnames(subgroup.est.att) <- c(
            "ATT",
            "S.E.",
            "CI.lower",
            "CI.upper",
            "p.value",
            "count"
          )
          rownames(subgroup.est.att) <- group.output.origin[[sub.name]]$time.on

          subgroup.att.bound <- cbind(
            subgroup.atts + qnorm(alpha) * subgroup.att.j$se,
            subgroup.atts + qnorm(1 - alpha) * subgroup.att.j$se
          )
          colnames(subgroup.att.bound) <- c("CI.lower", "CI.upper")
          rownames(subgroup.att.bound) <- group.output.origin[[
            sub.name
          ]]$time.on
        }

        subgroup.est.att.off <- NULL
        subgroup.att.off.bound <- NULL
        if (hasRevs == 1) {
          subgroup.atts.off <- group.output.origin[[sub.name]]$att.off
          subgroup.atts.off.boot <- group.atts.off.boot[[sub.name]]
          if (dim(subgroup.atts.off.boot)[1] > 0) {
            subgroup.att.off.j <- jackknifed(
              subgroup.atts.off,
              subgroup.atts.off.boot,
              alpha,
              quantile.CI = quantile.CI
            )
            subgroup.est.att.off <- cbind(
              subgroup.atts.off,
              subgroup.att.off.j$se,
              subgroup.att.off.j$CI.l,
              subgroup.att.off.j$CI.u,
              subgroup.att.off.j$P,
              group.output.origin[[sub.name]]$count.off
            )
            colnames(subgroup.est.att.off) <- c(
              "ATT",
              "S.E.",
              "CI.lower",
              "CI.upper",
              "p.value",
              "count"
            )
            rownames(subgroup.est.att.off) <- group.output.origin[[
              sub.name
            ]]$time.off

            subgroup.att.off.bound <- cbind(
              subgroup.atts.off + qnorm(alpha) * subgroup.att.off.j$se,
              subgroup.atts.off + qnorm(1 - alpha) * subgroup.att.off.j$se
            )
            colnames(subgroup.att.off.bound) <- c("CI.lower", "CI.upper")
            rownames(subgroup.att.off.bound) <- group.output.origin[[
              sub.name
            ]]$time.off
          }
        }

        subgroup.est.placebo <- NULL
        if (placeboTest) {
          subgroup.att.placebo <- group.output.origin[[sub.name]]$att.placebo
          if (length(subgroup.att.placebo) > 0) {
            subgroup.att.placebo.j <- jackknifed(
              subgroup.att.placebo,
              group.att.placebo.boot[[sub.name]],
              alpha,
              quantile.CI = quantile.CI
            )
            att.placebo.bound <- c(
              subgroup.att.placebo + qnorm(alpha) * subgroup.att.placebo.j$se,
              subgroup.att.placebo +
                qnorm(1 - alpha) * subgroup.att.placebo.j$se
            )

            subgroup.est.placebo <- t(as.matrix(c(
              subgroup.att.placebo,
              subgroup.att.placebo.j$se,
              subgroup.att.placebo.j$CI.l,
              subgroup.att.placebo.j$CI.u,
              subgroup.att.placebo.j$P,
              att.placebo.bound
            )))
            colnames(subgroup.est.placebo) <- c(
              "ATT.placebo",
              "S.E.",
              "CI.lower",
              "CI.upper",
              "p.value",
              "CI.lower(90%)",
              "CI.upper(90%)"
            )
          }
        }

        subgroup.est.carryover <- NULL
        if (carryoverTest) {
          subgroup.att.carryover <- group.output.origin[[
            sub.name
          ]]$att.carryover
          if (length(subgroup.att.carryover) > 0) {
            subgroup.att.carryover.j <- jackknifed(
              subgroup.att.carryover,
              group.att.carryover.boot[[sub.name]],
              alpha,
              quantile.CI = quantile.CI
            )
            att.carryover.bound <- c(
              subgroup.att.carryover +
                qnorm(alpha) * subgroup.att.carryover.j$se,
              subgroup.att.carryover +
                qnorm(1 - alpha) * subgroup.att.carryover.j$se
            )

            subgroup.est.carryover <- t(as.matrix(c(
              subgroup.att.carryover,
              subgroup.att.carryover.j$se,
              subgroup.att.carryover.j$CI.l,
              subgroup.att.carryover.j$CI.u,
              subgroup.att.carryover.j$P,
              att.carryover.bound
            )))
            colnames(subgroup.est.carryover) <- c(
              "ATT.carryover",
              "S.E.",
              "CI.lower",
              "CI.upper",
              "p.value",
              "CI.lower(90%)",
              "CI.upper(90%)"
            )
          }
        }

        est.group.out[[sub.name]] <- list(
          att.on = subgroup.est.att,
          att.on.bound = subgroup.att.bound,
          att.on.boot = group.atts.boot[[sub.name]],
          att.off = subgroup.est.att.off,
          att.off.bound = subgroup.att.off.bound,
          att.off.boot = group.atts.off.boot[[sub.name]],
          att.placebo = subgroup.est.placebo,
          att.carryover = subgroup.est.carryover
        )
      }
    }
  } else {
    ## Single source of truth for the parametric H0 -> H1 location shift
    ## downstream.  Used by every `quantile.CI == TRUE` branch in this
    ## section through the `.basic_ci_shifted()` / `.basic_ci_shifted_one()`
    ## helpers (defined near the top of this file).  See R/po-estimands.R
    ## commit b4e9fbf for the original shift fix on the estimand() side ---
    ## the helpers here apply the same shift to fect's built-in CI
    ## machinery so fit$est.* slots match estimand() byte-equally.
    .is_param <- isTRUE(vartype == "parametric")

    se.att <- apply(att.boot, 1, function(vec) sd(vec, na.rm = TRUE))
    if (quantile.CI == FALSE) {
      CI.att <- cbind(
        att - se.att * qnorm(1 - alpha / 2),
        att + se.att * qnorm(1 - alpha / 2)
      ) # normal approximation
      pvalue.att <- (1 - pnorm(abs(att / se.att))) * 2
    } else {
      CI.att <- .basic_ci_shifted(att, att.boot, alpha, .is_param)
      pvalue.att <- apply(att.boot, 1, get.pvalue)  # original (H0-centered)
    }

    #vcov.att <- cov(t(att.boot), use = "pairwise.complete.obs")
    vcov.att <- tryCatch(
      {
        cov(t(att.boot), use = "pairwise.complete.obs")
      },
      error = function(e) {
        NA
      }
    )
    # for equivalence test
    if (quantile.CI == FALSE) {
      att.bound <- cbind(
        att - se.att * qnorm(1 - alpha),
        att + se.att * qnorm(1 - alpha)
      ) # one-sided
    } else {
      att.bound <- .basic_ci_shifted(att, att.boot, 2 * alpha, .is_param)
    }

    colnames(att.bound) <- c("CI.lower", "CI.upper")
    rownames(att.bound) <- out$time
    est.att <- cbind(att, se.att, CI.att, pvalue.att, out$count)
    est.att90 <- cbind(att, se.att, att.bound, pvalue.att, out$count)
    colnames(est.att) <- c(
      "ATT",
      "S.E.",
      "CI.lower",
      "CI.upper",
      "p.value",
      "count"
    )
    colnames(est.att90) <- c(
      "ATT",
      "S.E.",
      "CI.lower",
      "CI.upper",
      "p.value",
      "count"
    )
    rownames(est.att) <- out$time
    rownames(est.att90) <- out$time

    if (hasRevs == 1) {
      se.att.off <- apply(att.off.boot, 1, function(vec) sd(vec, na.rm = TRUE))
      if (quantile.CI == FALSE) {
        CI.att.off <- cbind(
          att.off - se.att.off * qnorm(1 - alpha / 2),
          att.off + se.att.off * qnorm(1 - alpha / 2)
        )
        pvalue.att.off <- (1 - pnorm(abs(att.off / se.att.off))) * 2
      } else {
        CI.att.off <- .basic_ci_shifted(att.off, att.off.boot, alpha, .is_param)
        pvalue.att.off <- apply(att.off.boot, 1, get.pvalue)
      }

      #vcov.att.off <- cov(t(att.off.boot), use = "pairwise.complete.obs")
      vcov.att.off <- tryCatch(
        {
          cov(t(att.off.boot), use = "pairwise.complete.obs")
        },
        error = function(e) {
          NA
        }
      )

      est.att.off <- cbind(
        att.off,
        se.att.off,
        CI.att.off,
        pvalue.att.off,
        out$count.off
      )
      colnames(est.att.off) <- c(
        "ATT.OFF",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value",
        "count.off"
      )
      rownames(est.att.off) <- out$time.off
      # T0.off.l <- sum(out$time.off > 0)
      # norm.att.off.sq <- (att.off/se.att.off)^2
      # T0.off.p <- 1 - pchisq(sum(norm.att.off.sq[(length(out$time.off) - T0.off.l + 1):length(out$time.off)]), df = T0.off.l)
      if (quantile.CI == FALSE) {
        att.off.bound <- cbind(
          att.off - se.att.off * qnorm(1 - alpha),
          att.off + se.att.off * qnorm(1 - alpha)
        )
      } else {
        att.off.bound <- .basic_ci_shifted(att.off, att.off.boot, 2 * alpha, .is_param)
      }

      colnames(att.off.bound) <- c("CI.lower", "CI.upper")
      rownames(att.off.bound) <- out$time.off
    }

    if (!is.null(T.on.carry)) {
      se.carry.att <- apply(carry.att.boot, 1, function(vec) {
        sd(vec, na.rm = TRUE)
      })
      if (quantile.CI == FALSE) {
        CI.carry.att <- cbind(
          carry.att - se.carry.att * qnorm(1 - alpha / 2),
          carry.att + se.carry.att * qnorm(1 - alpha / 2)
        ) # normal approximation
        pvalue.carry.att <- (1 - pnorm(abs(carry.att / se.carry.att))) * 2
      } else {
        CI.carry.att <- .basic_ci_shifted(carry.att, carry.att.boot, alpha, .is_param)
        pvalue.carry.att <- apply(carry.att.boot, 1, get.pvalue)
      }

      est.carry.att <- cbind(
        carry.att,
        se.carry.att,
        CI.carry.att,
        pvalue.carry.att
      )

      colnames(est.carry.att) <- c(
        "ATT",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value"
      )
      rownames(est.carry.att) <- carry.time
    }

    if (!is.null(balance.period)) {
      se.balance.att <- apply(balance.att.boot, 1, function(vec) {
        sd(vec, na.rm = TRUE)
      })
      if (quantile.CI == FALSE) {
        CI.balance.att <- cbind(
          balance.att - se.balance.att * qnorm(1 - alpha / 2),
          balance.att + se.balance.att * qnorm(1 - alpha / 2)
        )
        pvalue.balance.att <- (1 - pnorm(abs(balance.att / se.balance.att))) * 2
      } else {
        CI.balance.att <- .basic_ci_shifted(balance.att, balance.att.boot, alpha, .is_param)
        pvalue.balance.att <- apply(balance.att.boot, 1, get.pvalue)
      }

      #vcov.balance.att <- cov(t(balance.att.boot), use = "pairwise.complete.obs")

      vcov.balance.att <- tryCatch(
        {
          cov(t(balance.att.boot), use = "pairwise.complete.obs")
        },
        error = function(e) {
          NA
        }
      )

      est.balance.att <- cbind(
        balance.att,
        se.balance.att,
        CI.balance.att,
        pvalue.balance.att,
        out$balance.count
      )
      colnames(est.balance.att) <- c(
        "ATT",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value",
        "count"
      )
      rownames(est.balance.att) <- out$balance.time

      se.balance.avg.att <- sd(balance.avg.att.boot, na.rm = TRUE)
      if (quantile.CI == FALSE) {
        CI.balance.avg.att <- c(
          balance.avg.att - se.balance.avg.att * qnorm(1 - alpha / 2),
          balance.avg.att + se.balance.avg.att * qnorm(1 - alpha / 2)
        )
        p.balance.avg.att <- (1 -
          pnorm(abs(balance.avg.att / se.balance.avg.att))) *
          2
      } else {
        CI.balance.avg.att <- .basic_ci_shifted_one(balance.avg.att,
                                                    balance.avg.att.boot,
                                                    alpha, .is_param)
        p.balance.avg.att <- get.pvalue(balance.avg.att.boot)
      }

      est.balance.avg <- t(as.matrix(c(
        balance.avg.att,
        se.balance.avg.att,
        CI.balance.avg.att,
        p.balance.avg.att
      )))
      colnames(est.balance.avg) <- c(
        "ATT.avg",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value"
      )
      if (quantile.CI == FALSE) {
        balance.att.bound <- cbind(
          balance.att - se.balance.att * qnorm(1 - alpha),
          balance.att + se.balance.att * qnorm(1 - alpha)
        )
      } else {
        balance.att.bound <- .basic_ci_shifted(balance.att, balance.att.boot,
                                               2 * alpha, .is_param)
      }

      colnames(balance.att.bound) <- c("CI.lower", "CI.upper")
      rownames(balance.att.bound) <- out$balance.time

      if (!is.null(placebo.period) & placeboTest == TRUE) {
        balance.att.placebo <- out$balance.att.placebo
        balance.se.placebo <- sd(balance.att.placebo.boot, na.rm = TRUE)
        if (quantile.CI == FALSE) {
          balance.CI.placebo <- c(
            balance.att.placebo - balance.se.placebo * qnorm(1 - alpha / 2),
            balance.att.placebo + balance.se.placebo * qnorm(1 - alpha / 2)
          )
          balance.CI.placebo.bound <- c(
            balance.att.placebo - balance.se.placebo * qnorm(1 - alpha),
            balance.att.placebo + balance.se.placebo * qnorm(1 - alpha)
          )
          balance.pvalue.placebo <- (1 -
            pnorm(abs(balance.att.placebo / balance.se.placebo))) *
            2
        } else {
          balance.CI.placebo <- .basic_ci_shifted_one(balance.att.placebo,
                                                      balance.att.placebo.boot,
                                                      alpha, .is_param)
          balance.CI.placebo.bound <- .basic_ci_shifted_one(balance.att.placebo,
                                                            balance.att.placebo.boot,
                                                            2 * alpha, .is_param)
          balance.pvalue.placebo <- get.pvalue(balance.att.placebo.boot)
        }

        est.balance.placebo <- t(as.matrix(c(
          balance.att.placebo,
          balance.se.placebo,
          balance.CI.placebo,
          balance.pvalue.placebo,
          balance.CI.placebo.bound
        )))
        colnames(est.balance.placebo) <- c(
          "ATT.placebo",
          "S.E.",
          "CI.lower",
          "CI.upper",
          "p.value",
          "CI.lower(90%)",
          "CI.upper(90%)"
        )
      }
    }

    if (!is.null(W)) {
      # att.avg.W.boot
      se.att.avg.W <- sd(att.avg.W.boot, na.rm = TRUE)
      if (quantile.CI == FALSE) {
        CI.att.avg.W <- c(
          att.avg.W - se.att.avg.W * qnorm(1 - alpha / 2),
          att.avg.W + se.att.avg.W * qnorm(1 - alpha / 2)
        )
        p.att.avg.W <- (1 - pnorm(abs(att.avg.W / se.att.avg.W))) * 2
      } else {
        CI.att.avg.W <- .basic_ci_shifted_one(att.avg.W, att.avg.W.boot,
                                               alpha, .is_param)
        p.att.avg.W <- get.pvalue(att.avg.W.boot)
      }

      est.avg.W <- t(as.matrix(c(
        att.avg.W,
        se.att.avg.W,
        CI.att.avg.W,
        p.att.avg.W
      )))
      colnames(est.avg.W) <- c(
        "ATT.avg",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value"
      )

      # att.on.W.boot
      se.att.W <- apply(att.on.W.boot, 1, function(vec) sd(vec, na.rm = TRUE))
      if (quantile.CI == FALSE) {
        CI.att.W <- cbind(
          att.on.W - se.att.W * qnorm(1 - alpha / 2),
          att.on.W + se.att.W * qnorm(1 - alpha / 2)
        )
        att.W.bound <- cbind(
          att.on.W - se.att.W * qnorm(1 - alpha),
          att.on.W + se.att.W * qnorm(1 - alpha)
        )
        pvalue.att.W <- (1 - pnorm(abs(att.on.W / se.att.W))) * 2
      } else {
        CI.att.W   <- .basic_ci_shifted(att.on.W, att.on.W.boot, alpha,     .is_param)
        att.W.bound <- .basic_ci_shifted(att.on.W, att.on.W.boot, 2 * alpha, .is_param)
        pvalue.att.W <- apply(att.on.W.boot, 1, get.pvalue)
      }

      #vcov.att.W <- cov(t(att.on.W.boot), use = "pairwise.complete.obs")
      vcov.att.W <- tryCatch(
        {
          cov(t(att.on.W.boot), use = "pairwise.complete.obs")
        },
        error = function(e) {
          NA
        }
      )

      est.att.W <- cbind(
        att.on.W,
        se.att.W,
        CI.att.W,
        pvalue.att.W,
        count.on.W
      )
      colnames(est.att.W) <- c(
        "ATT",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value",
        "count"
      )
      rownames(est.att.W) <- time.on.W

      colnames(att.W.bound) <- c("CI.lower", "CI.upper")
      rownames(att.W.bound) <- time.on.W

      est.att90.W <- cbind(
        att.on.W,
        se.att.W,
        att.W.bound,
        pvalue.att.W,
        count.on.W
      )
      colnames(est.att90.W) <- c(
        "ATT",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value",
        "count"
      )
      rownames(est.att90.W) <- time.on.W

      if (!is.null(placebo.period) & placeboTest == TRUE) {
        # att.placebo.W.boot
        se.placebo.W <- sd(att.placebo.W.boot, na.rm = TRUE)
        if (quantile.CI == FALSE) {
          CI.placebo.W <- c(
            att.placebo.W - se.placebo.W * qnorm(1 - alpha / 2),
            att.placebo.W + se.placebo.W * qnorm(1 - alpha / 2)
          )
          CI.placebo.bound.W <- c(
            att.placebo.W - se.placebo.W * qnorm(1 - alpha),
            att.placebo.W + se.placebo.W * qnorm(1 - alpha)
          )
          pvalue.placebo.w <- (1 - pnorm(abs(att.placebo.W / se.placebo.W))) * 2
        } else {
          CI.placebo.W       <- .basic_ci_shifted_one(att.placebo.W, att.placebo.W.boot, alpha,     .is_param)
          CI.placebo.bound.W <- .basic_ci_shifted_one(att.placebo.W, att.placebo.W.boot, 2 * alpha, .is_param)
          pvalue.placebo.w <- get.pvalue(att.placebo.W.boot)
        }

        est.placebo.W <- t(as.matrix(c(
          att.placebo.W,
          se.placebo.W,
          CI.placebo.W,
          pvalue.placebo.w,
          CI.placebo.bound.W
        )))
        colnames(est.placebo.W) <- c(
          "ATT.placebo",
          "S.E.",
          "CI.lower",
          "CI.upper",
          "p.value",
          "CI.lower(90%)",
          "CI.upper(90%)"
        )
      }
      if (hasRevs == 1) {
        # att.off.W.boot
        se.att.off.W <- apply(att.off.W.boot, 1, function(vec) {
          sd(vec, na.rm = TRUE)
        })
        if (quantile.CI == FALSE) {
          CI.att.off.W <- cbind(
            att.off.W - se.att.off.W * qnorm(1 - alpha / 2),
            att.off.W + se.att.off.W * qnorm(1 - alpha / 2)
          )
          att.off.W.bound <- cbind(
            att.off.W - se.att.off.W * qnorm(1 - alpha),
            att.off.W + se.att.off.W * qnorm(1 - alpha)
          )
          pvalue.att.off.W <- (1 - pnorm(abs(att.off.W / se.att.off.W))) * 2
        } else {
          CI.att.off.W   <- .basic_ci_shifted(att.off.W, att.off.W.boot, alpha,     .is_param)
          att.off.W.bound <- .basic_ci_shifted(att.off.W, att.off.W.boot, 2 * alpha, .is_param)
          pvalue.att.off.W <- apply(att.off.W.boot, 1, get.pvalue)
        }

        #vcov.att.off.W <- cov(t(att.off.W.boot), use = "pairwise.complete.obs")
        vcov.att.off.W <- tryCatch(
          {
            cov(t(att.off.W.boot), use = "pairwise.complete.obs")
          },
          error = function(e) {
            NA
          }
        )

        est.att.off.W <- cbind(
          att.off.W,
          se.att.off.W,
          CI.att.off.W,
          pvalue.att.off.W,
          count.off.W
        )
        colnames(est.att.off.W) <- c(
          "ATT",
          "S.E.",
          "CI.lower",
          "CI.upper",
          "p.value",
          "count"
        )
        rownames(est.att.off.W) <- time.off.W
        colnames(att.off.W.bound) <- c("CI.lower", "CI.upper")
        rownames(att.off.W.bound) <- time.off.W

        if (!is.null(carryover.period) & carryoverTest == TRUE) {
          # att.carryover.W.boot
          se.carryover.W <- sd(att.carryover.W.boot, na.rm = TRUE)
          if (quantile.CI == FALSE) {
            CI.carryover.W <- c(
              att.carryover.W - se.carryover.W * qnorm(1 - alpha / 2),
              att.carryover.W + se.carryover.W * qnorm(1 - alpha / 2)
            )
            CI.carryover.bound.W <- c(
              att.carryover.W - se.carryover.W * qnorm(1 - alpha),
              att.carryover.W + se.carryover.W * qnorm(1 - alpha)
            )
            pvalue.carryover.w <- (1 -
              pnorm(abs(att.carryover.W / se.carryover.W))) *
              2
          } else {
            CI.carryover.W       <- .basic_ci_shifted_one(att.carryover.W, att.carryover.W.boot, alpha,     .is_param)
            CI.carryover.bound.W <- .basic_ci_shifted_one(att.carryover.W, att.carryover.W.boot, 2 * alpha, .is_param)
            pvalue.carryover.w <- get.pvalue(att.carryover.W.boot)
          }

          est.carryover.W <- t(as.matrix(c(
            att.carryover.W,
            se.carryover.W,
            CI.carryover.W,
            pvalue.carryover.w,
            CI.carryover.bound.W
          )))
          colnames(est.carryover.W) <- c(
            "ATT.carryover",
            "S.E.",
            "CI.lower",
            "CI.upper",
            "p.value",
            "CI.lower(90%)",
            "CI.upper(90%)"
          )
        }
      }
    }

    ## average (over time) ATT
    se.avg <- sd(att.avg.boot, na.rm = TRUE)
    if (quantile.CI == FALSE) {
      CI.avg <- c(
        att.avg - se.avg * qnorm(1 - alpha / 2),
        att.avg + se.avg * qnorm(1 - alpha / 2)
      )
      pvalue.avg <- (1 - pnorm(abs(att.avg / se.avg))) * 2
    } else {
      ## ci.method = "basic": reflected pivot interval (with parametric
      ## location shift via .basic_ci_shifted_one).  NOTE: the legacy
      ## quantile.CI = TRUE path here previously returned raw percentile
      ## quantiles (an inconsistency with the per-event-time block, which
      ## already used basic).  v2.4.2 standardizes on basic at both sites.
      CI.avg <- .basic_ci_shifted_one(att.avg, att.avg.boot, alpha, .is_param)
      pvalue.avg <- get.pvalue(att.avg.boot)
    }

    est.avg <- t(as.matrix(c(att.avg, se.avg, CI.avg, pvalue.avg)))
    colnames(est.avg) <- c("ATT.avg", "S.E.", "CI.lower", "CI.upper", "p.value")

    se.avg.unit <- sd(att.avg.unit.boot, na.rm = TRUE)
    if (quantile.CI == FALSE) {
      CI.avg.unit <- c(
        att.avg.unit - se.avg.unit * qnorm(1 - alpha / 2),
        att.avg.unit + se.avg.unit * qnorm(1 - alpha / 2)
      )
      pvalue.avg.unit <- (1 - pnorm(abs(att.avg.unit / se.avg.unit))) * 2
    } else {
      CI.avg.unit <- .basic_ci_shifted_one(att.avg.unit, att.avg.unit.boot,
                                            alpha, .is_param)
      pvalue.avg.unit <- get.pvalue(att.avg.unit.boot)
    }

    est.avg.unit <- t(as.matrix(c(
      att.avg.unit,
      se.avg.unit,
      CI.avg.unit,
      pvalue.avg.unit
    )))
    colnames(est.avg.unit) <- c(
      "ATT.avg.unit",
      "S.E.",
      "CI.lower",
      "CI.upper",
      "p.value"
    )

    se.eff.calendar <- apply(calendar.eff.boot, 1, function(vec) {
      sd(vec, na.rm = TRUE)
    })
    if (quantile.CI == FALSE) {
      CI.eff.calendar <- cbind(
        calendar.eff - se.eff.calendar * qnorm(1 - alpha / 2),
        calendar.eff + se.eff.calendar * qnorm(1 - alpha / 2)
      )
      pvalue.eff.calendar <- (1 - pnorm(abs(calendar.eff / se.eff.calendar))) *
        2
    } else {
      CI.eff.calendar <- .basic_ci_shifted(calendar.eff, calendar.eff.boot,
                                            alpha, .is_param)
      pvalue.eff.calendar <- apply(calendar.eff.boot, 1, get.pvalue)
    }
    est.eff.calendar <- cbind(
      calendar.eff,
      se.eff.calendar,
      CI.eff.calendar,
      pvalue.eff.calendar,
      calendar.N
    )
    colnames(est.eff.calendar) <- c(
      "ATT-calendar",
      "S.E.",
      "CI.lower",
      "CI.upper",
      "p.value",
      "count"
    )

    se.eff.calendar.fit <- apply(calendar.eff.fit.boot, 1, function(vec) {
      sd(vec, na.rm = TRUE)
    })
    if (quantile.CI == FALSE) {
      CI.eff.calendar.fit <- cbind(
        calendar.eff.fit - se.eff.calendar.fit * qnorm(1 - alpha / 2),
        calendar.eff.fit + se.eff.calendar.fit * qnorm(1 - alpha / 2)
      )
      pvalue.eff.calendar.fit <- (1 -
        pnorm(abs(calendar.eff.fit / se.eff.calendar.fit))) *
        2
    } else {
      CI.eff.calendar.fit <- .basic_ci_shifted(calendar.eff.fit,
                                                calendar.eff.fit.boot,
                                                alpha, .is_param)
      pvalue.eff.calendar.fit <- apply(calendar.eff.fit.boot, 1, get.pvalue)
    }
    est.eff.calendar.fit <- cbind(
      calendar.eff.fit,
      se.eff.calendar.fit,
      CI.eff.calendar.fit,
      pvalue.eff.calendar.fit,
      calendar.N
    )
    colnames(est.eff.calendar.fit) <- c(
      "ATT-calendar Fitted",
      "S.E.",
      "CI.lower",
      "CI.upper",
      "p.value",
      "count"
    )

    ## regression coefficents
    if (p > 0) {
      se.beta <- apply(beta.boot, 1, function(vec) sd(vec, na.rm = TRUE))
      if (quantile.CI == FALSE) {
        CI.beta <- cbind(
          c(beta) - se.beta * qnorm(1 - alpha / 2),
          c(beta) + se.beta * qnorm(1 - alpha / 2)
        )
        pvalue.beta <- (1 - pnorm(abs(beta / se.beta))) * 2
      } else {
        CI.beta <- .basic_ci_shifted(c(beta), beta.boot, alpha, .is_param)
        pvalue.beta <- apply(beta.boot, 1, get.pvalue)
      }
      est.beta <- cbind(c(beta), se.beta, CI.beta, pvalue.beta)
      colnames(est.beta) <- c("Coef", "S.E.", "CI.lower", "CI.upper", "p.value")

      if (binary == TRUE) {
        out$marginal[na.pos] <- NA
        se.marginal <- apply(marginal.boot, 1, function(vec) {
          sd(vec, na.rm = TRUE)
        })
        CI.marginal <- cbind(
          c(out$marginal) - se.marginal * qnorm(1 - alpha / 2),
          c(out$marginal) + se.marginal * qnorm(1 - alpha / 2)
        )
        pvalue.marginal <- (1 - pnorm(abs(out$marginal / se.marginal))) * 2
        est.marginal <- cbind(
          out$marginal,
          se.marginal,
          CI.marginal,
          pvalue.marginal
        )
        colnames(est.marginal) <- c(
          "marginal",
          "S.E.",
          "CI.lower",
          "CI.upper",
          "p.value"
        )
      }
    }

    ## placebo test
    if (!is.null(placebo.period) & placeboTest == TRUE) {
      att.placebo <- out$att.placebo
      se.placebo <- sd(att.placebo.boot, na.rm = TRUE)
      if (quantile.CI == FALSE) {
        CI.placebo <- c(
          att.placebo - se.placebo * qnorm(1 - alpha / 2),
          att.placebo + se.placebo * qnorm(1 - alpha / 2)
        )
        CI.placebo.bound <- c(
          att.placebo - se.placebo * qnorm(1 - alpha),
          att.placebo + se.placebo * qnorm(1 - alpha)
        )
        pvalue.placebo <- (1 - pnorm(abs(att.placebo / se.placebo))) * 2
      } else {
        CI.placebo       <- .basic_ci_shifted_one(att.placebo, att.placebo.boot, alpha,     .is_param)
        CI.placebo.bound <- .basic_ci_shifted_one(att.placebo, att.placebo.boot, 2 * alpha, .is_param)
        pvalue.placebo <- get.pvalue(att.placebo.boot)
      }

      est.placebo <- t(as.matrix(c(
        att.placebo,
        se.placebo,
        CI.placebo,
        pvalue.placebo,
        CI.placebo.bound
      )))
      colnames(est.placebo) <- c(
        "ATT.placebo",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value",
        "CI.lower(90%)",
        "CI.upper(90%)"
      )
    }

    ## carryover test
    if (!is.null(carryover.period) & carryoverTest == TRUE) {
      att.carryover <- out$att.carryover
      se.carryover <- sd(att.carryover.boot, na.rm = TRUE)
      if (quantile.CI == FALSE) {
        CI.carryover <- c(
          att.carryover - se.carryover * qnorm(1 - alpha / 2),
          att.carryover + se.carryover * qnorm(1 - alpha / 2)
        )
        CI.carryover.bound <- c(
          att.carryover - se.carryover * qnorm(1 - alpha),
          att.carryover + se.carryover * qnorm(1 - alpha)
        )
        pvalue.carryover <- (1 - pnorm(abs(att.carryover / se.carryover))) * 2
      } else {
        CI.carryover       <- .basic_ci_shifted_one(att.carryover, att.carryover.boot, alpha,     .is_param)
        CI.carryover.bound <- .basic_ci_shifted_one(att.carryover, att.carryover.boot, 2 * alpha, .is_param)
        pvalue.carryover <- get.pvalue(att.carryover.boot)
      }
      est.carryover <- t(as.matrix(c(
        att.carryover,
        se.carryover,
        CI.carryover,
        pvalue.carryover,
        CI.carryover.bound
      )))
      colnames(est.carryover) <- c(
        "ATT.carryover",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value",
        "CI.lower(90%)",
        "CI.upper(90%)"
      )
    }

    ## group effect
    if (!is.null(group)) {
      se.group.att <- apply(group.att.boot, 1, function(vec) {
        sd(vec, na.rm = TRUE)
      })
      if (quantile.CI == TRUE) {
        CI.group.att <- cbind(
          c(out$group.att) - se.group.att * qnorm(1 - alpha / 2),
          c(out$group.att) + se.group.att * qnorm(1 - alpha / 2)
        )
        pvalue.group.att <- (1 - pnorm(abs(out$group.att / se.group.att))) * 2
      } else {
        CI.group.att <- .basic_ci_shifted(c(out$group.att), group.att.boot,
                                           alpha, .is_param)
        pvalue.group.att <- apply(group.att.boot, 1, get.pvalue)
      }

      est.group.att <- cbind(
        out$group.att,
        se.group.att,
        CI.group.att,
        pvalue.group.att
      )
      colnames(est.group.att) <- c(
        "ATT",
        "S.E.",
        "CI.lower",
        "CI.upper",
        "p.value"
      )

      est.group.out <- list()
      for (sub.name in group.output.name) {
        subgroup.atts <- group.output.origin[[sub.name]]$att.on
        subgroup.atts.boot <- group.atts.boot[[sub.name]]
        subgroup.est.att <- NULL
        subgroup.att.bound <- NULL
        if (dim(subgroup.atts.boot)[1] > 0) {
          subgroup.se.att <- apply(subgroup.atts.boot, 1, function(vec) {
            sd(vec, na.rm = TRUE)
          })
          if (quantile.CI == FALSE) {
            subgroup.CI.att <- cbind(
              subgroup.atts - subgroup.se.att * qnorm(1 - alpha / 2),
              subgroup.atts + subgroup.se.att * qnorm(1 - alpha / 2)
            )
            subgroup.pvalue.att <- (1 -
              pnorm(abs(subgroup.atts / subgroup.se.att))) *
              2
            subgroup.att.bound <- cbind(
              subgroup.atts - subgroup.se.att * qnorm(1 - alpha),
              subgroup.atts + subgroup.se.att * qnorm(1 - alpha)
            )
          } else {
            subgroup.CI.att   <- .basic_ci_shifted(subgroup.atts, subgroup.atts.boot, alpha,     .is_param)
            subgroup.pvalue.att <- apply(subgroup.atts.boot, 1, get.pvalue)
            subgroup.att.bound <- .basic_ci_shifted(subgroup.atts, subgroup.atts.boot, 2 * alpha, .is_param)
          }
          subgroup.est.att <- cbind(
            subgroup.atts,
            subgroup.se.att,
            subgroup.CI.att,
            subgroup.pvalue.att,
            group.output.origin[[sub.name]]$count.on
          )
          colnames(subgroup.est.att) <- c(
            "ATT",
            "S.E.",
            "CI.lower",
            "CI.upper",
            "p.value",
            "count"
          )
          rownames(subgroup.est.att) <- group.output.origin[[sub.name]]$time.on

          # for equivalence test
          colnames(subgroup.att.bound) <- c("CI.lower", "CI.upper")
          rownames(subgroup.att.bound) <- group.output.origin[[
            sub.name
          ]]$time.on
        }

        subgroup.att.off.bound <- NULL
        subgroup.est.att.off <- NULL
        if (hasRevs == 1) {
          subgroup.atts.off <- group.output.origin[[sub.name]]$att.off
          subgroup.atts.off.boot <- group.atts.off.boot[[sub.name]]

          if (dim(subgroup.atts.off.boot)[1] > 0) {
            subgroup.se.att.off <- apply(
              subgroup.atts.off.boot,
              1,
              function(vec) sd(vec, na.rm = TRUE)
            )
            if (quantile.CI == FALSE) {
              subgroup.CI.att.off <- cbind(
                subgroup.atts.off - subgroup.se.att.off * qnorm(1 - alpha / 2),
                subgroup.atts.off + subgroup.se.att.off * qnorm(1 - alpha / 2)
              )
              subgroup.pvalue.att.off <- apply(
                subgroup.atts.off.boot,
                1,
                get.pvalue
              )
              subgroup.att.off.bound <- cbind(
                subgroup.atts.off - subgroup.se.att.off * qnorm(1 - alpha),
                subgroup.atts.off + subgroup.se.att.off * qnorm(1 - alpha)
              )
            } else {
              subgroup.CI.att.off <- .basic_ci_shifted(subgroup.atts.off,
                                                        subgroup.atts.off.boot,
                                                        alpha, .is_param)
              subgroup.pvalue.att.off <- apply(subgroup.atts.off.boot, 1, get.pvalue)
              subgroup.att.off.bound <- .basic_ci_shifted(subgroup.atts.off,
                                                           subgroup.atts.off.boot,
                                                           2 * alpha, .is_param)
            }
            subgroup.est.att.off <- cbind(
              subgroup.atts.off,
              subgroup.se.att.off,
              subgroup.CI.att.off,
              subgroup.pvalue.att.off,
              group.output.origin[[sub.name]]$count.off
            )
            colnames(subgroup.est.att.off) <- c(
              "ATT.OFF",
              "S.E.",
              "CI.lower",
              "CI.upper",
              "p.value",
              "count.off"
            )
            rownames(subgroup.est.att.off) <- group.output.origin[[
              sub.name
            ]]$time.off

            colnames(subgroup.att.off.bound) <- c("CI.lower", "CI.upper")
            rownames(subgroup.att.off.bound) <- group.output.origin[[
              sub.name
            ]]$time.off
          }
        }

        ## placebo test
        subgroup.est.placebo <- NULL
        if (!is.null(placebo.period) & placeboTest == TRUE) {
          subgroup.att.placebo <- group.output.origin[[sub.name]]$att.placebo
          if (length(subgroup.att.placebo) > 0) {
            subgroup.se.placebo <- sd(
              group.att.placebo.boot[[sub.name]],
              na.rm = TRUE
            )
            if (quantile.CI == FALSE) {
              subgroup.CI.placebo <- c(
                subgroup.att.placebo -
                  subgroup.se.placebo * qnorm(1 - alpha / 2),
                subgroup.att.placebo +
                  subgroup.se.placebo * qnorm(1 - alpha / 2)
              )
              subgroup.CI.placebo.bound <- c(
                subgroup.att.placebo - subgroup.se.placebo * qnorm(1 - alpha),
                subgroup.att.placebo + subgroup.se.placebo * qnorm(1 - alpha)
              )
              subgroup.pvalue.placebo <- (1 -
                pnorm(abs(subgroup.att.placebo / subgroup.se.placebo))) *
                2
            } else {
              subgroup.CI.placebo       <- .basic_ci_shifted_one(subgroup.att.placebo,
                                                                  group.att.placebo.boot[[sub.name]],
                                                                  alpha, .is_param)
              subgroup.CI.placebo.bound <- .basic_ci_shifted_one(subgroup.att.placebo,
                                                                  group.att.placebo.boot[[sub.name]],
                                                                  2 * alpha, .is_param)
              subgroup.pvalue.placebo <- get.pvalue(group.att.placebo.boot[[sub.name]])
            }

            subgroup.est.placebo <- t(as.matrix(c(
              subgroup.att.placebo,
              subgroup.se.placebo,
              subgroup.CI.placebo,
              subgroup.pvalue.placebo,
              subgroup.CI.placebo.bound
            )))
            colnames(subgroup.est.placebo) <- c(
              "ATT.placebo",
              "S.E.",
              "CI.lower",
              "CI.upper",
              "p.value",
              "CI.lower(90%)",
              "CI.upper(90%)"
            )
          }
        }

        ## carryover test
        subgroup.est.carryover <- NULL
        if (!is.null(carryover.period) & carryoverTest == TRUE) {
          subgroup.att.carryover <- group.output.origin[[
            sub.name
          ]]$att.carryover
          if (length(subgroup.att.carryover) > 0) {
            subgroup.se.carryover <- sd(
              group.att.carryover.boot[[sub.name]],
              na.rm = TRUE
            )
            if (quantile.CI == FALSE) {
              subgroup.CI.carryover <- c(
                subgroup.att.carryover -
                  subgroup.se.carryover * qnorm(1 - alpha / 2),
                subgroup.att.carryover +
                  subgroup.se.carryover * qnorm(1 - alpha / 2)
              )
              subgroup.CI.carryover.bound <- c(
                subgroup.att.carryover -
                  subgroup.se.carryover * qnorm(1 - alpha),
                subgroup.att.carryover +
                  subgroup.se.carryover * qnorm(1 - alpha)
              )
              subgroup.pvalue.carryover <- (1 -
                pnorm(abs(subgroup.att.carryover / subgroup.se.carryover))) *
                2
            } else {
              subgroup.CI.carryover       <- .basic_ci_shifted_one(subgroup.att.carryover,
                                                                    group.att.carryover.boot[[sub.name]],
                                                                    alpha, .is_param)
              subgroup.CI.carryover.bound <- .basic_ci_shifted_one(subgroup.att.carryover,
                                                                    group.att.carryover.boot[[sub.name]],
                                                                    2 * alpha, .is_param)
              subgroup.pvalue.carryover <- get.pvalue(group.att.carryover.boot[[sub.name]])
            }

            subgroup.est.carryover <- t(as.matrix(c(
              subgroup.att.carryover,
              subgroup.se.carryover,
              subgroup.CI.carryover,
              subgroup.pvalue.carryover,
              subgroup.CI.carryover.bound
            )))
            colnames(subgroup.est.carryover) <- c(
              "ATT.carryover",
              "S.E.",
              "CI.lower",
              "CI.upper",
              "p.value",
              "CI.lower(90%)",
              "CI.upper(90%)"
            )
          }
        }

        est.group.out[[sub.name]] <- list(
          att.on = subgroup.est.att,
          att.on.bound = subgroup.att.bound,
          att.on.boot = group.atts.boot[[sub.name]],
          att.off = subgroup.est.att.off,
          att.off.bound = subgroup.att.off.bound,
          att.off.boot = group.atts.off.boot[[sub.name]],
          att.placebo = subgroup.est.placebo,
          att.carryover = subgroup.est.carryover
        )
      }
    }
  }

  ## storage
  ## When W is supplied AND aggregation should reflect those weights
  ## (W or W.agg supplied), route the W-weighted aggregations into the
  ## canonical slot names. When only W.est is supplied (W in fit only),
  ## the canonical aggregation stays unweighted. The fect.default() tail
  ## does the same role-gated routing for the per-method on/off/avg
  ## vectors.
  if (!is.null(W) && isTRUE(W.in.agg)) {
    est.avg      <- est.avg.W
    att.bound    <- att.W.bound
    att.avg.boot <- att.avg.W.boot
    est.att      <- est.att.W
    est.att90    <- est.att90.W
    att.boot     <- att.on.W.boot
    vcov.att     <- vcov.att.W
  }
  result <- list(
    est.avg = est.avg,
    att.bound = att.bound,
    att.avg.boot = att.avg.boot,
    est.avg.unit = est.avg.unit,
    att.avg.unit.boot = att.avg.unit.boot,
    est.eff.calendar = est.eff.calendar,
    est.eff.calendar.fit = est.eff.calendar.fit,
    est.att = est.att,
    est.att90 = est.att90,
    att.boot = att.boot,
    att.boot.original = att.boot.original,
    att.vcov = vcov.att,
    att.count.boot = att.count.boot,
    vartype = vartype,
    para.error = if (vartype == "parametric" && exists("para.error.resolved")) para.error.resolved else NULL
  )
  if (keep.sims) {
    result = c(
      result,
      list(
        eff.boot = eff.boot,
        D.boot = D.boot,
        I.boot = I.boot,
        colnames.boot = colnames.boot
      )
    )
  }

  if (p > 0) {
    result <- c(result, list(beta.boot = beta.boot))
    result <- c(result, list(est.beta = est.beta))
    if (binary == TRUE) {
      result <- c(result, list(est.marginal = est.marginal))
    }
  }
  if (hasRevs == 1) {
    if (!is.null(W) && isTRUE(W.in.agg)) {
      est.att.off  <- est.att.off.W
      att.off.boot <- att.off.W.boot
      vcov.att.off <- vcov.att.off.W
      att.off.bound <- att.off.W.bound
    }
    result <- c(
      result,
      list(
        est.att.off = est.att.off,
        att.off.boot = att.off.boot,
        att.off.vcov = vcov.att.off,
        att.off.bound = att.off.bound,
        att.off.count.boot = att.off.count.boot
      )
    )
    if (keep.sims) {
      result <- c(result, list(eff.off.boot = eff.off.boot))
    }
  }
  if (!is.null(T.on.carry)) {
    result <- c(result, list(est.carry.att = est.carry.att))
  }

  if (!is.null(balance.period)) {
    result <- c(result, list(est.balance.att = est.balance.att))
    result <- c(result, list(est.balance.avg = est.balance.avg))
    result <- c(
      result,
      list(
        balance.att.bound = balance.att.bound,
        balance.att.vcov = vcov.balance.att,
        balance.att.boot = balance.att.boot,
        balance.count.boot = balance.count.boot
      )
    )
    if (!is.null(placebo.period) & placeboTest == TRUE) {
      result <- c(
        result,
        list(
          est.balance.placebo = est.balance.placebo,
          balance.att.placebo.boot = balance.att.placebo.boot
        )
      )
    }
  }
  if (!is.null(placebo.period) & placeboTest == TRUE) {
    if (!is.null(W) && isTRUE(W.in.agg)) {
      est.placebo       <- est.placebo.W
      att.placebo.boot  <- att.placebo.W.boot
    }
    result <- c(
      result,
      list(est.placebo = est.placebo, att.placebo.boot = att.placebo.boot)
    )
  }

  if (!is.null(carryover.period) & carryoverTest == TRUE) {
    if (!is.null(W) && isTRUE(W.in.agg)) {
      est.carryover       <- est.carryover.W
      att.carryover.boot  <- att.carryover.W.boot
    }
    result <- c(
      result,
      list(
        est.carryover = est.carryover,
        att.carryover.boot = att.carryover.boot
      )
    )
  }

  if (!is.null(group)) {
    result <- c(
      result,
      list(
        est.group.att = est.group.att,
        est.group.output = est.group.out
      )
    )
  }

  return(c(out, result))
} ## end of boot


## jackknife se
jackknifed <- function(
  x, ## ols estimates
  y,
  alpha,
  quantile.CI = FALSE
) {
  ## sub-sample ols estimates)

  p <- length(x)
  N <- dim(y)[2] ## sample size

  X <- matrix(rep(c(x), N), p, N) * N
  Y <- X - y * (N - 1)

  Yvar <- apply(Y, 1, var, na.rm = TRUE)
  vn <- N - apply(is.na(y), 1, sum)

  Ysd <- sqrt(Yvar / vn) ## jackknife se

  vcov_matrix <- 1 / vn * cov(t(Y), use = "pairwise.complete.obs")

  if (quantile.CI == FALSE) {
    CI.l <- Ysd * qnorm(alpha / 2) + c(x)
    CI.u <- Ysd * qnorm(1 - alpha / 2) + c(x)
  } else {
    CI <- t(apply(y, 1, function(vec) {
      quantile(vec, c(0.05 / 2, 1 - 0.05 / 2), na.rm = TRUE)
    }))
    CI.l <- CI[, 1]
    CI.u <- CI[, 2]
  }

  ## wald test
  P <- NULL
  for (i in 1:p) {
    subz <- pnorm(c(x)[i] / Ysd[i])
    P <- c(P, 2 * min(1 - subz, subz))
  }

  ## P <- 2 * min(1 - pnorm(c(x)/Ysd), pnorm(c(x)/Ysd))

  out <- list(se = Ysd, CI.l = CI.l, CI.u = CI.u, P = P, vcov = vcov_matrix)

  return(out)
}

## ---------------------------------------------------------
## Phase tests for the time.component.from refactoring
##
## These tests define acceptance criteria for 5 phases:
##   Phase 1: time.component.from parameter (notyettreated / nevertreated)
##   Phase 2: gsynth merged into ife via time.component.from
##   Phase 3: parametric bootstrap unlocked for ife/cfe
##   Phase 4: polynomial removed (superseded by cfe)
##   Phase 5: input validation and safety guards
##
## Tests are written BEFORE implementation. Each test will fail
## until its phase is implemented, then pass permanently.
## ---------------------------------------------------------

## ---- helpers ----
make_staggered_data <- function(N = 40, TT = 20, Ntr = 15, tau = 3.0,
                                 seed = 42) {
  set.seed(seed)
  alpha_i <- rnorm(N, 0, 1)
  xi_t    <- rnorm(TT, 0, 0.5)

  ## staggered adoption: treated units adopt between T0=8 and T0=14
  T0_vec <- rep(Inf, N)
  if (Ntr > 0) {
    T0_vec[1:Ntr] <- sample(8:14, Ntr, replace = TRUE)
  }

  Y_vec <- D_vec <- numeric(N * TT)
  id_vec <- time_vec <- integer(N * TT)
  idx <- 1
  for (i in 1:N) {
    for (t in 1:TT) {
      treated <- (t >= T0_vec[i])
      D_vec[idx]    <- as.integer(treated)
      Y_vec[idx]    <- alpha_i[i] + xi_t[t] + tau * D_vec[idx] + rnorm(1, 0, 0.5)
      id_vec[idx]   <- i
      time_vec[idx] <- t
      idx <- idx + 1
    }
  }

  data.frame(id = id_vec, time = time_vec, Y = Y_vec, D = D_vec)
}

## DGP with factor structure for comparisons needing r > 0
make_factor_data <- function(N = 100, TT = 30, Ntr = 30, tau = 3.0,
                              r = 2, seed = 42) {
  set.seed(seed)
  F_mat <- matrix(rnorm(TT * r), TT, r)
  L_mat <- matrix(rnorm(N * r), N, r)
  alpha_i <- rnorm(N, 0, 1)
  xi_t <- rnorm(TT, 0, 0.5)

  T0_vec <- rep(Inf, N)
  if (Ntr > 0) {
    T0_vec[1:Ntr] <- sample(round(TT * 0.4):round(TT * 0.7), Ntr,
                             replace = TRUE)
  }

  Y_vec <- D_vec <- numeric(N * TT)
  id_vec <- time_vec <- integer(N * TT)
  idx <- 1
  for (i in 1:N) {
    for (t in 1:TT) {
      treated <- (t >= T0_vec[i])
      D_vec[idx] <- as.integer(treated)
      Y_vec[idx] <- alpha_i[i] + xi_t[t] +
        sum(F_mat[t, ] * L_mat[i, ]) +
        tau * D_vec[idx] + rnorm(1, 0, 0.5)
      id_vec[idx] <- i
      time_vec[idx] <- t
      idx <- idx + 1
    }
  }

  data.frame(id = id_vec, time = time_vec, Y = Y_vec, D = D_vec)
}

## ========================================================
## PHASE 6: em parameter and gsynth reroute
## ========================================================

test_that("Phase 6a: gsynth and ife+nevertreated produce identical ATT", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  set.seed(100)
  out_gs <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "gsynth", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", parallel = FALSE
  )))

  set.seed(100)
  out_ife <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_equal(out_gs$att.avg, out_ife$att.avg,
               info = "gsynth and ife+nevertreated must be the same estimator")
})

test_that("Phase 6b: em=FALSE + time.component.from='notyettreated' errors", {

  skip_on_cran()
  df <- make_staggered_data(N = 40, Ntr = 15)

  expect_error(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r = 2, CV = FALSE, se = FALSE,
      time.component.from = "notyettreated", em = FALSE,
      parallel = FALSE
    ),
    regexp = "em.*FALSE|notyettreated.*em"
  )
})

test_that("Phase 6c: method='gsynth' auto-sets em=FALSE in output", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "gsynth", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", parallel = FALSE
  )))

  expect_false(out$em)
})

test_that("Phase 6d: method='ife' defaults em=TRUE in output", {

  skip_on_cran()
  df <- make_staggered_data(N = 40, Ntr = 15)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 0, CV = FALSE, se = FALSE,
    parallel = FALSE
  )))

  expect_true(out$em)
})

test_that("Phase 6e: ife+nevertreated and ife+notyettreated both close to true tau", {

  skip_on_cran()
  df <- make_factor_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 2, seed = 42)

  set.seed(100)
  out_nyt <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "notyettreated",
    parallel = FALSE
  )))

  set.seed(100)
  out_nt <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  ## Both should be within 1.0 of true tau=3.0 (loose bound)
  expect_true(abs(out_nyt$att.avg - 3.0) < 1.0,
              info = paste("notyettreated ATT:", out_nyt$att.avg))
  expect_true(abs(out_nt$att.avg - 3.0) < 1.0,
              info = paste("nevertreated ATT:", out_nt$att.avg))

  ## They need not be equal — different predictive routines
  cat(sprintf("\n  [info] notyettreated ATT=%.4f  nevertreated ATT=%.4f  diff=%.4f\n",
              out_nyt$att.avg, out_nt$att.avg, out_nyt$att.avg - out_nt$att.avg))
})

## ========================================================
## PHASE 1: time.component.from parameter
## ========================================================

test_that("Phase 1a: fect accepts time.component.from='notyettreated' (default behavior)", {

  skip_on_cran()
  df <- make_staggered_data()

  out <- suppressWarnings(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 2, CV = FALSE, se = FALSE,
    time.component.from = "notyettreated",
    parallel = FALSE
  ))

  expect_s3_class(out, "fect")
  expect_true(is.numeric(out$att.avg))
  expect_true(!is.na(out$att.avg))
})

## Phase 1b-1g pruned (2026-05-03 simplification): each was a single-line
## smoke check on the time.component.from API surface; the substantive
## coverage lives in Phase 6a-6e (gsynth <-> ife+nevertreated equivalence)
## and the Phase 3a-* series (cfe+nevertreated thorough coverage). Phase 1a
## above retained as the single smoke-entry-point for the API.
##
##   1b: nevertreated acceptance        -> covered by Phase 6a-6e
##   1c: nyt vs nt produce different    -> non-regression-grade observation
##   1d: default = notyettreated        -> argument-matching is R semantics
##   1e: cfe + nevertreated works       -> covered by Phase 3a-B/C/D
##   1f: time.component.from in CV      -> covered by Phase 3a-G1/G3
##   1g: time.component.from in boot    -> covered by Phase 3a-F1/F2/I1

## ========================================================
## PHASE 2: gsynth merged into ife
## ========================================================

test_that("Phase 2a: gsynth produces same results via ife+time.component.from='nevertreated'", {

  skip_on_cran()
  suppressWarnings(try(data("simgsynth", package = "fect"), silent = TRUE))
  skip_if_not(exists("simgsynth"), "Dataset 'simgsynth' not available")

  set.seed(200)
  out_gs <- suppressWarnings(fect::fect(
    Y ~ D, data = simgsynth, index = c("id", "time"),
    method = "gsynth", r = 2, CV = FALSE, se = FALSE,
    force = "two-way",
    parallel = FALSE
  ))

  set.seed(200)
  out_ife <- suppressWarnings(fect::fect(
    Y ~ D, data = simgsynth, index = c("id", "time"),
    method = "ife", r = 2, CV = FALSE, se = FALSE,
    force = "two-way",
    time.component.from = "nevertreated",
    parallel = FALSE
  ))

  ## ATT estimates should be numerically very close
  expect_equal(out_gs$att.avg, out_ife$att.avg, tolerance = 0.05)
})

test_that("Phase 2b: gsynth still works (backward compatibility)", {

  skip_on_cran()
  suppressWarnings(try(data("simgsynth", package = "fect"), silent = TRUE))
  skip_if_not(exists("simgsynth"), "Dataset 'simgsynth' not available")

  set.seed(201)
  out <- suppressWarnings(fect::fect(
    Y ~ D, data = simgsynth, index = c("id", "time"),
    method = "gsynth", r = 2, CV = FALSE, se = FALSE,
    force = "two-way",
    parallel = FALSE
  ))

  expect_s3_class(out, "fect")
  expect_true(is.numeric(out$att.avg))
})

test_that("Phase 2c: gsynth with parametric bootstrap still works", {

  skip_on_cran()
  suppressWarnings(try(data("simgsynth", package = "fect"), silent = TRUE))
  skip_if_not(exists("simgsynth"), "Dataset 'simgsynth' not available")

  set.seed(202)
  out <- suppressWarnings(fect::fect(
    Y ~ D, data = simgsynth, index = c("id", "time"),
    method = "gsynth", r = 2, CV = FALSE,
    se = TRUE, vartype = "parametric", nboots = 30,
    force = "two-way", min.T0 = 2,
    parallel = FALSE
  ))

  expect_s3_class(out, "fect")
  expect_true(!is.null(out$est.att))
})

## ========================================================
## PHASE 3: parametric bootstrap unlocked for ife/cfe
## ========================================================

test_that("Phase 3a: vartype='parametric' + ife + default factors.from errors (notyettreated gate)", {

  skip_on_cran()
  df <- make_staggered_data(N = 40, Ntr = 15)

  expect_error(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r = 2, CV = FALSE,
      se = TRUE, vartype = "parametric", nboots = 30,
      parallel = FALSE
    ),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

test_that("Phase 3b: vartype='parametric' + fe (->ife) + default factors.from errors (notyettreated gate)", {

  skip_on_cran()
  df <- make_staggered_data(N = 40, Ntr = 15)

  expect_error(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "fe", CV = FALSE,
      se = TRUE, vartype = "parametric", nboots = 30,
      parallel = FALSE
    ),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

test_that("Phase 3c: vartype='parametric' + cfe + default factors.from errors (notyettreated gate)", {

  skip_on_cran()
  df <- make_staggered_data(N = 40, Ntr = 15)

  expect_error(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "cfe", r = 0, CV = FALSE,
      se = TRUE, vartype = "parametric", nboots = 30,
      parallel = FALSE
    ),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

test_that("Phase 3d: vartype='parametric' still errors for method='mc'", {

  skip_on_cran()
  df <- make_staggered_data(N = 40, Ntr = 15)

  expect_error(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "mc", CV = FALSE,
      se = TRUE, vartype = "parametric", nboots = 30,
      parallel = FALSE
    ),
    regexp = "parametric"
  )
})

## ========================================================
## PHASE 4: polynomial removed (superseded by cfe)
## ========================================================
## Phase 4a test removed: method="polynomial" no longer exists.

## ========================================================
## PHASE 5: input validation and safety guards
## ========================================================

test_that("Phase 5a: time.component.from rejects invalid values", {

  skip_on_cran()
  df <- make_staggered_data()

  expect_error(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r = 2, CV = FALSE, se = FALSE,
      time.component.from = "invalid_value",
      parallel = FALSE
    ),
    regexp = "time.component.from"
  )
})

test_that("Phase 5b: time.component.from='nevertreated' + method='mc' errors", {

  skip_on_cran()
  df <- make_staggered_data(N = 40, Ntr = 15)

  expect_error(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "mc", CV = FALSE, se = FALSE,
      time.component.from = "nevertreated",
      parallel = FALSE
    ),
    regexp = "time.component.from|nevertreated|mc"
  )
})

test_that("Phase 5c: time.component.from='nevertreated' errors when no never-treated units", {

  skip_on_cran()
  ## All units are treated
  df <- make_staggered_data(N = 20, Ntr = 20)

  expect_error(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r = 2, CV = FALSE, se = FALSE,
      time.component.from = "nevertreated",
      parallel = FALSE
    ),
    regexp = "never.treated|nevertreated|control"
  )
})

test_that("Phase 5d: time.component.from='nevertreated' errors when too few never-treated units", {

  skip_on_cran()
  ## Only 1 never-treated unit (insufficient for r=2)
  df <- make_staggered_data(N = 20, Ntr = 19)

  expect_error(
    suppressWarnings(fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r = 2, CV = FALSE, se = FALSE,
      time.component.from = "nevertreated",
      parallel = FALSE
    )),
    regexp = "never.treated|nevertreated|insuffic|too few"
  )
})

test_that("Phase 5e: output records time.component.from in return object", {

  skip_on_cran()
  df <- make_staggered_data(N = 40, Ntr = 15)

  out <- suppressWarnings(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 2, CV = FALSE, se = FALSE,
    time.component.from = "nevertreated",
    parallel = FALSE
  ))

  expect_true("time.component.from" %in% names(out))
  expect_equal(out$time.component.from, "nevertreated")
})

test_that("Phase 5f: time.component.from='nevertreated' + method='both' errors", {

  skip_on_cran()
  df <- make_staggered_data(N = 40, Ntr = 15)

  expect_error(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "both", CV = FALSE, se = FALSE,
      time.component.from = "nevertreated",
      parallel = FALSE
    ),
    regexp = "both"
  )
})

## ========================================================
## PHASE 3a: CFE bifurcation in fect_nevertreated
## ========================================================

## ---- Phase 3a DGP helpers ----

make_cfe_z_data <- function(N = 100, TT = 30, Ntr = 30, tau = 3.0,
                             r = 2, seed = 42) {
  set.seed(seed)
  F_mat <- matrix(rnorm(TT * r), TT, r)
  L_mat <- matrix(rnorm(N * r), N, r)
  alpha_i <- rnorm(N, 0, 1)
  xi_t <- rnorm(TT, 0, 0.5)

  ## Z: time-invariant covariate (baseline characteristic)
  Z_i <- rnorm(N, 0, 1)

  ## gamma: grouped time coefficient (one per time period)
  gamma_t <- rnorm(TT, 0, 0.5)

  ## Staggered treatment adoption
  T0_vec <- rep(Inf, N)
  if (Ntr > 0) {
    T0_vec[1:Ntr] <- sample(round(TT * 0.4):round(TT * 0.7), Ntr,
                             replace = TRUE)
  }

  Y_vec <- D_vec <- Z_vec <- numeric(N * TT)
  id_vec <- time_vec <- integer(N * TT)
  idx <- 1
  for (i in 1:N) {
    for (t in 1:TT) {
      treated <- (t >= T0_vec[i])
      D_vec[idx] <- as.integer(treated)
      factor_component <- if (r > 0) sum(F_mat[t, ] * L_mat[i, ]) else 0
      Y_vec[idx] <- alpha_i[i] + xi_t[t] + Z_i[i] * gamma_t[t] +
        factor_component + tau * D_vec[idx] + rnorm(1, 0, 0.5)
      Z_vec[idx] <- Z_i[i]
      id_vec[idx] <- i
      time_vec[idx] <- t
      idx <- idx + 1
    }
  }

  data.frame(id = id_vec, time = time_vec, Y = Y_vec, D = D_vec, Z = Z_vec)
}

make_cfe_q_data <- function(N = 100, TT = 30, Ntr = 30, tau = 3.0,
                             r = 2, seed = 42) {
  set.seed(seed)
  F_mat <- matrix(rnorm(TT * r), TT, r)
  L_mat <- matrix(rnorm(N * r), N, r)
  alpha_i <- rnorm(N, 0, 1)
  xi_t <- rnorm(TT, 0, 0.5)

  ## Q: time-varying basis (normalized time trend)
  Q_t <- (1:TT) / TT

  ## kappa: unit-specific coefficient on Q
  kappa_i <- rnorm(N, 0, 0.5)

  ## Staggered treatment adoption
  T0_vec <- rep(Inf, N)
  if (Ntr > 0) {
    T0_vec[1:Ntr] <- sample(round(TT * 0.4):round(TT * 0.7), Ntr,
                             replace = TRUE)
  }

  Y_vec <- D_vec <- Q_vec <- numeric(N * TT)
  id_vec <- time_vec <- integer(N * TT)
  idx <- 1
  for (i in 1:N) {
    for (t in 1:TT) {
      treated <- (t >= T0_vec[i])
      D_vec[idx] <- as.integer(treated)
      factor_component <- if (r > 0) sum(F_mat[t, ] * L_mat[i, ]) else 0
      Y_vec[idx] <- alpha_i[i] + xi_t[t] + Q_t[t] * kappa_i[i] +
        factor_component + tau * D_vec[idx] + rnorm(1, 0, 0.5)
      Q_vec[idx] <- Q_t[t]
      id_vec[idx] <- i
      time_vec[idx] <- t
      idx <- idx + 1
    }
  }

  data.frame(id = id_vec, time = time_vec, Y = Y_vec, D = D_vec, Q = Q_vec)
}

make_cfe_fe_data <- function(N = 100, TT = 30, Ntr = 30, tau = 3.0,
                              r = 2, seed = 42) {
  set.seed(seed)
  F_mat <- matrix(rnorm(TT * r), TT, r)
  L_mat <- matrix(rnorm(N * r), N, r)
  alpha_i <- rnorm(N, 0, 1)
  xi_t <- rnorm(TT, 0, 0.5)

  ## Industry: 4 industries, randomly assigned to ALL units (shared / Type B)
  industry_i <- sample(1:4, N, replace = TRUE)
  industry_fe <- c(-1.0, 0.5, 0.0, 1.5)

  ## Staggered treatment adoption
  T0_vec <- rep(Inf, N)
  if (Ntr > 0) {
    T0_vec[1:Ntr] <- sample(round(TT * 0.4):round(TT * 0.7), Ntr,
                             replace = TRUE)
  }

  Y_vec <- D_vec <- numeric(N * TT)
  id_vec <- time_vec <- industry_vec <- integer(N * TT)
  idx <- 1
  for (i in 1:N) {
    for (t in 1:TT) {
      treated <- (t >= T0_vec[i])
      D_vec[idx] <- as.integer(treated)
      factor_component <- if (r > 0) sum(F_mat[t, ] * L_mat[i, ]) else 0
      Y_vec[idx] <- alpha_i[i] + xi_t[t] + industry_fe[industry_i[i]] +
        factor_component + tau * D_vec[idx] + rnorm(1, 0, 0.5)
      industry_vec[idx] <- industry_i[i]
      id_vec[idx] <- i
      time_vec[idx] <- t
      idx <- idx + 1
    }
  }

  data.frame(id = id_vec, time = time_vec, Y = Y_vec, D = D_vec,
             industry = industry_vec)
}

make_cfe_fe_nesting_data <- function(N = 100, TT = 30, Ntr = 30, tau = 3.0,
                                      r = 2, seed = 42) {
  set.seed(seed)
  F_mat <- matrix(rnorm(TT * r), TT, r)
  L_mat <- matrix(rnorm(N * r), N, r)
  alpha_i <- rnorm(N, 0, 1)
  xi_t <- rnorm(TT, 0, 0.5)

  ## Industry: treated get 1-2, control get 3-4 (unit-nesting / Type A)
  industry_i <- integer(N)
  industry_i[1:Ntr] <- sample(1:2, Ntr, replace = TRUE)
  industry_i[(Ntr + 1):N] <- sample(3:4, N - Ntr, replace = TRUE)
  industry_fe <- c(-1.0, 0.5, -0.5, 1.0)

  ## Staggered treatment adoption
  T0_vec <- rep(Inf, N)
  if (Ntr > 0) {
    T0_vec[1:Ntr] <- sample(round(TT * 0.4):round(TT * 0.7), Ntr,
                             replace = TRUE)
  }

  Y_vec <- D_vec <- numeric(N * TT)
  id_vec <- time_vec <- industry_vec <- integer(N * TT)
  idx <- 1
  for (i in 1:N) {
    for (t in 1:TT) {
      treated <- (t >= T0_vec[i])
      D_vec[idx] <- as.integer(treated)
      factor_component <- if (r > 0) sum(F_mat[t, ] * L_mat[i, ]) else 0
      Y_vec[idx] <- alpha_i[i] + xi_t[t] + industry_fe[industry_i[i]] +
        factor_component + tau * D_vec[idx] + rnorm(1, 0, 0.5)
      industry_vec[idx] <- industry_i[i]
      id_vec[idx] <- i
      time_vec[idx] <- t
      idx <- idx + 1
    }
  }

  data.frame(id = id_vec, time = time_vec, Y = Y_vec, D = D_vec,
             industry = industry_vec)
}

make_cfe_full_data <- function(N = 100, TT = 30, Ntr = 30, tau = 3.0,
                                r = 2, seed = 42) {
  set.seed(seed)
  F_mat <- matrix(rnorm(TT * r), TT, r)
  L_mat <- matrix(rnorm(N * r), N, r)
  alpha_i <- rnorm(N, 0, 1)
  xi_t <- rnorm(TT, 0, 0.5)

  ## Z: time-invariant covariate
  Z_i <- rnorm(N, 0, 1)
  gamma_t <- rnorm(TT, 0, 0.5)

  ## Q: time-varying basis
  Q_t <- (1:TT) / TT
  kappa_i <- rnorm(N, 0, 0.5)

  ## Industry: shared (Type B) — all units draw from same pool
  industry_i <- sample(1:4, N, replace = TRUE)
  industry_fe <- c(-1.0, 0.5, 0.0, 1.5)

  ## Staggered treatment adoption
  T0_vec <- rep(Inf, N)
  if (Ntr > 0) {
    T0_vec[1:Ntr] <- sample(round(TT * 0.4):round(TT * 0.7), Ntr,
                             replace = TRUE)
  }

  Y_vec <- D_vec <- Z_vec <- Q_vec <- numeric(N * TT)
  id_vec <- time_vec <- industry_vec <- integer(N * TT)
  idx <- 1
  for (i in 1:N) {
    for (t in 1:TT) {
      treated <- (t >= T0_vec[i])
      D_vec[idx] <- as.integer(treated)
      factor_component <- if (r > 0) sum(F_mat[t, ] * L_mat[i, ]) else 0
      Y_vec[idx] <- alpha_i[i] + xi_t[t] + Z_i[i] * gamma_t[t] +
        Q_t[t] * kappa_i[i] + industry_fe[industry_i[i]] +
        factor_component + tau * D_vec[idx] + rnorm(1, 0, 0.5)
      Z_vec[idx] <- Z_i[i]
      Q_vec[idx] <- Q_t[t]
      industry_vec[idx] <- industry_i[i]
      id_vec[idx] <- i
      time_vec[idx] <- t
      idx <- idx + 1
    }
  }

  data.frame(id = id_vec, time = time_vec, Y = Y_vec, D = D_vec,
             Z = Z_vec, Q = Q_vec, industry = industry_vec)
}

## ---- Category B: Specification Equivalence ----

test_that("Phase 3a-B1: ife+nevertreated equals cfe+nevertreated (no extras)", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  set.seed(100)
  out_ife <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  set.seed(100)
  out_cfe <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out_ife$att.avg - out_cfe$att.avg) < 0.1,
              info = paste("ife ATT:", out_ife$att.avg, "cfe ATT:", out_cfe$att.avg))
})

test_that("Phase 3a-B2: fe+nevertreated equals cfe+nevertreated at r=0", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 0, seed = 42)

  out_ife <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 0, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  out_cfe <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", r = 0, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out_ife$att.avg - out_cfe$att.avg) < 0.1,
              info = paste("ife ATT:", out_ife$att.avg, "cfe ATT:", out_cfe$att.avg))
})

test_that("Phase 3a-B3: cfe+nevertreated is deterministic (same seed, same result)", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  set.seed(100)
  out1 <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  set.seed(100)
  out2 <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_identical(out1$att.avg, out2$att.avg)
})

test_that("Phase 3a-B4: gsynth equals cfe+nevertreated (no extras)", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  set.seed(100)
  out_gs <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "gsynth", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", parallel = FALSE
  )))

  set.seed(100)
  out_cfe <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out_gs$att.avg - out_cfe$att.avg) < 0.1,
              info = paste("gsynth ATT:", out_gs$att.avg, "cfe ATT:", out_cfe$att.avg))
})

## ---- Category C: Accuracy ----

test_that("Phase 3a-C1: accuracy with Z (time-invariant covariates)", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out$att.avg - 3.0) < 0.5,
              info = paste("ATT:", out$att.avg, "true tau: 3.0"))
})

test_that("Phase 3a-C2: accuracy with Q (unit-specific time trends)", {

  skip_on_cran()
  df <- make_cfe_q_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Q = "Q", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out$att.avg - 3.0) < 0.5,
              info = paste("ATT:", out$att.avg, "true tau: 3.0"))
})

test_that("Phase 3a-C3: accuracy with shared extra FE (Type B)", {

  skip_on_cran()
  df <- make_cfe_fe_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time", "industry"),
    method = "cfe", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out$att.avg - 3.0) < 0.5,
              info = paste("ATT:", out$att.avg, "true tau: 3.0"))
})

test_that("Phase 3a-C4: accuracy with unit-nesting extra FE (Type A)", {

  skip_on_cran()
  df <- make_cfe_fe_nesting_data(N = 200, TT = 30, Ntr = 60, tau = 3.0,
                                  r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time", "industry"),
    method = "cfe", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out$att.avg - 3.0) < 0.5,
              info = paste("ATT:", out$att.avg, "true tau: 3.0"))
})

test_that("Phase 3a-C5: accuracy with all CFE components", {

  skip_on_cran()
  df <- make_cfe_full_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time", "industry"),
    method = "cfe", Z = "Z", Q = "Q", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out$att.avg - 3.0) < 0.5,
              info = paste("ATT:", out$att.avg, "true tau: 3.0"))
})

test_that("Phase 3a-C6: notyettreated vs nevertreated both accurate", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 2, seed = 42)

  out_nyt <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "notyettreated",
    parallel = FALSE
  )))

  out_nt <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out_nyt$att.avg - 3.0) < 1.0,
              info = paste("notyettreated ATT:", out_nyt$att.avg))
  expect_true(abs(out_nt$att.avg - 3.0) < 1.0,
              info = paste("nevertreated ATT:", out_nt$att.avg))
})

test_that("Phase 3a-C7: accuracy with Z at r=0 (no factors)", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 0, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 0, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out$att.avg - 3.0) < 0.5,
              info = paste("ATT:", out$att.avg, "true tau: 3.0"))
})

## ---- Category D: Validation Guards ----

test_that("Phase 3a-D1: Type-B FE with missing level in co errors", {

  skip_on_cran()
  df <- make_cfe_fe_data(N = 100, TT = 30, Ntr = 30, seed = 42)
  ## Assign a unique industry level to treated unit 1 (not in controls)
  df$industry[df$id == 1] <- 99

  expect_error(
    suppressWarnings(suppressMessages(fect::fect(
      Y ~ D, data = df, index = c("id", "time", "industry"),
      method = "cfe", r = 2, CV = FALSE, se = FALSE,
      force = "two-way", time.component.from = "nevertreated",
      parallel = FALSE
    )))
  )
})

test_that("Phase 3a-D2: no never-treated units errors", {

  skip_on_cran()
  df <- make_staggered_data(N = 40, TT = 20, Ntr = 40)

  expect_error(
    suppressWarnings(suppressMessages(fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "cfe", r = 0, CV = FALSE, se = FALSE,
      force = "two-way", time.component.from = "nevertreated",
      parallel = FALSE
    ))),
    regexp = "never.treated|nevertreated|control"
  )
})

test_that("Phase 3a-D3: too few control units for r errors", {

  skip_on_cran()
  df <- make_factor_data(N = 35, TT = 30, Ntr = 33, tau = 3.0, r = 2, seed = 42)

  expect_error(
    suppressWarnings(suppressMessages(fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "cfe", r = 2, CV = FALSE, se = FALSE,
      force = "two-way", time.component.from = "nevertreated",
      parallel = FALSE
    ))),
    regexp = "never.treated|nevertreated|insuffic|too few"
  )
})

test_that("Phase 3a-D4: valid cfe+nevertreated combination is accepted", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  out <- expect_no_error(suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  ))))

  expect_false(is.null(out$att.avg))
})

test_that("Phase 3a-D5: mc+nevertreated is rejected", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, seed = 42)

  expect_error(
    suppressWarnings(suppressMessages(fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "mc", r = 2, CV = FALSE, se = FALSE,
      force = "two-way", time.component.from = "nevertreated",
      parallel = FALSE
    ))),
    regexp = "mc|nevertreated|time.component.from"
  )
})

## ---- Category A: Solver Equivalence ----

test_that("Phase 3a-A1: complex_fe_ub vs inter_fe_ub equivalence", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  set.seed(100)
  out_cfe <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  set.seed(100)
  out_ife <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out_cfe$att.avg - out_ife$att.avg) < 0.1,
              info = paste("cfe ATT:", out_cfe$att.avg, "ife ATT:", out_ife$att.avg))
})

test_that("Phase 3a-A2: solver equivalence at r=0 (FE only)", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 0, seed = 42)

  out_cfe <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", r = 0, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  out_ife <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 0, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out_cfe$att.avg - out_ife$att.avg) < 0.01,
              info = paste("cfe ATT:", out_cfe$att.avg, "ife ATT:", out_ife$att.avg))
})

test_that("Phase 3a-A3: solver equivalence with unbalanced data", {

  skip_on_cran()
  df <- make_staggered_data(N = 40, TT = 20, Ntr = 15, tau = 3.0, seed = 42)

  out_cfe <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", r = 0, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  out_ife <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 0, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(abs(out_cfe$att.avg - out_ife$att.avg) < 0.1,
              info = paste("cfe ATT:", out_cfe$att.avg, "ife ATT:", out_ife$att.avg))
})

## ---- Category F: Bootstrap / Inference ----

test_that("Phase 3a-F2: parametric bootstrap SE with cfe+nevertreated", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE,
    se = TRUE, vartype = "parametric", nboots = 30,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_false(is.null(out$est.att))
  expect_false(all(is.na(out$est.att[, "S.E."])))
})

test_that("Phase 3a-F1: jackknife SE with cfe+nevertreated", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE,
    se = TRUE, vartype = "jackknife", nboots = 30,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_false(is.null(out$est.att))
  expect_false(all(is.na(out$est.att[, "S.E."])))
})

test_that("Phase 3a-F3: parametric bootstrap with cfe+notyettreated errors (notyettreated gate)", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  expect_error(
    fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "cfe", Z = "Z", r = 2, CV = FALSE,
      se = TRUE, vartype = "parametric", nboots = 30,
      force = "two-way", time.component.from = "notyettreated",
      parallel = FALSE
    ),
    regexp = "parametric",
    ignore.case = TRUE
  )
})

test_that("Phase 3a-F4: ife+nevertreated SE unchanged (regression)", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 2, CV = FALSE,
    se = TRUE, nboots = 30,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_false(is.null(out$est.att))
  expect_false(all(is.na(out$est.att[, "S.E."])))
})

## ---- Category E: Output Completeness ----

## Phase 3a-E1 pruned (2026-05-03): subsumed by E3 (which already asserts
## core output fields non-NULL with dimensions, including gamma when CFE+Z).

test_that("Phase 3a-E2: time.component.from field in output", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_equal(out$time.component.from, "nevertreated")
})

test_that("Phase 3a-E3: core output fields non-NULL with correct dimensions", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_false(is.null(out$Y.ct))
  expect_false(is.null(out$eff))
  expect_false(is.null(out$att.avg))
  expect_true(is.numeric(out$att.avg))
  expect_false(is.na(out$att.avg))
  expect_equal(nrow(out$Y.ct), length(unique(df$time)))
  expect_equal(ncol(out$Y.ct), length(unique(df$id)))
})

## Phase 3a-E4 (plot smoke) and E5 (print smoke) pruned (2026-05-03):
## covered by dedicated test-plot-fect.R + test-plot-refactor.R for plot,
## and the print method is exercised by every regression test that prints
## a summary of fect() output.

## ---- Category G: Cross-Validation ----

test_that("Phase 3a-G1: CV selects a reasonable r with factors in data", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = c(0, 5), CV = TRUE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(out$r.cv >= 0)
  expect_true(is.numeric(out$r.cv))
})

test_that("Phase 3a-G2: CV selects r=0 on no-factor data", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 0, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = c(0, 3), CV = TRUE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_equal(out$r.cv, 0)
})

test_that("Phase 3a-G3: ife+nevertreated CV unchanged (regression)", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = c(0, 5), CV = TRUE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_true(out$r.cv >= 0)
})

## ---- Category H: Edge Cases ----

test_that("Phase 3a-H1: single treated unit", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 100, TT = 30, Ntr = 1, tau = 3.0, r = 2, seed = 42)

  expect_no_error(suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  ))))
})

test_that("Phase 3a-H2: single control unit (r=0)", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 31, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  ## With only 1 never-treated unit at r=0 this may either run or error clearly
  result <- tryCatch(
    suppressWarnings(suppressMessages(fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "cfe", r = 0, CV = FALSE, se = FALSE,
      force = "two-way", time.component.from = "nevertreated",
      parallel = FALSE
    ))),
    error = function(e) e
  )

  ## Either it runs successfully or produces a clear error
  if (inherits(result, "error")) {
    expect_true(grepl("never.treated|nevertreated|insuffic|too few|control",
                      result$message, ignore.case = TRUE),
                info = paste("Unexpected error:", result$message))
  } else {
    expect_false(is.null(result$att.avg))
  }
})

test_that("Phase 3a-H3: cfe+nevertreated with r=0 (no factors)", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 0, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 0, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  )))

  expect_no_error(out)
  expect_true(abs(out$att.avg - 3.0) < 1.0,
              info = paste("ATT:", out$att.avg, "true tau: 3.0"))
})

test_that("Phase 3a-H4: no covariates at all", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  expect_no_error(suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  ))))
})

test_that("Phase 3a-H5: treatment reversals", {

  skip_on_cran()
  df <- make_cfe_full_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)
  ## Create reversals: unit 1 reverts to control after period 20
  reversal_rows <- df$id == 1 & df$time > 20
  df$D[reversal_rows] <- 0

  expect_no_error(suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time", "industry"),
    method = "cfe", Z = "Z", Q = "Q", r = 2, CV = FALSE, se = FALSE,
    force = "two-way", time.component.from = "nevertreated",
    parallel = FALSE
  ))))
})

## ========================================================
## PHASE 3a-I: Parametric bootstrap under nevertreated
## with em=TRUE / em=FALSE for both cfe and ife
## Uses parallel computing with controlled seeds (doRNG)
## ========================================================

test_that("Phase 3a-I1: ife+nevertreated parametric bootstrap, em=TRUE, parallel", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 2, CV = FALSE, force = "two-way",
    time.component.from = "nevertreated", em = TRUE,
    se = TRUE, vartype = "bootstrap", nboots = 20,
    parallel = TRUE, cores = 2, seed = 123
  )))

  expect_false(is.na(out$att.avg),
               info = "att.avg should not be NA")
  expect_false(is.null(out$est.att),
               info = "est.att should not be NULL")
  expect_true(any(!is.na(out$est.att[, "S.E."])),
              info = "SE estimates should not all be NA")
})

## Phase 3a-I2 pruned (2026-05-03): em=FALSE variant of I1; em=TRUE/FALSE
## equivalence on nevertreated is asserted by I10. Smoke check for em=FALSE
## bootstrap path is implicit through I10's identity assertion.

test_that("Phase 3a-I3: cfe+nevertreated parametric bootstrap, em=TRUE, parallel", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, force = "two-way",
    time.component.from = "nevertreated", em = TRUE,
    se = TRUE, vartype = "bootstrap", nboots = 20,
    parallel = TRUE, cores = 2, seed = 123
  )))

  expect_false(is.na(out$att.avg),
               info = "att.avg should not be NA")
  expect_false(is.null(out$est.att),
               info = "est.att should not be NULL")
  expect_true(any(!is.na(out$est.att[, "S.E."])),
              info = "SE estimates should not all be NA")
})

## Phase 3a-I4 pruned (2026-05-03): cfe em=FALSE variant of I3; same
## reasoning as I2's prune (em=TRUE/FALSE equivalence on nevertreated
## covered by I10).

test_that("Phase 3a-I5: bootstrap reproducibility with same seed (ife+nevertreated)", {

  skip_on_cran()
  df <- make_factor_data(N = 80, TT = 20, Ntr = 25, tau = 3.0, r = 2, seed = 42)

  run_boot <- function(s) {
    suppressWarnings(suppressMessages(fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r = 2, CV = FALSE, force = "two-way",
      time.component.from = "nevertreated", em = TRUE,
      se = TRUE, vartype = "bootstrap", nboots = 10,
      parallel = TRUE, cores = 2, seed = s
    )))
  }

  out1 <- run_boot(999)
  out2 <- run_boot(999)

  expect_equal(out1$att.avg, out2$att.avg,
               info = "Same seed should give identical att.avg")
  expect_equal(out1$est.att[, "S.E."], out2$est.att[, "S.E."],
               info = "Same seed should give identical SE")
})

test_that("Phase 3a-I6: bootstrap reproducibility with same seed (cfe+nevertreated)", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 80, TT = 20, Ntr = 25, tau = 3.0, r = 2, seed = 42)

  run_boot <- function(s) {
    suppressWarnings(suppressMessages(fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "cfe", Z = "Z", r = 2, CV = FALSE, force = "two-way",
      time.component.from = "nevertreated", em = TRUE,
      se = TRUE, vartype = "bootstrap", nboots = 10,
      parallel = TRUE, cores = 2, seed = s
    )))
  }

  out1 <- run_boot(999)
  out2 <- run_boot(999)

  expect_equal(out1$att.avg, out2$att.avg,
               info = "Same seed should give identical att.avg")
  expect_equal(out1$est.att[, "S.E."], out2$est.att[, "S.E."],
               info = "Same seed should give identical SE")
})

test_that("Phase 3a-I7: different seeds give different bootstrap SE (ife+nevertreated)", {

  skip_on_cran()
  df <- make_factor_data(N = 80, TT = 20, Ntr = 25, tau = 3.0, r = 2, seed = 42)

  run_boot <- function(s) {
    suppressWarnings(suppressMessages(fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r = 2, CV = FALSE, force = "two-way",
      time.component.from = "nevertreated", em = TRUE,
      se = TRUE, vartype = "bootstrap", nboots = 10,
      parallel = TRUE, cores = 2, seed = s
    )))
  }

  out1 <- run_boot(111)
  out2 <- run_boot(222)

  ## Point estimate should be the same (not seed-dependent)
  expect_equal(out1$att.avg, out2$att.avg, tolerance = 1e-10,
               info = "Point estimate should not depend on bootstrap seed")
  ## SE should differ (different bootstrap draws)
  expect_false(identical(out1$est.att[, "S.E."], out2$est.att[, "S.E."]),
               info = "Different seeds should give different SE draws")
})

test_that("Phase 3a-I8: ATT accuracy check under bootstrap (cfe+nevertreated, em=TRUE)", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 150, TT = 30, Ntr = 50, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, force = "two-way",
    time.component.from = "nevertreated", em = TRUE,
    se = TRUE, vartype = "bootstrap", nboots = 20,
    parallel = TRUE, cores = 2, seed = 123
  )))

  expect_true(abs(out$att.avg - 3.0) < 1.0,
              info = paste("ATT:", out$att.avg, "should be near 3.0"))
  ## CI should cover the true value
  ci_lower <- out$est.avg[, "CI.lower"]
  ci_upper <- out$est.avg[, "CI.upper"]
  expect_true(ci_lower < 3.0 & ci_upper > 3.0,
              info = paste("95% CI [", ci_lower, ",", ci_upper,
                           "] should cover true tau=3.0"))
})

test_that("Phase 3a-I9: quantile.CI=TRUE bias-corrected reflection CI (cfe+nevertreated)", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, force = "two-way",
    time.component.from = "nevertreated",
    se = TRUE, vartype = "bootstrap", nboots = 20,
    quantile.CI = TRUE,
    parallel = TRUE, cores = 2, seed = 123
  )))

  expect_false(is.null(out$est.att),
               info = "est.att should not be NULL")
  expect_true("CI.lower" %in% colnames(out$est.att),
              info = "est.att should have CI.lower column")
  expect_true("CI.upper" %in% colnames(out$est.att),
              info = "est.att should have CI.upper column")
  expect_true(any(!is.na(out$est.att[, "CI.lower"])),
              info = "At least some CI.lower values should be non-NA")
  expect_false(is.null(out$est.avg),
               info = "est.avg should not be NULL")
  expect_true("CI.lower" %in% colnames(out$est.avg),
              info = "est.avg should have CI.lower column")
  expect_true("CI.upper" %in% colnames(out$est.avg),
              info = "est.avg should have CI.upper column")
  ci_lower <- out$est.avg[, "CI.lower"]
  ci_upper <- out$est.avg[, "CI.upper"]
  expect_true(ci_lower < 3.0 & ci_upper > 3.0,
              info = paste("Quantile CI [", ci_lower, ",", ci_upper,
                           "] should cover true tau=3.0"))
})

test_that("Phase 3a-I10: em=TRUE vs em=FALSE identical for nevertreated (ife)", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)

  run_em <- function(em_val) {
    suppressWarnings(suppressMessages(fect::fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r = 2, CV = FALSE, force = "two-way",
      time.component.from = "nevertreated", em = em_val,
      se = FALSE
    )))
  }

  out1 <- run_em(TRUE)
  out2 <- run_em(FALSE)

  expect_equal(out1$att.avg, out2$att.avg, tolerance = 1e-10,
               info = "em=TRUE and em=FALSE should give identical att.avg for nevertreated")
})

test_that("Phase 3a-I11: unbalanced data forces _ub/EM path in draw.error() bootstrap", {

  skip_on_cran()

  ## IFE on unbalanced data
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)
  set.seed(999)
  n_drop <- floor(nrow(df) * 0.05)
  drop_idx <- sample(seq_len(nrow(df)), n_drop, replace = FALSE)
  df_ub <- df[-drop_idx, ]

  out_ife <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df_ub, index = c("id", "time"),
    method = "ife", r = 2, CV = FALSE, force = "two-way",
    time.component.from = "nevertreated",
    se = TRUE, vartype = "bootstrap", nboots = 15,
    parallel = TRUE, cores = 2, seed = 123
  )))

  expect_false(is.na(out_ife$att.avg),
               info = "IFE unbalanced: att.avg should not be NA")
  expect_true(any(!is.na(out_ife$est.att[, "S.E."])),
              info = "IFE unbalanced: SE estimates should not all be NA")

  ## CFE on unbalanced data
  df2 <- make_cfe_z_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)
  set.seed(999)
  n_drop2 <- floor(nrow(df2) * 0.05)
  drop_idx2 <- sample(seq_len(nrow(df2)), n_drop2, replace = FALSE)
  df2_ub <- df2[-drop_idx2, ]

  out_cfe <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df2_ub, index = c("id", "time"),
    method = "cfe", Z = "Z", r = 2, CV = FALSE, force = "two-way",
    time.component.from = "nevertreated",
    se = TRUE, vartype = "bootstrap", nboots = 15,
    parallel = TRUE, cores = 2, seed = 123
  )))

  expect_false(is.na(out_cfe$att.avg),
               info = "CFE unbalanced: att.avg should not be NA")
  expect_true(any(!is.na(out_cfe$est.att[, "S.E."])),
              info = "CFE unbalanced: SE estimates should not all be NA")
})

test_that("Phase 3a-I12: r=0 invariance — time.component.from is a no-op when r=0", {

  skip_on_cran()
  df <- make_factor_data(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 0, seed = 42)

  out_nyt <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 0, CV = FALSE, force = "two-way",
    time.component.from = "notyettreated",
    se = FALSE
  )))

  out_nt <- suppressWarnings(suppressMessages(fect::fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "ife", r = 0, CV = FALSE, force = "two-way",
    time.component.from = "nevertreated",
    se = FALSE
  )))

  expect_equal(out_nyt$att.avg, out_nt$att.avg, tolerance = 1e-2,
               info = "time.component.from should be a no-op when r=0: att.avg must match")
})

## ========================================================
## CV routing for cfe+nevertreated (REQ-cv-gap-001)
## ========================================================

test_that("cfe+nevertreated CV selects r and runs without error", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = c(0, 3), CV = TRUE, se = FALSE,
    time.component.from = "nevertreated",
    force = "two-way", parallel = TRUE, cores = 10
  )))

  expect_s3_class(out, "fect")
  expect_true(is.numeric(out$r.cv), info = "r.cv should be set by CV")
  expect_true(out$r.cv >= 1 && out$r.cv <= 3,
              info = paste("r.cv should be >= 1 for DGP with r=2, got:", out$r.cv))
  expect_true(is.numeric(out$att.avg))
  expect_true(!is.na(out$att.avg))
})

## ========================================================
## CFE CV r-selection accuracy tests (REQ-cfe-cv-rselect-001)
## ========================================================

test_that("cfe+nevertreated+Z CV correctly selects r=2", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = c(0, 5), CV = TRUE, se = FALSE,
    time.component.from = "nevertreated",
    force = "two-way", parallel = TRUE, cores = 10
  )))

  expect_equal(out$r.cv, 2,
               info = paste("r.cv should be 2 for DGP with r=2, got:", out$r.cv))
  expect_true(abs(out$att.avg - 3.0) < 0.5,
              info = paste("ATT should be close to 3.0, got:", out$att.avg))
})

test_that("cfe+nevertreated+Z CV correctly selects r=0 on zero-factor data", {

  skip_on_cran()
  df <- make_cfe_z_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 0, seed = 42)

  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", Z = "Z", r = c(0, 5), CV = TRUE, se = FALSE,
    time.component.from = "nevertreated",
    force = "two-way", parallel = TRUE, cores = 10
  )))

  expect_equal(out$r.cv, 0,
               info = paste("r.cv should be 0 for DGP with r=0, got:", out$r.cv))
})

test_that("cfe+nevertreated CV selects r=2 on factor-only data (no Z in DGP)", {

  skip_on_cran()
  df <- make_factor_data(N = 200, TT = 30, Ntr = 60, tau = 3.0, r = 2, seed = 42)

  out <- suppressWarnings(suppressMessages(fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "cfe", r = c(0, 5), CV = TRUE, se = FALSE,
    time.component.from = "nevertreated",
    force = "two-way", parallel = TRUE, cores = 10
  )))

  expect_equal(out$r.cv, 2,
               info = paste("r.cv should be 2 for DGP with r=2, got:", out$r.cv))
})

## ---------------------------------------------------------------
## Tests for Phase 1: parallel CV in cv.R IFE notyettreated path
##
## Covers test-spec.md sections P, A, S, T, R, D, L, E, and G
## for REQ-parallel-cv-phase1.
##
## RNG discipline: set.seed() immediately before EVERY fect() call.
## cv.sample() consumes the global RNG for fold masks; any intervening
## RNG call between two fect() calls produces different fold splits.
##
## See: statsclaw-workspace/fect/runs/REQ-parallel-cv-phase1/test-spec.md
## ---------------------------------------------------------------

suppressWarnings(data("simdata", package = "fect"))

## =================================================================
## Section P: Numerical Identity (serial == parallel, IFE notyettreated)
## =================================================================

## -- P.1  IFE x notyettreated x all_units -------------------------

test_that("P.1: IFE notyettreated all_units serial == parallel", {

  skip_on_cran()

  set.seed(42)
  fit_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "all_units",
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  set.seed(42)
  fit_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "all_units",
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  ## r.cv: named integer — must be identical (including the "r" name attribute)
  expect_identical(fit_seq$r.cv, fit_par$r.cv)

  ## CV.out.ife scores: within 1e-10
  cv_diff <- max(abs(fit_seq$CV.out.ife - fit_par$CV.out.ife), na.rm = TRUE)
  expect_true(cv_diff < 1e-10,
    info = sprintf("CV.out.ife max diff = %.2e (tolerance 1e-10)", cv_diff))

  ## att.avg: same r.cv => same final model => identical
  expect_identical(fit_seq$att.avg, fit_par$att.avg)
})

## -- P.2  IFE x notyettreated x treated_units ---------------------

test_that("P.2: IFE notyettreated treated_units serial == parallel", {

  skip_on_cran()

  set.seed(42)
  fit_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "treated_units",
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  set.seed(42)
  fit_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "treated_units",
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  expect_identical(fit_seq$r.cv, fit_par$r.cv)

  cv_diff <- max(abs(fit_seq$CV.out.ife - fit_par$CV.out.ife), na.rm = TRUE)
  expect_true(cv_diff < 1e-10,
    info = sprintf("CV.out.ife max diff = %.2e (tolerance 1e-10)", cv_diff))

  expect_identical(fit_seq$att.avg, fit_par$att.avg)
})

## -- P.3  IFE x nevertreated x all_units (helper-migration fixture) --
## Note: simdata triggers the nevertreated early-exit path
## (insufficient pre-treatment records -> r.cv=0, CV.out=NULL).
## The G.1-G.4 tests in test-score-unify.R exercise the substantive
## path with synthetic data; this test pins the migration regression.

test_that("P.3: IFE nevertreated all_units — helper migration preserves output", {

  skip_on_cran()

  fixture <- readRDS(test_path("fixtures", "nevertreated_ife_all.rds"))

  set.seed(42)
  fit_new <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "all_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  expect_identical(fit_new$r.cv, fixture$r.cv)

  ## CV.out may be NULL when early-exit fires; handle that gracefully
  if (is.null(fixture$CV.out) && is.null(fit_new$CV.out)) {
    expect_true(TRUE)  ## both NULL: identical
  } else {
    cv_diff <- max(abs(fit_new$CV.out - fixture$CV.out), na.rm = TRUE)
    expect_true(cv_diff < 1e-10,
      info = sprintf("CV.out max diff = %.2e (tolerance 1e-10)", cv_diff))
  }

  expect_identical(fit_new$att.avg, fixture$att.avg)
})

## -- P.4  IFE x nevertreated x treated_units (helper-migration fixture) --

test_that("P.4: IFE nevertreated treated_units — helper migration preserves output", {

  skip_on_cran()

  fixture <- readRDS(test_path("fixtures", "nevertreated_ife_tr.rds"))

  set.seed(42)
  fit_new <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "treated_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  expect_identical(fit_new$r.cv, fixture$r.cv)

  if (is.null(fixture$CV.out) && is.null(fit_new$CV.out)) {
    expect_true(TRUE)
  } else {
    cv_diff <- max(abs(fit_new$CV.out - fixture$CV.out), na.rm = TRUE)
    expect_true(cv_diff < 1e-10,
      info = sprintf("CV.out max diff = %.2e (tolerance 1e-10)", cv_diff))
  }

  expect_identical(fit_new$att.avg, fixture$att.avg)
})

## -- P.5  CFE x nevertreated x all_units (helper-migration fixture) --

test_that("P.5: CFE nevertreated all_units — helper migration preserves output", {

  skip_on_cran()

  fixture <- readRDS(test_path("fixtures", "nevertreated_cfe_all.rds"))

  set.seed(42)
  fit_new <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "cfe",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "all_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  expect_identical(fit_new$r.cv, fixture$r.cv)

  if (is.null(fixture$CV.out) && is.null(fit_new$CV.out)) {
    expect_true(TRUE)
  } else {
    cv_diff <- max(abs(fit_new$CV.out - fixture$CV.out), na.rm = TRUE)
    expect_true(cv_diff < 1e-10,
      info = sprintf("CV.out max diff = %.2e (tolerance 1e-10)", cv_diff))
  }

  expect_identical(fit_new$att.avg, fixture$att.avg)
})

## -- P.6  CFE x nevertreated x treated_units (helper-migration fixture) --

test_that("P.6: CFE nevertreated treated_units — helper migration preserves output", {

  skip_on_cran()

  fixture <- readRDS(test_path("fixtures", "nevertreated_cfe_tr.rds"))

  set.seed(42)
  fit_new <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "cfe",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "treated_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  expect_identical(fit_new$r.cv, fixture$r.cv)

  if (is.null(fixture$CV.out) && is.null(fit_new$CV.out)) {
    expect_true(TRUE)
  } else {
    cv_diff <- max(abs(fit_new$CV.out - fixture$CV.out), na.rm = TRUE)
    expect_true(cv_diff < 1e-10,
      info = sprintf("CV.out max diff = %.2e (tolerance 1e-10)", cv_diff))
  }

  expect_identical(fit_new$att.avg, fixture$att.avg)
})

## =================================================================
## Section A: API Back-Compatibility
## =================================================================

## -- A.1  parallel = TRUE unchanged (threshold gates to serial on simdata) --

test_that("A.1: parallel=TRUE on small panel is backward-compatible (threshold gates to serial)", {

  skip_on_cran()

  set.seed(42)
  fit_ref <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "all_units",
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  ## simdata Nco * TT < 20000 => parallel=TRUE gates to serial => results identical
  set.seed(42)
  fit_true <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "all_units",
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = TRUE
    )
  ))

  expect_identical(fit_ref$r.cv, fit_true$r.cv)

  cv_diff <- max(abs(fit_ref$CV.out.ife - fit_true$CV.out.ife), na.rm = TRUE)
  expect_true(cv_diff < 1e-10,
    info = sprintf("CV.out.ife max diff = %.2e (tolerance 1e-10)", cv_diff))
})

## -- A.2  parallel = FALSE deterministic ----------------------------

test_that("A.2: parallel=FALSE two identical calls produce identical results", {

  skip_on_cran()

  set.seed(42)
  fit1 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  set.seed(42)
  fit2 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  expect_identical(fit1$r.cv, fit2$r.cv)

  cv_diff <- max(abs(fit1$CV.out.ife - fit2$CV.out.ife), na.rm = TRUE)
  expect_true(cv_diff < 1e-12,
    info = sprintf("CV.out.ife max diff = %.2e (tolerance 1e-12)", cv_diff))
})

## -- A.3  Invalid parallel value raises informative error -----------

test_that("A.3: invalid parallel value raises an informative error", {

  expect_error(
    fect::fect(
      Y ~ D,
      data     = simdata,
      index    = c("id", "time"),
      method   = "ife",
      CV       = TRUE,
      parallel = "all"
    ),
    regexp = "parallel.*must be"
  )

  expect_error(
    fect::fect(
      Y ~ D,
      data     = simdata,
      index    = c("id", "time"),
      method   = "ife",
      CV       = TRUE,
      parallel = 2L
    ),
    regexp = "parallel.*must be"
  )
})

## =================================================================
## Section S: Selective Switches
## =================================================================

## -- S.1  parallel = "cv" engages CV, leaves bootstrap serial ------

test_that("S.1: parallel='cv' emits CV parallel banner but not bootstrap banner", {

  skip_on_cran()

  msgs <- capture.output({
    set.seed(42)
    suppressWarnings(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 0:2,
        CV        = TRUE,
        k         = 3,
        cv.method = "all_units",
        time.component.from = "notyettreated",
        se        = TRUE,
        nboots    = 5,
        parallel  = "cv",
        cores     = 2
      )
    )
  }, type = "message")

  ## CV banner present
  expect_true(any(grepl("Parallel CV", msgs)),
    info = "Expected 'Parallel CV' banner when parallel='cv'")

  ## Bootstrap parallel banner absent
  expect_false(any(grepl("Parallel computing", msgs)),
    info = "Did not expect 'Parallel computing' banner when parallel='cv' (boot should be serial)")
})

## -- S.2  parallel = "boot" engages bootstrap, leaves CV serial ----

test_that("S.2: parallel='boot' emits bootstrap banner but not CV parallel banner", {

  skip_on_cran()

  msgs <- capture.output({
    set.seed(42)
    suppressWarnings(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 0:2,
        CV        = TRUE,
        k         = 3,
        cv.method = "all_units",
        time.component.from = "notyettreated",
        se        = TRUE,
        nboots    = 5,
        parallel  = "boot",
        cores     = 2
      )
    )
  }, type = "message")

  ## Bootstrap banner present
  expect_true(any(grepl("Parallel computing", msgs)),
    info = "Expected 'Parallel computing' banner when parallel='boot'")

  ## CV banner absent
  expect_false(any(grepl("Parallel CV", msgs)),
    info = "Did not expect 'Parallel CV' banner when parallel='boot' (CV should be serial)")
})

## =================================================================
## Section T: Threshold Gating and Override
## =================================================================

## -- T.1  parallel = TRUE on small panel -> serial (no CV banner) --

test_that("T.1: parallel=TRUE on small panel runs CV serially (threshold gate)", {

  skip_on_cran()

  ## simdata: Nco * TT well below 20,000 threshold => parallel auto-gates to serial
  msgs <- capture.output({
    set.seed(42)
    suppressWarnings(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 0:3,
        CV        = TRUE,
        k         = 5,
        cv.method = "all_units",
        time.component.from = "notyettreated",
        se        = FALSE,
        parallel  = TRUE
      )
    )
  }, type = "message")

  expect_false(any(grepl("Parallel CV", msgs)),
    info = "Threshold gate should suppress 'Parallel CV' banner for small simdata panel")
})

## -- T.2  parallel = "cv" on small panel -> parallel engages (override) --

test_that("T.2: parallel='cv' on small panel forces parallel (override)", {

  skip_on_cran()

  msgs <- capture.output({
    set.seed(42)
    suppressWarnings(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 0:3,
        CV        = TRUE,
        k         = 5,
        cv.method = "all_units",
        time.component.from = "notyettreated",
        se        = FALSE,
        parallel  = "cv",
        cores     = 2
      )
    )
  }, type = "message")

  expect_true(any(grepl("Parallel CV", msgs)),
    info = "Explicit parallel='cv' should override threshold and emit 'Parallel CV' banner")
})

## -- T.3  parallel = "cv" on small panel -> results correct --------

test_that("T.3: parallel='cv' on small panel produces numerically correct output", {

  skip_on_cran()

  set.seed(42)
  fit_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "all_units",
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  set.seed(42)
  fit_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "all_units",
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  expect_identical(fit_seq$r.cv, fit_par$r.cv)

  cv_diff <- max(abs(fit_seq$CV.out.ife - fit_par$CV.out.ife), na.rm = TRUE)
  expect_true(cv_diff < 1e-10,
    info = sprintf("CV.out.ife max diff = %.2e (tolerance 1e-10)", cv_diff))
})

## =================================================================
## Section R: Plan Restoration
## =================================================================

## -- R.1  Plan restored after normal return ------------------------

test_that("R.1: future plan restored after fect() with parallel='cv'", {

  skip_on_cran()

  future::plan(future::sequential)
  plan_before_class <- class(future::plan())

  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  plan_after_class <- class(future::plan())
  expect_identical(plan_before_class, plan_after_class)

  future::plan(future::sequential)  ## cleanup
})

## -- R.2  Plan restored even when fect errors internally -----------

test_that("R.2: future plan restored even when fect returns an error", {

  skip_on_cran()

  future::plan(future::sequential)
  plan_before_class <- class(future::plan())

  ## min.T0 = 9999L forces "All treated units have been removed" error
  tryCatch(
    suppressMessages(suppressWarnings(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 0:3,
        CV        = TRUE,
        k         = 5,
        se        = FALSE,
        parallel  = "cv",
        cores     = 2,
        min.T0    = 9999L
      )
    )),
    error = function(e) NULL
  )

  plan_after_class <- class(future::plan())
  expect_identical(plan_before_class, plan_after_class)

  future::plan(future::sequential)  ## cleanup
})

## =================================================================
## Section D: Reproducibility
## =================================================================

## -- D.1  Same seed -> bit-identical parallel results --------------

test_that("D.1: two parallel runs with same seed produce identical CV.out.ife and r.cv", {

  skip_on_cran()

  set.seed(42)
  fit_par1 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "all_units",
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  set.seed(42)
  fit_par2 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      k         = 5,
      cv.method = "all_units",
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  expect_identical(fit_par1$r.cv, fit_par2$r.cv)
  expect_identical(fit_par1$CV.out.ife, fit_par2$CV.out.ife)
})

## =================================================================
## Section L: LOO Path Unaffected
## =================================================================

## -- L.1  LOO path identical regardless of parallel ---------------
## Note: test-spec.md L.1 listed time.component.from = "notyettreated"
## with cv.method = "loo", which is an invalid combination (LOO is only
## supported by fect_nevertreated.R). Per tester audit, L.1 is a
## test-spec error. This test uses time.component.from = "nevertreated"
## (matching G.7 in test-score-unify.R which already covers this path
## substantively with ntdata). With simdata the nevertreated path fires
## an early-exit (insufficient pre-treatment records), so r.cv = 0 and
## CV.out = NULL in both modes — the test confirms no error and identity.

test_that("L.1: LOO path produces identical results for any parallel value (nevertreated)", {

  skip_on_cran()

  set.seed(42)
  fit_loo_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      cv.method = "loo",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  set.seed(42)
  fit_loo_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:3,
      CV        = TRUE,
      cv.method = "loo",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  expect_identical(fit_loo_seq$r.cv, fit_loo_par$r.cv)

  ## CV.out.ife may be NULL (early-exit); handle both cases
  if (is.null(fit_loo_seq$CV.out.ife) && is.null(fit_loo_par$CV.out.ife)) {
    expect_true(TRUE)
  } else {
    cv_diff <- max(abs(fit_loo_seq$CV.out.ife - fit_loo_par$CV.out.ife), na.rm = TRUE)
    expect_true(cv_diff < 1e-12,
      info = sprintf("CV.out.ife max diff = %.2e (tolerance 1e-12)", cv_diff))
  }
})

## =================================================================
## Section E: Edge Cases
## =================================================================

## -- E.1  k = 1 with parallel = "cv" runs serial silently ----------

test_that("E.1: k=1 with parallel='cv' runs serial silently (no error)", {

  skip_on_cran()

  expect_no_error({
    set.seed(42)
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 0:2,
        CV        = TRUE,
        k         = 1,
        se        = FALSE,
        parallel  = "cv",
        cores     = 2
      )
    ))
  })
})

## -- E.2  cores = 1 with parallel = "cv" runs without error --------

test_that("E.2: cores=1 with parallel='cv' runs without error", {

  skip_on_cran()

  expect_no_error({
    set.seed(42)
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 0:2,
        CV        = TRUE,
        k         = 3,
        se        = FALSE,
        parallel  = "cv",
        cores     = 1
      )
    ))
  })
})

## -- E.3  parallel = FALSE with cores = 4 ignores cores (no CV banner) --

test_that("E.3: parallel=FALSE with cores=4 runs sequentially (no CV banner)", {

  skip_on_cran()

  msgs <- capture.output({
    set.seed(42)
    suppressWarnings(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 0:2,
        CV        = TRUE,
        k         = 3,
        se        = FALSE,
        parallel  = FALSE,
        cores     = 4
      )
    )
  }, type = "message")

  expect_false(any(grepl("Parallel CV", msgs)),
    info = "parallel=FALSE should suppress CV parallelism even when cores=4")
})

## -- E.4  c("cv","boot") form is accepted without error ------------

test_that("E.4: parallel=c('cv','boot') is accepted without error", {

  skip_on_cran()

  expect_no_error({
    set.seed(42)
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 0:2,
        CV        = TRUE,
        k         = 3,
        se        = FALSE,
        parallel  = c("cv", "boot"),
        cores     = 2
      )
    ))
  })
})

## -- E.5  criterion = "pc" with parallel = "cv" — Bug #3 regression guard --
## Bug #3: cv.R parallel branch guard used `criterion != "PC"` (uppercase)
## while the canonical form is `"pc"` (lowercase). Fixed in 379751d.

test_that("E.5: criterion='pc' with parallel='cv' runs without error (Bug #3 guard)", {

  skip_on_cran()

  expect_no_error({
    set.seed(42)
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 0:3,
        CV        = TRUE,
        k         = 5,
        criterion = "pc",
        se        = FALSE,
        parallel  = "cv",
        cores     = 2
      )
    ))
  })
})

## -- E.6  Bootstrap + CV interaction: se=TRUE, nboots=5, parallel=TRUE — no deadlock --

test_that("E.6: bootstrap + CV interaction (se=TRUE, nboots=5, parallel=TRUE) runs to completion", {

  skip_on_cran()

  ## Defensive: clear any leaked future plan / parallel backend from
  ## upstream tests in the same suite run. Without this, an inherited
  ## doFuture / doParallel registration with stale cluster handles can
  ## cause "incorrect number of dimensions" errors in the bootstrap
  ## phase. Test passes 73/73 in isolation; this guard makes it pass
  ## inside the full devtools::test() ordering as well.
  suppressWarnings({
    try(future::plan(future::sequential), silent = TRUE)
    try(foreach::registerDoSEQ(),         silent = TRUE)
  })

  expect_no_error({
    set.seed(42)
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 0:2,
        CV        = TRUE,
        k         = 3,
        se        = TRUE,
        nboots    = 5,
        parallel  = TRUE
      )
    ))
  })
})

## -- E.7  Boot under stale future plan: simulate the CV→boot plan-leak
##         that produced the E.6 PSOCK race. Set up a cluster plan,
##         stop its workers behind future's back, then run boot. The
##         cluster-init retry loop in run_dopar_retry should recover.
test_that("E.7: bootstrap survives a stale future::plan() with dead workers", {

  skip_on_cran()

  ## Save / restore plan around the test.
  old_plan <- future::plan()
  on.exit(suppressWarnings(future::plan(old_plan)), add = TRUE)

  ## Install a fresh PSOCK cluster, then forcibly kill its workers
  ## while the plan still points at it. Mirrors the post-CV state the
  ## bootstrap path inherits when CV's on.exit restores a plan whose
  ## underlying cluster was stopped.
  cl <- parallelly::makeClusterPSOCK(
    workers      = 2L,
    rscript_libs = .libPaths(),
    autoStop     = TRUE
  )
  suppressWarnings(future::plan(future::cluster, workers = cl))
  try(parallel::stopCluster(cl), silent = TRUE)

  expect_no_error({
    set.seed(42)
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 0:1,
        CV        = FALSE,
        force     = "two-way",
        se        = TRUE,
        nboots    = 5,
        parallel  = TRUE
      )
    ))
  })
})

## =================================================================
## Section G: Regression guard — nevertreated parallel paths
## =================================================================
## G.1-G.4 in test-score-unify.R already cover the nevertreated IFE/CFE
## fold-parallel paths with synthetic data (ntdata, N=50 TT=20 Ntr=15)
## that exercises substantive CV scoring (r.cv > 0, CV.out non-NULL).
## No duplication needed here; the P.3-P.6 fixture tests above guard
## against the simdata early-exit case.

## =================================================================
## Section M: MC Parallel CV (Phase 2)
## =================================================================
## Tests for the MC notyettreated parallel branch added in Phase 2.
## Mirror the P-section style for numerical identity and the T/R-section
## style for threshold gating and plan restoration.
##
## RNG discipline: set.seed() immediately before EVERY fect() call.
## simdata is used throughout (small panel; Nco * TT < 20000 threshold).
## parallel = "cv" overrides the threshold for MC explicitly.
## =================================================================

## -- M.1  MC notyettreated serial == parallel (numerical identity) --

test_that("M.1: MC notyettreated serial == parallel (lambda.cv and CV.out.mc within 1e-10)", {

  skip_on_cran()

  set.seed(42)
  fit_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "mc",
      CV        = TRUE,
      k         = 5,
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  set.seed(42)
  fit_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "mc",
      CV        = TRUE,
      k         = 5,
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  ## lambda.cv: plain numeric (no name attribute needed for MC)
  expect_equal(fit_seq$lambda.cv, fit_par$lambda.cv, tolerance = 1e-10)

  ## CV.out.mc: parallel computes all lambdas; serial may break_check early, leaving some
  ## rows at the 1e20 sentinel. Compare only the rows both paths computed (MSPE < 1e19).
  ## The optimal lambda (lambda.cv) is selected from the rows both paths evaluated,
  ## so numerical identity of that selection is the key correctness property.
  both_computed <- fit_seq$CV.out.mc[, "MSPE"] < 1e19 & fit_par$CV.out.mc[, "MSPE"] < 1e19
  if (any(both_computed)) {
    cv_diff <- max(abs(
      fit_seq$CV.out.mc[both_computed, ] - fit_par$CV.out.mc[both_computed, ]
    ), na.rm = TRUE)
    expect_true(cv_diff < 1e-10,
      info = sprintf("CV.out.mc max diff (computed rows only) = %.2e (tolerance 1e-10)", cv_diff))
  }

  ## att.avg: same lambda.cv => same final model => identical
  expect_equal(fit_seq$att.avg, fit_par$att.avg, tolerance = 1e-10)
})

## -- M.2  MC parallel: banner contains "Parallel CV (MC)" string ----

test_that("M.2: parallel='cv' with method='mc' emits MC-specific banner", {

  skip_on_cran()

  msgs <- capture.output({
    set.seed(42)
    suppressWarnings(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "mc",
        CV        = TRUE,
        k         = 3,
        time.component.from = "notyettreated",
        se        = FALSE,
        parallel  = "cv",
        cores     = 2
      )
    )
  }, type = "message")

  ## Banner contains "Parallel CV (MC):" to distinguish from IFE banner
  expect_true(any(grepl("Parallel CV \\(MC\\)", msgs)),
    info = "Expected 'Parallel CV (MC)' banner when parallel='cv', method='mc'")
})

## -- M.3  MC parallel: threshold gate on small panel ---------------

test_that("M.3: parallel=TRUE on small MC panel runs serially (no MC banner)", {

  skip_on_cran()

  ## simdata: Nco * TT well below 20,000 threshold => parallel=TRUE gates to serial
  msgs <- capture.output({
    set.seed(42)
    suppressWarnings(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "mc",
        CV        = TRUE,
        k         = 5,
        time.component.from = "notyettreated",
        se        = FALSE,
        parallel  = TRUE
      )
    )
  }, type = "message")

  expect_false(any(grepl("Parallel CV \\(MC\\)", msgs)),
    info = "Threshold gate should suppress MC banner for small simdata panel")
})

## -- M.4  MC parallel: plan restored after normal return ------------

test_that("M.4: future plan restored after MC parallel CV call", {

  skip_on_cran()

  future::plan(future::sequential)
  plan_before_class <- class(future::plan())

  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "mc",
      CV        = TRUE,
      k         = 5,
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  plan_after_class <- class(future::plan())
  expect_identical(plan_before_class, plan_after_class)

  future::plan(future::sequential)  ## cleanup
})

## -- M.5  method = "both" with parallel = "cv" — LIFO plan restore --

test_that("M.5: method='both' parallel='cv' restores plan correctly (LIFO)", {

  skip_on_cran()

  future::plan(future::sequential)
  plan_before_class <- class(future::plan())

  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "both",
      r         = 0:2,
      CV        = TRUE,
      k         = 3,
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  plan_after_class <- class(future::plan())
  expect_identical(plan_before_class, plan_after_class)

  future::plan(future::sequential)  ## cleanup
})

## -- M.6  MC reproducibility: two parallel runs bit-identical -------

test_that("M.6: two MC parallel runs with same seed produce identical results", {

  skip_on_cran()

  set.seed(42)
  fit_par1 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "mc",
      CV        = TRUE,
      k         = 5,
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  set.seed(42)
  fit_par2 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "mc",
      CV        = TRUE,
      k         = 5,
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  expect_identical(fit_par1$lambda.cv, fit_par2$lambda.cv)
  expect_identical(fit_par1$CV.out.mc, fit_par2$CV.out.mc)
})

## =================================================================
## Section C: CFE Nevertreated Parallel CV (Phase 3)
## =================================================================
## Tests for the CFE nevertreated flat r×k parallel dispatch added in
## Phase 3. Mirror the M-section structure.
##
## RNG discipline: set.seed() immediately before EVERY fect() call.
## Synthetic data (ntdata_cfe) is used because simdata triggers an
## early-exit (insufficient pre-treatment records). ntdata_cfe is
## generated with a fixed seed inside each test so the test is
## self-contained and reproducible.
##
## See: statsclaw-workspace/fect/runs/REQ-parallel-cv-phase3/spec.md
## =================================================================

## Helper: generate reproducible synthetic panel for CFE tests
## N = 38 (Nco=30, Ntr=8), TT=30, treatment from period 20.
## Nco * TT = 900 < 60000 threshold: use parallel = "cv" to force.
.make_cfe_panel <- function(seed = 1234) {
  set.seed(seed)
  TT <- 30; Nco <- 30; Ntr <- 8; N <- Nco + Ntr
  panel <- expand.grid(id = 1:N, time = 1:TT)
  panel <- panel[order(panel$id, panel$time), ]
  fac1 <- sin(seq(0, pi, length.out = TT))
  fac2 <- cos(seq(0, pi, length.out = TT))
  for (i in 1:N) {
    lambda1_i <- rnorm(1, 0, 0.5); lambda2_i <- rnorm(1, 0, 0.5)
    panel$Y[panel$id == i] <- 2*lambda1_i*fac1 + lambda2_i*fac2 + rnorm(TT, 0, 0.3)
  }
  panel$D <- ifelse(panel$id > Nco & panel$time >= 20, 1, 0)
  panel
}

## -- C.1  CFE nevertreated all_units serial == parallel -------------

test_that("C.1: CFE nevertreated all_units serial == parallel (within 1e-10)", {

  skip_on_cran()

  ntdata_cfe <- .make_cfe_panel()

  set.seed(42)
  fit_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = ntdata_cfe,
      index     = c("id", "time"),
      method    = "cfe",
      r         = 0:2,
      CV        = TRUE,
      k         = 3,
      cv.method = "all_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  set.seed(42)
  fit_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = ntdata_cfe,
      index     = c("id", "time"),
      method    = "cfe",
      r         = 0:2,
      CV        = TRUE,
      k         = 3,
      cv.method = "all_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  ## r.cv must match
  expect_identical(fit_seq$r.cv, fit_par$r.cv)

  ## CV.out scores within 1e-10
  if (!is.null(fit_seq$CV.out) && !is.null(fit_par$CV.out)) {
    cv_diff <- max(abs(fit_seq$CV.out - fit_par$CV.out), na.rm = TRUE)
    expect_true(cv_diff < 1e-10,
      info = sprintf("CV.out max diff = %.2e (tolerance 1e-10)", cv_diff))
  }

  ## att.avg: same r.cv => same final model
  expect_equal(fit_seq$att.avg, fit_par$att.avg, tolerance = 1e-10)
})

## -- C.2  CFE parallel banner contains "Parallel CV (CFE)" ----------

test_that("C.2: parallel='cv' with method='cfe' emits CFE-specific banner", {

  skip_on_cran()

  ntdata_cfe <- .make_cfe_panel()

  msgs <- capture.output({
    set.seed(42)
    suppressWarnings(
      fect::fect(
        Y ~ D,
        data      = ntdata_cfe,
        index     = c("id", "time"),
        method    = "cfe",
        r         = 0:2,
        CV        = TRUE,
        k         = 3,
        cv.method = "all_units",
        time.component.from = "nevertreated",
        se        = FALSE,
        parallel  = "cv",
        cores     = 2
      )
    )
  }, type = "message")

  ## Banner contains "Parallel CV (CFE):" to distinguish from IFE banner
  expect_true(any(grepl("Parallel CV \\(CFE\\)", msgs)),
    info = "Expected 'Parallel CV (CFE)' banner when parallel='cv', method='cfe'")
})

## -- C.3  CFE threshold gate: parallel=TRUE on small panel -> serial --

test_that("C.3: parallel=TRUE on small CFE panel runs serially (CFE threshold gate)", {

  skip_on_cran()

  ntdata_cfe <- .make_cfe_panel()
  ## ntdata_cfe: Nco * TT = 30 * 30 = 900 < 60000 => parallel=TRUE gates to serial

  msgs <- capture.output({
    set.seed(42)
    suppressWarnings(
      fect::fect(
        Y ~ D,
        data      = ntdata_cfe,
        index     = c("id", "time"),
        method    = "cfe",
        r         = 0:2,
        CV        = TRUE,
        k         = 3,
        cv.method = "all_units",
        time.component.from = "nevertreated",
        se        = FALSE,
        parallel  = TRUE
      )
    )
  }, type = "message")

  expect_false(any(grepl("Parallel CV \\(CFE\\)", msgs)),
    info = "CFE threshold (60k) should suppress CFE parallel banner for small panel")
})

## -- C.4  CFE parallel: plan restored after normal return -----------

test_that("C.4: future plan restored after CFE parallel CV call", {

  skip_on_cran()

  ntdata_cfe <- .make_cfe_panel()

  future::plan(future::sequential)
  plan_before_class <- class(future::plan())

  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = ntdata_cfe,
      index     = c("id", "time"),
      method    = "cfe",
      r         = 0:2,
      CV        = TRUE,
      k         = 3,
      cv.method = "all_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  plan_after_class <- class(future::plan())
  expect_identical(plan_before_class, plan_after_class)

  future::plan(future::sequential)  ## cleanup
})

## -- C.5  CFE reproducibility: two parallel runs bit-identical ------

test_that("C.5: two CFE parallel runs with same seed produce identical results", {

  skip_on_cran()

  ntdata_cfe <- .make_cfe_panel()

  set.seed(42)
  fit_par1 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = ntdata_cfe,
      index     = c("id", "time"),
      method    = "cfe",
      r         = 0:2,
      CV        = TRUE,
      k         = 3,
      cv.method = "all_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  set.seed(42)
  fit_par2 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = ntdata_cfe,
      index     = c("id", "time"),
      method    = "cfe",
      r         = 0:2,
      CV        = TRUE,
      k         = 3,
      cv.method = "all_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  expect_identical(fit_par1$r.cv, fit_par2$r.cv)
  if (!is.null(fit_par1$CV.out) && !is.null(fit_par2$CV.out)) {
    expect_identical(fit_par1$CV.out, fit_par2$CV.out)
  }
})

## -- C.6  CFE notyettreated + CV = TRUE => CV forced to FALSE -------
## Architectural confirmation: planner's finding that method="cfe" with
## time.component.from="notyettreated" sets CV=FALSE in default.R.

test_that("C.6: CFE notyettreated+CV=TRUE is silently forced to CV=FALSE (no CV output)", {

  skip_on_cran()

  set.seed(42)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "cfe",
      CV        = TRUE,
      time.component.from = "notyettreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  ## CV forced to FALSE => no CV output, r.cv = 0 (single r candidate)
  expect_null(fit$CV.out.ife,
    info = "CV.out.ife should be NULL when CV is forced FALSE for cfe+notyettreated")
  expect_null(fit$CV.out,
    info = "CV.out should be NULL when CV is forced FALSE for cfe+notyettreated")
})

## =================================================================
## Section N: Nevertreated Flat-Dispatch Migration Regression (Phase 3)
## =================================================================
## Pin the post-migration serial == parallel identity for both
## IFE and CFE nevertreated paths using the Phase 3 fixture data.
##
## Fixtures generated from HEAD 5531166 (pre-refactor serial path)
## and stored in tests/testthat/fixtures/phase3-baseline-*.rds.
## Serial path remains identical; parallel path (future_lapply) must
## match within 1e-10.
##
## RNG discipline: set.seed(42) immediately before each fect() call.
## =================================================================

## -- N.1  IFE nevertreated all_units: parallel == serial (fixture) --

test_that("N.1: IFE nevertreated all_units parallel matches phase3 fixture (within 1e-10)", {

  skip_on_cran()

  fixture <- readRDS(test_path("fixtures", "phase3-baseline-ife-all-serial.rds"))
  ntdata <- fixture$panel  ## NOTE: if fixture was saved without $panel, use inline
  ## The fixture was saved with the panel as part of the environment — regenerate inline
  set.seed(if (!is.null(fixture$panel_seed)) fixture$panel_seed else 1234)
  TT <- 30; Nco <- 30; Ntr <- 8; N <- Nco + Ntr
  ntdata <- expand.grid(id = 1:N, time = 1:TT)
  ntdata <- ntdata[order(ntdata$id, ntdata$time), ]
  fac1 <- sin(seq(0, pi, length.out = TT)); fac2 <- cos(seq(0, pi, length.out = TT))
  for (i in 1:N) {
    lambda1_i <- rnorm(1, 0, 0.5); lambda2_i <- rnorm(1, 0, 0.5)
    ntdata$Y[ntdata$id == i] <- 2*lambda1_i*fac1 + lambda2_i*fac2 + rnorm(TT, 0, 0.3)
  }
  ntdata$D <- ifelse(ntdata$id > Nco & ntdata$time >= 20, 1, 0)

  set.seed(42)
  fit_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = ntdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:2,
      CV        = TRUE,
      cv.rule   = "1se",    # explicit pin (matches v2.3.0 default; future-proof)
      cv.prop   = 0.1,      # explicit pin (matches v2.3.0 default)
      cv.donut  = 1,        # explicit pin (matches v2.3.0 default)
      k         = 20,       # explicit pin (matches v2.3.0 default; fixture regen under k=20)
      cv.method = "all_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  expect_identical(fit_par$r.cv, fixture$fit$r.cv)
  att_diff <- max(abs(fit_par$att.avg - fixture$fit$att.avg), na.rm = TRUE)
  expect_true(att_diff < 1e-10,
    info = sprintf("IFE nevertreated all_units parallel vs fixture: att.avg diff = %.2e", att_diff))
})

## -- N.2  IFE nevertreated treated_units: serial still matches fixture --

test_that("N.2: IFE nevertreated treated_units serial matches phase3 fixture (within 1e-10)", {

  skip_on_cran()

  fixture <- readRDS(test_path("fixtures", "phase3-baseline-ife-tr-serial.rds"))
  set.seed(if (!is.null(fixture$panel_seed)) fixture$panel_seed else 1234)
  TT <- 30; Nco <- 30; Ntr <- 8; N <- Nco + Ntr
  ntdata <- expand.grid(id = 1:N, time = 1:TT)
  ntdata <- ntdata[order(ntdata$id, ntdata$time), ]
  fac1 <- sin(seq(0, pi, length.out = TT)); fac2 <- cos(seq(0, pi, length.out = TT))
  for (i in 1:N) {
    lambda1_i <- rnorm(1, 0, 0.5); lambda2_i <- rnorm(1, 0, 0.5)
    ntdata$Y[ntdata$id == i] <- 2*lambda1_i*fac1 + lambda2_i*fac2 + rnorm(TT, 0, 0.3)
  }
  ntdata$D <- ifelse(ntdata$id > Nco & ntdata$time >= 20, 1, 0)

  set.seed(42)
  fit_ser <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = ntdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:2,
      CV        = TRUE,
      cv.rule   = "1se",    # explicit pin (matches v2.3.0 default; future-proof)
      cv.prop   = 0.1,      # explicit pin (matches v2.3.0 default)
      cv.donut  = 1,        # explicit pin (matches v2.3.0 default)
      k         = 20,       # explicit pin (matches v2.3.0 default; fixture regen under k=20)
      cv.method = "treated_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  expect_identical(fit_ser$r.cv, fixture$fit$r.cv)
  att_diff <- max(abs(fit_ser$att.avg - fixture$fit$att.avg), na.rm = TRUE)
  expect_true(att_diff < 1e-10,
    info = sprintf("IFE nevertreated treated_units serial vs fixture: att.avg diff = %.2e", att_diff))
})

## -- N.3  CFE nevertreated all_units: parallel == serial (fixture) --

test_that("N.3: CFE nevertreated all_units parallel matches phase3 fixture (within 1e-10)", {

  skip_on_cran()

  fixture <- readRDS(test_path("fixtures", "phase3-baseline-cfe-all-serial.rds"))
  set.seed(if (!is.null(fixture$panel_seed)) fixture$panel_seed else 1234)
  TT <- 30; Nco <- 30; Ntr <- 8; N <- Nco + Ntr
  ntdata <- expand.grid(id = 1:N, time = 1:TT)
  ntdata <- ntdata[order(ntdata$id, ntdata$time), ]
  fac1 <- sin(seq(0, pi, length.out = TT)); fac2 <- cos(seq(0, pi, length.out = TT))
  for (i in 1:N) {
    lambda1_i <- rnorm(1, 0, 0.5); lambda2_i <- rnorm(1, 0, 0.5)
    ntdata$Y[ntdata$id == i] <- 2*lambda1_i*fac1 + lambda2_i*fac2 + rnorm(TT, 0, 0.3)
  }
  ntdata$D <- ifelse(ntdata$id > Nco & ntdata$time >= 20, 1, 0)

  set.seed(42)
  fit_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = ntdata,
      index     = c("id", "time"),
      method    = "cfe",
      r         = 0:2,
      CV        = TRUE,
      cv.rule   = "1se",    # explicit pin (matches v2.3.0 default; future-proof)
      cv.prop   = 0.1,      # explicit pin (matches v2.3.0 default)
      cv.donut  = 1,        # explicit pin (matches v2.3.0 default)
      k         = 20,       # explicit pin (matches v2.3.0 default; fixture regen under k=20)
      cv.method = "all_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  expect_identical(fit_par$r.cv, fixture$fit$r.cv)
  att_diff <- max(abs(fit_par$att.avg - fixture$fit$att.avg), na.rm = TRUE)
  expect_true(att_diff < 1e-10,
    info = sprintf("CFE nevertreated all_units parallel vs fixture: att.avg diff = %.2e", att_diff))
})

## -- N.4  CFE nevertreated treated_units: serial still matches fixture --

test_that("N.4: CFE nevertreated treated_units serial matches phase3 fixture (within 1e-10)", {

  skip_on_cran()

  fixture <- readRDS(test_path("fixtures", "phase3-baseline-cfe-tr-serial.rds"))
  set.seed(if (!is.null(fixture$panel_seed)) fixture$panel_seed else 1234)
  TT <- 30; Nco <- 30; Ntr <- 8; N <- Nco + Ntr
  ntdata <- expand.grid(id = 1:N, time = 1:TT)
  ntdata <- ntdata[order(ntdata$id, ntdata$time), ]
  fac1 <- sin(seq(0, pi, length.out = TT)); fac2 <- cos(seq(0, pi, length.out = TT))
  for (i in 1:N) {
    lambda1_i <- rnorm(1, 0, 0.5); lambda2_i <- rnorm(1, 0, 0.5)
    ntdata$Y[ntdata$id == i] <- 2*lambda1_i*fac1 + lambda2_i*fac2 + rnorm(TT, 0, 0.3)
  }
  ntdata$D <- ifelse(ntdata$id > Nco & ntdata$time >= 20, 1, 0)

  set.seed(42)
  fit_ser <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = ntdata,
      index     = c("id", "time"),
      method    = "cfe",
      r         = 0:2,
      CV        = TRUE,
      k         = 20,       # explicit pin (matches v2.3.0 default; fixture regen under k=20)
      cv.method = "treated_units",
      cv.rule   = "1se",    # explicit pin (matches v2.3.0 default; future-proof)
      cv.prop   = 0.1,      # explicit pin (matches v2.3.0 default)
      cv.donut  = 1,        # explicit pin (matches v2.3.0 default)
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = FALSE
    )
  ))

  expect_identical(fit_ser$r.cv, fixture$fit$r.cv)
  att_diff <- max(abs(fit_ser$att.avg - fixture$fit$att.avg), na.rm = TRUE)
  expect_true(att_diff < 1e-10,
    info = sprintf("CFE nevertreated treated_units serial vs fixture: att.avg diff = %.2e", att_diff))
})

## -- N.6  method = "both" (sequential IFE + CFE) plan restoration ----
## Formalized from Phase 3 audit spot-check. Runs IFE nevertreated parallel
## CV followed immediately by CFE nevertreated parallel CV. Each call
## registers an on.exit() that restores the caller's sequential plan.
## After both calls complete, the outer plan must be unchanged.

test_that("N.6: plan restored after sequential IFE and CFE nevertreated parallel CV", {

  skip_on_cran()

  ## Nevertreated panel: Nco=30, Ntr=8, TT=30.  Nco*TT=900 < both thresholds;
  ## use parallel="cv" to force parallel dispatch regardless of size.
  set.seed(1234)
  TT <- 30; Nco <- 30; Ntr <- 8; N <- Nco + Ntr
  ntdata <- expand.grid(id = 1:N, time = 1:TT)
  ntdata <- ntdata[order(ntdata$id, ntdata$time), ]
  fac1 <- sin(seq(0, pi, length.out = TT))
  fac2 <- cos(seq(0, pi, length.out = TT))
  for (i in 1:N) {
    lambda1_i <- rnorm(1, 0, 0.5); lambda2_i <- rnorm(1, 0, 0.5)
    ntdata$Y[ntdata$id == i] <- 2*lambda1_i*fac1 + lambda2_i*fac2 + rnorm(TT, 0, 0.3)
  }
  ntdata$D <- ifelse(ntdata$id > Nco & ntdata$time >= 20, 1, 0)

  plan_before <- class(future::plan())

  ## First call: IFE nevertreated parallel CV
  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = ntdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:2,
      CV        = TRUE,
      k         = 3,
      cv.method = "all_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  plan_mid <- class(future::plan())
  expect_equal(plan_before, plan_mid,
    info = "plan should be restored to sequential after IFE nevertreated parallel CV")

  ## Second call: CFE nevertreated parallel CV
  set.seed(42)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = ntdata,
      index     = c("id", "time"),
      method    = "cfe",
      r         = 0:2,
      CV        = TRUE,
      k         = 3,
      cv.method = "all_units",
      time.component.from = "nevertreated",
      se        = FALSE,
      parallel  = "cv",
      cores     = 2
    )
  ))

  plan_after <- class(future::plan())
  expect_equal(plan_before, plan_after,
    info = "plan should be restored to sequential after CFE nevertreated parallel CV")
})


## =================================================================
## Section B: v2.4.3 — quiet_nonpara closure-leak regression
## =================================================================

## -- B.1  Parallel bootstrap on simdata under a tight future.globals.maxSize
##         cap. With the v2.4.3 trim of quiet_nonpara and the local 2 GiB
##         bump, no "future.globals.maxSize" warning should fire even when
##         the caller's pre-block cap is small.

test_that("B.1: parallel bootstrap on simdata does not trip future.globals.maxSize", {

  skip_on_cran()

  ## Force a tight pre-block cap to simulate the failure environment.
  ## The v2.4.3 local bump inside do_parallel_boot raises this to 2 GiB
  ## for the duration of the parallel block (with on.exit restore).
  old_max <- getOption("future.globals.maxSize", 500 * 1024^2)
  options(future.globals.maxSize = 50 * 1024^2)  # 50 MiB pre-block
  on.exit(options(future.globals.maxSize = old_max), add = TRUE)

  ## Clear leaked plan state (defensive, mirrors E.6 / E.7).
  suppressWarnings({
    try(future::plan(future::sequential), silent = TRUE)
    try(foreach::registerDoSEQ(),         silent = TRUE)
  })

  warns <- character(0)

  expect_no_error(
    withCallingHandlers(
      {
        set.seed(42)
        suppressMessages(
          fect::fect(
            Y ~ D,
            data      = simdata,
            index     = c("id", "time"),
            method    = "ife",
            r         = 1,
            CV        = FALSE,
            se        = TRUE,
            vartype   = "bootstrap",
            nboots    = 20,
            parallel  = TRUE
          )
        )
      },
      warning = function(w) {
        warns <<- c(warns, conditionMessage(w))
        invokeRestart("muffleWarning")
      }
    )
  )

  expect_false(any(grepl("future\\.globals\\.maxSize", warns)),
               info = paste("warns:", paste(warns, collapse = " | ")))
  expect_false(any(grepl("Future backend failed", warns)),
               info = paste("warns:", paste(warns, collapse = " | ")))
})

## -- B.2  Structural contract: trim_closure_env(), when applied to a
##         quiet_nonpara-shaped wrapper defined inside a heavy outer
##         frame, keeps ONLY the two referenced locals (one.nonpara,
##         boot.seq) and drops the rest. Locks in the invariant the
##         v2.4.3 fix depends on.

test_that("B.2: trim_closure_env strips fect_boot frame from quiet_nonpara-shape closure", {

  ## Reproduce the shape of fect_boot()'s local frame: heavy locals
  ## around a wrapper that only references `one.nonpara` and `boot.seq`.
  outer_fn <- function() {
    Y           <- matrix(rnorm(1e4), 100, 100)
    D           <- matrix(rnorm(1e4), 100, 100)
    big_out     <- as.list(seq_len(1e4))
    boot.seq    <- 1:20
    one.nonpara <- function(j) j
    quiet_nonpara <- function(j) {
      suppressMessages(suppressWarnings(one.nonpara(boot.seq[j])))
    }
    quiet_nonpara <- fect:::trim_closure_env(quiet_nonpara)
    quiet_nonpara
  }

  qn <- outer_fn()
  env_names <- ls(environment(qn), all.names = TRUE)

  ## The trimmed env should hold ONLY the two referenced locals.
  expect_setequal(env_names, c("one.nonpara", "boot.seq"))

  ## And it should NOT carry the heavy frame objects.
  expect_false("Y"       %in% env_names)
  expect_false("D"       %in% env_names)
  expect_false("big_out" %in% env_names)

  ## Size sanity: trimmed closure stays small (< 1 MiB).
  expect_lt(as.numeric(object.size(qn)), 1 * 1024^2)
})

## impute_Y0() — unified Y0-imputer dispatcher for parametric bootstrap
##
## Dispatches to the correct underlying estimator based on (method, predictive).
## Pure dispatcher — introduces NO RNG calls.  try() wrapping is the caller's
## responsibility (see draw.error() and one.nonpara() in boot.R).
##
## Dispatch table:
##   gsynth  OR (ife, nevertreated)   -> fect_nevertreated(method="ife",  r=tuning)
##   ife,  notyettreated              -> fect_fe(r.cv=tuning)
##   cfe,  nevertreated               -> fect_nevertreated(method="cfe",  r=tuning)
##   cfe,  notyettreated              -> fect_cfe(r.cv=tuning)
##   mc    (any)                      -> stop()  [Phase 4 – not yet implemented]

impute_Y0 <- function(
  method,           # "gsynth", "ife", "cfe"  (dispatch key 1)
  predictive,       # "notyettreated", "nevertreated"  (dispatch key 2)

  ## Data arguments (passed through to callee)
  Y,                # TT x N outcome matrix
  X,                # TT x N x p covariate array, or NULL
  D,                # TT x N treatment indicator
  W,                # TT x N weight matrix or NULL
  I,                # TT x N observation indicator
  II,               # TT x N estimation indicator
  T.on,             # TT x N integer matrix: periods since onset

  ## Extra-FE arguments — forwarded only to cfe-family callees; ignored by fect_fe
  X.extra.FE = NULL,
  X.Z        = NULL,
  X.Q        = NULL,
  X.gamma    = NULL,
  X.kappa    = NULL,
  Zgamma.id  = NULL,
  kappaQ.id  = NULL,

  ## Tuning and fit control
  tuning,           # scalar: r.cv (pre-selected rank). Forwarded as r= or r.cv= per callee.
  boot = 1,         # always 1 for bootstrap refits; kept for signature symmetry

  ## Aggregation sequences (forwarded to callee)
  T.on.balance     = NULL,
  balance.period   = NULL,
  placeboTest      = 0,
  placebo.period   = NULL,
  carryoverTest    = 0,
  carryover.period = NULL,
  calendar.enp.seq = NULL,
  time.on.seq      = NULL,
  time.off.seq     = NULL,
  time.on.seq.W    = NULL,
  time.off.seq.W   = NULL,
  time.on.seq.group = NULL,
  time.off.seq.group = NULL,
  time.on.balance.seq = NULL,

  ## Estimation control (forwarded to callee)
  force,
  hasRevs      = 1,
  tol,
  max.iteration = 1000,
  norm.para    = NULL,
  group.level  = NULL,
  group        = NULL
) {

  if (method == "gsynth" || (method == "ife" && predictive == "nevertreated")) {
    ## Branch 1: gsynth and ife+nevertreated both call fect_nevertreated(method="ife")
    fect_nevertreated(
      Y              = Y,
      X              = X,
      D              = D,
      W              = W,
      I              = I,
      II             = II,
      T.on           = T.on,
      T.on.balance   = T.on.balance,
      balance.period = balance.period,
      hasRevs        = hasRevs,
      force          = force,
      r              = tuning,
      CV             = 0,
      boot           = boot,
      placeboTest    = placeboTest,
      placebo.period = placebo.period,
      carryover.period = carryover.period,
      carryoverTest  = carryoverTest,
      calendar.enp.seq = calendar.enp.seq,
      time.on.seq    = time.on.seq,
      time.off.seq   = time.off.seq,
      time.on.seq.W  = time.on.seq.W,
      time.off.seq.W = time.off.seq.W,
      time.on.seq.group  = time.on.seq.group,
      time.off.seq.group = time.off.seq.group,
      time.on.balance.seq = time.on.balance.seq,
      norm.para      = norm.para,
      tol            = tol,
      max.iteration  = max.iteration,
      group.level    = group.level,
      group          = group,
      method         = "ife"
    )

  } else if (method == "ife" && predictive == "notyettreated") {
    ## Branch 2: ife+notyettreated -> fect_fe
    fect_fe(
      Y              = Y,
      X              = X,
      D              = D,
      W              = W,
      I              = I,
      II             = II,
      T.on           = T.on,
      T.on.balance   = T.on.balance,
      balance.period = balance.period,
      hasRevs        = hasRevs,
      force          = force,
      r.cv           = tuning,
      boot           = boot,
      placeboTest    = placeboTest,
      placebo.period = placebo.period,
      carryover.period = carryover.period,
      carryoverTest  = carryoverTest,
      calendar.enp.seq = calendar.enp.seq,
      time.on.seq    = time.on.seq,
      time.off.seq   = time.off.seq,
      time.on.seq.W  = time.on.seq.W,
      time.off.seq.W = time.off.seq.W,
      time.on.seq.group  = time.on.seq.group,
      time.off.seq.group = time.off.seq.group,
      time.on.balance.seq = time.on.balance.seq,
      tol            = tol,
      max.iteration  = max.iteration,
      norm.para      = norm.para,
      group.level    = group.level,
      group          = group
    )

  } else if (method == "cfe" && predictive == "nevertreated") {
    ## Branch 3: cfe+nevertreated -> fect_nevertreated(method="cfe")
    fect_nevertreated(
      Y              = Y,
      X              = X,
      D              = D,
      W              = W,
      I              = I,
      II             = II,
      T.on           = T.on,
      T.on.balance   = T.on.balance,
      balance.period = balance.period,
      hasRevs        = hasRevs,
      force          = force,
      r              = tuning,
      CV             = 0,
      boot           = boot,
      placeboTest    = placeboTest,
      placebo.period = placebo.period,
      carryover.period = carryover.period,
      carryoverTest  = carryoverTest,
      calendar.enp.seq = calendar.enp.seq,
      time.on.seq    = time.on.seq,
      time.off.seq   = time.off.seq,
      time.on.seq.W  = time.on.seq.W,
      time.off.seq.W = time.off.seq.W,
      time.on.seq.group  = time.on.seq.group,
      time.off.seq.group = time.off.seq.group,
      time.on.balance.seq = time.on.balance.seq,
      norm.para      = norm.para,
      tol            = tol,
      max.iteration  = max.iteration,
      group.level    = group.level,
      group          = group,
      method         = "cfe",
      X.extra.FE    = X.extra.FE,
      X.Z           = X.Z,
      X.Q           = X.Q,
      X.gamma       = X.gamma,
      X.kappa       = X.kappa,
      Zgamma.id     = Zgamma.id,
      kappaQ.id     = kappaQ.id
    )

  } else if (method == "cfe" && predictive == "notyettreated") {
    ## Branch 4: cfe+notyettreated -> fect_cfe
    fect_cfe(
      Y              = Y,
      X              = X,
      D              = D,
      W              = W,
      X.extra.FE    = X.extra.FE,
      X.Z           = X.Z,
      X.Q           = X.Q,
      X.gamma       = X.gamma,
      X.kappa       = X.kappa,
      Zgamma.id     = Zgamma.id,
      kappaQ.id     = kappaQ.id,
      I              = I,
      II             = II,
      T.on           = T.on,
      T.on.balance   = T.on.balance,
      balance.period = balance.period,
      hasRevs        = hasRevs,
      force          = force,
      r.cv           = tuning,
      boot           = boot,
      placeboTest    = placeboTest,
      placebo.period = placebo.period,
      carryover.period = carryover.period,
      carryoverTest  = carryoverTest,
      calendar.enp.seq = calendar.enp.seq,
      time.on.seq    = time.on.seq,
      time.off.seq   = time.off.seq,
      time.on.seq.W  = time.on.seq.W,
      time.off.seq.W = time.off.seq.W,
      time.on.seq.group  = time.on.seq.group,
      time.off.seq.group = time.off.seq.group,
      time.on.balance.seq = time.on.balance.seq,
      tol            = tol,
      max.iteration  = max.iteration,
      norm.para      = norm.para,
      group.level    = group.level,
      group          = group
    )

  } else {
    stop(paste0(
      "impute_Y0: unsupported combination method='", method,
      "', predictive='", predictive, "'. ",
      "MC support is planned for Phase 4."
    ))
  }
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

wE_adj <- function(E, FE, W, I) {
    .Call(`_fect_wE_adj`, E, FE, W, I)
}

loglh <- function(Y_fit, Y) {
    .Call(`_fect_loglh`, Y_fit, Y)
}

loglh_ub <- function(Y_fit, Y, I) {
    .Call(`_fect_loglh_ub`, Y_fit, Y, I)
}

data_ub_adj <- function(I_data, data) {
    .Call(`_fect_data_ub_adj`, I_data, data)
}

XXinv <- function(X) {
    .Call(`_fect_XXinv`, X)
}

wXXinv <- function(X, w) {
    .Call(`_fect_wXXinv`, X, w)
}

panel_beta <- function(X, xxinv, Y, FE) {
    .Call(`_fect_panel_beta`, X, xxinv, Y, FE)
}

wpanel_beta <- function(X, xwxinv, w, Y, FE) {
    .Call(`_fect_wpanel_beta`, X, xwxinv, w, Y, FE)
}

panel_est <- function(X, Y, MF) {
    .Call(`_fect_panel_est`, X, Y, MF)
}

inter_fe_d_qr <- function(Y, Y_fit0, FE0, factor0, xi0, X, r, force, mniter = 5000L, w = 1.0, tol = 1e-5) {
    .Call(`_fect_inter_fe_d_qr`, Y, Y_fit0, FE0, factor0, xi0, X, r, force, mniter, w, tol)
}

inter_fe_d_qr_ub <- function(Y, Y_fit0, FE0, factor0, xi0, X, I, r, force, mniter = 5000L, w = 1.0, tol = 1e-5) {
    .Call(`_fect_inter_fe_d_qr_ub`, Y, Y_fit0, FE0, factor0, xi0, X, I, r, force, mniter, w, tol)
}

qr_factor <- function(F, L) {
    .Call(`_fect_qr_factor`, F, L)
}

IND <- function(I) {
    .Call(`_fect_IND`, I)
}

subfe <- function(Y, X, I, intercept) {
    .Call(`_fect_subfe`, Y, X, I, intercept)
}

l_ub <- function(Y, F, I, r, force) {
    .Call(`_fect_l_ub`, Y, F, I, r, force)
}

f_ub <- function(Y, L, I, r, force) {
    .Call(`_fect_f_ub`, Y, L, I, r, force)
}

fe <- function(E, F_old_in, xi_old, force, r) {
    .Call(`_fect_fe`, E, F_old_in, xi_old, force, r)
}

fe_ub <- function(E, I, F_old, xi_old, force, r) {
    .Call(`_fect_fe_ub`, E, I, F_old, xi_old, force, r)
}

inter_fe_d <- function(Y, Y_fit0, FE0, X, r, force, mniter = 5000L, w = 1.0, tol = 1e-5) {
    .Call(`_fect_inter_fe_d`, Y, Y_fit0, FE0, X, r, force, mniter, w, tol)
}

inter_fe_d_ub <- function(Y, Y_fit0, FE0, X, I, r, force, mniter = 5000L, w = 1.0, tol = 1e-5) {
    .Call(`_fect_inter_fe_d_ub`, Y, Y_fit0, FE0, X, I, r, force, mniter, w, tol)
}

complex_fe_ub <- function(Y, Y0, X_covariates, X_extra_FE, X_Z, X_Q, X_gamma, X_kappa, Zgamma_id, kappaQ_id, I, W_in, beta0, r, force, tol = 1e-5, max_iter = 1000L, fit_init = NULL) {
    .Call(`_fect_complex_fe_ub`, Y, Y0, X_covariates, X_extra_FE, X_Z, X_Q, X_gamma, X_kappa, Zgamma_id, kappaQ_id, I, W_in, beta0, r, force, tol, max_iter, fit_init)
}

YY_adj <- function(YYYY, EEE, I, use_weight, W) {
    .Call(`_fect_YY_adj`, YYYY, EEE, I, use_weight, W)
}

Demean <- function(E, force, X_extra_FE, extra_FE_index_cache) {
    .Call(`_fect_Demean`, E, force, X_extra_FE, extra_FE_index_cache)
}

fixed_effects_part <- function(E, force, X_extra_FE, extra_FE_index_cache) {
    .Call(`_fect_fixed_effects_part`, E, force, X_extra_FE, extra_FE_index_cache)
}

Gamma <- function(E, Z, zzinv, gamma_t_group) {
    .Call(`_fect_Gamma`, E, Z, zzinv, gamma_t_group)
}

gamma_part <- function(E, Z, zzinv, gamma_t_group) {
    .Call(`_fect_gamma_part`, E, Z, zzinv, gamma_t_group)
}

Kappa <- function(E, Q, qqinv, kappa_i_group) {
    .Call(`_fect_Kappa`, E, Q, qqinv, kappa_i_group)
}

kappa_part <- function(E, Q, qqinv, kappa_i_group) {
    .Call(`_fect_kappa_part`, E, Q, qqinv, kappa_i_group)
}

Beta <- function(X, xxinv, E) {
    .Call(`_fect_Beta`, X, xxinv, E)
}

WBeta <- function(X, xwxinv, w, E) {
    .Call(`_fect_WBeta`, X, xwxinv, w, E)
}

beta_part <- function(E, XX, xxinv, W, use_weight) {
    .Call(`_fect_beta_part`, E, XX, xxinv, W, use_weight)
}

ife_part <- function(E, r) {
    .Call(`_fect_ife_part`, E, r)
}

cfe_iter <- function(XX, xxinv, X_extra_FE, X_Z, X_Q, X_gamma, X_kappa, Zgamma_id, kappaQ_id, Y, Y0, I, W, beta0, force, r, tolerate, max_iter, fit_init = NULL) {
    .Call(`_fect_cfe_iter`, XX, xxinv, X_extra_FE, X_Z, X_Q, X_gamma, X_kappa, Zgamma_id, kappaQ_id, Y, Y0, I, W, beta0, force, r, tolerate, max_iter, fit_init)
}

Y_demean <- function(Y, force) {
    .Call(`_fect_Y_demean`, Y, force)
}

Y_wdemean <- function(Y, W, force) {
    .Call(`_fect_Y_wdemean`, Y, W, force)
}

fe_add <- function(alpha_Y, xi_Y, mu_Y, T, N, force) {
    .Call(`_fect_fe_add`, alpha_Y, xi_Y, mu_Y, T, N, force)
}

panel_factor <- function(E, r) {
    .Call(`_fect_panel_factor`, E, r)
}

panel_FE <- function(E, lambda, hard) {
    .Call(`_fect_panel_FE`, E, lambda, hard)
}

ife <- function(E, force, mc, r, hard, lambda) {
    .Call(`_fect_ife`, E, force, mc, r, hard, lambda)
}

inter_fe <- function(Y, X, r, force, beta0_in, tol = 1e-5, max_iter = 500L) {
    .Call(`_fect_inter_fe`, Y, X, r, force, beta0_in, tol, max_iter)
}

inter_fe_ub <- function(Y, Y0, X, I, W_in, beta0, r, force, tol = 1e-5, max_iter = 1000L, fit_init = NULL) {
    .Call(`_fect_inter_fe_ub`, Y, Y0, X, I, W_in, beta0, r, force, tol, max_iter, fit_init)
}

fe_ad_iter <- function(Y, Y0, I, W, force, tolerate, max_iter = 500L) {
    .Call(`_fect_fe_ad_iter`, Y, Y0, I, W, force, tolerate, max_iter)
}

fe_ad_covar_iter <- function(XX, xxinv, Y, Y0, I, beta0, W, force, tolerate, max_iter = 500L) {
    .Call(`_fect_fe_ad_covar_iter`, XX, xxinv, Y, Y0, I, beta0, W, force, tolerate, max_iter)
}

fe_ad_inter_iter <- function(Y, Y0, I, W, force, mc, r, hard, lambda, tolerate, max_iter = 1000L, fit_init = NULL) {
    .Call(`_fect_fe_ad_inter_iter`, Y, Y0, I, W, force, mc, r, hard, lambda, tolerate, max_iter, fit_init)
}

fe_ad_inter_covar_iter <- function(XX, xxinv, Y, Y0, I, W, beta0, force, mc, r, hard, lambda, tolerate, max_iter = 1000L, fit_init = NULL) {
    .Call(`_fect_fe_ad_inter_covar_iter`, XX, xxinv, Y, Y0, I, W, beta0, force, mc, r, hard, lambda, tolerate, max_iter, fit_init)
}

beta_iter <- function(X, xxinv, Y, r, tolerate, beta0, max_iter) {
    .Call(`_fect_beta_iter`, X, xxinv, Y, r, tolerate, beta0, max_iter)
}

inter_fe_mc <- function(Y, Y0, X, I, W_in, beta0, r, lambda, force, tol = 1e-5, max_iter = 1000L, fit_init = NULL) {
    .Call(`_fect_inter_fe_mc`, Y, Y0, X, I, W_in, beta0, r, lambda, force, tol, max_iter, fit_init)
}

###################################################################
## Standard rolling-window cross-validation for rank selection
##
## Standalone user-facing wrapper. For each of `k` folds:
##   - samples `cv.prop` fraction of eligible units (controls + treated
##     pre-treatment); only sampled units carry a mask. Other units
##     stay fully observed and contribute training data at every
##     period --- preserves factor identification at the masked tails.
##   - per sampled unit, samples a random anchor time `t*` and:
##       - holds out `cv.nobs` observations starting at `t*` (scored for MSPE)
##       - drops `cv.buffer` observations immediately before `t*` from training
##         (buffer, attenuates AR-leakage from training to holdout)
##       - drops ALL observations from `t* + cv.nobs` onward from training
##         (the rolling-window step --- training cannot see the future of
##         the held-out block)
## For treated units, eligible times are the pre-treatment observations
## only (post-treatment is the imputation target, never masked).
## Refits fect with the masked panel for each candidate `r`, scores
## MSPE at the held-out positions via `fit$Y.ct.full`, and aggregates
## across folds. Applies the requested `cv.rule` to pick `r.cv`.
##
## Closes the forward-leakage channel that the existing
## `cv.method = "all_units" / "treated_units"` (random contiguous-block
## masking) leaves open at `cv.donut = 0 / 1` under serially correlated
## residuals: under rolling-window CV, the train/test boundary on the
## future side is closed by construction.
##
## Supports method = "ife" (IFE-EM, time.component.from = "notyettreated"),
## method = "gsynth" (GSC, time.component.from = "nevertreated"), and
## method = "cfe" (Complex Fixed Effects, time.component.from =
## "notyettreated"). All three paths populate Y.ct.full at masked
## positions, so MSPE scoring works uniformly. CFE-specific arguments
## (Z, gamma, Q, Q.type, kappa, extra index columns) are forwarded to
## the inner fect() call via `...`. CFE rolling CV picks `r` only;
## the user holds non-`r` CFE components fixed at their spec.
###################################################################

#' Rolling-window cross-validation for rank selection
#'
#' Picks the number of factors `r` for an interactive-fixed-effects
#' model via standard rolling-window cross-validation. For each of `k`
#' folds, a fraction `cv.prop` of eligible units (controls plus
#' treated pre-treatment) is sampled; only sampled units carry a mask.
#' For each sampled unit, a random anchor time `t*` is drawn and the
#' fold's training set excludes `t* - cv.buffer, ..., t* - 1` (buffer),
#' `t*, ..., t* + cv.nobs - 1` (the held-out, scored block), and
#' `t* + cv.nobs, ..., end_of_eligible(t)` (rolling-window future drop;
#' for treated units, `end_of_eligible` is the cell strictly before
#' treatment onset). MSPE is scored at the held-out block only and
#' averaged across folds.
#'
#' Per-fold unit sampling is required: masking every eligible unit at
#' the same time leaves no donor data at the masked time points and
#' breaks factor identification. Sampling `cv.prop` of units per fold
#' keeps unsampled units fully observed at all periods.
#'
#' This is the standard time-series CV design (cf. `forecast::tsCV`,
#' `tidymodels::sliding_window`, `caret::createTimeSlices`) adapted to
#' panel data: each sampled unit gets its own anchor per fold, drawn
#' uniformly from valid positions.
#'
#' @param formula A model formula, e.g. `Y ~ D + X1 + X2`.
#' @param data A long-format data frame.
#' @param index Character vector identifying the panel structure. For
#'   `method = "ife"` and `method = "gsynth"`, length 2:
#'   `c("unit", "time")`. For `method = "cfe"`, length >= 2: the first
#'   two entries are `c("unit", "time")` and any additional entries are
#'   extra grouping fixed-effect columns forwarded to the inner
#'   `fect()` call's `index =` argument.
#' @param method Estimator. One of `"ife"` (IFE-EM, internally
#'   `time.component.from = "notyettreated"`), `"gsynth"` (GSC,
#'   internally `time.component.from = "nevertreated"`), or `"cfe"`
#'   (Complex Fixed Effects, internally
#'   `time.component.from = "notyettreated"`). All three paths populate
#'   `Y.ct.full` at masked positions, so MSPE scoring works uniformly.
#'   For CFE, rolling CV picks `r` only; CFE-specific arguments
#'   (`Z`, `gamma`, `Q`, `Q.type`, `kappa`, extra index columns) are
#'   forwarded via `...` and held fixed at their user-supplied values.
#' @param r.max Largest candidate rank to evaluate. CV is run over
#'   `0:r.max`.
#' @param cv.nobs Length of the held-out (scored) block per unit per
#'   fold. Default 3.
#' @param cv.buffer Number of observations immediately BEFORE the
#'   held-out block to drop from training (the past-side buffer that
#'   attenuates AR-leakage). Default 1. Analogous to `cv.donut` in the
#'   existing `cv.method = "all_units"` / `"treated_units"` strategies,
#'   but applied only on the past side: the future side is dropped by
#'   construction.
#' @param k Number of folds. Each fold draws a fresh sample of units
#'   and a fresh set of per-unit anchors; the per-r MSPE is averaged
#'   across folds and the SE used by the `"1se"` rule reflects
#'   fold-to-fold variability. Default 20 (matches the default for
#'   the existing CV strategies).
#' @param cv.prop Fraction of eligible units sampled per fold. Only
#'   sampled units receive a mask in that fold; the rest stay fully
#'   observed and contribute training data at every period. Default
#'   0.1 (paired with `k = 20` for ~2x coverage of every eligible
#'   unit across folds). Across `k` folds, every eligible unit lands
#'   in the holdout roughly `k * cv.prop` times in expectation. Must
#'   satisfy `0 < cv.prop <= 1`. On small panels (n_eligible < 30)
#'   consider raising further, since per-fold MSPE precision scales
#'   with `cv.prop * n_eligible * cv.nobs`.
#' @param cv.rule Rule for picking `r` from the MSPE curve: `"1se"`
#'   (default), `"min"`, or `"1pct"`.
#' @param min.T0 Minimum observations required strictly before the
#'   anchor. Sets the lower bound on valid anchor positions. Default 5.
#' @param force One of `"none"`, `"unit"`, `"time"`, `"two-way"`. Default
#'   `"unit"`.
#' @param seed Optional integer base seed; per-fold seeds derive from
#'   `seed + fold_id` for reproducibility. Default `NULL` (use the
#'   ambient RNG).
#' @param verbose If TRUE (default), print per-fold per-r MSPE.
#' @param ... Additional arguments forwarded to `fect()`. For
#'   `method = "cfe"`, the user holds CFE structural arguments (`Z`,
#'   `gamma`, `Q`, `Q.type`, `Q.bspline.degree`, `kappa`, etc.) fixed
#'   at their spec via `...`; rolling CV varies only `r`.
#'
#' @return List with components:
#'   - `r.cv`: chosen rank.
#'   - `cv.rule`: rule applied.
#'   - `mspe`: data.frame of per-r MSPE (averaged across folds), SE
#'     across folds, and held-out cell counts.
#'   - `mspe.per.fold`: r-by-k matrix of per-fold MSPE.
#'   - `k`, `cv.nobs`, `cv.buffer`, `cv.prop`: parameters used.
#'   - `n.units.masked`: distinct units that contributed to at least
#'     one fold's holdout.
#'
#' @examples
#' \dontrun{
#'   library(fect)
#'   data(simdata)
#'   res <- r.cv.rolling(Y ~ D, data = simdata, index = c("id", "time"),
#'                       method = "ife", r.max = 5,
#'                       cv.nobs = 3, cv.buffer = 1, k = 20)
#'   res$r.cv
#'   ## then use the chosen r in a CV-disabled fit:
#'   fit <- fect(Y ~ D, data = simdata, index = c("id", "time"),
#'               method = "ife", time.component.from = "notyettreated",
#'               CV = FALSE, r = res$r.cv, se = TRUE)
#' }
#'
#' @export
r.cv.rolling <- function(formula,
                          data,
                          index,
                          method = c("ife", "gsynth", "cfe"),
                          r.max = 5L,
                          cv.nobs = 3L,
                          cv.buffer = 1L,
                          k = 20L,
                          cv.prop = 0.1,
                          cv.rule = c("1se", "min", "1pct"),
                          min.T0 = 5L,
                          force = "unit",
                          seed = NULL,
                          verbose = TRUE,
                          ...) {
    cv.rule <- match.arg(cv.rule)
    method  <- match.arg(method)
    fect_method <- method
    ## CFE on the notyettreated path also populates Y.ct.full at masked
    ## cells, so MSPE scoring works uniformly. Only "gsynth" uses
    ## the nevertreated factor estimation sample.
    fect_tcf <- if (identical(method, "gsynth")) "nevertreated" else "notyettreated"

    r.max     <- as.integer(r.max);     if (r.max     < 0L) stop("r.max must be >= 0.")
    cv.nobs   <- as.integer(cv.nobs);   if (cv.nobs   < 1L) stop("cv.nobs must be >= 1.")
    cv.buffer <- as.integer(cv.buffer); if (cv.buffer < 0L) stop("cv.buffer must be >= 0.")
    k         <- as.integer(k);         if (k         < 1L) stop("k must be >= 1.")
    min.T0    <- as.integer(min.T0);    if (min.T0    < 1L) stop("min.T0 must be >= 1.")
    cv.prop   <- as.numeric(cv.prop)
    if (!is.finite(cv.prop) || cv.prop <= 0 || cv.prop > 1) {
        stop("cv.prop must satisfy 0 < cv.prop <= 1.")
    }

    ## CFE may carry extra index columns (additional grouping FEs); IFE
    ## and gsynth still require length-2 (unit, time).
    if (length(index) < 2L) {
        stop("index must be a character vector of length >= 2 ",
             "(c(unit, time) for ife/gsynth; c(unit, time, ...) for cfe).")
    }
    if (!all(index %in% colnames(data))) {
        stop("index columns not found in data: ",
             paste(setdiff(index, colnames(data)), collapse = ", "))
    }
    Yname <- as.character(formula[[2L]])
    if (!Yname %in% colnames(data)) {
        stop("Outcome '", Yname, "' not found in data.")
    }

    ## Identify treated units and their treatment-onset times.
    ## Treated units' eligible holdout window is the pre-treatment
    ## observed times only (post-treatment is the imputation target,
    ## never masked).
    Dname <- NULL
    rhs_terms <- attr(stats::terms(formula), "term.labels")
    if (length(rhs_terms) >= 1L) Dname <- rhs_terms[1L]
    treat_onset <- list()  # named: unit -> first treated time
    if (!is.null(Dname) && Dname %in% colnames(data)) {
        for (u in unique(as.character(data[[index[1L]]]))) {
            rows <- which(as.character(data[[index[1L]]]) == u)
            d_u <- data[[Dname]][rows]
            t_u <- data[[index[2L]]][rows]
            treated_rows <- which(d_u >= 1)
            if (length(treated_rows) > 0L) {
                treat_onset[[u]] <- min(t_u[treated_rows], na.rm = TRUE)
            }
        }
    }

    ## Build per-unit eligible-time vectors:
    ##   - controls: all observed times
    ##   - treated:  observed times strictly before treatment onset
    elig_obs_times <- by(data, data[[index[1L]]], function(d) {
        u <- as.character(d[[index[1L]]][1L])
        t_all <- sort(d[[index[2L]]])
        if (!is.null(treat_onset[[u]])) t_all[t_all < treat_onset[[u]]]
        else                            t_all
    })
    ## Eligible: at least min.T0 + cv.nobs observations.
    eligible <- vapply(elig_obs_times, length,
                       integer(1)) >= (min.T0 + cv.nobs)
    elig_obs_times <- elig_obs_times[eligible]
    if (length(elig_obs_times) == 0L) {
        stop("r.cv.rolling: no eligible units have enough pre-treatment ",
             "observations (need >= min.T0 + cv.nobs = ",
             min.T0 + cv.nobs, ").")
    }

    n_eligible_units <- length(elig_obs_times)
    n_sample_per_fold <- max(1L, as.integer(round(cv.prop * n_eligible_units)))
    if (isTRUE(verbose)) {
        message(sprintf(
            "r.cv.rolling: %d eligible units (controls + treated pre-treatment); k = %d folds; cv.nobs = %d, cv.buffer = %d, cv.prop = %.3g (-> %d units sampled per fold).",
            n_eligible_units, k, cv.nobs, cv.buffer, cv.prop, n_sample_per_fold
        ))
    }

    ## --------------------------------------------------------------------
    ## Per-fold per-r CV loop.
    ## --------------------------------------------------------------------
    r.grid <- 0L:r.max
    fold_mspe <- matrix(NA_real_, nrow = length(r.grid), ncol = k,
                        dimnames = list(paste0("r", r.grid),
                                        paste0("fold", seq_len(k))))
    fold_n    <- matrix(0L, nrow = length(r.grid), ncol = k,
                        dimnames = dimnames(fold_mspe))
    units_seen <- character(0)
    data_keys  <- paste(data[[index[1L]]], data[[index[2L]]], sep = "_")

    for (fold_id in seq_len(k)) {
        if (!is.null(seed)) set.seed(as.integer(seed) + fold_id)

        ## Sample cv.prop fraction of eligible units to mask in this fold.
        ## Unsampled units stay fully observed and contribute training
        ## data at every period (preserves factor identification at the
        ## masked time points).
        sampled_idx <- sort(sample.int(n_eligible_units, n_sample_per_fold))
        sampled_units <- names(elig_obs_times)[sampled_idx]

        ## Sample an anchor for each sampled unit.
        anchor_records <- lapply(sampled_units, function(u) {
            obs_t <- elig_obs_times[[u]]
            n_obs <- length(obs_t)
            ## Anchor index in obs_t such that:
            ##   - >= min.T0 obs before  -> a_idx >= min.T0 + 1
            ##   - >= cv.nobs obs from a -> a_idx <= n_obs - cv.nobs + 1
            ## For treated units, obs_t already excludes post-treatment
            ## cells, so the holdout block is guaranteed to stay in the
            ## pre-treatment window.
            valid <- seq.int(min.T0 + 1L, n_obs - cv.nobs + 1L)
            if (length(valid) == 0L) return(NULL)
            a_idx <- valid[sample.int(length(valid), 1L)]
            holdout_t <- obs_t[a_idx:(a_idx + cv.nobs - 1L)]
            buf_t <- if (cv.buffer > 0L) {
                bs <- max(1L, a_idx - cv.buffer)
                obs_t[bs:(a_idx - 1L)]
            } else integer(0)
            drop_t <- if ((a_idx + cv.nobs) <= n_obs) {
                obs_t[(a_idx + cv.nobs):n_obs]
            } else integer(0)
            list(unit = u,
                 holdout_t = holdout_t,
                 buf_t = buf_t,
                 drop_t = drop_t)
        })
        anchor_records <- anchor_records[!vapply(anchor_records, is.null,
                                                 logical(1))]
        if (length(anchor_records) == 0L) {
            if (isTRUE(verbose)) message(sprintf("  fold %d: no valid anchors; skipped.", fold_id))
            next
        }

        ## Build mask key sets.
        scored_keys <- unlist(lapply(anchor_records, function(r) {
            paste(r$unit, r$holdout_t, sep = "_")
        }))
        buffer_keys <- unlist(lapply(anchor_records, function(r) {
            if (length(r$buf_t) > 0L) paste(r$unit, r$buf_t, sep = "_") else character(0)
        }))
        dropped_keys <- unlist(lapply(anchor_records, function(r) {
            if (length(r$drop_t) > 0L) paste(r$unit, r$drop_t, sep = "_") else character(0)
        }))
        all_mask_keys <- c(scored_keys, buffer_keys, dropped_keys)
        units_seen <- union(units_seen, vapply(anchor_records,
                                               function(r) r$unit,
                                               character(1)))

        ## Apply mask: set Y to NA at all masked cells.
        data_masked <- data
        data_masked[[Yname]][data_keys %in% all_mask_keys] <- NA_real_

        ## Score positions and observed Y.
        score_rows <- which(data_keys %in% scored_keys)
        Y_obs_score <- data[[Yname]][score_rows]

        ## Per-r fits for this fold.
        for (idx in seq_along(r.grid)) {
            r_try <- r.grid[idx]
            fit <- tryCatch(
                suppressWarnings(suppressMessages(fect(
                    formula = formula, data = data_masked, index = index,
                    method = fect_method, time.component.from = fect_tcf,
                    force = force, CV = FALSE, r = r_try,
                    min.T0 = min.T0, se = FALSE, na.rm = FALSE, ...))),
                error = function(e) {
                    if (isTRUE(verbose)) message(sprintf(
                        "  fold %d, r=%d: fit error: %s",
                        fold_id, r_try, conditionMessage(e)))
                    NULL
                })
            if (is.null(fit) || is.null(fit$Y.ct.full)) next

            ## Match by raw labels: fit$id and fit$rawtime are the unit/time
            ## labels in fect's coordinate system (integer for ife/gsynth, can
            ## be character/integer for cfe). Match against the same column
            ## values from `data` and accept whichever type comes back.
            unit_ord <- fit$id
            year_ord <- fit$rawtime
            unit_ord_int <- suppressWarnings(as.integer(unit_ord))
            year_ord_int <- suppressWarnings(as.integer(year_ord))
            unit_match <- if (any(is.na(unit_ord_int))) unit_ord else unit_ord_int
            year_match <- if (any(is.na(year_ord_int))) year_ord else year_ord_int
            Y_pred <- numeric(length(score_rows))
            for (kk in seq_along(score_rows)) {
                rr <- score_rows[kk]
                ui <- match(data[[index[1L]]][rr], unit_match)
                ti <- match(data[[index[2L]]][rr], year_match)
                Y_pred[kk] <- if (!is.na(ui) && !is.na(ti)) {
                    fit$Y.ct.full[ti, ui]
                } else NA_real_
            }
            e2 <- (Y_obs_score - Y_pred)^2
            e2 <- e2[is.finite(e2)]
            if (length(e2) == 0L) next
            fold_mspe[idx, fold_id] <- mean(e2)
            fold_n[idx, fold_id]    <- length(e2)
            if (isTRUE(verbose)) {
                message(sprintf("  fold %d, r=%d: MSPE=%.4g (n=%d)",
                                fold_id, r_try, fold_mspe[idx, fold_id],
                                length(e2)))
            }
        }
    }

    ## Aggregate per-r across folds.
    mspe_per_r <- rowMeans(fold_mspe, na.rm = TRUE)
    n_folds_used <- rowSums(!is.na(fold_mspe))
    se_per_r <- vapply(seq_along(r.grid), function(i) {
        x <- fold_mspe[i, ]
        x <- x[is.finite(x)]
        if (length(x) <= 1L) NA_real_
        else stats::sd(x) / sqrt(length(x))
    }, numeric(1))
    n_per_r <- rowSums(fold_n)

    chosen_idx <- .fect_apply_cv_rule(mspe_per_r, ses = se_per_r,
                                      rule = cv.rule)
    if (is.na(chosen_idx)) {
        stop("r.cv.rolling: all candidate fits failed across all folds.")
    }
    r_cv <- r.grid[chosen_idx]
    if (isTRUE(verbose)) {
        message(sprintf(
            "r.cv.rolling: r.cv = %d (rule = %s; aggregated over %d folds).",
            r_cv, cv.rule, k
        ))
    }

    list(r.cv          = as.integer(r_cv),
         cv.rule       = cv.rule,
         mspe          = data.frame(r = r.grid,
                                    mspe = mspe_per_r,
                                    se = se_per_r,
                                    n_holdout = n_per_r,
                                    n_folds_used = n_folds_used),
         mspe.per.fold = fold_mspe,
         k             = k,
         cv.nobs       = cv.nobs,
         cv.buffer     = cv.buffer,
         cv.prop       = cv.prop,
         n.units.masked = length(units_seen))
}

## ---------------------------------------------------------------
## Regression test: standard rolling-window CV in r.cv.rolling.
##
## After the rewrite, r.cv.rolling implements the standard time-series
## rolling-window design (random anchors per unit per fold, with
## training restricted to observations strictly before each anchor;
## an optional cv.buffer excludes a buffer immediately before the anchor;
## all observations after anchor + cv.nobs are dropped from training).
##
## Tests cover: new parameters (cv.buffer, k, seed), reproducibility under
## a fixed seed, fold-aggregated MSPE structure, identical CV behavior
## across method = "ife" and method = "gsynth".
## ---------------------------------------------------------------

.make_unbalanced_panel <- function(seed = 11) {
  set.seed(seed)
  N  <- 30L; T <- 25L
  ids   <- rep(seq_len(N), each = T)
  times <- rep(seq_len(T), N)
  F1 <- rnorm(T); L1 <- rnorm(N)
  err <- rnorm(N * T, sd = 0.3)
  mu  <- rep(L1, each = T) * rep(F1, N)
  Y   <- mu + err
  treated      <- 26L:30L
  treat_starts <- c(18L, 19L, 20L, 21L, 22L)
  D <- rep(0L, N * T)
  for (k in seq_along(treated)) {
    u <- treated[k]
    D[ids == u & times >= treat_starts[k]] <- 1L
  }
  Y <- Y + D * 1.5
  df <- data.frame(id = ids, time = times, Y = Y, D = D)
  drop_keys <- c(
    paste(27L, c(24L, 25L), sep = "_"),
    paste(28L, c(22L, 23L, 24L, 25L), sep = "_"),
    paste(1L,  c(23L, 24L, 25L), sep = "_"),
    paste(2L,  c(24L, 25L), sep = "_")
  )
  df_keys <- paste(df$id, df$time, sep = "_")
  df[!df_keys %in% drop_keys, ]
}

test_that("k and cv.buffer parameters are honored", {

  skip_on_cran()
  df <- .make_unbalanced_panel(seed = 11)

  set.seed(99)
  res_k5 <- suppressWarnings(suppressMessages(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r.max = 2, cv.nobs = 2, cv.buffer = 1,
      k = 5L, seed = 99,
      verbose = FALSE
    )
  ))
  expect_equal(res_k5$k, 5L)
  expect_equal(res_k5$cv.buffer, 1L)
  expect_equal(res_k5$cv.nobs, 2L)
  expect_equal(dim(res_k5$mspe.per.fold),
               c(3L, 5L))    # (r.max + 1) rows x k columns
  expect_true(all(res_k5$mspe$n_folds_used <= 5L))
})

test_that("seed gives reproducible per-fold MSPE", {

  skip_on_cran()
  df <- .make_unbalanced_panel(seed = 11)

  res_a <- suppressWarnings(suppressMessages(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r.max = 2, cv.nobs = 2, cv.buffer = 1,
      k = 3L, seed = 42L, verbose = FALSE
    )
  ))
  res_b <- suppressWarnings(suppressMessages(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r.max = 2, cv.nobs = 2, cv.buffer = 1,
      k = 3L, seed = 42L, verbose = FALSE
    )
  ))
  ## The seed is per-fold; identical inputs should give identical fold MSPE.
  expect_equal(res_a$mspe.per.fold, res_b$mspe.per.fold)
  expect_equal(res_a$r.cv, res_b$r.cv)
})

test_that("rolling-window CV identical for method = 'ife' and 'gsynth' inner-call paths", {

  skip_on_cran()
  ## The two methods route to fect with the same masking logic;
  ## with the same seed they should sample the same per-fold anchors.
  ## Numerical results may differ because of different inner estimators,
  ## but the anchor sets and units_seen should match.
  df <- .make_unbalanced_panel(seed = 11)

  res_ife <- suppressWarnings(suppressMessages(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r.max = 1, cv.nobs = 2, cv.buffer = 0,
      k = 2L, seed = 7L, verbose = FALSE
    )
  ))
  res_gsc <- suppressWarnings(suppressMessages(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "gsynth", r.max = 1, cv.nobs = 2, cv.buffer = 0,
      k = 2L, seed = 7L, verbose = FALSE
    )
  ))
  expect_equal(res_ife$n.units.masked, res_gsc$n.units.masked)
  expect_equal(res_ife$k, res_gsc$k)
  expect_equal(res_ife$cv.nobs, res_gsc$cv.nobs)
  expect_equal(res_ife$cv.buffer, res_gsc$cv.buffer)
})

test_that("invalid arguments are rejected", {

  skip_on_cran()
  df <- .make_unbalanced_panel(seed = 11)

  expect_error(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r.max = 1, cv.nobs = 0, verbose = FALSE
    ),
    regexp = "cv\\.nobs must be"
  )
  expect_error(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r.max = 1, cv.buffer = -1, verbose = FALSE
    ),
    regexp = "cv\\.buffer must be"
  )
  expect_error(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r.max = 1, k = 0, verbose = FALSE
    ),
    regexp = "k must be"
  )
  expect_error(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r.max = 1, cv.prop = 0, verbose = FALSE
    ),
    regexp = "cv\\.prop"
  )
  expect_error(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r.max = 1, cv.prop = 1.5, verbose = FALSE
    ),
    regexp = "cv\\.prop"
  )
})

test_that("cv.prop controls per-fold unit sampling", {

  skip_on_cran()
  ## With 30 units (25 controls + 5 treated, all 5 treated have at
  ## least min.T0 + cv.nobs pre-treatment cells in the fixture), the
  ## eligible pool covers controls AND treated pre-treatment.
  df <- .make_unbalanced_panel(seed = 11)

  ## Small cv.prop -> few units sampled per fold; capped at >= 1.
  res_small <- suppressWarnings(suppressMessages(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r.max = 1, cv.nobs = 2, cv.buffer = 1,
      k = 2L, cv.prop = 0.05, seed = 17L, verbose = FALSE
    )
  ))
  expect_equal(res_small$cv.prop, 0.05)
  ## At cv.prop = 0.05 with ~30 eligible units -> max(1, round(1.5)) = 2
  ## sampled per fold; with k = 2 folds, units_seen <= 4.
  expect_lte(res_small$n.units.masked, 4L)
  expect_gte(res_small$n.units.masked, 1L)

  ## Full cv.prop -> every eligible unit sampled every fold.
  res_full <- suppressWarnings(suppressMessages(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r.max = 1, cv.nobs = 2, cv.buffer = 1,
      k = 2L, cv.prop = 1.0, seed = 17L, verbose = FALSE
    )
  ))
  expect_equal(res_full$cv.prop, 1.0)
  ## n.units.masked at cv.prop = 1.0 should exceed n.units.masked at
  ## cv.prop = 0.05 by a wide margin.
  expect_gt(res_full$n.units.masked, res_small$n.units.masked)
})

test_that("treated pre-treatment cells are eligible for masking", {

  skip_on_cran()
  ## With cv.prop = 1.0, every eligible unit gets masked. We then
  ## check that at least one treated unit appears in the holdout
  ## (would be impossible under the old controls-only design).
  df <- .make_unbalanced_panel(seed = 11)
  treated_ids <- 26:30  # from .make_unbalanced_panel

  res <- suppressWarnings(suppressMessages(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", r.max = 1, cv.nobs = 2, cv.buffer = 1,
      k = 2L, cv.prop = 1.0, seed = 21L, verbose = FALSE
    )
  ))
  ## n.units.masked should include at least one treated unit (their
  ## pre-treatment cells are eligible). Total eligible >= 25
  ## controls + the treated units with sufficient pre-period.
  expect_gt(res$n.units.masked, 25L)
})

## ---------------------------------------------------------------
## Regression test: r.cv.rolling supports both method = "ife"
## (IFE-EM, time.component.from = "notyettreated") and
## method = "gsynth" (GSC, time.component.from = "nevertreated").
##
## Pre-fold-in (PR #119 as merged): r.cv.rolling guarded on
## method == "ife" and hard-coded the inner fect call to
## time.component.from = "notyettreated", because the GSC path did not
## populate Y.ct.full at masked control positions.
##
## Stage 2 (committed separately) fixed Y.ct.full on the GSC path.
## Stage 3 fold-in (this branch):
##   - Drops the method == "ife" guard in r.cv.rolling.
##   - Accepts method = "gsynth" and routes the inner fect call to
##     time.component.from = "nevertreated".
##
## The bundled fect demo datasets (simdata, simgsynth) are balanced
## panels: masking the last cv.nobs observations of every control
## unit drops those time points entirely, so cv.sample.rolling cannot
## exercise method = "gsynth" cleanly on them. We synthesize a small
## unbalanced panel inline for the gsynth test, and use simdata for
## the IFE-EM test (works because simdata has many ineligible /
## treated units that retain data at the masked time points).
## ---------------------------------------------------------------

.make_unbalanced_panel <- function(seed = 11) {
  set.seed(seed)
  N  <- 30L; T <- 25L
  ids   <- rep(seq_len(N), each = T)
  times <- rep(seq_len(T), N)
  F1 <- rnorm(T); L1 <- rnorm(N)
  err <- rnorm(N * T, sd = 0.3)
  mu  <- rep(L1, each = T) * rep(F1, N)
  Y   <- mu + err
  treated      <- 26L:30L
  treat_starts <- c(18L, 19L, 20L, 21L, 22L)
  D <- rep(0L, N * T)
  for (k in seq_along(treated)) {
    u <- treated[k]
    D[ids == u & times >= treat_starts[k]] <- 1L
  }
  Y <- Y + D * 1.5
  df <- data.frame(id = ids, time = times, Y = Y, D = D)
  ## Drop trailing observations from a few units to make panel unbalanced
  ## so the rolling-CV mask doesn't drop entire time points.
  drop_keys <- c(
    paste(27L, c(24L, 25L), sep = "_"),
    paste(28L, c(22L, 23L, 24L, 25L), sep = "_"),
    paste(1L,  c(23L, 24L, 25L), sep = "_"),
    paste(2L,  c(24L, 25L), sep = "_")
  )
  df_keys <- paste(df$id, df$time, sep = "_")
  df[!df_keys %in% drop_keys, ]
}

test_that("r.cv.rolling works for method = 'ife' (IFE-EM)", {

  skip_on_cran()

  if (!exists("simdata", inherits = TRUE)) {
    suppressWarnings(try(utils::data("simdata", package = "fect"),
                         silent = TRUE))
  }
  skip_if_not(exists("simdata", inherits = TRUE),
              "simdata dataset not available")

  set.seed(42)
  res <- suppressWarnings(suppressMessages(
    fect::r.cv.rolling(
      Y ~ D, data = simdata, index = c("id", "time"),
      method  = "ife",
      r.max   = 3,
      cv.nobs = 3,
      cv.rule = "1se",
      verbose = FALSE
    )
  ))

  expect_true(is.list(res))
  expect_true(is.integer(res$r.cv))
  expect_true(res$r.cv >= 0L && res$r.cv <= 3L)
  expect_true(is.data.frame(res$mspe))
  expect_equal(nrow(res$mspe), 4L)
  expect_true(any(is.finite(res$mspe$mspe)))
})

test_that("r.cv.rolling works for method = 'gsynth' (GSC)", {

  skip_on_cran()

  df <- .make_unbalanced_panel(seed = 11)

  set.seed(42)
  res <- suppressWarnings(suppressMessages(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method  = "gsynth",
      r.max   = 2,
      cv.nobs = 2,
      cv.rule = "1se",
      force   = "two-way",
      verbose = FALSE
    )
  ))

  expect_true(is.list(res))
  expect_true(is.integer(res$r.cv))
  expect_true(res$r.cv >= 0L && res$r.cv <= 2L)
  expect_true(is.data.frame(res$mspe))
  expect_equal(nrow(res$mspe), 3L)
  ## At least one r must have produced a finite MSPE (otherwise the GSC
  ## inner fits silently fell through and the dispatch is broken).
  expect_true(any(is.finite(res$mspe$mspe)))
})

test_that("r.cv.rolling rejects unknown method", {

  skip_on_cran()

  df <- .make_unbalanced_panel(seed = 11)

  expect_error(
    fect::r.cv.rolling(
      Y ~ D, data = df, index = c("id", "time"),
      method  = "mc",
      r.max   = 1,
      verbose = FALSE
    ),
    regexp = "should be one of"
  )
})

## ---------------------------------------------------------------
## CFE coverage: r.cv.rolling(method = "cfe", ...) accepts CFE
## structural args via `...` (Z, gamma, Q, kappa, extra index
## columns), holds them fixed, and varies only `r`.
## ---------------------------------------------------------------

test_that("r.cv.rolling works for method = 'cfe' with extra-FE index", {

  skip_on_cran()

  if (!exists("sim_region", inherits = TRUE)) {
    suppressWarnings(try(utils::data("sim_region", package = "fect"),
                         silent = TRUE))
  }
  skip_if_not(exists("sim_region", inherits = TRUE),
              "sim_region dataset not available")

  set.seed(42)
  res <- suppressWarnings(suppressMessages(
    fect::r.cv.rolling(
      Y ~ D, data = sim_region,
      index   = c("id", "time", "region_time"),
      method  = "cfe",
      r.max   = 1,
      cv.nobs = 2,
      cv.buffer = 1,
      k = 2,
      cv.rule = "1se",
      force   = "two-way",
      verbose = FALSE
    )
  ))

  expect_true(is.list(res))
  expect_true(is.integer(res$r.cv))
  expect_true(res$r.cv >= 0L && res$r.cv <= 1L)
  expect_true(is.data.frame(res$mspe))
  expect_equal(nrow(res$mspe), 2L)
  expect_true(any(is.finite(res$mspe$mspe)))
})

test_that("r.cv.rolling works for method = 'cfe' with Z/gamma", {

  skip_on_cran()

  if (!exists("simdata", inherits = TRUE)) {
    suppressWarnings(try(utils::data("simdata", package = "fect"),
                         silent = TRUE))
  }
  skip_if_not(exists("simdata", inherits = TRUE),
              "simdata dataset not available")
  ## simdata must carry an L1 column for the Z/gamma example
  skip_if_not("L1" %in% colnames(simdata),
              "simdata does not have L1 column for Z/gamma example")

  set.seed(42)
  ## CFE's `gamma` argument expects a column name (the time-group index
  ## for time-group-varying coefficients), not a numeric scalar. Build a
  ## simple gamma_t column by binning time into 3 groups.
  simdata_local <- simdata
  simdata_local$gamma_t <- ((simdata_local$time - 1L) %/%
                            ceiling(max(simdata_local$time) / 3L)) + 1L
  res <- suppressWarnings(suppressMessages(
    fect::r.cv.rolling(
      Y ~ D, data = simdata_local, index = c("id", "time"),
      method  = "cfe",
      r.max   = 1,
      cv.nobs = 2,
      cv.buffer = 1,
      k = 2,
      Z = "L1", gamma = "gamma_t",
      cv.rule = "1se",
      force   = "two-way",
      verbose = FALSE
    )
  ))

  expect_true(is.list(res))
  expect_true(is.integer(res$r.cv))
  expect_true(res$r.cv >= 0L && res$r.cv <= 1L)
  expect_true(is.data.frame(res$mspe))
  expect_equal(nrow(res$mspe), 2L)
  expect_true(any(is.finite(res$mspe$mspe)))
})

test_that("r.cv.rolling accepts length-3 index for method = 'ife' (regression)", {

  skip_on_cran()

  ## Regression test for the index length check: length-3 index used to
  ## error on `length(index) != 2L`. We don't actually feed a third FE
  ## column to IFE (IFE doesn't use it), but the length check should
  ## not block dispatch when the user mistakenly passes one. The
  ## downstream fect() call may or may not accept it; here we just
  ## confirm the helper doesn't preemptively error.
  if (!exists("simdata", inherits = TRUE)) {
    suppressWarnings(try(utils::data("simdata", package = "fect"),
                         silent = TRUE))
  }
  skip_if_not(exists("simdata", inherits = TRUE),
              "simdata dataset not available")

  ## We only verify that the input validation accepts length-3 index,
  ## not that the inner fect() call succeeds (IFE doesn't use index[3]).
  ## So we trap any inner error here.
  set.seed(42)
  err <- tryCatch({
    suppressWarnings(suppressMessages(
      fect::r.cv.rolling(
        Y ~ D, data = simdata, index = c("id", "time"),
        method  = "ife",
        r.max   = 1,
        cv.nobs = 2,
        k = 2,
        verbose = FALSE
      )
    ))
    NULL
  }, error = function(e) conditionMessage(e))
  ## Length-2 index for IFE: should still work.
  expect_null(err)
})

##---------------------------------------##
## Cumulative treatment effect           ##
##---------------------------------------##
att.cumu <- function(x, ## a fect object
                     period = NULL, ## range, length = 2
                     weighted = TRUE, ## weighted cumulative effect
                     alpha = 0.05,
                     type = "on", ## switch on or switch off
                     plot = FALSE
                    ) {

    ## v2.4.0: soft-deprecated in favor of estimand(). Continues to work
    ## byte-identical to today; one-time-per-session message points
    ## users at the unified API. Removal not before v3.0.0. Suppressed
    ## when called internally from .compute_att_cumu_overall().
    .fect_deprecation_message_once("att.cumu")

    end <- catt <- CI.lower <- CI.upper <- NULL
    if (is.null(period)) {
        period <- c(1, max(x$time))
    }

    if (length(period) != 2) {
        stop(" Length of \"period\" should equal 2.\n")
    }

    if(period[1] > period[2]){
        stop("period[1] should be smaller than period[2].\n")
    }

    ## start point
    se <- 0
    time <- x$time
    att <- x$att
    est.att <- x$est.att

    if (!is.null(est.att) & sum(abs(c(x$att.boot)),na.rm = TRUE) != 0) {
        se <- 1
    }

    if (type == "off") {
        time <- x$time.off
        att <- x$att.off
        est.att <- x$est.att.off
        if (is.null(time)) {
            stop("No switch-off effect.")
        }
    }

    p.start <- max(period[1], min(time))
    p.end <- min(period[2], max(time))

    pl <- p.end - p.start + 1 

    if(se == 0){
        cumu.att <- matrix(NA, pl, 3)
        cumu.att[1, c(1,2)] <- p.start
        cumu.att[1, 3] <- att[which(time == p.start)]
        colnames(cumu.att) <- c("start", "end", "catt")
    }
    else if (se == 1) {
        cumu.att <- matrix(NA, pl, 7)
        cumu.att[1, c(1,2)] <- p.start
        cumu.att[1, 3] <- att[which(time == p.start)]
        cumu.att[1, 4:7] <- est.att[which(time == p.start), c("S.E.", "CI.lower", "CI.upper", "p.value")]
        colnames(cumu.att) <- c("start", "end", "catt", "S.E.", "CI.lower", "CI.upper", "p.value")
    }

    for(i in 2:pl) {
        cumu.att[i, ] <- att.cumu.sub(x, c(p.start, p.start + i - 1), weighted, alpha, type)
    }

    ## plot 
    if (plot == TRUE) {
        data <- as.data.frame(cumu.att)
        data <- data[, -1] ## remove the first column
        data <- rbind(0, data)

        lp.start <- p.start - 1

        data[1, 1] <- lp.start

        p <- ggplot(data)
        p <- p + geom_hline(yintercept = 0, colour="#FFFFFF", linewidth = 1)
        p <- p + geom_line(data = data, aes(end, catt), linewidth = 1.2)
        if (se == 1) {
            p <- p + geom_ribbon(data = data, aes(x = end, ymin = CI.lower, ymax = CI.upper), alpha = 0.2) 
        }

        p <- p + xlab("Relative Period") +  ylab("Estimated Cumulative Effects")

        suppressWarnings(print(p))
    }

    return(cumu.att)
}





att.cumu.sub <- function(x, ## a fect object
                         period, ## range, length = 2
                         weighted = TRUE, ## weighted cumulative effect
                         alpha = 0.05, 
                         type = "on" ## switch on or switch off
                        ) {
  
    if (length(period) != 2) {
        stop(" Length of \"period\" should equal 2.")
    }
    
    ## period length
    pl <- period[2] - period[1] + 1  
    se <- 0

    if (type == "on") {
        time <- x$time
        att <- x$att
        count <- x$count

        if (!is.null(x$est.att)) {
            se <- 1
            att.boot <- x$att.boot
            count.boot <- x$att.count.boot
        }
    } else {
        time <- x$time.off
        att <- x$att.off
        count <- x$count.off

        if (is.null(time)) {
            stop("No switch-off effect.")
        }

        if (!is.null(x$est.att.off)) {
            se <- 1
            att.boot <- x$att.off.boot
            count.boot <- x$att.off.count.boot
        }
    }


    att.pos <- which(time>=period[1]&time<=period[2])
    att <- att[att.pos]
    count <- count[att.pos]
    rm.pos1 <- which(is.na(att))
    rm.pos2 <- which(is.na(count))
    if (NA %in% att | NA %in% count) {
        att <- att[-c(rm.pos1, rm.pos2)]
        count <- count[-c(rm.pos1, rm.pos2)]
    }
    catt <- sum(att*count*(length(count)/sum(count)))


    if (se == 1) {
        att.boot <- as.matrix(att.boot[att.pos,])
        count.boot <- as.matrix(count.boot[att.pos,])
        
        nboots <- dim(att.boot)[2]
        catt.boot <- rep(NA, nboots)
        
        for (i in 1:nboots) {
            att.sub <- att.boot[,i]
            count.sub <- count.boot[,i]
            rm.pos1 <- which(is.na(att.sub))
            rm.pos2 <- which(is.na(count.sub))
            if (NA %in% att.sub | NA %in% count.sub) {
                att.sub <- att.sub[-c(rm.pos1, rm.pos2)]
                count.sub <- count.sub[-c(rm.pos1, rm.pos2)]
            }
            catt.boot[i] <- sum(att.sub*count.sub*(length(count.sub)/sum(count.sub)))*(pl)/(length(count.sub))
        }

        catt.se <- sd(catt.boot, na.rm = TRUE)
        catt.ci <- quantile(catt.boot, c(alpha/2, 1 - alpha/2), na.rm = TRUE)
        catt.p <- get.pvalue(catt.boot)
    }
  
    if (se == 0) {
        result <- t(matrix(c(period, catt)))
        colnames(result) <- c("start", "end", "cumulative att")
    } else { ## with uncertainties
        result <- t(matrix(c(period, catt, catt.se, catt.ci, catt.p)))
        colnames(result) <- c("start", "end", "cumulative att", "S.E.", "CI.lower", "CI.upper", "p.value")
    }
  
    return(result)
} 

##############################
##   equivalence test       ##
##############################

## function to get two-sided p-values
get.pvalue <- function(vec) {
  if (NaN%in%vec|NA%in%vec) {
    nan.pos <- is.nan(vec)
    na.pos <- is.na(vec)
    pos <- c(which(nan.pos),which(na.pos))
    vec.a <- vec[-pos]
    a <- sum(vec.a >= 0)/(length(vec)-sum(nan.pos|na.pos)) * 2
    b <- sum(vec.a <= 0)/(length(vec)-sum(nan.pos|na.pos)) * 2  
  } else {
    a <- sum(vec >= 0)/length(vec) * 2
    b <- sum(vec <= 0)/length(vec) * 2  
  }
  return(min(as.numeric(min(a, b)),1))
}

## Causal inference using counterfactual estimators
## (fect: fixed effects counterfactuals)

## MAIN FUNCTION
## fect.formula()
## fect.default()

## DEPENDENT FUNCTIONS
## fect_fe() ## interactive fixed effects model
## fect_mc() ## matrix completion
## fect_boot() ## bootstrap

## fitness test
## fect_test ## wild bootstrap

## METHODS
## print.fect()
## plot.fect()

## several supporting functions

#####################################################################
## A Shell Function
#####################################################################

## generic function
fect <- function(
    formula = NULL,
    data, # a data frame (long-form)
    Y, # outcome
    D, # treatment
    X = NULL, # time-varying covariates
    W = NULL, # weight column name; convenience default for both roles
    W.est = NULL, # weight column for outcome-model fit (overrides W);
                  # NULL falls back to W
    W.agg = NULL, # weight column for ATT aggregation (overrides W);
                  # NULL falls back to W
    group = NULL, # cohort
    na.rm = FALSE, # remove missing values
    index, # c(unit, time) indicators
    force = "two-way", # fixed effects demeaning
    time.component.from = "notyettreated", # factor estimation sample: "notyettreated" or "nevertreated"
    em = TRUE, # EM algorithm for missing data; FALSE uses direct SVD (requires complete estimation sample)
    r = 0, # number of factors
    lambda = NULL, # mc method: regularization parameter
    nlambda = 10, ## mc method: regularization parameter
    CV = NULL, # cross-validation
    k = 20, # times of CV
    cv.prop = 0.1, ## fraction of eligible units sampled per fold (rolling) / proportion of obs masked per round (block)
    cv.method = "rolling", ## CV masking strategy (default flipped from "all_units" in v2.3.0)
    cv.nobs = 3, ## cv taking consecutive units
    cv.donut = 1, ## cv mspe (default flipped from 0 to match cv.buffer = 1 in v2.3.0)
    cv.buffer = 1, ## past-side buffer for cv.method = "rolling"
    criterion = "mspe", # for ife model: mspe, pc or both
    binary = FALSE, # probit model
    QR = FALSE, # QR or SVD for binary probit
    method = "fe", # method: e for fixed effects; ife for interactive fe; mc for matrix completion
    se = FALSE, # report uncertainties
    vartype = "bootstrap", # bootstrap or jackknife
    para.error = "auto", # parametric bootstrap error strategy: "auto", "ar", "empirical", "wild"
    cl = NULL,
    ci.method = "normal", # CI method for fect's est.* slots: "normal" (Wald: theta_hat +- z * SE) or "basic" (reflected pivot, Davison-Hinkley 1997 Sec. 5.2.1). For percentile / bc / bca on alternative estimands (att.cumu, aptt, log.att), call estimand(fit, type, ci.method) post-fit
    quantile.CI = NULL, # DEPRECATED: use ci.method instead. NULL sentinel = "not supplied"; legacy FALSE -> ci.method = "normal", legacy TRUE -> ci.method = "basic"
    nboots = 200, # number of bootstraps (sufficient for SE / normal CI; bump to 1000+ for tail-quantile CIs in estimand())
    alpha = 0.05, # significance level
    parallel = TRUE, # parallel computing
    cores = NULL, # number of cores
    tol = 1e-5, # tolerance level (tightened from 1e-3 in v2.4.3)
    max.iteration = 5000,
    seed = NULL, # set seed
    min.T0 = NULL, # minimum T0
    max.missing = NULL, # maximum missing
    proportion = 0.3, # use to fit the f test and equivalence test
    pre.periods = NULL, # fit test period
    f.threshold = 0.5, # equiv
    tost.threshold = NULL, # equiv
    knots = NULL,
    degree = 2, # wald = FALSE, # fit test
    group.fe = NULL, # additional additive FE columns absorbed via CFE machinery
    cfe = NULL,
    Z = NULL,
    gamma = NULL,
    Q = NULL,
    kappa = NULL,
    Q.type = NULL, # c(1,2,3,"bspline") or c("linear", "quadratic", "cubic", "bspline")
    Q.bspline.degree = NULL, # used only when Q.type includes "bspline"
    Z.param = NULL,
    Q.param = NULL,
    balance.period = NULL, # the pre and post periods for balanced samples
    fill.missing = FALSE, # whether to balance missing observations
    placeboTest = FALSE, # placebo test
    placebo.period = NULL, # placebo test period
    carryoverTest = FALSE, # carry-over test
    carryover.period = NULL, # carry-over period
    carryover.rm = NULL,
    loo = FALSE, # leave one period out placebo
    permute = FALSE, ## permutation test
    m = 2, ## block length
    normalize = FALSE, # accelerate option
    keep.sims = FALSE, # keep individual bootstrap/jackknife simulations
    cm = FALSE, # causal moderation
    loading.bound = "none",           # simplex projection of treated loadings
    gamma.loading = NULL,             # scalar gamma for simplex; NULL = CV
    gamma.loading.grid = NULL,        # optional grid for gamma CV
    cv.rule = "1se"                   # CV selection rule: "1se", "min", "1pct"
) {
    UseMethod("fect")
}

## formula method

fect.formula <- function(
    formula = NULL,
    data, # a data frame (long-form)
    Y, # outcome
    D, # treatment
    X = NULL, # time-varying covariates
    W = NULL, # weights; convenience default for both roles
    W.est = NULL, # weight column for outcome-model fit; see fect()
    W.agg = NULL, # weight column for ATT aggregation; see fect()
    group = NULL, # cohort
    na.rm = FALSE, # remove missing values
    index, # c(unit, time) indicators
    force = "two-way", # fixed effects demeaning
    time.component.from = "notyettreated", # factor estimation sample: "notyettreated" or "nevertreated"
    em = TRUE, # EM algorithm for missing data; FALSE uses direct SVD (requires complete estimation sample)
    r = 0, # nubmer of factors
    lambda = NULL, # mc method: regularization parameter
    nlambda = 10, ## mc method: regularization parameter
    CV = NULL, # cross-validation
    k = 20, # times of CV
    cv.prop = 0.1, ## fraction of eligible units sampled per fold (rolling) / proportion of obs masked per round (block)
    cv.method = "rolling",
    cv.nobs = 3,
    cv.donut = 1, ## cv mspe (default flipped from 0 to match cv.buffer = 1 in v2.3.0)
    cv.buffer = 1, ## past-side buffer for cv.method = "rolling"
    criterion = "mspe", # for ife model: mspe, pc or both
    binary = FALSE, # probit model
    QR = FALSE, # QR or SVD for binary probit
    method = "fe", # method: fe for fixed effects; ife for interactive fe; mc for matrix completion
    se = FALSE, # report uncertainties
    vartype = "bootstrap", # bootstrap or jackknife
    para.error = "auto", # parametric bootstrap error strategy: "auto", "ar", "empirical", "wild"
    cl = NULL,
    ci.method = "normal", # CI method for fect's est.* slots: "normal" or "basic"
    quantile.CI = NULL, # DEPRECATED: use ci.method instead
    nboots = 200, # number of bootstraps (sufficient for SE / normal CI; bump to 1000+ for tail-quantile CIs in estimand())
    alpha = 0.05, # significance level
    parallel = TRUE, # parallel computing
    cores = NULL, # number of cores
    tol = 1e-5, # tolerance level (tightened from 1e-3 in v2.4.3)
    max.iteration = 5000,
    seed = NULL, # set seed
    min.T0 = NULL,
    max.missing = NULL,
    proportion = 0.3,
    pre.periods = NULL,
    f.threshold = 0.5, # equiv
    tost.threshold = NULL,
    knots = NULL,
    degree = 2, # wald = FALSE,
    group.fe = NULL,
    cfe = NULL,
    Z = NULL,
    gamma = NULL,
    Q = NULL,
    kappa = NULL,
    Q.type = NULL,
    Q.bspline.degree = NULL,
    Z.param = NULL,
    Q.param = NULL,
    balance.period = NULL, # the pre and post periods for balanced samples
    fill.missing = FALSE, # whether to balance missing observations
    placeboTest = FALSE, # placebo test
    placebo.period = NULL, # placebo test period
    carryoverTest = FALSE, # carry-over test
    carryover.period = NULL, # carry-over period
    carryover.rm = NULL,
    loo = FALSE, # leave one period out placebo
    permute = FALSE, ## permutation test
    m = 2, ## block length
    normalize = FALSE,
    keep.sims = FALSE,
    cm = FALSE,
    loading.bound = "none",
    gamma.loading = NULL,
    gamma.loading.grid = NULL,
    cv.rule = "1se"
) {
    ## parsing
    varnames <- all.vars(formula)
    Yname <- varnames[1]
    Dname <- varnames[2]
    if (length(varnames) > 2) {
        Xname <- varnames[3:length(varnames)]
    } else {
        Xname <- NULL
    }

    namesData <- colnames(data)
    for (i in 1:length(varnames)) {
        if (!varnames[i] %in% namesData) {
            stop(paste(
                "variable \"",
                varnames[i],
                "\" is not in the data set.",
                sep = ""
            ))
        }
    }

    ## check binary outcome
    if (binary == TRUE) {
        unique_y <- sort(unique(data[, Yname]))
        if (length(unique_y) != 2) {
            stop("Outcome should only contain 0 and 1.")
        } else {
            if (sum(unique_y == c(0, 1)) != 2) {
                stop("Outcome should only contain 0 and 1.")
            }
        }
    }

    ## run the model
    out <- fect.default(
        formula = NULL,
        data = data,
        Y = Yname,
        D = Dname,
        X = Xname,
        W = W,
        W.est = W.est,
        W.agg = W.agg,
        group = group,
        na.rm = na.rm,
        balance.period = balance.period,
        fill.missing = fill.missing,
        index = index,
        force = force,
        time.component.from = time.component.from,
        em = em,
        r = r,
        lambda = lambda,
        nlambda = nlambda,
        CV = CV,
        k = k,
        cv.prop = cv.prop,
        cv.method = cv.method,
        cv.nobs = cv.nobs,
        cv.donut = cv.donut,
        cv.buffer = cv.buffer,
        criterion = criterion,
        binary = binary,
        QR = QR,
        method = method,
        se = se,
        vartype = vartype,
        para.error = para.error,
        cl = cl,
        ci.method = ci.method,
        quantile.CI = quantile.CI,
        nboots = nboots,
        alpha = alpha,
        parallel = parallel,
        cores = cores,
        tol = tol,
        max.iteration = max.iteration,
        seed = seed,
        min.T0 = min.T0,
        max.missing = max.missing,
        proportion = proportion,
        pre.periods = pre.periods,
        f.threshold = f.threshold,
        tost.threshold = tost.threshold,
        knots = knots,
        degree = degree,
        group.fe = group.fe,
        cfe = cfe,
        Z = Z,
        gamma = gamma,
        Q = Q,
        kappa = kappa,
        Q.type = Q.type,
        Q.bspline.degree = Q.bspline.degree,
        Z.param = Z.param,
        Q.param = Q.param,
        placebo.period = placebo.period,
        placeboTest = placeboTest,
        carryoverTest = carryoverTest,
        carryover.period = carryover.period,
        carryover.rm = carryover.rm,
        loo = loo,
        permute = permute,
        m = m,
        normalize = normalize,
        keep.sims = keep.sims,
        cm = cm,
        loading.bound      = loading.bound,
        gamma.loading      = gamma.loading,
        gamma.loading.grid = gamma.loading.grid,
        cv.rule            = cv.rule
    )

    out$call <- match.call()
    out$formula <- formula
    return(out)
}


## default function

fect.default <- function(
    formula = NULL,
    data, # a data frame (long-form)
    Y, # outcome
    D, # treatment
    X = NULL, # time-varying covariates
    W = NULL, # weights; convenience default for both roles
    W.est = NULL, # weight column for outcome-model fit; see fect()
    W.agg = NULL, # weight column for ATT aggregation; see fect()
    group = NULL, # cohort
    na.rm = FALSE, # remove missing values
    index, # c(unit, time) indicators
    force = "two-way", # fixed effects demeaning
    time.component.from = "notyettreated", # factor estimation sample: "notyettreated" or "nevertreated"
    em = TRUE, # EM algorithm for missing data; FALSE uses direct SVD (requires complete estimation sample)
    r = 0, # nubmer of factors
    lambda = NULL, ## mc method: regularization parameter
    nlambda = 0,
    CV = NULL, # cross-validation
    k = 20, # times of CV
    cv.prop = 0.1,
    cv.method = "all_units",
    cv.nobs = 3,
    cv.donut = 1, ## cv mspe
    cv.buffer = 1, ## past-side buffer for cv.method = "rolling"
    criterion = "mspe",
    binary = FALSE, # probit model
    QR = FALSE, # QR or SVD for binary probit
    method = "fe", # method: ife for interactive fe; mc for matrix completion
    se = FALSE, # report uncertainties
    vartype = "bootstrap", # bootstrap or jackknife
    para.error = "auto", # parametric bootstrap error strategy: "auto", "ar", "empirical", "wild"
    cl = NULL,
    ci.method = "normal", # CI method for fect's est.* slots: "normal" or "basic"
    quantile.CI = NULL, # DEPRECATED: use ci.method instead
    nboots = 200, # number of bootstraps (sufficient for SE / normal CI; bump to 1000+ for tail-quantile CIs in estimand())
    alpha = 0.05, # significance level
    parallel = TRUE, # parallel computing
    cores = NULL, # number of cores
    tol = 1e-5, # tolerance level (tightened from 1e-3 in v2.4.3)
    max.iteration = 5000,
    seed = NULL, # set seed
    min.T0 = NULL,
    max.missing = NULL,
    proportion = 0.3,
    pre.periods = NULL,
    f.threshold = 0.5, # equiv
    tost.threshold = NULL,
    knots = NULL,
    degree = 2, # wald = FALSE,
    group.fe = NULL,
    cfe = NULL,
    Z = NULL,
    gamma = NULL,
    Q = NULL,
    kappa = NULL,
    Q.type = NULL,
    Q.bspline.degree = NULL,
    Z.param = NULL,
    Q.param = NULL,
    balance.period = NULL, # the pre and post periods for balanced samples
    fill.missing = FALSE, # whether to balance missing observations
    placeboTest = FALSE, # placebo test
    placebo.period = NULL, # placebo test period
    carryoverTest = FALSE, # carry-over test
    carryover.period = NULL, # carry-over period
    carryover.rm = NULL,
    loo = FALSE, # leave one period out placebo
    permute = FALSE, ## permutation test
    m = 2, ## block length
    normalize = FALSE,
    keep.sims = FALSE,
    cm=FALSE,
    loading.bound = "none",
    gamma.loading = NULL,
    gamma.loading.grid = NULL,
    cv.rule = "1se"
) {
    ## -------------------------------##
    ## Checking Parameters
    ## -------------------------------##
    cv.rule <- .fect_validate_cv_rule(cv.rule)
    placeboEquiv <- loo
    permu.dimension <- "time"

    ## read data
    if (is.data.frame(data) == FALSE || length(class(data)) > 1) {
        data <- as.data.frame(data)
        ## warning("Not a data frame.")
    }

    ## ----------------------------------------------------------------
    ## group.fe: validate, auto-route method, normalize into index[3:]
    ## ----------------------------------------------------------------
    ## group.fe is a character vector of column names naming additive simple
    ## FE groupings absorbed via the existing CFE pipeline (X.extra.FE). It
    ## is a discoverable surface for the legacy index = c(unit, time, extra,
    ## extra, ...) syntax. See statsclaw-workspace/fect/runs/2026-05-21-
    ## higher-level-fe.md for the design memo.
    if (!is.null(group.fe)) {
        if (!is.character(group.fe)) {
            stop("\"group.fe\" must be a character vector of column names.",
                 call. = FALSE)
        }
        ## (a) columns exist
        missing.cols <- setdiff(group.fe, colnames(data))
        if (length(missing.cols) > 0) {
            stop("group.fe column(s) not found in data: ",
                 paste(missing.cols, collapse = ", "), call. = FALSE)
        }
        ## (b) conflict with legacy index[3:] form: hard error
        if (length(index) > 2) {
            stop("Specify additional FE via group.fe OR extra index slots, ",
                 "not both. index[3:] is the legacy form; group.fe is preferred.",
                 call. = FALSE)
        }
        ## (c) overlap with index[1:2]: warn and drop the duplicate entries
        dup.idx <- intersect(group.fe, index[1:2])
        if (length(dup.idx) > 0) {
            warning("group.fe entry/entries (",
                    paste(dup.idx, collapse = ", "),
                    ") duplicate index[1:2]; dropping from group.fe.",
                    call. = FALSE)
            group.fe <- setdiff(group.fe, index[1:2])
        }
        ## (d) auto-route method (per D3 in design memo). When routing fe ->
        ## cfe, also pre-set r = 0 and CV = FALSE so the call behaves like
        ## the user's intended "FE only" model. (The standard fe -> ife r=0
        ## coercion at L711-714 of this file uses ife as its backing path;
        ## the cfe backing requires explicit r/CV to avoid hitting the
        ## interactive-FE code that expects a populated cfe= list.)
        if (length(group.fe) > 0) {
            if (method == "fe") {
                method <- "cfe"
                r <- 0
                if (is.null(CV)) CV <- FALSE
            } else if (method %in% c("ife", "mc", "both", "gsynth")) {
                stop("group.fe with method = \"", method, "\" is not supported. ",
                     "Use method = \"cfe\", r = N explicitly to get free latent ",
                     "factors with group-level FE.",
                     call. = FALSE)
            }
            ## method == "cfe": keep as-is, no message
        }
        ## (e) normalize: append group.fe to index for downstream X.extra.FE pipeline
        if (length(group.fe) > 0) {
            index <- c(index, group.fe)
        }
        ## (f) default cl from group.fe[1] when cl is NULL (the placeholder
        ## value) and group.fe is single-column. We use is.null(cl) here
        ## because missing(cl) doesn't survive the fect.formula -> fect.default
        ## dispatch (cl is explicitly forwarded). Documented: cl = NULL is
        ## the default placeholder and does NOT suppress clustering; use
        ## cl = FALSE to suppress explicitly.
        if (is.null(cl) && length(group.fe) == 1) {
            cl <- group.fe[1]
        } else if (length(group.fe) > 1 && is.null(cl)) {
            stop("Multi-column group.fe requires explicit cl (e.g., cl = '",
                 group.fe[1], "', or cl = index[1] for unit-level clustering).",
                 call. = FALSE)
        }
    }
    ## (g) cl must be a single column name. cl = FALSE is rejected to avoid
    ## the misleading "no clustering" framing: even without an explicit
    ## cluster column the case bootstrap still resamples units, which is
    ## unit-level clustering. Users who want explicit unit-level clustering
    ## should pass cl = index[1] (e.g., cl = "id").
    if (isFALSE(cl)) {
        stop("cl = FALSE is not supported. The case bootstrap always ",
             "resamples units; to cluster at the unit level explicitly, ",
             "pass cl = '", index[1], "' (or cl = index[1]).",
             call. = FALSE)
    }

    ## Snapshot the cl column name now (before downstream reshape to matrix)
    ## so print(fit) / summary(fit) can show the user-facing cluster label.
    cl.label <- if (is.character(cl) && length(cl) == 1) cl else NULL

    ## ----------------------------------------------------------------
    ## Nesting check: each group.fe column must be constant within index[1].
    ## This is required because group.fe is documented as "additive simple FE
    ## on a coarsening of the unit identifier" --- e.g., state on county.
    ## Without this check, default.R:1672 silently reshapes a non-constant
    ## column into a TT*N matrix and produces wrong fits.
    ##
    ## We do NOT apply this check to the legacy `index = c(unit, time, extra,
    ## ...)` form because that form has historically supported BOTH (a)
    ## nested additive FE (the group.fe-style use) AND (b) cell-level
    ## interactions like region_time (which by design vary within unit). The
    ## strict modern surface (group.fe) is for case (a) only; the legacy
    ## surface keeps its full original scope.
    ## ----------------------------------------------------------------
    if (length(group.fe) > 0 && all(index[1] %in% colnames(data))) {
        for (gfe in group.fe) {
            nest <- tapply(data[[gfe]], data[[index[1]]],
                           function(x) length(unique(stats::na.omit(x))))
            bad <- names(nest)[!is.na(nest) & nest > 1]
            if (length(bad) > 0) {
                bad.show <- if (length(bad) > 10)
                    paste0(paste(bad[1:10], collapse = ", "),
                           ", and ", length(bad) - 10, " more")
                else paste(bad, collapse = ", ")
                stop("group.fe column '", gfe,
                     "' is not constant within index[1] = '", index[1],
                     "'. Offending units: ", bad.show,
                     "\n  (group.fe is for additive simple FE on a coarsening ",
                     "of the unit identifier. For cell-level interactions ",
                     "that vary within unit, use the legacy `index = c(unit, ",
                     "time, ...)` form or the `cfe = list(...)` interactive-",
                     "FE argument instead.)",
                     call. = FALSE)
            }
        }
    }

    ## index
    if (
        (length(index) != 2 | sum(index %in% colnames(data)) != 2) &
            method != "cfe"
    ) {
        stop(
            "\"index\" option misspecified. Try, for example, index = c(\"unit.id\", \"time\")."
        )
    }

    if (!is.null(Z.param)) {
        Zgamma <- setNames(as.list(names(Z.param)), unlist(Z.param))
    } else {
        Zgamma <- NULL
    }

    if (!is.null(Q.param)) {
        kappaQ <- setNames(as.list(names(Q.param)), unlist(Q.param))
    } else {
        kappaQ <- NULL
    }

    id <- index[1]
    time <- index[2]


    if (cm == TRUE & ! method %in% c("fe", "ife")) {
        stop("\"cm\" option is only available for the \"fe\" and \"ife\" methods.")
    }

    ## Save user's literal method argument before any silent coercion (e.g.
    ## fe -> ife r=0 below). Referenced by the parametric/nevertreated gate
    ## further down so error messages name the user's actual method = "fe".
    method_arg <- method

    if (se == 1) {
        if (!vartype %in% c("bootstrap", "jackknife", "parametric")) {
            stop(
                "\"vartype\" must be one of \"bootstrap\", \"jackknife\", or \"parametric\".",
                call. = FALSE
            )
        }
        if (vartype == "parametric" && method %in% c("mc", "both")) {
            stop(
                "The \"parametric\" option is not available for the \"mc\" or \"both\" methods."
            )
        }
        if (vartype == "jackknife" && !is.null(cl)) {
            warning(
                "vartype = \"jackknife\" with cl = ... : the cl argument is ignored. ",
                "fect's jackknife is leave-one-unit-out and does not support a ",
                "cluster (block) jackknife. The resulting SEs do not account for ",
                "within-cluster correlation. Use vartype = \"bootstrap\" with cl ",
                "for cluster-aware inference.",
                call. = FALSE
            )
        }
        if (!para.error %in% c("auto", "ar", "empirical", "wild")) {
            stop(
                "\"para.error\" must be one of \"auto\", \"ar\", \"empirical\", or \"wild\".",
                call. = FALSE
            )
        }
    }

    ## Default: W populates both roles (back-compat for callers that just set W).
    if (is.null(W.est)) W.est <- W
    if (is.null(W.agg)) W.agg <- W

    ## Validate the resolved weight columns.
    validate_weight_col <- function(col, label) {
        if (is.null(col)) return(invisible())
        if (length(col) != 1 || !is.character(col)) {
            stop("`", label, "` must be a single column name.")
        }
        if (!col %in% colnames(data)) {
            stop("`", label, "` (\"", col, "\") is not in the dataset.")
        }
        if (!is.numeric(data[, col])) {
            stop("`", label, "` (\"", col, "\") must be numeric.")
        }
        if (any(data[, col] < 0, na.rm = TRUE)) {
            stop("`", label, "` (\"", col, "\") must be non-negative.")
        }
    }
    validate_weight_col(W,     "W")
    validate_weight_col(W.est, "W.est")
    validate_weight_col(W.agg, "W.agg")

    ## v2.3.1 supports only a single weight column (with role gating). Truly
    ## different columns for fit vs aggregation (e.g. survey weights x IPW)
    ## requires carrying two parallel matrices through the dispatcher and is
    ## scheduled for v2.4.0. Stop with an informative message if requested now.
    if (!is.null(W.est) && !is.null(W.agg) && !identical(W.est, W.agg)) {
        stop(
            "Distinct weight columns for `W.est` and `W.agg` are scheduled ",
            "for fect v2.4.0. v2.3.1 supports a single weight column with ",
            "role-gated routing: set `W.est = W.agg = \"col\"` (or `W = \"col\"`) ",
            "for survey/sample weights, `W.agg = \"col\"` alone for IPW / ",
            "balancing weights, or `W.est = \"col\"` alone for ",
            "robust-regression / heteroskedasticity weights."
        )
    }

    ## Resolve a single column name `Wname` for downstream code paths. The
    ## per-role flags `use.W.in.fit` and `use.W.in.agg` carry the role
    ## information.
    if (is.null(W) && !is.null(W.est)) W <- W.est
    if (is.null(W) && !is.null(W.agg)) W <- W.agg

    ## Drop rows where the weight column is exactly 0.
    if (!is.null(W) && any(data[, W] == 0, na.rm = TRUE)) {
        data <- data[data[, W] > 0, , drop = FALSE]
    }

    use.W.in.fit <- !is.null(W.est)
    use.W.in.agg <- !is.null(W.agg)

    ## check duplicated observations
    unique_label <- unique(paste(
        data[, index[1]],
        "_",
        data[, index[2]],
        sep = ""
    ))
    if (length(unique_label) != dim(data)[1]) {
        stop(
            "Observations are not uniquely defined by unit and time indicators."
        )
    }

    ## force
    if (force == "none") {
        # force = 0 "none": no additive fixed effects imposed
        force <- 0
    } else if (force == "unit") {
        # force = 1 "unit": unit fixed-effect (default)
        force <- 1
    } else if (force == "time") {
        # force = 2 "time": time fixed-effect
        force <- 2
    } else if (force == "two-way") {
        # force = 3 "two-way": two-way fixed-effect
        force <- 3
    }

    if (!force %in% c(0, 1, 2, 3)) {
        stop(
            "\"force\" option misspecified; choose from c(\"none\", \"unit\", \"time\", \"two-way\")."
        )
    }

    ## binary
    if (binary == 1) {
        method <- "ife"
        normalize <- FALSE
    }

    ## method
    if (
        !method %in%
            c(
                "fe",
                "ife",
                "mc",
                "both",
                "gsynth",
                "cfe"
            )
    ) {
        stop(
            "\"method\" option misspecified; choose from c(\"fe\",\"gsynth\", \"ife\", \"mc\", \"both\", \"cfe\")."
        )
    }

    ## validate time.component.from
    if (!time.component.from %in% c("notyettreated", "nevertreated")) {
        stop("\"time.component.from\" must be \"notyettreated\" or \"nevertreated\".")
    }

    ## mc does not use explicit factor estimation
    if (method %in% c("mc", "both") && time.component.from == "nevertreated") {
        stop("\"time.component.from = 'nevertreated'\" is not supported for the \"mc\" or \"both\" methods. ",
             "Matrix completion does not use explicit factor estimation.")
    }

    ## em=FALSE requires a complete estimation sample
    if (isFALSE(em) && time.component.from == "notyettreated") {
        stop("\"em = FALSE\" is not compatible with \"time.component.from = 'notyettreated'\". ",
             "The not-yet-treated estimation sample has missing data by construction.")
    }

    ## gsynth always uses never-treated units only
    if (method == "gsynth") {
        if (time.component.from == "notyettreated") {
            time.component.from <- "nevertreated"
        }
        em <- FALSE
    }

    if (is.null(min.T0)) {
        if (method == "fe") {
            min.T0 <- 1
        }
        if (method %in% c("ife", "mc", "both", "gsynth", "cfe")) {
            min.T0 <- 5
        }
    } else {
        if (min.T0 <= 0) {
            stop("\"min.T0\" option should be larger than 0.\n")
        }
    }

    ## the default setting of CV
    if (is.null(CV)) {
        if (method == "fe") {
            r <- 0
            CV <- FALSE
            method <- "ife"
        } else if (method == "cfe") {
            if (time.component.from == "nevertreated") {
                if (length(r) == 1) {
                    CV <- FALSE
                } else if (length(r) > 1) {
                    CV <- TRUE
                }
            } else {
                CV <- FALSE
            }
        } else if (method == "both") {
            CV <- TRUE
            if (length(r) == 1 & r == 0) {
                r <- c(0:5)
            }
        } else if (method %in% c("ife", "gsynth")) {
            if (length(r) == 1) {
                CV <- FALSE
            } else if (length(r) > 1) {
                CV <- TRUE
            }
        } else if (method == "mc") {
            if (length(lambda) == 1) {
                CV <- FALSE
            } else if (length(lambda) > 1 | is.null(lambda)) {
                CV <- TRUE
            }
        }
    } else {
        if (method == "fe") {
            r <- 0
            CV <- FALSE
            method <- "ife"
        } else if (method == "cfe") {
            if (time.component.from != "nevertreated") {
                CV <- FALSE
            }
        } else if (method == "both") {
            CV <- TRUE
        }
    }

    if (!criterion %in% c("mspe", "gmspe", "moment", "pc")) {
        stop(
            "\"criterion\" option misspecified; choose from c(\"mspe\", \"gmspe\",\"moment\", \"pc\")."
        )
    }

    # if (!criterion %in% c("mspe","pc")) {
    #    stop("\"criterion\" option misspecified; choose from c(\"mspe\", \"pc\").")
    # }

    if (method == "both" && criterion == "pc") {
        stop(
            "\"ife\" and \"mc\" method cannot be compared; try using \"mspe\" as criteria."
        )
    }

    ## r
    if (method %in% c("ife", "both", "cfe") & r[1] < 0) {
        stop(
            "\"r\" option misspecified. The number of factors must be non-negative."
        )
    }

    ## lambda
    if (method %in% c("mc", "both")) {
        if (!is.null(lambda)) {
            if (sum(lambda < 0) > 0) {
                stop("\"lambda\" option misspecified. It must be non-negative.")
            }
        }
        if (CV == FALSE & is.null(lambda)) {
            message("No lambda is supplied. FEct is applied.")
            method <- "ife"
            r <- 0
        }
    }

    ## leave one period out placebo
    if (placeboEquiv == TRUE) {
        if (se != TRUE) {
            message(
                "For leave one period out placebo test, automatically set \"se\" to TRUE."
            )
            se <- TRUE
        }
        if (placeboTest == TRUE) {
            message(
                "For leave one period out placebo test, automatically set \"placeboTest\" to FALSE."
            )
            placeboTest <- FALSE
        }
        if (carryoverTest == TRUE) {
            message(
                "For leave one period out placebo test, automatically set \"carryoverTest\" to FALSE."
            )
            carryoverTest <- FALSE
        }
    }

    if (placeboTest == TRUE & carryoverTest == TRUE) {
        stop(
            "\"placeboTest\" and \"carryoverTest\" can't be performed simultaneously."
        )
    }

    ## CV

    if (CV == TRUE) {
        if (placeboTest == TRUE) {
            stop(
                "Placebo test cannot be performed while doing cross-validation."
            )
        }

        if (carryoverTest == TRUE) {
            stop(
                "Carry-over test cannot be performed while doing cross-validation."
            )
        }

        if (method %in% c("ife", "both", "gsynth", "cfe")) {
            if (length(r) == 2 & r[1] > r[2]) {
                stop(
                    "\"r\" option misspecified. The first element should be smaller than the second element in r().\n"
                )
            }
        }
        if (method %in% c("mc", "both")) {
            if (nlambda <= 0) {
                stop("\"nlambda\" option misspecified.\n")
            }
        }
    } else {
        if (
            !method %in%
                c("gsynth", "ife", "mc", "cfe")
        ) {
            stop(
                "\"method\" option misspecified; please choose from c(\"gsynth\",\"ife\", \"mc\", \"cfe\")."
            )
        }
    }


    if (permute == 1) {
        if (placeboTest == TRUE) {
            stop("\"permute\" can't be used with \"placeboTest\".")
        }
        if (carryoverTest == TRUE) {
            stop("\"permute\" can't be used with \"carryoverTest\".")
        }
        if (loo == TRUE) {
            stop("\"permute\" can't be used with \"loo\".")
        }
        if (!permu.dimension %in% c("time", "unit")) {
            stop("\"permu.dimension\" must be in c(\"unit\",\"time\") .")
        }
    }

    if (length(r) == 1) {
        # if (r>=5) {
        #    r.end <- r
        # }
        # else {
        #    r.end <- 5
        # }
        r.end <- r
    } else {
        r.end <- max(r)
        r <- min(r)
    }

    ## uncertainty estimates
    if (is.logical(se) == FALSE & !se %in% c(0, 1)) {
        stop("\"se\" is not a logical flag.")
    }

    ## ci.method: "normal" (Wald: theta_hat +- z * SE) or "basic" (reflected
    ## pivot: 2*theta_hat - quantile(boot, c(1-alpha/2, alpha/2))). Wider 5-method
    ## surface (basic, percentile, bc, bca, normal) lives on estimand() for
    ## alternative estimands (att.cumu, aptt, log.att); see ?estimand and
    ## chapter 7 of the fect User Manual.
    ##
    ## quantile.CI is the deprecated legacy arg; map it to ci.method with a
    ## one-time warning when the user supplied it explicitly. The NULL sentinel
    ## (default) means "user did not supply it" so the warning stays silent
    ## under the modern API path.
    if (!is.null(quantile.CI)) {
        warning(
            "Argument `quantile.CI` is deprecated as of fect 2.4.2 and will be ",
            "removed in a future release. Use `ci.method = \"normal\"` (the ",
            "default; equivalent to `quantile.CI = FALSE`) or `ci.method = \"basic\"` ",
            "(equivalent to `quantile.CI = TRUE`) instead.\n\n",
            "For percentile / bc / bca CIs on alternative estimands ",
            "(att.cumu, aptt, log.att), call `estimand(fit, type, ci.method)` ",
            "post-fit; see chapter 7 of the fect User Manual.",
            call. = FALSE
        )
        if (!is.logical(quantile.CI) && !quantile.CI %in% c(0, 1)) {
            stop("\"quantile.CI\" is not a logical flag.", call. = FALSE)
        }
        ci.method <- if (isTRUE(as.logical(quantile.CI))) "basic" else "normal"
    }
    ## fect's built-in CI machinery for ci.method = "basic" reads raw
    ## quantiles of the bootstrap distribution.  This is calibrated for
    ## vartype = "bootstrap" only --- the case bootstrap is naturally
    ## centered at theta-hat.  vartype = "parametric" stores eff.boot
    ## centered at 0 (under H0); the reflected CI 2*theta_hat - quantile(boot)
    ## then collapses around 2*theta_hat instead of theta_hat, giving 0%
    ## coverage.  vartype = "jackknife" leave-one-out values are not
    ## exchangeable draws from the sampling distribution either (E&T 1993
    ## ch11; D&H 1997 Sec. 3.2.1).  estimand() applies a location-shift fix for
    ## parametric and a hard-error for jackknife on non-normal ci.methods;
    ## fect's built-in path does neither, so we hard-error here and point
    ## users to estimand().
    .ci.method.allowed <- c("normal", "basic")
    if (length(ci.method) != 1L || !is.character(ci.method) ||
        !ci.method %in% .ci.method.allowed) {
        if (is.character(ci.method) && length(ci.method) == 1L &&
            ci.method %in% c("percentile", "bc", "bca")) {
            stop(
                "ci.method = \"", ci.method, "\" is not supported in fect(); ",
                "fect()'s built-in CI machinery offers only \"normal\" (Wald) ",
                "and \"basic\" (reflected pivot).\n\n",
                "For \"percentile\", \"bc\", and \"bca\" CIs, fit with the ",
                "default ci.method and call `estimand(fit, type, ci.method = \"",
                ci.method, "\")` post-fit. The full 5-method surface is the ",
                "estimand() path; see ?estimand and chapter 7 of the fect ",
                "User Manual.",
                call. = FALSE
            )
        }
        stop(
            "`ci.method` must be one of \"normal\" or \"basic\"; got \"",
            ci.method, "\".",
            call. = FALSE
        )
    }
    ## Reject ci.method = "basic" on jackknife fits.  Jackknife pseudo-values
    ## are influence-function-flavored leave-one-out quantities, not exchangeable
    ## draws from the sampling distribution of theta-hat (E&T 1993 ch11; D&H
    ## 1997 Sec. 3.2.1).  The reflected pivot interval has no defensible meaning
    ## here.  Matches estimand()'s behaviour --- estimand hard-errors on
    ## non-"normal" ci.methods for jackknife fits.
    if (ci.method == "basic" && !is.null(vartype) &&
        identical(vartype, "jackknife")) {
        stop(
            "ci.method = \"basic\" is not supported for vartype = \"jackknife\". ",
            "Jackknife produces an SE estimate via the Tukey pseudo-value ",
            "formula, not a sampling distribution of theta-hat; reflection-based ",
            "CIs require exchangeable draws from a sampling distribution ",
            "(Efron & Tibshirani 1993, Chapter 11; Davison & Hinkley 1997, ",
            "Section 3.2.1).\n\n",
            "Use ci.method = \"normal\" (the only valid CI for jackknife: ",
            "theta-hat +- z * SE_jack) or refit with vartype = \"bootstrap\" ",
            "for the basic interval.",
            call. = FALSE
        )
    }
    ## ci.method = "basic" on vartype = "parametric" is supported via a
    ## location-shift fix in the CI computation downstream (R/boot.R, around
    ## line 3590): the parametric path stores eff.boot centered at 0 (under
    ## H0), and the reflected pivot CI 2*theta_hat - quantile(boot) collapses
    ## around 2*theta_hat without a shift.  fect() applies the same shift
    ## that R/po-estimands.R applies inside estimand() (commit b4e9fbf), so
    ## fit$est.avg with ci.method = "basic" on a parametric fit matches
    ## estimand(fit, "att", ci.method = "basic") byte-equally for the
    ## avg-level + per-event-time CIs.  The shift is currently applied at
    ## those two slots only; for other slots (calendar, cohort, subgroup,
    ## balanced, by-W, placebo, carryover), basic on parametric is not
    ## yet patched and may produce 0% coverage CIs --- call estimand() for
    ## those slots.
    ## Bridge to the existing internal dispatch in fect_boot, which is gated by
    ## a logical `quantile.CI`. After this resolution, .quantile.CI.bool is the
    ## single source of truth for the bootstrap-CI branch downstream.
    .quantile.CI.bool <- (ci.method == "basic")
    ## Tail-CI replicate warning (mirrors estimand's .check_tail_ci_replicates):
    ## `basic` reads tail quantiles of the bootstrap distribution; the literature
    ## floor for tail-quantile CIs is B >= 1000 (Efron 1987 Sec. 3; DiCiccio & Efron
    ## 1996 Sec. 4). Warn at fit time so users don't need to refit.
    if (ci.method == "basic" && isTRUE(se) && !is.null(nboots) &&
        is.numeric(nboots) && nboots < 1000 &&
        !identical(Sys.getenv("TESTTHAT"), "true")) {
        warning(
            "ci.method = \"basic\" reads tail quantiles of the bootstrap ",
            "distribution; with nboots = ", nboots, " (< 1000) the 5th / 195th ",
            "order statistics that the basic interval depends on may be ",
            "unstable (Efron 1987 Sec. 3; DiCiccio & Efron 1996 Sec. 4 recommend ",
            "B >= 1000). Refit with `fect(..., nboots = 1000)` for ",
            "publication-grade CIs. The point estimate and SE are unaffected.",
            call. = FALSE
        )
    }

    ## normalize
    if (is.logical(normalize) == FALSE & !normalize %in% c(0, 1)) {
        stop("\"normalize\" is not a logical flag.")
    }

    ## nboots
    if (se == TRUE & nboots <= 0) {
        stop("\"nboots\" option misspecified. Try, for example, nboots = 1000.")
    }

    ## parallel & cores
    ## parallel accepts: TRUE, FALSE, "cv", "boot", c("cv","boot")
    valid_parallel_strings <- c("cv", "boot")
    if (!isTRUE(parallel) && !identical(parallel, FALSE)) {
        parallel_chars <- as.character(parallel)
        bad <- setdiff(parallel_chars, valid_parallel_strings)
        if (length(bad) > 0) {
            stop(
                "\"parallel\" must be TRUE, FALSE, \"cv\", \"boot\", or c(\"cv\",\"boot\"). ",
                "Got: ", paste0("\"", bad, "\"", collapse = ", ")
            )
        }
    }
    do_parallel_cv   <- isTRUE(parallel) || "cv"   %in% as.character(parallel)
    do_parallel_boot <- isTRUE(parallel) || "boot" %in% as.character(parallel)

    if ((do_parallel_cv || do_parallel_boot) && !is.null(cores)) {
        if (!is.numeric(cores) || cores <= 0) {
            stop("\"cores\" option misspecified. Try, for example, cores = 2.")
        }
    }

    ## --- loading.bound validation (REQ-bounded-loadings) ---
    if (!is.character(loading.bound) || length(loading.bound) != 1L ||
        !(loading.bound %in% c("none", "simplex"))) {
        stop("'loading.bound' must be one of 'none', 'simplex'.")
    }
    if (loading.bound == "simplex") {
        if (!(method %in% c("ife", "gsynth"))) {
            stop("'loading.bound = \"simplex\"' is only supported for method = \"ife\" or \"gsynth\" in this version.")
        }
        if (!identical(time.component.from, "nevertreated")) {
            stop("'loading.bound = \"simplex\"' requires time.component.from = \"nevertreated\". ",
                 "The not-yet-treated dispatch is not supported in this version.")
        }
    }
    if (!is.null(gamma.loading)) {
        if (!is.numeric(gamma.loading) || length(gamma.loading) != 1L ||
            !is.finite(gamma.loading) || gamma.loading <= 0) {
            stop("'gamma.loading' must be a single positive finite numeric, or NULL.")
        }
    }
    if (!is.null(gamma.loading.grid)) {
        if (!is.numeric(gamma.loading.grid) || any(!is.finite(gamma.loading.grid)) ||
            any(gamma.loading.grid <= 0)) {
            stop("'gamma.loading.grid' must be a positive finite numeric vector, or NULL.")
        }
    }

    ## tol
    if (tol <= 0) {
        stop("\"tol\" option misspecified. Try using the default option.")
    }

    ## seed
    if (is.null(seed) == FALSE) {
        if (is.numeric(seed) == FALSE) {
            stop("\"seed\" should be a number.")
        }
    }

    ## remove missing values
    if (is.logical(na.rm) == FALSE & !na.rm %in% c(0, 1)) {
        stop("\"na.rm\" is not a logical flag.")
    }

    if (!is.null(balance.period)) {
        if (length(balance.period) != 2) {
            stop(" should be of length 2.")
        }
        if (balance.period[1] > 0) {
            stop(
                "The first element in \"balance.period\" should be no larger than 0."
            )
        }
        if (balance.period[2] < 1) {
            stop(
                "The second element in \"balance.period\" should be no smaller than 1."
            )
        }
        if (is.logical(fill.missing) == FALSE & !fill.missing %in% c(0, 1)) {
            stop("\"fill.missing\" is not a logical flag.")
        }
        if (!is.null(group)) {
            stop("\"group\" should not be used with \"balance.period\".\n")
        }
        if (!is.null(carryover.rm)) {
            stop(
                "\"balance.period\" should not be used with \"carryover.rm\".\n"
            )
        }
        balance.periods <- c(balance.period[1]:balance.period[2])
        # treat the units with history balance.periods as a certain group
    }

    # cohort
    if (!is.null(group)) {
        if (!group %in% names(data)) {
            stop("\"group\" misspecified.\n")
        }
        ## warn if group coincides with unit id but some units never treated
        if (!is.null(index) && length(index) >= 1 && group == index[1]) {
            if (!is.null(D) && D %in% names(data)) {
                id.vec <- data[, index[1]]
                d.vec <- data[, D]
                ## treat anything different from 0/NA as treated
                if (is.factor(d.vec) || is.character(d.vec)) {
                    d.numeric <- suppressWarnings(as.numeric(as.character(d.vec)))
                } else {
                    d.numeric <- suppressWarnings(as.numeric(d.vec))
                }
                if (all(is.na(d.numeric))) {
                    d.numeric <- suppressWarnings(as.numeric(d.vec))
                }
                treated.flag <- ifelse(is.na(d.numeric), 0, d.numeric != 0)
                treat.count <- tapply(treated.flag, id.vec, function(x) sum(x, na.rm = TRUE))
                never.tr <- names(treat.count)[is.na(treat.count) | treat.count == 0]
                if (length(never.tr) > 0) {
                    sample.ids <- utils::head(never.tr, 3)
                    warning(
                        paste0(
                            "The \"group\" option is set to the unit identifier (",
                            group,
                            "), but ",
                            length(never.tr),
                            " unit(s) never receive treatment (e.g., ",
                            paste(sample.ids, collapse = ", "),
                            "). ",
                            "Consider omitting the \"group\" option or using a cohort indicator instead."
                        ),
                        call. = FALSE
                    )
                }
            }
        }
    }

    if (!is.null(cl)) {
        if (!cl %in% names(data)) {
            stop("\"cl\" misspecified.\n")
        }
        if (length(cl) != 1) {
            stop("Length of \"cl\" must be 1.\n")
        }
        if (cl == index[1]) {
            cl <- NULL
        }
    }


    if (method == "cfe") {
        if (length(index) > 2) {
            for (extra.FE in index[3:length(index)]) {
                if (!extra.FE %in% names(data)) {
                    stop("\"index\" misspecified.\n")
                }
                if (extra.FE %in% index[1:2]) {
                    stop("\"index\" misspecified.\n")
                }
            }
        }

        if (!is.null(Q.type)) {
            if (!is.null(Q)) {
                stop("\"Q\" and \"Q.type\" cannot be used simultaneously.")
            }
            Q <- c()
            time.raw <- data[, time]
            if (is.factor(time.raw)) {
                time.num <- suppressWarnings(as.numeric(as.character(time.raw)))
            } else {
                time.num <- suppressWarnings(as.numeric(time.raw))
            }
            if (all(is.na(time.num))) {
                ## fallback: use ordered time index if not numeric
                time.num <- as.numeric(factor(time.raw, levels = sort(unique(time.raw))))
            }
            for (i in seq_along(Q.type)) {
                Q.i <- tolower(as.character(Q.type[i]))
                if (Q.i == "linear" || Q.i == "1") {
                    data[, paste(time, "1", sep = ".")] <- time.num**1
                    Q <- c(Q, paste(time, "1", sep = "."))
                } else if (Q.i == "quadratic" || Q.i == "2") {
                    data[, paste(time, "2", sep = ".")] <- time.num**2
                    Q <- c(Q, paste(time, "2", sep = "."))
                } else if (Q.i == "cubic" || Q.i == "3") {
                    data[, paste(time, "3", sep = ".")] <- time.num**3
                    Q <- c(Q, paste(time, "3", sep = "."))
                } else if (Q.i == "bspline" || Q.i == "b" || Q.i == "bs") {
                    n.t <- length(unique(time.num))
                    if (n.t < 2) {
                        stop("\"Q.type='bspline'\" requires at least 2 distinct time values.")
                    }
                    if (is.null(Q.bspline.degree)) {
                        degree.bs <- min(3, n.t - 1)
                    } else {
                        degree.bs <- suppressWarnings(as.integer(Q.bspline.degree))
                        if (length(degree.bs) != 1 || is.na(degree.bs)) {
                            stop("\"Q.bspline.degree\" must be a single numeric/integer value.")
                        }
                        degree.bs <- min(max(1L, degree.bs), n.t - 1L)
                    }
                    df.bs <- max(degree.bs + 1, min(degree.bs + 2, n.t))
                    bs.mat <- splines::bs(time.num, df = df.bs, degree = degree.bs, intercept = FALSE)
                    bs.names <- paste0(time, ".bs", seq_len(ncol(bs.mat)))
                    for (j in seq_len(ncol(bs.mat))) {
                        data[, bs.names[j]] <- bs.mat[, j]
                    }
                    Q <- c(Q, bs.names)
                } else {
                    stop(
                        "\"Q.type\" must be in c(1, 2, 3, \"bspline\") or c(\"linear\", \"quadratic\", \"cubic\", \"bspline\")."
                    )
                }
            }
        }

        if (!is.null(Zgamma)) {
            if (is.null(Z)) {
                stop("\"Z\" is not specified while \"Zgamma\" is specified.")
                # Z = unique(unname(unlist(Zgamma)))
            }
            if (is.null(gamma)) {
                gamma = names(Zgamma)
            }
        }
        if (!is.null(kappaQ)) {
            if (is.null(Q)) {
                stop("\"Q\" is not specified while \"kappaQ\" is specified.")
                # Q = unique(unname(unlist(kappaQ)))
            }
            if (is.null(kappa)) {
                kappa = names(kappaQ)
            }
        }
    }

    if (method != "cfe") {
        if (!is.null(group)) {
            data <- data[, unique(c(index, Y, D, X, W, group, cl))]
        } else {
            data <- data[, unique(c(index, Y, D, X, W, cl))] ## some variables may not be used
        }
    } else if (method == "cfe") {
        all.var <- unique(c(
            index,
            Z,
            Q,
            gamma,
            kappa,
            Y,
            D,
            X,
            W,
            group,
            cl
        ))
        data <- data[, all.var]
    }

    if (na.rm == TRUE) {
        data <- na.omit(data)
    } else {
        if (length(index) > 2) {
            for (extra.FE in index[3:length(index)]) {
                if (sum(is.na(data[, extra.FE])) >= 1) {
                    stop(
                        "\"index\" should not have missing values when setting \"na.rm\" to FALSE."
                    )
                }
            }
        }
        if (
            sum(is.na(data[, D])) >= 1 |
                sum(is.na(data[, index[1]])) >= 1 |
                sum(is.na(data[, index[2]])) >= 1
        ) {
            stop(
                "\"D\" or \"index\" should not have missing values when setting \"na.rm\" to FALSE."
            )
        }
    }

    ## placebo period
    if (placeboTest == TRUE && !is.null(placebo.period)) {
        if (sum(placebo.period > 0) > 0) {
            stop("\"placebo.period\" should not be greater than 0.")
        } else {
            if (length(placebo.period) > 2) {
                stop("\"placebo.period\" option misspecified. ")
            }
        }
    }

    ## carry-over period
    if (carryoverTest == TRUE && !is.null(carryover.period)) {
        if (sum(carryover.period <= 0) > 0) {
            stop("\"carryover.period\" should be larger than 0.")
        } else {
            if (length(carryover.period) > 2) {
                stop("\"carryover.period\" option misspecified. ")
            }
        }
    }

    ## deal with Inf and -Inf
    if (Inf %in% data[, Y] | -Inf %in% data[, Y]) {
        data[which(data[, Y] == Inf), Y] <- NA
        data[which(data[, Y] == -Inf), Y] <- NA
        warning(
            "Detect infinite values in outcome, automatically replace them with NA.\n"
        )
    }

    ## -------------------------------##
    ## Parsing raw data
    ## -------------------------------##

    ## store data and variable names
    data.old <- data
    Yname <- Y
    Dname <- D
    Xname <- X
    clname <- cl
    Wname <- W

    if (!is.null(clname)) {
        if (!clname %in% index) {
            data[, clname] <- as.numeric(as.factor(data[, clname]))
        }
    }

    ## normalize
    norm.para <- NULL
    if (normalize == TRUE) {
        sd.Y <- sd(as.matrix(data[, Yname]), na.rm = TRUE)
        data[, c(Yname, Xname)] <- data[, c(Yname, Xname)] / sd.Y
        norm.para <- sd.Y ## normalized parameter
    }

    ## check index and treatment indicator
    if (!class(data[, Dname]) %in% c("numeric", "integer")) {
        ## data[, Dname] <- as.numeric(as.character(data[, Dname]))
        stop("Treatment indicator should be a numeric value.")
    }

    ## check missingness
    # if (sum(is.na(data[, Yname])) > 0 & na.rm == TRUE) {
    #    stop(paste("Missing values in variable \"", Yname,"\".", sep = ""))
    # }
    if (sum(is.na(data[, Dname])) > 0) {
        stop(paste("Missing values in variable \"", Dname, "\".", sep = ""))
    }
    if (
        !(1 %in%
            data[, Dname] &
            0 %in% data[, Dname] &
            length(unique(data[, Dname])) == 2)
    ) {
        stop(paste("Error values in variable \"", Dname, "\".", sep = ""))
    }

    if (class(data[, index[1]])[1] == "factor") {
        data[, index[1]] <- as.character(data[, index[1]])
    }

    if (class(data[, index[2]])[1] == "factor") {
        data[, index[2]] <- as.character(data[, index[2]])
    }

    TT.old <- TT <- length(unique(data[, time]))
    N.old <- N <- length(unique(data[, id]))
    p <- length(Xname)
    id.series <- unique(sort(data[, id])) ## unit id
    time.uni <- unique(sort(data[, time])) ## period

    if (!is.null(knots)) {
        for (i in 1:length(knots)) {
            knots[i] <- which(time.uni == knots[i])
        }
    }

    ## sort data
    data <- data[order(data[, id], data[, time]), ]
    if (na.rm == FALSE) {
        # if X and Y have missing values while D doesn't have missing values, save the value of D
        # check if some units or periods are completely missing after drop missing values of X or Y
        data.full <- data[, c(Dname, id, time)]
        data <- na.omit(data)
        data.old <- data
        TT <- length(unique(data[, time]))
        N <- length(unique(data[, id]))
        if (TT != TT.old) {
            message(
                "Some periods are totally removed after drop missing values of the outcome or covariates.\n"
            )
        }
        if (N != N.old) {
            message(
                "Some units are totally removed after drop missing values of the outcome or covariates.\n"
            )
        }
        id.series <- unique(sort(data[, id])) ## unit id
        time.uni <- unique(sort(data[, time])) ## period

        # remove these dropped units or periods in data.full
        if (length(setdiff(data.full[, id], data[, id])) > 0) {
            rm.na.id <- setdiff(data.full[, id], data[, id])
            data.full <- data.full[which(!data.full[, id] %in% rm.na.id), ]
        } else {
            rm.na.id <- NULL
        }

        if (length(setdiff(data.full[, time], data[, time])) > 0) {
            rm.na.time <- setdiff(data.full[, time], data[, time])
            data.full <- data.full[which(!data.full[, time] %in% rm.na.time), ]
        } else {
            rm.na.time <- NULL
        }
        # here the size of data.full should be smaller than TT*N, larger than length(data)
    }

    ## max.missing
    if (is.null(max.missing)) {
        max.missing <- TT
    }

    if (method == "cfe") {
        if (length(index) > 2) {
            for (extra.FE in index[3:length(index)]) {
                data[, extra.FE] <- as.numeric(as.factor(data[, extra.FE]))
            }
        }
    }

    ## gen group matrix
    if (!is.null(group)) {
        rawgroup <- data[, group]
        newgroup <- as.numeric(as.factor(rawgroup))
        data[, group] <- newgroup
        rawgroup <- cbind.data.frame(rawgroup, newgroup)
        rawgroup <- rawgroup[!duplicated(rawgroup[, 1]), ]
    }

    ## message("\nOK1\n")
    ## check variation in x
    if (p > 0) {
        for (i in 1:p) {
            if (sum(is.na(data[, Xname[i]])) > 0 & na.rm == TRUE) {
                stop(paste(
                    "Missing values in variable \"",
                    Xname[i],
                    "\".",
                    sep = ""
                ))
            }
            if (force %in% c(1, 3)) {
                if (
                    sum(
                        tapply(data[, Xname[i]], data[, id], var),
                        na.rm = TRUE
                    ) ==
                        0
                ) {
                    stop(paste(
                        "Variable \"",
                        Xname[i],
                        "\" is unit-invariant. Try to remove it.",
                        sep = ""
                    ))
                }
            }
            if (force %in% c(2, 3)) {
                if (
                    sum(
                        tapply(data[, Xname[i]], data[, time], var),
                        na.rm = TRUE
                    ) ==
                        0
                ) {
                    stop(paste(
                        "Variable \"",
                        Xname[i],
                        "\" is time-invariant. Try to remove it.",
                        sep = ""
                    ))
                }
            }
        }
    }

    ## check index
    if (sum(is.na(data[, id])) > 0) {
        stop(paste("Missing values in variable \"", id, "\".", sep = ""))
    }
    if (sum(is.na(data[, time])) > 0) {
        stop(paste("Missing values in variable \"", time, "\".", sep = ""))
    }

    ## check balanced panel and fill unbalanced panel
    if (dim(data)[1] < TT * N) {
        data[, time] <- as.numeric(as.factor(data[, time]))
        data[, id] <- as.numeric(as.factor(data[, id]))

        ob.indicator <- data[, time]
        id.indicator <- table(data[, id])
        sub.start <- 1
        for (i in 1:(N - 1)) {
            sub.start <- sub.start + id.indicator[i]
            sub.end <- sub.start + id.indicator[i + 1] - 1
            ob.indicator[sub.start:sub.end] <- ob.indicator[sub.start:sub.end] +
                i * TT
        }

        variable <- c(Yname, Dname, Xname, id, time)

        if (!is.null(group)) {
            variable <- c(variable, group)
        }

        if (!is.null(W)) {
            variable <- c(variable, Wname)
        }

        if (method == "cfe") {
            if (length(index) > 2) {
                variable <- unique(c(
                    index[3:length(index)],
                    Z,
                    Q,
                    gamma,
                    kappa,
                    variable
                ))
            } else {
                variable <- unique(c(Z, Q, gamma, kappa, variable))
            }
        }

        if (!is.null(cl)) {
            variable <- unique(c(variable, cl))
        }

        data_I <- matrix(0, N * TT, 1)
        data_I[ob.indicator, 1] <- 1
        data_ub <- as.matrix(data[, variable])
        data <- data_ub_adj(data_I, data_ub)
        colnames(data) <- variable
        ## data is a TT*N matrix filled with observed pairs (Y/X, D).

        ## if these exists observations whose Y/X is missing but D is observed.
        if (na.rm == FALSE) {
            data.full[, time] <- as.numeric(as.factor(data.full[, time]))
            ob.indicator.full <- data.full[, time]
            id.indicator.full <- table(data.full[, id])
            sub.start <- 1
            for (i in 1:(N - 1)) {
                sub.start <- sub.start + id.indicator.full[i]
                sub.end <- sub.start + id.indicator.full[i + 1] - 1
                ob.indicator.full[sub.start:sub.end] <- ob.indicator.full[
                    sub.start:sub.end
                ] +
                    i * TT
            }
            data_I.full <- matrix(0, N * TT, 1)
            data_I.full[ob.indicator.full, 1] <- 1
            data_ub.full <- as.matrix(data.full[, Dname])
            data.D.full <- data_ub_adj(data_I.full, data_ub.full)
            colnames(data.D.full) <- Dname
            ## replace D in data with D in data.full
            data[, Dname] <- data.D.full[, Dname]
        }
    }

    # recode treatment status for osbervations exposed to carryover effects
    hasCarryover <- 0
    if (!is.null(carryover.rm)) {
        if (length(carryover.rm) == 1 & class(carryover.rm)[1] == "numeric") {
            if (carryover.rm > 0) {
                newT <- c(1:TT)
                data <- data[order(data[, id], data[, time]), ]
                tempID <- unique(data[, id])
                for (i in tempID) {
                    subpos <- which(data[, id] == i)
                    subtime <- data[subpos, time]
                    subd <- data[subpos, Dname]
                    if (sum(subd) >= 1) {
                        tr.time <- subtime[which(subd == 1)]
                        cr.time <- c() # carryover period
                        for (k in 1:carryover.rm) {
                            cr.time <- c(cr.time, tr.time + k)
                        }
                        # note: if a period has both treatment effect and carryover effect,
                        # regard carryover effect as 0
                        cr.time <- unique(cr.time)
                        cr.time <- setdiff(cr.time, tr.time)

                        cr.pos <- subpos[which(subtime %in% cr.time)]

                        if (length(cr.pos) > 0) {
                            data[cr.pos, Dname] <- 2
                        }
                    }
                }
            }
        }
    }

    if (2 %in% data[, Dname]) {
        hasCarryover <- 1
    }

    ## indicator matrix: index matrix that indicates if data is observed
    I.D <- I <- matrix(1, TT, N)
    Y.ind <- matrix(data[, Yname], TT, N)
    D.ind <- matrix(data[, Dname], TT, N)
    I[is.nan(Y.ind)] <- 0
    I.D[is.nan(D.ind)] <- 0
    ## I has more zeros than I.D
    ## I.D is used in the function get_term

    if (0 %in% I) {
        data[is.nan(data)] <- 0
    }

    ## group indicator
    G.old <- G <- NULL
    if (!is.null(group)) {
        G <- matrix(data[, group], TT, N)
        ## replace group index 0(missing) for each unit
        if (!0 %in% G) {
            if (sum(apply(G, 2, var)) > 0) {
                stop(
                    "A unit in different periods should have the same group index.\n"
                )
            }
        } else {
            if (
                max(apply(G, 2, function(vec) {
                    return(length(table(vec)))
                })) >
                    2
            ) {
                stop(
                    "A unit in different periods should have the same group index.\n"
                )
            }
            G <- apply(G, 2, function(vec) {
                return(rep(max(vec), length(vec)))
            })
        }
        G.old <- G
    }

    if (!is.null(W)) {
        W <- matrix(data[, Wname], TT, N)
    }

    if (!is.null(cl)) {
        cl <- matrix(data[, clname], TT, N)
        # for each column, replace 0 with the mean of non-zero values
        cl <- apply(cl, 2, function(column) {
            column[column == 0] <- mean(column[column != 0])
            return(column)
        })
    }

    ## treatment indicator: incorporates reversal treatments
    ## D==1 -> treatment
    D.origin <- D <- matrix(data[, Dname], TT, N)
    ## outcome variable
    Y <- matrix(data[, Yname], TT, N)
    ## time-varying covariates
    X <- array(0, dim = c(TT, N, p))
    if (p > 0) {
        for (i in 1:p) {
            X[,, i] <- matrix(data[, Xname[i]], TT, N)
        }
    }

    index.matrix <- list()

    data.colnames <- colnames(data)

    X.extra.FE <- array(0, dim = c(TT, N, length(index) - 2))

    X.Z <- array(0, dim = c(TT, N, length(Z)))
    if (is.null(gamma)) {
        if (!is.null(Z)) {
            gamma <- index[2]
            X.gamma <- array(0, dim = c(TT, N, 1))
            if (is.null(Zgamma)) {
                Zgamma <- list()
                Zgamma[[gamma]] <- Z
            }
        } else {
            gamma <- NULL # Explicitly set gamma to NULL when Z is NULL
            X.gamma <- array(0, dim = c(TT, N, 0))
            if (is.null(Zgamma)) {
                Zgamma <- list()
            }
        }
    } else {
        if (is.null(Z)) {
            stop("\"gamma\" cannot be specified while \"Z\" is not specified.")
        } else {
            X.gamma <- array(0, dim = c(TT, N, length(gamma)))
            if (is.null(Zgamma)) {
                Zgamma <- list()
                Zgamma[[gamma]] <- Z
            }
        }
    }

    X.Q <- array(0, dim = c(TT, N, length(Q)))
    if (is.null(kappa)) {
        if (!is.null(Q)) {
            kappa <- index[1]
            X.kappa <- array(0, dim = c(TT, N, 1))
            if (is.null(kappaQ)) {
                kappaQ <- list()
                kappaQ[[kappa]] <- Q
            }
        } else {
            kappa <- NULL # Explicitly set kappa to NULL when Q is NULL
            X.kappa <- array(0, dim = c(TT, N, 0))
            if (is.null(kappaQ)) {
                kappaQ <- list()
            }
        }
    } else {
        if (is.null(Q)) {
            stop("\"kappa\" cannot be specified while \"Q\" is not specified.")
        } else {
            X.kappa <- array(0, dim = c(TT, N, length(kappa)))
            if (is.null(kappaQ)) {
                kappaQ <- list()
                kappaQ[[kappa]] <- Q
            }
        }
    }

    if (method == "cfe") {
        if (length(index) > 2) {
            for (i in 1:(length(index) - 2)) {
                X.extra.FE[,, i] <- matrix(data[, index[i + 2]], TT, N)
            }
        }

        if (!is.null(Z)) {
            for (i in 1:length(Z)) {
                X.Z[,, i] <- matrix(data[, Z[i]], TT, N)
            }
        }

        if (!is.null(Q)) {
            for (i in 1:length(Q)) {
                X.Q[,, i] <- matrix(data[, Q[i]], TT, N)
            }
        }

        if (length(gamma) > 1) {
            for (i in 1:length(gamma)) {
                X.gamma[,, i] <- matrix(data[, gamma[i]], TT, N)
            }
        } else if (length(gamma) == 1) {
            X.gamma[,, 1] <- matrix(data[, gamma], TT, N)
        }

        if (length(kappa) > 1) {
            for (i in seq_along(kappa)) {
                kappa.val <- data[, kappa[i]]
                if (!is.numeric(kappa.val)) {
                    kappa.val <- suppressWarnings(as.numeric(as.character(kappa.val)))
                    if (all(is.na(kappa.val))) {
                        kappa.val <- as.numeric(factor(data[, kappa[i]]))
                    }
                }
                X.kappa[,, i] <- matrix(kappa.val, TT, N)
            }
        } else if (length(kappa) == 1) {
            kappa.val <- data[, kappa]
            if (!is.numeric(kappa.val)) {
                kappa.val <- suppressWarnings(as.numeric(as.character(kappa.val)))
                if (all(is.na(kappa.val))) {
                    kappa.val <- as.numeric(factor(data[, kappa]))
                }
            }
            X.kappa[,, 1] <- matrix(kappa.val, TT, N)
        }

        Zgamma.id <- list()
        for (key in names(Zgamma)) {
            value <- c()
            for (i in 1:length(Zgamma[[key]])) {
                value <- c(value, which(Z == Zgamma[[key]][i]))
            }
            Zgamma.id[[which(gamma == key)]] <- value
        }

        kappaQ.id <- list()
        for (key in names(kappaQ)) {
            value <- c()
            for (i in 1:length(kappaQ[[key]])) {
                value <- c(value, which(Q == kappaQ[[key]][i]))
            }
            kappaQ.id[[which(kappa == key)]] <- value
        }
    } else {
        Zgamma.id <- list()
        kappaQ.id <- list()
    }

    ## ----------------------------------------------------------- ##
    II <- I
    II[which(D == 1)] <- 0 ## regard treated values as missing

    II.cm <- matrix(0, TT, N)
    II.cm[which(D == 1)] <- 1
    II.cm[is.nan(Y.ind)] <- 0

    # Unbalance Check
    ## 1. remove units that have too control status
    T0 <- apply(II, 2, sum)
    T0.min <- min(T0)

    if (time.component.from == "nevertreated") {
        ## Validate: enough never-treated units for factor estimation
        D.cum <- apply(D, 2, function(vec) { cummax(vec) })
        D.unit.sum <- colSums(D.cum)
        co.never <- which(D.unit.sum == 0)

        if (length(co.never) == 0) {
            stop("\"time.component.from\" is set to \"nevertreated\" but no never-treated units found in the data.")
        }
        if (length(co.never) < r + 1) {
            stop(paste0("\"time.component.from\" is set to \"nevertreated\" but only ",
                         length(co.never), " never-treated units found. Need at least ",
                         r + 1, " for r = ", r, "."))
        }

        if (!(method %in% c("gsynth", "ife") ||
              (method == "cfe" && time.component.from == "nevertreated"))) {
            ## Zero out II for methods that don't do their own co/tr split
            ## (gsynth, ife+nevertreated, cfe+nevertreated route to fect_nevertreated which splits internally)
            ever.treated <- setdiff(1:ncol(II), co.never)
            II[, ever.treated] <- 0
        }
    }

    # Unbalance Check
    ## 1. remove units that have too control status

    if (sum(T0[which(apply(D, 2, sum) > 0)] >= min.T0) == 0) {
        stop(
            "All treated units have been removed. Please specify a smaller min.T0.\n"
        )
    }
    ## T0.min : minimum T0
    ## min.T0: manually set
    ## rm.tr.id: relative location of treated units (within all treated units)
    ## that will be removed
    if (T0.min < min.T0) {
        message(paste0(
            "For identification purposes, units whose number of untreated periods <",
            min.T0,
            " are dropped automatically.\n"
        ))
    }

    rm.id <- sort(unique(c(
        which((TT - apply(I, 2, sum)) > max.missing),
        which(T0 < min.T0)
    )))
    ## rm.id <- which(T0 < min.T0) ## removed id
    ## rem.id <- which(T0 >= min.T0) ## remaining id
    rem.id <- setdiff(1:N, rm.id)

    if (length(rm.id) == N) {
        stop("All units have been removed.\n")
    }

    if (length(rm.id) > 0) {
        n.keep <- length(rem.id)
        if (n.keep <= 10) {
            pct.keep <- round(100 * n.keep / N, 1)
            message(paste0(
                "After dropping units (max.missing/min.T0): retained ",
                pct.keep, "% of units.\n"
            ))
        }
    }

    if (length(rm.id) > 0) {
        X.old <- X
        if (p > 0) {
            X <- array(0, dim = c(TT, (N - length(rm.id)), p))
            for (i in 1:p) {
                subX <- X.old[,, i]
                X[,, i] <- as.matrix(subX[, -rm.id])
            }
        } else {
            X <- array(0, dim = c(TT, (N - length(rm.id)), 0))
        }

        # N <- N - length(rm.id)
        Y <- as.matrix(Y[, -rm.id])
        D <- as.matrix(D[, -rm.id])
        I <- as.matrix(I[, -rm.id]) ## after removing
        I.D <- as.matrix(I.D[, -rm.id])
        II <- as.matrix(II[, -rm.id])
        II.cm <- as.matrix(II.cm[, -rm.id])
        if (!is.null(group)) {
            G <- as.matrix(G[, -rm.id])
        }
        if (!is.null(cl)) {
            cl <- as.matrix(cl[, -rm.id])
        }
        if (!is.null(W)) {
            W <- as.matrix(W[, -rm.id])
        }
        if (method == "cfe") {
            rm.id.diff <- setdiff(1:dim(X.extra.FE)[2], rm.id)
            X.extra.FE <- X.extra.FE[, rm.id.diff, , drop = FALSE]
            X.Z <- X.Z[, rm.id.diff, , drop = FALSE]
            X.Q <- X.Q[, rm.id.diff, , drop = FALSE]
            X.gamma <- X.gamma[, rm.id.diff, , drop = FALSE]
            X.kappa <- X.kappa[, rm.id.diff, , drop = FALSE]
        }
    }

    ## message("\nOK1\n")

    ## 2. check if some periods when all units are missing or treated
    I.use <- apply(II, 1, sum)
    if (0 %in% I.use) {
        for (i in 1:TT) {
            if (I.use[i] == 0) {
                message(
                    "\nThere are not any observations under control at ",
                    time.uni[i],
                    ", drop that period.\n"
                )
            }
        }
        TT <- TT - sum(I.use == 0)
        time.uni <- time.uni[-which(I.use == 0)]

        I <- I[-which(I.use == 0), ] ## remove that period
        I.D <- I.D[-which(I.use == 0), ]
        II <- II[-which(I.use == 0), ] ## remove that period
        II.cm <- II.cm[-which(I.use == 0), ] ## remove that period
        D <- D[-which(I.use == 0), ] ## remove that period
        Y <- Y[-which(I.use == 0), ] ## remove that period

        if (!is.null(group)) {
            G <- G[-which(I.use == 0), ]
        }

        if (!is.null(W)) {
            W <- W[-which(I.use == 0), ]
        }

        if (!is.null(cl)) {
            cl <- cl[-which(I.use == 0), ]
        }

        if (method == "cfe") {
            X.extra.FE <- X.extra.FE[-which(I.use == 0), , , drop = FALSE]
            X.Z <- X.Z[-which(I.use == 0), , , drop = FALSE]
            X.Q <- X.Q[-which(I.use == 0), , , drop = FALSE]
            X.gamma <- X.gamma[-which(I.use == 0), , , drop = FALSE]
            X.kappa <- X.kappa[-which(I.use == 0), , , drop = FALSE]
        }

        X.old <- X
        if (p > 0) {
            X <- array(0, dim = c(TT, (N - length(rm.id)), p))
            for (i in 1:p) {
                subX <- X.old[,, i]
                X[,, i] <- as.matrix(subX[-which(I.use == 0), ])
            }
        } else {
            X <- array(0, dim = c(TT, (N - length(rm.id)), 0))
        }
    }

    ## message("\nOK2\n")
    ## 3. relative period
    T.on <- matrix(NA, TT, (N - length(rm.id)))
    D1 <- D2 <- NULL
    T.on.carry <- NULL
    if (hasCarryover == 0) {
        for (i in 1:(N - length(rm.id))) {
            T.on[, i] <- get_term(D[, i], I.D[, i], type = "on")
        }
    } else {
        # separate T.on and T.on.carry
        D1 <- D2 <- D
        D1[which(D1 == 2)] <- 0
        D2[which(D2 == 1)] <- 0
        D2[which(D2 == 2)] <- 1

        T.on.carry <- matrix(NA, TT, (N - length(rm.id)))

        for (i in 1:(N - length(rm.id))) {
            T.on[, i] <- get_term(D1[, i], I.D[, i], type = "on")
            T.on.carry[, i] <- get_term(D2[, i], I.D[, i], type = "on")
        }
        T.on[which(D == 2)] <- NA ## remove carryover effect
        T.on.carry[which(T.on.carry <= 0)] <- NA ## only keep carryover effect
    }
    rm(D1, D2)
    # calendar.time <- as.matrix(replicate((N - length(rm.id)), c(time.uni)))

    ## 3.1 balance samples
    ## for balance group, add group indicator
    T.on.balance <- matrix(NA, TT, (N - length(rm.id)))
    if (!is.null(balance.period)) {
        if (fill.missing == FALSE) {
            T.on.miss <- T.on
            T.on.miss[which(I == 0)] <- NA
            T.on.balance <- apply(T.on.miss, 2, function(x) {
                v_replace(balance.periods, x)
            })
        }
        if (fill.missing == TRUE) {
            T.on.balance <- apply(T.on, 2, function(x) {
                v_replace(balance.periods, x)
            })
        }
        if (sum(!is.na(T.on.balance)) == 0) {
            stop("No Balanced Sample Found.\n")
        }
    }

    ## 4. check reversals
    D1 <- D
    if (hasCarryover == 1) {
        D1[which(D == 2)] <- 1
    }
    D.fake <- apply(D1, 2, function(vec) {
        cumsum(vec)
    })
    D.fake <- ifelse(D.fake > 0, 1, 0)
    D.fake[which(I.D == 0)] <- 0
    Nrev <- sum(apply(D.fake == D1, 2, sum) != TT)
    hasRevs <- ifelse(Nrev > 0, 1, 0)
    if (hasRevs == FALSE & carryoverTest == TRUE) {
        stop(
            "Treatment status have no reversals. Cannot perform \"carryoverTest\" in this case."
        )
    }
    if (hasRevs == TRUE & method == "gsynth") {
        stop("Gsynth can't be used when treatments have reversals.")
    }
    if (se == 1 && vartype == "parametric" && hasRevs == TRUE) {
        stop(
            "Parametric bootstrap is not valid when treatment reversal is present. ",
            "Use vartype='bootstrap' or 'jackknife'."
        )
    }
    ## Parametric bootstrap requires nevertreated control-pool isolation.
    ## Placed AFTER the reversal-check gate so reversal users get the more
    ## actionable reversal message first. Uses method_arg (saved before the
    ## silent fe -> ife coercion) so the message names the user's literal
    ## method argument.
    if (se == 1 && vartype == "parametric" && time.component.from != "nevertreated") {
        stop(sprintf(
            paste0(
                "vartype = \"parametric\" requires time.component.from = \"nevertreated\".\n",
                "  Your call: method = \"%s\", time.component.from = \"%s\".\n\n",
                "The parametric pseudo-treated bootstrap requires a control pool ",
                "isolated from treated-unit pre-treatment cells. Pass ",
                "time.component.from = \"nevertreated\" (if never-treated controls ",
                "exist) or use vartype = \"bootstrap\" or \"jackknife\"."
            ),
            method_arg, time.component.from
        ), call. = FALSE)
    }
    ## para.error = "empirical" or "wild" requires fully-observed panel.
    if (se == 1 && vartype == "parametric" &&
        para.error %in% c("empirical", "wild") &&
        (0 %in% I)) {
        stop(
            "para.error = \"", para.error, "\" requires a fully-observed panel ",
            "(no missing cells in the observation matrix I). ",
            "The current data has unobserved cells. ",
            "Use para.error = \"ar\" or para.error = \"auto\" (which selects \"ar\" automatically ",
            "when missing data are present).",
            call. = FALSE
        )
    }

    ## 5. switch-off periods
    T.off <- NULL
    if (hasRevs == 1) {
        T.off <- matrix(NA, TT, (N - length(rm.id)))
        for (i in 1:(N - length(rm.id))) {
            T.off[, i] <- get_term(D1[, i], I.D[, i], type = "off")
        }
    }
    rm(D1)

    ## 6. regard placebo period as under treatment
    if (placeboTest == TRUE) {
        II.origin <- II
        if (length(placebo.period) == 1) {
            placebo.pos <- which(T.on == placebo.period)
            II[placebo.pos] <- 0
        } else {
            placebo.pos <- which(
                T.on >= placebo.period[1] & T.on <= placebo.period[2]
            )
            II[placebo.pos] <- 0
        }

        ## remove treated units that have too few observations
        T0.2 <- apply(II, 2, sum)

        if (sum(T0.2[which(apply(D, 2, sum) > 0)] >= min.T0) == 0) {
            stop("All treated units have been removed in placebo test.\n")
        }

        rm.id.2.pos <- sort(which(T0.2 < min.T0))
        rm.id.2 <- rem.id[rm.id.2.pos]
        rem.id.2 <- setdiff(rem.id, rm.id.2)

        rem.id <- rem.id.2
        rm.id <- setdiff(1:N, rem.id)

        if (length(rm.id.2) > 0) {
            X.old <- X
            if (p > 0) {
                X <- array(0, dim = c(TT, (N - length(rm.id)), p))
                for (i in 1:p) {
                    subX <- X.old[,, i]
                    X[,, i] <- as.matrix(subX[, -rm.id.2.pos])
                }
            } else {
                X <- array(0, dim = c(TT, (N - length(rm.id)), 0))
            }

            # N <- N - length(rm.id)
            Y <- as.matrix(Y[, -rm.id.2.pos])
            D <- as.matrix(D[, -rm.id.2.pos])
            I <- as.matrix(I[, -rm.id.2.pos]) ## after removing
            I.D <- as.matrix(I.D[, -rm.id.2.pos])
            II <- as.matrix(II[, -rm.id.2.pos])
            II.origin <- as.matrix(II.origin[, -rm.id.2.pos])
            T.on <- as.matrix(T.on[, -rm.id.2.pos])
            if (hasCarryover) {
                T.on.carry <- as.matrix(T.on.carry[, -rm.id.2.pos])
            }
            if (hasRevs) {
                T.off <- as.matrix(T.off[, -rm.id.2.pos])
            }
            if (!is.null(group)) {
                G <- as.matrix(G[, -rm.id.2.pos])
            }
            if (!is.null(W)) {
                W <- as.matrix(W[, -rm.id.2.pos])
            }
            if (!is.null(cl)) {
                cl <- as.matrix(cl[, -rm.id.2.pos])
            }

            if (method == "cfe") {
                rm.id.2.diff <- setdiff(1:dim(X.extra.FE)[2], rm.id.2.pos)
                X.extra.FE <- X.extra.FE[, rm.id.2.diff, , drop = FALSE]
                X.Z <- X.Z[, rm.id.2.diff, , drop = FALSE]
                X.Q <- X.Q[, rm.id.2.diff, , drop = FALSE]
                X.gamma <- X.gamma[, rm.id.2.diff, , drop = FALSE]
                X.kappa <- X.kappa[, rm.id.2.diff, , drop = FALSE]
            }
        }
    }

    ## 7. Carryover Test
    ## testcarryover.period = c(1,3)
    if (
        hasRevs == 1 &
            carryoverTest == TRUE &
            is.null(carryover.period) == FALSE
    ) {
        II.origin <- II
        if (length(carryover.period) == 1) {
            carryover.pos <- which(T.off == carryover.period)
            II[carryover.pos] <- 0
        } else {
            carryover.pos <- which(
                T.off >= carryover.period[1] & T.off <= carryover.period[2]
            )
            II[carryover.pos] <- 0
        }
        ## remove treated units that have too few observations
        T0.3 <- apply(II, 2, sum)

        if (sum(T0.3[which(apply(D, 2, sum) > 0)] >= min.T0) == 0) {
            stop("All treated units have been removed in carryover test.\n")
        }

        rm.id.3.pos <- sort(which(T0.3 < min.T0))
        rm.id.3 <- rem.id[rm.id.3.pos]
        rem.id.3 <- setdiff(rem.id, rm.id.3)

        rem.id <- rem.id.3
        rm.id <- setdiff(1:N, rem.id)

        if (length(rm.id.3) > 0) {
            X.old <- X
            if (p > 0) {
                X <- array(0, dim = c(TT, (N - length(rm.id)), p))
                for (i in 1:p) {
                    subX <- X.old[,, i]
                    X[,, i] <- as.matrix(subX[, -rm.id.3.pos])
                }
            } else {
                X <- array(0, dim = c(TT, (N - length(rm.id)), 0))
            }

            # N <- N - length(rm.id)
            Y <- as.matrix(Y[, -rm.id.3.pos])
            D <- as.matrix(D[, -rm.id.3.pos])
            I <- as.matrix(I[, -rm.id.3.pos]) ## after removing
            I.D <- as.matrix(I.D[, -rm.id.3.pos])
            II <- as.matrix(II[, -rm.id.3.pos])
            II.origin <- as.matrix(II.origin[, -rm.id.3.pos])
            T.on <- as.matrix(T.on[, -rm.id.3.pos])
            if (hasCarryover) {
                T.on.carry <- as.matrix(T.on.carry[, -rm.id.3.pos])
            }
            if (hasRevs) {
                T.off <- as.matrix(T.off[, -rm.id.3.pos])
            }
            if (!is.null(group)) {
                G <- as.matrix(G[, -rm.id.3.pos])
            }
            if (!is.null(W)) {
                W <- as.matrix(W[, -rm.id.3.pos])
            }
            if (!is.null(cl)) {
                cl <- as.matrix(cl[, -rm.id.3.pos])
            }
            if (method == "cfe") {
                rm.id.3.diff <- setdiff(1:dim(X.extra.FE)[2], rm.id.3.pos)
                X.extra.FE <- X.extra.FE[, rm.id.3.diff, , drop = FALSE]
                X.Z <- X.Z[, rm.id.3.diff, , drop = FALSE]
                X.Q <- X.Q[, rm.id.3.diff, , drop = FALSE]
                X.gamma <- X.gamma[, rm.id.3.diff, , drop = FALSE]
                X.kappa <- X.kappa[, rm.id.3.diff, , drop = FALSE]
            }
        }
    }

    ## recover treatment indicators
    D.pos <- NULL
    ## DD <- D
    if (hasCarryover) {
        D.pos <- which(D == 2)
        if (length(D.pos) > 0) {
            D[D.pos] <- 0
        }
    }

    ## 8. Finally, check enough observations
    if (min(apply(II, 1, sum)) == 0) {
        if (placeboTest == 1) {
            stop(
                "Some periods do not have any observations. Please set a smaller range for placebo period."
            )
        } else if (carryoverTest == 1) {
            stop(
                "Some periods do not have any observations. Please set a smaller range for carryover period."
            )
        } else {
            stop("Some periods do not have any observations.")
        }
    }

    if (time.component.from != "nevertreated" && min(apply(II, 2, sum)) == 0) {
        if (placeboTest == 1) {
            stop(
                "Some units do not have any observations. Please set a smaller range for placebo period."
            )
        } else if (carryoverTest == 1) {
            stop(
                "Some units do not have any observations. Please set a smaller range for carryover period."
            )
        } else {
            stop("Some units do not have any untreated observations.")
        }
    }
    ## cohort
    g.level <- NULL
    if (!is.null(group)) {
        g.level <- unique(c(G))
        g.level <- g.level[!is.na(g.level)]
        rownames(rawgroup) <- rawgroup[, "newgroup"]
        names(g.level) <- rawgroup[as.character(g.level), "rawgroup"]
        g.level <- sort(g.level)
    }

    ## -------------------------------##
    ## Register clusters
    ## -------------------------------##

    old.future.plan <- NULL

    if ((se == TRUE | permute == TRUE) & !do_parallel_boot) {
        ## set seed
        if (is.null(seed) == FALSE) {
            set.seed(seed + 1)
        }
    }

    if ((se == TRUE | permute == TRUE) & do_parallel_boot) {
        ## set seed
        if (is.null(seed) == FALSE) {
            set.seed(seed)
        }
        if (is.null(cores) == TRUE) {
            cores <- max(1L, min(parallelly::availableCores(omit = 2L), 8L))
        }
        old.future.plan <- future::plan()
        suppressWarnings(suppressPackageStartupMessages({
            ## v2.4.2+: route through .fect_make_future_cluster so workers
            ## pre-load mvtnorm / future / etc. silently. Avoids the
            ## "package was built under R version X.Y.Z" warnings firing
            ## per worker on first parallel package use (especially under
            ## parametric bootstrap where mvtnorm::rmvnorm fires on every
            ## worker for every replicate).
            future::plan(future::cluster,
                         workers = .fect_make_future_cluster(cores))
            doFuture::registerDoFuture()
        }))
        if (is.null(seed) == FALSE) {
            registerDoRNG(seed)
        }
        avail <- parallelly::availableCores()
        msg_line <- sprintf("Parallel computing: using %d of %d available cores.", cores, avail)
        pad <- strrep(" ", max(0, 56 - nchar(msg_line)))
        message("\n",
            " +----------------------------------------------------------+\n",
            " | ", msg_line, pad, " |\n",
            " |                                                          |\n",
            " | To change: set cores = <n> in fect().                    |\n",
            " | Default: min(available - 2, 8).                          |\n",
            " +----------------------------------------------------------+\n")
    }

    ## -------------------------------##
    ## run main program
    ## -------------------------------##
    if (se == FALSE) {
        if (CV == TRUE) {
            if (binary == FALSE) {
                out <- fect_cv(
                    Y = Y,
                    D = D,
                    X = X,
                    W = W,
                    W.in.fit = use.W.in.fit,
                    I = I,
                    II = II,
                    T.on = T.on,
                    T.off = T.off,
                    T.on.carry = T.on.carry,
                    T.on.balance = T.on.balance,
                    balance.period = balance.period,
                    method = method,
                    criterion = criterion,
                    k = k,
                    cv.prop = cv.prop,
                    cv.method = cv.method,
                    cv.nobs = cv.nobs,
                    cv.donut = cv.donut,
                    cv.buffer = cv.buffer,
                    min.T0 = min.T0,
                    r = r,
                    r.end = r.end,
                    proportion = proportion,
                    nlambda = nlambda,
                    lambda = lambda,
                    force = force,
                    hasRevs = hasRevs,
                    tol = tol,
                    max.iteration = max.iteration,
                    norm.para = norm.para,
                    group.level = g.level,
                    group = G,
                    time.component.from = time.component.from,
                    X.extra.FE = X.extra.FE,
                    X.Z = X.Z,
                    X.Q = X.Q,
                    X.gamma = X.gamma,
                    X.kappa = X.kappa,
                    Zgamma.id = Zgamma.id,
                    kappaQ.id = kappaQ.id,
                    parallel = parallel,
                    cores = cores,
                    do_parallel_cv   = do_parallel_cv,
                    do_parallel_boot = do_parallel_boot,
                    cv.rule = cv.rule
                )
            } else {
                out <- fect_binary_cv(
                    Y = Y,
                    D = D,
                    X = X,
                    I = I,
                    II = II,
                    T.on = T.on,
                    T.off = T.off,
                    k = k,
                    cv.prop = cv.prop,
                    cv.method = cv.method,
                    cv.nobs = cv.nobs,
                    cv.buffer = cv.buffer,
                    min.T0 = min.T0,
                    r = r,
                    r.end = r.end,
                    QR = QR,
                    force = force,
                    hasRevs = hasRevs,
                    tol = tol,
                    group.level = g.level,
                    group = G
                )
            }
        } else {
            ## non-binary case
            if (method == "ife" && time.component.from == "nevertreated") {
                ## nevertreated: route to fect_nevertreated (the nevertreated estimator)
                out <- fect_nevertreated(
                    Y = Y,
                    D = D,
                    X = X,
                    W = W,
                    W.in.fit = use.W.in.fit,
                    I = I,
                    II = II,
                    cm=cm,
                    II.cm = II.cm,
                    T.on = T.on,
                    T.off = T.off,
                    r = r,
                    CV = 0,
                    T.on.balance = T.on.balance,
                    balance.period = balance.period,
                    binary = binary,
                    QR = QR,
                    force = force,
                    hasRevs = hasRevs,
                    tol = tol,
                    max.iteration = max.iteration,
                    boot = 0,
                    norm.para = norm.para,
                    placeboTest = placeboTest,
                    placebo.period = placebo.period,
                    carryoverTest = carryoverTest,
                    carryover.period = carryover.period,
                    group.level = g.level,
                    group = G,
                    parallel = parallel,
                    cores = cores,
                    loading.bound      = loading.bound,
                    gamma.loading      = gamma.loading,
                    gamma.loading.grid = gamma.loading.grid
                )
            } else if (method == "ife") {
                out <- fect_fe(
                    Y = Y,
                    D = D,
                    X = X,
                    W = W,
                    W.in.fit = use.W.in.fit,
                    I = I,
                    II = II,
                    T.on = T.on,
                    T.off = T.off,
                    r.cv = r,
                    T.on.carry = T.on.carry,
                    T.on.balance = T.on.balance,
                    balance.period = balance.period,
                    binary = binary,
                    QR = QR,
                    force = force,
                    hasRevs = hasRevs,
                    tol = tol,
                    max.iteration = max.iteration,
                    boot = 0,
                    norm.para = norm.para,
                    placeboTest = placeboTest,
                    placebo.period = placebo.period,
                    carryoverTest = carryoverTest,
                    carryover.period = carryover.period,
                    group.level = g.level,
                    group = G
                )
            } else if (method == "cfe" && time.component.from == "nevertreated") {
                ## cfe + nevertreated: route to fect_nevertreated with method="cfe"
                out <- fect_nevertreated(
                    Y = Y,
                    D = D,
                    X = X,
                    W = W,
                    W.in.fit = use.W.in.fit,
                    I = I,
                    II = II,
                    T.on = T.on,
                    T.off = T.off,
                    r = r,
                    CV = 0,
                    T.on.balance = T.on.balance,
                    balance.period = balance.period,
                    binary = binary,
                    QR = QR,
                    force = force,
                    hasRevs = hasRevs,
                    tol = tol,
                    max.iteration = max.iteration,
                    boot = 0,
                    norm.para = norm.para,
                    placeboTest = placeboTest,
                    placebo.period = placebo.period,
                    carryoverTest = carryoverTest,
                    carryover.period = carryover.period,
                    group.level = g.level,
                    group = G,
                    ## CFE-specific parameters
                    method = "cfe",
                    X.extra.FE = X.extra.FE,
                    X.Z = X.Z,
                    X.Q = X.Q,
                    X.gamma = X.gamma,
                    X.kappa = X.kappa,
                    Zgamma.id = Zgamma.id,
                    kappaQ.id = kappaQ.id,
                    parallel = parallel,
                    cores = cores
                )
            } else if (method == "cfe") {
                out <- fect_cfe(
                    Y = Y,
                    D = D,
                    X = X,
                    W = W,
                    W.in.fit = use.W.in.fit,
                    X.extra.FE = X.extra.FE,
                    X.Z = X.Z,
                    X.Q = X.Q,
                    X.gamma = X.gamma,
                    X.kappa = X.kappa,
                    Zgamma.id = Zgamma.id,
                    kappaQ.id = kappaQ.id,
                    I = I,
                    II = II,
                    T.on = T.on,
                    T.off = T.off,
                    r.cv = r,
                    T.on.carry = T.on.carry,
                    T.on.balance = T.on.balance,
                    balance.period = balance.period,
                    binary = FALSE,
                    QR = QR,
                    force = force,
                    hasRevs = hasRevs,
                    tol = tol,
                    max.iteration = max.iteration,
                    boot = 0,
                    norm.para = norm.para,
                    placeboTest = placeboTest,
                    placebo.period = placebo.period,
                    carryoverTest = carryoverTest,
                    carryover.period = carryover.period,
                    group.level = g.level,
                    group = G
                )
            } else if (method == "gsynth") {
                out <- fect_nevertreated(
                    Y = Y,
                    D = D,
                    X = X,
                    W = W,
                    W.in.fit = use.W.in.fit,
                    I = I,
                    II = II,
                    T.on = T.on,
                    T.off = T.off,
                    r = r,
                    CV = 0,
                    T.on.balance = T.on.balance,
                    balance.period = balance.period,
                    binary = binary,
                    QR = QR,
                    force = force,
                    hasRevs = hasRevs,
                    tol = tol,
                    max.iteration = max.iteration,
                    boot = 0,
                    norm.para = norm.para,
                    placeboTest = placeboTest,
                    placebo.period = placebo.period,
                    carryoverTest = carryoverTest,
                    carryover.period = carryover.period,
                    group.level = g.level,
                    group = G,
                    parallel = parallel,
                    cores = cores
                )
            } else if (method == "mc") {
                out <- fect_mc(
                    Y = Y,
                    D = D,
                    X = X,
                    W = W,
                    W.in.fit = use.W.in.fit,
                    I = I,
                    II = II,
                    T.on = T.on,
                    T.off = T.off,
                    T.on.carry = T.on.carry,
                    T.on.balance = T.on.balance,
                    balance.period = balance.period,
                    lambda.cv = lambda,
                    force = force,
                    hasRevs = hasRevs,
                    tol = tol,
                    max.iteration = max.iteration,
                    boot = 0,
                    norm.para = norm.para,
                    placeboTest = placeboTest,
                    placebo.period = placebo.period,
                    carryoverTest = carryoverTest,
                    carryover.period = carryover.period,
                    group.level = g.level,
                    group = G
                )
            }

            if ("try-error" %in% class(out)) {
                stop("\nCannot estimate.\n")
            }
        }
    } else {
        # SE == TRUE

        out <- fect_boot(
            Y = Y,
            D = D,
            X = X,
            W = W,
            W.in.fit = use.W.in.fit,
            W.in.agg = use.W.in.agg,
            I = I,
            II = II,
            cm = cm,
            II.cm = II.cm,
            T.on = T.on,
            T.off = T.off,
            T.on.carry = T.on.carry,
            cl = cl,
            T.on.balance = T.on.balance,
            balance.period = balance.period,
            method = method,
            degree = degree,
            cfe = cfe,
            X.extra.FE = X.extra.FE,
            X.Z = X.Z,
            X.Q = X.Q,
            X.gamma = X.gamma,
            X.kappa = X.kappa,
            Zgamma.id = Zgamma.id,
            kappaQ.id = kappaQ.id,
            ind.matrix = index.matrix,
            knots = knots,
            criterion = criterion,
            CV = CV,
            k = k,
            cv.prop = cv.prop,
            cv.method = cv.method,
            cv.nobs = cv.nobs,
            r = r,
            r.end = r.end,
            nlambda = nlambda,
            lambda = lambda,
            alpha = alpha,
            binary = binary,
            QR = QR,
            force = force,
            hasRevs = hasRevs,
            tol = tol,
            max.iteration = max.iteration,
            norm.para = norm.para,
            placeboTest = placeboTest,
            placebo.period = placebo.period,
            carryoverTest = carryoverTest,
            carryover.period = carryover.period,
            vartype = vartype,
            para.error = para.error,
            quantile.CI = .quantile.CI.bool,
            nboots = nboots,
            parallel = parallel,
            cores = cores,
            do_parallel_cv = do_parallel_cv,
            group.level = g.level,
            group = G,
            keep.sims = keep.sims,
            time.component.from = time.component.from,
            loading.bound      = loading.bound,
            gamma.loading      = gamma.loading,
            gamma.loading.grid = gamma.loading.grid
        )

    }

    if ((out$validX == 0) & (p != 0)) {
        warning(
            "Multi-colinearity among covariates. Try removing some of them.\r"
        )
    }

    pre.est.att <- pre.att.bound <- NULL
    pre.term <- NULL
    N_bar <- NULL
    ## leave one period out placebo test for pre-treatment periods

    if (is.null(proportion) == TRUE) {
        proportion <- 0
    }
    max.count <- max(out$count)
    max.pre.periods <- out$time[which(
        out$count >= max.count * proportion & out$time <= 0
    )]
    all.pre.periods <- out$time[which(out$time <= 0)]
    if (is.null(pre.periods) == TRUE) {
        pre.periods <- max.pre.periods
    } else if (length(pre.periods) > 0) {
        pre.periods <- intersect(
            pre.periods[1]:pre.periods[length(pre.periods)],
            max.pre.periods
        )
    } else {
        pre.periods <- NULL
    }

    if (!is.null(pre.periods)) {
        pre.term <- pre.periods
        N_bar <- max(out$count[which(out$time %in% pre.periods)])
    }

    if (placeboEquiv == TRUE) {
        pre.term <- all.pre.periods
        r.cv <- out$r.cv
        lambda.cv <- out$lambda.cv
        method <- out$method
        if (method == "fe") {
            method <- "ife"
        }

        message("\nOut-of-Sample Test...\n")
        # if (is.null(placebo.period)) {
        #    pre.term <- pre.term.min:0
        # } else {
        #    pre.term.bound <- sort(placebo.period)
        #    pre.term <- pre.term.bound[1]:pre.term.bound[length(pre.term.bound)]
        # }

        placebo.period <- pre.term[1]:pre.term[length(pre.term)]

        pre.est.att <- matrix(NA, length(pre.term), 6)
        pre.att.bound <- matrix(NA, length(pre.term), 2)
        pre.att.boot <- matrix(NA, length(pre.term), nboots)

        rownames(pre.est.att) <- rownames(pre.att.bound) <- pre.term
        colnames(pre.est.att) <- c(
            "ATT",
            "S.E.",
            "CI.lower",
            "CI.upper",
            "p.value",
            "count.on"
        )
        colnames(pre.att.bound) <- c("CI.lower", "CI.upper")

        if (!is.null(group)) {
            all_group_names <- names(g.level)
            pre.est.att.group <- list()
            pre.att.bound.group <- list()
            pre.att.boot.group <- list()

            # only use the selected pre.periods before
            for (group.name in all_group_names) {
                pre.term.sub <- intersect(
                    out$group.output[[group.name]]$time.on,
                    pre.term
                )
                pre.est.att.sub <- matrix(NA, length(pre.term.sub), 6)
                pre.att.bound.sub <- matrix(NA, length(pre.term.sub), 2)
                pre.att.boot.sub <- matrix(NA, length(pre.term.sub), nboots)
                if (length(pre.term.sub) > 0) {
                    rownames(pre.att.boot.sub) <- rownames(
                        pre.est.att.sub
                    ) <- rownames(pre.att.bound.sub) <- pre.term.sub
                }
                colnames(pre.est.att.sub) <- c(
                    "ATT",
                    "S.E.",
                    "CI.lower",
                    "CI.upper",
                    "p.value",
                    "count.on"
                )
                colnames(pre.att.bound.sub) <- c("CI.lower", "CI.upper")

                pre.est.att.group[[group.name]] <- pre.est.att.sub
                pre.att.bound.group[[group.name]] <- pre.att.bound.sub
                pre.att.boot.group[[group.name]] <- pre.att.boot.sub
            }
        }

        jj <- length(pre.term)
        pre.term <- sort(pre.term, decreasing = TRUE)

        for (kk in pre.term) {
            placebo.pos <- which(T.on == kk)

            pX <- X
            pY <- Y
            pD <- D
            pI <- I
            pII <- II
            pT.on <- T.on
            pT.off <- T.off
            pG <- G
            pW <- W
            if (!is.null(cl)) {
                p.cl <- cl
            } else {
                p.cl <- NULL
            }

            pII[placebo.pos] <- 0 ## placebo treatment

            ## remove treated units that have too few observations
            T0.2 <- apply(pII, 2, sum)

            if (sum(T0.2[which(apply(D, 2, sum) > 0)] >= min.T0) == 0) {
                message("\n")
                message(paste(
                    "All treated units have been removed for period ",
                    kk,
                    sep = ""
                ))
                message("\n")
                jj <- jj - 1
            } else {
                te <- paste("Pre-period ", kk, sep = "")
                if (kk == 0) {
                    te <- paste(te, "(one period before treatment)", sep = " ")
                }
                message("\n")
                message(te)
                message("\n")

                rem.id.new <- rem.id
                rm.id.2.pos <- sort(which(T0.2 < min.T0))
                rm.id.2 <- rem.id.new[rm.id.2.pos]
                rem.id.2 <- setdiff(rem.id.new, rm.id.2)

                rem.id.new <- rem.id.2
                rm.id.new <- setdiff(1:N, rem.id.new)

                if (length(rm.id.2) > 0) {
                    X.old <- pX
                    if (p > 0) {
                        pX <- array(0, dim = c(TT, (N - length(rm.id.new)), p))
                        for (i in 1:p) {
                            subX <- X.old[,, i]
                            pX[,, i] <- as.matrix(subX[, -rm.id.2.pos])
                        }
                    } else {
                        pX <- array(0, dim = c(TT, (N - length(rm.id.new)), 0))
                    }

                    # N <- N - length(rm.id)
                    pY <- as.matrix(Y[, -rm.id.2.pos])
                    pD <- as.matrix(D[, -rm.id.2.pos])
                    pI <- as.matrix(I[, -rm.id.2.pos]) ## after removing
                    pII <- as.matrix(II[, -rm.id.2.pos])
                    pT.on <- as.matrix(T.on[, -rm.id.2.pos])
                    if (!is.null(cl)) {
                        p.cl <- cl[-rm.id.2.pos]
                    } else {
                        p.cl <- NULL
                    }
                    if (hasRevs) {
                        pT.off <- as.matrix(T.off[, -rm.id.2.pos])
                    }
                    if (!is.null(group)) {
                        ## group <- group[-rm.id.2.pos]
                        ## rawgroup <- rawgroup[-rm.id.2.pos]
                        pG <- as.matrix(G[, -rm.id.2.pos])
                    }
                    if (!is.null(W)) {
                        ## group <- group[-rm.id.2.pos]
                        ## rawgroup <- rawgroup[-rm.id.2.pos]
                        pW <- as.matrix(W[, -rm.id.2.pos])
                    }
                }

                p.out <- fect_boot(
                    Y = pY,
                    D = pD,
                    X = pX,
                    W = pW,
                    W.in.fit = use.W.in.fit,
                    W.in.agg = use.W.in.agg,
                    I = pI,
                    II = pII,
                    T.on = pT.on,
                    T.off = pT.off,
                    cl = p.cl,
                    T.on.carry = T.on.carry,
                    method = method,
                    degree = degree,
                    X.extra.FE = X.extra.FE,
                    X.Z = X.Z,
                    X.Q = X.Q,
                    X.gamma = X.gamma,
                    X.kappa = X.kappa,
                    Zgamma.id = Zgamma.id,
                    kappaQ.id = kappaQ.id,
                    knots = knots,
                    criterion = criterion,
                    CV = 0,
                    k = k,
                    cv.prop = cv.prop,
                    cv.method = cv.method,
                    cv.nobs = cv.nobs,
                    r = r.cv,
                    r.end = r.end,
                    nlambda = nlambda,
                    lambda = lambda.cv,
                    alpha = alpha,
                    binary = binary,
                    QR = QR,
                    force = force,
                    hasRevs = 0,
                    tol = tol,
                    max.iteration = max.iteration,
                    norm.para = norm.para,
                    placeboTest = 0,
                    placebo.period = NULL,
                    carryoverTest = 0,
                    carryover.period = NULL,
                    vartype = vartype,
                    nboots = nboots,
                    parallel = parallel,
                    quantile.CI = .quantile.CI.bool,
                    cores = cores,
                    group.level = g.level,
                    group = pG,
                    dis = FALSE
                )

                p.est.att <- p.out$est.att
                p.att.bound <- p.out$att.bound
                p.pos <- which(as.numeric(rownames(p.est.att)) == kk)
                pre.est.att[jj, ] <- p.est.att[p.pos, ]
                pre.att.bound[jj, ] <- p.att.bound[p.pos, ]
                pre.att.boot[jj, ] <- p.out$att.boot.original[p.pos, ]
                pre.period.name <- rownames(pre.est.att)[jj]

                if (!is.null(group)) {
                    for (group.name in all_group_names) {
                        check.sub <- p.out$group.output[[group.name]]$att.on
                        if (!is.null(check.sub)) {
                            p.est.att.sub <- p.out$est.group.output[[
                                group.name
                            ]]$att.on
                            p.att.bound.sub <- p.out$est.group.output[[
                                group.name
                            ]]$att.on.bound
                            p.att.boot.sub <- p.out$est.group.output[[
                                group.name
                            ]]$att.on.boot
                            p.pos.sub <- which(
                                as.numeric(rownames(p.est.att.sub)) == kk
                            )
                            if (length(p.pos.sub) == 1) {
                                pre.est.att.group[[group.name]][
                                    pre.period.name,
                                ] <- p.est.att.sub[p.pos.sub, ]
                                pre.att.bound.group[[group.name]][
                                    pre.period.name,
                                ] <- p.att.bound.sub[p.pos.sub, ]
                                pre.att.boot.group[[group.name]][
                                    pre.period.name,
                                ] <- p.att.boot.sub[p.pos.sub, ]
                            }
                        }
                    }
                }
                jj <- jj - 1
            }
        }

        pre.est.group.output <- NULL
        if (!is.null(group)) {
            pre.est.group.output <- list()
            for (group.name in all_group_names) {
                sub.output <- list()
                sub.output$pre.est.att <- pre.est.att.group[[group.name]]
                sub.output$pre.att.bound <- pre.att.bound.group[[group.name]]
                sub.output$pre.att.boot <- pre.att.boot.group[[group.name]]
                pre.est.group.output[[group.name]] <- sub.output
            }
        }
    }

    ## permutation test
    if (permute == TRUE) {
        message("Permuting under sharp null hypothesis ... ")

        out.permute <- fect_permu(
            Y = Y,
            X = X,
            D = D,
            I = I,
            r.cv = out$r.cv,
            lambda.cv = out$lambda.cv,
            m = m,
            permu.dimension = permu.dimension,
            method = out$method,
            degree = degree,
            knots = knots,
            force = force,
            tol = tol,
            norm.para = norm.para,
            nboots = nboots,
            parallel = parallel,
            cores = cores
        )

        permute.p <- sum(out.permute > abs(out$att.avg)) / length(out.permute)

        permute.result <- list(permute.att.avg = out.permute, p = permute.p)
    }

    if (!is.null(old.future.plan)) {
        future::plan(old.future.plan)
    }

    ## message("\nOK4\n")

    ## -------------------------------##
    ## storage
    ## -------------------------------##

    iname.old <- iname <- unique(sort(data.old[, id]))
    ## tname.old <- tname <- unique(sort(data.old[,time]))
    if (!0 %in% I.use) {
        tname.old <- tname <- unique(sort(data.old[, time]))
    } else {
        tname.old <- tname <- unique(sort(data.old[, time]))[which(I.use != 0)]
    }

    if (length(rm.id) > 0) {
        remove.id <- iname[rm.id]
        iname <- iname[-rm.id]
    }

    N.rem <- dim(D)[2]
    unit.type <- rep(NA, N.rem) ## 1 for control; 2 for treated; 3 for reversal;
    for (i in 1:N.rem) {
        di <- D[, i]
        ii <- I[, i]
        if (length(unique(di[which(ii == 1)])) == 1) {
            ## treated or control
            if (0 %in% unique(di[which(ii == 1)])) {
                unit.type[i] <- 1 ## control
            } else {
                unit.type[i] <- 2 ## treated
            }
        } else {
            unit.type[i] <- 3 ## reversal
        }
    }

    # 1 treated
    # 2 control
    # 3 missing
    # 4 removed
    # 5 placebo or carryover
    obs.missing <- matrix(0, TT, N) ## not under treatment
    obs.missing[, rem.id] <- D + as.matrix(abs(I - 1)) * 3 ## under treatment
    obs.missing[which(obs.missing == 0)] <- 2
    if (placeboTest | carryoverTest) {
        obs.missing[, rem.id] <- obs.missing[, rem.id] + 3 * (II.origin != II) ## placebo or carryover
    }
    if (carryoverTest & !is.null(carryover.rm)) {
        obs.missing[intersect(which(D.origin == 2), which(I == 1))] <- 6
    }
    obs.missing[which(obs.missing == 4)] <- 3 # in case if I.D!=I
    obs.missing[, rm.id] <- 4 ## removed

    colnames(obs.missing) <- unique(sort(data.old[, id]))
    rownames(obs.missing) <- tname

    obs.missing.balance <- NA
    if (!is.null(balance.period)) {
        obs.missing.balance <- obs.missing
        obs.missing.sub <- obs.missing[, rem.id]
        obs.missing.sub[which(T.on.balance > 0)] <- 7
        obs.missing.sub[which(T.on.balance <= 0)] <- 8
        obs.missing.sub[which(I == 0)] <- 3
        obs.missing.balance[, rem.id] <- obs.missing.sub
    }

    sample <- matrix(obs.missing %in% c(1L, 2L, 5L), nrow = TT, ncol = N,
                     dimnames = dimnames(obs.missing))

    # if cross-validation:

    if (p > 0) {
        Xname.tmp <- Xname
        rownames(out$beta) <- Xname.tmp
        colnames(out$beta) <- c("Coef")
        if (binary == TRUE) {
            rownames(out$marginal) <- Xname.tmp
        }
        if (se == TRUE) {
            rownames(out$est.beta) <- Xname.tmp
            colnames(out$est.beta) <- c(
                "Coef",
                "S.E.",
                "CI.lower",
                "CI.upper",
                "p.value"
            )
            if (binary == TRUE) {
                rownames(out$est.marginal) <- Xname.tmp
            }
            if (placeboTest == TRUE) {
                colnames(out$est.placebo) <- c(
                    "Coef",
                    "S.E.",
                    "CI.lower",
                    "CI.upper",
                    "p.value",
                    "CI.lower(90%)",
                    "CI.upper(90%)"
                )
                rownames(out$est.placebo) <- c("Placebo effect")
            }
            if (carryoverTest == TRUE) {
                colnames(out$est.carryover) <- c(
                    "Coef",
                    "S.E.",
                    "CI.lower",
                    "CI.upper",
                    "p.value",
                    "CI.lower(90%)",
                    "CI.upper(90%)"
                )
                rownames(out$est.carryover) <- c("Carryover effect")
            }
        }
    }
    colnames(out$eff) <- iname
    rownames(out$eff) <- tname
    out$eff.calendar <- cbind(
        matrix(out$eff.calendar, ncol = 1),
        out$N.calendar
    )
    out$eff.calendar.fit <- cbind(
        matrix(out$eff.calendar.fit, ncol = 1),
        out$N.calendar
    )
    rownames(out$eff.calendar) <- tname
    rownames(out$eff.calendar.fit) <- tname
    colnames(out$eff.calendar) <- c("ATT-calendar", "count")
    colnames(out$eff.calendar.fit) <- c("ATT-calendar Fitted", "count")

    if (se == TRUE) {
        rownames(out$est.eff.calendar) <- tname
        rownames(out$est.eff.calendar.fit) <- tname
    }

    ## cohort effect
    if (!is.null(group)) {
        out$group <- rawgroup
        # out$G <- group
        # out$group2 <- rawgroup2
        if (se == 1) {
            rownames(out$est.group.att) <- names(g.level)
        }
        out$g.level <- g.level
    }

    if (is.null(tost.threshold) == TRUE) {
        tost.threshold <- 0.36 * sqrt(out$sigma2.fect)
    }

    output <- c(
        list(
            Y.dat = Y,
            D.dat = D,
            I.dat = I,
            Y = Yname,
            D = Dname,
            X = Xname,
            W = Wname,
            W.est.col = if (use.W.in.fit) Wname else NULL,
            W.agg.col = if (use.W.in.agg) Wname else NULL,
            W.in.fit = use.W.in.fit,
            W.in.agg = use.W.in.agg,
            T.on = T.on,
            G = G.old,
            balance.period = balance.period,
            hasRevs = hasRevs,
            T.off = T.off,
            T.on.balance = T.on.balance,
            index = index,
            id = iname,
            rawtime = tname,
            binary = binary,
            loo = loo,
            proportion = proportion,
            pre.periods = pre.periods,
            tost.threshold = tost.threshold,
            placeboTest = placeboTest,
            placebo.period = placebo.period,
            carryoverTest = carryoverTest,
            carryover.period = carryover.period,
            ## Stored on the fit object so plot logic does not need to
            ## re-parse `x$call` --- robust under do.call(), positional
            ## args, and call-rewriting wrappers.
            carryover.rm = carryover.rm,
            ## Reserved slot for the post-hoc estimand API (v2.4.0+).
            ## NULL for plain imputation estimators (FE / IFE / MC / CFE /
            ## GSC); future doubly-robust estimators will populate this
            ## with the per-cell debias correction so that
            ## eff = (Y_obs - Y0_hat) + eff_debias is the cell-level
            ## score. See ref/po-estimands-contract.md Sec. 3.
            eff_debias = NULL,
            unit.type = unit.type,
            obs.missing = obs.missing,
            obs.missing.balance = obs.missing.balance,
            sample = sample,
            ## Original input panel (pre-drop), preserved so
            ## panelView::panelview(fit) can render the full set of
            ## units --- including those fect dropped (always-treated,
            ## insufficient pre-period, etc.) --- as "Not used" cells.
            data.long = data.old[, c(index, Yname, Dname), drop = FALSE],
            time.component.from = time.component.from,
            em = em
        ),
        out
    )

    if (1 %in% rm.id) {
        output <- c(output, list(remove.id = remove.id))
        ## message("list of removed units:",remove.id)
        ## message("\n\n")
    }

    # if (se == TRUE) {
    #    suppressWarnings(test.out <- diagtest(output, pre.periods = pre.periods,
    #        f.threshold = f.threshold, tost.threshold = tost.threshold))
    #    output <- c(output, list(test.out = test.out))
    # }

    if (permute == TRUE) {
        output <- c(output, list(permute = permute.result))
    }

    if (placeboEquiv == TRUE) {
        output <- c(
            output,
            list(
                pre.est.att = pre.est.att,
                pre.att.bound = pre.att.bound,
                pre.att.boot = pre.att.boot,
                pre.est.group.output = pre.est.group.output
            )
        )
    }

    # if (placeboEquiv || placeboTest || carryoverTest) {
    # classic equivalence test, placeboTest, and carryoverTest
    # this can also be used in placeboTest

    # classic equivalence test
    if (loo == TRUE) {
        output$loo <- FALSE
    }

    if (se == 1) {
        suppressWarnings(
            test.out <- diagtest(
                output,
                pre.periods = pre.periods,
                f.threshold = f.threshold,
                tost.threshold = tost.threshold,
                N_bar = N_bar
            )
        )
        output <- c(output, list(test.out = test.out))
    }

    # loo equivalence test
    if (loo == TRUE) {
        output$loo <- TRUE
    }
    if (loo == TRUE && se == 1) {
        suppressWarnings(
            test.out <- diagtest(
                output,
                pre.periods = pre.periods,
                f.threshold = f.threshold,
                tost.threshold = tost.threshold,
                N_bar = N_bar
            )
        )
        output <- c(output, list(loo.test.out = test.out))
    }

    output <- c(output, list(call = match.call()))

    ## When W is supplied AND the aggregation surface should be weighted
    ## (W or W.agg supplied), route the W-weighted aggregations into the
    ## canonical slot names so fit$att, fit$time, fit$count, fit$att.avg,
    ## fit$att.off etc. agree with fit$est.att, fit$est.avg, plot(fit), and
    ## print(fit). When only W.est is supplied, the aggregation surface
    ## stays unweighted (canonical slots untouched). Either way, strip the
    ## redundant *.W slots from the user-facing fit object so there is only
    ## one canonical aggregation. (Inferential slots --- est.att, est.avg,
    ## est.att90, att.bound, att.boot, att.vcov, est.placebo, est.carryover,
    ## off variants --- are routed inside fect_boot() before reaching here.)
    if (!is.null(Wname) && isTRUE(use.W.in.agg)) {
        if (!is.null(output$att.on.W))      output$att        <- output$att.on.W
        if (!is.null(output$time.on.W))     output$time       <- output$time.on.W
        if (!is.null(output$count.on.W))    output$count      <- output$count.on.W
        if (!is.null(output$att.avg.W))     output$att.avg    <- output$att.avg.W
        if (!is.null(output$att.off.W))     output$att.off    <- output$att.off.W
        if (!is.null(output$time.off.W))    output$time.off   <- output$time.off.W
        if (!is.null(output$count.off.W))   output$count.off  <- output$count.off.W
        if (!is.null(output$att.placebo.W))   output$att.placebo   <- output$att.placebo.W
        if (!is.null(output$att.carryover.W)) output$att.carryover <- output$att.carryover.W
    }
    if (!is.null(Wname)) {
        output[c(
            "att.on.W", "time.on.W", "count.on.W", "att.avg.W",
            "att.on.sum.W", "W.on.sum",
            "att.off.W", "time.off.W", "count.off.W",
            "att.off.sum.W", "W.off.sum",
            "att.placebo.W", "att.carryover.W"
        )] <- NULL
    }

    ## v2.4.3+: convergence diagnostic. The EM check is
    ##   ||fit_new - fit_old||_F / ||fit_old||_F < tol
    ## When niter >= max.iteration, the EM was halted before satisfying
    ## the tol gate. The reported att.avg may not reflect a converged
    ## minimum; users should inspect (and ideally raise max.iteration)
    ## before publishing.
    if (!is.null(output$niter) && length(output$niter) == 1 &&
        is.finite(output$niter) && output$niter >= max.iteration) {
        warning(sprintf(
            paste(
                "EM did not converge within max.iteration = %d (final niter = %d).",
                "Estimates may be under-converged. Consider raising max.iteration",
                "and/or tightening tol. If att.avg shifts substantially under",
                "longer runs, the data may be poorly conditioned for this model",
                "specification (e.g. very small Ntr, near-collinear factors,",
                "or a misspecified rank).", sep = "\n  "),
            as.integer(max.iteration), as.integer(output$niter)
        ), call. = FALSE)
    }

    ## Surface the user's group.fe input on the fit object so print/summary
    ## can show the absorbed FE composition (per D8 in design memo). Also
    ## store the user-facing cl column name (cl gets reshaped to a matrix
    ## internally for the bootstrap; we want the original label here).
    output$group.fe <- group.fe
    output$cl.label <- cl.label

    class(output) <- "fect"
    return(output)
} ## Program fect ends

## ---------------------------------------------------------
## Normalization regression test
## Verifies that sigma2 (est.fect) is consistent with and
## without normalize=TRUE (the sigma2 normalization bug fix)
## ---------------------------------------------------------

test_that("sigma2 is consistent with and without normalization (FE)", {

  skip_on_cran()

    set.seed(6001)

    N <- 30
    TT <- 15
    T0 <- 10
    Ntr <- 10
    tau <- 3.0

    alpha_i <- rnorm(N, 0, 1)
    xi_t <- rnorm(TT, 0, 0.5)

    Y_vec <- numeric(N * TT)
    D_vec <- integer(N * TT)
    id_vec <- integer(N * TT)
    time_vec <- integer(N * TT)

    idx <- 1
    for (i in 1:N) {
        for (t in 1:TT) {
            treated <- (i <= Ntr) && (t > T0)
            D_vec[idx] <- as.integer(treated)
            eps <- rnorm(1, 0, 1)
            Y_vec[idx] <- alpha_i[i] + xi_t[t] + tau * D_vec[idx] + eps
            id_vec[idx] <- i
            time_vec[idx] <- t
            idx <- idx + 1
        }
    }

    simdf <- data.frame(
        id = id_vec,
        time = time_vec,
        Y = Y_vec,
        D = D_vec
    )

    out_raw <- suppressWarnings(fect::fect(
        Y ~ D,
        data = simdf,
        index = c("id", "time"),
        method = "fe",
        force = "two-way",
        normalize = FALSE,
        se = FALSE,
        parallel = FALSE
    ))

    out_norm <- suppressWarnings(fect::fect(
        Y ~ D,
        data = simdf,
        index = c("id", "time"),
        method = "fe",
        force = "two-way",
        normalize = TRUE,
        se = FALSE,
        parallel = FALSE
    ))

    ## ATT should be essentially the same
    expect_equal(out_raw$att.avg, out_norm$att.avg, tolerance = 0.01,
        label = "ATT should match with/without normalization")

    ## sigma2.fect should be close (within 10% relative error)
    ## This is the regression test for the normalization bug fix
    if (!is.null(out_raw$sigma2.fect) && !is.null(out_norm$sigma2.fect)) {
        rel_diff <- abs(out_raw$sigma2.fect - out_norm$sigma2.fect) /
            max(abs(out_raw$sigma2.fect), 1e-10)
        expect_lt(rel_diff, 0.1,
            label = paste0("sigma2.fect relative difference = ",
                           round(rel_diff, 6),
                           " (raw=", round(out_raw$sigma2.fect, 4),
                           ", norm=", round(out_norm$sigma2.fect, 4), ")"))
    }
})

test_that("sigma2 is consistent with and without normalization (IFE)", {

  skip_on_cran()

    set.seed(6002)

    N <- 30
    TT <- 15
    T0 <- 10
    Ntr <- 10
    tau <- 2.0

    alpha_i <- rnorm(N, 0, 1)
    xi_t <- rnorm(TT, 0, 0.5)
    lambda_i <- rnorm(N, 0, 0.5)
    f_t <- rnorm(TT, 0, 0.5)

    Y_vec <- numeric(N * TT)
    D_vec <- integer(N * TT)
    id_vec <- integer(N * TT)
    time_vec <- integer(N * TT)

    idx <- 1
    for (i in 1:N) {
        for (t in 1:TT) {
            treated <- (i <= Ntr) && (t > T0)
            D_vec[idx] <- as.integer(treated)
            eps <- rnorm(1, 0, 1)
            Y_vec[idx] <- alpha_i[i] + xi_t[t] +
                lambda_i[i] * f_t[t] +
                tau * D_vec[idx] + eps
            id_vec[idx] <- i
            time_vec[idx] <- t
            idx <- idx + 1
        }
    }

    simdf <- data.frame(
        id = id_vec,
        time = time_vec,
        Y = Y_vec,
        D = D_vec
    )

    out_raw <- suppressWarnings(fect::fect(
        Y ~ D,
        data = simdf,
        index = c("id", "time"),
        method = "ife",
        r = 1,
        CV = FALSE,
        force = "two-way",
        normalize = FALSE,
        se = FALSE,
        parallel = FALSE
    ))

    out_norm <- suppressWarnings(fect::fect(
        Y ~ D,
        data = simdf,
        index = c("id", "time"),
        method = "ife",
        r = 1,
        CV = FALSE,
        force = "two-way",
        normalize = TRUE,
        se = FALSE,
        parallel = FALSE
    ))

    ## ATT should be essentially the same
    expect_equal(out_raw$att.avg, out_norm$att.avg, tolerance = 0.05,
        label = "IFE ATT should match with/without normalization")

    ## sigma2.fect should be close
    if (!is.null(out_raw$sigma2.fect) && !is.null(out_norm$sigma2.fect)) {
        rel_diff <- abs(out_raw$sigma2.fect - out_norm$sigma2.fect) /
            max(abs(out_raw$sigma2.fect), 1e-10)
        expect_lt(rel_diff, 0.15,
            label = paste0("IFE sigma2.fect relative difference = ",
                           round(rel_diff, 6)))
    }
})

## ---------------------------------------------------------
## Tests derived from Quarto book documentation claims
## Each test validates a specific assertion made in the user-facing docs.
## ---------------------------------------------------------

## =========================================================
## Shared fixtures (computed once, reused across test blocks)
## =========================================================

## DGP helper: two-way FE + optional factors + optional covariates
make_panel <- function(N = 40, TT = 20, T0 = 12, Ntr = 12,
                       tau = 3.0, r = 0, beta_x = 0,
                       mean_shift = 0, seed = 8001,
                       stagger = FALSE, reversals = FALSE,
                       add_Z = FALSE, add_group = FALSE) {
    set.seed(seed)
    alpha_i <- rnorm(N, 0, 2)
    xi_t <- rnorm(TT, 0, 1)
    ## factors
    FL <- matrix(0, TT, N)
    if (r > 0) {
        Fmat <- matrix(rnorm(TT * r), TT, r)
        Lmat <- matrix(rnorm(N * r), N, r)
        FL <- Fmat %*% t(Lmat)
    }
    ## treatment assignment
    D <- matrix(0L, TT, N)
    tr_units <- 1:Ntr
    if (stagger) {
        ## 3 cohorts
        for (i in tr_units) {
            cohort <- ((i - 1) %% 3) + 1
            t0_i <- T0 + (cohort - 1) * 2
            if (t0_i < TT) D[(t0_i + 1):TT, i] <- 1L
        }
    } else {
        D[(T0 + 1):TT, tr_units] <- 1L
    }
    if (reversals) {
        ## flip last 2 periods off for first 3 treated
        for (i in 1:min(3, Ntr)) {
            D[(TT - 1):TT, i] <- 0L
        }
    }
    ## covariates
    X1 <- matrix(rnorm(N * TT, 0, 1), TT, N)
    eps <- matrix(rnorm(N * TT, 0, 1), TT, N)
    Y <- mean_shift +
        outer(xi_t, rep(1, N)) +
        outer(rep(1, TT), alpha_i) +
        FL +
        beta_x * X1 +
        tau * D +
        eps
    df <- data.frame(
        id = rep(1:N, each = TT),
        time = rep(1:TT, N),
        Y = as.vector(Y),
        D = as.vector(D),
        X1 = as.vector(X1)
    )
    if (add_Z) {
        ## time-invariant covariate (same across all periods for a unit)
        Z_vals <- rnorm(N, 0, 1)
        df$Z <- rep(Z_vals, each = TT)
    }
    if (add_group) {
        ## group variable: 3 groups by unit
        df$grp <- rep(((1:N - 1) %% 3) + 1, each = TT)
    }
    df
}


## =========================================================
## Section A: Method Equivalences (Book: index.qmd, 04-gsynth.Rmd)
## =========================================================

test_that("A1: gsynth == ife + time.component.from='nevertreated' (exact)", {

  skip_on_cran()
    d <- make_panel(N = 50, TT = 20, T0 = 12, Ntr = 10, r = 2, seed = 8010)
    out_g <- fect(Y ~ D + X1, data = d, index = c("id", "time"),
                  method = "gsynth", r = 2, se = FALSE, CV = FALSE)
    out_i <- fect(Y ~ D + X1, data = d, index = c("id", "time"),
                  method = "ife", time.component.from = "nevertreated",
                  r = 2, se = FALSE, CV = FALSE)
    expect_equal(out_g$att.avg, out_i$att.avg, tolerance = 1e-8)
    expect_equal(as.vector(out_g$eff), as.vector(out_i$eff), tolerance = 1e-8)
})

test_that("A2: IFE(r=0) == FE (exact equivalence)", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 10, r = 0, seed = 8020)
    out_fe <- fect(Y ~ D + X1, data = d, index = c("id", "time"),
                   method = "fe", se = FALSE, CV = FALSE)
    out_ife <- fect(Y ~ D + X1, data = d, index = c("id", "time"),
                    method = "ife", r = 0, se = FALSE, CV = FALSE)
    expect_equal(out_fe$att.avg, out_ife$att.avg, tolerance = 1e-6)
    expect_equal(as.vector(out_fe$eff), as.vector(out_ife$eff), tolerance = 1e-6)
})


## =========================================================
## Section B: force Parameter Variations (Book: 02-fect.Rmd)
## =========================================================

test_that("B1: force='unit' runs without error", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 10, seed = 8030)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", force = "unit", se = FALSE, CV = FALSE)
    expect_true(is.numeric(out$att.avg))
    expect_false(is.na(out$att.avg))
})

test_that("B2: force='time' runs without error", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 10, seed = 8031)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", force = "time", se = FALSE, CV = FALSE)
    expect_true(is.numeric(out$att.avg))
    expect_false(is.na(out$att.avg))
})

test_that("B3: force='none' runs without error", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 10, seed = 8032)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", force = "none", se = FALSE, CV = FALSE)
    expect_true(is.numeric(out$att.avg))
    expect_false(is.na(out$att.avg))
})

test_that("B4: force variations work for IFE with r > 0", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12, r = 1, seed = 8033)
    for (f in c("unit", "time", "none", "two-way")) {
        out <- fect(Y ~ D, data = d, index = c("id", "time"),
                    method = "ife", r = 1, force = f, se = FALSE, CV = FALSE)
        expect_true(is.numeric(out$att.avg),
                    info = paste("force =", f))
    }
})


## =========================================================
## Section C: Diagnostic Tests (Book: 02-fect.Rmd)
## =========================================================

test_that("C1: placeboTest produces p-value and test output", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12, seed = 8040)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", se = TRUE, nboots = 50,
                placeboTest = TRUE, placebo.period = c(-2, 0),
                CV = FALSE)
    ## Book: placebo p-value should exist
    expect_true(!is.null(out$placebo.p))
    expect_true(is.numeric(out$placebo.p))
})

test_that("C2: carryoverTest produces test output", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12,
                    reversals = TRUE, seed = 8041)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", se = TRUE, nboots = 50,
                carryoverTest = TRUE, carryover.period = c(1, 2),
                CV = FALSE)
    expect_true(!is.null(out$carryover.p))
    expect_true(is.numeric(out$carryover.p))
})

test_that("C3: loo=TRUE produces pre-trend leave-one-out estimates", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12, seed = 8042)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", se = TRUE, nboots = 50,
                loo = TRUE, CV = FALSE)
    ## LOO test output should exist
    expect_true(!is.null(out$loo.test.out))
    expect_true(!is.null(out$loo.test.out$f.stat))
    expect_true(!is.null(out$loo.test.out$f.p))
})

test_that("C4: equivalence test output (pre-trend F-test)", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12, seed = 8043)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", se = TRUE, nboots = 50, loo = TRUE, CV = FALSE)
    ## F-test lives inside loo.test.out
    expect_true(!is.null(out$loo.test.out$f.p))
    expect_true(is.numeric(out$loo.test.out$f.p))
    ## Book: p > 0.05 means good pre-trend fitting; for clean DGP expect pass
    expect_true(out$loo.test.out$f.p > 0.01,
                info = "Pre-trend F-test should not reject for clean DGP")
})


## =========================================================
## Section D: Output Structure (Book: 02-fect.Rmd, 04-gsynth.Rmd)
## =========================================================

test_that("D1: est.att structure (periods × 6 columns)", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12, seed = 8050)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", se = TRUE, nboots = 50, CV = FALSE)
    expect_true(!is.null(out$est.att))
    expect_true(is.matrix(out$est.att) || is.data.frame(out$est.att))
    ## Book: columns include ATT, S.E., CI.lower, CI.upper, p.value, count
    cnames <- colnames(out$est.att)
    expect_true("ATT" %in% cnames)
    expect_true("S.E." %in% cnames)
    expect_true("CI.lower" %in% cnames)
    expect_true("CI.upper" %in% cnames)
    expect_true("count" %in% cnames)
})

test_that("D2: est.avg vs est.avg.unit differ with unequal exposure", {

  skip_on_cran()
    ## staggered adoption → unequal treated-period counts
    d <- make_panel(N = 40, TT = 20, T0 = 10, Ntr = 12,
                    stagger = TRUE, seed = 8051)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", se = FALSE, CV = FALSE)
    ## Book: est.avg weights observations equally; est.avg.unit weights units equally
    ## With staggered adoption these should differ
    expect_true(is.numeric(out$att.avg))
    expect_true(!is.null(out$att.avg.unit))
    ## They won't be exactly equal due to stagger
    expect_false(isTRUE(all.equal(out$att.avg, out$att.avg.unit)),
                 info = "est.avg and est.avg.unit should differ under staggered adoption")
})

test_that("D3: eff.boot dimensions = periods × treated × nboots", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 8, seed = 8052)
    nboots <- 30
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", se = TRUE, nboots = nboots,
                keep.sims = TRUE, CV = FALSE)
    ## Book: eff.boot is array(periods × treated × nboots)
    expect_true(!is.null(out$eff.boot))
    expect_true(is.array(out$eff.boot))
    dims <- dim(out$eff.boot)
    expect_equal(length(dims), 3)
    ## dim 3 should be nboots
    expect_equal(dims[3], nboots)
})

test_that("D4: est.beta present with covariates", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12,
                    beta_x = 1.5, seed = 8053)
    out <- fect(Y ~ D + X1, data = d, index = c("id", "time"),
                method = "fe", se = FALSE, CV = FALSE)
    ## Book: beta reports time-varying covariate coefficients
    expect_true(!is.null(out$beta))
    expect_true(is.numeric(out$beta))
})

test_that("D5: sigma2 is estimated error variance", {

  skip_on_cran()
    d <- make_panel(N = 50, TT = 20, T0 = 12, Ntr = 15, seed = 8054)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", se = FALSE, CV = FALSE)
    ## DGP has eps ~ N(0,1) so sigma2 should be close to 1
    expect_true(!is.null(out$sigma2))
    expect_true(out$sigma2 > 0)
    expect_true(out$sigma2 < 3,
                info = "sigma2 should be in reasonable range for sd=1 DGP")
})

test_that("D6: wgt.implied dimensions = Nco × Ntr for gsynth", {

  skip_on_cran()
    d <- make_panel(N = 50, TT = 20, T0 = 12, Ntr = 10, r = 1, seed = 8055)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "gsynth", r = 1, se = FALSE, CV = FALSE)
    if (!is.null(out$wgt.implied)) {
        ## Book: dimensions = N_co × N_tr
        expect_equal(ncol(out$wgt.implied), 10)  ## Ntr
        expect_equal(nrow(out$wgt.implied), 40)   ## N - Ntr
        ## Book: weights can be positive and negative
        expect_true(any(out$wgt.implied > 0))
    }
})


## =========================================================
## Section E: Edge Cases from Book (02-fect.Rmd, 04-gsynth.Rmd)
## =========================================================

test_that("E1: single treated unit + parametric bootstrap", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 20, T0 = 12, Ntr = 1, r = 1, seed = 8060)
    ## Book: parametric bootstrap works even with Ntr = 1
    out <- suppressWarnings(fect(
        Y ~ D, data = d, index = c("id", "time"),
        method = "gsynth", r = 1, se = TRUE,
        vartype = "parametric", nboots = 30, CV = FALSE
    ))
    expect_true(is.numeric(out$att.avg))
    expect_false(is.na(out$att.avg))
    ## CI should exist
    expect_true(!is.null(out$est.att))
})

test_that("E2: missing Y allowed, missing D errors", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 8, seed = 8061)
    ## Introduce missing Y
    d$Y[sample(nrow(d), 10)] <- NA
    out <- suppressWarnings(fect(
        Y ~ D, data = d, index = c("id", "time"),
        method = "fe", se = FALSE, CV = FALSE, na.rm = TRUE
    ))
    expect_true(is.numeric(out$att.avg))

    ## Missing D should error
    d2 <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 8, seed = 8062)
    d2$D[5] <- NA
    expect_error(
        fect(Y ~ D, data = d2, index = c("id", "time"),
             method = "fe", se = FALSE, CV = FALSE),
        info = "Missing D should produce an error"
    )
})

test_that("E3: unbalanced panel (missing time periods) works", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12, seed = 8063)
    ## Drop ~10% of observations to create unbalanced panel
    keep <- sample(nrow(d), round(nrow(d) * 0.9))
    d_unbal <- d[keep, ]
    out <- suppressWarnings(fect(
        Y ~ D, data = d_unbal, index = c("id", "time"),
        method = "ife", r = 1, se = FALSE, CV = FALSE
    ))
    expect_true(is.numeric(out$att.avg))
    expect_false(is.na(out$att.avg))
})

test_that("E4: balance.period restricts included units", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 10, Ntr = 12,
                    stagger = TRUE, seed = 8064)
    out_full <- fect(Y ~ D, data = d, index = c("id", "time"),
                     method = "fe", se = FALSE, CV = FALSE)
    out_bal <- fect(Y ~ D, data = d, index = c("id", "time"),
                    method = "fe", se = FALSE, CV = FALSE,
                    balance.period = c(-3, 3))
    ## Balanced subset should have fewer or equal treated obs
    expect_true(sum(out_bal$count, na.rm = TRUE) <=
                sum(out_full$count, na.rm = TRUE))
})

test_that("E5: min.T0 drops units with few control observations", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12, seed = 8065)
    out5 <- fect(Y ~ D, data = d, index = c("id", "time"),
                 method = "fe", se = FALSE, CV = FALSE, min.T0 = 5)
    out10 <- fect(Y ~ D, data = d, index = c("id", "time"),
                  method = "fe", se = FALSE, CV = FALSE, min.T0 = 10)
    ## Stricter min.T0 should drop more or equal treated units
    expect_true(sum(out10$count, na.rm = TRUE) <=
                sum(out5$count, na.rm = TRUE))
})

test_that("E6: seed produces reproducible results", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 8, seed = 8066)
    out1 <- fect(Y ~ D, data = d, index = c("id", "time"),
                 method = "fe", se = TRUE, nboots = 30, seed = 1234, CV = FALSE)
    out2 <- fect(Y ~ D, data = d, index = c("id", "time"),
                 method = "fe", se = TRUE, nboots = 30, seed = 1234, CV = FALSE)
    expect_equal(out1$att.avg, out2$att.avg, tolerance = 1e-10)
    expect_equal(out1$est.att, out2$est.att, tolerance = 1e-10)
})


## =========================================================
## Section F: MC Method (Book: 04-gsynth.Rmd cheatsheet)
## =========================================================

test_that("F1: method='mc' basic functionality", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12, r = 1, seed = 8070)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "mc", se = FALSE, CV = FALSE)
    expect_true(is.numeric(out$att.avg))
    expect_false(is.na(out$att.avg))
    ## ATT should be in reasonable range of true tau=3.0
    expect_true(abs(out$att.avg - 3.0) < 2.0)
})

test_that("F2: mc rejects time.component.from='nevertreated'", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 8, seed = 8071)
    expect_error(
        fect(Y ~ D, data = d, index = c("id", "time"),
             method = "mc", time.component.from = "nevertreated",
             se = FALSE, CV = FALSE)
    )
})

test_that("F3: mc with CV selects lambda", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12, r = 1, seed = 8072)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "mc", se = FALSE, CV = TRUE)
    ## Should have selected a lambda
    expect_true(!is.null(out$lambda.cv) || !is.null(out$lambda))
})


## =========================================================
## Section G: CFE Features (Book: 07-cfe.Rmd)
## =========================================================

test_that("G1: CFE with unit-specific linear time trend (Q.type)", {

  skip_on_cran()
    d <- make_panel(N = 200, TT = 20, T0 = 12, Ntr = 60, seed = 8080)
    out <- fect(Y ~ D + X1, data = d, index = c("id", "time"),
                method = "cfe", r = 0, se = FALSE, CV = FALSE,
                Q.type = "linear")
    expect_true(is.numeric(out$att.avg))
    expect_false(is.na(out$att.avg))
})

test_that("G2: CFE with time-invariant covariate Z and gamma", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12,
                    add_Z = TRUE, seed = 8081)
    ## Create time group for gamma: early vs late
    d$time_group <- ifelse(d$time <= 10, 1, 2)
    out <- fect(Y ~ D + X1, data = d, index = c("id", "time"),
                method = "cfe", r = 0, se = FALSE, CV = FALSE,
                Z = "Z", gamma = "time_group")
    expect_true(is.numeric(out$att.avg))
    expect_false(is.na(out$att.avg))
})

test_that("G3: CFE with extra group FE via index[3]", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12,
                    add_group = TRUE, seed = 8082)
    ## Book: extra FE specified via index with 3+ elements
    out <- fect(Y ~ D + X1, data = d, index = c("id", "time", "grp"),
                method = "cfe", r = 0, se = FALSE, CV = FALSE)
    expect_true(is.numeric(out$att.avg))
    expect_false(is.na(out$att.avg))
})

test_that("G4: CFE defaults to time.component.from='notyettreated'", {

  skip_on_cran()
    d <- make_panel(N = 40, TT = 20, T0 = 12, Ntr = 12, seed = 8083)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "cfe", r = 0, se = FALSE, CV = FALSE)
    ## Book: CFE defaults to notyettreated, unlike gsynth
    expect_equal(out$time.component.from, "notyettreated")
})

test_that("G5: CFE with r > 0 adds latent factors", {

  skip_on_cran()
    d <- make_panel(N = 50, TT = 20, T0 = 12, Ntr = 10,
                    r = 2, seed = 8084)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "cfe", r = 2, se = FALSE, CV = FALSE,
                time.component.from = "nevertreated")
    expect_true(is.numeric(out$att.avg))
    ## ATT should be in reasonable range
    expect_true(abs(out$att.avg - 3.0) < 2.0)
})


## =========================================================
## Section H: Inference Methods (Book: cheatsheet, 02-fect.Rmd)
## =========================================================

test_that("H1: vartype='parametric' only for gsynth/nevertreated", {

  skip_on_cran()
    d <- make_panel(N = 50, TT = 20, T0 = 12, Ntr = 10, r = 1, seed = 8090)
    ## Book: parametric specific to gsynth
    out <- suppressWarnings(fect(Y ~ D, data = d, index = c("id", "time"),
                method = "gsynth", r = 1, se = TRUE,
                vartype = "parametric", nboots = 30, CV = FALSE))
    expect_true(!is.null(out$est.att))
    ## Also works for ife with time.component.from='nevertreated' (Phase 3 unlocked)
    out2 <- suppressWarnings(fect(Y ~ D, data = d, index = c("id", "time"),
                 method = "ife", time.component.from = "nevertreated",
                 r = 1, se = TRUE, vartype = "parametric",
                 nboots = 30, CV = FALSE))
    expect_true(!is.null(out2$est.att))
})

test_that("H2: vartype='jackknife' works for FE and IFE", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 8, seed = 8091)
    out_fe <- fect(Y ~ D, data = d, index = c("id", "time"),
                   method = "fe", se = TRUE, vartype = "jackknife", CV = FALSE)
    expect_true(!is.null(out_fe$est.att))

    out_ife <- fect(Y ~ D, data = d, index = c("id", "time"),
                    method = "ife", r = 0, se = TRUE,
                    vartype = "jackknife", CV = FALSE)
    expect_true(!is.null(out_ife$est.att))
})

test_that("H3: nboots default is 200 (tail-CI warning fires at < 1000)", {

  skip_on_cran()
    ## v2.4.2: default stays at 200 (sufficient for SE / normal CI).
    ## A runtime warning fires when ci.method = "basic" and nboots < 1000,
    ## redirecting users to bump nboots for tail-quantile CIs.
    defs <- formals(fect)
    expect_equal(defs$nboots, 200)
})


## =========================================================
## Section I: Plot Types (Book: 03-plots.Rmd, 04-gsynth.Rmd)
## =========================================================

test_that("I1: gap plot (default) returns ggplot", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 8, seed = 8100)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", se = TRUE, nboots = 30, CV = FALSE)
    p <- plot(out, type = "gap")
    expect_true(inherits(p, "gg") || inherits(p, "ggplot"))
})

test_that("I2: counterfactual plot returns ggplot", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 8, seed = 8101)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", se = TRUE, nboots = 30, CV = FALSE)
    p <- plot(out, type = "ct", id = 1)
    expect_true(inherits(p, "gg") || inherits(p, "ggplot") || is.list(p))
})

test_that("I3: equivalence plot returns ggplot", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 8, seed = 8102)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", se = TRUE, nboots = 30, CV = FALSE)
    p <- plot(out, type = "equiv")
    expect_true(inherits(p, "gg") || inherits(p, "ggplot") || is.list(p))
})

test_that("I4: status plot returns ggplot", {

  skip_on_cran()
    d <- make_panel(N = 30, TT = 15, T0 = 10, Ntr = 8, seed = 8103)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "fe", se = FALSE, CV = FALSE)
    p <- plot(out, type = "status")
    expect_true(inherits(p, "gg") || inherits(p, "ggplot") || is.list(p))
})

test_that("I5: gsynth factor and loading plots", {

  skip_on_cran()
    d <- make_panel(N = 50, TT = 20, T0 = 12, Ntr = 10, r = 2, seed = 8104)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "gsynth", r = 2, se = FALSE, CV = FALSE)
    pf <- plot(out, type = "factors")
    expect_true(inherits(pf, "gg") || inherits(pf, "ggplot") || is.list(pf))
    pl <- plot(out, type = "loadings")
    expect_true(inherits(pl, "gg") || inherits(pl, "ggplot") ||
                inherits(pl, "ggmatrix") || is.list(pl))
})


## =========================================================
## Section J: CV and Criterion Selection (Book: 02-fect.Rmd)
## =========================================================

test_that("J1: CV selects r minimizing MSPE", {

  skip_on_cran()
    d <- make_panel(N = 50, TT = 20, T0 = 12, Ntr = 10, r = 2, seed = 8110)
    out <- fect(Y ~ D, data = d, index = c("id", "time"),
                method = "ife", r = c(0, 5), se = FALSE, CV = TRUE)
    ## Should have selected an r
    expect_true(!is.null(out$r.cv))
    expect_true(out$r.cv >= 0 && out$r.cv <= 5)
})

test_that("J2: criterion='pc' runs and may select different r", {

  skip_on_cran()
    d <- make_panel(N = 50, TT = 20, T0 = 12, Ntr = 10, r = 2, seed = 8111)
    out_mspe <- fect(Y ~ D, data = d, index = c("id", "time"),
                     method = "ife", r = c(0, 5), se = FALSE,
                     CV = TRUE, criterion = "mspe")
    out_pc <- fect(Y ~ D, data = d, index = c("id", "time"),
                   method = "ife", r = c(0, 5), se = FALSE,
                   CV = TRUE, criterion = "pc")
    ## Both should produce valid r.cv
    expect_true(!is.null(out_mspe$r.cv))
    expect_true(!is.null(out_pc$r.cv))
})


## =========================================================
## Section K: Bundled Dataset Properties (Book: 01-start.Rmd)
## =========================================================

test_that("K1: simdata has 200 units and 30+ time periods", {

  skip_on_cran()
  data(simdata, package = "fect")
    expect_equal(length(unique(simdata$id)), 200)
    expect_true(length(unique(simdata$time)) >= 30)
})

test_that("K2: simdata has treatment reversals", {

  skip_on_cran()
  data(simdata, package = "fect")
    ## Check that some units switch D from 1 back to 0
    has_reversal <- FALSE
    for (uid in unique(simdata$id)) {
        dvec <- simdata$D[simdata$id == uid]
        diffs <- diff(dvec)
        if (any(diffs == -1)) {
            has_reversal <- TRUE
            break
        }
    }
    expect_true(has_reversal, info = "simdata should contain treatment reversals")
})

test_that("K3: simgsynth has 5 treated, 45 control, 30 periods", {

  skip_on_cran()
  data(simgsynth, package = "fect")
    uid <- unique(simgsynth$id)
    expect_equal(length(uid), 50)
    expect_equal(length(unique(simgsynth$time)), 30)
    ## 5 treated units (check treatment ever on)
    ever_treated <- tapply(simgsynth$D, simgsynth$id, max)
    expect_equal(sum(ever_treated == 1), 5)
    expect_equal(sum(ever_treated == 0), 45)
})

test_that("K4: simgsynth treatment starts at period 21", {

  skip_on_cran()
  data(simgsynth, package = "fect")
    ## For treated units, first treatment period should be 21
    ever_treated <- tapply(simgsynth$D, simgsynth$id, max)
    tr_ids <- names(ever_treated[ever_treated == 1])
    for (uid in tr_ids) {
        subset <- simgsynth[simgsynth$id == uid, ]
        first_treat <- min(subset$time[subset$D == 1])
        expect_equal(first_treat, 21,
                     info = paste("Unit", uid, "first treatment at", first_treat))
    }
})

# Test: parametric bootstrap on unbalanced AR(1) panels now preserves within-unit
# serial correlation (fix at R/boot.R:900, 904; cov.ar = 0 -> cov.ar = TT - 1L).
#
# Strategy: run fect twice on the same DGP, once with the fix (branch HEAD)
# and once with res.vcov monkey-patched to force cov.ar = 0 (simulating the
# pre-fix behavior). Compare the bootstrap sampling SD of the post-period
# time-averaged ATT — this is the quantity affected by within-unit serial
# correlation, because cross-period correlation in residuals shows up only
# when the estimator aggregates across time.
#
# Observed magnitudes (empirical, at nboots=200, N_co=30, N_tr=6, T=40, T_0=20):
#   rho=0.0 : ratio ~= 1.0  (no serial correlation to miss)
#   rho=0.8 : ratio ~= 1.57 (fix restores cross-period cov; dampened by IFE refit)
#   rho=0.9 : ratio ~= 2.17
# The naive asymptotic VIF sqrt((1+rho)/(1-rho)) = 3 / 4.36 is an upper bound;
# the IFE factor projection absorbs part of the serial pattern.

# ---- helpers ----

.gen_ar1 <- function(T, rho, n, seed) {
  set.seed(seed)
  out <- matrix(0, T, n)
  e <- matrix(rnorm(T * n), T, n)
  if (abs(rho) < 1) {
    out[1, ] <- e[1, ] / sqrt(1 - rho^2)
  } else {
    out[1, ] <- e[1, ]
  }
  if (T > 1) {
    for (t in 2:T) out[t, ] <- rho * out[t - 1, ] + e[t, ]
  }
  out
}

.make_dgp <- function(N_co, N_tr, T, T_0, rho, miss_rate, seed) {
  N <- N_co + N_tr
  D_mat <- matrix(0, T, N)
  D_mat[(T_0 + 1):T, (N_co + 1):N] <- 1

  set.seed(seed + 1)
  alpha <- matrix(rep(rnorm(N), each = T), T, N)
  lambda <- matrix(rep(rnorm(T), times = N), T, N)
  eps <- .gen_ar1(T, rho, N, seed + 2)
  Y_mat <- alpha + lambda + eps + 0.5 * D_mat

  if (miss_rate > 0) {
    set.seed(seed + 3)
    miss <- matrix(runif(T * N) < miss_rate, T, N)
    # never blank treated post-period
    miss[(T_0 + 1):T, (N_co + 1):N] <- FALSE
    Y_mat[miss] <- NA
  }

  data.frame(
    id   = rep(1:N, each = T),
    time = rep(1:T, times = N),
    Y    = as.vector(Y_mat),
    D    = as.vector(D_mat)
  )
}

.fit_one <- function(df, nboots, seed) {
  ## suppressWarnings covers the "EM did not converge within
  ## max.iteration = 5000" warning. The DGP here is intentionally
  ## ill-conditioned (small N_co/N_tr + AR(1) rho = 0.8) and these tests
  ## verify the bootstrap SD ratio, not EM convergence.
  suppressWarnings(
    fect(
      Y ~ D, data = df, index = c("id", "time"),
      method = "ife", force = "two-way",
      CV = FALSE, r = 1, se = TRUE,
      vartype = "parametric", nboots = nboots, parallel = FALSE, seed = seed,
      time.component.from = "nevertreated"
    )
  )
}

.with_forced_cov_ar_zero <- function(expr) {
  orig <- fect:::res.vcov
  # Capture `orig` in a closure so the patched function can find it when
  # dispatched from within the fect namespace (the raw `orig` binding is not
  # visible from that scope).
  .mk <- local({
    .orig <- orig
    function() function(res, cov.ar = 1) .orig(res = res, cov.ar = 0)
  })
  patched <- .mk()
  assignInNamespace("res.vcov", patched, ns = "fect")
  on.exit(assignInNamespace("res.vcov", orig, ns = "fect"), add = TRUE)
  force(expr)
}

.se_post_avg <- function(fit) {
  # SD of the time-averaged bootstrap ATT — this is the quantity where
  # within-unit serial correlation in residuals matters.
  b <- as.vector(fit$att.avg.boot)
  sd(b, na.rm = TRUE)
}

# ---- tests ----

test_that("T2a: unbalanced AR(1), rho=0.8, fix inflates att.avg bootstrap SD", {
  skip_on_cran()
  df <- .make_dgp(N_co = 30, N_tr = 6, T = 40, T_0 = 20, rho = 0.8,
                  miss_rate = 0.20, seed = 42)
  fit_new <- .fit_one(df, nboots = 200, seed = 42)
  fit_old <- .with_forced_cov_ar_zero(.fit_one(df, nboots = 200, seed = 42))
  se_new <- .se_post_avg(fit_new)
  se_old <- .se_post_avg(fit_old)
  ratio <- se_new / se_old
  message(sprintf("[T2a] att.avg SD  new=%.4f  old=%.4f  ratio=%.3f", se_new, se_old, ratio))
  # Expect ratio clearly above 1: fix restores the cross-period covariance
  # component that the old code zeroed. IFE refit damps the naive
  # sqrt((1+rho)/(1-rho))=3 asymptote to ~1.5-1.7 in finite samples.
  expect_gt(ratio, 1.25)
  expect_lt(ratio, 3.5)   # guard against runaway inflation too
})

test_that("T2b: balanced-panel parametric-boot SE is unchanged by the fix", {
  skip_on_cran()
  df <- .make_dgp(N_co = 30, N_tr = 6, T = 20, T_0 = 12, rho = 0.8,
                  miss_rate = 0.0, seed = 42)
  expect_true(!anyNA(df$Y))  # truly balanced
  fit_new <- .fit_one(df, nboots = 200, seed = 42)
  fit_old <- .with_forced_cov_ar_zero(.fit_one(df, nboots = 200, seed = 42))
  # Balanced-panel branch is the `else` side of `if (0 %in% I)` at
  # R/boot.R:906; uses direct vector resampling and never calls res.vcov.
  # Must match bit-for-bit.
  se_new <- .se_post_avg(fit_new)
  se_old <- .se_post_avg(fit_old)
  message(sprintf("[T2b] balanced att.avg SD new=%.6f old=%.6f diff=%.3e",
                  se_new, se_old, se_new - se_old))
  expect_equal(se_new, se_old, tolerance = 1e-12)

  # Also per-period SE bitwise-identical
  post_rows <- as.numeric(rownames(fit_new$est.att)) >= 1
  expect_equal(fit_new$est.att[post_rows, "S.E."],
               fit_old$est.att[post_rows, "S.E."],
               tolerance = 1e-12)
})

test_that("T2c: unbalanced rho=0, fix leaves att.avg bootstrap SD ~unchanged", {
  skip_on_cran()
  df <- .make_dgp(N_co = 30, N_tr = 6, T = 40, T_0 = 20, rho = 0.0,
                  miss_rate = 0.20, seed = 42)
  fit_new <- .fit_one(df, nboots = 200, seed = 42)
  fit_old <- .with_forced_cov_ar_zero(.fit_one(df, nboots = 200, seed = 42))
  se_new <- .se_post_avg(fit_new)
  se_old <- .se_post_avg(fit_old)
  ratio <- se_new / se_old
  message(sprintf("[T2c] rho=0 att.avg SD new=%.4f old=%.4f ratio=%.3f",
                  se_new, se_old, ratio))
  # No serial correlation -> no cross-period covariance to restore. Any
  # observed ratio is MC noise between two bootstrap runs with the same
  # seed but slightly different covariance-estimation code paths.
  expect_gt(ratio, 0.80)
  expect_lt(ratio, 1.25)
})

test_that("plot.fect executes for gap plot without error", {
  suppressWarnings(try(data("simdata", package = "fect"), silent = TRUE))
  expect_true(exists("simdata"))
  set.seed(4)
  out <- fect::fect(
    Y ~ D + X1 + X2,
    data = simdata,
    index = c("id", "time"),
    method = "ife",
    r = 1,
    CV = FALSE,
    se = FALSE,
    parallel = FALSE
  )
  # Just ensure the plotting call produces a ggplot object
  p <- plot.fect(out, type = "gap", show.count = FALSE)
  expect_s3_class(p, "ggplot")
})

test_that("plot.fect executes for heterogeneous (hte) plot without error", {
  suppressWarnings(try(data("simdata", package = "fect"), silent = TRUE))
  expect_true(exists("simdata"))
  set.seed(4)
  out <- fect::fect(
    Y ~ D + X1 + X2,
    data = simdata,
    index = c("id", "time"),
    method = "ife",
    r = 1,
    CV = FALSE,
    se = FALSE,
    parallel = FALSE
  )
  p <- plot.fect(out, type = "hte", covariate = "X1", show.count = FALSE)
  expect_s3_class(p, "ggplot")
})

test_that("plot.fect executes for calendar plot with relative.time=TRUE without error", {
  suppressWarnings(try(data("simdata", package = "fect"), silent = TRUE))
  expect_true(exists("simdata"))
  set.seed(4)
  out <- fect::fect(
    Y ~ D + X1 + X2,
    data = simdata,
    index = c("id", "time"),
    method = "ife",
    r = 1,
    CV = FALSE,
    se = FALSE,
    parallel = FALSE
  )
  p <- plot.fect(out, type = "calendar", relative.time = TRUE, show.count = TRUE)
  expect_s3_class(p, "ggplot")
})

test_that("plot.fect executes for heterogeneous pretreatment plot with num.pretreatment without error", {
  suppressWarnings(try(data("simdata", package = "fect"), silent = TRUE))
  expect_true(exists("simdata"))
  set.seed(4)
  out <- fect::fect(
    Y ~ D + X1 + X2,
    data = simdata,
    index = c("id", "time"),
    method = "ife",
    r = 1,
    CV = FALSE,
    se = FALSE,
    parallel = FALSE
  )
  p <- plot.fect(out, type = "hte", covariate = "X1", pretreatment = TRUE, num.pretreatment = 3, show.count = FALSE)
  expect_s3_class(p, "ggplot")
})

test_that("fect formula basic runs on simdata and returns expected slots", {
  suppressWarnings(try(data("simdata", package = "fect"), silent = TRUE))
  expect_true(exists("simdata"))

  set.seed(1)
  out <- fect::fect(
    Y ~ D + X1 + X2,
    data = simdata,
    index = c("id", "time"),
    method = "ife",
    r = 2,
    CV = FALSE,
    se = FALSE,
    parallel = FALSE
  )

  expect_s3_class(out, "fect")
  expect_true(is.matrix(out$eff))
  expect_true(is.numeric(out$att.avg))
  expect_true(length(out$time) >= 1)
})

test_that("fect returns est.att and att.vcov when se=TRUE", {

  ## Bootstrap-based test: too slow for CRAN's test budget. The fast
  ## smoke test above already exercises end-to-end fitting; this one
  ## additionally validates the SE / CI surface.
  skip_on_cran()

  suppressWarnings(try(data("simdata", package = "fect"), silent = TRUE))
  expect_true(exists("simdata"))

  set.seed(2)
  out <- fect::fect(
    Y ~ D + X1 + X2,
    data = simdata,
    index = c("id", "time"),
    method = "ife",
    r = 1,
    CV = FALSE,
    se = TRUE,
    nboots = 30,
    parallel = FALSE
  )

  expect_true(!is.null(out$est.att))
  expect_true(!is.null(out$att.vcov))
  expect_true(all(colnames(out$est.att) %in% c("ATT", "S.E.", "CI.lower", "CI.upper", "p.value", "count")))
})

## ---------------------------------------------------------------
## Tests for the explicit carryover.rm slot on the fit object.
##
## Pre-fix (v2.3.2 and earlier), plot.R recovered carryover.rm from
## `as.list(x$call)$carryover.rm` and `eval(arg, envir = parent.frame())`.
## That call-parsing path silently gave the wrong answer when the fit
## was constructed via do.call(), wrappers that rewrote the call, or
## when x$call was missing/altered. Storing carryover.rm directly on
## the fit object makes the plot logic robust to all those paths.
## ---------------------------------------------------------------

suppressWarnings(data("simdata", package = "fect"))

## -- C.1  Slot exists and matches the value passed at fit time --

test_that("C.1: fit$carryover.rm is populated with the value passed to fect()", {

  skip_on_cran()

  set.seed(42)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = simdata, index = c("id", "time"),
      method = "ife", r = 1, CV = FALSE,
      force = "two-way",
      se = FALSE,
      carryover.rm = 2
    )
  ))

  expect_true("carryover.rm" %in% names(fit))
  expect_equal(fit$carryover.rm, 2)
})


## -- C.2  Slot is NULL when the user does not pass carryover.rm --

test_that("C.2: fit$carryover.rm is NULL when carryover.rm is not used", {

  skip_on_cran()

  set.seed(42)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = simdata, index = c("id", "time"),
      method = "ife", r = 1, CV = FALSE,
      force = "two-way",
      se = FALSE
    )
  ))

  expect_null(fit$carryover.rm)
})


## -- C.3  Plot logic uses the slot, not x$call. Wipe x$call after
## fitting and confirm that plot(fit) still recovers carryover_rm_K.
## This is the regression test for the call-parsing fragility: the
## old code would have read 0 from a call-less fit and rendered the
## carryover plot with a missing K-shift; the new code reads from
## x$carryover.rm and is unaffected.

test_that("C.3: plot logic recovers carryover.rm even when x$call is wiped", {

  skip_on_cran()

  set.seed(42)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D, data = simdata, index = c("id", "time"),
      method = "ife", r = 1, CV = FALSE,
      force = "two-way",
      se = FALSE,
      carryover.rm = 2
    )
  ))

  ## Wipe x$call to defeat the legacy call-parsing path.
  fit$call <- NULL

  ## The slot should still report K = 2, and plot.fect should not
  ## throw. (We do not inspect the rendered plot's pixel content;
  ## a successful build is the contract guard.)
  expect_equal(fit$carryover.rm, 2)
  expect_no_error(suppressWarnings(suppressMessages(plot(fit, type = "exit"))))
})

## ---------------------------------------------------------------
## Runtime benchmarks for cv.method timing comparison.
## Heavy: each test fits the same DGP under {loo, all_units, treated_units}
## and prints elapsed time per call. Skip on CRAN.
##
## Originally Section I of test-score-unify.R; split out 2026-05-03.
## Shared fixtures live in helper-score-unify.R.
## ---------------------------------------------------------------

## =================================================================
## Section I: Runtime Benchmarks (cv.method timing comparison)
## Informational only — skip on CRAN
## =================================================================

test_that("BENCH1: IFE timing comparison (loo vs all_units vs treated_units)", {

  skip_on_cran()

  cat("\n=== Runtime Benchmark: IFE nevertreated cv.method timing ===\n")

  set.seed(42)
  t_loo <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data            = ntdata,
        index           = c("id", "time"),
        method          = "ife",
        force           = "two-way",
        time.component.from    = "nevertreated",
        CV              = TRUE,
        r               = c(0, 3),
        cv.method       = "loo",
        se              = FALSE,
        parallel        = FALSE
      )
    ))
  )
  cat(sprintf("  IFE loo:            %6.2f sec (elapsed)\n", t_loo["elapsed"]))

  set.seed(42)
  t_au <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data            = ntdata,
        index           = c("id", "time"),
        method          = "ife",
        force           = "two-way",
        time.component.from    = "nevertreated",
        CV              = TRUE,
        r               = c(0, 3),
        cv.method       = "all_units",
        se              = FALSE,
        parallel        = FALSE
      )
    ))
  )
  cat(sprintf("  IFE all_units:      %6.2f sec (elapsed)\n", t_au["elapsed"]))

  set.seed(42)
  t_tu <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data            = ntdata,
        index           = c("id", "time"),
        method          = "ife",
        force           = "two-way",
        time.component.from    = "nevertreated",
        CV              = TRUE,
        r               = c(0, 3),
        cv.method       = "treated_units",
        se              = FALSE,
        parallel        = FALSE
      )
    ))
  )
  cat(sprintf("  IFE treated_units:  %6.2f sec (elapsed)\n", t_tu["elapsed"]))

  cat(sprintf("  Speedup (loo/all_units):      %.2fx\n",
              t_loo["elapsed"] / max(t_au["elapsed"], 0.001)))
  cat(sprintf("  Speedup (loo/treated_units):  %.2fx\n",
              t_loo["elapsed"] / max(t_tu["elapsed"], 0.001)))
  cat("=== End IFE Benchmark ===\n")

  # All three must complete — that's the real test
  expect_true(TRUE)
})

test_that("BENCH2: CFE timing comparison (loo vs all_units vs treated_units)", {

  skip_on_cran()

  cat("\n=== Runtime Benchmark: CFE nevertreated cv.method timing ===\n")

  set.seed(42)
  t_loo <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data            = ntdata,
        index           = c("id", "time"),
        method          = "cfe",
        force           = "two-way",
        time.component.from    = "nevertreated",
        CV              = TRUE,
        r               = c(0, 3),
        cv.method       = "loo",
        se              = FALSE,
        parallel        = FALSE
      )
    ))
  )
  cat(sprintf("  CFE loo:            %6.2f sec (elapsed)\n", t_loo["elapsed"]))

  set.seed(42)
  t_au <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data            = ntdata,
        index           = c("id", "time"),
        method          = "cfe",
        force           = "two-way",
        time.component.from    = "nevertreated",
        CV              = TRUE,
        r               = c(0, 3),
        cv.method       = "all_units",
        se              = FALSE,
        parallel        = FALSE
      )
    ))
  )
  cat(sprintf("  CFE all_units:      %6.2f sec (elapsed)\n", t_au["elapsed"]))

  set.seed(42)
  t_tu <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data            = ntdata,
        index           = c("id", "time"),
        method          = "cfe",
        force           = "two-way",
        time.component.from    = "nevertreated",
        CV              = TRUE,
        r               = c(0, 3),
        cv.method       = "treated_units",
        se              = FALSE,
        parallel        = FALSE
      )
    ))
  )
  cat(sprintf("  CFE treated_units:  %6.2f sec (elapsed)\n", t_tu["elapsed"]))

  cat(sprintf("  Speedup (loo/all_units):      %.2fx\n",
              t_loo["elapsed"] / max(t_au["elapsed"], 0.001)))
  cat(sprintf("  Speedup (loo/treated_units):  %.2fx\n",
              t_loo["elapsed"] / max(t_tu["elapsed"], 0.001)))
  cat("=== End CFE Benchmark ===\n")

  expect_true(TRUE)
})

## Regression tests for the v2.3.1 modern-theme overhaul.
##
## Goals:
##  1. Modern (default): theme.bw = TRUE + legacy.style = FALSE produces a plot
##     with the modern visual recipe (plain title, no x major grid).
##  2. Classic: legacy.style = TRUE produces pre-2.3.1 visuals (bold centered
##     title, x major grid present) regardless of theme.bw.
##  3. Gray (theme.bw = FALSE) emits a soft-deprecation message exactly once
##     per session.
##  4. Each mode runs without error on a small, real fit.

skip_if_not_installed("ggplot2")

local_fect_fit <- function(seed = 1) {
  set.seed(seed)
  N <- 30; TT <- 12; Ntr <- 6; T0 <- 7
  T0_vec <- rep(Inf, N); T0_vec[seq_len(Ntr)] <- T0
  Y <- D <- numeric(N * TT)
  id <- time <- integer(N * TT)
  alpha <- rnorm(N); xi <- rnorm(TT, 0, 0.3)
  idx <- 1L
  for (i in seq_len(N)) for (t in seq_len(TT)) {
    treated <- (t >= T0_vec[i])
    D[idx] <- as.integer(treated)
    Y[idx] <- alpha[i] + xi[t] + 1.0 * D[idx] + rnorm(1, 0, 0.3)
    id[idx] <- i; time[idx] <- t; idx <- idx + 1L
  }
  df <- data.frame(id = id, time = time, Y = Y, D = D)
  suppressMessages(suppressWarnings(fect(
    Y ~ D, data = df, index = c("id", "time"),
    method = "fe", force = "two-way", se = FALSE, parallel = FALSE
  )))
}

extract_title_face <- function(p) {
  th <- ggplot2::theme_get()
  pt <- ggplot2::ggplot_build(p)$plot$theme$plot.title
  if (is.null(pt) || is.null(pt$face)) return(NA_character_)
  pt$face
}

extract_title_hjust <- function(p) {
  pt <- ggplot2::ggplot_build(p)$plot$theme$plot.title
  if (is.null(pt) || is.null(pt$hjust)) return(NA_real_)
  pt$hjust
}

extract_grid_major_x <- function(p) {
  pg <- ggplot2::ggplot_build(p)$plot$theme$panel.grid.major.x
  if (is.null(pg)) return(NA_character_)
  ## Modern recipe sets it to element_blank(); legacy leaves the default.
  if (inherits(pg, "element_blank")) return("blank") else return("not_blank")
}

test_that("modern recipe (default) produces plain left-aligned title with no x major grid", {
  fit <- local_fect_fit()
  p <- plot(fit, type = "gap", main = "modern test")
  expect_equal(extract_title_face(p), "plain")
  expect_equal(extract_title_hjust(p), 0)
  expect_equal(extract_grid_major_x(p), "blank")
})

test_that("legacy.style = TRUE produces bold centered title (theme.bw = TRUE)", {
  fit <- local_fect_fit()
  p <- plot(fit, type = "gap", main = "classic test", legacy.style = TRUE)
  expect_equal(extract_title_face(p), "bold")
  expect_equal(extract_title_hjust(p), 0.5)
})

test_that("legacy.style = TRUE produces bold centered title even with theme.bw = FALSE", {
  fit <- local_fect_fit()
  ## Reset deprecation flag so any messages from theme.bw=FALSE don't leak
  options(fect.theme.bw.deprecated.notified = TRUE)
  p <- plot(fit, type = "gap", main = "gray-classic test",
            legacy.style = TRUE, theme.bw = FALSE)
  expect_equal(extract_title_face(p), "bold")
  expect_equal(extract_title_hjust(p), 0.5)
})

test_that("theme.bw = FALSE emits soft-deprecation message exactly once per session", {
  fit <- local_fect_fit()
  ## Force re-arming
  options(fect.theme.bw.deprecated.notified = FALSE)
  expect_message(
    plot(fit, type = "gap", theme.bw = FALSE),
    "soft-deprecated"
  )
  ## Second call should NOT emit
  expect_silent({
    suppressMessages(plot(fit, type = "gap", theme.bw = FALSE))
  })
})

test_that("all three modes run without error", {
  fit <- local_fect_fit()
  options(fect.theme.bw.deprecated.notified = TRUE)
  expect_no_error(plot(fit, type = "gap"))                                           # modern
  expect_no_error(plot(fit, type = "gap", legacy.style = TRUE))                      # classic
  expect_no_error(plot(fit, type = "gap", theme.bw = FALSE))                         # gray
  expect_no_error(plot(fit, type = "gap", legacy.style = TRUE, theme.bw = FALSE))    # gray-classic
})

## Integration tests for loading.bound argument in fect().
## See statsclaw-workspace/fect/runs/REQ-bounded-loadings/test-spec.md section I.

## ---------------------------------------------------------------------------
## Helper DGP: factor-structure panel with never-treated controls
## ---------------------------------------------------------------------------

make_factor_panel <- function(N = 100, TT = 30, Ntr = 30, tau = 3.0,
                              r = 2, seed = 42) {
    set.seed(seed)
    F_mat    <- matrix(rnorm(TT * r), TT, r)
    L_mat    <- matrix(rnorm(N * r), N, r)
    alpha_i  <- rnorm(N, 0, 1)
    xi_t     <- rnorm(TT, 0, 0.5)
    T0_vec   <- rep(Inf, N)
    if (Ntr > 0) {
        T0_vec[1:Ntr] <- sample(round(TT * 0.4):round(TT * 0.7), Ntr,
                                replace = TRUE)
    }
    Y_vec <- D_vec <- numeric(N * TT)
    id_vec <- time_vec <- integer(N * TT)
    idx <- 1
    for (i in seq_len(N)) {
        for (t in seq_len(TT)) {
            treated       <- (t >= T0_vec[i])
            D_vec[idx]    <- as.integer(treated)
            Y_vec[idx]    <- alpha_i[i] + xi_t[t] +
                             sum(F_mat[t, ] * L_mat[i, ]) +
                             tau * D_vec[idx] + rnorm(1, 0, 0.5)
            id_vec[idx]   <- i
            time_vec[idx] <- t
            idx           <- idx + 1
        }
    }
    data.frame(id = id_vec, time = time_vec, Y = Y_vec, D = D_vec)
}

common_fit <- function(df, ...) {
    suppressMessages(suppressWarnings(fect::fect(
        Y ~ D, data = df, index = c("id", "time"),
        method = "ife", r = 2, CV = FALSE, se = FALSE,
        parallel = FALSE, force = "two-way",
        time.component.from = "nevertreated", ...
    )))
}

## ---------------------------------------------------------------------------
## I.1: back-compat bit-identity with pinned dev@69bf243 output
## ---------------------------------------------------------------------------

test_that("I.1: default loading.bound = 'none' matches pinned dev@69bf243 output", {
    fixture_path <- test_path("fixtures", "bounded-loadings-backcompat-dev69bf243.rds")
    skip_if_not(file.exists(fixture_path), "Back-compat fixture not pinned.")
    pinned <- readRDS(fixture_path)

    df  <- make_factor_panel(N = 100, TT = 30, Ntr = 30, tau = 3.0, r = 2, seed = 42)
    set.seed(100)
    fit <- common_fit(df)
    expect_equal(fit$att.avg,   pinned$att.avg,   tolerance = 1e-10)
    expect_equal(fit$lambda.tr, pinned$lambda.tr, tolerance = 1e-10)
    expect_equal(fit$lambda.co, pinned$lambda.co, tolerance = 1e-10)
    expect_equal(fit$wgt.implied, pinned$wgt.implied, tolerance = 1e-10)
})

## ---------------------------------------------------------------------------
## I.2: simplex bound yields weights on simplex and lambda.tr = Lambda.co' %*% w
## ---------------------------------------------------------------------------

test_that("I.2: simplex bound produces lambda.tr in conv(lambda.co)", {
    df <- make_factor_panel(seed = 42)
    set.seed(100)
    fit <- common_fit(df, loading.bound = "simplex", gamma.loading = 1.0)

    expect_identical(fit$loading.bound, "simplex")
    expect_equal(fit$gamma.loading, 1.0, tolerance = 1e-12)

    for (i in seq_len(nrow(fit$lambda.tr))) {
        w_i <- fit$wgt.implied[i, ]
        expect_equal(sum(w_i), 1, tolerance = 1e-6)
        expect_true(all(w_i >= -1e-10))
        expect_equal(as.numeric(fit$lambda.tr[i, ]),
                     as.numeric(crossprod(fit$lambda.co, w_i)),
                     tolerance = 1e-6)
    }
})

## ---------------------------------------------------------------------------
## I.3: pointwise hull of fitted factor-component counterfactual
## ---------------------------------------------------------------------------

test_that("I.3: Y_hat(0) factor component lies in pointwise hull of F %*% lambda.co", {
    df <- make_factor_panel(seed = 42)
    set.seed(100)
    fit <- common_fit(df, loading.bound = "simplex", gamma.loading = 1.0)

    F_hat    <- fit$factor                       # TT x r
    Lam_co   <- fit$lambda.co                    # Nco x r
    ctrl_fit <- F_hat %*% t(Lam_co)              # TT x Nco
    Y_factor <- F_hat %*% t(fit$lambda.tr)       # TT x Ntr

    for (t in seq_len(nrow(ctrl_fit))) {
        lo <- min(ctrl_fit[t, ]); hi <- max(ctrl_fit[t, ])
        expect_true(all(Y_factor[t, ] >= lo - 1e-6 &
                        Y_factor[t, ] <= hi + 1e-6))
    }
})

## ---------------------------------------------------------------------------
## I.4: method restriction
## ---------------------------------------------------------------------------

test_that("I.4: loading.bound='simplex' errors with unsupported method/dispatch", {
    df <- make_factor_panel(N = 60, TT = 20, Ntr = 15, seed = 3)
    ## method = "mc" is independently blocked by the method check, but because
    ## the default `time.component.from = "notyettreated"`, either check may fire
    ## first. Both are acceptable errors indicating the configuration is invalid.
    expect_error(
        suppressMessages(suppressWarnings(fect(
            Y ~ D, data = df, index = c("id","time"),
            method = "mc", loading.bound = "simplex",
            CV = FALSE, se = FALSE, parallel = FALSE
        ))),
        regexp = "only supported for method|nevertreated"
    )
    ## method = "fe" is rewritten internally to method = "ife" with r = 0,
    ## so the effective failure mode is the time.component.from check.
    expect_error(
        suppressMessages(suppressWarnings(fect(
            Y ~ D, data = df, index = c("id","time"),
            method = "fe", loading.bound = "simplex",
            CV = FALSE, se = FALSE, parallel = FALSE
        ))),
        regexp = "only supported for method|nevertreated"
    )
})

test_that("I.4b: loading.bound='simplex' requires time.component.from='nevertreated'", {
    df <- make_factor_panel(seed = 5)
    expect_error(
        suppressMessages(suppressWarnings(fect(
            Y ~ D, data = df, index = c("id","time"),
            method = "ife", loading.bound = "simplex",
            time.component.from = "notyettreated",
            CV = FALSE, se = FALSE, parallel = FALSE, r = 2
        ))),
        regexp = "nevertreated"
    )
})

## ---------------------------------------------------------------------------
## I.5: argument validation
## ---------------------------------------------------------------------------

test_that("I.5: invalid loading.bound / gamma.loading / grid raise", {
    df <- make_factor_panel(seed = 7)
    expect_error(common_fit(df, loading.bound = "foo"),
                 regexp = "must be one of")
    expect_error(common_fit(df, loading.bound = "simplex", gamma.loading = -1),
                 regexp = "positive")
    expect_error(common_fit(df, loading.bound = "simplex",
                            gamma.loading.grid = c(1, -2, 3)),
                 regexp = "positive")
})

## ---------------------------------------------------------------------------
## I.6: gamma CV selects from grid
## ---------------------------------------------------------------------------

test_that("I.6: gamma.loading = NULL selects gamma from provided grid", {
    df <- make_factor_panel(seed = 9)
    grid <- c(0.1, 1.0, 10.0)
    set.seed(100)
    fit <- common_fit(df, loading.bound = "simplex",
                       gamma.loading = NULL, gamma.loading.grid = grid)
    expect_true(fit$gamma.loading %in% grid)
})

## ---------------------------------------------------------------------------
## I.7: reproducibility with fixed gamma
## ---------------------------------------------------------------------------

test_that("I.7: gamma fixed + same seed produces identical lambda.tr and wgt.implied", {
    df <- make_factor_panel(seed = 11)
    set.seed(100)
    fit1 <- common_fit(df, loading.bound = "simplex", gamma.loading = 2.5)
    set.seed(100)
    fit2 <- common_fit(df, loading.bound = "simplex", gamma.loading = 2.5)
    expect_equal(fit1$att.avg, fit2$att.avg, tolerance = 1e-10)
    expect_equal(fit1$lambda.tr, fit2$lambda.tr, tolerance = 1e-10)
    expect_equal(fit1$wgt.implied, fit2$wgt.implied, tolerance = 1e-10)
})

## ---------------------------------------------------------------------------
## I.8: diagnostic populated
## ---------------------------------------------------------------------------

test_that("I.8: loading.proj.resid is populated and sensible", {
    df <- make_factor_panel(seed = 13)
    set.seed(100)
    fit <- common_fit(df, loading.bound = "simplex", gamma.loading = 1.0)
    expect_false(is.null(fit$loading.proj.resid))
    expect_length(fit$loading.proj.resid, nrow(fit$lambda.tr))
    expect_true(all(fit$loading.proj.resid >= 0))
    expect_true(all(is.finite(fit$loading.proj.resid)))
})

## ---------------------------------------------------------------------------
## I.9: smoke test with bootstrap
## ---------------------------------------------------------------------------

test_that("I.9: fect() with se=TRUE runs with bounded loading (smoke)", {
    df <- make_factor_panel(N = 60, TT = 25, Ntr = 20, seed = 15)
    set.seed(100)
    fit <- suppressMessages(suppressWarnings(fect(
        Y ~ D, data = df, index = c("id","time"),
        method = "ife", r = 2, CV = FALSE, se = TRUE, nboots = 20,
        parallel = FALSE, force = "two-way",
        time.component.from = "nevertreated",
        loading.bound = "simplex", gamma.loading = 1.0
    )))
    expect_false(is.null(fit$est.att))
    expect_true(is.finite(fit$att.avg))
})

## ---------------------------------------------------------------------------
## I.10: loading.overlap plot type
## ---------------------------------------------------------------------------

test_that("I.10a: loading.overlap returns a ggplot for r >= 2", {
    df <- make_factor_panel(N = 80, TT = 25, Ntr = 20, r = 2, seed = 7)
    fit <- common_fit(df)
    p <- plot(fit, type = "loading.overlap")
    expect_s3_class(p, "ggplot")

    rd <- plot(fit, type = "loading.overlap", return.data = TRUE)
    expect_true(is.list(rd) || inherits(rd, "data.frame"))
})

test_that("I.10b: loading.overlap returns a ggplot for r == 1 (mirror histogram)", {
    df <- make_factor_panel(N = 80, TT = 25, Ntr = 20, r = 1, seed = 8)
    fit <- suppressMessages(suppressWarnings(fect::fect(
        Y ~ D, data = df, index = c("id", "time"),
        method = "ife", r = 1, CV = FALSE, se = FALSE,
        parallel = FALSE, force = "two-way",
        time.component.from = "nevertreated"
    )))
    p <- plot(fit, type = "loading.overlap")
    expect_s3_class(p, "ggplot")
})

test_that("I.10c: loading.overlap errors when r == 0", {
    df  <- make_factor_panel(N = 60, TT = 20, Ntr = 15, r = 1, seed = 9)
    fit <- suppressMessages(suppressWarnings(fect::fect(
        Y ~ D, data = df, index = c("id", "time"),
        method = "fe", r = 0, CV = FALSE, se = FALSE,
        parallel = FALSE, force = "two-way"
    )))
    expect_error(plot(fit, type = "loading.overlap"), regexp = "r >= 1")
})

## ---------------------------------------------------------------
## Regression test: Y.ct.full populated at control positions on GSC.
##
## Pre-fix (HANDOFF 2026-04-25): On the GSC path
## (method = "ife" + time.component.from = "nevertreated"), Y.ct.full
## at id_co positions was set via `Y.co - residuals`, which leaves NA
## at any position where Y.co is NA. This blocks user-space rolling CV
## from scoring MSPE at masked control observations -- the channel
## that R/cv-rolling.R uses to evaluate held-out positions.
##
## Fix (Stage 2, 2026-04-25): After Y.ct.full <- Y.ct on the shared
## post-bifurcation path of fect_nevertreated.R, on the IFE method
## branch (the GSC estimator) overwrite control columns with the
## model-implied factor product:
##     Y.ct.full[, id_co] <- F * t(lambda_co)
## sourced directly from est.co.best$factor and est.co.best$lambda.
##
## Tests:
##   1. fit$Y.ct.full[, fit$co] matches F * t(lambda) element-wise.
##   2. ATT.avg matches the unmodified-dev baseline (4.639593 on
##      simgsynth, set.seed(11), r=2, CV=FALSE, force="two-way").
## ---------------------------------------------------------------

test_that("GSC populates Y.ct.full[, co] with F * t(lambda)", {

  skip_on_cran()

  if (!exists("simgsynth", inherits = TRUE)) {
    suppressWarnings(try(utils::data("simgsynth", package = "fect"),
                         silent = TRUE))
  }
  skip_if_not(exists("simgsynth", inherits = TRUE),
              "simgsynth dataset not available")

  set.seed(11)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simgsynth,
      index     = c("id", "time"),
      method    = "ife",
      time.component.from = "nevertreated",
      r         = 2,
      CV        = FALSE,
      force     = "two-way",
      se        = FALSE
    )
  ))

  expect_true(!is.null(fit$factor))
  expect_true(!is.null(fit$lambda.co))
  expect_true(!is.null(fit$Y.ct.full))
  expect_true(!is.null(fit$co))

  F.hat <- as.matrix(fit$factor)
  L.co  <- as.matrix(fit$lambda.co)
  expected <- F.hat %*% t(L.co)

  observed <- fit$Y.ct.full[, fit$co, drop = FALSE]

  expect_equal(dim(observed), dim(expected))
  expect_equal(observed, expected, tolerance = 1e-10,
               ignore_attr = TRUE)
})

test_that("GSC ATT.avg unchanged after Stage 2 patch (simgsynth anchor)", {

  skip_on_cran()

  if (!exists("simgsynth", inherits = TRUE)) {
    suppressWarnings(try(utils::data("simgsynth", package = "fect"),
                         silent = TRUE))
  }
  skip_if_not(exists("simgsynth", inherits = TRUE),
              "simgsynth dataset not available")

  set.seed(11)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simgsynth,
      index     = c("id", "time"),
      method    = "ife",
      time.component.from = "nevertreated",
      r         = 2,
      CV        = FALSE,
      force     = "two-way",
      se        = FALSE
    )
  ))

  ## Anchor: 4.639593 on unmodified dev, set.seed(11), simgsynth.
  expect_equal(fit$att.avg, 4.639593, tolerance = 1e-5)
})

## ---------------------------------------------------------
## Simulation test: Interactive Fixed Effects (IFE) estimator
## DGP: Factor model with one factor and tau=2.0
## ---------------------------------------------------------

test_that("IFE estimator recovers ATT under factor model DGP", {

  skip_on_cran()

    set.seed(3001)

    N <- 30
    TT <- 15
    T0 <- 10
    Ntr <- 10
    tau <- 2.0
    nsim <- 50

    estimates <- numeric(nsim)

    for (sim in 1:nsim) {
        ## Unit and time fixed effects
        alpha_i <- rnorm(N, 0, 1)
        xi_t <- rnorm(TT, 0, 0.5)

        ## Single factor structure
        lambda_i <- rnorm(N, 0, 1)
        f_t <- rnorm(TT, 0, 1)

        Y_vec <- numeric(N * TT)
        D_vec <- integer(N * TT)
        id_vec <- integer(N * TT)
        time_vec <- integer(N * TT)

        idx <- 1
        for (i in 1:N) {
            for (t in 1:TT) {
                treated <- (i <= Ntr) && (t > T0)
                D_vec[idx] <- as.integer(treated)
                eps <- rnorm(1, 0, 1)
                Y_vec[idx] <- alpha_i[i] + xi_t[t] +
                    lambda_i[i] * f_t[t] +
                    tau * D_vec[idx] + eps
                id_vec[idx] <- i
                time_vec[idx] <- t
                idx <- idx + 1
            }
        }

        simdf <- data.frame(
            id = id_vec,
            time = time_vec,
            Y = Y_vec,
            D = D_vec
        )

        out <- suppressWarnings(fect::fect(
            Y ~ D,
            data = simdf,
            index = c("id", "time"),
            method = "ife",
            r = 1,
            CV = FALSE,
            force = "two-way",
            se = FALSE,
            parallel = FALSE
        ))

        estimates[sim] <- out$att.avg
    }

    mean_est <- mean(estimates, na.rm = TRUE)
    bias <- abs(mean_est - tau)
    expect_lt(bias, 0.8,
        label = paste0("IFE bias = ", round(bias, 4),
                       ", mean_est = ", round(mean_est, 4),
                       ", tau = ", tau))

    sd_est <- sd(estimates, na.rm = TRUE)
    expect_lt(sd_est, 2.0,
        label = paste0("IFE SD = ", round(sd_est, 4)))
})

test_that("IFE with r=0 reduces to FE and produces consistent estimates", {

  skip_on_cran()

    set.seed(3002)

    N <- 30
    TT <- 15
    T0 <- 10
    Ntr <- 10
    tau <- 2.5

    ## Pure two-way FE DGP (no factors)
    alpha_i <- rnorm(N, 0, 1)
    xi_t <- rnorm(TT, 0, 0.5)

    Y_vec <- numeric(N * TT)
    D_vec <- integer(N * TT)
    id_vec <- integer(N * TT)
    time_vec <- integer(N * TT)

    idx <- 1
    for (i in 1:N) {
        for (t in 1:TT) {
            treated <- (i <= Ntr) && (t > T0)
            D_vec[idx] <- as.integer(treated)
            eps <- rnorm(1, 0, 0.5)
            Y_vec[idx] <- alpha_i[i] + xi_t[t] + tau * D_vec[idx] + eps
            id_vec[idx] <- i
            time_vec[idx] <- t
            idx <- idx + 1
        }
    }

    simdf <- data.frame(
        id = id_vec,
        time = time_vec,
        Y = Y_vec,
        D = D_vec
    )

    out_fe <- suppressWarnings(fect::fect(
        Y ~ D,
        data = simdf,
        index = c("id", "time"),
        method = "fe",
        force = "two-way",
        se = FALSE,
        parallel = FALSE
    ))

    out_ife <- suppressWarnings(fect::fect(
        Y ~ D,
        data = simdf,
        index = c("id", "time"),
        method = "ife",
        r = 0,
        CV = FALSE,
        force = "two-way",
        se = FALSE,
        parallel = FALSE
    ))

    ## With r=0, IFE should produce the same result as FE
    expect_equal(out_fe$att.avg, out_ife$att.avg, tolerance = 0.01,
        label = "IFE(r=0) should match FE")
})

## ---------------------------------------------------------
## Simulation test: Complex Fixed Effects (CFE) estimator
## DGP: Two-way FE with a time-varying covariate and tau=2.5
## ---------------------------------------------------------

test_that("CFE estimator recovers ATT with time-varying covariate", {

  skip_on_cran()

    set.seed(4001)

    N <- 30
    TT <- 15
    T0 <- 10
    Ntr <- 10
    tau <- 2.5
    beta_x <- 1.5  # true coefficient on X
    nsim <- 50

    estimates <- numeric(nsim)

    for (sim in 1:nsim) {
        alpha_i <- rnorm(N, 0, 1)
        xi_t <- rnorm(TT, 0, 0.5)

        Y_vec <- numeric(N * TT)
        D_vec <- integer(N * TT)
        X1_vec <- numeric(N * TT)
        id_vec <- integer(N * TT)
        time_vec <- integer(N * TT)

        idx <- 1
        for (i in 1:N) {
            for (t in 1:TT) {
                treated <- (i <= Ntr) && (t > T0)
                D_vec[idx] <- as.integer(treated)
                X1_vec[idx] <- rnorm(1, 0, 1)
                eps <- rnorm(1, 0, 1)
                Y_vec[idx] <- alpha_i[i] + xi_t[t] +
                    beta_x * X1_vec[idx] +
                    tau * D_vec[idx] + eps
                id_vec[idx] <- i
                time_vec[idx] <- t
                idx <- idx + 1
            }
        }

        simdf <- data.frame(
            id = id_vec,
            time = time_vec,
            Y = Y_vec,
            D = D_vec,
            X1 = X1_vec
        )

        out <- suppressWarnings(fect::fect(
            Y ~ D + X1,
            data = simdf,
            index = c("id", "time"),
            method = "fe",
            force = "two-way",
            se = FALSE,
            parallel = FALSE
        ))

        estimates[sim] <- out$att.avg
    }

    mean_est <- mean(estimates, na.rm = TRUE)
    bias <- abs(mean_est - tau)
    expect_lt(bias, 0.5,
        label = paste0("CFE bias = ", round(bias, 4),
                       ", mean_est = ", round(mean_est, 4),
                       ", tau = ", tau))

    sd_est <- sd(estimates, na.rm = TRUE)
    expect_lt(sd_est, 1.5,
        label = paste0("CFE SD = ", round(sd_est, 4)))
})

test_that("CFE estimator handles correlated covariate correctly", {

  skip_on_cran()

    set.seed(4002)

    N <- 40
    TT <- 15
    T0 <- 10
    Ntr <- 15
    tau <- 2.5
    beta_x <- 2.0

    alpha_i <- rnorm(N, 0, 1)
    xi_t <- rnorm(TT, 0, 0.5)

    Y_vec <- numeric(N * TT)
    D_vec <- integer(N * TT)
    X1_vec <- numeric(N * TT)
    id_vec <- integer(N * TT)
    time_vec <- integer(N * TT)

    idx <- 1
    for (i in 1:N) {
        for (t in 1:TT) {
            treated <- (i <= Ntr) && (t > T0)
            D_vec[idx] <- as.integer(treated)
            ## Covariate correlated with unit FE
            X1_vec[idx] <- 0.5 * alpha_i[i] + rnorm(1, 0, 1)
            eps <- rnorm(1, 0, 1)
            Y_vec[idx] <- alpha_i[i] + xi_t[t] +
                beta_x * X1_vec[idx] +
                tau * D_vec[idx] + eps
            id_vec[idx] <- i
            time_vec[idx] <- t
            idx <- idx + 1
        }
    }

    simdf <- data.frame(
        id = id_vec,
        time = time_vec,
        Y = Y_vec,
        D = D_vec,
        X1 = X1_vec
    )

    out <- suppressWarnings(fect::fect(
        Y ~ D + X1,
        data = simdf,
        index = c("id", "time"),
        method = "fe",
        force = "two-way",
        se = FALSE,
        parallel = FALSE
    ))

    ## ATT should still be close to true tau
    expect_lt(abs(out$att.avg - tau), 1.0,
        label = paste0("CFE with correlated X: att.avg = ",
                       round(out$att.avg, 4), ", tau = ", tau))

    ## Output structure checks
    expect_true(is.numeric(out$att.avg))
    expect_true(is.matrix(out$eff))
    expect_true(!is.null(out$beta))
})

## ---------------------------------------------------------
## Edge case and boundary condition tests
## ---------------------------------------------------------

test_that("fect handles single treated unit", {

  skip_on_cran()

    set.seed(7001)

    N <- 20
    TT <- 15
    T0 <- 10
    Ntr <- 1    # only one treated unit
    tau <- 3.0

    alpha_i <- rnorm(N, 0, 1)
    xi_t <- rnorm(TT, 0, 0.5)

    Y_vec <- numeric(N * TT)
    D_vec <- integer(N * TT)
    id_vec <- integer(N * TT)
    time_vec <- integer(N * TT)

    idx <- 1
    for (i in 1:N) {
        for (t in 1:TT) {
            treated <- (i <= Ntr) && (t > T0)
            D_vec[idx] <- as.integer(treated)
            eps <- rnorm(1, 0, 0.5)
            Y_vec[idx] <- alpha_i[i] + xi_t[t] + tau * D_vec[idx] + eps
            id_vec[idx] <- i
            time_vec[idx] <- t
            idx <- idx + 1
        }
    }

    simdf <- data.frame(
        id = id_vec,
        time = time_vec,
        Y = Y_vec,
        D = D_vec
    )

    ## Should not error with a single treated unit
    out <- suppressWarnings(fect::fect(
        Y ~ D,
        data = simdf,
        index = c("id", "time"),
        method = "fe",
        force = "two-way",
        se = FALSE,
        parallel = FALSE
    ))

    expect_true(is.numeric(out$att.avg))
    expect_true(!is.na(out$att.avg))
    expect_true(is.matrix(out$eff))
})

test_that("fect handles missing data (scattered NAs in Y)", {

  skip_on_cran()

    set.seed(7002)

    N <- 25
    TT <- 15
    T0 <- 10
    Ntr <- 8
    tau <- 2.0

    alpha_i <- rnorm(N, 0, 1)
    xi_t <- rnorm(TT, 0, 0.5)

    Y_vec <- numeric(N * TT)
    D_vec <- integer(N * TT)
    id_vec <- integer(N * TT)
    time_vec <- integer(N * TT)

    idx <- 1
    for (i in 1:N) {
        for (t in 1:TT) {
            treated <- (i <= Ntr) && (t > T0)
            D_vec[idx] <- as.integer(treated)
            eps <- rnorm(1, 0, 1)
            Y_vec[idx] <- alpha_i[i] + xi_t[t] + tau * D_vec[idx] + eps
            id_vec[idx] <- i
            time_vec[idx] <- t
            idx <- idx + 1
        }
    }

    ## Introduce ~5% missing values in Y (but NOT in treated post-period)
    miss_candidates <- which(D_vec == 0)
    n_miss <- min(floor(0.05 * length(Y_vec)), length(miss_candidates))
    miss_idx <- sample(miss_candidates, n_miss)
    Y_vec[miss_idx] <- NA

    simdf <- data.frame(
        id = id_vec,
        time = time_vec,
        Y = Y_vec,
        D = D_vec
    )

    ## Should handle missing data without error
    out <- suppressWarnings(fect::fect(
        Y ~ D,
        data = simdf,
        index = c("id", "time"),
        method = "fe",
        force = "two-way",
        se = FALSE,
        parallel = FALSE,
        na.rm = TRUE
    ))

    expect_true(is.numeric(out$att.avg))
    expect_true(!is.na(out$att.avg))
    ## ATT should still be roughly correct
    expect_lt(abs(out$att.avg - tau), 1.5,
        label = paste0("Missing data FE: att.avg = ",
                       round(out$att.avg, 4), ", tau = ", tau))
})

test_that("fect output structure is complete (FE)", {

  skip_on_cran()

    set.seed(7003)

    N <- 20
    TT <- 10
    T0 <- 6
    Ntr <- 8

    Y_vec <- numeric(N * TT)
    D_vec <- integer(N * TT)
    id_vec <- integer(N * TT)
    time_vec <- integer(N * TT)

    idx <- 1
    for (i in 1:N) {
        for (t in 1:TT) {
            treated <- (i <= Ntr) && (t > T0)
            D_vec[idx] <- as.integer(treated)
            Y_vec[idx] <- rnorm(1) + 2.0 * D_vec[idx]
            id_vec[idx] <- i
            time_vec[idx] <- t
            idx <- idx + 1
        }
    }

    simdf <- data.frame(
        id = id_vec,
        time = time_vec,
        Y = Y_vec,
        D = D_vec
    )

    out <- suppressWarnings(fect::fect(
        Y ~ D,
        data = simdf,
        index = c("id", "time"),
        method = "fe",
        force = "two-way",
        se = FALSE,
        parallel = FALSE
    ))

    ## Check required output slots exist
    expect_true(!is.null(out$att.avg))
    expect_true(!is.null(out$att.avg.unit))
    expect_true(!is.null(out$eff))
    expect_true(!is.null(out$Y.ct))
    expect_true(!is.null(out$time))
    expect_true(!is.null(out$att))
    expect_true(!is.null(out$count))
    expect_true(!is.null(out$sigma2))
})

test_that("fect with many treated units still works", {

  skip_on_cran()

    set.seed(7004)

    N <- 25
    TT <- 12
    T0 <- 8
    Ntr <- 20  # 80% treated
    tau <- 1.5

    alpha_i <- rnorm(N, 0, 0.5)
    xi_t <- rnorm(TT, 0, 0.3)

    Y_vec <- numeric(N * TT)
    D_vec <- integer(N * TT)
    id_vec <- integer(N * TT)
    time_vec <- integer(N * TT)

    idx <- 1
    for (i in 1:N) {
        for (t in 1:TT) {
            treated <- (i <= Ntr) && (t > T0)
            D_vec[idx] <- as.integer(treated)
            eps <- rnorm(1, 0, 1)
            Y_vec[idx] <- alpha_i[i] + xi_t[t] + tau * D_vec[idx] + eps
            id_vec[idx] <- i
            time_vec[idx] <- t
            idx <- idx + 1
        }
    }

    simdf <- data.frame(
        id = id_vec,
        time = time_vec,
        Y = Y_vec,
        D = D_vec
    )

    ## Should work even with many treated / few control
    out <- suppressWarnings(fect::fect(
        Y ~ D,
        data = simdf,
        index = c("id", "time"),
        method = "fe",
        force = "two-way",
        se = FALSE,
        parallel = FALSE
    ))

    expect_true(is.numeric(out$att.avg))
    expect_true(!is.na(out$att.avg))
})

# Regression test for the "Unsupported bootstrap method: fe" crash when CV
# selects r = 0 under method = "gsynth".
#
# Chain of the original bug (fect <= 2.2.0, diagnosed 2026-04-21):
#   1. fect(method = "gsynth", CV = TRUE, r = c(0, ...), se = TRUE) dispatches
#      to fect_nevertreated.
#   2. When CV picks r.cv == 0, fect_nevertreated used to relabel the outgoing
#      `method` slot to "fe".
#   3. fect_boot reads out$method and dispatches per-iteration bootstrap. It
#      has branches for gsynth / ife / mc / cfe but no "fe" branch, so it
#      would stop() with "Unsupported bootstrap method: fe".
#
# The fix preserves method = "gsynth" (or "cfe") regardless of r.cv. The
# fixture below is a pure two-way-FE DGP that deterministically makes CV
# prefer r = 0 over r >= 1, so this test exercises the r.cv == 0 code path.

test_that("gsynth + CV + se survives when r.cv == 0 (regression for dispatcher crash)", {

  ## Bootstrap regression test: heavy for CRAN's test budget. Runs in
  ## the standard test environment (devtools::test() / R CMD check
  ## without --as-cran sandboxing).
  skip_on_cran()

  set.seed(20260421)
  N <- 20
  TT <- 15
  id <- rep(seq_len(N), each = TT)
  time <- rep(seq_len(TT), times = N)
  alpha <- rnorm(N)
  xi <- rnorm(TT)
  D <- as.integer(id <= 5 & time >= 10)
  Y <- alpha[id] + xi[time] + 2 * D + rnorm(N * TT, sd = 0.5)
  dat <- data.frame(id = id, time = time, Y = Y, D = D)

  # Before the fix this call raised:
  #   "Unsupported bootstrap method: fe"
  out <- fect::fect(
    Y ~ D,
    data = dat,
    index = c("id", "time"),
    method = "gsynth",
    CV = TRUE,
    r = c(0, 3),
    se = TRUE,
    nboots = 5,
    parallel = FALSE
  )

  expect_s3_class(out, "fect")
  expect_equal(out$r.cv, 0)
  # Method must remain "gsynth" after CV; relabel to "fe" was the bug.
  expect_equal(out$method, "gsynth")
  expect_true(is.numeric(out$att.avg))
  expect_true(!is.null(out$est.avg))
})

## ---------------------------------------------------------------
## Regression test: Phase A bootstrap parallel state across calls.
##
## Pre-fix bug (HANDOFF 2026-04-25): boot.R Phase A's foreach %dopar%
## inherited whatever foreach backend was registered in the global
## registry. After any prior fect parallel call, run_dopar_retry's
## on.exit re-registered doFuture, polluting subsequent calls --
## variant (iii) Phase A in a forest session ran 8x slower than the
## same call in isolation.
##
## Fix (Stage 1, 2026-04-25):
##   1. Phase A migrated from foreach %dopar% to future_lapply
##      (boot.R:884). Backend is now governed by the active future plan
##      set in default.R, not by a global foreach registry.
##   2. Removed `doFuture::registerDoFuture()` from run_dopar_retry's
##      on.exit (boot.R:1553).
##
## Test: two consecutive fect(parallel = TRUE) calls in the same R
## process. The second call's wall-time must be within 2x of the
## first; pre-fix this could blow up to 8x.
## ---------------------------------------------------------------

test_that("Phase A wall-time is stable across consecutive parallel calls", {

  skip_on_cran()

  one_call <- function(seed) {
    set.seed(seed)
    t0 <- Sys.time()
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data      = simdata,
        index     = c("id", "time"),
        method    = "ife",
        r         = 1,
        CV        = FALSE,
        force     = "two-way",
        time.component.from = "nevertreated",
        se        = TRUE,
        vartype   = "bootstrap",
        nboots    = 20,
        parallel  = TRUE,
        cores     = 2
      )
    ))
    as.numeric(difftime(Sys.time(), t0, units = "secs"))
  }

  ## Warm-up: pay future plan / package-load cost outside the measurement.
  invisible(one_call(seed = 1))

  ## First measured call.
  t1 <- one_call(seed = 2)
  ## Second measured call -- must not blow up due to backend pollution.
  t2 <- one_call(seed = 3)

  ratio <- t2 / max(t1, 0.5)
  ## 2x is the plan budget; allow 3x ceiling to absorb GC / OS jitter on CI.
  expect_lt(ratio, 3)
})

## ---------------------------------------------------------------
## Tests for the $sample logical mask on the fect() return value.
##
## $sample is a logical matrix (same dims as $Y.dat) that is TRUE for
## every cell used in any part of the estimation procedure:
##   obs.missing == 1  (treated)
##   obs.missing == 2  (control / pre-treatment)
##   obs.missing == 5  (placebo or carryover period)
## Cells with obs.missing == 3 (missing/unbalanced) or 4 (removed unit)
## are FALSE.
##
## Derived via:  sample <- obs.missing %in% c(1L, 2L, 5L)
## ---------------------------------------------------------------

suppressWarnings(data("simdata", package = "fect"))

## -- S.1  Slot exists and is a logical matrix --

test_that("S.1: fit$sample exists and is a logical matrix", {

  skip_on_cran()

  set.seed(1)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
      method = "fe", CV = FALSE, se = FALSE, parallel = FALSE
    )
  ))

  expect_true("sample" %in% names(fit))
  expect_true(is.matrix(fit$sample))
  expect_true(is.logical(fit$sample))
})


## -- S.2  Dimensions match Y.dat --

test_that("S.2: fit$sample has the same dimensions as fit$Y.dat", {

  skip_on_cran()

  set.seed(1)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
      method = "fe", CV = FALSE, se = FALSE, parallel = FALSE
    )
  ))

  expect_equal(dim(fit$sample), dim(fit$Y.dat))
})


## -- S.3  Consistency with obs.missing codes --
## sample must equal obs.missing %in% c(1L, 2L, 5L) cell-for-cell.

test_that("S.3: fit$sample equals obs.missing %in% c(1L, 2L, 5L)", {

  skip_on_cran()

  set.seed(1)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
      method = "fe", CV = FALSE, se = FALSE, parallel = FALSE
    )
  ))

  expected <- matrix(fit$obs.missing %in% c(1L, 2L, 5L),
                     nrow = nrow(fit$obs.missing), ncol = ncol(fit$obs.missing),
                     dimnames = dimnames(fit$obs.missing))
  expect_identical(fit$sample, expected)
})


## -- S.4  Removed units are FALSE --
## Units with rm.id (obs.missing == 4) must be entirely FALSE in $sample.

test_that("S.4: cells with obs.missing == 4 (removed) are FALSE in $sample", {

  skip_on_cran()

  set.seed(1)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
      method = "fe", CV = FALSE, se = FALSE, parallel = FALSE
    )
  ))

  removed_cells <- fit$obs.missing == 4L
  ## If no removed units, skip the body (not an error — just no removed units).
  if (any(removed_cells)) {
    expect_true(all(!fit$sample[removed_cells]))
  } else {
    expect_true(TRUE)  ## no removed units — pass trivially
  }
})


## -- S.5  Missing cells are FALSE --
## obs.missing == 3 cells must be FALSE in $sample.

test_that("S.5: cells with obs.missing == 3 (missing) are FALSE in $sample", {

  skip_on_cran()

  set.seed(1)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
      method = "fe", CV = FALSE, se = FALSE, parallel = FALSE
    )
  ))

  missing_cells <- fit$obs.missing == 3L
  if (any(missing_cells)) {
    expect_true(all(!fit$sample[missing_cells]))
  } else {
    expect_true(TRUE)
  }
})


## -- S.6  Placebo period cells are TRUE --
## When placeboTest = TRUE, pre-treatment placebo cells (obs.missing == 5)
## must appear as TRUE in $sample.

test_that("S.6: placebo cells (obs.missing == 5) are TRUE in $sample", {

  skip_on_cran()

  set.seed(2)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
      method = "fe", CV = FALSE, se = FALSE, parallel = FALSE,
      placeboTest = TRUE, placebo.period = c(-2, -1)
    )
  ))

  placebo_cells <- fit$obs.missing == 5L
  if (any(placebo_cells)) {
    expect_true(all(fit$sample[placebo_cells]))
  } else {
    ## If no placebo cells found, the placeboTest may not have produced code 5
    ## under this data/seed — at minimum check slot still exists and is logical.
    expect_true(is.logical(fit$sample))
  }
})


## -- S.7  at_least_some_TRUE: basic sanity that $sample is not all FALSE --

test_that("S.7: fit$sample has at least some TRUE cells (treated + control)", {

  skip_on_cran()

  set.seed(1)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
      method = "fe", CV = FALSE, se = FALSE, parallel = FALSE
    )
  ))

  expect_true(any(fit$sample))
})


## -- S.8  Treated cells are TRUE --
## obs.missing == 1 cells must be TRUE in $sample.

test_that("S.8: treated cells (obs.missing == 1) are TRUE in $sample", {

  skip_on_cran()

  set.seed(1)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
      method = "fe", CV = FALSE, se = FALSE, parallel = FALSE
    )
  ))

  treated_cells <- fit$obs.missing == 1L
  expect_true(any(treated_cells))
  expect_true(all(fit$sample[treated_cells]))
})


## -- S.9  Control cells are TRUE --
## obs.missing == 2 cells must be TRUE in $sample.

test_that("S.9: control cells (obs.missing == 2) are TRUE in $sample", {

  skip_on_cran()

  set.seed(1)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
      method = "fe", CV = FALSE, se = FALSE, parallel = FALSE
    )
  ))

  control_cells <- fit$obs.missing == 2L
  expect_true(any(control_cells))
  expect_true(all(fit$sample[control_cells]))
})


## -- S.10  Method = ife gives same slot structure --

test_that("S.10: fit$sample present and logical for method='ife'", {

  skip_on_cran()

  set.seed(3)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D + X1 + X2, data = simdata, index = c("id", "time"),
      method = "ife", r = 2, CV = FALSE, se = FALSE, parallel = FALSE
    )
  ))

  expect_true("sample" %in% names(fit))
  expect_true(is.logical(fit$sample))
  expect_equal(dim(fit$sample), dim(fit$Y.dat))
})

## ---------------------------------------------------------------
## Tests for parallel CV folds in fect_nevertreated.
## Section originally also labeled "G" (now disambiguated via filename
## as test-score-parallel-cv.R; the integration "G" is in
## test-score-nevertreated.R).
##
## Originally Section G "Parallel CV Folds" of test-score-unify.R;
## split out 2026-05-03. Shared fixtures live in helper-score-unify.R.
## ---------------------------------------------------------------

## ---------------------------------------------------------------
## Section G: Parallel CV Folds in fect_nevertreated
##
## Tests for REQ-parallel-cv: verifies that parallel=TRUE produces
## identical results to parallel=FALSE (sequential), that
## reproducibility holds under parallelism with fixed seeds, that
## default behavior is unchanged, that the LOO path is unaffected,
## and that edge cases work correctly.
##
## Follows test-spec.md for REQ-parallel-cv.
## Tolerances: 1e-10 for CV score differences (per test-spec.md).
## ---------------------------------------------------------------

## -- G.1  Sequential-Parallel Equivalence: IFE, all_units ----------

test_that("G.1: parallel CV matches sequential — IFE, all_units", {

  skip_on_cran()

  dat <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

  set.seed(123)
  result_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = FALSE
    )
  ))

  set.seed(123)
  result_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = TRUE,
      cores        = 2
    )
  ))

  ## r.cv must be identical
  expect_identical(result_seq$r.cv, result_par$r.cv)

  ## CV.out matrix must match within tolerance 1e-10
  cv_diff <- max(abs(result_seq$CV.out - result_par$CV.out))
  expect_true(cv_diff < 1e-10,
    info = sprintf("CV.out max diff = %.2e (tolerance = 1e-10)", cv_diff))
})

## -- G.2  Sequential-Parallel Equivalence: IFE, treated_units ------

test_that("G.2: parallel CV matches sequential — IFE, treated_units", {

  skip_on_cran()

  dat <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

  set.seed(123)
  result_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "treated_units",
      se           = FALSE,
      parallel     = FALSE
    )
  ))

  set.seed(123)
  result_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "treated_units",
      se           = FALSE,
      parallel     = TRUE,
      cores        = 2
    )
  ))

  expect_identical(result_seq$r.cv, result_par$r.cv)

  cv_diff <- max(abs(result_seq$CV.out - result_par$CV.out))
  expect_true(cv_diff < 1e-10,
    info = sprintf("CV.out max diff = %.2e (tolerance = 1e-10)", cv_diff))
})

## -- G.3  Sequential-Parallel Equivalence: CFE, all_units ----------

test_that("G.3: parallel CV matches sequential — CFE, all_units", {

  skip_on_cran()

  dat <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

  set.seed(123)
  result_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "cfe",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = FALSE
    )
  ))

  set.seed(123)
  result_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "cfe",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = TRUE,
      cores        = 2
    )
  ))

  expect_identical(result_seq$r.cv, result_par$r.cv)

  cv_diff <- max(abs(result_seq$CV.out - result_par$CV.out))
  expect_true(cv_diff < 1e-10,
    info = sprintf("CV.out max diff = %.2e (tolerance = 1e-10)", cv_diff))
})

## -- G.4  Sequential-Parallel Equivalence: CFE, treated_units ------

test_that("G.4: parallel CV matches sequential — CFE, treated_units", {

  skip_on_cran()

  dat <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

  set.seed(123)
  result_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "cfe",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "treated_units",
      se           = FALSE,
      parallel     = FALSE
    )
  ))

  set.seed(123)
  result_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "cfe",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "treated_units",
      se           = FALSE,
      parallel     = TRUE,
      cores        = 2
    )
  ))

  expect_identical(result_seq$r.cv, result_par$r.cv)

  cv_diff <- max(abs(result_seq$CV.out - result_par$CV.out))
  expect_true(cv_diff < 1e-10,
    info = sprintf("CV.out max diff = %.2e (tolerance = 1e-10)", cv_diff))
})

## -- G.5  Reproducibility Under Parallelism ------------------------

test_that("G.5: parallel CV is reproducible with same seed", {

  skip_on_cran()

  dat <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

  set.seed(123)
  result_par1 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = TRUE,
      cores        = 2,
      seed         = 12345
    )
  ))

  set.seed(123)
  result_par2 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = TRUE,
      cores        = 2,
      seed         = 12345
    )
  ))

  expect_identical(result_par1$r.cv, result_par2$r.cv)
  expect_identical(result_par1$CV.out, result_par2$CV.out)
})

## -- G.6  Default Behavior Unchanged -------------------------------

test_that("G.6: default (no parallel arg) behaves as sequential", {

  skip_on_cran()

  dat <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

  ## Call without specifying parallel or cores — should default to sequential
  set.seed(123)
  result_default <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE
    )
  ))

  set.seed(123)
  result_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = FALSE
    )
  ))

  expect_identical(result_default$r.cv, result_seq$r.cv)
  expect_identical(result_default$CV.out, result_seq$CV.out)
})

## -- G.7  LOO Path Unaffected --------------------------------------

test_that("G.7: LOO path is unaffected by parallel flag", {

  skip_on_cran()

  dat <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

  set.seed(123)
  result_loo_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      cv.method    = "loo",
      se           = FALSE,
      parallel     = FALSE
    )
  ))

  set.seed(123)
  result_loo_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      cv.method    = "loo",
      se           = FALSE,
      parallel     = TRUE,
      cores        = 2
    )
  ))

  ## LOO should produce identical results regardless of parallel flag
  expect_identical(result_loo_seq$r.cv, result_loo_par$r.cv)
  expect_identical(result_loo_seq$CV.out, result_loo_par$CV.out)
})

## -- G.8  Edge Case: k = 1 ----------------------------------------

test_that("G.8: edge case — k = 1 with parallel=TRUE runs without error", {

  skip_on_cran()

  dat <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

  set.seed(123)
  result_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 1,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = TRUE,
      cores        = 2
    )
  ))

  set.seed(123)
  result_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 1,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = FALSE
    )
  ))

  expect_identical(result_par$r.cv, result_seq$r.cv)
})

## -- G.9  Edge Case: cores = 1 ------------------------------------

test_that("G.9: edge case — cores = 1 behaves as sequential", {

  skip_on_cran()

  dat <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

  set.seed(123)
  result_c1 <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = TRUE,
      cores        = 1
    )
  ))

  set.seed(123)
  result_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = FALSE
    )
  ))

  cv_diff <- max(abs(result_c1$CV.out - result_seq$CV.out))
  expect_true(cv_diff < 1e-10,
    info = sprintf("CV.out max diff = %.2e (tolerance = 1e-10)", cv_diff))
})

## -- G.10  Edge Case: cores = NULL (auto-detect) ------------------

test_that("G.10: edge case — cores = NULL auto-detects and runs", {

  skip_on_cran()

  dat <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

  set.seed(123)
  result_auto <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = TRUE,
      cores        = NULL
    )
  ))

  set.seed(123)
  result_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = FALSE
    )
  ))

  cv_diff <- max(abs(result_auto$CV.out - result_seq$CV.out))
  expect_true(cv_diff < 1e-10,
    info = sprintf("CV.out max diff = %.2e (tolerance = 1e-10)", cv_diff))
})

## -- G.11  Edge Case: parallel=FALSE with cores specified ----------

test_that("G.11: edge case — parallel=FALSE ignores cores", {

  skip_on_cran()

  dat <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

  set.seed(123)
  result_no_par <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = FALSE,
      cores        = 4
    )
  ))

  set.seed(123)
  result_seq <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = FALSE
    )
  ))

  expect_identical(result_no_par$r.cv, result_seq$r.cv)
  expect_identical(result_no_par$CV.out, result_seq$CV.out)
})

## -- G.12  Timing Benchmark: Parallel vs Sequential ----------------
## This is an informational benchmark; we report speedup but do not
## hard-fail on specific speedup thresholds (per test-spec.md:
## "Not a pass/fail test, but auditor should measure and report").

test_that("G.12: timing benchmark — parallel vs sequential with 10 cores", {

  skip_on_cran()

  ## Larger dataset for meaningful timing differences
  dat <- make_factor_data(N = 100, TT = 30, Ntr = 25, r = 2, seed = 99)

  cat("\n=== Parallel CV Timing Benchmark ===\n")

  ## --- IFE, all_units ---
  set.seed(123)
  t_seq_ife_au <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data         = dat,
        index        = c("id", "time"),
        method       = "ife",
        force        = "two-way",
        time.component.from = "nevertreated",
        CV           = TRUE,
        r            = c(0, 5),
        k            = 10,
        cv.method    = "all_units",
        se           = FALSE,
        parallel     = FALSE
      )
    ))
  )

  set.seed(123)
  t_par_ife_au <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data         = dat,
        index        = c("id", "time"),
        method       = "ife",
        force        = "two-way",
        time.component.from = "nevertreated",
        CV           = TRUE,
        r            = c(0, 5),
        k            = 10,
        cv.method    = "all_units",
        se           = FALSE,
        parallel     = TRUE,
        cores        = 10
      )
    ))
  )

  cat(sprintf("  IFE all_units  sequential: %6.2f sec\n", t_seq_ife_au["elapsed"]))
  cat(sprintf("  IFE all_units  parallel:   %6.2f sec (10 cores)\n", t_par_ife_au["elapsed"]))
  speedup_ife_au <- t_seq_ife_au["elapsed"] / max(t_par_ife_au["elapsed"], 0.001)
  cat(sprintf("  Speedup: %.2fx\n\n", speedup_ife_au))

  ## --- IFE, treated_units ---
  set.seed(123)
  t_seq_ife_tu <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data         = dat,
        index        = c("id", "time"),
        method       = "ife",
        force        = "two-way",
        time.component.from = "nevertreated",
        CV           = TRUE,
        r            = c(0, 5),
        k            = 10,
        cv.method    = "treated_units",
        se           = FALSE,
        parallel     = FALSE
      )
    ))
  )

  set.seed(123)
  t_par_ife_tu <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data         = dat,
        index        = c("id", "time"),
        method       = "ife",
        force        = "two-way",
        time.component.from = "nevertreated",
        CV           = TRUE,
        r            = c(0, 5),
        k            = 10,
        cv.method    = "treated_units",
        se           = FALSE,
        parallel     = TRUE,
        cores        = 10
      )
    ))
  )

  cat(sprintf("  IFE treated_units  sequential: %6.2f sec\n", t_seq_ife_tu["elapsed"]))
  cat(sprintf("  IFE treated_units  parallel:   %6.2f sec (10 cores)\n", t_par_ife_tu["elapsed"]))
  speedup_ife_tu <- t_seq_ife_tu["elapsed"] / max(t_par_ife_tu["elapsed"], 0.001)
  cat(sprintf("  Speedup: %.2fx\n\n", speedup_ife_tu))

  ## --- CFE, all_units ---
  set.seed(123)
  t_seq_cfe_au <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data         = dat,
        index        = c("id", "time"),
        method       = "cfe",
        force        = "two-way",
        time.component.from = "nevertreated",
        CV           = TRUE,
        r            = c(0, 5),
        k            = 10,
        cv.method    = "all_units",
        se           = FALSE,
        parallel     = FALSE
      )
    ))
  )

  set.seed(123)
  t_par_cfe_au <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data         = dat,
        index        = c("id", "time"),
        method       = "cfe",
        force        = "two-way",
        time.component.from = "nevertreated",
        CV           = TRUE,
        r            = c(0, 5),
        k            = 10,
        cv.method    = "all_units",
        se           = FALSE,
        parallel     = TRUE,
        cores        = 10
      )
    ))
  )

  cat(sprintf("  CFE all_units  sequential: %6.2f sec\n", t_seq_cfe_au["elapsed"]))
  cat(sprintf("  CFE all_units  parallel:   %6.2f sec (10 cores)\n", t_par_cfe_au["elapsed"]))
  speedup_cfe_au <- t_seq_cfe_au["elapsed"] / max(t_par_cfe_au["elapsed"], 0.001)
  cat(sprintf("  Speedup: %.2fx\n\n", speedup_cfe_au))

  ## --- CFE, treated_units ---
  set.seed(123)
  t_seq_cfe_tu <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data         = dat,
        index        = c("id", "time"),
        method       = "cfe",
        force        = "two-way",
        time.component.from = "nevertreated",
        CV           = TRUE,
        r            = c(0, 5),
        k            = 10,
        cv.method    = "treated_units",
        se           = FALSE,
        parallel     = FALSE
      )
    ))
  )

  set.seed(123)
  t_par_cfe_tu <- system.time(
    suppressWarnings(suppressMessages(
      fect::fect(
        Y ~ D,
        data         = dat,
        index        = c("id", "time"),
        method       = "cfe",
        force        = "two-way",
        time.component.from = "nevertreated",
        CV           = TRUE,
        r            = c(0, 5),
        k            = 10,
        cv.method    = "treated_units",
        se           = FALSE,
        parallel     = TRUE,
        cores        = 10
      )
    ))
  )

  cat(sprintf("  CFE treated_units  sequential: %6.2f sec\n", t_seq_cfe_tu["elapsed"]))
  cat(sprintf("  CFE treated_units  parallel:   %6.2f sec (10 cores)\n", t_par_cfe_tu["elapsed"]))
  speedup_cfe_tu <- t_seq_cfe_tu["elapsed"] / max(t_par_cfe_tu["elapsed"], 0.001)
  cat(sprintf("  Speedup: %.2fx\n\n", speedup_cfe_tu))

  cat("=== End Parallel CV Timing Benchmark ===\n")

  ## Informational — always passes; speedup is reported in test output
  expect_true(TRUE)
})

## -- G.13  Property: Backend Cleanup After Parallel CV -------------

test_that("G.13: parallel backend is restored after fect() returns", {

  skip_on_cran()

  dat <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

  old_plan <- future::plan()

  set.seed(123)
  suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data         = dat,
      index        = c("id", "time"),
      method       = "ife",
      force        = "two-way",
      time.component.from = "nevertreated",
      CV           = TRUE,
      r            = c(0, 3),
      k            = 5,
      cv.method    = "all_units",
      se           = FALSE,
      parallel     = TRUE,
      cores        = 2
    )
  ))

  new_plan <- future::plan()

  ## The future plan class should be restored
  expect_identical(class(old_plan), class(new_plan))
})

## v2.3.1 regression: when W is supplied, fit$est.att / fit$est.avg /
## plot(fit) / print(fit) all agree (W-weighted), and the redundant *.W
## slots are stripped from the user-facing fit object.

test_that("weighted fit: plot trajectory == fit$est.att (W-weighted)", {
  skip_on_cran()

  suppressWarnings(try(data("turnout", package = "fect"), silent = TRUE))
  expect_true(exists("turnout"))

  set.seed(123456)
  ub <- turnout[
    -c(which(turnout$abb == "WY")[1:15],
       sample(seq_len(nrow(turnout)), 50, replace = FALSE)),
  ]
  ub$weight <- runif(nrow(ub), 1, 10)

  fit <- suppressMessages(suppressWarnings(fect::fect(
    turnout ~ policy_edr + policy_mail_in + policy_motor,
    data = ub, index = c("abb", "year"),
    method = "mc", CV = TRUE, k = 5, se = TRUE,
    nboots = 30, min.T0 = 5, seed = 42,
    W = "weight", na.rm = TRUE, parallel = FALSE
  )))

  ## the plot data layer must match est.att exactly
  p <- plot(fit, show.points = FALSE)
  plot_dat <- p$data
  expect_true(all(c("Period", "ATT", "CI.lower", "CI.upper") %in% colnames(plot_dat)))

  table <- fit$est.att
  table_periods <- as.numeric(rownames(table))

  ## intersect on shared periods (plot may restrict event-time window)
  shared <- intersect(plot_dat$Period, table_periods)
  expect_gt(length(shared), 0L)

  plot_atts  <- plot_dat$ATT[match(shared, plot_dat$Period)]
  table_atts <- table[as.character(shared), "ATT"]
  expect_equal(unname(plot_atts), unname(table_atts), tolerance = 1e-12)
})

test_that("weighted fit: *.W parallel slots are stripped", {
  skip_on_cran()

  suppressWarnings(try(data("turnout", package = "fect"), silent = TRUE))
  set.seed(123456)
  ub <- turnout[
    -c(which(turnout$abb == "WY")[1:15],
       sample(seq_len(nrow(turnout)), 50, replace = FALSE)),
  ]
  ub$weight <- runif(nrow(ub), 1, 10)

  fit <- suppressMessages(suppressWarnings(fect::fect(
    turnout ~ policy_edr + policy_mail_in + policy_motor,
    data = ub, index = c("abb", "year"),
    method = "mc", CV = TRUE, k = 5, se = TRUE,
    nboots = 30, min.T0 = 5, seed = 42,
    W = "weight", na.rm = TRUE, parallel = FALSE
  )))

  ## Inferential *.W slots removed
  expect_null(fit$est.att.W)
  expect_null(fit$est.avg.W)
  expect_null(fit$att.W.bound)
  expect_null(fit$att.W.boot)
  expect_null(fit$att.W.vcov)
  expect_null(fit$est.placebo.W)
  expect_null(fit$est.carryover.W)

  ## Per-method *.W slots removed
  expect_null(fit$att.on.W)
  expect_null(fit$att.avg.W)
  expect_null(fit$time.on.W)
  expect_null(fit$count.on.W)
  expect_null(fit$att.on.sum.W)
  expect_null(fit$W.on.sum)
  expect_null(fit$att.placebo.W)
  expect_null(fit$att.carryover.W)
})

test_that("weighted fit: print() aggregate equals est.avg (W-weighted)", {
  skip_on_cran()

  suppressWarnings(try(data("turnout", package = "fect"), silent = TRUE))
  set.seed(123456)
  ub <- turnout[
    -c(which(turnout$abb == "WY")[1:15],
       sample(seq_len(nrow(turnout)), 50, replace = FALSE)),
  ]
  ub$weight <- runif(nrow(ub), 1, 10)

  fit <- suppressMessages(suppressWarnings(fect::fect(
    turnout ~ policy_edr + policy_mail_in + policy_motor,
    data = ub, index = c("abb", "year"),
    method = "mc", CV = TRUE, k = 5, se = TRUE,
    nboots = 30, min.T0 = 5, seed = 42,
    W = "weight", na.rm = TRUE, parallel = FALSE
  )))

  out_lines <- capture.output(print(fit))
  ## label change confirms est.avg is now W-weighted
  expect_true(any(grepl("Tr obs sample-weighted \\(W\\)", out_lines)))
  expect_false(any(grepl("Tr obs equally weighted", out_lines, fixed = TRUE)))
})

test_that("W.est alone: aggregation unweighted, label unweighted", {
  skip_on_cran()

  suppressWarnings(try(data("turnout", package = "fect"), silent = TRUE))
  set.seed(123456)
  ub <- turnout[
    -c(which(turnout$abb == "WY")[1:15],
       sample(seq_len(nrow(turnout)), 50, replace = FALSE)),
  ]
  ub$weight <- runif(nrow(ub), 1, 10)

  fit <- suppressMessages(suppressWarnings(fect::fect(
    turnout ~ policy_edr + policy_mail_in + policy_motor,
    data = ub, index = c("abb", "year"),
    method = "mc", CV = TRUE, k = 5, se = TRUE,
    nboots = 30, min.T0 = 5, seed = 42,
    W.est = "weight",
    na.rm = TRUE, parallel = FALSE
  )))

  ## Print label is unweighted
  out_lines <- capture.output(print(fit))
  expect_true(any(grepl("Tr obs equally weighted", out_lines, fixed = TRUE)))
  expect_false(any(grepl("sample-weighted", out_lines, fixed = TRUE)))

  ## per-role flags stashed on the fit
  expect_true(isTRUE(fit$W.in.fit))
  expect_false(isTRUE(fit$W.in.agg))

  ## Slots stripped
  expect_null(fit$est.att.W)
  expect_null(fit$att.on.W)
})

test_that("W.agg alone: fit matches W=NULL fit; aggregation is W-weighted", {
  skip_on_cran()

  suppressWarnings(try(data("turnout", package = "fect"), silent = TRUE))
  set.seed(123456)
  ub <- turnout[
    -c(which(turnout$abb == "WY")[1:15],
       sample(seq_len(nrow(turnout)), 50, replace = FALSE)),
  ]
  ub$weight <- runif(nrow(ub), 1, 10)

  fit_agg <- suppressMessages(suppressWarnings(fect::fect(
    turnout ~ policy_edr + policy_mail_in + policy_motor,
    data = ub, index = c("abb", "year"),
    method = "mc", CV = TRUE, k = 5, se = TRUE,
    nboots = 30, min.T0 = 5, seed = 42,
    W.agg = "weight",
    na.rm = TRUE, parallel = FALSE
  )))

  fit_null <- suppressMessages(suppressWarnings(fect::fect(
    turnout ~ policy_edr + policy_mail_in + policy_motor,
    data = ub, index = c("abb", "year"),
    method = "mc", CV = TRUE, k = 5, se = TRUE,
    nboots = 30, min.T0 = 5, seed = 42,
    na.rm = TRUE, parallel = FALSE
  )))

  ## Same outcome model (both fit unweighted): residuals/eff matrix identical
  expect_equal(fit_agg$eff, fit_null$eff, tolerance = 1e-10)

  ## But aggregation differs (W-weighted vs unweighted)
  expect_false(isTRUE(all.equal(
    unname(fit_agg$est.avg[1, "ATT.avg"]),
    unname(fit_null$est.avg[1, "ATT.avg"]),
    tolerance = 1e-6
  )))

  ## Print label is sample-weighted
  out_lines <- capture.output(print(fit_agg))
  expect_true(any(grepl("Tr obs sample-weighted \\(W\\)", out_lines)))

  expect_false(isTRUE(fit_agg$W.in.fit))
  expect_true(isTRUE(fit_agg$W.in.agg))
})

test_that("Distinct W.est and W.agg columns: errors in v2.3.1", {
  skip_on_cran()

  suppressWarnings(try(data("turnout", package = "fect"), silent = TRUE))
  set.seed(123456)
  ub <- turnout[
    -c(which(turnout$abb == "WY")[1:15],
       sample(seq_len(nrow(turnout)), 50, replace = FALSE)),
  ]
  ub$weight     <- runif(nrow(ub), 1, 10)
  ub$weight_alt <- runif(nrow(ub), 1, 10)

  expect_error(
    suppressMessages(suppressWarnings(fect::fect(
      turnout ~ policy_edr + policy_mail_in + policy_motor,
      data = ub, index = c("abb", "year"),
      method = "mc", se = FALSE, CV = FALSE,
      W.est = "weight", W.agg = "weight_alt",
      na.rm = TRUE, parallel = FALSE
    ))),
    "v2\\.4\\.0"
  )
})


test_that("unweighted fit (W = NULL): unchanged label, no *.W slots either", {
  skip_on_cran()

  suppressWarnings(try(data("turnout", package = "fect"), silent = TRUE))
  set.seed(123456)
  ub <- turnout[
    -c(which(turnout$abb == "WY")[1:15],
       sample(seq_len(nrow(turnout)), 50, replace = FALSE)),
  ]

  fit <- suppressMessages(suppressWarnings(fect::fect(
    turnout ~ policy_edr + policy_mail_in + policy_motor,
    data = ub, index = c("abb", "year"),
    method = "mc", CV = TRUE, k = 5, se = TRUE,
    nboots = 30, min.T0 = 5, seed = 42,
    na.rm = TRUE, parallel = FALSE
  )))

  out_lines <- capture.output(print(fit))
  expect_true(any(grepl("Tr obs equally weighted", out_lines, fixed = TRUE)))
  expect_false(any(grepl("sample-weighted", out_lines, fixed = TRUE)))

  expect_null(fit$est.att.W)
  expect_null(fit$att.on.W)
})

## ---------------------------------------------------------------
## Regression test: parallel = c("cv", "boot") must not error.
##
## Phase 1-3 updated the CV dispatch (cv.R, fect_nevertreated.R) to
## handle the vector form via the do_parallel_cv / do_parallel_boot
## flags in default.R. Several downstream paths (fect_boot, fit_test,
## permutation, fect_sens, did_wrapper) kept legacy `parallel == TRUE`
## or `if (parallel)` checks that error with
##   "the condition has length > 1"
## when the user passes a length > 1 character vector.
##
## Caught by an Alsaadi (2025) real-workload smoke test on 2026-04-22.
## ---------------------------------------------------------------

suppressWarnings(data("simdata", package = "fect"))

test_that("parallel = c('cv','boot') does not crash fect_boot", {

  skip_on_cran()

  set.seed(42)
  fit <- suppressWarnings(suppressMessages(
    fect::fect(
      Y ~ D,
      data      = simdata,
      index     = c("id", "time"),
      method    = "ife",
      r         = 0:1,
      CV        = TRUE,
      k         = 2,
      cv.method = "all_units",
      time.component.from = "nevertreated",
      se        = TRUE,
      vartype   = "bootstrap",
      nboots    = 5,
      parallel  = c("cv", "boot"),
      cores     = 2
    )
  ))

  expect_true(!is.null(fit$att.avg))
  expect_true(!is.null(fit$est.att))
})

## ---------------------------------------------------------
## Simulation test: Fixed Effects (FE) estimator
## DGP: Two-way FE model with known treatment effect tau=3.0
## ---------------------------------------------------------

test_that("FE estimator recovers ATT under simple two-way FE DGP", {

  skip_on_cran()

    set.seed(2024)

    N <- 30       # units
    TT <- 15      # time periods
    T0 <- 10      # treatment starts at period 11 for treated units
    Ntr <- 10     # number of treated units
    tau <- 3.0    # true treatment effect
    nsim <- 50    # number of Monte Carlo replications

    estimates <- numeric(nsim)

    for (sim in 1:nsim) {
        ## Unit and time fixed effects
        alpha_i <- rnorm(N, 0, 1)
        xi_t <- rnorm(TT, 0, 0.5)

        ## Generate panel data
        Y_vec <- numeric(N * TT)
        D_vec <- integer(N * TT)
        id_vec <- integer(N * TT)
        time_vec <- integer(N * TT)

        idx <- 1
        for (i in 1:N) {
            for (t in 1:TT) {
                treated <- (i <= Ntr) && (t > T0)
                D_vec[idx] <- as.integer(treated)
                eps <- rnorm(1, 0, 1)
                Y_vec[idx] <- alpha_i[i] + xi_t[t] + tau * D_vec[idx] + eps
                id_vec[idx] <- i
                time_vec[idx] <- t
                idx <- idx + 1
            }
        }

        simdf <- data.frame(
            id = id_vec,
            time = time_vec,
            Y = Y_vec,
            D = D_vec
        )

        out <- suppressWarnings(fect::fect(
            Y ~ D,
            data = simdf,
            index = c("id", "time"),
            method = "fe",
            force = "two-way",
            se = FALSE,
            parallel = FALSE
        ))

        estimates[sim] <- out$att.avg
    }

    ## Check bias: mean estimate should be close to true tau
    mean_est <- mean(estimates, na.rm = TRUE)
    bias <- abs(mean_est - tau)
    expect_lt(bias, 0.5,
        label = paste0("FE bias = ", round(bias, 4),
                       ", mean_est = ", round(mean_est, 4),
                       ", tau = ", tau))

    ## Check dispersion: SD should be reasonable
    sd_est <- sd(estimates, na.rm = TRUE)
    expect_lt(sd_est, 1.5,
        label = paste0("FE SD = ", round(sd_est, 4)))

    ## Check that estimates are centered around tau
    ## (median should also be near tau)
    median_est <- median(estimates, na.rm = TRUE)
    expect_lt(abs(median_est - tau), 0.6,
        label = paste0("FE median bias = ", round(abs(median_est - tau), 4)))
})

test_that("FE estimator produces near-zero pre-treatment effects", {

  skip_on_cran()

    set.seed(2025)

    N <- 30
    TT <- 15
    T0 <- 10
    Ntr <- 10
    tau <- 3.0

    alpha_i <- rnorm(N, 0, 1)
    xi_t <- rnorm(TT, 0, 0.5)

    Y_vec <- numeric(N * TT)
    D_vec <- integer(N * TT)
    id_vec <- integer(N * TT)
    time_vec <- integer(N * TT)

    idx <- 1
    for (i in 1:N) {
        for (t in 1:TT) {
            treated <- (i <= Ntr) && (t > T0)
            D_vec[idx] <- as.integer(treated)
            eps <- rnorm(1, 0, 1)
            Y_vec[idx] <- alpha_i[i] + xi_t[t] + tau * D_vec[idx] + eps
            id_vec[idx] <- i
            time_vec[idx] <- t
            idx <- idx + 1
        }
    }

    simdf <- data.frame(
        id = id_vec,
        time = time_vec,
        Y = Y_vec,
        D = D_vec
    )

    out <- suppressWarnings(fect::fect(
        Y ~ D,
        data = simdf,
        index = c("id", "time"),
        method = "fe",
        force = "two-way",
        se = FALSE,
        parallel = FALSE
    ))

    ## Pre-treatment dynamic effects should be near zero
    pre_idx <- which(out$time <= 0)
    if (length(pre_idx) > 0) {
        pre_effects <- out$att[pre_idx]
        pre_effects <- pre_effects[!is.na(pre_effects)]
        if (length(pre_effects) > 0) {
            mean_pre <- mean(abs(pre_effects))
            expect_lt(mean_pre, 1.5,
                label = paste0("Mean |pre-treatment effect| = ",
                               round(mean_pre, 4)))
        }
    }
})

## Tests for the group.fe argument (added v2.4.5; closes #139).
## Design memo: statsclaw-workspace/fect/runs/2026-05-21-higher-level-fe.md

## ----------------------------------------------------------------------
## Helper: small nested panel (county within state, state-level treatment).
## Returns a balanced TT*N panel with non-trivial state structure.
## ----------------------------------------------------------------------
.make_nested_panel <- function(N = 40, TT = 10, n_states = 4, seed = 42) {
    set.seed(seed)
    df <- expand.grid(id = 1:N, time = 1:TT)
    df$state <- paste0("S", (df$id - 1) %% n_states + 1)
    df$D <- as.integer(df$state %in% c("S1", "S2") & df$time >= 6)
    df$Y <- 1 + 0.5 * df$D + rnorm(nrow(df), sd = 0.5)
    df
}

## ----------------------------------------------------------------------
## Point-estimate equivalence
## ----------------------------------------------------------------------

test_that("group.fe = 'state' is byte-equivalent to legacy index[3]", {
    df <- .make_nested_panel()
    fit_new <- fect(Y ~ D, data = df, index = c("id", "time"),
                    group.fe = "state", force = "time", se = FALSE)
    fit_old <- fect(Y ~ D, data = df, index = c("id", "time", "state"),
                    method = "cfe", r = 0, CV = FALSE,
                    force = "time", se = FALSE)
    expect_equal(fit_new$att.avg, fit_old$att.avg)
})

test_that("auto-route from method='fe' preserves the result", {
    df <- .make_nested_panel()
    fit_default <- fect(Y ~ D, data = df, index = c("id", "time"),
                        group.fe = "state", force = "time", se = FALSE)
    fit_explicit <- fect(Y ~ D, data = df, index = c("id", "time"),
                         group.fe = "state", method = "cfe", r = 0,
                         CV = FALSE, force = "time", se = FALSE)
    expect_equal(fit_default$att.avg, fit_explicit$att.avg)
})

## ----------------------------------------------------------------------
## Hard errors on unsupported methods (D3)
## ----------------------------------------------------------------------

test_that("method='ife' + group.fe hard-errors with guidance", {
    df <- .make_nested_panel()
    expect_error(
        fect(Y ~ D, data = df, index = c("id", "time"),
             group.fe = "state", method = "ife", se = FALSE),
        "not supported"
    )
})

test_that("method='mc' + group.fe hard-errors", {
    df <- .make_nested_panel()
    expect_error(
        fect(Y ~ D, data = df, index = c("id", "time"),
             group.fe = "state", method = "mc", se = FALSE),
        "not supported"
    )
})

test_that("method='both' + group.fe hard-errors", {
    df <- .make_nested_panel()
    expect_error(
        fect(Y ~ D, data = df, index = c("id", "time"),
             group.fe = "state", method = "both", se = FALSE),
        "not supported"
    )
})

test_that("method='gsynth' + group.fe hard-errors", {
    df <- .make_nested_panel()
    expect_error(
        fect(Y ~ D, data = df, index = c("id", "time"),
             group.fe = "state", method = "gsynth", se = FALSE),
        "not supported"
    )
})

## ----------------------------------------------------------------------
## Edge cases (D5)
## ----------------------------------------------------------------------

test_that("group.fe column missing from data hard-errors", {
    df <- .make_nested_panel()
    expect_error(
        fect(Y ~ D, data = df, index = c("id", "time"),
             group.fe = "nonexistent_col", se = FALSE),
        "not found"
    )
})

test_that("group.fe AND legacy index[3:] both used hard-errors", {
    df <- .make_nested_panel()
    expect_error(
        fect(Y ~ D, data = df, index = c("id", "time", "state"),
             group.fe = "state", se = FALSE),
        "OR extra index slots"
    )
})

test_that("non-character group.fe hard-errors", {
    df <- .make_nested_panel()
    expect_error(
        fect(Y ~ D, data = df, index = c("id", "time"),
             group.fe = 1L, se = FALSE),
        "character vector"
    )
})

test_that("group.fe overlapping index[1:2] is warned and dropped", {
    df <- .make_nested_panel()
    expect_warning(
        fect(Y ~ D, data = df, index = c("id", "time"),
             group.fe = c("id", "state"), force = "time", se = FALSE),
        "duplicate index"
    )
})

## ----------------------------------------------------------------------
## Nesting check (D5 edge 5, D3f) -- applies to BOTH group.fe and legacy
## ----------------------------------------------------------------------

test_that("non-nested group.fe hard-errors with offending units listed", {
    df <- .make_nested_panel()
    df$state[1:5] <- "BAD"  # county 1 now appears in two states across time
    expect_error(
        fect(Y ~ D, data = df, index = c("id", "time"),
             group.fe = "state", force = "time", se = FALSE),
        "not constant within"
    )
})

test_that("legacy index[3:] does NOT enforce nesting (supports cell-level interactions)", {
    ## Legacy index = c(unit, time, extra) syntax has historically supported
    ## both nested groupings AND cell-level interactions like region_time.
    ## The nesting check only fires for group.fe; legacy form keeps full scope.
    df <- .make_nested_panel()
    df$region_time <- as.numeric(df$state == "S1") + df$time / 100  # varies within id
    expect_no_error(
        fect(Y ~ D, data = df, index = c("id", "time", "region_time"),
             method = "cfe", force = "two-way", se = FALSE)
    )
})

## ----------------------------------------------------------------------
## cl auto-default + FALSE sentinel (D6)
## ----------------------------------------------------------------------

test_that("single-column group.fe auto-defaults cl to group.fe[1]", {
    df <- .make_nested_panel()
    fit <- fect(Y ~ D, data = df, index = c("id", "time"),
                group.fe = "state", force = "time", se = FALSE)
    expect_equal(fit$cl.label, "state")
})

test_that("cl = FALSE is rejected with a guiding error", {
    df <- .make_nested_panel()
    expect_error(
        fect(Y ~ D, data = df, index = c("id", "time"),
             group.fe = "state", force = "time", cl = FALSE, se = FALSE),
        "cl = FALSE is not supported"
    )
})

test_that("cl = index[1] explicitly clusters at the unit level", {
    df <- .make_nested_panel()
    fit <- fect(Y ~ D, data = df, index = c("id", "time"),
                group.fe = "state", force = "time", cl = "id", se = FALSE)
    expect_equal(fit$cl.label, "id")
})

test_that("cl = 'other_col' overrides the auto-default", {
    df <- .make_nested_panel()
    df$region <- substr(df$state, 1, 1)
    fit <- fect(Y ~ D, data = df, index = c("id", "time"),
                group.fe = "state", force = "time", cl = "region", se = FALSE)
    expect_equal(fit$cl.label, "region")
})

test_that("multi-column group.fe requires explicit cl", {
    df <- .make_nested_panel()
    df$region <- substr(df$state, 1, 1)
    expect_error(
        fect(Y ~ D, data = df, index = c("id", "time"),
             group.fe = c("state", "region"), force = "time", se = FALSE),
        "Multi-column group.fe requires explicit cl"
    )
})

## ----------------------------------------------------------------------
## Fit slots for print (D8)
## ----------------------------------------------------------------------

test_that("fit$group.fe and fit$cl.label are populated for downstream print", {
    df <- .make_nested_panel()
    fit <- fect(Y ~ D, data = df, index = c("id", "time"),
                group.fe = "state", force = "time", se = FALSE)
    expect_equal(fit$group.fe, "state")
    expect_equal(fit$cl.label, "state")
})

test_that("print(fit) surfaces Estimator + Fixed effects + Cluster SE", {
    df <- .make_nested_panel()
    fit <- fect(Y ~ D, data = df, index = c("id", "time"),
                group.fe = "state", force = "time", se = FALSE)
    out <- capture.output(print(fit))
    expect_true(any(grepl("^Estimator:", out)))
    expect_true(any(grepl("^Fixed effects:.*state", out)))
    expect_true(any(grepl("^Cluster SE:.*state", out)))
})

test_that("print(fit) shows the user's chosen cl in the Cluster SE line", {
    df <- .make_nested_panel()
    fit <- fect(Y ~ D, data = df, index = c("id", "time"),
                group.fe = "state", force = "time", cl = "id", se = FALSE)
    out <- capture.output(print(fit))
    expect_true(any(grepl("^Cluster SE:.*id", out)))
})

fect_nevertreated <- function(Y, # Outcome variable, (T*N) matrix
                        X, # Explanatory variables:  (T*N*p) array
                        D, #  Indicator for treated unit (tr==1)
                        W,
                        I,
                        II,
                        cm = FALSE,
                        II.cm = NULL,
                        T.on,
                        T.off = NULL,
                        T.on.carry = NULL,
                        T.on.balance = NULL,
                        balance.period = NULL,
                        CV = TRUE,
                        criterion = "mspe",
                        cv.method = "rolling",
                        k = 20,
                        cv.prop = 0.1,
                        cv.nobs = 3,
                        cv.donut = 1,
                        cv.buffer = 1,
                        min.T0 = 5,
                        r = 0, # r.end when CV==TRUE
                        r.end = 3,
                        binary = FALSE,
                        QR = FALSE,
                        force,
                        hasRevs = 1,
                        tol, # tolerance level
                        max.iteration = 1000,
                        boot = FALSE, # bootstrapped sample
                        placeboTest = 0,
                        placebo.period = NULL,
                        carryoverTest = 0,
                        carryover.period = NULL,
                        norm.para = NULL,
                        calendar.enp.seq = NULL,
                        time.on.seq = NULL,
                        time.off.seq = NULL,
                        time.on.balance.seq = NULL,
                        time.on.seq.W = NULL,
                        time.off.seq.W = NULL,
                        group.level = NULL,
                        group = NULL,
                        time.on.seq.group = NULL,
                        time.off.seq.group = NULL,
                        ## CFE-specific parameters (new)
                        method = "ife",           ## "ife" (existing) or "cfe" (new)
                        X.extra.FE = NULL,        ## TT x N x n_extra array
                        X.Z = NULL,               ## TT x N x n_z array
                        X.Q = NULL,               ## TT x N x n_q array
                        X.gamma = NULL,           ## TT x N x n_gamma array
                        X.kappa = NULL,           ## TT x N x n_kappa array
                        Zgamma.id = NULL,         ## list mapping gamma groups to Z columns
                        kappaQ.id = NULL,         ## list mapping kappa groups to Q columns
                        parallel = TRUE,
                        cores = NULL,
                        do_parallel_cv = NULL,  ## pre-computed flag from default.R; NULL means derive from parallel
                        loading.bound = "none",
                        gamma.loading = NULL,
                        gamma.loading.grid = NULL,
                        cv.rule = "1se",
                        W.in.fit = TRUE,
                        fit.init = NULL ## warm-start aux surface (T x N_boot); v2.4.3+
                        ) {
    ## -------------------------------##
    ## Parsing data
    ## -------------------------------##
    cv.rule <- .fect_validate_cv_rule(cv.rule)
    carryover.pos <- placebo.pos <- na.pos <- NULL
    res.sd1 <- res.sd2 <- NULL
    ## unit id and time
    TT <- dim(Y)[1]
    N <- dim(Y)[2]
    id <- 1:N
    time <- 1:TT
    if (is.null(X) == FALSE) {
        p <- dim(X)[3]
    } else {
        p <- 0
        X <- array(0, dim = c(1, 1, 0))
    }

    ## replicate data
    YY <- Y
    YY[which(II == 0)] <- 0 ## reset to 0

    ## once treated, always treated
    ## careful unbalanced case: calculate treated units
    ## treat reversals as always treated
    D <- apply(D, 2, function(vec) {
        cumsum(vec)
    })
    D <- ifelse(D > 0, 1, 0)
    D.sum <- colSums(D)
    tr <- which(D.sum >= 1)
    Ntr <- length(tr)
    co <- which(D.sum == 0)
    Nco <- length(co)
    r <- min(r, TT, Nco)

    ## Slice warm-start aux surface to control units (the only units fed
    ## to complex_fe_ub / inter_fe_ub here). NULL preserves cold-start.
    fit.init.co <- NULL
    if (!is.null(fit.init)) {
        if (is.matrix(fit.init) && nrow(fit.init) == TT && ncol(fit.init) >= max(co)) {
            fit.init.co <- fit.init[, co, drop = FALSE]
        }
    }

    I.tr <- as.matrix(I[, tr]) ## maybe only 1 treated unit
    I.co <- I[, co]
    II.tr <- as.matrix(II[, tr])
    II.co <- II[, co]
    Y.tr <- as.matrix(Y[, tr])
    Y.co <- as.matrix(Y[, co])
    YY.tr <- as.matrix(YY[, tr])
    YY.co <- as.matrix(YY[, co])
    Ntr <- length(tr)
    Nco <- length(co)
    if (p == 0) {
        X.tr <- array(0, dim = c(TT, Ntr, 0))
        X.co <- array(0, dim = c(TT, Nco, 0))
    } else {
        X.tr <- array(NA, dim = c(TT, Ntr, p))
        X.co <- array(NA, dim = c(TT, Nco, p))
        for (j in 1:p) {
            X.tr[, , j] <- X[, tr, j]
            X.co[, , j] <- X[, co, j]
        }
    }

    ## ---- CFE array subsetting (co/tr split) ----
    if (method == "cfe") {
        ## helper to split a TT x N x k array into co/tr slices
        .split_array <- function(arr, co_idx, tr_idx, TT, Nco, Ntr) {
            if (!is.null(arr) && length(dim(arr)) == 3 && dim(arr)[3] > 0) {
                list(co = arr[, co_idx, , drop = FALSE],
                     tr = arr[, tr_idx, , drop = FALSE])
            } else {
                list(co = array(0, dim = c(TT, Nco, 0)),
                     tr = array(0, dim = c(TT, Ntr, 0)))
            }
        }
        efe.split  <- .split_array(X.extra.FE, co, tr, TT, Nco, Ntr)
        X.extra.FE.co <- efe.split$co;  X.extra.FE.tr <- efe.split$tr
        xz.split   <- .split_array(X.Z, co, tr, TT, Nco, Ntr)
        X.Z.co <- xz.split$co;  X.Z.tr <- xz.split$tr
        xq.split   <- .split_array(X.Q, co, tr, TT, Nco, Ntr)
        X.Q.co <- xq.split$co;  X.Q.tr <- xq.split$tr
        xg.split   <- .split_array(X.gamma, co, tr, TT, Nco, Ntr)
        X.gamma.co <- xg.split$co;  X.gamma.tr <- xg.split$tr
        xk.split   <- .split_array(X.kappa, co, tr, TT, Nco, Ntr)
        X.kappa.co <- xk.split$co;  X.kappa.tr <- xk.split$tr

        ## ---- Extra FE Type A/B classification ----
        n_extra <- dim(X.extra.FE.co)[3]
        fe_type <- character(0)
        typeA_idx <- integer(0)
        typeB_idx <- integer(0)

        if (n_extra > 0) {
            fe_type <- character(n_extra)
            for (k in 1:n_extra) {
                co_levels <- unique(X.extra.FE[1, co, k])
                tr_levels <- unique(X.extra.FE[1, tr, k])
                overlap <- intersect(co_levels, tr_levels)
                if (length(overlap) == 0) {
                    fe_type[k] <- "A"
                } else {
                    fe_type[k] <- "B"
                    missing <- setdiff(tr_levels, co_levels)
                    if (length(missing) > 0) {
                        stop(paste0("Extra fixed effect dimension ", k,
                             " has levels in treated units not found in controls: ",
                             paste(missing, collapse = ", "),
                             ". Cannot estimate from controls."))
                    }
                }
            }
            typeA_idx <- which(fe_type == "A")
            typeB_idx <- which(fe_type == "B")
        }

        ## Build co-only Type-B FE array for complex_fe_ub
        if (length(typeB_idx) > 0) {
            X.extra.FE.co.B <- X.extra.FE.co[, , typeB_idx, drop = FALSE]
        } else {
            X.extra.FE.co.B <- array(0, dim = c(TT, Nco, 0))
        }
    }

    if (is.null(W) || !W.in.fit) {
        W.use <- as.matrix(0)
    } else {
        W.use <- as.matrix(W[, co, drop = FALSE])
        W.use[which(II.co == 0)] <- 0
    }

    ## ---- cv.method validation ---- ##
    cv.method <- .fect_normalize_cv_method(
        cv.method,
        allowed = c("rolling", "block", "all_units", "treated_units", "loo")
    )

    ## ---- W for treated units (scoring) ---- ##
    if (!is.null(W)) {
        W.tr <- as.matrix(W[, tr, drop = FALSE])
    } else {
        W.tr <- NULL
    }

    if (!0 %in% I.tr) {
        ## a (TT*Ntr) matrix, time dimension: before treatment
        pre <- as.matrix(D[, tr] == 0 & II[, tr] == 1)
    } else {
        pre <- as.matrix(D[, tr] == 0 & I[, tr] == 1 & II[, tr] == 1)
    }

    T0 <- apply(pre, 2, sum)
    T0.min <- min(T0)
    pre.v <- as.vector(pre) ## vectorized "pre-treatment" indicator
    id.tr.pre.v <- rep(id, each = TT)[which(pre.v == 1)] ## vectorized pre-treatment grouping variable for the treated
    time.pre <- split(rep(time, Ntr)[which(pre.v == 1)], id.tr.pre.v) ## a list of pre-treatment periods
    sameT0 <- length(unique(T0)) == 1

    ## ---- count.T.cv construction ---- ##
    count.T.cv <- NULL
    if (!is.null(T.on)) {
        t.on.full <- c(T.on)
        count.T.cv <- table(t.on.full)
        count.T.cv <- count.T.cv[which(as.numeric(names(count.T.cv)) <= 0)]
        if (length(count.T.cv) > 0) {
            count.T.cv <- count.T.cv / mean(count.T.cv)
            nm <- names(count.T.cv)
            count.T.cv <- c(count.T.cv, median(count.T.cv))
            names(count.T.cv) <- c(nm, "Control")
        }
    }

    ## ====================================================================
    ## Solver dispatch wrapper
    ## ====================================================================
    ## Dispatches to inter_fe (balanced) / inter_fe_ub (unbalanced) for IFE,
    ## or complex_fe_ub for CFE. Eliminates balanced/unbalanced branching
    ## at each call site.
    .estimate_co <- function(Y, Y0, X, I_obs, W_in, beta0_in, r, force,
                             tol, max.iteration, use_cfe = FALSE,
                             center = TRUE, fit_init = NULL) {
        ## Center Y to improve convergence conditioning:
        ## removes grand mean from fit so tol applies to variation, not level.
        mu_init <- 0
        if (center && 0 %in% I_obs) {
            ## Only center for unbalanced panels (where EM iterates)
            mu_init <- sum(Y * I_obs) / sum(I_obs)
            Y  <- Y - mu_init * I_obs   ## observed positions centered, zeros stay
            Y0 <- Y0 - mu_init          ## initial fit centered
            ## Centering the warm-start surface keeps it on the same scale
            ## as the centered Y; otherwise the EM would unwind centering on
            ## the very first iteration. Only apply when fit_init is non-null.
            if (!is.null(fit_init)) fit_init <- fit_init - mu_init
        }

        if (use_cfe) {
            out <- complex_fe_ub(Y, Y0, X,
                X.extra.FE.co.B, X.Z.co, X.Q.co, X.gamma.co, X.kappa.co,
                Zgamma.id, kappaQ.id,
                I_obs, W_in, beta0_in, r, force = force, tol, max.iteration,
                fit_init = fit_init)
        } else if (!0 %in% I_obs) {
            ## Balanced inter_fe does not yet expose fit_init (would require
            ## a separate C++ extension). Cold-start here is the safe default.
            out <- inter_fe(Y, X, r, force = force, beta0_in = beta0_in, tol, max.iteration)
        } else {
            out <- inter_fe_ub(Y, Y0, X, I_obs, W_in, beta0_in, r,
                        force = force, tol, max.iteration,
                        fit_init = fit_init)
        }

        ## Undo centering
        if (mu_init != 0) {
            out$mu  <- out$mu + mu_init
            out$fit <- out$fit + mu_init
        }
        out
    }

  if (method == "ife") {
    ## ====================================================================
    ## IFE PATH
    ## ====================================================================

    beta0 <- matrix(0, p, 1)
    # initial fit using Y.co
    if (0 %in% I.co) {
        data.ini <- matrix(NA, Nco * TT, (p + 3))
        data.ini[, 1] <- c(Y.co)
        data.ini[, 2] <- rep(1:Nco, each = TT)
        data.ini[, 3] <- rep(1:TT, Nco)
        if (p > 0) {
            for (i in 1:p) {
                data.ini[, (3 + i)] <- c(X.co[, , i])
            }
        }
        ## observed Y0 indicator:
        initialOut <- Y0.co <- beta0 <- FE0 <- xi0 <- factor0 <- NULL
        oci <- which(c(II.co) == 1)
        if (is.null(W) || !W.in.fit) {
            initialOut <- initialFit(data = data.ini, force = force, oci = oci)
        } else {
            initialOut <- initialFit(data = data.ini, force = force, w = c(W.use), oci = oci)
        }
        Y0.co <- initialOut$Y0
        beta0 <- initialOut$beta0
        if (p > 0 && sum(is.na(beta0)) > 0) {
            beta0[which(is.na(beta0))] <- 0
        }
    }


    validX <- 1 ## no multi-colinearity

    ## cross-validation only for gsynth
    if (CV == TRUE) {
        ## Two-tier tolerance: CV uses looser tol for r-selection speed,
        ## final estimation (after CV) uses the user's tol for precision.
        cv_tol <- max(tol, 1e-3)

        ## starting r
        if ((r > (T0.min - 1) & force %in% c(0, 2)) | (r > (T0.min - 2) & force %in% c(1, 3))) {
            message("r is too big compared with T0; reset to 0.")
            r <- 0
        }

        ## store all MSPE
        if (force %in% c(0, 2)) {
            r.max <- max(min((T0.min - 1), r.end), 0)
        } else {
            r.max <- max(min((T0.min - 2), r.end), 0)
        }

        if (r.max == 0) {
            r.cv <- 0
            message("Cross validation cannot be performed since available pre-treatment records of treated units are too few. So set r.cv = 0.")
            est.co.best <- .estimate_co(YY.co, Y0.co, X.co, I.co, W.use, beta0, 0, force, cv_tol, max.iteration,
                                        fit_init = fit.init.co)
        } else {
            r.old <- r ## save the minimal number of factors

            message("Cross-validating ...", "\r")
            score_names <- c("MSPE", "WMSPE", "GMSPE", "WGMSPE",
                             "MAD", "Moment", "GMoment", "RMSE", "Bias")
            CV.out <- matrix(NA, (r.max - r.old + 1), 4 + length(score_names))
            colnames(CV.out) <- c("r", "sigma2", "IC", "PC", score_names)
            CV.out[, "r"] <- c(r.old:r.max)
            CV.out[, score_names] <- 1e10
            CV.out[, "PC"] <- 1e10
            r.pc <- est.co.pc.best <- NULL

            ## Per-fold SE matrix parallel to CV.out (added v2.3.0). Populated
            ## below in both parallel and serial CV branches; consumed at the
            ## end of the IFE CV block to apply `cv.rule` (default "1se").
            CV.out.se <- matrix(NA_real_, nrow(CV.out), ncol(CV.out))
            colnames(CV.out.se) <- colnames(CV.out)
            CV.out.se[, "r"] <- CV.out[, "r"]

            crit_col <- switch(criterion,
                mspe = "MSPE", wmspe = "WMSPE", gmspe = "GMSPE", wgmspe = "WGMSPE",
                mad = "MAD", moment = "Moment", gmoment = "GMoment", "MSPE")

            ## ---- cv.sample pre-computation (IFE) ---- ##
            if (cv.method != "loo" && r.max > 0) {
                if (cv.method %in% c("all_units", "rolling")) {
                    rm.count.co <- floor(sum(II.co) * cv.prop)
                    if (rm.count.co == 0 && cv.method == "all_units") {
                        message("cv.prop too small for control panel; falling back to LOO.")
                        cv.method <- "loo"
                    } else {
                        D.co.fake <- matrix(0, TT, Nco)
                        oci.co <- which(c(II.co) == 1)
                        ## Ensure data.ini exists (balanced IFE case skips its creation)
                        if (!exists("data.ini", inherits = FALSE) || is.null(data.ini)) {
                            data.ini <- matrix(NA, Nco * TT, (p + 3))
                            data.ini[, 1] <- c(Y.co)
                            data.ini[, 2] <- rep(1:Nco, each = TT)
                            data.ini[, 3] <- rep(1:TT, Nco)
                            if (p > 0) {
                                for (i.tmp in 1:p) {
                                    data.ini[, (3 + i.tmp)] <- c(X.co[, , i.tmp])
                                }
                            }
                        }
                        rmCV <- list()
                        ociCV <- list()
                        estCV <- list()
                        Y0CV.co <- array(NA, dim = c(TT, Nco, k))
                        if (p > 0) {
                            beta0CV.co <- array(NA, dim = c(p, 1, k))
                        } else {
                            beta0CV.co <- array(0, dim = c(1, 0, k))
                        }
                        flag.cv <- 0

                        ## ---- rolling-window pre-computation ---- ##
                        ## When rolling, masks come from .build_cv_mask_rolling
                        ## (per-fold sampling of cv.prop of eligible control
                        ## units; per-unit random anchor + cv.nobs scored
                        ## holdout + cv.buffer past-side buffer + drop-future
                        ## as the rolling-window step). Skip the con1/con2
                        ## block-CV feasibility checks --- rolling preserves
                        ## per-time donor coverage by construction via
                        ## per-fold unit sampling.
                        rolling_folds <- NULL
                        if (cv.method == "rolling") {
                            rolling_folds <- .build_cv_mask_rolling(
                                II = II.co, D = D.co.fake, k = k,
                                cv.nobs = cv.nobs, cv.buffer = cv.buffer,
                                cv.prop = cv.prop, min.T0 = min.T0, seed = NULL
                            )
                        }

                        for (i.cv in 1:k) {
                            if (cv.method == "rolling") {
                                cv.n <- 0
                                cv.id  <- rolling_folds[[i.cv]]$cv.id
                                est.id <- rolling_folds[[i.cv]]$est.id
                            } else {
                                cv.n <- 0
                                repeat {
                                    cv.n <- cv.n + 1
                                    get.cv <- cv.sample(II.co, D.co.fake,
                                        count = rm.count.co,
                                        cv.count = cv.nobs,
                                        cv.treat = FALSE,
                                        cv.donut = cv.donut)
                                    cv.id <- get.cv$cv.id
                                    II.co.cv <- II.co
                                    II.co.cv[cv.id] <- 0
                                    II.co.cv.valid <- II.co
                                    II.co.cv.valid[cv.id] <- -1
                                    con1 <- sum(apply(II.co.cv, 1, sum) >= 1) == TT
                                    con2 <- sum(apply(II.co.cv, 2, sum) >= min.T0) == Nco
                                    if (con1 && con2) break
                                    if (cv.n >= 200) {
                                        flag.cv <- 1
                                        keep.1 <- which(apply(II.co.cv, 1, sum) < 1)
                                        keep.2 <- which(apply(II.co.cv, 2, sum) < min.T0)
                                        II.co.cv[keep.1, ] <- II.co[keep.1, ]
                                        II.co.cv[, keep.2] <- II.co[, keep.2]
                                        II.co.cv.valid[keep.1, ] <- II.co[keep.1, ]
                                        II.co.cv.valid[, keep.2] <- II.co[, keep.2]
                                        cv.id <- which(II.co.cv.valid != II.co)
                                        break
                                    }
                                }
                            }
                            rmCV[[i.cv]] <- cv.id
                            ociCV[[i.cv]] <- setdiff(oci.co, cv.id)
                            if (cv.method == "rolling") {
                                estCV[[i.cv]] <- est.id
                            } else if (cv.n < 200) {
                                estCV[[i.cv]] <- get.cv$est.id
                            } else {
                                cv.diff <- setdiff(get.cv$cv.id, cv.id)
                                estCV[[i.cv]] <- setdiff(get.cv$est.id, cv.diff)
                            }
                            if (is.null(W) || !W.in.fit) {
                                initialOutCv <- initialFit(data = data.ini, force = force, oci = ociCV[[i.cv]])
                            } else {
                                initialOutCv <- initialFit(data = data.ini, force = force, w = c(W.use), oci = ociCV[[i.cv]])
                            }
                            Y0CV.co[, , i.cv] <- initialOutCv$Y0
                            if (p > 0) {
                                beta0cv <- initialOutCv$beta0
                                if (sum(is.na(beta0cv)) > 0) {
                                    beta0cv[which(is.na(beta0cv))] <- 0
                                }
                                beta0CV.co[, , i.cv] <- beta0cv
                            }
                        }
                        if (flag.cv == 1) {
                            message("Some control units have too few observations. Removed automatically in CV.\n")
                        }
                    }
                } else if (cv.method == "treated_units") {
                    rm.count.tr <- floor(sum(pre) * cv.prop)
                    if (rm.count.tr == 0) {
                        message("cv.prop too small for treated pre-treatment panel; falling back to LOO.")
                        cv.method <- "loo"
                    } else {
                        D.tr.fake <- matrix(0, TT, Ntr)
                        rmCV.tr <- list()
                        estCV.tr <- list()
                        flag.cv <- 0
                        for (i.cv in 1:k) {
                            cv.n <- 0
                            repeat {
                                cv.n <- cv.n + 1
                                get.cv <- cv.sample(pre, D.tr.fake,
                                    count = rm.count.tr,
                                    cv.count = cv.nobs,
                                    cv.treat = FALSE,
                                    cv.donut = cv.donut)
                                cv.id <- get.cv$cv.id
                                pre.cv <- pre
                                pre.cv[cv.id] <- 0
                                con1 <- TRUE
                                pre.rows <- which(rowSums(pre) > 0)
                                if (length(pre.rows) > 0) {
                                    con1 <- all(rowSums(pre.cv[pre.rows, , drop = FALSE]) >= 1)
                                }
                                con2 <- all(colSums(pre.cv) >= min(min.T0, 2))
                                if (con1 && con2) break
                                if (cv.n >= 200) {
                                    flag.cv <- 1
                                    pre.cv.valid <- pre
                                    pre.cv.valid[cv.id] <- -1
                                    keep.1 <- pre.rows[rowSums(pre.cv[pre.rows, , drop = FALSE]) < 1]
                                    keep.2 <- which(colSums(pre.cv) < min(min.T0, 2))
                                    if (length(keep.1) > 0) {
                                        pre.cv[keep.1, ] <- pre[keep.1, ]
                                        pre.cv.valid[keep.1, ] <- pre[keep.1, ]
                                    }
                                    if (length(keep.2) > 0) {
                                        pre.cv[, keep.2] <- pre[, keep.2]
                                        pre.cv.valid[, keep.2] <- pre[, keep.2]
                                    }
                                    cv.id <- which(pre.cv.valid != pre)
                                    break
                                }
                            }
                            rmCV.tr[[i.cv]] <- cv.id
                            if (cv.n < 200) {
                                estCV.tr[[i.cv]] <- get.cv$est.id
                            } else {
                                cv.diff <- setdiff(get.cv$cv.id, cv.id)
                                estCV.tr[[i.cv]] <- setdiff(get.cv$est.id, cv.diff)
                            }
                        }
                        if (flag.cv == 1) {
                            message("Some treated units have too few pre-treatment observations. Removed automatically in CV.\n")
                        }
                    }
                }
            }

            ## ---- Parallel backend setup (IFE CV) — Phase 3 flat dispatch ---- ##
            ## do_parallel_cv: pre-computed flag from default.R/cv.R (NULL means derive from parallel).
            ## parallel=TRUE (default from fect): auto-enable for all_units
            ##   when control panel is large enough (Nco*TT > .CV_PARALLEL_THRESH$ife)
            ## parallel=FALSE: force sequential (user override)
            ## parallel="cv": force parallel regardless of threshold
            if (is.null(do_parallel_cv)) {
                ## backwards-compat: derive from parallel if not pre-computed
                do_parallel_cv <- isTRUE(parallel) || "cv" %in% as.character(parallel)
            }
            if (!do_parallel_cv) {
                cv_ife_parallel <- FALSE
            } else {
                ## Explicit "cv" override: bypass threshold; auto mode: threshold gates
                use_explicit_cv_ife <- "cv" %in% as.character(parallel) && !isTRUE(parallel)
                ## Centralized threshold gate — replaces bespoke Nco * TT > 20000 check
                ife_threshold_met <- (Nco * TT) > .CV_PARALLEL_THRESH$ife
                cv_ife_parallel <- (cv.method %in% c("all_units", "rolling")) &&
                                   (ife_threshold_met || use_explicit_cv_ife) &&
                                   (k > 1)
            }
            if (cv_ife_parallel) {
                if (is.null(cores)) {
                    cores <- max(1L, min(parallelly::availableCores(omit = 2L), 8L))
                }
                old.future.plan.ife <- future::plan()
                on.exit(future::plan(old.future.plan.ife), add = TRUE, after = FALSE)
                future::plan(future::cluster, workers = .fect_make_future_cluster(cores))
                ## doFuture::registerDoFuture() removed — not needed for future_lapply dispatch
                avail <- parallelly::availableCores()
                msg_line <- sprintf("Parallel CV: using %d of %d available cores.", cores, avail)
                pad <- strrep(" ", max(0, 56 - nchar(msg_line)))
                message("\n",
                    " +----------------------------------------------------------+\n",
                    " | ", msg_line, pad, " |\n",
                    " |                                                          |\n",
                    " | To change: set cores = <n> in fect().                    |\n",
                    " | Default: min(available - 2, 8).                          |\n",
                    " +----------------------------------------------------------+\n")
            }

            ## ---- Task list construction (IFE flat r×k dispatch) ---- ##
            if (cv_ife_parallel) {
                r_seq_ife <- CV.out[, "r"]
                tasks_ife <- vector("list", length(r_seq_ife) * k)
                idx <- 1L
                for (ri in seq_along(r_seq_ife)) {
                    for (ii in 1:k) {
                        tasks_ife[[idx]] <- list(r = r_seq_ife[ri], ii = ii, ri = ri)
                        idx <- idx + 1L
                    }
                }
                ## Capture helper in closure for worker serialization
                .score_fn_ife_all <- .fect_cv_score_one_ife_nt_all
            }

            if (cv_ife_parallel) {
            ## ---- PARALLEL BRANCH: flat r×k future_lapply dispatch (IFE all_units) ---- ##

                ## Step 1: dispatch all (r, fold) scoring tasks
                fold_scores_ife <- future.apply::future_lapply(
                    tasks_ife,
                    FUN = function(task) {
                        .score_fn_ife_all(
                            ii            = task$ii,
                            YY.co         = YY.co,
                            Y0CV.co       = Y0CV.co,
                            X.co          = X.co,
                            II.co         = II.co,
                            W.use         = W.use,
                            W             = W,
                            beta0CV.co    = beta0CV.co,
                            rmCV          = rmCV,
                            estCV         = estCV,
                            r             = task$r,
                            force         = force,
                            cv_tol        = cv_tol,
                            max.iteration = max.iteration
                        )
                    },
                    future.seed     = TRUE,
                    future.packages = "fect"
                )

                ## Step 2: sequential master walk — apply 1% rule in rank order
                n_r_ife <- length(r_seq_ife)
                for (i in seq_len(n_r_ife)) {
                    r <- unname(r_seq_ife[i])
                    ## Full-data fit (sequential in master) — needed for sigma2/IC/PC
                    est.co <- .estimate_co(YY.co, Y0.co, X.co, I.co, W.use, beta0, r, force, cv_tol, max.iteration)

                    if (p > 0) {
                        na.pos <- is.nan(est.co$beta)
                        beta <- est.co$beta
                        beta[is.nan(est.co$beta)] <- 0
                    }
                    if (is.null(norm.para)) {
                        sigma2 <- est.co$sigma2; IC <- est.co$IC; PC <- est.co$PC
                    } else {
                        sigma2 <- est.co$sigma2 * (norm.para[1]^2)
                        IC <- est.co$IC - log(est.co$sigma2) + log(sigma2)
                        PC <- est.co$PC * (norm.para[1]^2)
                    }

                    ## Aggregate fold scores for this rank
                    task_idx <- which(vapply(tasks_ife, function(t) t$ri == i, logical(1)))
                    all_resid    <- unlist(lapply(fold_scores_ife[task_idx], `[[`, "resid"))
                    all_time_idx <- unlist(lapply(fold_scores_ife[task_idx], `[[`, "time_idx"))
                    all_obs_w    <- if (!is.null(W)) unlist(lapply(fold_scores_ife[task_idx], `[[`, "obs_w")) else c()

                    if (length(all_resid) == 0) {
                        scores <- c(MSPE = Inf, WMSPE = Inf, GMSPE = Inf, WGMSPE = Inf,
                                    MAD = Inf, Moment = Inf, GMoment = Inf, RMSE = Inf, Bias = Inf)
                        se_v <- setNames(rep(NA_real_, length(scores)), names(scores))
                    } else {
                        agg <- .fect_cv_aggregate_folds(
                            fold_list  = fold_scores_ife[task_idx],
                            count.T.cv = count.T.cv,
                            use_weight = as.integer(!is.null(W)),
                            norm.para  = NULL
                        )
                        scores <- agg$pooled
                        se_v   <- agg$se
                    }

                    ## 1% rule — identical logic to serial path
                    if ((min(CV.out[, crit_col]) - scores[crit_col]) > 0.01 * min(CV.out[, crit_col])) {
                        est.co.best <- est.co
                        r.cv <- r
                    } else {
                        if (r == r.cv + 1) message("*")
                    }
                    if (PC < min(CV.out[, "PC"])) {
                        r.pc <- r
                        est.co.pc.best <- est.co
                    }
                    CV.out[i, 2:4] <- c(sigma2, IC, PC)
                    CV.out[i, score_names] <- scores[score_names]
                    ## Persist per-fold SEs for end-of-loop cv.rule application
                    for (cn in score_names) {
                        if (cn %in% names(se_v)) CV.out.se[i, cn] <- se_v[cn]
                    }
                    message("r = ", r, "; sigma2 = ",
                        sprintf("%.5f", sigma2), "; IC = ",
                        sprintf("%.5f", IC), "; PC = ",
                        sprintf("%.5f", PC), "; MSPE = ",
                        sprintf("%.5f", scores["MSPE"]), sep = "")
                } ## end per-r master walk (IFE parallel)

            } else {
            ## ---- SERIAL BRANCH (existing r-loop, all cv.method values) ---- ##

            for (i in 1:dim(CV.out)[1]) {
                r <- unname(CV.out[i, "r"])
                est.co <- .estimate_co(YY.co, Y0.co, X.co, I.co, W.use, beta0, r, force, cv_tol, max.iteration)

                if (p > 0) {
                    na.pos <- is.nan(est.co$beta)
                    beta <- est.co$beta
                    beta[is.nan(est.co$beta)] <- 0
                }

                if (is.null(norm.para)) {
                    sigma2 <- est.co$sigma2
                    IC <- est.co$IC
                    PC <- est.co$PC
                } else {
                    sigma2 <- est.co$sigma2 * (norm.para[1]^2)
                    IC <- est.co$IC - log(est.co$sigma2) + log(sigma2)
                    PC <- est.co$PC * (norm.para[1]^2)
                }

              if (cv.method == "loo") {
                ## ---- LOO CV (existing code) ---- ##
                if (r != 0) {
                    F.hat <- as.matrix(est.co$factor)
                    if (force %in% c(1, 3)) {
                        F.hat <- cbind(F.hat, rep(1, TT))
                    }
                }
                U.tr <- Y.tr
                if (p > 0) {
                    for (j in 1:p) {
                        U.tr <- U.tr - X.tr[, , j] * beta[j]
                    }
                }
                if (force != 0) {
                    U.tr <- U.tr - matrix(est.co$mu, TT, Ntr) ## grand mean
                }
                if (force %in% c(2, 3)) {
                    U.tr <- U.tr - matrix(est.co$xi, TT, Ntr, byrow = FALSE)
                }

                if (0 %in% I.tr) {
                    U.tr[which(I.tr == 0)] <- 0
                }

                U.sav <- U.tr

                resid_all <- c()
                for (lv in unique(unlist(time.pre))) {
                    U.tr <- U.sav
                    if (max(T0) == T0.min & (!0 %in% I.tr)) {
                        U.lv <- as.matrix(U.tr[setdiff(c(1:T0.min), lv), ]) ## setdiff : x
                    } else {
                        U.tr.pre.v <- as.vector(U.tr)[which(pre.v == 1)] ## pre-treatment residual in a vector
                        U.tr.pre <- split(U.tr.pre.v, id.tr.pre.v) ##  a list of pretreatment residuals
                        if (!0 %in% I.tr) {
                            U.lv <- lapply(U.tr.pre, function(vec) {
                                return(vec[-lv])
                            }) ## a list
                        } else {
                            ## U.tr.pre.sav <- U.tr.pre
                            for (i.tr in 1:Ntr) {
                                U.tmp <- U.tr.pre[[i.tr]]
                                U.tr.pre[[i.tr]] <- U.tmp[!time.pre[[i.tr]] == lv]
                            }
                            U.lv <- U.tr.pre
                        }
                    }

                    if (r == 0) {
                        if (force %in% c(1, 3)) { ## take out unit fixed effect
                            if (max(T0) == T0.min & (!0 %in% I.tr)) {
                                alpha.tr.lv <- colMeans(U.lv)
                                U.tr <- U.tr - matrix(alpha.tr.lv, TT, Ntr, byrow = TRUE)
                            } else {
                                alpha.tr.lv <- sapply(U.lv, mean)
                                U.tr <- U.tr - matrix(alpha.tr.lv, TT, Ntr, byrow = TRUE)
                            }
                        }
                        e <- U.tr[which(time == lv), ] ## that period
                    } else {
                        F.lv <- as.matrix(F.hat[which(time != lv), ])
                        if (max(T0) == T0.min & (!0 %in% I.tr)) {
                            F.lv.pre <- F.hat[setdiff(c(1:T0.min), lv), ]
                            lambda.lv <- try(
                                solve(t(F.lv.pre) %*% F.lv.pre) %*% t(F.lv.pre) %*% U.lv,
                                silent = TRUE
                            )
                            if ("try-error" %in% class(lambda.lv)) {
                                break
                            }
                        } else {
                            if (!0 %in% I.tr) {
                                lambda.lv <- try(as.matrix(sapply(U.lv, function(vec) {
                                    F.lv.pre <- as.matrix(F.lv[1:length(vec), ])
                                    l.lv.tr <- solve(t(F.lv.pre) %*% F.lv.pre) %*% t(F.lv.pre) %*% vec
                                    return(l.lv.tr)
                                })), silent = TRUE)
                                if ("try-error" %in% class(lambda.lv)) {
                                    break
                                } else {
                                    if ((r == 1) & (force %in% c(0, 2))) {
                                        lambda.lv <- t(lambda.lv)
                                    }
                                }
                            } else {
                                if (force %in% c(1, 3)) {
                                    lambda.lv <- matrix(NA, (r + 1), Ntr)
                                } else {
                                    lambda.lv <- matrix(NA, r, Ntr)
                                }
                                test <- try(
                                    for (i.tr in 1:Ntr) {
                                        F.lv.pre <- as.matrix(F.hat[setdiff(time.pre[[i.tr]], lv), ])
                                        lambda.lv[, i.tr] <- solve(t(F.lv.pre) %*% F.lv.pre) %*% t(F.lv.pre) %*% as.matrix(U.lv[[i.tr]])
                                    },
                                    silent = TRUE
                                )
                                if ("try-error" %in% class(test)) {
                                    break
                                }
                            }
                        }
                        lambda.lv <- t(lambda.lv) ## N*r
                        e <- U.tr[which(time == lv), ] - c(F.hat[which(time == lv), ] %*% t(lambda.lv))
                    }
                    if (sameT0 == FALSE | 0 %in% I.tr) {
                        e <- e[which(pre[which(time == lv), ] == TRUE)]
                    }
                    ## accumulate residuals
                    resid_all <- c(resid_all, e)
                } ## end of leave-one-out

                if (length(resid_all) == 0) {
                    scores <- c(MSPE = Inf, WMSPE = Inf, GMSPE = Inf, WGMSPE = Inf,
                                MAD = Inf, Moment = Inf, GMoment = Inf, RMSE = Inf, Bias = Inf)
                } else {
                    ## Build time indices for LOO residuals
                    time_idx_loo <- NULL
                    obs_w_loo <- NULL
                    if (!is.null(count.T.cv)) {
                        time_idx_loo <- c()
                        for (lv in unique(unlist(time.pre))) {
                            if (sameT0 == FALSE | 0 %in% I.tr) {
                                n_resid_lv <- sum(pre[which(time == lv), ] == TRUE)
                            } else {
                                n_resid_lv <- Ntr
                            }
                            if (n_resid_lv > 0) {
                                t.on.tr <- T.on[, tr, drop = FALSE]
                                t.on.lv <- unique(t.on.tr[lv, ])
                                t.on.lv <- t.on.lv[!is.na(t.on.lv)]
                                if (length(t.on.lv) > 0) {
                                    time_idx_loo <- c(time_idx_loo, rep(as.character(t.on.lv[1]), n_resid_lv))
                                } else {
                                    time_idx_loo <- c(time_idx_loo, rep("Control", n_resid_lv))
                                }
                            }
                        }
                    }
                    if (!is.null(W.tr)) {
                        obs_w_loo <- c()
                        for (lv in unique(unlist(time.pre))) {
                            if (sameT0 == FALSE | 0 %in% I.tr) {
                                w_lv <- W.tr[lv, which(pre[which(time == lv), ] == TRUE)]
                            } else {
                                w_lv <- W.tr[lv, ]
                            }
                            obs_w_loo <- c(obs_w_loo, w_lv)
                        }
                    }
                    scores <- .score_residuals(
                        resid_all,
                        obs_weights = obs_w_loo,
                        time_index = time_idx_loo,
                        count_weights = count.T.cv,
                        norm.para = norm.para
                    )
                }

              } else if (cv.method %in% c("all_units", "rolling")) {
                ## ---- cv.sample "all_units" / "rolling" IFE CV (serial path — lapply only) ---- ##
                ## Rolling reuses the all_units scoring helper; only the rmCV/estCV
                ## fold construction differs (built via .build_cv_mask_rolling above).
                fold_results <- lapply(1:k, function(ii) {
                    .fect_cv_score_one_ife_nt_all(
                        ii       = ii,
                        YY.co    = YY.co,
                        Y0CV.co  = Y0CV.co,
                        X.co     = X.co,
                        II.co    = II.co,
                        W.use    = W.use,
                        W        = W,
                        beta0CV.co = beta0CV.co,
                        rmCV     = rmCV,
                        estCV    = estCV,
                        r        = r,
                        force    = force,
                        cv_tol   = cv_tol,
                        max.iteration = max.iteration
                    )
                })
                if (length(fold_results) == 0L) {
                    scores <- c(MSPE = Inf, WMSPE = Inf, GMSPE = Inf, WGMSPE = Inf,
                                MAD = Inf, Moment = Inf, GMoment = Inf, RMSE = Inf, Bias = Inf)
                    se_v <- setNames(rep(NA_real_, length(scores)), names(scores))
                } else {
                    agg <- .fect_cv_aggregate_folds(
                        fold_list  = fold_results,
                        count.T.cv = count.T.cv,
                        use_weight = as.integer(!is.null(W)),
                        norm.para  = NULL
                    )
                    scores <- agg$pooled
                    se_v   <- agg$se
                }

              } else {
                ## ---- cv.sample "treated_units" IFE CV (serial path — lapply only) ---- ##
                if (r != 0) {
                    F.hat <- as.matrix(est.co$factor)
                    if (force %in% c(1, 3)) {
                        F.hat <- cbind(F.hat, rep(1, TT))
                    }
                }
                U.tr <- Y.tr
                if (p > 0) {
                    for (j in 1:p) {
                        U.tr <- U.tr - X.tr[, , j] * beta[j]
                    }
                }
                if (force != 0) {
                    U.tr <- U.tr - matrix(est.co$mu, TT, Ntr)
                }
                if (force %in% c(2, 3)) {
                    U.tr <- U.tr - matrix(est.co$xi, TT, Ntr, byrow = FALSE)
                }
                if (0 %in% I.tr) {
                    U.tr[which(I.tr == 0)] <- 0
                }

                fold_results <- lapply(1:k, function(ii) {
                    .fect_cv_score_one_ife_nt_tr(
                        ii       = ii,
                        U.tr     = U.tr,
                        F.hat    = F.hat,
                        pre      = pre,
                        r        = r,
                        force    = force,
                        rmCV.tr  = rmCV.tr,
                        estCV.tr = estCV.tr,
                        W.tr     = W.tr,
                        T.on     = T.on,
                        tr       = tr,
                        TT       = TT,
                        Ntr      = Ntr
                    )
                })
                if (length(fold_results) == 0L) {
                    scores <- c(MSPE = Inf, WMSPE = Inf, GMSPE = Inf, WGMSPE = Inf,
                                MAD = Inf, Moment = Inf, GMoment = Inf, RMSE = Inf, Bias = Inf)
                    se_v <- setNames(rep(NA_real_, length(scores)), names(scores))
                } else {
                    agg <- .fect_cv_aggregate_folds(
                        fold_list  = fold_results,
                        count.T.cv = count.T.cv,
                        use_weight = as.integer(!is.null(W.tr)),
                        norm.para  = norm.para
                    )
                    scores <- agg$pooled
                    se_v   <- agg$se
                }

              } ## end cv.method branching

                if ((min(CV.out[, crit_col]) - scores[crit_col]) > 0.01 * min(CV.out[, crit_col])) {
                    ## at least 1% improvement for selected criterion
                    est.co.best <- est.co ## interFE result with the best r
                    r.cv <- r
                } else {
                    if (r == r.cv + 1) message("*")
                }

                if (PC < min(CV.out[, "PC"])) {
                    r.pc <- r
                    est.co.pc.best <- est.co
                }
                CV.out[i, 2:4] <- c(sigma2, IC, PC)
                CV.out[i, score_names] <- scores[score_names]
                ## Persist per-fold SEs for end-of-loop cv.rule application.
                ## Some serial paths (e.g., LOO branch) may leave se_v undefined;
                ## existsCheck protects those cases.
                if (exists("se_v", inherits = FALSE) && !is.null(se_v)) {
                    for (cn in score_names) {
                        if (cn %in% names(se_v)) CV.out.se[i, cn] <- se_v[cn]
                    }
                }
                message("r = ", r, "; sigma2 = ",
                    sprintf("%.5f", sigma2), "; IC = ",
                    sprintf("%.5f", IC), "; PC = ",
                    sprintf("%.5f", PC), "; MSPE = ",
                    sprintf("%.5f", scores["MSPE"]),
                    sep = ""
                )
            } ## end of while: search for r_star over

            } ## end SERIAL BRANCH (IFE)

            MSPE.best <- min(CV.out[, "MSPE"])

            ## --- Apply cv.rule (added v2.3.0) -----------------------------
            ## Override r.cv based on the user-selected rule. The default "1se"
            ## picks the smallest r within one fold-SE of the minimum-CV-error r;
            ## "min" picks the argmin; "1pct" preserves the legacy 1% rule from
            ## the in-loop assignments above.
            if (criterion %in% c("mspe","wmspe","gmspe","wgmspe","mad","moment","gmoment")) {
                means <- CV.out[, crit_col]
                ses   <- CV.out.se[, crit_col]
                ## Treat sentinels (1e10, Inf, NA) as missing for selection.
                means[!is.finite(means) | means >= 1e9] <- NA_real_
                i_pick <- .fect_apply_cv_rule(means, ses, rule = cv.rule)
                if (!is.na(i_pick) && i_pick >= 1L && i_pick <= nrow(CV.out)) {
                    new_r_cv <- unname(CV.out[i_pick, "r"])
                    if (!is.null(new_r_cv) && is.finite(new_r_cv)) {
                        if (new_r_cv != as.integer(unname(r.cv))) {
                            message(sprintf(
                                "  [cv.rule = %s] r.cv adjusted from %d to %d (1-SE band)",
                                cv.rule,
                                as.integer(unname(r.cv)),
                                as.integer(new_r_cv)
                            ))
                            est.co.best <- .estimate_co(
                                YY.co, Y0.co, X.co, I.co, W.use, beta0,
                                as.integer(new_r_cv), force, cv_tol, max.iteration
                            )
                        }
                        ## Preserve the in-loop names convention: serial path
                        ## leaves r.cv unnamed; parallel path sets names "r".
                        had_name <- !is.null(names(r.cv))
                        r.cv <- new_r_cv
                        if (had_name) names(r.cv) <- "r"
                    }
                }
            }
            ## --------------------------------------------------------------

            if (r > (T0.min - 1)) {
                message(" (r hits maximum)")
            }
            message("\n r* = ", r.cv, sep = "")
            message("\n")
        }
    } else {
        r.cv <- r
        r.min <- r.max <- r
    }

    est.co.fect <- NULL

    est.co.best <- .estimate_co(YY.co, Y0.co, X.co, II.co, W.use, beta0, r.cv, force, tol, max.iteration,
                                fit_init = fit.init.co)

    if (boot == FALSE) {
        if (r.cv == 0) {
            est.co.fect <- est.co.best
        } else {
            est.co.fect <- .estimate_co(YY.co, Y0.co, X.co, II.co, W.use, beta0, 0, force, tol, max.iteration)
        }
    }
    validX <- est.co.best$validX
    validF <- ifelse(r.cv > 0, 1, 0)

    # get the counterfactual of X
    ## ## take out the effect of X
    U.tr.r0 <- U.tr <- Y.tr
    if (p > 0) {
        beta <- est.co.best$beta

        if (est.co.best$validX == 0) {
            beta <- matrix(0, p, 1)
        } else {
            beta <- est.co.best$beta
            beta[is.nan(est.co.best$beta)] <- 0
        }
        for (j in 1:p) {
            U.tr <- U.tr - X.tr[, , j] * beta[j]
        }

        if (boot == FALSE) {
            beta.r0 <- est.co.fect$beta
            if (est.co.fect$validX == 0) {
                beta.r0 <- matrix(0, p, 1)
            } else {
                beta.r0 <- est.co.fect$beta
                beta.r0[is.nan(est.co.fect$beta)] <- 0
            }
            for (j in 1:p) {
                U.tr.r0 <- U.tr.r0 - X.tr[, , j] * beta.r0[j]
            }
        }
    } else {
        beta <- NA
        beta.r0 <- NA
    }

    mu <- est.co.best$mu
    U.tr <- U.tr - matrix(mu, TT, Ntr) ## grand mean
    Y.fe.bar <- rep(mu, TT)

    if (boot == FALSE) {
        mu.r0 <- est.co.fect$mu
        U.tr.r0 <- U.tr.r0 - matrix(mu.r0, TT, Ntr)
        Y.fe.bar.r0 <- rep(mu.r0, TT)
    }

    if (force %in% c(2, 3)) {
        xi <- est.co.best$xi ## a (TT*1) matrix
        U.tr <- U.tr - matrix(c(xi), TT, Ntr, byrow = FALSE) ## will be adjusted at last
        Y.fe.bar <- Y.fe.bar + xi
        if (boot == FALSE) {
            xi.r0 <- est.co.fect$xi ## a (TT*1) matrix
            U.tr.r0 <- U.tr.r0 - matrix(c(xi.r0), TT, Ntr, byrow = FALSE)
            Y.fe.bar.r0 <- Y.fe.bar.r0 + xi.r0
        }
    }

    if (max(T0) == T0.min & (!0 %in% I.tr)) {
        U.tr.pre <- as.matrix(U.tr[1:T0.min, ])
        if (boot == FALSE) {
            U.tr.pre.r0 <- as.matrix(U.tr.r0[1:T0.min, ])
        }
    } else {
        ## not necessary to reset utr for ub data for pre.v doesn't include them
        U.tr.pre.v <- as.vector(U.tr)[which(pre.v == 1)] # pre-treatment residual in a vector
        U.tr.pre <- split(U.tr.pre.v, id.tr.pre.v) ##  a list of pretreatment residuals

        if (boot == FALSE) {
            U.tr.pre.v.r0 <- as.vector(U.tr.r0)[which(pre.v == 1)] # pre-treatment residual in a vector
            U.tr.pre.r0 <- split(U.tr.pre.v.r0, id.tr.pre.v) ##  a list of pretreatment residuals
        }
    }

    ## the error structure

    # for r=0
    if (force %in% c(1, 3)) { ## take out unit fixed effect
        if ((max(T0) == T0.min) & (!0 %in% I.tr)) {
            if (boot == FALSE) {
                alpha.tr.r0 <- as.matrix(colMeans(U.tr.pre.r0))
                U.tr.r0 <- U.tr.r0 - matrix(alpha.tr.r0, TT, Ntr, byrow = TRUE)
            }
        } else {
            if (boot == FALSE) {
                alpha.tr.r0 <- as.matrix(sapply(U.tr.pre.r0, mean))
                U.tr.r0 <- U.tr.r0 - matrix(alpha.tr.r0, TT, Ntr, byrow = TRUE)
            }
        }
    }
    if (boot == FALSE) {
        eff.r0 <- U.tr.r0
    }



    if (r.cv == 0) {
        if (force %in% c(1, 3)) { ## take out unit fixed effect
            if ((max(T0) == T0.min) & (!0 %in% I.tr)) {
                alpha.tr <- as.matrix(colMeans(U.tr.pre))
                U.tr <- U.tr - matrix(alpha.tr, TT, Ntr, byrow = TRUE)
            } else {
                alpha.tr <- as.matrix(sapply(U.tr.pre, mean))
                U.tr <- U.tr - matrix(alpha.tr, TT, Ntr, byrow = TRUE)
            }
        }
        eff <- U.tr
        lambda.tr <- NULL
        lambda.co <- NULL
    } else { ## Factors
        F.hat <- as.matrix(est.co.best$factor)
        if (force %in% c(1, 3)) {
            F.hat <- cbind(F.hat, rep(1, TT))
        }

        ## Bounded-loading dispatch: resolves use_bounded, gamma_use, and
        ## optional W_tr / loading.proj.resid diagnostics. See statsclaw-
        ## workspace/fect/runs/REQ-bounded-loadings/spec.md for the design.
        use_bounded <- identical(loading.bound, "simplex")
        W_tr <- NULL
        loading.proj.resid <- NULL

        if (use_bounded) {
            Lambda.co.raw <- as.matrix(est.co.best$lambda)   # Nco x r.cv (no intercept col)
            F.hat.raw     <- as.matrix(est.co.best$factor)   # TT  x r.cv (no intercept col)
            Nco_b         <- nrow(Lambda.co.raw)
            gamma_grid_use <- if (!is.null(gamma.loading.grid)) gamma.loading.grid
                              else .default_gamma_grid()
            gamma_use <- gamma.loading
            if (is.null(gamma_use)) {
                balanced_pre <- max(T0) == T0.min & !0 %in% I.tr
                if (balanced_pre) {
                    cv_F_pre <- F.hat.raw[1:T0.min, , drop = FALSE]
                    cv_U_pre <- U.tr.pre
                } else if (!0 %in% I.tr) {
                    ## Different T0 per unit; U.tr.pre is a list
                    cv_F_pre <- lapply(U.tr.pre, function(vec)
                        F.hat.raw[seq_along(vec), , drop = FALSE]
                    )
                    cv_U_pre <- lapply(U.tr.pre, as.numeric)
                } else {
                    ## Some missing observations in treated pre-period
                    cv_F_pre <- lapply(seq_along(U.tr.pre), function(i.tr)
                        F.hat.raw[time.pre[[i.tr]], , drop = FALSE]
                    )
                    cv_U_pre <- lapply(U.tr.pre, as.numeric)
                }
                gamma_use <- .cv_gamma_loading(
                    U_tr_pre   = cv_U_pre,
                    F_hat_pre  = cv_F_pre,
                    Lambda_co  = Lambda.co.raw,
                    gamma_grid = gamma_grid_use,
                    cv_k       = 5L
                )$gamma_cv
            }
        }

        ## Lambda_tr (Ntr*r) or (Ntr*(r+1))
        if (max(T0) == T0.min & (!0 %in% I.tr)) {
            F.hat.pre <- F.hat[1:T0.min, ]
            if (!use_bounded) {
                lambda.tr <- try(solve(t(F.hat.pre) %*% F.hat.pre) %*% t(F.hat.pre) %*% U.tr.pre,
                    silent = TRUE
                )
                if ("try-error" %in% class(lambda.tr)) {
                    return(list(att = rep(NA, TT), att.avg = NA, beta = matrix(NA, p, 1)))
                }
            } else {
                ## Bounded: solve simplex QP per treated unit on the r-col F block only
                F.pre.r <- F.hat.raw[1:T0.min, , drop = FALSE]
                lambda.tr.r <- matrix(NA_real_, nrow = r.cv, ncol = Ntr)
                W_tr <- matrix(NA_real_, nrow = Ntr, ncol = Nco_b)
                loading.proj.resid <- numeric(Ntr)
                for (i.tr in seq_len(Ntr)) {
                    sol <- .solve_bounded_loading(
                        u_pre     = U.tr.pre[, i.tr],
                        F_pre     = F.pre.r,
                        Lambda_co = Lambda.co.raw,
                        gamma     = gamma_use
                    )
                    lambda.tr.r[, i.tr] <- sol$lambda_hat
                    W_tr[i.tr, ] <- sol$w
                    loading.proj.resid[i.tr] <- sqrt(sum(
                        (U.tr.pre[, i.tr] - F.pre.r %*% sol$lambda_hat)^2
                    ))
                }
                if (force %in% c(1, 3)) {
                    ## alpha.tr is residual-mean (NOT bounded; per locked decision)
                    resid_mat <- U.tr.pre - F.pre.r %*% lambda.tr.r   # T0.min x Ntr
                    alpha_row <- matrix(colMeans(resid_mat), nrow = 1L)
                    lambda.tr <- rbind(lambda.tr.r, alpha_row)
                } else {
                    lambda.tr <- lambda.tr.r
                }
            }
        } else {
            if (!0 %in% I.tr) {
                if (!use_bounded) {
                    lambda.tr <- try(as.matrix(sapply(U.tr.pre, function(vec) {
                        F.hat.pre <- as.matrix(F.hat[1:length(vec), ])
                        l.tr <- solve(t(F.hat.pre) %*% F.hat.pre) %*% t(F.hat.pre) %*% vec
                        return(l.tr) ## a vector of each individual lambdas
                    })), silent = TRUE)
                    if ("try-error" %in% class(lambda.tr)) {
                        return(list(att = rep(NA, TT), att.avg = NA, beta = matrix(NA, p, 1)))
                        ## stop("Error occurs. Please set a smaller value of factor number.")
                    }
                    if ((r.cv == 1) & (force %in% c(0, 2))) {
                        lambda.tr <- t(lambda.tr)
                    }
                } else {
                    lambda.tr.r <- matrix(NA_real_, nrow = r.cv, ncol = Ntr)
                    W_tr <- matrix(NA_real_, nrow = Ntr, ncol = Nco_b)
                    loading.proj.resid <- numeric(Ntr)
                    alpha_vec <- numeric(Ntr)
                    for (i.tr in seq_len(Ntr)) {
                        vec <- U.tr.pre[[i.tr]]
                        F.pre.i <- F.hat.raw[seq_along(vec), , drop = FALSE]
                        sol <- .solve_bounded_loading(
                            u_pre     = vec,
                            F_pre     = F.pre.i,
                            Lambda_co = Lambda.co.raw,
                            gamma     = gamma_use
                        )
                        lambda.tr.r[, i.tr] <- sol$lambda_hat
                        W_tr[i.tr, ] <- sol$w
                        loading.proj.resid[i.tr] <- sqrt(sum(
                            (vec - F.pre.i %*% sol$lambda_hat)^2
                        ))
                        alpha_vec[i.tr] <- mean(vec - F.pre.i %*% sol$lambda_hat)
                    }
                    if (force %in% c(1, 3)) {
                        lambda.tr <- rbind(lambda.tr.r, matrix(alpha_vec, nrow = 1L))
                    } else {
                        lambda.tr <- lambda.tr.r
                    }
                }
            } else {
                if (!use_bounded) {
                    if (force %in% c(1, 3)) {
                        lambda.tr <- matrix(NA, (r.cv + 1), Ntr)
                    } else {
                        lambda.tr <- matrix(NA, r.cv, Ntr)
                    }
                    test <- try(
                        for (i.tr in 1:Ntr) {
                            F.hat.pre <- as.matrix(F.hat[time.pre[[i.tr]], ])
                            lambda.tr[, i.tr] <- solve(t(F.hat.pre) %*% F.hat.pre) %*% t(F.hat.pre) %*% as.matrix(U.tr.pre[[i.tr]])
                        },
                        silent = TRUE
                    )
                    if ("try-error" %in% class(test)) {
                        return(list(att = rep(NA, TT), att.avg = NA, beta = matrix(NA, p, 1), eff = matrix(NA, TT, Ntr)))
                        ## stop("Error occurs. Please set a smaller value of factor number.")
                    }
                } else {
                    lambda.tr.r <- matrix(NA_real_, nrow = r.cv, ncol = Ntr)
                    W_tr <- matrix(NA_real_, nrow = Ntr, ncol = Nco_b)
                    loading.proj.resid <- numeric(Ntr)
                    alpha_vec <- numeric(Ntr)
                    for (i.tr in seq_len(Ntr)) {
                        vec    <- as.numeric(U.tr.pre[[i.tr]])
                        F.pre.i <- F.hat.raw[time.pre[[i.tr]], , drop = FALSE]
                        sol <- .solve_bounded_loading(
                            u_pre     = vec,
                            F_pre     = F.pre.i,
                            Lambda_co = Lambda.co.raw,
                            gamma     = gamma_use
                        )
                        lambda.tr.r[, i.tr] <- sol$lambda_hat
                        W_tr[i.tr, ] <- sol$w
                        loading.proj.resid[i.tr] <- sqrt(sum(
                            (vec - F.pre.i %*% sol$lambda_hat)^2
                        ))
                        alpha_vec[i.tr] <- mean(vec - F.pre.i %*% sol$lambda_hat)
                    }
                    if (force %in% c(1, 3)) {
                        lambda.tr <- rbind(lambda.tr.r, matrix(alpha_vec, nrow = 1L))
                    } else {
                        lambda.tr <- lambda.tr.r
                    }
                }
            }
        }

        lambda.tr <- t(lambda.tr)
        eff <- U.tr - F.hat %*% t(lambda.tr)
        if (force %in% c(1, 3)) {
            alpha.tr <- as.matrix(lambda.tr[, (r.cv + 1), drop = FALSE])
            lambda.tr <- lambda.tr[, 1:r.cv, drop = FALSE]
        }
        if (boot == 0) {
            if (use_bounded && !is.null(W_tr)) {
                wgt.implied <- W_tr
            } else {
                inv.tr <- try(
                    ginv(t(as.matrix(lambda.tr))),
                    silent = TRUE
                )

                if (!"try-error" %in% class(inv.tr)) {
                    wgt.implied <- t(inv.tr %*% t(as.matrix(est.co.best$lambda)))
                }
            }
        }
    } ## end of r!=0 case


    if (0 %in% I.tr) {
        eff[which(I.tr == 0)] <- 0 ## adjust
        if (boot == FALSE) {
            eff.r0[which(I.tr == 0)] <- 0
        }
    } ## missing data will be adjusted to NA finally

  } else if (method == "cfe") {
    ## ====================================================================
    ## CFE PATH (new code)
    ## ====================================================================

    ## ---- Validate sufficient control units ----
    if (Nco < 2) {
        stop("Too few never-treated (control) units for CFE estimation. ",
             "At least 2 control units are required, but only ", Nco, " found.")
    }

    ## ---- Initial fit for co-only data ----
    beta0 <- matrix(0, p, 1)
    data.ini <- matrix(NA, Nco * TT, (p + 3))
    data.ini[, 1] <- c(Y.co)
    data.ini[, 2] <- rep(1:Nco, each = TT)
    data.ini[, 3] <- rep(1:TT, Nco)
    if (p > 0) {
        for (i in 1:p) {
            data.ini[, (3 + i)] <- c(X.co[, , i])
        }
    }
    initialOut <- Y0.co <- NULL
    oci <- which(c(II.co) == 1)
    if (is.null(W)) {
        initialOut <- initialFit(data = data.ini, force = force, oci = oci)
    } else {
        initialOut <- initialFit(data = data.ini, force = force, w = c(W.use), oci = oci)
    }
    Y0.co <- initialOut$Y0
    beta0 <- initialOut$beta0
    if (p > 0 && sum(is.na(beta0)) > 0) {
        beta0[which(is.na(beta0))] <- 0
    }

    validX <- 1

    ## ---- CV loop for CFE ----
    if (CV == TRUE) {
        ## Two-tier tolerance for CFE CV
        cv_tol <- max(tol, 1e-3)

        ## starting r
        if ((r > (T0.min - 1) & force %in% c(0, 2)) | (r > (T0.min - 2) & force %in% c(1, 3))) {
            message("r is too big compared with T0; reset to 0.")
            r <- 0
        }
        if (force %in% c(0, 2)) {
            r.max <- max(min((T0.min - 1), r.end), 0)
        } else {
            r.max <- max(min((T0.min - 2), r.end), 0)
        }

        if (r.max == 0) {
            r.cv <- 0
            message("Cross validation cannot be performed since available pre-treatment records of treated units are too few. So set r.cv = 0.")
            est.co.best <- complex_fe_ub(YY.co, Y0.co, X.co,
                X.extra.FE.co.B, X.Z.co, X.Q.co, X.gamma.co, X.kappa.co,
                Zgamma.id, kappaQ.id,
                II.co, W.use, beta0, 0, force = force, cv_tol, max.iteration,
                fit_init = fit.init.co)
        } else {
            r.old <- r
            message("Cross-validating ...", "\r")
            score_names <- c("MSPE", "WMSPE", "GMSPE", "WGMSPE",
                             "MAD", "Moment", "GMoment", "RMSE", "Bias")
            CV.out <- matrix(NA, (r.max - r.old + 1), 4 + length(score_names))
            colnames(CV.out) <- c("r", "sigma2", "IC", "PC", score_names)
            CV.out[, "r"] <- c(r.old:r.max)
            CV.out[, score_names] <- 1e10
            CV.out[, "PC"] <- 1e10
            r.pc <- est.co.pc.best <- NULL

            crit_col <- switch(criterion,
                mspe = "MSPE", wmspe = "WMSPE", gmspe = "GMSPE", wgmspe = "WGMSPE",
                mad = "MAD", moment = "Moment", gmoment = "GMoment", "MSPE")

            ## ---- cv.sample pre-computation (CFE) ---- ##
            if (cv.method != "loo" && r.max > 0) {
                if (cv.method %in% c("all_units", "rolling")) {
                    rm.count.co <- floor(sum(II.co) * cv.prop)
                    if (rm.count.co == 0 && cv.method == "all_units") {
                        message("cv.prop too small for control panel; falling back to LOO.")
                        cv.method <- "loo"
                    } else {
                        D.co.fake <- matrix(0, TT, Nco)
                        oci.co <- which(c(II.co) == 1)
                        rmCV <- list()
                        ociCV <- list()
                        estCV <- list()
                        Y0CV.co <- array(NA, dim = c(TT, Nco, k))
                        if (p > 0) {
                            beta0CV.co <- array(NA, dim = c(p, 1, k))
                        } else {
                            beta0CV.co <- array(0, dim = c(1, 0, k))
                        }
                        flag.cv <- 0

                        ## ---- rolling-window pre-computation (CFE) ---- ##
                        rolling_folds <- NULL
                        if (cv.method == "rolling") {
                            rolling_folds <- .build_cv_mask_rolling(
                                II = II.co, D = D.co.fake, k = k,
                                cv.nobs = cv.nobs, cv.buffer = cv.buffer,
                                cv.prop = cv.prop, min.T0 = min.T0, seed = NULL
                            )
                        }

                        for (i.cv in 1:k) {
                            if (cv.method == "rolling") {
                                cv.n <- 0
                                cv.id  <- rolling_folds[[i.cv]]$cv.id
                                est.id <- rolling_folds[[i.cv]]$est.id
                            } else {
                                cv.n <- 0
                                repeat {
                                    cv.n <- cv.n + 1
                                    get.cv <- cv.sample(II.co, D.co.fake,
                                        count = rm.count.co,
                                        cv.count = cv.nobs,
                                        cv.treat = FALSE,
                                        cv.donut = cv.donut)
                                    cv.id <- get.cv$cv.id
                                    II.co.cv <- II.co
                                    II.co.cv[cv.id] <- 0
                                    II.co.cv.valid <- II.co
                                    II.co.cv.valid[cv.id] <- -1
                                    con1 <- sum(apply(II.co.cv, 1, sum) >= 1) == TT
                                    con2 <- sum(apply(II.co.cv, 2, sum) >= min.T0) == Nco
                                    if (con1 && con2) break
                                    if (cv.n >= 200) {
                                        flag.cv <- 1
                                        keep.1 <- which(apply(II.co.cv, 1, sum) < 1)
                                        keep.2 <- which(apply(II.co.cv, 2, sum) < min.T0)
                                        II.co.cv[keep.1, ] <- II.co[keep.1, ]
                                        II.co.cv[, keep.2] <- II.co[, keep.2]
                                        II.co.cv.valid[keep.1, ] <- II.co[keep.1, ]
                                        II.co.cv.valid[, keep.2] <- II.co[, keep.2]
                                        cv.id <- which(II.co.cv.valid != II.co)
                                        break
                                    }
                                }
                            }
                            rmCV[[i.cv]] <- cv.id
                            ociCV[[i.cv]] <- setdiff(oci.co, cv.id)
                            if (cv.method == "rolling") {
                                estCV[[i.cv]] <- est.id
                            } else if (cv.n < 200) {
                                estCV[[i.cv]] <- get.cv$est.id
                            } else {
                                cv.diff <- setdiff(get.cv$cv.id, cv.id)
                                estCV[[i.cv]] <- setdiff(get.cv$est.id, cv.diff)
                            }
                            if (is.null(W) || !W.in.fit) {
                                initialOutCv <- initialFit(data = data.ini, force = force, oci = ociCV[[i.cv]])
                            } else {
                                initialOutCv <- initialFit(data = data.ini, force = force, w = c(W.use), oci = ociCV[[i.cv]])
                            }
                            Y0CV.co[, , i.cv] <- initialOutCv$Y0
                            if (p > 0) {
                                beta0cv <- initialOutCv$beta0
                                if (sum(is.na(beta0cv)) > 0) {
                                    beta0cv[which(is.na(beta0cv))] <- 0
                                }
                                beta0CV.co[, , i.cv] <- beta0cv
                            }
                        }
                        if (flag.cv == 1) {
                            message("Some control units have too few observations. Removed automatically in CV.\n")
                        }
                    }
                } else if (cv.method == "treated_units") {
                    rm.count.tr <- floor(sum(pre) * cv.prop)
                    if (rm.count.tr == 0) {
                        message("cv.prop too small for treated pre-treatment panel; falling back to LOO.")
                        cv.method <- "loo"
                    } else {
                        D.tr.fake <- matrix(0, TT, Ntr)
                        rmCV.tr <- list()
                        estCV.tr <- list()
                        flag.cv <- 0
                        for (i.cv in 1:k) {
                            cv.n <- 0
                            repeat {
                                cv.n <- cv.n + 1
                                get.cv <- cv.sample(pre, D.tr.fake,
                                    count = rm.count.tr,
                                    cv.count = cv.nobs,
                                    cv.treat = FALSE,
                                    cv.donut = cv.donut)
                                cv.id <- get.cv$cv.id
                                pre.cv <- pre
                                pre.cv[cv.id] <- 0
                                con1 <- TRUE
                                pre.rows <- which(rowSums(pre) > 0)
                                if (length(pre.rows) > 0) {
                                    con1 <- all(rowSums(pre.cv[pre.rows, , drop = FALSE]) >= 1)
                                }
                                con2 <- all(colSums(pre.cv) >= min(min.T0, 2))
                                if (con1 && con2) break
                                if (cv.n >= 200) {
                                    flag.cv <- 1
                                    pre.cv.valid <- pre
                                    pre.cv.valid[cv.id] <- -1
                                    keep.1 <- pre.rows[rowSums(pre.cv[pre.rows, , drop = FALSE]) < 1]
                                    keep.2 <- which(colSums(pre.cv) < min(min.T0, 2))
                                    if (length(keep.1) > 0) {
                                        pre.cv[keep.1, ] <- pre[keep.1, ]
                                        pre.cv.valid[keep.1, ] <- pre[keep.1, ]
                                    }
                                    if (length(keep.2) > 0) {
                                        pre.cv[, keep.2] <- pre[, keep.2]
                                        pre.cv.valid[, keep.2] <- pre[, keep.2]
                                    }
                                    cv.id <- which(pre.cv.valid != pre)
                                    break
                                }
                            }
                            rmCV.tr[[i.cv]] <- cv.id
                            if (cv.n < 200) {
                                estCV.tr[[i.cv]] <- get.cv$est.id
                            } else {
                                cv.diff <- setdiff(get.cv$cv.id, cv.id)
                                estCV.tr[[i.cv]] <- setdiff(get.cv$est.id, cv.diff)
                            }
                        }
                        if (flag.cv == 1) {
                            message("Some treated units have too few pre-treatment observations. Removed automatically in CV.\n")
                        }
                    }
                }
            }

            ## ---- Parallel backend setup (CFE CV) — Phase 3 flat dispatch ---- ##
            ## Reuse do_parallel_cv derived at IFE setup; if not yet set (CFE-only call), derive now.
            if (is.null(do_parallel_cv)) {
                do_parallel_cv <- isTRUE(parallel) || "cv" %in% as.character(parallel)
            }
            if (!do_parallel_cv) {
                cv_cfe_parallel <- FALSE
            } else {
                ## Re-derive explicitly for CFE block (do NOT reuse use_explicit_cv_ife)
                use_explicit_cv_cfe <- "cv" %in% as.character(parallel) && !isTRUE(parallel)
                ## Centralized threshold gate: CFE threshold is 60000L (higher overhead per fit)
                cfe_threshold_met <- (Nco * TT) > .CV_PARALLEL_THRESH$cfe
                cv_cfe_parallel <- (cv.method %in% c("all_units", "rolling")) &&
                                   (cfe_threshold_met || use_explicit_cv_cfe) &&
                                   (k > 1)
            }
            if (cv_cfe_parallel) {
                if (is.null(cores)) {
                    cores <- max(1L, min(parallelly::availableCores(omit = 2L), 8L))
                }
                old.future.plan.cfe <- future::plan()
                on.exit(future::plan(old.future.plan.cfe), add = TRUE, after = FALSE)
                future::plan(future::cluster, workers = .fect_make_future_cluster(cores))
                ## doFuture::registerDoFuture() removed — not needed for future_lapply dispatch
                avail <- parallelly::availableCores()
                msg_line <- sprintf("Parallel CV (CFE): using %d of %d available cores.", cores, avail)
                pad <- strrep(" ", max(0, 56 - nchar(msg_line)))
                message("\n",
                    " +----------------------------------------------------------+\n",
                    " | ", msg_line, pad, " |\n",
                    " |                                                          |\n",
                    " | To change: set cores = <n> in fect().                    |\n",
                    " | Default: min(available - 2, 8).                          |\n",
                    " +----------------------------------------------------------+\n")
            }

            ## ---- Task list construction (CFE flat r×k dispatch) ---- ##
            if (cv_cfe_parallel) {
                r_seq_cfe <- CV.out[, "r"]
                tasks_cfe <- vector("list", length(r_seq_cfe) * k)
                idx <- 1L
                for (ri in seq_along(r_seq_cfe)) {
                    for (ii in 1:k) {
                        tasks_cfe[[idx]] <- list(r = r_seq_cfe[ri], ii = ii, ri = ri)
                        idx <- idx + 1L
                    }
                }
                ## Capture helper in closure for worker serialization
                .score_fn_cfe_all <- .fect_cv_score_one_cfe_nt_all
            }

            if (cv_cfe_parallel) {
            ## ---- PARALLEL BRANCH: flat r×k future_lapply dispatch (CFE all_units) ---- ##

                ## Step 1: dispatch all (r, fold) scoring tasks
                fold_scores_cfe <- future.apply::future_lapply(
                    tasks_cfe,
                    FUN = function(task) {
                        .score_fn_cfe_all(
                            ii              = task$ii,
                            YY.co           = YY.co,
                            Y0CV.co         = Y0CV.co,
                            X.co            = X.co,
                            II.co           = II.co,
                            W.use           = W.use,
                            W               = W,
                            beta0CV.co      = beta0CV.co,
                            X.extra.FE.co.B = X.extra.FE.co.B,
                            X.Z.co          = X.Z.co,
                            X.Q.co          = X.Q.co,
                            X.gamma.co      = X.gamma.co,
                            X.kappa.co      = X.kappa.co,
                            Zgamma.id       = Zgamma.id,
                            kappaQ.id       = kappaQ.id,
                            rmCV            = rmCV,
                            estCV           = estCV,
                            r               = task$r,
                            force           = force,
                            cv_tol          = cv_tol,
                            max.iteration   = max.iteration
                        )
                    },
                    future.seed     = TRUE,
                    future.packages = "fect"
                )

                ## Step 2: sequential master walk — apply 1% rule in rank order
                n_r_cfe <- length(r_seq_cfe)
                for (i in seq_len(n_r_cfe)) {
                    r <- unname(r_seq_cfe[i])
                    ## Full-data fit (sequential in master) — needed for sigma2/IC/PC
                    est.co <- complex_fe_ub(YY.co, Y0.co, X.co,
                        X.extra.FE.co.B, X.Z.co, X.Q.co, X.gamma.co, X.kappa.co,
                        Zgamma.id, kappaQ.id,
                        II.co, W.use, beta0, r, force = force, cv_tol, max.iteration)

                    if (p > 0) {
                        na.pos <- is.nan(est.co$beta)
                        beta <- est.co$beta
                        beta[is.nan(est.co$beta)] <- 0
                    }
                    if (is.null(norm.para)) {
                        sigma2 <- est.co$sigma2; IC <- est.co$IC; PC <- est.co$PC
                    } else {
                        sigma2 <- est.co$sigma2 * (norm.para[1]^2)
                        IC <- est.co$IC - log(est.co$sigma2) + log(sigma2)
                        PC <- est.co$PC * (norm.para[1]^2)
                    }

                    ## Aggregate fold scores for this rank
                    task_idx <- which(vapply(tasks_cfe, function(t) t$ri == i, logical(1)))
                    all_resid    <- unlist(lapply(fold_scores_cfe[task_idx], `[[`, "resid"))
                    all_time_idx <- unlist(lapply(fold_scores_cfe[task_idx], `[[`, "time_idx"))
                    all_obs_w    <- if (!is.null(W)) unlist(lapply(fold_scores_cfe[task_idx], `[[`, "obs_w")) else c()

                    if (length(all_resid) == 0) {
                        scores <- c(MSPE = Inf, WMSPE = Inf, GMSPE = Inf, WGMSPE = Inf,
                                    MAD = Inf, Moment = Inf, GMoment = Inf, RMSE = Inf, Bias = Inf)
                    } else {
                        scores <- .score_residuals(
                            all_resid,
                            obs_weights   = if (!is.null(W)) all_obs_w else NULL,
                            time_index    = all_time_idx,
                            count_weights = count.T.cv,
                            norm.para     = NULL
                        )
                    }

                    ## 1% rule — identical logic to serial path
                    if ((min(CV.out[, crit_col]) - scores[crit_col]) > 0.01 * min(CV.out[, crit_col])) {
                        est.co.best <- est.co
                        r.cv <- r
                    } else {
                        if (r == r.cv + 1) message("*")
                    }
                    if (PC < min(CV.out[, "PC"])) {
                        r.pc <- r
                        est.co.pc.best <- est.co
                    }
                    CV.out[i, 2:4] <- c(sigma2, IC, PC)
                    CV.out[i, score_names] <- scores[score_names]
                    message("r = ", r, "; sigma2 = ",
                        sprintf("%.5f", sigma2), "; IC = ",
                        sprintf("%.5f", IC), "; PC = ",
                        sprintf("%.5f", PC), "; MSPE = ",
                        sprintf("%.5f", scores["MSPE"]), sep = "")
                } ## end per-r master walk (CFE parallel)

            } else {
            ## ---- SERIAL BRANCH (existing r-loop, all cv.method values) ---- ##

            for (i in 1:dim(CV.out)[1]) {
                r <- unname(CV.out[i, "r"])
                est.co <- complex_fe_ub(YY.co, Y0.co, X.co,
                    X.extra.FE.co.B, X.Z.co, X.Q.co, X.gamma.co, X.kappa.co,
                    Zgamma.id, kappaQ.id,
                    II.co, W.use, beta0, r, force = force, cv_tol, max.iteration)

                if (p > 0) {
                    na.pos <- is.nan(est.co$beta)
                    beta <- est.co$beta
                    beta[is.nan(est.co$beta)] <- 0
                }

                if (is.null(norm.para)) {
                    sigma2 <- est.co$sigma2
                    IC <- est.co$IC
                    PC <- est.co$PC
                } else {
                    sigma2 <- est.co$sigma2 * (norm.para[1]^2)
                    IC <- est.co$IC - log(est.co$sigma2) + log(sigma2)
                    PC <- est.co$PC * (norm.para[1]^2)
                }

              if (cv.method == "loo") {
                ## ---- LOO CV (existing code) ---- ##
                ## Build U.tr: subtract Layer 1 from Y.tr
                if (r != 0) {
                    F.hat <- as.matrix(est.co$factor)
                    if (force %in% c(1, 3)) {
                        F.hat <- cbind(F.hat, rep(1, TT))
                    }
                }
                U.tr <- Y.tr
                if (p > 0) {
                    for (j in 1:p) {
                        U.tr <- U.tr - X.tr[, , j] * beta[j]
                    }
                }
                if (force != 0) {
                    U.tr <- U.tr - matrix(est.co$mu, TT, Ntr)
                }
                if (force %in% c(2, 3)) {
                    U.tr <- U.tr - matrix(est.co$xi, TT, Ntr, byrow = FALSE)
                }

                ## Subtract gamma for treated (Layer 1)
                if (!is.null(est.co$gamma) && length(est.co$gamma) > 0) {
                    for (k_g in seq_along(est.co$gamma)) {
                        gamma.fit.tr.k <- .reconstruct_gamma_fit_tr(
                            est.co$gamma[[k_g]], X.Z.tr, X.gamma.tr[, , k_g, drop = FALSE],
                            Zgamma.id[[k_g]], TT, Ntr)
                        U.tr <- U.tr - gamma.fit.tr.k
                    }
                }

                ## Subtract Type-B extra FE for treated (Layer 1)
                if (length(typeB_idx) > 0) {
                    typeB.fit.tr <- .extract_and_apply_typeB_fe(
                        est.co, X.co, X.extra.FE.co.B, X.extra.FE.tr,
                        typeB_idx, X.Z.co, X.gamma.co, X.Q.co, X.kappa.co,
                        Zgamma.id, kappaQ.id,
                        TT, Nco, Ntr, p, r, force)
                    U.tr <- U.tr - typeB.fit.tr
                }

                if (0 %in% I.tr) {
                    U.tr[which(I.tr == 0)] <- 0
                }

                U.sav <- U.tr

                ## Leave-one-out CV (same structure as IFE)
                resid_all <- c()
                for (lv in unique(unlist(time.pre))) {
                    U.tr <- U.sav
                    if (max(T0) == T0.min & (!0 %in% I.tr)) {
                        U.lv <- as.matrix(U.tr[setdiff(c(1:T0.min), lv), ])
                    } else {
                        U.tr.pre.v <- as.vector(U.tr)[which(pre.v == 1)]
                        U.tr.pre <- split(U.tr.pre.v, id.tr.pre.v)
                        if (!0 %in% I.tr) {
                            U.lv <- lapply(U.tr.pre, function(vec) {
                                return(vec[-lv])
                            })
                        } else {
                            for (i.tr in 1:Ntr) {
                                U.tmp <- U.tr.pre[[i.tr]]
                                U.tr.pre[[i.tr]] <- U.tmp[!time.pre[[i.tr]] == lv]
                            }
                            U.lv <- U.tr.pre
                        }
                    }

                    if (r == 0) {
                        if (force %in% c(1, 3)) {
                            if (max(T0) == T0.min & (!0 %in% I.tr)) {
                                alpha.tr.lv <- colMeans(U.lv)
                                U.tr <- U.tr - matrix(alpha.tr.lv, TT, Ntr, byrow = TRUE)
                            } else {
                                alpha.tr.lv <- sapply(U.lv, mean)
                                U.tr <- U.tr - matrix(alpha.tr.lv, TT, Ntr, byrow = TRUE)
                            }
                        }
                        e <- U.tr[which(time == lv), ]
                    } else {
                        F.lv <- as.matrix(F.hat[which(time != lv), ])
                        if (max(T0) == T0.min & (!0 %in% I.tr)) {
                            F.lv.pre <- F.hat[setdiff(c(1:T0.min), lv), ]
                            lambda.lv <- try(
                                solve(t(F.lv.pre) %*% F.lv.pre) %*% t(F.lv.pre) %*% U.lv,
                                silent = TRUE
                            )
                            if ("try-error" %in% class(lambda.lv)) {
                                break
                            }
                        } else {
                            if (!0 %in% I.tr) {
                                lambda.lv <- try(as.matrix(sapply(U.lv, function(vec) {
                                    F.lv.pre <- as.matrix(F.lv[1:length(vec), ])
                                    l.lv.tr <- solve(t(F.lv.pre) %*% F.lv.pre) %*% t(F.lv.pre) %*% vec
                                    return(l.lv.tr)
                                })), silent = TRUE)
                                if ("try-error" %in% class(lambda.lv)) {
                                    break
                                } else {
                                    if ((r == 1) & (force %in% c(0, 2))) {
                                        lambda.lv <- t(lambda.lv)
                                    }
                                }
                            } else {
                                if (force %in% c(1, 3)) {
                                    lambda.lv <- matrix(NA, (r + 1), Ntr)
                                } else {
                                    lambda.lv <- matrix(NA, r, Ntr)
                                }
                                test <- try(
                                    for (i.tr in 1:Ntr) {
                                        F.lv.pre <- as.matrix(F.hat[setdiff(time.pre[[i.tr]], lv), ])
                                        lambda.lv[, i.tr] <- solve(t(F.lv.pre) %*% F.lv.pre) %*%
                                            t(F.lv.pre) %*% as.matrix(U.lv[[i.tr]])
                                    },
                                    silent = TRUE
                                )
                                if ("try-error" %in% class(test)) {
                                    break
                                }
                            }
                        }
                        lambda.lv <- t(lambda.lv)
                        e <- U.tr[which(time == lv), ] - c(F.hat[which(time == lv), ] %*% t(lambda.lv))
                    }
                    if (sameT0 == FALSE | 0 %in% I.tr) {
                        e <- e[which(pre[which(time == lv), ] == TRUE)]
                    }
                    ## accumulate residuals
                    resid_all <- c(resid_all, e)
                } ## end of leave-one-out

                if (length(resid_all) == 0) {
                    scores <- c(MSPE = Inf, WMSPE = Inf, GMSPE = Inf, WGMSPE = Inf,
                                MAD = Inf, Moment = Inf, GMoment = Inf, RMSE = Inf, Bias = Inf)
                } else {
                    ## Build time indices for LOO residuals (CFE)
                    time_idx_loo <- NULL
                    obs_w_loo <- NULL
                    if (!is.null(count.T.cv)) {
                        time_idx_loo <- c()
                        for (lv in unique(unlist(time.pre))) {
                            if (sameT0 == FALSE | 0 %in% I.tr) {
                                n_resid_lv <- sum(pre[which(time == lv), ] == TRUE)
                            } else {
                                n_resid_lv <- Ntr
                            }
                            if (n_resid_lv > 0) {
                                t.on.tr <- T.on[, tr, drop = FALSE]
                                t.on.lv <- unique(t.on.tr[lv, ])
                                t.on.lv <- t.on.lv[!is.na(t.on.lv)]
                                if (length(t.on.lv) > 0) {
                                    time_idx_loo <- c(time_idx_loo, rep(as.character(t.on.lv[1]), n_resid_lv))
                                } else {
                                    time_idx_loo <- c(time_idx_loo, rep("Control", n_resid_lv))
                                }
                            }
                        }
                    }
                    if (!is.null(W.tr)) {
                        obs_w_loo <- c()
                        for (lv in unique(unlist(time.pre))) {
                            if (sameT0 == FALSE | 0 %in% I.tr) {
                                w_lv <- W.tr[lv, which(pre[which(time == lv), ] == TRUE)]
                            } else {
                                w_lv <- W.tr[lv, ]
                            }
                            obs_w_loo <- c(obs_w_loo, w_lv)
                        }
                    }
                    scores <- .score_residuals(
                        resid_all,
                        obs_weights = obs_w_loo,
                        time_index = time_idx_loo,
                        count_weights = count.T.cv,
                        norm.para = norm.para
                    )
                }

              } else if (cv.method %in% c("all_units", "rolling")) {
                ## ---- cv.sample "all_units" / "rolling" CFE CV (serial path — lapply only) ---- ##
                ## Rolling reuses the all_units CFE scoring helper; only rmCV/estCV
                ## fold construction differs (built via .build_cv_mask_rolling above).
                fold_results <- lapply(1:k, function(ii) {
                    .fect_cv_score_one_cfe_nt_all(
                        ii              = ii,
                        YY.co           = YY.co,
                        Y0CV.co         = Y0CV.co,
                        X.co            = X.co,
                        II.co           = II.co,
                        W.use           = W.use,
                        W               = W,
                        beta0CV.co      = beta0CV.co,
                        X.extra.FE.co.B = X.extra.FE.co.B,
                        X.Z.co          = X.Z.co,
                        X.Q.co          = X.Q.co,
                        X.gamma.co      = X.gamma.co,
                        X.kappa.co      = X.kappa.co,
                        Zgamma.id       = Zgamma.id,
                        kappaQ.id       = kappaQ.id,
                        rmCV            = rmCV,
                        estCV           = estCV,
                        r               = r,
                        force           = force,
                        cv_tol          = cv_tol,
                        max.iteration   = max.iteration
                    )
                })
                all_resid    <- unlist(lapply(fold_results, `[[`, "resid"))
                all_time_idx <- unlist(lapply(fold_results, `[[`, "time_idx"))
                all_obs_w    <- if (!is.null(W)) unlist(lapply(fold_results, `[[`, "obs_w")) else c()
                if (length(all_resid) == 0) {
                    scores <- c(MSPE = Inf, WMSPE = Inf, GMSPE = Inf, WGMSPE = Inf,
                                MAD = Inf, Moment = Inf, GMoment = Inf, RMSE = Inf, Bias = Inf)
                } else {
                    scores <- .score_residuals(
                        all_resid,
                        obs_weights = if (!is.null(W)) all_obs_w else NULL,
                        time_index = all_time_idx,
                        count_weights = count.T.cv,
                        norm.para = NULL
                    )
                }

              } else {
                ## ---- cv.sample "treated_units" CFE CV (serial path — lapply only) ---- ##
                ## Build U.tr: subtract Layer 1 from Y.tr
                if (r != 0) {
                    F.hat <- as.matrix(est.co$factor)
                    if (force %in% c(1, 3)) {
                        F.hat <- cbind(F.hat, rep(1, TT))
                    }
                }
                U.tr <- Y.tr
                if (p > 0) {
                    for (j in 1:p) {
                        U.tr <- U.tr - X.tr[, , j] * beta[j]
                    }
                }
                if (force != 0) {
                    U.tr <- U.tr - matrix(est.co$mu, TT, Ntr)
                }
                if (force %in% c(2, 3)) {
                    U.tr <- U.tr - matrix(est.co$xi, TT, Ntr, byrow = FALSE)
                }

                ## Subtract gamma for treated (Layer 1)
                if (!is.null(est.co$gamma) && length(est.co$gamma) > 0) {
                    for (k_g in seq_along(est.co$gamma)) {
                        gamma.fit.tr.k <- .reconstruct_gamma_fit_tr(
                            est.co$gamma[[k_g]], X.Z.tr, X.gamma.tr[, , k_g, drop = FALSE],
                            Zgamma.id[[k_g]], TT, Ntr)
                        U.tr <- U.tr - gamma.fit.tr.k
                    }
                }

                ## Subtract Type-B extra FE for treated (Layer 1)
                if (length(typeB_idx) > 0) {
                    typeB.fit.tr <- .extract_and_apply_typeB_fe(
                        est.co, X.co, X.extra.FE.co.B, X.extra.FE.tr,
                        typeB_idx, X.Z.co, X.gamma.co, X.Q.co, X.kappa.co,
                        Zgamma.id, kappaQ.id,
                        TT, Nco, Ntr, p, r, force)
                    U.tr <- U.tr - typeB.fit.tr
                }

                if (0 %in% I.tr) {
                    U.tr[which(I.tr == 0)] <- 0
                }

                fold_results <- lapply(1:k, function(ii) {
                    .fect_cv_score_one_cfe_nt_tr(
                        ii       = ii,
                        U.tr     = U.tr,
                        F.hat    = F.hat,
                        pre      = pre,
                        r        = r,
                        force    = force,
                        rmCV.tr  = rmCV.tr,
                        estCV.tr = estCV.tr,
                        W.tr     = W.tr,
                        T.on     = T.on,
                        tr       = tr,
                        TT       = TT,
                        Ntr      = Ntr
                    )
                })
                all_resid    <- unlist(lapply(fold_results, `[[`, "resid"))
                all_time_idx <- unlist(lapply(fold_results, `[[`, "time_idx"))
                all_obs_w    <- if (!is.null(W.tr)) unlist(lapply(fold_results, `[[`, "obs_w")) else c()
                if (length(all_resid) == 0) {
                    scores <- c(MSPE = Inf, WMSPE = Inf, GMSPE = Inf, WGMSPE = Inf,
                                MAD = Inf, Moment = Inf, GMoment = Inf, RMSE = Inf, Bias = Inf)
                } else {
                    scores <- .score_residuals(
                        all_resid,
                        obs_weights = if (!is.null(W.tr)) all_obs_w else NULL,
                        time_index = if (length(all_time_idx) > 0) all_time_idx else NULL,
                        count_weights = count.T.cv,
                        norm.para = norm.para
                    )
                }

              } ## end cv.method branching

                if ((min(CV.out[, crit_col]) - scores[crit_col]) > 0.01 * min(CV.out[, crit_col])) {
                    est.co.best <- est.co
                    r.cv <- r
                } else {
                    if (r == r.cv + 1) message("*")
                }

                if (PC < min(CV.out[, "PC"])) {
                    r.pc <- r
                    est.co.pc.best <- est.co
                }
                CV.out[i, 2:4] <- c(sigma2, IC, PC)
                CV.out[i, score_names] <- scores[score_names]
                message("r = ", r, "; sigma2 = ",
                    sprintf("%.5f", sigma2), "; IC = ",
                    sprintf("%.5f", IC), "; PC = ",
                    sprintf("%.5f", PC), "; MSPE = ",
                    sprintf("%.5f", scores["MSPE"]),
                    sep = ""
                )
            } ## end of CV loop

            } ## end SERIAL BRANCH (CFE)

            MSPE.best <- min(CV.out[, "MSPE"])
            if (r > (T0.min - 1)) {
                message(" (r hits maximum)")
            }
            message("\n r* = ", r.cv, sep = "")
            message("\n")
        }
    } else {
        r.cv <- r
        r.min <- r.max <- r
    }

    ## ---- Final fit with selected r.cv ----
    est.co.best <- complex_fe_ub(YY.co, Y0.co, X.co,
        X.extra.FE.co.B, X.Z.co, X.Q.co, X.gamma.co, X.kappa.co,
        Zgamma.id, kappaQ.id,
        II.co, W.use, beta0, r.cv, force = force, tol, max.iteration,
        fit_init = fit.init.co)

    ## Convergence check
    if (!is.null(est.co.best$niter) && est.co.best$niter >= max.iteration) {
        warning(paste0("CFE optimization did not converge within ", max.iteration,
            " iterations. Results may be unreliable."))
    }

    est.co.fect <- NULL
    if (boot == FALSE) {
        if (r.cv == 0) {
            est.co.fect <- est.co.best
        } else {
            est.co.fect <- complex_fe_ub(YY.co, Y0.co, X.co,
                X.extra.FE.co.B, X.Z.co, X.Q.co, X.gamma.co, X.kappa.co,
                Zgamma.id, kappaQ.id,
                II.co, W.use, beta0, 0, force = force, tol, max.iteration)
        }
    }

    validX <- est.co.best$validX
    validF <- ifelse(r.cv > 0, 1, 0)

    ## ---- Three-Layer Projection ----
    ## Layer 1: subtract shared parameters from Y.tr
    U.tr.r0 <- U.tr <- Y.tr
    if (p > 0) {
        beta <- est.co.best$beta
        if (est.co.best$validX == 0) {
            beta <- matrix(0, p, 1)
        } else {
            beta <- est.co.best$beta
            beta[is.nan(est.co.best$beta)] <- 0
        }
        for (j in 1:p) {
            U.tr <- U.tr - X.tr[, , j] * beta[j]
        }
        if (boot == FALSE) {
            beta.r0 <- est.co.fect$beta
            if (est.co.fect$validX == 0) {
                beta.r0 <- matrix(0, p, 1)
            } else {
                beta.r0 <- est.co.fect$beta
                beta.r0[is.nan(est.co.fect$beta)] <- 0
            }
            for (j in 1:p) {
                U.tr.r0 <- U.tr.r0 - X.tr[, , j] * beta.r0[j]
            }
        }
    } else {
        beta <- NA
        beta.r0 <- NA
    }

    mu <- est.co.best$mu
    U.tr <- U.tr - matrix(mu, TT, Ntr)
    Y.fe.bar <- rep(mu, TT)

    if (boot == FALSE) {
        mu.r0 <- est.co.fect$mu
        U.tr.r0 <- U.tr.r0 - matrix(mu.r0, TT, Ntr)
        Y.fe.bar.r0 <- rep(mu.r0, TT)
    }

    if (force %in% c(2, 3)) {
        xi <- est.co.best$xi
        U.tr <- U.tr - matrix(c(xi), TT, Ntr, byrow = FALSE)
        Y.fe.bar <- Y.fe.bar + xi
        if (boot == FALSE) {
            xi.r0 <- est.co.fect$xi
            U.tr.r0 <- U.tr.r0 - matrix(c(xi.r0), TT, Ntr, byrow = FALSE)
            Y.fe.bar.r0 <- Y.fe.bar.r0 + xi.r0
        }
    }

    ## Subtract gamma for treated (Layer 1)
    if (!is.null(est.co.best$gamma) && length(est.co.best$gamma) > 0) {
        for (k_g in seq_along(est.co.best$gamma)) {
            gamma.fit.tr.k <- .reconstruct_gamma_fit_tr(
                est.co.best$gamma[[k_g]], X.Z.tr, X.gamma.tr[, , k_g, drop = FALSE],
                Zgamma.id[[k_g]], TT, Ntr)
            U.tr <- U.tr - gamma.fit.tr.k
        }
    }

    ## Subtract Type-B extra FE for treated (Layer 1)
    if (length(typeB_idx) > 0) {
        typeB.fit.tr <- .extract_and_apply_typeB_fe(
            est.co.best, X.co, X.extra.FE.co.B, X.extra.FE.tr,
            typeB_idx, X.Z.co, X.gamma.co, X.Q.co, X.kappa.co,
            Zgamma.id, kappaQ.id,
            TT, Nco, Ntr, p, r.cv, force)
        U.tr <- U.tr - typeB.fit.tr
    }

    ## Layer 2: estimate alpha, kappa, Type-A FE, lambda from pre-treatment
    ## Save Layer 1 residuals (before any Layer 2 subtractions)
    U.tr.L1 <- U.tr

    has_kappa <- !is.null(est.co.best$kappa) && length(est.co.best$kappa) > 0

    ## --- Helper: compute kappa_fit (TT x Ntr) from current residuals ---
    .estimate_kappa_fit <- function(U.cur) {
        kappa_fit <- matrix(0, TT, Ntr)
        if (!has_kappa) return(kappa_fit)
        for (k_k in seq_along(est.co.best$kappa)) {
            q_cols <- kappaQ.id[[k_k]]
            kappa_groups <- X.kappa.tr[1, , k_k]
            unique_kgroups <- sort(unique(kappa_groups))
            for (g_idx in seq_along(unique_kgroups)) {
                g <- unique_kgroups[g_idx]
                units_in_group <- which(kappa_groups == g)
                Q.pre.list <- list()
                U.pre.list <- list()
                for (i_idx in seq_along(units_in_group)) {
                    ii <- units_in_group[i_idx]
                    pre_t <- which(pre[, ii])
                    if (length(pre_t) > 0) {
                        Q.mat <- matrix(0, length(pre_t), length(q_cols))
                        for (jj in seq_along(q_cols)) {
                            Q.mat[, jj] <- X.Q.tr[pre_t, ii, q_cols[jj]]
                        }
                        Q.pre.list[[length(Q.pre.list) + 1]] <- Q.mat
                        U.pre.list[[length(U.pre.list) + 1]] <- U.cur[pre_t, ii]
                    }
                }
                if (length(U.pre.list) > 0) {
                    Q.pre.all <- do.call(rbind, Q.pre.list)
                    U.pre.all <- unlist(U.pre.list)
                    if (length(U.pre.all) > length(q_cols)) {
                        kappa.hat <- try(solve(t(Q.pre.all) %*% Q.pre.all) %*%
                            t(Q.pre.all) %*% U.pre.all, silent = TRUE)
                        if (!"try-error" %in% class(kappa.hat)) {
                            for (ii in units_in_group) {
                                Q.full <- matrix(0, TT, length(q_cols))
                                for (jj in seq_along(q_cols)) {
                                    Q.full[, jj] <- X.Q.tr[, ii, q_cols[jj]]
                                }
                                kappa_fit[, ii] <- kappa_fit[, ii] + Q.full %*% kappa.hat
                            }
                        }
                    }
                }
            }
        }
        return(kappa_fit)
    }

    ## --- Helper: compute alpha (Ntr x 1) from current residuals ---
    .estimate_alpha <- function(U.cur) {
        if (!force %in% c(1, 3)) return(matrix(0, Ntr, 1))
        if (max(T0) == T0.min & (!0 %in% I.tr)) {
            return(as.matrix(colMeans(U.cur[1:T0.min, ])))
        } else {
            U.pre.v <- as.vector(U.cur)[which(pre.v == 1)]
            U.pre.l <- split(U.pre.v, id.tr.pre.v)
            return(as.matrix(sapply(U.pre.l, mean)))
        }
    }

    ## --- Helper: estimate Type-A extra FE fit (TT x Ntr) from residuals ---
    .estimate_typeA_fit <- function(U.cur) {
        fe_fit <- matrix(0, TT, Ntr)
        if (length(typeA_idx) == 0) return(fe_fit)
        for (k_a in typeA_idx) {
            labels.tr <- X.extra.FE.tr[1, , k_a]
            unique_levels <- sort(unique(labels.tr))
            for (g in unique_levels) {
                units_in_level <- which(labels.tr == g)
                pre_vals <- c()
                for (ii in units_in_level) {
                    pre_t <- which(pre[, ii])
                    pre_vals <- c(pre_vals, U.cur[pre_t, ii])
                }
                if (length(pre_vals) > 0) {
                    fe_mean <- mean(pre_vals)
                    for (ii in units_in_level) {
                        fe_fit[, ii] <- fe_fit[, ii] + fe_mean
                    }
                }
            }
        }
        return(fe_fit)
    }

    ## --- Helper: estimate lambda fit (TT x Ntr) from residuals ---
    ## When force %in% c(1,3), F.hat.aug includes intercept column;
    ## alpha is embedded as the last column of lambda.
    ## Returns list(fit = TT x Ntr, lambda = Ntr x ncol(F.hat.aug))
    .estimate_lambda_fit <- function(U.cur, F.hat.aug) {
        ncol_f <- ncol(F.hat.aug)
        if (max(T0) == T0.min & (!0 %in% I.tr)) {
            F.pre <- F.hat.aug[1:T0.min, , drop = FALSE]
            U.pre <- as.matrix(U.cur[1:T0.min, ])
            lam <- try(solve(t(F.pre) %*% F.pre) %*% t(F.pre) %*% U.pre,
                       silent = TRUE)
            if ("try-error" %in% class(lam)) {
                return(list(fit = matrix(0, TT, Ntr),
                            lambda = matrix(0, Ntr, ncol_f), ok = FALSE))
            }
        } else if (!0 %in% I.tr) {
            lam <- try(as.matrix(sapply(seq_len(Ntr), function(j) {
                pre_t <- which(pre[, j])
                F.pre <- as.matrix(F.hat.aug[pre_t, , drop = FALSE])
                solve(t(F.pre) %*% F.pre) %*% t(F.pre) %*% U.cur[pre_t, j]
            })), silent = TRUE)
            if ("try-error" %in% class(lam)) {
                return(list(fit = matrix(0, TT, Ntr),
                            lambda = matrix(0, Ntr, ncol_f), ok = FALSE))
            }
            if (ncol_f == 1) lam <- t(lam)
        } else {
            lam <- matrix(NA, ncol_f, Ntr)
            test <- try(
                for (i.tr in 1:Ntr) {
                    F.pre <- as.matrix(F.hat.aug[time.pre[[i.tr]], , drop = FALSE])
                    lam[, i.tr] <- solve(t(F.pre) %*% F.pre) %*%
                        t(F.pre) %*% as.matrix(U.cur[time.pre[[i.tr]], i.tr])
                }, silent = TRUE)
            if ("try-error" %in% class(test)) {
                return(list(fit = matrix(0, TT, Ntr),
                            lambda = matrix(0, Ntr, ncol_f), ok = FALSE))
            }
        }
        lam_mat <- t(lam)  ## Ntr x ncol_f
        fit <- F.hat.aug %*% t(lam_mat)  ## TT x Ntr
        return(list(fit = fit, lambda = lam_mat, ok = TRUE))
    }

    ## ============================================================
    ## Block coordinate descent: jointly estimate all unit-specific
    ## parameters (alpha, kappa, Type-A FE, lambda) from
    ## pre-treatment residuals. Mirrors C++ cfe_iter logic.
    ## ============================================================
    has_alpha  <- force %in% c(1, 3)
    has_typeA  <- length(typeA_idx) > 0
    has_factor <- r.cv > 0

    ## Build augmented factor matrix (factors + intercept for alpha)
    F.hat.aug <- NULL
    if (has_factor) {
        F.hat.aug <- as.matrix(est.co.best$factor)
        if (has_alpha) F.hat.aug <- cbind(F.hat.aug, rep(1, TT))
    }

    ## Initialize all component fits to zero
    ## NOTE: When has_factor && has_alpha, alpha is embedded in lambda_fit
    ## (via the augmented intercept column in F.hat.aug). In that case,
    ## alpha_mat is NOT used in residual computation to avoid double subtraction.
    ## alpha_mat is only used when has_alpha && !has_factor.
    alpha_mat  <- matrix(0, TT, Ntr)   ## alpha broadcast to TT x Ntr
    kappa_fit  <- matrix(0, TT, Ntr)
    typeA_fit  <- matrix(0, TT, Ntr)
    lambda_fit <- matrix(0, TT, Ntr)   ## F %*% t(lambda), includes alpha when augmented
    alpha.tr   <- matrix(0, Ntr, 1)
    lambda.tr  <- NULL
    lambda.co  <- NULL

    ## When has_factor && has_alpha, alpha lives inside lambda_fit.
    ## alpha_mat is separate only when !has_factor.
    alpha_in_lambda <- has_factor && has_alpha

    n_components <- has_kappa + has_typeA + (has_factor || has_alpha)

    if (n_components >= 2) {
        ## Multiple components: iterate to convergence
        max_iter_bcd <- 100
        tol_bcd <- 1e-8

        for (iter_bcd in 1:max_iter_bcd) {
            old_kappa  <- kappa_fit
            old_typeA  <- typeA_fit
            old_alpha  <- alpha_mat
            old_lambda <- lambda_fit

            ## Step 1: estimate kappa from residual
            if (has_kappa) {
                if (alpha_in_lambda) {
                    resid <- U.tr.L1 - typeA_fit - lambda_fit
                } else {
                    resid <- U.tr.L1 - alpha_mat - typeA_fit - lambda_fit
                }
                kappa_fit <- .estimate_kappa_fit(resid)
            }

            ## Step 2: estimate Type-A FE from residual
            if (has_typeA) {
                if (alpha_in_lambda) {
                    resid <- U.tr.L1 - kappa_fit - lambda_fit
                } else {
                    resid <- U.tr.L1 - alpha_mat - kappa_fit - lambda_fit
                }
                typeA_fit <- .estimate_typeA_fit(resid)
            }

            ## Step 3: estimate alpha (+lambda if r.cv > 0)
            if (has_factor) {
                ## alpha embedded in augmented F.hat → lambda_fit includes alpha
                resid <- U.tr.L1 - kappa_fit - typeA_fit
                result <- .estimate_lambda_fit(resid, F.hat.aug)
                if (!result$ok) {
                    return(list(att = rep(NA, TT), att.avg = NA,
                                beta = matrix(NA, p, 1)))
                }
                lambda_fit <- result$fit
                lambda.tr <- result$lambda
                if (has_alpha) {
                    alpha.tr <- as.matrix(lambda.tr[, ncol(F.hat.aug), drop = FALSE])
                    ## Do NOT update alpha_mat — alpha is inside lambda_fit
                }
            } else if (has_alpha) {
                resid <- U.tr.L1 - kappa_fit - typeA_fit
                alpha.tr <- .estimate_alpha(resid)
                alpha_mat <- matrix(alpha.tr, TT, Ntr, byrow = TRUE)
            }

            ## Check convergence of all components
            delta <- max(
                max(abs(kappa_fit  - old_kappa)),
                max(abs(typeA_fit  - old_typeA)),
                if (alpha_in_lambda) 0 else max(abs(alpha_mat - old_alpha)),
                max(abs(lambda_fit - old_lambda))
            )
            if (delta < tol_bcd) break
        }
    } else {
        ## Single component (or none): one-pass estimation
        if (has_kappa) {
            kappa_fit <- .estimate_kappa_fit(U.tr.L1)
        }
        if (has_factor) {
            resid <- U.tr.L1 - kappa_fit - typeA_fit
            result <- .estimate_lambda_fit(resid, F.hat.aug)
            if (!result$ok) {
                return(list(att = rep(NA, TT), att.avg = NA,
                            beta = matrix(NA, p, 1)))
            }
            lambda_fit <- result$fit
            lambda.tr <- result$lambda
            if (has_alpha) {
                alpha.tr <- as.matrix(lambda.tr[, ncol(F.hat.aug), drop = FALSE])
            }
        } else if (has_alpha) {
            resid <- U.tr.L1 - kappa_fit - typeA_fit
            alpha.tr <- .estimate_alpha(resid)
            alpha_mat <- matrix(alpha.tr, TT, Ntr, byrow = TRUE)
        }
        if (has_typeA) {
            if (alpha_in_lambda) {
                resid <- U.tr.L1 - kappa_fit - lambda_fit
            } else {
                resid <- U.tr.L1 - alpha_mat - kappa_fit - lambda_fit
            }
            typeA_fit <- .estimate_typeA_fit(resid)
        }
    }

    ## Final residual = treatment effect
    if (alpha_in_lambda) {
        ## alpha is inside lambda_fit — don't subtract alpha_mat
        U.tr <- U.tr.L1 - kappa_fit - typeA_fit - lambda_fit
    } else {
        U.tr <- U.tr.L1 - alpha_mat - kappa_fit - typeA_fit - lambda_fit
    }
    eff <- U.tr

    ## Extract clean lambda.tr (remove alpha column if embedded)
    if (has_factor && has_alpha && !is.null(lambda.tr)) {
        alpha.tr <- as.matrix(lambda.tr[, ncol(F.hat.aug), drop = FALSE])
        lambda.tr <- lambda.tr[, 1:r.cv, drop = FALSE]
    } else if (has_factor && !is.null(lambda.tr)) {
        ## lambda.tr already clean
    } else {
        lambda.tr <- NULL
    }
    lambda.co <- if (has_factor) est.co.best$lambda else NULL

    ## Implied weights
    if (has_factor && boot == 0 && !is.null(lambda.tr)) {
        inv.tr <- try(ginv(t(as.matrix(lambda.tr))), silent = TRUE)
        if (!"try-error" %in% class(inv.tr)) {
            wgt.implied <- t(inv.tr %*% t(as.matrix(est.co.best$lambda)))
        }
    }

    ## r=0 path (for equivalence test baseline — uses FE-only model)
    if (force %in% c(1, 3)) {
        if (boot == FALSE) {
            if ((max(T0) == T0.min) & (!0 %in% I.tr)) {
                alpha.tr.r0 <- as.matrix(colMeans(as.matrix(U.tr.r0[1:T0.min, ])))
            } else {
                U.tr.pre.v.r0 <- as.vector(U.tr.r0)[which(pre.v == 1)]
                U.tr.pre.r0 <- split(U.tr.pre.v.r0, id.tr.pre.v)
                alpha.tr.r0 <- as.matrix(sapply(U.tr.pre.r0, mean))
            }
            U.tr.r0 <- U.tr.r0 - matrix(alpha.tr.r0, TT, Ntr, byrow = TRUE)
        }
    }
    if (boot == FALSE) {
        eff.r0 <- U.tr.r0
    }

    if (0 %in% I.tr) {
        eff[which(I.tr == 0)] <- 0
        if (boot == FALSE) {
            eff.r0[which(I.tr == 0)] <- 0
        }
    }

  } ## end of method bifurcation

    ## -------------------------------##
    ## Summarize
    ## -------------------------------##

    ## counterfactuals for treated units
    Y.ct.tr <- as.matrix(Y.tr - eff)
    Y.ct.co <- Y.co - est.co.best$residuals
    # print(Y.ct.co)
    Y.ct <- Y
    Y.ct[, tr] <- Y.ct.tr
    Y.ct[, co] <- Y.ct.co

    if (boot == FALSE) {
        Y.ct.tr.r0 <- as.matrix(Y.tr - eff.r0)
        Y.ct.co.r0 <- Y.co - est.co.fect$residuals
        Y.ct.r0 <- Y
        Y.ct.r0[, tr] <- Y.ct.tr.r0
        Y.ct.r0[, co] <- Y.ct.co.r0
    }

    ## we first adjustment for normalization
    if (!is.null(norm.para)) {
        Y <- Y * norm.para[1]
        ## variance of the error term
        sigma2 <- est.co.best$sigma2 <- est.co.best$sigma2 * (norm.para[1]^2)
        IC <- est.co.best$IC <- est.co.best$IC - log(est.co.best$sigma2) + log(sigma2)
        PC <- est.co.best$PC <- est.co.best$PC * (norm.para[1]^2)

        ## output of estimates
        mu <- est.co.best$mu <- est.co.best$mu * norm.para[1]
        if (r.cv > 0) {
            est.co.best$lambda <- est.co.best$lambda * norm.para[1]
            lambda.tr <- lambda.tr * norm.para[1]
            est.co.best$VNT <- est.co.best$VNT * norm.para[1]
        }
        if (force %in% c(1, 3)) {
            est.co.best$alpha <- est.co.best$alpha * norm.para[1]
            alpha.tr <- alpha.tr * norm.para[1]
        }
        if (force %in% c(2, 3)) {
            xi <- est.co.best$xi <- est.co.best$xi * norm.para[1]
        }
        est.co.best$residuals <- est.co.best$residuals * norm.para[1]
        est.co.best$fit <- est.co.best$fit * norm.para[1]
        if (boot == FALSE) {
            est.co.fect$fit <- est.co.fect$fit * norm.para[1]
        }
        est.co.fect$sigma2 <- est.co.fect$sigma2 * norm.para[1]

        # estimated counterfactual
        Y.tr <- Y.tr * norm.para[1]
        Y.co <- Y.co * norm.para[1]

        Y.ct <- Y.ct * norm.para[1]
        Y.ct.tr <- Y.ct.tr * norm.para[1]
        Y.ct.co <- Y.ct.co * norm.para[1]
        eff <- eff * norm.para[1]

        if (boot == FALSE) {
            Y.ct.r0 <- Y.ct.r0 * norm.para[1]
            Y.ct.tr.r0 <- Y.ct.tr.r0 * norm.para[1]
            Y.ct.co.r0 <- Y.ct.co.r0 * norm.para[1]
            eff.r0 <- eff.r0 * norm.para[1]
        }
    }

    ## 0. relevant parameters
    IC <- est.co.best$IC
    sigma2 <- est.co.best$sigma2
    PC <- est.co.best$PC
    loglikelihood <- NULL

    if (p > 0) {
        na.pos <- is.nan(est.co.best$beta)
        beta <- est.co.best$beta
        if (sum(na.pos) > 0) {
            beta[na.pos] <- NA
        }
    } else {
        beta <- NA
    }

    ## 1. estimated att and counterfactuals
    if (boot == FALSE) {
        Y.ct.equiv <- Y.ct.r0
    } else {
        Y.ct.equiv <- NULL
    }

    eff <- Y - Y.ct

    missing.index <- which(is.na(eff))
    if (length(missing.index) > 0) {
        I[missing.index] <- 0
        II[missing.index] <- 0
    }
    if (0 %in% I) {
        eff[which(I == 0)] <- NA
    }
    complete.index <- which(!is.na(eff))
    att.avg <- sum(eff[complete.index] * D[complete.index]) / (sum(D[complete.index]))
    marginal <- NULL

    att.avg.balance <- NA
    if (!is.null(balance.period)) {
        complete.index2 <- which(!is.na(T.on.balance))
        att.avg.balance <- sum(eff[complete.index2] * D[complete.index2]) / (sum(D[complete.index2]))
    }

    # weighted effect
    att.avg.W <- NA
    if (!is.null(W)) {
        att.avg.W <- sum(eff[complete.index] * D[complete.index] * W[complete.index]) / (sum(D[complete.index] * W[complete.index]))
    }

    ## att.avg.unit
    tr.pos <- which(apply(D, 2, sum) > 0)
    att.unit <- sapply(1:length(tr.pos), function(vec) {
        return(sum(eff[, tr.pos[vec]] * D[, tr.pos[vec]]) / sum(D[, tr.pos[vec]]))
    })
    att.avg.unit <- mean(att.unit, na.rm = TRUE)

    equiv.att.avg <- eff.equiv <- NULL
    if (boot == FALSE) {
        eff.equiv <- Y - Y.ct.equiv
        if (0 %in% I) {
            eff.equiv[which(I == 0)] <- NA
        }
        complete.index <- which(!is.na(eff.equiv))
        equiv.att.avg <- sum(eff.equiv[complete.index] * D[complete.index]) / (sum(D[complete.index]))
    }

    ## 2. rmse for treated units' observations under control
    if (binary == 0) {
        tr <- which(apply(D, 2, sum) > 0)
        tr.co <- which((as.matrix(1 - D[, tr]) * as.matrix(II[, tr])) == 1)
        eff.tr <- as.matrix(eff[, tr])
        v.eff.tr <- eff.tr[tr.co]
        rmse <- sqrt(mean(v.eff.tr^2, na.rm = TRUE))
    }

    ## 3. unbalanced output
    Y.ct.full <- Y.ct
    ## Stage 2: populate Y.ct.full at control positions on the GSC path.
    ## On the IFE-EM (notyettreated) path EM imputation already fills in
    ## masked control positions, so user-space rolling CV can read off
    ## fit$Y.ct.full there. On GSC the residual recipe `Y.co - residuals`
    ## leaves NA at masked positions because Y.co is NA. Overwriting with
    ## the model-implied F * t(lambda_co) closes that gap. ATT, gap, and
    ## est.avg are computed from treated units only and are unaffected.
    if (method == "ife" &&
        !is.null(est.co.best$factor) &&
        !is.null(est.co.best$lambda)) {
        F.hat.ctf     <- as.matrix(est.co.best$factor)
        Lambda.co.ctf <- as.matrix(est.co.best$lambda)
        if (ncol(F.hat.ctf) > 0 &&
            ncol(F.hat.ctf) == ncol(Lambda.co.ctf) &&
            nrow(F.hat.ctf) == TT && nrow(Lambda.co.ctf) == Nco) {
            Y.ct.full[, co] <- F.hat.ctf %*% t(Lambda.co.ctf)
        }
    }
    res.full <- Y - Y.ct
    if (0 %in% I) {
        eff[which(I == 0)] <- NA
        Y.ct[which(I == 0)] <- NA
    }
    if (binary == FALSE) {
        res.full[which(II == 0)] <- NA
    }


    ## 4. dynamic effects
    t.on <- c(T.on)
    eff.v <- c(eff) ## a vector
    eff.equiv.v <- NULL
    if (binary == FALSE && boot == FALSE) {
        eff.equiv.v <- c(eff.equiv)
    }

    rm.pos1 <- which(is.na(eff.v))
    rm.pos2 <- which(is.na(t.on))

    eff.v.use1 <- eff.v
    t.on.use <- t.on
    n.on.use <- rep(1:N, each = TT)

    if (NA %in% eff.v | NA %in% t.on) {
        eff.v.use1 <- eff.v[-c(rm.pos1, rm.pos2)]
        t.on.use <- t.on[-c(rm.pos1, rm.pos2)]
        n.on.use <- n.on.use[-c(rm.pos1, rm.pos2)]
        if (binary == FALSE && boot == FALSE) {
            eff.equiv.v <- eff.equiv.v[-c(rm.pos1, rm.pos2)]
        }
    }

    pre.pos <- which(t.on.use <= 0)
    eff.pre <- cbind(eff.v.use1[pre.pos], t.on.use[pre.pos], n.on.use[pre.pos])
    colnames(eff.pre) <- c("eff", "period", "unit")

    pre.sd <- eff.pre.equiv <- NULL
    if (binary == FALSE && boot == FALSE) {
        eff.pre.equiv <- cbind(eff.equiv.v[pre.pos], t.on.use[pre.pos], n.on.use[pre.pos])
        colnames(eff.pre.equiv) <- c("eff.equiv", "period", "unit")

        pre.sd <- tapply(eff.pre.equiv[, 1], eff.pre.equiv[, 2], sd)
        pre.sd <- cbind(pre.sd, sort(unique(eff.pre.equiv[, 2])), table(eff.pre.equiv[, 2]))
        colnames(pre.sd) <- c("sd", "period", "count")
    }

    time.on <- sort(unique(t.on.use))
    att.on <- as.numeric(tapply(eff.v.use1, t.on.use, mean)) ## NA already removed
    count.on <- as.numeric(table(t.on.use))

    if (!is.null(time.on.seq)) {
        count.on.med <- att.on.med <- rep(NA, length(time.on.seq))
        att.on.med[which(time.on.seq %in% time.on)] <- att.on
        count.on.med[which(time.on.seq %in% time.on)] <- count.on
        att.on <- att.on.med
        count.on <- count.on.med
        time.on <- time.on.seq
    }

    if (!is.null(W)) {
        W.v <- c(W)
        rm.pos.W <- which(is.na(W))
        if (NA %in% eff.v | NA %in% t.on | NA %in% W.v) {
            eff.v.use.W <- eff.v[-c(rm.pos1, rm.pos2, rm.pos.W)]
            W.v.use <- W.v[-c(rm.pos1, rm.pos2, rm.pos.W)]
            t.on.use.W <- t.on[-c(rm.pos1, rm.pos2, rm.pos.W)]
            n.on.use.W <- n.on.use[-c(rm.pos1, rm.pos2, rm.pos.W)]
        } else {
            eff.v.use.W <- eff.v.use1
            t.on.use.W <- t.on.use
            n.on.use.W <- n.on.use
            W.v.use <- W.v
        }
        time.on.W <- sort(unique(t.on.use.W))
        att.on.sum.W <- as.numeric(tapply(eff.v.use.W * W.v.use, t.on.use.W, sum)) ## NA already removed
        W.on.sum <- as.numeric(tapply(W.v.use, t.on.use.W, sum))
        att.on.W <- att.on.sum.W / W.on.sum
        count.on.W <- as.numeric(table(t.on.use.W))

        if (!is.null(time.on.seq.W)) {
            att.on.sum.med.W <- W.on.sum.med <- count.on.med.W <- att.on.med.W <- rep(NA, length(time.on.seq.W))
            att.on.sum.med.W[which(time.on.seq.W %in% time.on.W)] <- att.on.sum.W
            att.on.med.W[which(time.on.seq.W %in% time.on.W)] <- att.on.W
            count.on.med.W[which(time.on.seq.W %in% time.on.W)] <- count.on.W
            W.on.sum.med[which(time.on.seq.W %in% time.on.W)] <- W.on.sum
            att.on.sum.W <- att.on.sum.med.W
            att.on.W <- att.on.med.W
            count.on.W <- count.on.med.W
            time.on.W <- time.on.seq.W
            W.on.sum <- W.on.sum.med
        }
    } else {
        att.on.sum.med.W <- att.on.sum.W <- count.on.med.W <- att.on.med.W <- W.on.sum.med <- att.on.W <- count.on.W <- time.on.W <- W.on.sum <- NULL
    }

    ## 4.2 balance effect
    balance.att <- NULL
    if (!is.null(balance.period)) {
        t.on.balance <- c(T.on.balance)
        rm.pos4 <- which(is.na(t.on.balance))
        t.on.balance.use <- t.on.balance

        if (NA %in% eff.v | NA %in% t.on.balance) {
            eff.v.use3 <- eff.v[-c(rm.pos1, rm.pos4)]
            t.on.balance.use <- t.on.balance[-c(rm.pos1, rm.pos4)]
        }

        balance.time <- sort(unique(t.on.balance.use))
        balance.att <- as.numeric(tapply(eff.v.use3, t.on.balance.use, mean)) ## NA already removed
        balance.count <- as.numeric(table(t.on.balance.use))

        if (!is.null(time.on.balance.seq)) {
            balance.att.med <- rep(NA, length(time.on.balance.seq))
            balance.count.med <- rep(0, length(time.on.balance.seq))
            balance.att.med[which(time.on.balance.seq %in% balance.time)] <- balance.att
            if (length(balance.count) > 0) {
                balance.count.med[which(time.on.balance.seq %in% balance.time)] <- balance.count
            }
            balance.count <- balance.count.med
            balance.att <- balance.att.med
            balance.time <- time.on.balance.seq
        }

        # placebo for balanced samples
        if (!is.null(placebo.period) && placeboTest == 1) {
            if (length(placebo.period) == 1) {
                balance.placebo.pos <- which(balance.time == placebo.period)
                balance.att.placebo <- balance.att[balance.placebo.pos]
            } else {
                balance.placebo.pos <- which(balance.time >= placebo.period[1] & balance.time <= placebo.period[2])
                balance.att.placebo <- sum(balance.att[balance.placebo.pos] * balance.count[balance.placebo.pos]) / sum(balance.count[balance.placebo.pos])
            }
        }
    }

    ## 5. placebo effect, if placeboTest == 1
    if (!is.null(placebo.period) && placeboTest == 1) {
        if (length(placebo.period) == 1) {
            placebo.pos <- which(time.on == placebo.period)
            att.placebo <- att.on[placebo.pos]
        } else {
            placebo.pos <- which(time.on >= placebo.period[1] & time.on <= placebo.period[2])
            att.placebo <- sum(att.on[placebo.pos] * count.on[placebo.pos]) / sum(count.on[placebo.pos])
        }

        if (!is.null(W)) {
            if (length(placebo.period) == 1) {
                placebo.pos.W <- which(time.on.W == placebo.period)
                att.placebo.W <- att.on.W[placebo.pos.W]
            } else {
                placebo.pos.W <- which(time.on.W >= placebo.period[1] & time.on.W <= placebo.period[2])
                att.placebo.W <- sum(att.on.sum.W[placebo.pos.W]) / sum(W.on.sum[placebo.pos.W])
            }
        }
    }

    ## 6. switch-off effects
    eff.off.equiv <- off.sd <- eff.off <- NULL
    if (hasRevs == 1) {
        t.off <- c(T.off)
        rm.pos3 <- which(is.na(t.off))
        eff.v.use2 <- eff.v
        t.off.use <- t.off
        if (NA %in% eff.v | NA %in% t.off) {
            eff.v.use2 <- eff.v[-c(rm.pos1, rm.pos3)]
            t.off.use <- t.off[-c(rm.pos1, rm.pos3)]
        }

        off.pos <- which(t.off.use > 0)
        eff.off <- cbind(eff.v.use2[off.pos], t.off.use[off.pos], n.on.use[off.pos])
        colnames(eff.off) <- c("eff", "period", "unit")

        if (binary == FALSE && boot == FALSE) {
            eff.off.equiv <- cbind(eff.equiv.v[off.pos], t.off.use[off.pos], n.on.use[off.pos])
            colnames(eff.off.equiv) <- c("off.equiv", "period", "unit")

            off.sd <- tapply(eff.off.equiv[, 1], eff.off.equiv[, 2], sd)
            off.sd <- cbind(off.sd, sort(unique(eff.off.equiv[, 2])), table(eff.off.equiv[, 2]))
            colnames(off.sd) <- c("sd", "period", "count")
        }

        time.off <- sort(unique(t.off.use))

        att.off <- as.numeric(tapply(eff.v.use2, t.off.use, mean)) ## NA already removed
        count.off <- as.numeric(table(t.off.use))

        if (!is.null(time.off.seq)) {
            count.off.med <- att.off.med <- rep(NA, length(time.off.seq))
            att.off.med[which(time.off.seq %in% time.off)] <- att.off
            count.off.med[which(time.off.seq %in% time.off)] <- count.off
            att.off <- att.off.med
            count.off <- count.off.med
            time.off <- time.off.seq
        }

        if (!is.null(W)) {
            if (NA %in% eff.v | NA %in% t.off | NA %in% W.v) {
                eff.v.use2.W <- eff.v[-c(rm.pos1, rm.pos3, rm.pos.W)]
                W.v.use2 <- W.v[-c(rm.pos1, rm.pos3, rm.pos.W)]
                t.off.use.W <- t.off[-c(rm.pos1, rm.pos3, rm.pos.W)]
            } else {
                eff.v.use2.W <- eff.v.use2
                t.off.use.W <- t.off.use
                W.v.use2 <- W.v
            }

            time.off.W <- sort(unique(t.off.use.W))
            att.off.sum.W <- as.numeric(tapply(eff.v.use2.W * W.v.use2, t.off.use.W, sum))
            W.off.sum <- as.numeric(tapply(W.v.use2, t.off.use.W, sum))
            att.off.W <- att.off.sum.W / W.off.sum ## NA already removed
            count.off.W <- as.numeric(table(t.off.use.W))

            if (!is.null(time.off.seq.W)) {
                att.off.sum.med.W <- W.off.sum.med <- count.off.med.W <- att.off.med.W <- rep(NA, length(time.off.seq.W))
                att.off.sum.med.W[which(time.off.seq.W %in% time.off.W)] <- att.off.sum.W
                att.off.med.W[which(time.off.seq.W %in% time.off.W)] <- att.off.W
                count.off.med.W[which(time.off.seq.W %in% time.off.W)] <- count.off.W
                W.off.sum.med[which(time.off.seq.W %in% time.off.W)] <- W.off.sum
                att.off.sum.W <- att.off.sum.med.W
                att.off.W <- att.off.med.W
                count.off.W <- count.off.med.W
                time.off.W <- time.off.seq.W
                W.off.sum <- W.off.sum.med
            }
        } else {
            W.off.sum.med <- W.off.sum <- att.off.sum.W <- att.off.sum.med.W <- count.off.med.W <- att.off.med.W <- count.off.med.W <- att.off.W <- count.off.W <- time.off.W <- NULL
        }
    }

    ## 7. carryover effects
    if (!is.null(carryover.period) && carryoverTest == 1 && hasRevs == 1) {
        ## construct att.carryover
        ## eff is derived from eff.v
        ## period and Num.Units are derived from T.off
        if (length(carryover.period) == 1) {
            carryover.pos <- which(time.off == carryover.period)
            att.carryover <- att.off[carryover.pos]
        } else {
            carryover.pos <- which(time.off >= carryover.period[1] & time.off <= carryover.period[2])
            att.carryover <- sum(att.off[carryover.pos] * count.off[carryover.pos]) / sum(count.off[carryover.pos])
        }
    }

    ## 9. loess HTE by time
    D.missing <- D
    D.missing[which(D == 0)] <- NA
    eff.calendar <- apply(eff * D.missing, 1, mean, na.rm = TRUE)
    N.calendar <- apply(!is.na(eff * D.missing), 1, sum)
    T.calendar <- c(1:TT)
    if (sum(!is.na(eff.calendar)) > 1) {
        # loess fit
        if (!is.null(calendar.enp.seq)) {
            if (length(calendar.enp.seq) == 1 & is.na(calendar.enp.seq)) {
                calendar.enp.seq <- NULL
            }
        }
        if (is.null(calendar.enp.seq)) {
            loess.fit <- suppressWarnings(try(loess(eff.calendar ~ T.calendar, weights = N.calendar), silent = TRUE))
        } else {
            loess.fit <- suppressWarnings(try(loess(eff.calendar ~ T.calendar, weights = N.calendar, enp.target = calendar.enp.seq), silent = TRUE))
        }
        if ("try-error" %in% class(loess.fit)) {
            eff.calendar.fit <- eff.calendar
            calendar.enp <- NULL
        } else {
            eff.calendar.fit <- eff.calendar
            eff.calendar.fit[which(!is.na(eff.calendar))] <- loess.fit$fit
            calendar.enp <- loess.fit$enp
        }
    } else {
        eff.calendar.fit <- eff.calendar
        calendar.enp <- NULL
    }


    ## 8. cohort effects
    if (!is.null(group)) {
        cohort <- cbind(c(group), c(D), c(eff.v))
        rm.pos <- unique(c(rm.pos1, which(cohort[, 2] == 0)))
        cohort <- cohort[-rm.pos, ]

        g.level <- sort(unique(cohort[, 1]))
        raw.group.att <- as.numeric(tapply(cohort[, 3], cohort[, 1], mean))

        group.att <- rep(NA, length(group.level))
        group.att[which(group.level %in% g.level)] <- raw.group.att

        # by-group dynamic effects
        group.level.name <- names(group.level)

        group.output <- list()
        for (i in c(1:length(group.level))) {
            sub.group <- group.level[i]
            sub.group.name <- group.level.name[i]

            ## by-group dynamic effects
            t.on.sub <- c(T.on[which(group == sub.group)])
            eff.v.sub <- c(eff[which(group == sub.group)]) ## a vector
            rm.pos1.sub <- which(is.na(eff.v.sub))
            rm.pos2.sub <- which(is.na(t.on.sub))
            eff.v.use1.sub <- eff.v.sub
            t.on.use.sub <- t.on.sub
            if (NA %in% eff.v.sub | NA %in% t.on.sub) {
                eff.v.use1.sub <- eff.v.sub[-c(rm.pos1.sub, rm.pos2.sub)]
                t.on.use.sub <- t.on.sub[-c(rm.pos1.sub, rm.pos2.sub)]
            }
            if (length(t.on.use.sub) > 0) {
                time.on.sub <- sort(unique(t.on.use.sub))
                att.on.sub <- as.numeric(tapply(
                    eff.v.use1.sub,
                    t.on.use.sub,
                    mean
                )) ## NA already removed
                count.on.sub <- as.numeric(table(t.on.use.sub))
            } else {
                time.on.sub <- att.on.sub <- count.on.sub <- NULL
            }

            if (!is.null(time.on.seq.group)) {
                count.on.med.sub <- att.on.med.sub <- rep(NA, length(time.on.seq.group[[sub.group.name]]))
                time.on.seq.sub <- time.on.seq.group[[sub.group.name]]
                att.on.med.sub[which(time.on.seq.sub %in% time.on.sub)] <- att.on.sub
                count.on.med.sub[which(time.on.seq.sub %in% time.on.sub)] <- count.on.sub
                att.on.sub <- att.on.med.sub
                count.on.sub <- count.on.med.sub
                time.on.sub <- time.on.seq.sub
            }
            if (length(att.on.sub) == 0) {
                att.on.sub <- NULL
            }
            if (length(time.on.sub) == 0) {
                time.on.sub <- NULL
            }
            if (length(count.on.sub) == 0) {
                count.on.sub <- NULL
            }
            suboutput <- list(
                att.on = att.on.sub,
                time.on = time.on.sub,
                count.on = count.on.sub
            )

            ## placebo effect, if placeboTest == 1
            if (!is.null(placebo.period) && placeboTest == 1) {
                if (length(placebo.period) == 1) {
                    placebo.pos.sub <- which(time.on.sub == placebo.period)
                    if (length(placebo.pos.sub) > 0) {
                        att.placebo.sub <- att.on.sub[placebo.pos.sub]
                    } else {
                        att.placebo.sub <- NULL
                    }
                } else {
                    placebo.pos.sub <- which(time.on.sub >= placebo.period[1] & time.on.sub <= placebo.period[2])
                    if (length(placebo.pos.sub) > 0) {
                        att.placebo.sub <- sum(att.on.sub[placebo.pos.sub] * count.on.sub[placebo.pos.sub]) / sum(count.on.sub[placebo.pos.sub])
                    } else {
                        att.placebo.sub <- NULL
                    }
                }
                if (length(att.placebo.sub) == 0) {
                    att.placebo.sub <- NULL
                }
                suboutput <- c(suboutput, list(att.placebo = att.placebo.sub))
            }

            ## T.off
            if (hasRevs == 1) {
                t.off.sub <- c(T.off[which(group == sub.group)])
                rm.pos3.sub <- which(is.na(t.off.sub))
                eff.v.use2.sub <- eff.v.sub
                t.off.use.sub <- t.off.sub
                if (NA %in% eff.v.sub | NA %in% t.off.sub) {
                    eff.v.use2.sub <- eff.v.sub[-c(rm.pos1.sub, rm.pos3.sub)]
                    t.off.use.sub <- t.off.sub[-c(rm.pos1.sub, rm.pos3.sub)]
                }
                if (length(t.off.use.sub) > 0) {
                    time.off.sub <- sort(unique(t.off.use.sub))
                    att.off.sub <- as.numeric(tapply(eff.v.use2.sub, t.off.use.sub, mean)) ## NA already removed
                    count.off.sub <- as.numeric(table(t.off.use.sub))
                } else {
                    time.off.sub <- att.off.sub <- count.off.sub <- NULL
                }

                if (!is.null(time.off.seq.group)) {
                    count.off.med.sub <- att.off.med.sub <- rep(NA, length(time.off.seq.group[[sub.group.name]]))
                    time.off.seq.sub <- time.off.seq.group[[sub.group.name]]
                    att.off.med.sub[which(time.off.seq.sub %in% time.off.sub)] <- att.off.sub
                    count.off.med.sub[which(time.off.seq.sub %in% time.off.sub)] <- count.off.sub
                    att.off.sub <- att.off.med.sub
                    count.off.sub <- count.off.med.sub
                    time.off.sub <- time.off.seq.sub
                }
                if (length(att.off.sub) == 0) {
                    att.off.sub <- NULL
                }
                if (length(time.off.sub) == 0) {
                    time.off.sub <- NULL
                }
                if (length(count.off.sub) == 0) {
                    count.off.sub <- NULL
                }
                suboutput <- c(suboutput, list(
                    att.off = att.off.sub,
                    count.off = count.off.sub,
                    time.off = time.off.sub
                ))

                if (!is.null(carryover.period) && carryoverTest == 1) {
                    if (length(carryover.period) == 1) {
                        carryover.pos.sub <- which(time.off.sub == carryover.period)
                        if (length(carryover.pos.sub) > 0) {
                            att.carryover.sub <- att.off.sub[carryover.pos.sub]
                        } else {
                            att.carryover.sub <- NULL
                        }
                    } else {
                        carryover.pos.sub <- which(time.off.sub >= carryover.period[1] & time.off.sub <= carryover.period[2])
                        if (length(carryover.pos.sub) > 0) {
                            att.carryover.sub <- sum(att.off.sub[carryover.pos.sub] * count.off.sub[carryover.pos.sub]) / sum(count.off.sub[carryover.pos.sub])
                        } else {
                            att.carryover.sub <- NULL
                        }
                    }
                    if (length(att.carryover.sub) == 0) {
                        att.carryover.sub <- NULL
                    }
                    suboutput <- c(suboutput, list(att.carryover = att.carryover.sub))
                }
            }
            group.output[[sub.group.name]] <- suboutput
        }
    }

    ## Preserve method regardless of r.cv: the bootstrap dispatcher in fect_boot
    ## has branches for gsynth/ife/mc/cfe but not for "fe", so relabelling to
    ## "fe" when r.cv == 0 caused stop("Unsupported bootstrap method: fe") whenever
    ## CV selected the two-way FE model. The gsynth/cfe branches already handle
    ## r = 0 correctly on the estimation side (inside fect_nevertreated itself).
    if (method == "cfe") {
        method <- "cfe"
    } else {
        method <- "gsynth"
    }

    ## -------------------------------##
    ##            Storage            ##
    ## -------------------------------##
    out <- list(
        ## main results
        method = method,
        Y.ct = Y.ct,
        Y.tr.cnt = Y.ct.tr,
        Y.ct.cnt = Y.ct.co,
        Y.ct.full = Y.ct.full,
        Y = Y,
        D = D,
        tr = tr,
        co = co,
        eff = eff,
        eff.tr = eff[, tr],
        I = I,
        II = II,
        att.avg = att.avg,
        att.avg.unit = att.avg.unit,
        ## supporting
        force = force,
        T = TT,
        N = N,
        Ntr = Ntr,
        Nco = Nco,
        p = p,
        r.cv = r.cv,
        IC = IC,
        beta = beta,
        est = est.co.best,
        mu = est.co.best$mu,
        niter = est.co.best$niter,
        validX = validX,
        validF = validF,
        time = time.on,
        att = att.on,
        count = count.on,
        eff.calendar = eff.calendar,
        N.calendar = N.calendar,
        eff.calendar.fit = eff.calendar.fit,
        eff.pre = eff.pre,
        eff.pre.equiv = eff.pre.equiv,
        pre.sd = pre.sd,
        gamma = est.co.best$gamma,
        kappa = est.co.best$kappa
    )


    out <- c(out, list(
        PC = PC,
        sigma2 = sigma2,
        sigma2.fect = est.co.fect$sigma2,
        res = est.co.best$residuals,
        res.full = res.full,
        rmse = rmse
    ))


    if (hasRevs == 1) {
        out <- c(out, list(
            time.off = time.off,
            att.off = att.off,
            count.off = count.off,
            eff.off = eff.off,
            eff.off.equiv = eff.off.equiv,
            off.sd = off.sd
        ))
    }
    ## Include CV.out in the return list when cross-validation was performed
    if (exists("CV.out", inherits = FALSE)) {
        out <- c(out, list(CV.out = CV.out))
    }

    if (r.cv > 0) {
        lambda.co <- as.matrix(est.co.best$lambda)
        rownames(lambda.co) <- co
        rownames(lambda.tr) <- tr
        out <- c(out, list(
            factor = as.matrix(est.co.best$factor),
            lambda.co = as.matrix(lambda.co),
            lambda.tr = as.matrix(lambda.tr)
        ))
        if (boot == 0) {
            if (exists("use_bounded", inherits = FALSE) && isTRUE(use_bounded)) {
                out <- c(out, list(wgt.implied = wgt.implied))
            } else if (exists("inv.tr", inherits = FALSE) &&
                       !inherits(inv.tr, "try-error")) {
                out <- c(out, list(wgt.implied = wgt.implied))
            }
        }
        if (exists("use_bounded", inherits = FALSE) && isTRUE(use_bounded)) {
            out <- c(out, list(
                loading.bound       = "simplex",
                gamma.loading       = gamma_use,
                loading.proj.resid  = loading.proj.resid
            ))
        } else {
            out <- c(out, list(loading.bound = loading.bound))
        }
    }

    if (force == 1) {
        out <- c(out, list(
            alpha.tr = alpha.tr,
            alpha.co = est.co.best$alpha
        ))
    } else if (force == 2) {
        out <- c(out, list(xi = est.co.best$xi))
    } else if (force == 3) {
        out <- c(out, list(
            alpha.tr = alpha.tr,
            alpha.co = est.co.best$alpha,
            xi = est.co.best$xi
        ))
    }

    if (!is.null(placebo.period) && placeboTest == 1) {
        out <- c(out, list(att.placebo = att.placebo))
    }

    if (!is.null(W)) {
        out <- c(out, list(
            W = W,
            att.avg.W = att.avg.W,
            att.on.sum.W = att.on.sum.W,
            att.on.W = att.on.W,
            count.on.W = count.on.W,
            time.on.W = time.on.W,
            W.on.sum = W.on.sum
        ))
        if (hasRevs == 1) {
            out <- c(out, list(
                att.off.sum.W = att.off.sum.W,
                att.off.W = att.off.W,
                count.off.W = count.off.W,
                time.off.W = time.off.W,
                W.off.sum = W.off.sum
            ))
        }
        if (!is.null(placebo.period) && placeboTest == 1) {
            out <- c(out, list(att.placebo.W = att.placebo.W))
        }
    }

    if (!is.null(balance.period)) {
        out <- c(out, list(balance.att = balance.att, balance.time = balance.time, balance.count = balance.count, balance.avg.att = att.avg.balance))
        if (!is.null(placebo.period) && placeboTest == 1) {
            out <- c(out, list(balance.att.placebo = balance.att.placebo))
        }
    }

    if (!is.null(carryover.period) && carryoverTest == 1) {
        out <- c(out, list(att.carryover = att.carryover))
    }

    if (!is.null(group)) {
        out <- c(out, list(
            group.att = group.att,
            group.output = group.output
        ))
    }
    return(out)
}


## ====================================================================
## Helper functions for CFE nevertreated path
## ====================================================================

## Reconstruct gamma fitted values for treated units using co-estimated gamma
## gamma_coef: coefficient matrix from C++ Gamma(), n_groups x n_z_per_gamma
## X.Z.tr: TT x Ntr x n_z array
## gamma_groups_mat: TT x Ntr x 1 array (or TT x Ntr matrix via drop)
## z_cols: integer vector of which Z columns belong to this gamma
## Returns: TT x Ntr matrix
.reconstruct_gamma_fit_tr <- function(gamma_coef, X.Z.tr, gamma_groups_arr,
                                       z_cols, TT, Ntr) {
    ## gamma_groups_arr: TT x Ntr x 1 array (from X.gamma.tr[,,k,drop=FALSE])
    ## extract the TT x Ntr matrix
    if (length(dim(gamma_groups_arr)) == 3) {
        gamma_groups_mat <- gamma_groups_arr[, , 1, drop = FALSE]
        dim(gamma_groups_mat) <- dim(gamma_groups_mat)[1:2]
    } else {
        gamma_groups_mat <- gamma_groups_arr
    }
    if (is.null(dim(gamma_groups_mat))) {
        gamma_groups_mat <- matrix(gamma_groups_mat, nrow = TT, ncol = Ntr)
    }
    gamma_groups <- gamma_groups_mat[, 1]  ## all units share same time grouping
    unique_groups <- sort(unique(gamma_groups))

    fit <- matrix(0, TT, Ntr)

    ## Extract Z values for treated (time-invariant, take t=1)
    Z.tr.sub <- matrix(0, Ntr, length(z_cols))
    for (j in seq_along(z_cols)) {
        Z.tr.sub[, j] <- X.Z.tr[1, , z_cols[j]]
    }

    for (g_idx in seq_along(unique_groups)) {
        t_in_group <- which(gamma_groups == unique_groups[g_idx])
        if (g_idx <= nrow(gamma_coef)) {
            coef_g <- gamma_coef[g_idx, , drop = FALSE]  ## 1 x n_z
            contribution <- Z.tr.sub %*% t(coef_g)  ## Ntr x 1
            for (tt in t_in_group) {
                fit[tt, ] <- c(contribution)
            }
        }
    }
    return(fit)
}

## Reconstruct kappa fitted values
## kappa_coef: coefficient matrix from C++ Kappa()
## X.Q: TT x N x n_q array
## kappa_groups_mat: from X.kappa[,,k] -- unit-level grouping
## q_cols: integer vector
## Returns: TT x N matrix
.reconstruct_kappa_fit <- function(kappa_coef, X.Q, kappa_groups_mat, q_cols, TT, N) {
    ## kappa_coef: q x N matrix from C++ Kappa().t() — each column is a unit's
    ## kappa coefficients (q basis dimensions); units in the same group share
    ## identical columns.
    ## kappa_groups_mat may be a TT x N x 1 array (from X.kappa[,,k,drop=FALSE])
    ## extract the TT x N matrix
    if (length(dim(kappa_groups_mat)) == 3) {
        kappa_groups_mat <- kappa_groups_mat[, , 1, drop = FALSE]
        dim(kappa_groups_mat) <- dim(kappa_groups_mat)[1:2]
    }
    if (is.null(dim(kappa_groups_mat))) {
        kappa_groups_mat <- matrix(kappa_groups_mat, nrow = TT, ncol = N)
    }
    kappa_groups <- kappa_groups_mat[1, ]  ## N-length, unit grouping
    unique_groups <- sort(unique(kappa_groups))

    ## Ensure kappa_coef is a matrix (handles q=1 case)
    kappa_coef <- as.matrix(kappa_coef)

    fit <- matrix(0, TT, N)
    for (g_idx in seq_along(unique_groups)) {
        g <- unique_groups[g_idx]
        units_in_group <- which(kappa_groups == g)
        ## Extract kappa coefficients for this group from any unit in the group
        ## kappa_coef is q x N; column = unit's coefficients
        rep_unit <- units_in_group[1]
        if (rep_unit <= ncol(kappa_coef)) {
            kappa_g <- kappa_coef[, rep_unit, drop = FALSE]  ## q x 1
            ## Q basis values (time-varying)
            Q.basis <- matrix(0, TT, length(q_cols))
            for (j in seq_along(q_cols)) {
                Q.basis[, j] <- X.Q[, units_in_group[1], q_cols[j]]
            }
            contribution <- Q.basis %*% kappa_g  ## TT x 1
            for (ii in units_in_group) {
                fit[, ii] <- c(contribution)
            }
        }
    }
    return(fit)
}

## Extract Type-B FE from co estimation and apply to treated
## Returns: TT x Ntr matrix of Type-B FE fitted values for treated
.extract_and_apply_typeB_fe <- function(est.co, X.co, X.extra.FE.co.B,
                                         X.extra.FE.tr, typeB_idx,
                                         X.Z.co, X.gamma.co, X.Q.co, X.kappa.co,
                                         Zgamma.id, kappaQ.id,
                                         TT, Nco, Ntr, p, r, force) {
    ## Reconstruct non-FE fit for co
    beta <- est.co$beta
    if (!is.null(beta) && any(is.nan(beta))) beta[is.nan(beta)] <- 0

    fit_no_fe.co <- matrix(est.co$mu, TT, Nco)
    if (force %in% c(1, 3))
        fit_no_fe.co <- fit_no_fe.co + matrix(c(est.co$alpha), TT, Nco, byrow = TRUE)
    if (force %in% c(2, 3))
        fit_no_fe.co <- fit_no_fe.co + matrix(c(est.co$xi), TT, Nco, byrow = FALSE)
    if (p > 0 && !is.null(beta)) {
        for (j in 1:p) fit_no_fe.co <- fit_no_fe.co + X.co[, , j] * beta[j]
    }
    ## Add gamma for co
    if (!is.null(est.co$gamma) && length(est.co$gamma) > 0) {
        for (k_g in seq_along(est.co$gamma)) {
            gamma.fit.co.k <- .reconstruct_gamma_fit_tr(
                est.co$gamma[[k_g]], X.Z.co,
                X.gamma.co[, , k_g, drop = FALSE],
                Zgamma.id[[k_g]], TT, Nco)
            fit_no_fe.co <- fit_no_fe.co + gamma.fit.co.k
        }
    }
    ## Add kappa for co
    if (!is.null(est.co$kappa) && length(est.co$kappa) > 0) {
        for (k_k in seq_along(est.co$kappa)) {
            kappa.fit.co.k <- .reconstruct_kappa_fit(
                est.co$kappa[[k_k]], X.Q.co,
                X.kappa.co[, , k_k, drop = FALSE],
                kappaQ.id[[k_k]], TT, Nco)
            fit_no_fe.co <- fit_no_fe.co + kappa.fit.co.k
        }
    }
    ## Add factors for co
    if (r > 0 && !is.null(est.co$factor)) {
        fit_no_fe.co <- fit_no_fe.co + est.co$factor %*% t(est.co$lambda)
    }

    fe_fit.co <- est.co$fit - fit_no_fe.co  ## TT x Nco: extra FE contribution

    ## For each Type-B FE dimension, compute group means and apply to treated
    typeB.fit.tr <- matrix(0, TT, Ntr)
    for (k_b_idx in seq_along(typeB_idx)) {
        k <- typeB_idx[k_b_idx]
        labels.co <- X.extra.FE.co.B[1, , k_b_idx]
        labels.tr <- X.extra.FE.tr[1, , k]
        levels.all <- sort(unique(labels.co))

        for (g in levels.all) {
            co_mask <- which(labels.co == g)
            tr_mask <- which(labels.tr == g)
            if (length(tr_mask) > 0 && length(co_mask) > 0) {
                fe_val <- mean(fe_fit.co[, co_mask])
                typeB.fit.tr[, tr_mask] <- typeB.fit.tr[, tr_mask] + fe_val
            }
        }
    }
    return(typeB.fit.tr)
}

print.fect_plot_return <- function(x, ...) {
  # When return.data=TRUE we return a list; printing it should behave like a plot:
  # draw the figure, but don't dump $data to console.
  if (!is.null(x$p)) {
    suppressWarnings(print(x$p))
  }
  invisible(x)
}

get.cohort <- function(data,
                       D,
                       index,
                       varname = NULL, # length of 4 c("FirstTreated","Cohort",'Time_to_Treatment',"Time_to_Exit")
                       start0 = FALSE, # default 0 as the last pre-treatment period, 1 as the first post-treatment period
                       entry.time = NULL,
                       drop.always.treat = FALSE) {
    data.raw <- data
    if (length(D) > 1) {
        stop("Treatment indicator \"D\" should have length 1.")
    }
    if (is.data.frame(data) == FALSE || length(class(data)) > 1) {
        data <- as.data.frame(data)
    }
    varnames <- colnames(data)
    if (!D %in% varnames) {
        stop("\"D\" misspecified.\n")
    }

    if (is.logical(start0) == FALSE & !start0 %in% c(0, 1)) {
        stop("\"start0\" is not a logical flag.")
    }

    if (is.logical(drop.always.treat) == FALSE & !drop.always.treat %in% c(0, 1)) {
        stop("\"drop.always.treat\" is not a logical flag.")
    }


    if (is.null(varname)) {
        varname1 <- "FirstTreat"
        if (varname1 %in% colnames(data)) {
            warnings("The variable \"FirstTreat\" will be replaced.\n")
            data[, varname1] <- NULL
        }
        varname2 <- "Cohort"
        if (varname2 %in% colnames(data)) {
            warnings("The variable \"Cohort\" will be replaced.\n")
            data[, varname2] <- NULL
        }
        varname3 <- "Time_to_Treatment"
        if (varname3 %in% colnames(data)) {
            warnings("The variable \"Time_to_Treatment\" will be replaced.\n")
            data[, varname3] <- NULL
        }
        varname4 <- "Time_to_Exit"
        if (varname4 %in% colnames(data)) {
            warnings("The variable \"Time_to_Exit\" will be replaced.\n")
            data[, varname4] <- NULL
        }
    } else {
        if (length(varname) != 4) {
            stop("\"varname\" should have three elements.")
        }
        if (varname[1] %in% colnames(data)) {
            warnings(paste0("Variable ", varname[1], " will be replaced.\n"))
            data[, varname1] <- NULL
        }
        if (varname[2] %in% colnames(data)) {
            warnings(paste0("Variable ", varname[2], " will be replaced.\n"))
            data[, varname2] <- NULL
        }
        if (varname[3] %in% colnames(data)) {
            warnings(paste0("Variable ", varname[3], " will be replaced.\n"))
            data[, varname3] <- NULL
        }
        if (varname[4] %in% colnames(data)) {
            warnings(paste0("Variable ", varname[4], " will be replaced.\n"))
            data[, varname4] <- NULL
        }
    }

    if (length(index) != 2) {
        stop("\"index\" should have length 2.\n")
    }

    for (sub.index in index) {
        if (!sub.index %in% index) {
            stop(paste0(sub.index, "is not in the dataset.\n"))
        }
    }

    n.obs <- dim(data)[1]

    ## D and index should not have NA
    for (sub.var in c(D, index)) {
        if (length(which(is.na(data[, sub.var]))) > 0) {
            stop(paste0(sub.var, "contains missing values.\n"))
        }
    }

    ## check if uniquely identified
    unique_label <- unique(paste(data[, index[1]], "_", data[, index[2]], sep = ""))
    if (length(unique_label) != dim(data)[1]) {
        stop("Observations are not uniquely defined by unit and time indicators.")
    }

    if (!(class(data[, D]) %in% c("numeric", "integer"))) {
        stop("Treatment indicator should be a numeric value.")
    }

    if (!(1 %in% data[, D] & 0 %in% data[, D] & length(unique(data[, D])) == 2)) {
        stop(paste("Error values in variable \"", D, "\".", sep = ""))
    }


    ## index names
    index.id <- index[1]
    index.time <- index[2]
    data.raw[, index.id] <- as.character(data.raw[, index.id])

    ## raw id and time
    raw.id <- sort(unique(data[, index[1]]))
    raw.time <- sort(unique(data[, index[2]]))
    N <- length(raw.id)
    TT <- length(raw.time)

    ## entry time
    if (is.null(entry.time)) {
        staggered <- 1
    } else {
        staggered <- 0
        if (!is.list(entry.time)) {
            stop("\"entry.time\" should be a list.\n")
        }
        all.entry.time <- c()
        for (sub.time in entry.time) {
            if (!is.numeric(sub.time)) {
                stop("Elements in \"entry.time\" are misspecified.\n")
            }
            if (length(sub.time) != 2) {
                stop("Elements in \"entry.time\" are misspecified. Each element should have length 2.\n")
            }
            if (sub.time[1] > sub.time[2]) {
                stop("Elements in \"entry.time\" are misspecified. For each element, the second should not be smaller than the first.\n")
            }
            all.entry.time <- c(all.entry.time, sub.time)
        }
        if (length(all.entry.time) > length(unique(all.entry.time))) {
            stop("\"entry.time\" has overlapped periods.\n")
        }
    }

    ## sort data
    data <- data[order(data[, index.id], data[, index.time]), ]
    data.raw <- data
    id.match <- as.numeric(as.factor(raw.id))
    names(id.match) <- as.character(raw.id)
    time.match <- as.numeric(as.factor(raw.time))
    names(time.match) <- as.character(raw.time)

    ## generate first treated
    ## check balanced panel and fill unbalanced panel
    Dname <- D
    T.on <- I <- D <- NULL
    if (dim(data)[1] != TT * N) {
        data[, index.id] <- as.numeric(as.factor(data[, index.id]))
        data[, index.time] <- as.numeric(as.factor(data[, index.time]))
        ob.indicator <- data[, index.time]
        id.indicator <- table(data[, index.id])
        sub.start <- 1
        for (i in 1:(N - 1)) {
            sub.start <- sub.start + id.indicator[i]
            sub.end <- sub.start + id.indicator[i + 1] - 1
            ob.indicator[sub.start:sub.end] <- ob.indicator[sub.start:sub.end] + i * TT
        }
        variable <- c(Dname)
        data_I <- matrix(0, N * TT, 1)
        data_I[ob.indicator, 1] <- 1

        data_D <- matrix(NaN, N * TT, 1)
        data_D[ob.indicator, 1] <- as.matrix(data[, variable])
        colnames(data_D) <- variable
        I <- matrix(1, TT, N)
        D <- matrix(data_D[, Dname], TT, N)
        I[is.nan(D)] <- 0
        D[is.nan(D)] <- 0
    } else {
        data[, index.id] <- as.numeric(as.factor(data[, index.id]))
        data[, index.time] <- as.numeric(as.factor(data[, index.time]))
        I <- matrix(1, TT, N)
        D <- matrix(data[, Dname], TT, N)
    }

    T.on <- matrix(NA, TT, N)
    for (i in 1:N) {
        T.on[, i] <- get_term(D[, i], I[, i], type = "on")
    }
    T.off <- matrix(NA, TT, N)
    for (i in 1:N) {
        T.off[, i] <- get_term(D[, i], I[, i], type = "off")
    }

    t.on <- c(T.on)
    t.off <- c(T.off)
    use.index <- (data[, index.id] - 1) * TT + data[, index.time]
    if (start0 == TRUE) {
        t.on <- t.on[use.index] - 1
        t.off <- t.off[use.index] - 1
    } else {
        t.on <- t.on[use.index]
        t.off <- t.off[use.index]
    }

    tr.pos <- which(apply(D, 2, sum) > 0)
    co.pos <- which(apply(D, 2, sum) == 0)
    tr.name <- names(id.match)[tr.pos]
    co.name <- names(id.match)[co.pos]

    D.cum1 <- apply(D, 2, cumsum)
    D.cum2 <- apply(D.cum1, 2, cumsum)
    T0.tr <- apply(matrix(D.cum2[, tr.pos], nrow = dim(D.cum2)[1]), 2, function(vec) {
        which(vec == 1)
    })
    T0.tr.origin <- as.numeric(sapply(T0.tr, function(x) {
        names(time.match)[which(time.match == x)]
    }))
    T0.co.origin <- rep(NA, length(co.name))


    first.treat <- cbind.data.frame(
        index.id = c(tr.name, co.name),
        FirstTreat = c(T0.tr.origin, T0.co.origin)
    )

    first.treat[, varname1] <- first.treat[, "FirstTreat"]
    if (varname1 != "FirstTreat") {
        first.treat[, "FirstTreat"] <- NULL
    }
    data.raw <- merge(data.raw, first.treat, by.x = index.id, by.y = "index.id")
    data.raw[, varname2] <- "Control"
    data.raw[, varname3] <- t.on
    if (sum(!is.na(t.off)) > 0) {
        data.raw[, varname4] <- t.off
    }

    if (staggered == 1) {
        all.first.treat <- sort(unique(data.raw[, varname1]))
        for (sub.first in all.first.treat) {
            cohort.name <- paste0("Cohort:", sub.first)
            data.raw[which(data.raw[, varname1] == sub.first), varname2] <- cohort.name
        }
    } else {
        for (sub.time in entry.time) {
            cohort.name <- paste0("Cohort:", sub.time[1], "-", sub.time[2])
            data.raw[intersect(which(data.raw[, varname1] >= sub.time[1]), which(data.raw[, varname1] <= sub.time[2])), varname2] <- cohort.name
        }
        cohort.name <- "Cohort:Other"
        data.raw[which(!is.na(data.raw[, varname1]) & data.raw[, varname2] == "Control"), varname2] <- cohort.name
    }


    if (drop.always.treat) {
        size.0 <- dim(data.raw)[1]
        unit.index <- index[1]
        data.raw <- as.data.frame(data.raw %>% group_by(get(unit.index)) %>% mutate(treatment_mean = mean(get(Dname), na.rm = TRUE)))
        data.raw <- data.raw[which(data.raw$treatment_mean < 1), ]
        size.1 <- dim(data.raw)[1]
        diff.1 <- size.0 - size.1
        message(paste0("Number of Always Treated Observations Removed:", diff.1, ".\n"))
        data.raw[, "treatment_mean"] <- NULL
    }

    return(data.raw)
}

test_that("get.cohort adds cohort columns and handles options", {
  skip_on_cran()
  suppressWarnings(try(data("simdata", package = "fect"), silent = TRUE))
  expect_true(exists("simdata"))
  df <- simdata

  out <- get.cohort(
    data = df,
    D = "D",
    index = c("id", "time"),
    start0 = TRUE,
    drop.always.treat = TRUE
  )

  expect_true(all(c("Cohort", "Time_to_Treatment") %in% names(out)))
  expect_true(all(out$Cohort %in% c("Control", unique(out$Cohort))))
})

## ---------------------------------------------------------------
## Shared fixtures for the score-unify test family.
##
## testthat auto-loads helper-*.R files before any test-*.R, so the
## fixtures defined here (make_factor_data, ntdata, out_base) are
## visible to every test-score-*.R file. This avoids duplicating the
## ~60 lines of fixture setup across the 7 split test files.
##
## Originally lived at the top of test-score-unify.R; extracted on
## 2026-05-03 when that file was split for readability + progress
## visibility under reporter = "summary".
## ---------------------------------------------------------------

suppressWarnings(data("simdata", package = "fect"))

## DGP with factor structure and sufficient never-treated units for CV.
## N=50, TT=20, Ntr=15 => 35 never-treated units with 20 pre-treatment
## periods --- plenty for cross-validation with r up to 3.
make_factor_data <- function(N = 50, TT = 20, Ntr = 15, tau = 3.0,
                              r = 2, seed = 42) {
  set.seed(seed)
  F_mat <- matrix(rnorm(TT * r), TT, r)
  L_mat <- matrix(rnorm(N * r), N, r)
  alpha_i <- rnorm(N, 0, 1)
  xi_t <- rnorm(TT, 0, 0.5)

  T0_vec <- rep(Inf, N)
  if (Ntr > 0) {
    T0_vec[1:Ntr] <- sample(round(TT * 0.4):round(TT * 0.7), Ntr,
                             replace = TRUE)
  }

  Y_vec <- D_vec <- numeric(N * TT)
  id_vec <- time_vec <- integer(N * TT)
  idx <- 1
  for (i in 1:N) {
    for (t in 1:TT) {
      treated <- (t >= T0_vec[i])
      D_vec[idx] <- as.integer(treated)
      Y_vec[idx] <- alpha_i[i] + xi_t[t] +
        sum(F_mat[t, ] * L_mat[i, ]) +
        tau * D_vec[idx] + rnorm(1, 0, 0.5)
      id_vec[idx] <- i
      time_vec[idx] <- t
      idx <- idx + 1
    }
  }

  data.frame(id = id_vec, time = time_vec, Y = Y_vec, D = D_vec)
}

## Shared never-treated fixture for Section C / E / F tests.
ntdata <- make_factor_data(N = 50, TT = 20, Ntr = 15, r = 2, seed = 42)

## Shared fitted object for the .score_residuals() / fect_cv / fect_mspe
## test families. Fit once at session start; all test-*.R files that need
## it can reference the global symbol.
out_base <- suppressWarnings(suppressMessages(
  fect::fect(
    Y ~ D + X1 + X2,
    data    = simdata,
    index   = c("id", "time"),
    method  = "ife",
    r       = 2,
    CV      = FALSE,
    se      = FALSE,
    parallel = FALSE
  )
))

\name{fect_mspe}
\alias{fect_mspe}
\title{Mean Squared Prediction Error Evaluation for fect Objects}
\description{
Evaluates the prediction accuracy of one or more \code{fect} model fits by hiding a subset of control observations and comparing counterfactual predictions against actual values.
}
\usage{
fect_mspe(
  out.fect,
  seed = NULL,
  cv.method = "rolling",
  cv.nobs = 3,
  cv.donut = 1,
  cv.buffer = 1,
  cv.prop = 0.1,
  min.T0 = 5,
  k = 20,
  criterion = "mspe",
  W = NULL,
  norm.para = NULL,
  proportion = 0
)
}
\arguments{
  \item{out.fect}{A fitted \code{fect} object or a named list of fitted \code{fect} objects to compare.}
  \item{seed}{Optional integer; random seed for reproducibility.}
  \item{cv.method}{Character; cross-validation masking strategy. One of \code{"rolling"} (default), \code{"block"} (random scattered anchors with contiguous-block masking), or \code{"loo"}. Legacy aliases \code{"all_units"} (= \code{"block"}) and \code{"treated_units"} are still accepted but emit a deprecation message.}
  \item{cv.nobs}{Integer; number of observations to mask per unit per fold. Default is 3.}
  \item{cv.donut}{Integer; number of periods around treatment onset to exclude from masking. Used by block CV. Default is 1.}
  \item{cv.buffer}{Integer; number of past-side cells masked from training (but not scored) before each rolling-window holdout. Used only by \code{cv.method = "rolling"}. Default is 1.}
  \item{cv.prop}{Numeric; for block CV, proportion of observations to mask per round; for rolling CV, fraction of eligible units sampled per fold. Default is 0.2.}
  \item{min.T0}{Integer; minimum number of pre-treatment periods required. Default is 5.}
  \item{k}{Integer; number of cross-validation folds. Default is 5.}
  \item{criterion}{Character; scoring criterion. One of \code{"mspe"}, \code{"wmspe"}, \code{"gmspe"}, \code{"wgmspe"}, \code{"mad"}, \code{"moment"}, \code{"gmoment"}. Default is \code{"mspe"}.}
  \item{W}{Optional TT x N observation weight matrix. Default is \code{NULL}.}
  \item{norm.para}{Optional normalization vector. Default is \code{NULL}.}
  \item{proportion}{Numeric; proportion cutoff for count.T.cv (same as fect_cv). Default is 0.}
}
\value{
A list containing:
\item{summary}{A data frame with mean scores across replications for each model, including MSPE, WMSPE, GMSPE, WGMSPE, MAD, Moment, GMoment, RMSE, and Bias.}
\item{records}{A data frame with per-replication results including Rep, Model, Hidden_N, and all score columns.}
\item{fits}{The refitted \code{fect} objects from the last replication.}
\item{criterion}{The scoring criterion used.}
\item{scores}{The scores from the last replication.}
}
\examples{
\dontrun{
out <- fect(Y ~ D, data = df, index = c("unit", "time"), method = "ife", r = 2)
mspe <- fect_mspe(out.fect = out, hide_n = 10, seed = 42, n_rep = 5)
mspe$summary
}
}

\name{esplot}
\alias{esplot}
\title{Event Study Visualization}
\description{
  Visualize dynamic treatment effects and create an event study plot.
  This function offers flexibility in displaying estimates, confidence intervals,
  and various annotations. It can handle data directly or from `did_wrapper` objects,
  calculate confidence intervals from standard errors if needed, and allows for
  connected (line/ribbon) or point-range style plots.
}
\usage{
esplot(data, Period = NULL, Estimate = "ATT", SE = NULL,
       CI.lower = "CI.lower", CI.upper = "CI.upper", Count = NULL,
       proportion = 0.3, est.lwidth = NULL, est.pointsize = NULL,
       show.points = TRUE, fill.gap = TRUE, start0 = FALSE,
       only.pre = FALSE, only.post = FALSE, show.count = TRUE,
       stats = NULL, stats.labs = NULL, highlight.periods = NULL,
       highlight.colors = NULL, highlight.shapes = NULL,
       highlight.fill = FALSE,
       lcolor = NULL, lwidth = NULL,
       ltype = NULL, connected = FALSE, ci.outline = FALSE,
       main = NULL, xlim = NULL, ylim = NULL, xlab = NULL, ylab = NULL,
       gridOff = FALSE, stats.pos = NULL, theme.bw = TRUE,
       legacy.style = FALSE,
       cex.main = NULL, cex.axis = NULL, cex.lab = NULL,
       cex.text = NULL, axis.adjust = FALSE, color = "#000000",
       pre.color = NULL, post.color = NULL,
       count.color = "gray70", count.alpha = 0.4,
       count.outline.color = "grey69",
       xangle = NULL, yangle = NULL,
       xbreaks = NULL, ybreaks = NULL,
       legendOff = FALSE, cex.legend = NULL,
       pre.label = "Pre-treatment", post.label = "Post-treatment")
}
\arguments{
  \item{data}{The input data for the event study plot. Can be a \code{data.frame} or an object of class \code{did_wrapper} (in which case, the \code{est.att} component will be used).}
  \item{Period}{The name of the column in \code{data} representing the relative time period. If \code{NULL} (the default), the function will attempt to automatically identify the period column from common names (e.g., 'time', 'Time', 'period', 'Period', 'event.time', 'event_time', 'rel_time') or use numeric rownames if available.}
  \item{Estimate}{The name of the column in \code{data} containing the point estimates (e.g., Average Treatment Effect on the Treated). Default is \code{"ATT"}.}
  \item{SE}{The name of the column in \code{data} containing the standard errors. If columns for \code{CI.lower} and \code{CI.upper} (as specified by their respective arguments, defaulting to \code{"CI.lower"} and \code{"CI.upper"}) are not found in \code{data}, and an \code{SE} column is provided and found, 95\% confidence intervals will be calculated using \code{Estimate +/- 1.96 * SE}. Default is \code{NULL}.}
  \item{CI.lower}{The name of the column in \code{data} for the lower bound of the confidence interval. Default is \code{"CI.lower"}. If this column is not found, it may be calculated from \code{Estimate} and \code{SE} (if \code{SE} is provided and found in \code{data}).}
  \item{CI.upper}{The name of the column in \code{data} for the upper bound of the confidence interval. Default is \code{"CI.upper"}. If this column is not found, it may be calculated from \code{Estimate} and \code{SE} (if \code{SE} is provided and found in \code{data}).}
  \item{Count}{Optional. The name of the column in \code{data} indicating a count measure (e.g., number of observations) for each time period. Used if \code{show.count = TRUE} or for \code{xlim} determination based on \code{proportion}. Default is \code{NULL}.}
  \item{proportion}{Numeric, between 0 and 1. If \code{Count} is specified and \code{xlim} is not, this proportion is used to determine the default x-axis limits. Periods where the count is below \code{proportion * max(Count)} might be excluded from the default view. Default is \code{0.3}.}
  \item{est.lwidth}{Numeric. The line width for the estimate line (if \code{connected = TRUE}) or the main vertical line of the point-range (if \code{connected = FALSE}). Default is \code{NULL}, which resolves to \code{0.6} if \code{connected = FALSE} (the default for \code{connected}), \code{1.2} if \code{connected = TRUE} and \code{show.points = FALSE} (the default for \code{show.points}), and \code{0.7} if \code{connected = TRUE} and \code{show.points = TRUE}.}
  \item{est.pointsize}{Numeric. The size of the points. If \code{connected = TRUE} and \code{show.points = TRUE}, this is the size of points at integer time periods. If \code{connected = FALSE}, this controls the size of the central point in the point-range (via \code{fatten} aesthetic). Default is \code{NULL}, which resolves to \code{2} if \code{connected = FALSE} (the default for \code{connected}), \code{3} if \code{connected = TRUE} and \code{show.points = FALSE} (the default for \code{show.points}), and \code{1.2} if \code{connected = TRUE} and \code{show.points = TRUE}.}
  \item{show.points}{Logical. If \code{connected = TRUE}, whether to display points at integer time periods on top of the line and ribbon. Default is \code{TRUE}.}
  \item{fill.gap}{Logical. If \code{connected = FALSE}, whether to fill gaps in the sequence of time periods with an estimate and confidence interval of 0. This is useful when some integer time periods are missing from the input data. Default is \code{TRUE}.}
  \item{start0}{Logical. If \code{TRUE}, the vertical line separating pre- and post-treatment periods is drawn at \code{x = -0.5}, implying period 0 is the first post-treatment period. If \code{FALSE} (default), the line is at \code{x = 0.5}, implying period 0 is the last pre-treatment period.}
  \item{only.pre}{Logical. If \code{TRUE}, the plot will only display pre-treatment periods. The vertical separator line will be omitted. Default is \code{FALSE}.}
  \item{only.post}{Logical. If \code{TRUE}, the plot will only display post-treatment periods. The vertical separator line will be omitted. Default is \code{FALSE}.}
  \item{show.count}{Logical. Whether to display a bar plot of the values from the \code{Count} column at the bottom of the main plot. Default is \code{TRUE}.}
  \item{stats}{Optional. A numeric vector of statistics (e.g., p-values) to be printed on the plot.}
  \item{stats.labs}{Optional. A character vector of labels corresponding to the \code{stats} values. Must be the same length as \code{stats}.}
  \item{highlight.periods}{Optional. A numeric vector of time periods to highlight with different colors. For \code{connected = TRUE}, these define intervals from \code{period - 0.5} to \code{period + 0.5}. For \code{connected = FALSE}, individual points at these periods are highlighted.}
  \item{highlight.colors}{Optional. A character vector of colors corresponding to \code{highlight.periods}. If \code{NULL} and \code{highlight.periods} is provided, default rainbow colors are used. Must be the same length as \code{highlight.periods}.}
  \item{highlight.shapes}{Optional. An integer vector of point shapes corresponding to \code{highlight.periods} (e.g., 17 = triangle, 18 = diamond). Must be the same length as \code{highlight.periods}. Default is \code{NULL}.}
  \item{highlight.fill}{Logical. If \code{TRUE}, draws a lightened-tone background rectangle behind each highlighted period (auto-derived from \code{highlight.colors}). If \code{FALSE} (the default), only the colored glyph at each highlight period is drawn. The colored glyph alone is usually sufficient signal that a period belongs to a test window, and the glyph-only look survives grayscale printing without surprise. Set \code{TRUE} for slide / talk figures where the rectangle helps a remote audience locate the window.}
  \item{lcolor}{Optional. Color(s) for the reference lines. Can be a single color (applied to both horizontal y=0 line and vertical pre/post separator line) or a vector of two colors (first for horizontal, second for vertical). If \code{NULL}, defaults to \code{"#aaaaaa"} if \code{theme.bw = TRUE}, otherwise \code{"white"}. Default is \code{NULL}.}
  \item{lwidth}{Optional. Line width(s) for the reference lines. Can be a single width or a vector of two widths (similar to \code{lcolor}). If \code{NULL}, defaults to \code{1.5} if \code{theme.bw = TRUE}, otherwise \code{2}. Default is \code{NULL}.}
  \item{ltype}{Optional. Linetype(s) for the reference lines. Can be a single linetype (applied to both horizontal y=0 line and vertical pre/post separator line) or a vector of two linetypes (first for horizontal, second for vertical). Default is \code{NULL}, which resolves to \code{c("solid", "dashed")} under the modern recipe (\code{theme.bw = TRUE} with \code{legacy.style = FALSE}) and to \code{c("solid", "solid")} otherwise.}
  \item{connected}{Logical. If \code{TRUE}, estimates and confidence intervals are plotted as a connected line with a ribbon. This involves interpolating values between observed time points (at 0.5 steps by default) to create a smoother appearance. If \code{FALSE} (default), \code{geom_pointrange} is used, showing individual estimates and their CIs as points with ranges for each observed time period.}
  \item{ci.outline}{Logical. If \code{connected = TRUE}, whether to draw an outline around the confidence interval ribbon. The outline color is a slightly darker version of the fill color. Default is \code{FALSE}.}
  \item{main}{Optional. The main title for the plot. If \code{NULL} (default), a default title "Estimated Dynamic Treatment Effects" is used. If an empty string \code{""} is provided, no title is displayed.}
  \item{xlim}{Optional. A numeric vector of length 2 specifying the x-axis limits (\code{c(min, max)}). If \code{NULL} (default), limits are determined automatically based on the data range, potentially filtered by \code{proportion} if \code{Count} is used.}
  \item{ylim}{Optional. A numeric vector of length 2 specifying the y-axis limits (\code{c(min, max)}). If \code{NULL} (default), limits are determined automatically to encompass all estimates and confidence intervals, with potential expansion if \code{show.count = TRUE}.}
  \item{xlab}{Optional. The label for the x-axis. If \code{NULL} (default), "Time Relative to Treatment" is used. If an empty string \code{""} is provided, no label is displayed.}
  \item{ylab}{Optional. The label for the y-axis. If \code{NULL} (default), "Effect on Y" is used. If an empty string \code{""} is provided, no label is displayed.}
  \item{gridOff}{Logical. Whether to turn off major and minor grid lines. Default is \code{TRUE}.}
  \item{stats.pos}{Optional. A numeric vector of length 2 (\code{c(x, y)}) specifying the coordinates for the top-left position of the \code{stats} text block. If \code{NULL} (default), the position is automatically determined.}
  \item{theme.bw}{Logical. Whether to use \code{ggplot2::theme_bw()}. Default is \code{TRUE}.}
  \item{legacy.style}{Logical. If \code{TRUE}, reproduces the pre-2.3.1 visual defaults (bold centered title, larger axis text, solid vline, blue placebo triangles, no peach highlight rectangle) regardless of the value of \code{theme.bw}. If \code{FALSE} (the default) and \code{theme.bw = TRUE}, the modern v2.3.1 visual recipe is used. Pass \code{legacy.style = TRUE} to byte-identical-reproduce figures rendered under earlier fect versions.}
  \item{cex.main}{Optional. Numeric scaling factor for the plot title font size. The base size used by ggplot is 16. Default is \code{NULL} (uses base size 16).}
  \item{cex.axis}{Optional. Numeric scaling factor for the axis tick mark labels font size. The base size used by ggplot is 15. Default is \code{NULL} (uses base size 15).}
  \item{cex.lab}{Optional. Numeric scaling factor for the axis title (x and y labels) font size. The base size used by ggplot is 15. Default is \code{NULL} (uses base size 15).}
  \item{cex.text}{Optional. Numeric scaling factor for annotated text elements (e.g., \code{stats} text, count label). The base size used by ggplot for annotation is 5. Default is \code{NULL} (uses base size 5).}
  \item{axis.adjust}{Logical. If \code{TRUE}, x-axis tick labels are rotated 45 degrees for better readability with many labels. Default is \code{FALSE}.}
  \item{color}{Character. The primary color for plotting estimates, points, lines, and confidence interval fills/lines (unless overridden by \code{highlight.colors} for specific periods). Default is \code{"#000000"} (black).}
  \item{pre.color}{Optional. Character. Color for pre-treatment period estimates. If \code{NULL} (default), uses the value of \code{color}.}
  \item{post.color}{Optional. Character. Color for post-treatment period estimates. If \code{NULL} (default), uses the value of \code{color}.}
  \item{count.color}{Character. The fill color for the bars if \code{show.count = TRUE}. Default is \code{"gray70"}.}
  \item{count.alpha}{Numeric. Alpha transparency for the count bars if \code{show.count = TRUE}. Default is \code{0.4}.}
  \item{count.outline.color}{Character. The color for the outline of count bars if \code{show.count = TRUE}. Default is \code{"grey69"}.}
  \item{xangle}{Optional. Numeric. Rotation angle (in degrees) for x-axis tick labels. Default is \code{NULL} (no rotation).}
  \item{yangle}{Optional. Numeric. Rotation angle (in degrees) for y-axis tick labels. Default is \code{NULL} (no rotation).}
  \item{xbreaks}{Optional. A numeric vector of break points for the x-axis. If \code{NULL} (default), breaks are determined automatically.}
  \item{ybreaks}{Optional. A numeric vector of break points for the y-axis. If \code{NULL} (default), breaks are determined automatically.}
  \item{legendOff}{Logical. If \code{TRUE}, the legend is suppressed. Default is \code{FALSE}.}
  \item{cex.legend}{Optional. Numeric scaling factor for the legend text size. Default is \code{NULL} (uses ggplot2 defaults).}
  \item{pre.label}{Character. Label for the pre-treatment period in the legend. Default is \code{"Pre-treatment"}.}
  \item{post.label}{Character. Label for the post-treatment period in the legend. Default is \code{"Post-treatment"}.}
}
\value{
  \item{p}{A \code{ggplot} object representing the event study plot.}
}
\author{
  Licheng Liu, Yiqing Xu, Ziyi Liu, Zhongyu Yin, Rivka Lipkovitz
}
\examples{
# Basic example with simulated data
set.seed(123)
event_data <- data.frame(
  time = -5:5,
  ATT = cumsum(rnorm(11, 0, 0.2)) + c(rep(0,5), 0, 0.5, 1, 1.2, 1.5, 1.3),
  SE = runif(11, 0.1, 0.3)
)
event_data$CI.lower <- event_data$ATT - 1.96 * event_data$SE
event_data$CI.upper <- event_data$ATT + 1.96 * event_data$SE
event_data$count <- sample(50:150, 11, replace = TRUE)
event_data$count[event_data$time == -5 | event_data$time == 5] <- 20 # for proportion demo

# Default plot (point-range)
esplot(event_data, Period = "time", Estimate = "ATT",
       CI.lower = "CI.lower", CI.upper = "CI.upper")

# # Connected plot with ribbon
# esplot(event_data, Period = "time", Estimate = "ATT",
#        CI.lower = "CI.lower", CI.upper = "CI.upper",
#        connected = TRUE, show.points = TRUE)

# # Connected plot using SE for CI calculation
# event_data_no_ci <- event_data[, c("time", "ATT", "SE", "count")]
# esplot(event_data_no_ci, Period = "time", Estimate = "ATT", SE = "SE",
#        connected = TRUE, ci.outline = TRUE, color = "blue")

# # Show count bars and stats
# esplot(event_data, Period = "time", Estimate = "ATT",
#        CI.lower = "CI.lower", CI.upper = "CI.upper", Count = "count",
#        show.count = TRUE, stats = c(0.03, 0.12), stats.labs = c("P-val Pre", "P-val Post"),
#        main = "Event Study with Counts and Stats", proportion = 0.2)

# # Highlight specific periods (connected)
# esplot(event_data, Period = "time", Estimate = "ATT", SE = "SE",
#        connected = TRUE, highlight.periods = c(-1, 2),
#        highlight.colors = c("orange", "green"),
#        main = "Highlighted Periods (Connected)")

# # Highlight specific periods (point-range)
# esplot(event_data, Period = "time", Estimate = "ATT", SE = "SE",
#        connected = FALSE, highlight.periods = c(-1, 2),
#        highlight.colors = c("orange", "green"),
#        main = "Highlighted Periods (Point-Range)")

# # Only post-treatment period, custom labels
# esplot(event_data, Period = "time", Estimate = "ATT", SE = "SE",
#        only.post = TRUE, xlab = "Years Post-Intervention", ylab = "Impact Metric",
#        start0 = TRUE, color = "darkred", est.lwidth = 1.5)

# Using did_wrapper object (conceptual example, requires `did` package and setup)
# if (requireNamespace("did", quietly = TRUE)) {
#   # Assume `did_out` is an output from `did::att_gt` or similar
#   # and `did_wrapper_obj` is created, e.g.,
#   # did_wrapper_obj <- list(est.att = event_data) # Simplified for example
#   # class(did_wrapper_obj) <- "did_wrapper"
#   # esplot(did_wrapper_obj) # Would use defaults: Period="time", Estimate="ATT"
# }

}

\name{imputed_outcomes}
\alias{imputed_outcomes}
\title{Imputed Potential Outcomes Accessor}
\description{
  Returns the cell-level imputed potential-outcome surface from a fitted
  \code{fect} object as a long-form data frame. This is the rock-bottom
  accessor for any post-hoc estimand: read off the columns and aggregate
  however you like. Used internally by \code{\link{estimand}} for the
  shipped estimand types; exposed publicly so users can compose custom
  estimands the package does not ship.
}

\usage{
imputed_outcomes(
  fit,
  cells = NULL,
  replicates = FALSE,
  direction = c("on", "off")
)
}

\arguments{
  \item{fit}{A \code{fect} object (output of \code{\link{fect}}).}

  \item{cells}{Optional filter on which treated cells to return. Accepts
    \code{NULL} (default; all treated cells), a logical vector aligned
    with the unfiltered row order, or a one-sided formula evaluated
    against the long-form data
    (e.g., \code{~ event.time \%in\% 1:5 & !id \%in\% bad_ids}).}

  \item{replicates}{Logical. \code{FALSE} (default) returns one row per
    treated cell with the point-estimate \code{Y0_hat}. \code{TRUE}
    expands by bootstrap/jackknife replicate; requires the fit to have
    been built with \code{keep.sims = TRUE}.}

  \item{direction}{Either \code{"on"} (default; event time relative to
    treatment onset) or \code{"off"} (relative to treatment exit; only
    meaningful on reversal panels).}
}

\value{
  A data frame with one row per treated cell or per (treated cell,
  replicate). Columns:
  \describe{
    \item{\code{id}}{Unit identifier from \code{fit$id}.}
    \item{\code{time}}{Calendar time from \code{fit$rawtime}.}
    \item{\code{event.time}}{Event time at this cell (relative to onset
      or exit per \code{direction}).}
    \item{\code{cohort}}{First-treatment calendar time for the unit
      (\code{direction = "on"}) or first-exit calendar time
      (\code{direction = "off"}).}
    \item{\code{treated}}{Logical; always \code{TRUE} in returned rows.}
    \item{\code{Y_obs}}{Observed outcome at this cell.}
    \item{\code{Y0_hat}}{Imputed counterfactual outcome
      (\code{Y_obs - eff}).}
    \item{\code{eff}}{Cell-level score: contribution to the ATT estimator.
      For imputation estimators this equals \code{Y_obs - Y0_hat}; for
      doubly-robust estimators it equals
      \code{(Y_obs - Y0_hat) + eff_debias}.}
    \item{\code{eff_debias}}{Debias correction; 0 for plain imputation
      estimators, populated for DR estimators.}
    \item{\code{W.agg}}{Aggregation weight at this cell; 1 if the fit
      was built without \code{W} or \code{W.agg}.}
    \item{\code{replicate}}{(only when \code{replicates = TRUE})
      1..\code{nboots} for bootstrap, or the dropped-unit index for
      jackknife.}
  }
}

\section{Memory cost}{
  With \code{replicates = TRUE} the returned data frame has
  \code{n_treated_cells * nboots} rows. For typical panels this is
  manageable; for large panels (\eqn{TT \times N \ge 50{,}000} and
  \code{nboots} \eqn{\ge 500}) consider filtering via \code{cells}
  before expansion.
}

\section{Slot contract}{
  This function reads from \code{fit$Y.dat}, \code{fit$D.dat},
  \code{fit$T.on} (or \code{T.off}), \code{fit$eff},
  \code{fit$eff.boot} (when \code{replicates = TRUE}),
  \code{fit$eff_debias} (when populated), \code{fit$W.agg},
  \code{fit$id}, \code{fit$rawtime}. These slots have a documented
  stable shape; see the package design notes.
}

\seealso{
  \code{\link{estimand}} for the typed dispatcher built on top of this
  accessor; \code{\link{fect}} for the fitting interface.
}

\examples{
\dontrun{
library(fect)
fit <- fect(Y ~ D, data = simdata, index = c("id", "time"),
            method = "fe", force = "two-way",
            se = TRUE, nboots = 200, parallel = FALSE,
            keep.sims = TRUE)

## Point-estimate long form: one row per treated cell.
po <- imputed_outcomes(fit)
head(po)

## Filter to first 5 event times.
po5 <- imputed_outcomes(fit, cells = ~ event.time \%in\% 1:5)

## Bootstrap replicate expansion (requires keep.sims = TRUE).
po_b <- imputed_outcomes(fit, replicates = TRUE)
nrow(po_b) == nrow(po) * 200    # one row per (cell, replicate)
}
}

\name{gs2020}
\alias{gs2020}
\docType{data}
\title{Simulated Gsynth-like Panel Data (No Reversal)}
\description{A simulated dataset for demonstration and testing.}
\format{data.frame}
\keyword{datasets}

\encoding{UTF-8}
\name{simdata}
\docType{data}
\alias{simdata}
\title{Simulated panel data with two latent factors}
\description{
A simulated panel dataset with continuous outcomes used throughout the
package vignettes to demonstrate factor-augmented counterfactual
estimators. The data-generating process follows
\href{https://onlinelibrary.wiley.com/doi/10.1111/ajps.12723}{Liu, Wang,
and Xu (2024)} with one modification (see Format).

The panel has \eqn{N = 200} units and \eqn{T = 35} time periods.
Treatment switches on and off over time (99 of 150 treated units
experience at least one reversal), reflecting a general treatment
pattern rather than simple staggered adoption. The outcome includes two
latent factors (\eqn{r = 2}), so the parallel-trends assumption is
violated and the standard fixed-effects estimator is biased. Treatment
assignment loads on the same factors and fixed effects that enter the
outcome---units with larger \eqn{\lambda_i} and \eqn{\alpha_i} are more
likely to be treated---so the confounding is structural and cannot be
removed by two-way fixed effects alone.
}
\format{
A data frame with the following columns:
\describe{
  \item{id}{unit identifier (1--200)}
  \item{time}{time period (1--35)}
  \item{Y}{observed outcome}
  \item{error}{idiosyncratic error \eqn{\varepsilon_{it} \sim N(0, 2)}}
  \item{eff}{realized treatment effect \eqn{\tau_{it}}}
  \item{tr_cum, tr_prob}{treatment-probability constructions}
  \item{D}{treatment indicator}
  \item{X1, X2}{observed time-varying covariates \eqn{\sim N(0, 1)}
        with coefficients 1 and 3}
  \item{alpha}{unit fixed effect \eqn{\alpha_i \sim N(0, 1)}}
  \item{xi}{time fixed effect \eqn{\xi_t} (AR(1) with drift)}
  \item{F1, F2}{latent time factors \eqn{f_t \in \mathbb{R}^2}
        (one trending, one white noise)}
  \item{L1, L2}{unit-specific factor loadings
        \eqn{\lambda_i \sim N(0.5, 1)}}
  \item{FL1, FL2}{per-cell factor-loading products
        \eqn{\lambda_{i,k} \cdot f_{t,k}} (\eqn{k = 1, 2})}
}

The DGP is
\deqn{Y_{it} = \tau_{it} D_{it} + X_{1,it} + 3 X_{2,it} + \mu
              + 3\alpha_i + \xi_t
              + 2\, \lambda_i' f_t + \varepsilon_{it},}
with grand mean \eqn{\mu = 5} and treatment effect
\eqn{\tau_{it} \sim N(0.4 \cdot \mathrm{tr\_cum}_{it}/T,\; 0.2)}.

The \eqn{2\, \lambda_i' f_t} term doubles the latent factor contribution
relative to the original Liu, Wang, and Xu (2024) DGP. The doubling
strengthens the factor signal-to-noise ratio (variance of the factor
contribution to variance of the residual) from approximately 2.7 to
10.9, which makes the factor structure clearly recoverable by
cross-validated rank-selection procedures on this dataset. The
unmodified DGP is preserved in earlier package versions; see
\code{git log data/simdata.rda} for the prior file.
}
\references{
  Liu, L., Wang, Y., and Xu, Y. (2024).
  A Practical Guide to Counterfactual Estimators for Causal Inference
  with Time-Series Cross-Sectional Data.
  \emph{American Journal of Political Science}, 68(1), 160--176.
}
\keyword{datasets}

\encoding{UTF-8}
\name{print.fect}
\alias{print.fect}
\title{Print Results}
\description{Print results of the matrix completion method.}
\usage{\method{print}{fect}(x, switch.on = TRUE,
       switch.off = FALSE,time.on.lim = NULL, time.off.lim = NULL, \dots)} 
\arguments{
  \item{x}{a \code{\link{fect}} object.}
  \item{switch.on}{logical; if \code{TRUE}, print switch-on effect.}
  \item{switch.off}{logical; if \code{TRUE}, print switch-off effect.}
  \item{time.on.lim}{two-element numeric vector specifying the switch-on effect range.}
  \item{time.off.lim}{two-element numeric vector specifying the switch-off effect range.}
  \item{\dots}{other arguments.}
}
\value{No return value.}
\author{
  Licheng Liu; Ye Wang; Yiqing Xu; Ziyi Liu
}
\references{  
  Athey, S., Bayati, M., Doudchenko, N., Imbens, G., and Khosravi, K. (2021). 
  Matrix completion methods for causal panel data models. 
  \emph{Journal of the American Statistical Association, 116}(536), 1716-1730. 
  
  Bai, J. (2009). 
  Panel data models with interactive fixed effects. 
  \emph{Econometrica, 77}(4), 1229-1279. 
    
  Liu, L., Wang, Y., and Xu, Y. (2022).
  A Practical Guide to Counterfactual Estimators for Causal Inference with Time-Series Cross-Sectional Data. 
  \emph{American Journal of Political Science, 68}(1), 160-176. 
  
  Xu, Y. (2017). 
  Generalized Synthetic Control Method: Causal Inference with Interactive Fixed Effects Models. 
  \emph{Political Analysis, 25}(1), 57-76. 
}
\seealso{
  \code{\link{fect}}, \code{\link{plot.fect}}
}
\examples{
library(fect)
data(simdata)
out <- fect(Y ~ D + X1 + X2, data = simdata, 
            index = c("id","time"), force = "two-way",
            CV = TRUE, r = c(0, 5), se = 0, parallel = FALSE) 
print(out)
}

\name{estimand}
\alias{estimand}
\title{Post-hoc Estimand Dispatcher}
\description{
  Computes a post-hoc estimand from a \code{fect} fit, with bootstrap or
  jackknife uncertainty. The \code{type} argument selects from a closed
  enum of mathematically-defined estimands; the \code{by} argument
  selects the grouping axis. The accessor \code{\link{imputed_outcomes}}
  is the underlying long-form data source for any estimand the
  dispatcher does not ship natively.
}

\usage{
estimand(
  fit,
  type   = c("att", "att.cumu", "aptt", "log.att"),
  by     = c("event.time", "cohort", "calendar.time", "overall"),
  test        = c("none", "placebo", "carryover"),
  cells       = NULL,
  weights     = NULL,
  window      = NULL,
  direction   = c("on", "off"),
  vartype     = c("bootstrap", "jackknife", "parametric", "none"),
  conf.level  = 0.95,
  ci.method   = NULL
)
}

\arguments{
  \item{fit}{A \code{fect} object.}

  \item{type}{Estimand type. \code{"att"} returns per-cell mean
    treatment effect aggregated per group; \code{"att.cumu"} returns
    cumulative ATT through each event time (replaces \code{\link{effect}});
    \code{"aptt"} returns the average proportional treatment effect on
    the treated (Chen and Roth 2024 QJE); \code{"log.att"} returns the
    mean log-scale treatment effect. \code{"aptt"} and \code{"log.att"}
    require \code{keep.sims = TRUE} at fit time for SE/CI.}

  \item{by}{Grouping axis. One of \code{"event.time"} (default),
    \code{"cohort"}, \code{"calendar.time"}, \code{"overall"}, or any
    column name resolvable in the fit's panel data.}

  \item{test}{Selects which subset of cells to aggregate over.
    \code{"none"} (default) uses the standard treated post-treatment
    cells. \code{"placebo"} restricts aggregation to the
    pre-treatment cells in \code{fit$placebo.period} that were
    masked-and-imputed during the placebo fit; e.g.\
    \code{estimand(fit, "aptt", "event.time", test = "placebo")}
    returns a per-event-time placebo APTT series. Requires
    \code{placeboTest = TRUE} at fit time and forces
    \code{direction = "on"}. \code{"carryover"} is the analogous
    extension for early post-reversal cells in
    \code{fit$carryover.period}; requires \code{carryoverTest = TRUE}
    at fit time, a panel with reversals, and forces
    \code{direction = "off"}. Both are incompatible with
    \code{type = "att.cumu"}.}

  \item{cells}{Optional filter on which treated cells to include.
    Accepts \code{NULL} (default), a logical vector, or a one-sided
    formula. See \code{\link{imputed_outcomes}}.}

  \item{weights}{Aggregation-weight handling. \code{NULL} (default) uses
    \code{fit$W.agg} if the fit was built with \code{W} or \code{W.agg};
    otherwise uniform.}

  \item{window}{Optional event-time window \code{c(L, R)}; convenience
    sugar for \code{cells = ~ event.time >= L & event.time <= R}.}

  \item{direction}{Either \code{"on"} (default) or \code{"off"}; see
    \code{\link{imputed_outcomes}}.}

  \item{vartype}{\code{"bootstrap"} (default), \code{"jackknife"},
    \code{"parametric"}, or \code{"none"}. Selects which variance
    method to source replicates from. The output \code{vartype}
    column reports the method actually used at fit time (read from
    \code{fit$vartype}); this argument is informational and does not
    re-aggregate replicates. Wild-bootstrap fits (\code{vartype =
    "wild"} at fit time) are consumed transparently and reported as
    \code{"wild"} in the output column.}

  \item{conf.level}{Two-sided confidence level. Defaults to 0.95.}

  \item{ci.method}{One of \code{"basic"}, \code{"percentile"},
    \code{"bc"} (bias-corrected percentile), \code{"bca"}
    (bias-corrected accelerated; Efron 1987 in full), or
    \code{"normal"} (Wald: \eqn{\hat\theta \pm z \cdot SE}). Default
    is \code{NULL}, which triggers a per-type default:
    \code{"att"} -> \code{"normal"} (matches what \code{fit$est.att}
    already uses internally), \code{"att.cumu"} -> \code{"percentile"}
    (matches what \code{att.cumu()} does internally),
    \code{"aptt"} -> \code{"bca"} and \code{"log.att"} ->
    \code{"bca"} (ratio / log estimators benefit from joint
    bias and acceleration corrections when the bootstrap distribution
    is skewed; \code{"bc"} can degenerate at the boundary when the
    point estimate falls outside the bootstrap support). Pass an
    explicit value to override. \code{"bca"} requires bootstrap- or
    wild-bootstrap-sourced replicates.}
}

\value{
  A data frame with columns \code{<by_key>}, \code{estimate}, \code{se},
  \code{ci.lo}, \code{ci.hi}, \code{n_cells}, and \code{vartype}.
  Always tidy regardless of \code{type} or \code{by}.
}

\section{Numerical equality with existing slots}{
  \code{estimand(fit, "att", "event.time")} returns \code{estimate},
  \code{se}, \code{ci.lo}, \code{ci.hi} byte-identical to columns
  \code{ATT}, \code{S.E.}, \code{CI.lower}, \code{CI.upper} of
  \code{fit$est.att}, when default arguments are used. This invariant
  is asserted by package tests.
}

\seealso{
  \code{\link{imputed_outcomes}} for the underlying long-form accessor;
  \code{\link{fect}} for the fitting interface.
}

\examples{
\dontrun{
library(fect)
fit <- fect(Y ~ D, data = simdata, index = c("id", "time"),
            method = "fe", force = "two-way",
            se = TRUE, nboots = 200, parallel = FALSE)

## Default: per-event-time ATT (matches fit$est.att numerically).
est <- estimand(fit, "att", "event.time")
head(est)
}
}

\references{
  Chen, Jiafeng, and Jonathan Roth. 2024. "Logs with Zeros? Some Problems
  and Solutions." \emph{Quarterly Journal of Economics} 139 (2):
  891--936.
}

\encoding{UTF-8}
\name{fect-package}
\alias{fect-package}
\docType{package}
\title{Fixed Effects Counterfactual Estimators}
\description{
  The \code{fect} package implements counterfactual estimators for Time-Series Cross-Sectional (TSCS) data analysis, along with statistical tools to test their identification assumptions.
}
\details{
  The \code{fect} package provides tools for estimating treatment effects in TSCS datasets using a range of counterfactual estimators. These estimators first impute counterfactuals for treated observations by fitting an outcome model (fixed effects model, interactive fixed effects model, or matrix completion) to untreated observations. Then, the individual treatment effect for each treated observation is calculated as the difference between the observed and predicted counterfactual outcomes.

  The package supports:
  \itemize{
    \item Calculation of the Average Treatment Effect on the Treated (ATT) and period-specific ATTs.
    \item Placebo tests and equivalence tests to evaluate the validity of the identification assumptions.
    \item Robust estimation techniques for unbalanced panel datasets.
  }

  See \code{\link{fect}} for details.
}
\author{
  Licheng Liu (\email{liulch@mit.edu}), Massachusetts Institute of Technology
  
  Ye Wang (\email{yw1576@nyu.edu}), New York University
  
  Yiqing Xu (\email{yiqingxu@stanford.edu}), Stanford University
  
  Ziyi Liu (\email{zyliu2020@uchicago.edu}), University of Chicago
}
\references{
  Athey, S., Bayati, M., Doudchenko, N., Imbens, G., and Khosravi, K. (2021). 
  Matrix completion methods for causal panel data models. 
  \emph{Journal of the American Statistical Association, 116}(536), 1716-1730. 
  
  Bai, J. (2009). 
  Panel data models with interactive fixed effects. 
  \emph{Econometrica, 77}(4), 1229-1279.
    
  Liu, L., Wang, Y., and Xu, Y. (2022).
  A Practical Guide to Counterfactual Estimators for Causal Inference with Time-Series Cross-Sectional Data. 
  \emph{American Journal of Political Science, 68}(1), 160-176.
  
  Xu, Y. (2017). 
  Generalized Synthetic Control Method: Causal Inference with Interactive Fixed Effects Models. 
  \emph{Political Analysis, 25}(1), 57-76.
}

\encoding{UTF-8}
\name{plot.fect}
\alias{plot.fect}
\title{Plot Method for 'fect' Objects}
\description{
  Visualizes results from a \code{\link{fect}} estimation.
}
\usage{
\method{plot}{fect}(
    x,
    type = NULL,
    restrict = "rm",
    loo = FALSE,
    highlight = NULL,
    highlight.fill = FALSE,
    plot.ci = NULL,
    show.points = TRUE,
    loess.fit = TRUE,
    show.group = NULL,
    bound = NULL,
    show.count = TRUE,
    proportion = 0.3,
    pre.periods = NULL,
    f.threshold = NULL,
    tost.threshold = NULL,
    effect.bound.ratio = FALSE,
    stats = NULL,
    stats.labs = NULL,
    raw = "none",
    vis = NULL,
    main = NULL,
    xlim = NULL,
    ylim = NULL,
    xlab = NULL,
    ylab = NULL,
    xangle = NULL,
    yangle = NULL,
    xbreaks = NULL,
    ybreaks = NULL,
    xticklabels = NULL,
    yticklabels = NULL,
    gridOff = NULL,
    legendOff = FALSE,
    legend.pos = NULL,
    legend.nrow = NULL,
    legend.labs = NULL,
    stats.pos = NULL,
    show.stats = TRUE,
    theme.bw = TRUE,
    nfactors = NULL,
    include.FE = TRUE,
    id = NULL,
    cex.main = NULL,
    cex.main.sub = NULL,
    cex.axis = NULL,
    cex.lab = NULL,
    cex.legend = NULL,
    cex.text = NULL,
    axis.adjust = FALSE,
    axis.lab = "both",
    axis.lab.gap = c(0, 0),
    shade.post = FALSE,
    start0 = FALSE,
    return.test = FALSE,
    return.data = FALSE,
    balance = NULL,
    weight = NULL,
    lcolor = NULL,
    lwidth = NULL,
    ltype = NULL,
    line.color = NULL,
    line.width = NULL,
    count = NULL,
    preset = NULL,
    connected = NULL,
    ci.outline = FALSE,
    color =  NULL,
    pre.color = NULL,
    post.color = NULL,
    est.lwidth = NULL,
    est.pointsize = NULL,
    count.color =  NULL,
    count.alpha = NULL,
    count.outline.color = NULL,
    placebo.color =  NULL,
    carryover.color =  NULL,
    carryover.rm.color = NULL,
    sens.original.color = NULL,
    sens.colors =  NULL,
    counterfactual.color =  NULL,
    counterfactual.raw.controls.color =  NULL,
    counterfactual.raw.treated.color =  NULL,
    counterfactual.linetype =  NULL,
    box.control =  NULL,
    box.treat =  NULL,
    calendar.color = NULL,
    calendar.lcolor = NULL,
    calendar.cicolor = NULL,
    heterogeneous.color = NULL,
    heterogeneous.lcolor = NULL,
    heterogeneous.cicolor = NULL,
    equiv.color =  NULL,
    status.treat.color =  NULL,
    status.control.color = NULL,
    status.missing.color =  NULL,
    status.removed.color =  NULL,
    status.placebo.color = NULL,
    status.carryover.color =  NULL,
    status.carryover.rm.color =  NULL,
    status.balanced.post.color = NULL,
    status.balanced.pre.color = NULL,
    status.background.color = NULL,
    covariate = NULL,
    covariate.labels = NULL,
    covariate.value = NULL,
    covariate.value.range = FALSE,
    relative.time = FALSE,
    pretreatment = FALSE,
    num.pretreatment = 3,
    cm = FALSE,
    legacy.style = FALSE,
    ...
)
}
\arguments{
  \item{x}{A fitted \code{\link{fect}} object.}
  \item{type}{Plot type. Options include: \code{"gap"}, \code{"equiv"} (equivalence test), \code{"status"} (treatment status), \code{"exit"} (exiting treatment effects),
    \code{"factors"}, \code{"loadings"}, \code{"calendar"} (ATT by calendar time), \code{"box"} (individual effects distribution), \code{"counterfactual"},
    \code{"sens"} (sensitivity analysis plot, e.g., for Rambachan & Roth bounds), \code{"sens_es"} (event-study style sensitivity plot),
    or \code{"cumul"} (cumulative effect plot). Default is context-dependent (e.g., "gap", or "equiv" if \code{loo=TRUE}).}
  \item{restrict}{For sensitivity plots (\code{type = "sens"} or \code{type = "sens_es"}). Specifies the type of restriction: \code{"rm"} (restriction on M, for relative magnitude bounds) or \code{"sm"} (smoothness restriction, for bounds based on second differences). Default is \code{"rm"}.}
  \item{loo}{Logical; if \code{TRUE} (default is \code{FALSE}), use leave-one-out estimates for pre-treatment periods in equivalence tests or gap plots.}
  \item{highlight}{Controls which test-period types receive the colored-glyph treatment in placebo / carryover plots. Accepts:
    \itemize{
      \item \code{NULL} (default): auto --- highlight every test type that ran at fit time (placebo, carryover, carryover.rm).
      \item \code{TRUE}: same as \code{NULL} but explicit.
      \item \code{FALSE}: no highlighting; all periods render as plain pre/post circles.
      \item A character subset of \code{c("placebo", "carryover", "carryover.rm")}: highlight only the named test types; other test periods render as plain circles. For example, \code{highlight = "placebo"} on a fit with both \code{placeboTest = TRUE} and \code{carryoverTest = TRUE} draws orange triangles at placebo periods but renders carryover periods as ordinary post-treatment circles.
    }
  }
  \item{highlight.fill}{Logical. If \code{TRUE}, draws a lightened-tone background rectangle behind each highlighted period (in the same hue as the period's accent glyph). If \code{FALSE} (the default), only the colored glyph is drawn. The glyph alone is usually sufficient signal that a period belongs to a test window, and the glyph-only look survives grayscale printing without surprise. Set \code{TRUE} for slide / talk figures where the rectangle helps locate the window for a remote audience.}
  \item{plot.ci}{Specifies the confidence interval to be plotted. Options include:
    \code{"0.9"}, \code{"0.95"}, or \code{"none"}. Default is \code{NULL}, which is context-dependent (e.g., "0.95" for gap plots, "0.9" for equivalence plots, "none" if no CIs are available in the \code{fect} object).}
  \item{show.points}{Logical; if \code{TRUE} (default), shows point estimates on event study style plots (e.g., gap, equiv, exit plots).}
  \item{show.group}{Optional character string or \code{NULL} (default); if specified, plots results for a particular subgroup defined in the \code{fect} call.}
  \item{bound}{Which bounds to display in bounding/equivalence tests:
    \code{"none"} (default for most plots), \code{"min"} (minimal effect bound), \code{"equiv"} (equivalence range), or \code{"both"} (default for \code{type="equiv"}).}
  \item{show.count}{Logical; if \code{TRUE} (default), shows a histogram of observation counts (number of treated units) in relevant plots like gap, calendar, or box plots.}
  \item{proportion}{Numeric (0 to 1); for event study plots, restricts plotted time points to those where the number of treated units is at least this fraction of the maximum number of treated units observed across all periods. Default is \code{0.3}.}
  \item{pre.periods}{Optional numeric vector; specifies pre-treatment periods to be used for bounding and equivalence tests. Default is \code{NULL}, which uses the pre-treatment periods defined in the \code{fect} object.}
  \item{f.threshold}{Numeric or \code{NULL} (default); F-test threshold for equivalence checks. If \code{NULL}, uses the value from the \code{fect} object.}
  \item{tost.threshold}{Numeric or \code{NULL} (default); threshold for TOST-based equivalence checks (e.g., for pre-trend, placebo, or carryover tests). If \code{NULL}, uses the value from the \code{fect} object.}
  \item{effect.bound.ratio}{Logical; if \code{TRUE} (default is \code{FALSE}), shows the ratio of ATT to the minimal bound in equivalence plots.}
  \item{stats}{Character vector or \code{NULL} (default); specifies which test statistics to display on the plot (e.g., \code{"F.p"}, \code{"equiv.p"}, \code{"placebo.p"}). Default is context-dependent based on plot type and tests performed.}
  \item{stats.labs}{Character vector or \code{NULL} (default); custom labels for the displayed statistics. Must match the length of \code{stats}.}
  \item{raw}{For \code{type = "counterfactual"} plot: \code{"none"} (default; shows average treated and counterfactual lines), \code{"band"} (shows 5-95\% quantile band for raw control/treated data), or \code{"all"} (shows all raw control/treated unit trajectories).}
  \item{vis}{(Deprecated) Formerly controlled line display for placebo/carryover tests. This is now handled by \code{highlight} and internal logic.}
  \item{main}{Character string or \code{NULL} (default); main plot title. If \code{NULL}, a default title is used based on the plot type.}
  \item{xlim}{Numeric vector of length 2 or \code{NULL} (default); specifies x-axis limits.}
  \item{ylim}{Numeric vector of length 2 or \code{NULL} (default); specifies y-axis limits.}
  \item{xlab}{Character string or \code{NULL} (default); x-axis label. If \code{NULL}, a default label is used based on the plot type.}
  \item{ylab}{Character string or \code{NULL} (default); y-axis label. If \code{NULL}, a default label is used based on the plot type.}
  \item{xangle}{Numeric or \code{NULL} (default); angle (in degrees) for x-axis tick labels.}
  \item{yangle}{Numeric or \code{NULL} (default); angle (in degrees) for y-axis tick labels.}
  \item{xbreaks}{Numeric vector or \code{NULL} (default); positions of major ticks on the x-axis.}
  \item{ybreaks}{Numeric vector or \code{NULL} (default); positions of major ticks on the y-axis.}
  \item{xticklabels}{Character vector or \code{NULL} (default); custom labels for the x-axis ticks. Primarily used in status plots.}
  \item{yticklabels}{Character vector or \code{NULL} (default); custom labels for the y-axis ticks. Primarily used in status plots.}
  \item{gridOff}{Logical or \code{NULL} (default). If \code{TRUE}, removes major and minor grid lines. If \code{NULL} (which defaults to \code{FALSE} internally for this argument), the grid is off for most plots but on for \code{type = "status"} (where it forms tile outlines).}
  \item{legendOff}{Logical; if \code{TRUE} (default is \code{FALSE}), hides the legend.}
  \item{legend.pos}{Character string or \code{NULL} (default); position of the legend (e.g., "bottom", "top", "right", "left", "none", or c(x,y) coordinates). Default is typically "bottom".}
  \item{legend.nrow}{Integer or \code{NULL} (default); number of rows in the legend.}
  \item{legend.labs}{Character vector or \code{NULL} (default); custom legend labels.}
  \item{stats.pos}{Numeric vector of length 2 or \code{NULL} (default); (x,y) coordinates for displaying the text of test statistics on the plot.}
  \item{show.stats}{Logical; if \code{TRUE} (default), displays test statistics on the plot if \code{stats} is not "none".}
  \item{theme.bw}{Logical; if \code{TRUE} (default), uses a black-and-white ggplot2 theme (\code{theme_bw()}). If \code{FALSE} and no \code{preset} is specified, uses \code{theme_grey()}.}
  \item{nfactors}{Integer or \code{NULL} (default); number of factors to plot when \code{type} is \code{"factors"} or \code{"loadings"}. Defaults to \code{min(r.cv, 4)}, where \code{r.cv} is the number of estimated factors.}
  \item{include.FE}{Logical; if \code{TRUE} (default), includes fixed effects as additional "factors" in factor/loading plots if applicable (i.e., if unit or time fixed effects were estimated as part of the factor model).}
  \item{id}{Character vector or \code{NULL} (default); when specified, plots only the chosen unit(s). For counterfactual plots (\code{type="counterfactual"}), only one ID is typically supported for individual unit display; otherwise, averages are shown.}
  \item{cex.main}{Numeric or \code{NULL} (default); font size multiplier for the main plot title. Base size is 16pt.}
  \item{cex.main.sub}{Numeric or \code{NULL} (default); font size multiplier for a possible subtitle. Base size is 16pt.}
  \item{cex.axis}{Numeric or \code{NULL} (default); font size multiplier for axis tick labels. Base size is 15pt.}
  \item{cex.lab}{Numeric or \code{NULL} (default); font size multiplier for axis labels. Base size is 15pt.}
  \item{cex.legend}{Numeric or \code{NULL} (default); font size multiplier for legend text. Base size is 15pt.}
  \item{cex.text}{Numeric or \code{NULL} (default); font size multiplier for text annotations on the plot (e.g., statistics values). Base size is 5pt.}
  \item{axis.adjust}{Logical; if \code{TRUE} (default is \code{FALSE}), attempts to adjust axis labels for plots with many time points (e.g., by rotating x-axis labels).}
  \item{axis.lab}{Controls axis labeling in \code{type="status"} plots. Options: \code{"both"} (default), \code{"time"}, \code{"unit"}, or \code{"off"}.}
  \item{axis.lab.gap}{Numeric vector of length 1 or 2; gap spacing for axis labels in \code{type="status"} plots (sets frequency of labels). Default is \code{c(0, 0)} (label every tick).}
  \item{shade.post}{Logical; if \code{FALSE} (default). If \code{TRUE}, shades post-treatment periods in certain plots. For counterfactual plots, if \code{NULL} (the default for the argument is \code{FALSE}), it effectively becomes \code{TRUE} internally unless explicitly set to \code{FALSE}.}
  \item{start0}{Logical; if \code{TRUE} (default is \code{FALSE}), re-indexes the first post-treatment period as 0 in event-time plots (gap, equiv, exit). The period before treatment becomes -1.}
  \item{return.test}{Logical; if \code{TRUE} (default is \code{FALSE}), returns a list containing the plot object and relevant test statistics instead of just the plot.}
  \item{return.data}{Logical; if \code{TRUE} (default is \code{FALSE}), returns a data frame with the data used to construct the plot instead of returning a plot object. If \code{return.test=TRUE} as well, returns a list with \code{data} and \code{test.out}.}
  \item{balance}{Logical or \code{NULL} (default); if \code{TRUE}, uses a "balance sample" (if available from the \code{fect} object, e.g., from setting \code{balanced.sample = TRUE} in \code{fect}) for plotting.}
  \item{weight}{Logical or \code{NULL} (default); if \code{TRUE}, applies weighting (if available from the \code{fect} object, e.g., from providing weights in \code{fect}) to the estimates being plotted.}
  \item{lcolor}{Character vector of length 1 or 2, or \code{NULL} (default). If \code{NULL}, uses a default color (e.g., \code{"#AAAAAA70"} if \code{theme.bw=TRUE}, else \code{"white"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{lwidth}{Numeric vector of length 1 or 2, or \code{NULL} (default). If \code{NULL}, uses a default width (e.g., \code{1.5} if \code{theme.bw=TRUE}, else \code{2} for the base style). The \code{preset} argument may define a different width. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{ltype}{Character vector of length 1 or 2, or \code{NULL} (default). If \code{NULL}, uses a default linetype (e.g., \code{c("solid", "solid")} for the base style). The \code{preset} argument may define a different linetype (e.g., \code{preset="vibrant"} uses \code{c("solid", "dashed")}). Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{line.color}{(Deprecated) Use \code{color} for main estimate lines or \code{lcolor} for reference lines instead.}
  \item{line.width}{(Deprecated) Use \code{est.lwidth} for main estimate lines or \code{lwidth} for reference lines instead.}
  \item{count}{(Deprecated) Use \code{show.count} to control visibility and \code{count.color}, \code{count.alpha}, \code{count.outline.color} for appearance.}
  \item{preset}{A character string specifying a color and style preset: \code{NULL} (default, uses a base style, often with \code{theme_bw} if \code{theme.bw=TRUE}), \code{"vibrant"}, or \code{"grayscale"}. This sets a group of defaults for many plot elements. Individual appearance arguments (like \code{color}, \code{lcolor}, etc.) can override settings from a \code{preset} or the base style.}
  \item{connected}{Logical or \code{NULL} (default). For event study type plots (e.g., \code{type="gap"}), if \code{TRUE}, connects points with lines; if \code{FALSE}, shows point-ranges. If \code{NULL}, defaults to \code{FALSE} for the base style, and \code{TRUE} if \code{preset = "vibrant"}.}
  \item{ci.outline}{Logical; if \code{FALSE} (default). For event study plots where CIs are shown as ribbons, if \code{TRUE}, adds an outline to the CI ribbon.}
  \item{color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for main plot lines/points (e.g., \code{"black"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{pre.color}{Character string or \code{NULL} (default). Color for pre-treatment period estimates. If \code{NULL}, uses the value of \code{color}.}
  \item{post.color}{Character string or \code{NULL} (default). Color for post-treatment period estimates. If \code{NULL}, uses the value of \code{color}.}
  \item{est.lwidth}{Numeric or \code{NULL}; line width for the main estimate line in event study plots. If \code{NULL} (default), the width is determined based on \code{connected} and \code{show.points} (e.g., `0.6` for point-ranges with the base style, `0.7` for connected points with markers with the base style). The \code{preset} argument may influence this default. Specifying a numeric value directly overrides any other setting.}
  \item{est.pointsize}{Numeric or \code{NULL}; size of points for the main estimate in event study plots. If \code{NULL} (default), the size is determined based on \code{connected} and \code{show.points} (e.g., `2` for point-ranges with the base style, `1.2` for connected points with markers with the base style). The \code{preset} argument may influence this default. Specifying a numeric value directly overrides any other setting.}
  \item{count.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for count bars (e.g., \code{"grey70"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{count.alpha}{Numeric (0-1) or \code{NULL} (default). If \code{NULL}, uses a default alpha for count bars (e.g., \code{0.4} for the base style). The \code{preset} argument may define a different alpha. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{count.outline.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default outline color for count bars (e.g., \code{"grey69"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{placebo.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for placebo elements (e.g., \code{"blue"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{carryover.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for carryover elements (e.g., \code{"red"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{carryover.rm.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for removed carryover elements (e.g., \code{"blue"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{sens.original.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for original estimates in sensitivity plots (e.g., \code{"darkblue"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{sens.colors}{Vector of colors or \code{NULL} (default). If \code{NULL}, uses a default palette for sensitivity bounds (e.g., \code{c("#218C23","#FF34B4","#FF521B","#2B59C3")} for the base style). The \code{preset} argument may define a different palette. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{counterfactual.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for counterfactual lines (e.g., \code{"steelblue"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{counterfactual.raw.controls.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for raw control lines (e.g., \code{"#4682B420"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{counterfactual.raw.treated.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for raw treated lines (e.g., \code{"#77777750"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{counterfactual.linetype}{Character string or \code{NULL} (default). If \code{NULL}, uses a default linetype for counterfactual lines (e.g., \code{"longdash"} for the base style). The \code{preset} argument may define a different linetype. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{box.control}{Character string or \code{NULL} (default). If \code{NULL}, uses a default fill color for control boxplots (e.g., \code{"skyblue"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{box.treat}{Character string or \code{NULL} (default). If \code{NULL}, uses a default fill color for treated boxplots (e.g., \code{"pink"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{calendar.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for fitted lines in calendar plots (e.g., \code{"skyblue"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{calendar.lcolor}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for average ATT lines in calendar plots (e.g., \code{"red"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{calendar.cicolor}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for CI in calendar plots (e.g., \code{"red"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{heterogeneous.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for fitted lines in heterogeneous plots (e.g., \code{"red"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{heterogeneous.lcolor}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for average ATT lines in heterogeneous plots (e.g., \code{"red"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{heterogeneous.cicolor}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for CI in heterogeneous plots (e.g., \code{"red"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{equiv.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for equivalence bounds (e.g., \code{"red"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{status.treat.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for treated status (e.g., \code{"#06266F"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{status.control.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for control status (e.g., \code{"#B0C4DE"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{status.missing.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for missing status (e.g., \code{"#FFFFFF"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{status.removed.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for removed status (e.g., \code{"#A9A9A9"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{status.placebo.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for placebo status (e.g., \code{"#66C2A5"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{status.carryover.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for carryover status (e.g., \code{"#E78AC3"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{status.carryover.rm.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for removed carryover status (e.g., \code{"#ffc425"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{status.balanced.post.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for balanced post-treatment status (e.g., \code{"#00852B"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{status.balanced.pre.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default color for balanced pre-treatment status (e.g., \code{"#A5CA18"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{status.background.color}{Character string or \code{NULL} (default). If \code{NULL}, uses a default background color for status plots (e.g., \code{"gray90"} for the base style). The \code{preset} argument may define a different color. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{covariate}{Character string or \code{NULL} (default). If \code{NULL}, uses a default covariate for heterogeneous plots (e.g., \code{"X"} for the base style). The \code{preset} argument may define a different covariate. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{covariate.labels}{Character vector or \code{NULL} (default). If \code{NULL}, uses a default covariate labels for heterogeneous plots (e.g., \code{"X1"} for the base style). The \code{preset} argument may define a different covariate labels. Specifying this argument directly overrides any \code{preset} or base style setting.}
  \item{covariate.value}{Numeric or \code{NULL} (default); a specific covariate value at which to evaluate the HTE plot. If \code{NULL}, the full range is used.}
  \item{covariate.value.range}{Logical; if \code{TRUE}, plots HTE over the range of covariate values. Default is \code{FALSE}.}
  \item{relative.time}{Logical; if \code{TRUE}, uses relative time (time since treatment) on the x-axis for HTE plots. Default is \code{FALSE}.}
  \item{pretreatment}{Logical; if \code{TRUE}, restricts HTE analysis to pre-treatment periods only. Default is \code{FALSE}.}
  \item{num.pretreatment}{Integer; number of pre-treatment periods to include when \code{pretreatment = TRUE}. Default is \code{3}.}
  \item{cm}{Logical; if \code{TRUE}, plots causal moderation estimates instead of effect modification estimates for HTE plots. Requires the \code{fect} object to have been estimated with \code{cm = TRUE}. Default is \code{FALSE}.}
  \item{legacy.style}{Logical; if \code{TRUE}, reproduces the pre-2.3.1 visual defaults (bold centered title, larger axis text, solid treatment-onset line, blue placebo triangles, no peach highlight rectangle) regardless of the value of \code{theme.bw}. If \code{FALSE} (the default) and \code{theme.bw = TRUE}, the modern v2.3.1 visual recipe is used. Pass \code{legacy.style = TRUE} for byte-identical reproduction of figures rendered under earlier fect versions.}
  \item{loess.fit}{Logical; if \code{TRUE} (default), overlays a LOESS smoothing curve on HTE scatter plots.}
  \item{...}{Additional graphical parameters passed to internal plotting routines, primarily those accepted by \code{esplot} for event study style plots (gap, equiv, exit, sens_es, cumul).}
}
\details{
  \code{plot.fect} generates various visualizations for objects estimated via \code{\link{fect}}. Depending on
  the selected \code{type}, it can show treatment-effect dynamics (\code{"gap"}, \code{"exit"}, \code{"cumul"}),
  equivalence test outcomes (\code{"equiv"}), factor/loadings for factor models (\code{"factors"}, \code{"loadings"}),
  raw vs. counterfactual trajectories (\code{"counterfactual"}), treatment status (\code{"status"}),
  effects by calendar time (\code{"calendar"}), distribution of individual effects (\code{"box"}),
  or sensitivity analysis results (\code{"sens"}, \code{"sens_es"}).
  The \code{preset} argument allows for quick customization of plot aesthetics using predefined color schemes.
  Many individual color and style parameters can be set to further customize the appearance, and these will override any settings from a \code{preset} or the base style.
  The function heavily relies on \code{ggplot2} for plotting and \code{esplot} for event-study style visualizations.
}
\value{
  By default, a \code{ggplot2} object representing the plot.

  If \code{return.data=TRUE}, returns a list containing:
  \item{p}{The \code{ggplot2} object.}
  \item{data}{A named list of data frames containing the data used to construct each layer of the plot (no extra all-NA columns are added).}
  If \code{return.test=TRUE} as well, the list also contains:
  \item{test.out}{A list or data frame containing relevant test statistics (such as F-tests, equivalence p-values, placebo test p-values, etc.), if applicable to the plot type and options chosen.}

  If \code{return.test=TRUE} and \code{return.data=FALSE}, returns a list containing \code{p} and \code{test.out}.
}
\author{
  Licheng Liu, Ye Wang, Yiqing Xu, Ziyi Liu, Rivka Lipkovitz
}
\references{
  Xu, Y. (2017). Generalized Synthetic Control Method: Causal Inference with Interactive Fixed Effects Models.
  \emph{Political Analysis, 25}(1), 57–76.

  Liu, L., Wang, Y., & Xu, Y. (2022). A Practical Guide to Counterfactual Estimators for Causal Inference with Time-Series Cross-Sectional Data.
  \emph{American Journal of Political Science, 66}(1), 220-237. (Provides context for \code{esplot} which is used internally for many plot types)

  Rambachan, A., & Roth, J. (2023). A More Credible Approach to Parallel Trends.
  \emph{Review of Economic Studies, 90}(5), 2555-2591. (Provides context for sensitivity analysis plots like \code{type="sens"} and \code{type="sens_es"})
}
\examples{
library(fect)
# For CRAN checks, use a small number of bootstraps
# In practice, use a larger number (e.g., nboots = 200 or more)
if(requireNamespace("ggplot2") && requireNamespace("ggrepel")) {
  data(simdata)
  # Estimate with fixed effects method
  out.fect <- fect(
    Y ~ D + X1 + X2,
    data = simdata,
    index = c("id","time"),
    method = "fe",
    force = "two-way",
    se = TRUE,
    parallel = FALSE,
    nboots = 5 # nboots low for example
  )

  # Default gap plot
  plot(out.fect, main = "Estimated ATT (FEct)", ylab = "Effect of D on Y")

  # Gap plot with vibrant preset and custom line color
  # plot(out.fect, preset = "vibrant", color = "darkgreen",
  #      main = "Estimated ATT (Vibrant Preset, Custom Line)")

  # Counterfactual plot for the first treated unit
  # Need to know the ID of a treated unit. Let's find one.
  treated_ids <- unique(simdata$id[simdata$D == 1])
  if (length(treated_ids) > 0) {
    plot(out.fect, type = "counterfactual", id = treated_ids[1],
         main = paste("Counterfactual for Unit", treated_ids[1]))
  }

  # Status plot
  plot(out.fect, type = "status")

  # Cumulative effect plot (if est.eff is available from fect call)
  # This example might not have it by default, but showing how to call
  # out.fect.cumul <- fect(Y ~ D, data = simdata, index = c("id","time"), method = "fe",
  #                        cumulative = TRUE, se = TRUE, parallel = FALSE, nboots = 5)
  # if (exists("out.fect.cumul")) {
  #   plot(out.fect.cumul, type = "cumul", main = "Cumulative ATT")
  # }

  # Example for sensitivity plot (requires IFE/GSYNTH method and sensitivity analysis)
  # \donttest{
  #   out.ife <- fect(Y ~ D, data = simdata, index = c("id","time"),
  #                   method = "ife", se = TRUE, r = 2,
  #                   sensitivity.analysis = TRUE, sensitivity.plot = FALSE, # run analysis
  #                   parallel = FALSE, nboots = 5) # nboots low for example
  #   if (!is.null(out.ife$sensitivity.rm)) {
  #     plot(out.ife, type = "sens", restrict = "rm",
  #          main = "Sensitivity Analysis (Relative Magnitude)")
  #     plot(out.ife, type = "sens_es", restrict = "rm",
  #          main = "Event-Study Sensitivity (Relative Magnitude)")
  #   }
  # }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cv-rolling.R
\name{r.cv.rolling}
\alias{r.cv.rolling}
\title{Rolling-window cross-validation for rank selection}
\usage{
r.cv.rolling(
  formula,
  data,
  index,
  method = c("ife", "gsynth", "cfe"),
  r.max = 5L,
  cv.nobs = 3L,
  cv.buffer = 1L,
  k = 20L,
  cv.prop = 0.1,
  cv.rule = c("1se", "min", "1pct"),
  min.T0 = 5L,
  force = "unit",
  seed = NULL,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{formula}{A model formula, e.g. `Y ~ D + X1 + X2`.}

\item{data}{A long-format data frame.}

\item{index}{Character vector identifying the panel structure. For
`method = "ife"` and `method = "gsynth"`, length 2:
`c("unit", "time")`. For `method = "cfe"`, length >= 2: the first
two entries are `c("unit", "time")` and any additional entries are
extra grouping fixed-effect columns forwarded to the inner
`fect()` call's `index =` argument.}

\item{method}{Estimator. One of `"ife"` (IFE-EM, internally
`time.component.from = "notyettreated"`), `"gsynth"` (GSC,
internally `time.component.from = "nevertreated"`), or `"cfe"`
(Complex Fixed Effects, internally
`time.component.from = "notyettreated"`). All three paths populate
`Y.ct.full` at masked positions, so MSPE scoring works uniformly.
For CFE, rolling CV picks `r` only; CFE-specific arguments
(`Z`, `gamma`, `Q`, `Q.type`, `kappa`, extra index columns) are
forwarded via `...` and held fixed at their user-supplied values.}

\item{r.max}{Largest candidate rank to evaluate. CV is run over
`0:r.max`.}

\item{cv.nobs}{Length of the held-out (scored) block per unit per
fold. Default 3.}

\item{cv.buffer}{Number of observations immediately BEFORE the
held-out block to drop from training (the past-side buffer that
attenuates AR-leakage). Default 1. Analogous to `cv.donut` in the
existing `cv.method = "all_units"` / `"treated_units"` strategies,
but applied only on the past side: the future side is dropped by
construction.}

\item{k}{Number of folds. Each fold draws a fresh sample of units
and a fresh set of per-unit anchors; the per-r MSPE is averaged
across folds and the SE used by the `"1se"` rule reflects
fold-to-fold variability. Default 20 (matches the default for
the existing CV strategies).}

\item{cv.prop}{Fraction of eligible units sampled per fold. Only
sampled units receive a mask in that fold; the rest stay fully
observed and contribute training data at every period. Default
0.1 (paired with `k = 20` for ~2x coverage of every eligible
unit across folds). Across `k` folds, every eligible unit lands
in the holdout roughly `k * cv.prop` times in expectation. Must
satisfy `0 < cv.prop <= 1`. On small panels (n_eligible < 30)
consider raising further, since per-fold MSPE precision scales
with `cv.prop * n_eligible * cv.nobs`.}

\item{cv.rule}{Rule for picking `r` from the MSPE curve: `"1se"`
(default), `"min"`, or `"1pct"`.}

\item{min.T0}{Minimum observations required strictly before the
anchor. Sets the lower bound on valid anchor positions. Default 5.}

\item{force}{One of `"none"`, `"unit"`, `"time"`, `"two-way"`. Default
`"unit"`.}

\item{seed}{Optional integer base seed; per-fold seeds derive from
`seed + fold_id` for reproducibility. Default `NULL` (use the
ambient RNG).}

\item{verbose}{If TRUE (default), print per-fold per-r MSPE.}

\item{...}{Additional arguments forwarded to `fect()`. For
`method = "cfe"`, the user holds CFE structural arguments (`Z`,
`gamma`, `Q`, `Q.type`, `Q.bspline.degree`, `kappa`, etc.) fixed
at their spec via `...`; rolling CV varies only `r`.}
}
\value{
List with components:
  - `r.cv`: chosen rank.
  - `cv.rule`: rule applied.
  - `mspe`: data.frame of per-r MSPE (averaged across folds), SE
    across folds, and held-out cell counts.
  - `mspe.per.fold`: r-by-k matrix of per-fold MSPE.
  - `k`, `cv.nobs`, `cv.buffer`, `cv.prop`: parameters used.
  - `n.units.masked`: distinct units that contributed to at least
    one fold's holdout.
}
\description{
Picks the number of factors `r` for an interactive-fixed-effects
model via standard rolling-window cross-validation. For each of `k`
folds, a fraction `cv.prop` of eligible units (controls plus
treated pre-treatment) is sampled; only sampled units carry a mask.
For each sampled unit, a random anchor time `t*` is drawn and the
fold's training set excludes `t* - cv.buffer, ..., t* - 1` (buffer),
`t*, ..., t* + cv.nobs - 1` (the held-out, scored block), and
`t* + cv.nobs, ..., end_of_eligible(t)` (rolling-window future drop;
for treated units, `end_of_eligible` is the cell strictly before
treatment onset). MSPE is scored at the held-out block only and
averaged across folds.
}
\details{
Per-fold unit sampling is required: masking every eligible unit at
the same time leaves no donor data at the masked time points and
breaks factor identification. Sampling `cv.prop` of units per fold
keeps unsampled units fully observed at all periods.

This is the standard time-series CV design (cf. `forecast::tsCV`,
`tidymodels::sliding_window`, `caret::createTimeSlices`) adapted to
panel data: each sampled unit gets its own anchor per fold, drawn
uniformly from valid positions.
}
\examples{
\dontrun{
  library(fect)
  data(simdata)
  res <- r.cv.rolling(Y ~ D, data = simdata, index = c("id", "time"),
                      method = "ife", r.max = 5,
                      cv.nobs = 3, cv.buffer = 1, k = 20)
  res$r.cv
  ## then use the chosen r in a CV-disabled fit:
  fit <- fect(Y ~ D, data = simdata, index = c("id", "time"),
              method = "ife", time.component.from = "notyettreated",
              CV = FALSE, r = res$r.cv, se = TRUE)
}

}

\name{fect_sens}
\alias{fect_sens}
\title{Sensitivity Analysis for fect Objects under Relative Magnitude and Smoothness Restrictions}
\description{
Conducts sensitivity analyses on \code{fect} model objects under relative magnitude (RM) and smoothness (C-LF) assumptions, producing robust confidence intervals and parameter sets.
}
\usage{
fect_sens(
  fect.out,
  post.periods = NA,
  l_vec = NA,
  Mbarvec = seq(0, 1, by = 0.1),
  Mvec = seq(0, 0.25, 0.05),
  periodMbarvec = c(0, 0.5),
  periodMvec = c(0, 0.1),
  parallel = FALSE,
  cores = NULL
)
}
\arguments{
  \item{fect.out}{A fitted \code{fect} object.}
  \item{post.periods}{Vector of post-treatment periods for sensitivity analysis. Default uses all available post-treatment periods.}
  \item{l_vec}{Optional weighting vector for averaging ATT across post-treatment periods. Default weights by treated-unit counts.}
  \item{Mbarvec}{Values of \code{Mbar} for overall RM-based sensitivity analysis.}
  \item{Mvec}{Values of \code{M} for overall smoothness-based sensitivity analysis.}
  \item{periodMbarvec}{Values of \code{Mbar} for period-specific RM sensitivity analysis.}
  \item{periodMvec}{Values of \code{M} for period-specific smoothness sensitivity analysis.}
  \item{parallel}{Logical; if \code{TRUE}, uses parallel computation where supported. Default is \code{FALSE}.}
  \item{cores}{Optional integer; number of workers when \code{parallel = TRUE}.}
}
\details{
This function:
\enumerate{
  \item Extracts ATT estimates and variance-covariance matrices from \code{fect.out}.
  \item Constructs weighted averages across post-treatment periods.
  \item Computes robust confidence sets under RM and C-LF assumptions.
  \item Optionally computes robust bounds for each post-treatment period separately.
}
Robust sets are computed using functions from the \pkg{HonestDiDFEct} package.
}
\value{
An updated \code{fect} object including:
\item{sensitivity.rm}{Relative Magnitude (RM) sensitivity results for average and period-by-period ATT.}
\item{sensitivity.smooth}{Smoothness (C-LF) sensitivity results for average and period-by-period ATT.}
}
\author{
Rivka Lipkovitz
}
\examples{
\dontrun{
out <- fect(Y ~ D, data = df, index = c("unit", "time"), method = "fe", se = TRUE)
out_sens <- fect_sens(
  fect.out = out,
  post.periods = c(1, 2, 3, 4),
  Mbarvec = seq(0, 1, by = 0.2),
  Mvec = c(0, 0.05, 0.1),
  periodMbarvec = c(0, 0.5),
  periodMvec = c(0, 0.1)
)
names(out_sens$sensitivity.rm)
names(out_sens$sensitivity.smooth)
}
}

\encoding{UTF-8}
\name{interFE}
\alias{interFE}
\title{Interactive Fixed Effects Models}
\description{Estimating interactive fixed effect models.}
\usage{interFE(formula = NULL, data, Y, X = NULL, W = NULL,
         index, r = 0, force = "two-way",
         se = FALSE, nboots = 500, seed = NULL, 
         tol = 1e-3, max_iteration = 500, 
         binary = FALSE, QR = FALSE, normalize = FALSE)
}
\arguments{
  \item{formula}{an object of class "formula": a symbolic description of the model to be fitted.}
  \item{data}{a data frame.}
  \item{Y}{outcome variable.}
  \item{X}{time-varying covariates.}
  \item{W}{weights.}
  \item{index}{a two-element string vector specifying the unit (group) and time indicators.}
  \item{r}{an integer specifying the number of factors.}
  \item{force}{a string indicating whether unit/time/both fixed effects will be imposed: "none", "unit", "time", or "two-way".}
  \item{se}{logical; if \code{TRUE}, computes bootstrap-based standard errors.}
  \item{nboots}{number of bootstrap runs (ignored if \code{se=FALSE}).}
  \item{seed}{random seed.}
  \item{tol}{tolerance for EM algorithm.}
  \item{max_iteration}{max number of EM iterations.}
  \item{binary}{logical flag for a probit link (not fully supported here).}
  \item{QR}{logical flag for QR-based factor analysis in probit model (not fully supported).}
  \item{normalize}{logical; if \code{TRUE}, scales outcome/covariates.}
}
\details{
  \code{interFE} estimates interactive fixed effect models as in Bai (2009).
}
\value{
  \item{beta}{estimated coefficients.}
  \item{mu}{estimated grand mean.}
  \item{factor}{estimated time-varying factors.}
  \item{lambda}{estimated factor loadings.}
  \item{VNT}{diagonal matrix of r eigenvalues.}
  \item{niter}{number of iterations before convergence.}
  \item{alpha}{estimated unit fixed effects (if imposed).}
  \item{xi}{estimated time fixed effects (if imposed).}
  \item{residuals}{model residuals.}
  \item{sigma2}{residual mean squared error.}
  \item{IC}{information criterion.}
  \item{ValidX}{logical for whether valid covariates exist.}
  \item{dat.Y}{matrix of outcome data.}
  \item{dat.X}{array of independent variables.}
  \item{Y}{name of the outcome variable.}
  \item{X}{name of time-varying control variables.}
  \item{index}{name of unit/time indicators.}
  \item{est.table}{table of final estimates.}
  \item{est.boot}{matrix of bootstrap results.}
}
\author{
  Licheng Liu; Ye Wang; Yiqing Xu
}
\references{  
  Bai, J. (2009). 
  Panel data models with interactive fixed effects. 
  \emph{Econometrica, 77}(4), 1229-1279.
}
\seealso{
  \code{\link{print.interFE}}, \code{\link{fect}}
}
\examples{
library(fect)
data(simdata)
d <- simdata[-(1:150),] # remove the treated units
out <- interFE(Y ~ X1 + X2, data = d, index=c("id","time"),
               r = 2, force = "two-way", nboots = 50)
}

\encoding{UTF-8}
\name{att.cumu}
\alias{att.cumu}
\title{Calculate Cumulative Treatment Effects}
\description{
  Calculate cumulative treatment effects based on the results of a \code{\link{fect}} object.
}
\usage{
att.cumu(x, period = NULL, weighted = TRUE, alpha = 0.05, type = "on", plot = FALSE)
}
\arguments{
  \item{x}{A \code{\link{fect}} object.}
  \item{period}{A two-element numeric vector specifying the range of terms during which treatment effects are to be accumulated, e.g., \code{period = c(-1, 1)}.}
  \item{weighted}{A logical flag specifying whether to calculate weighted cumulative treatment effects based on counts at each period. Default is \code{TRUE}.}
  \item{alpha}{A numerical value specifying the significance level. Default is \code{0.05}.}
  \item{type}{A string that specifies the type of effect to calculate. Must be one of the following: \code{"on"} (switch-on treatment effect) or \code{"off"} (switch-off treatment effect). Default is \code{"on"}.}
  \item{plot}{A logical flag indicating whether to plot cumulative effects. Default is \code{FALSE}.}
}
\author{
  Licheng Liu, Ye Wang, and Yiqing Xu
}
\references{
  Athey, S., Bayati, M., Doudchenko, N., Imbens, G., and Khosravi, K. (2021).
  Matrix completion methods for causal panel data models.
  \emph{Journal of the American Statistical Association}, 116(536), 1716-1730.

  Bai, J. (2009).
  Panel data models with interactive fixed effects.
  \emph{Econometrica}, 77(4), 1229-1279.

  Liu, L., Wang, Y., and Xu, Y. (2022).
  A Practical Guide to Counterfactual Estimators for Causal Inference with Time-Series Cross-Sectional Data.
  \emph{American Journal of Political Science}, 68(1), 160-176.

  Xu, Y. (2017).
  Generalized Synthetic Control Method: Causal Inference with Interactive Fixed Effects Models.
  \emph{Political Analysis}, 25(1), 57-76.
}
\seealso{
  \code{\link{fect}}, \code{\link{plot.fect}}
}

\encoding{UTF-8}
\name{fect-internal}
\alias{XXinv}
\alias{beta_iter}
\alias{beta_iter_ub}
\alias{data_ub_adj}
\alias{Y_demean}
\alias{fe_add}
\alias{fe_add2}
\alias{fe_ad_iter}
\alias{fe_ad_covar_iter}
\alias{fe_ad_inter_iter}
\alias{fe_ad_inter_covar_iter}
\alias{_gsynth_XXinv}
\alias{_gsynth_beta_iter}
\alias{_gsynth_beta_iter_ub}
\alias{_gsynth_inter_fe}
\alias{_gsynth_inter_fe_ub}
\alias{_gsynth_inter_fe_mc}
\alias{_gsynth_data_ub_adj}
\alias{_gsynth_panel_beta}
\alias{_gsynth_panel_est}
\alias{_gsynth_panel_factor}
\alias{_gsynth_panel_factor_ub}
\alias{_gsynth_panel_fe}
\alias{_gsynth_panel_fe_ub}
\alias{_gsynth_Y_demean}
\alias{_gsynth_fe_add}
\alias{_gsynth_fe_add2}
\alias{_gsynth_fe_ad_iter}
\alias{_gsynth_fe_ad_covar_iter}
\alias{_gsynth_fe_ad_inter_iter}
\alias{_gsynth_fe_ad_inter_covar_iter}
\alias{inter_fe}
\alias{inter_fe_ub}
\alias{inter_fe_mc}
\alias{panel_beta}
\alias{panel_est}
\alias{panel_factor}
\alias{panel_factor_ub}
\alias{panel_fe}
\alias{panel_fe_ub}
\alias{fect.default}
\alias{fect.formula}
\alias{fect_fe}
\alias{fect_mc}
\alias{interFE.default}
\alias{interFE.formula}
\alias{cv.sample}
\alias{equiv_test}
\title{Internal FEct Functions}
\description{
 Internal fect functions
}
\details{
These are not to be called by the user (or in some cases are just
waiting for proper documentation to be written :).
}
\keyword{ts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fect_iden.R
\name{fect_iden}
\alias{fect_iden}
\title{Over-identification test for causal moderation (cm)}
\usage{
fect_iden(
  x,
  moderator,
  covariates = NULL,
  quadratic = TRUE,
  interaction = TRUE
)
}
\arguments{
\item{x}{A `fect` object. Must be estimated with `cm=TRUE` so that `x$est.cm` exists.}

\item{moderator}{Character scalar. Name of the moderator \eqn{M_{it}} (must be in `x$X`).}

\item{covariates}{Optional character vector. Names of other covariates \eqn{X_{it}}.
Default is all covariates in `x$X` except `moderator`.}

\item{quadratic}{Logical. Include quadratic terms \eqn{M_{it}^2} and \eqn{X_{it}^2}. Default TRUE.}

\item{interaction}{Logical. Include interaction terms \eqn{M_{it} \times X_{it}}. Default TRUE.}
}
\value{
A list with elements `e1` and `e0`, each containing:
  - `n`: sample size used in the auxiliary regression
  - `r2`: R-squared of the auxiliary regression
  - `stat`: \eqn{n \times R^2}
  - `df`: degrees of freedom (# nonlinear terms kept)
  - `p`: p-value from chi-square test
  - `model`: the fitted `fixest` model
}
\description{
Implements the over-identification test described in the slide:
run two auxiliary regressions of residualized outcomes on the moderator
and covariates (including quadratic and interaction terms) with unit and
time fixed effects, then test using \eqn{n \times R^2 \sim \chi^2(df)},
where df is the number of *nonlinear* (quadratic + interaction) terms that
remain in the fitted model.
}

\name{get.cohort}
\alias{get.cohort}
\title{Obatin the Cohort Index}
\description{Obtain the cohort index based on the provided panel data.}
\usage{
  get.cohort(data, D, index, 
                  varname = NULL, 
                  start0 = FALSE,
                  entry.time = NULL,
                  drop.always.treat = FALSE)
}
\arguments{
    \item{data}{a data frame, can be a balanced or unbalanced panel data.}
    \item{D}{the treatment indicator. The treatment should be binary (0 and 1).}
    \item{index}{a two-element string vector specifying the unit and time indicators. Must be of length 2. Every observation should be uniquely defined by the pair of the unit and time indicator.}
    \item{varname}{a vector of strings that specifies the variable names for the cohort index and relative periods to be generated.}
    \item{start0}{a logical flag that indicates whether period 0 is the first post-treatment period. By default, it is set to FALSE, meaning that period 1 is considered the initial post-treatment period.}
    \item{entry.time}{a list of intervals representing the initial calendar time of treatment for each cohort. Units that received treatment within each specified interval are categorized as belonging to a single cohort.}  
    \item{drop.always.treat}{a logical flag that indicates whether to drop all always-treated units.}  
}
\details{
  \code{get.cohort} function preprocesses the data and generates the cohort index for different groups. If not specified in \code{varname}, the function automatically creates four new variables: 'FirstTreat' stores the initial calendar time when a unit is first treated, 'Cohort' stores the cohort time, 'Time_to_Treatment' tracks the period relative to the treatment, and 'Time_to_Exit' tracks the period relative to the treatment exit. Users have the option to specify custom names for these newly generated variables using the \code{varname} parameter. By default, the function sets \code{start0=FALSE}, where the relative period 1 corresponds to the first post-treatment/post-exit-treatment period. If \code{start0=TRUE}, period 0 becomes the first post-treatment/post-exit-treatment period. Additionally, users can categorize units into specific cohorts using the \code{entry.time} parameter. For instance, if \code{entry.time=list(c(1,5),c(6,10))}, units that receive their first treatment within the calendar intervals \code{c(1,5)} and \code{c(6,10)} are grouped into two distinct cohorts. Units that have never been treated are always categorized as the "control cohort".  
}
\value{
  \item{data}{a new data frame containing the cohort index.}
}
\author{
  Licheng Liu; Yiqing Xu, Ziyi Liu; Zhongyu Yin
}

\name{effect}
\alias{effect}
\title{Calculate Cumulative or Sub-group Treatment Effects}
\description{
  Calculates cumulative or average treatment effects for specified units and time periods based on a fitted \code{fect} object. The function supports both cumulative effects over time and period-specific average treatment effects, with bootstrap-based uncertainty estimates.
}

\usage{
effect(
  x,
  cumu = TRUE,
  id = NULL,
  period = NULL,
  plot = FALSE,
  count = TRUE,
  xlab = NULL,
  ylab = NULL,
  main = NULL
)
}

\arguments{
  \item{x}{A \code{fect} object containing treatment effect estimates and bootstrap results.}

  \item{cumu}{Logical. If \code{TRUE} (default), calculates cumulative treatment effects. If \code{FALSE}, calculates period-specific average treatment effects.}

  \item{id}{Character vector or NULL. Unit identifiers to include in the analysis. If \code{NULL} (default), all treated units are included.}

  \item{period}{Numeric vector of length 2 specifying the time window \code{c(start, end)} for effect calculation. If \code{NULL}, uses the maximum possible window based on the data.}

  \item{plot}{Logical. If \code{TRUE}, creates a visualization of the cumulative treatment effects with confidence intervals and a bar chart showing the number of treated units at each time point. Default is \code{FALSE}.}

  \item{count}{Logical. If \code{TRUE}, shows the count bars in the plot.}

  \item{xlab}{Character. X-axis label for the plot.}

  \item{ylab}{Character. Y-axis label for the plot.}

  \item{main}{Character. Main title for the plot.}
}

\details{
  The function processes treatment effects in several steps:

  1. Selects units based on the \code{id} parameter or includes all treated units if \code{id = NULL}.

  2. Calculates relative time to treatment for each unit.

  3. If \code{cumu = TRUE}, computes cumulative effects by summing average effects up to each period.

  4. Performs bootstrap analysis to estimate uncertainty (standard errors, confidence intervals, and p-values).

  The function supports different inference methods (bootstrap, jackknife, parametric) and adjusts calculations accordingly.

  Note: The function requires bootstrap results in the input \code{fect} object (\code{keep.sims = TRUE} must be set when fitting the model).
}

\value{
  Returns a list containing:
  \item{eff}{Vector of point estimates for cumulative or average treatment effects.}
  \item{est.eff}{Matrix containing the following columns:
    \itemize{
      \item ATT: Point estimates
      \item S.E.: Standard errors
      \item CI.lower: Lower bound of confidence interval
      \item CI.upper: Upper bound of confidence interval
      \item p.value: Two-sided p-values
    }
  }
}

\section{Warning}{
  The function will stop with an error if:
  \itemize{
    \item No bootstrap results are available in the input object
    \item The panel contains treatment reversals
    \item The specified ending period exceeds the maximum available period
  }
}

\examples{
\dontrun{
# Fit fect model with bootstrap
fit <- fect(Y ~ D + X, data = panel_data, keep.sims = TRUE)

# Calculate cumulative effects for all treated units
results <- effect(fit)

# Calculate period-specific effects for specific units
results_specific <- effect(fit,
                          cumu = FALSE,
                          id = c("unit1", "unit2"),
                          period = c(1, 4))

# View results
print(results$est.catt)
}
}

\seealso{
  \code{\link{fect}}, \code{\link{plot.fect}}
}

\references{
  Liu, L., Wang, Y., & Xu, Y. (2022).
  A Practical Guide to Counterfactual Estimators for Causal Inference with Time-Series Cross-Sectional Data.
  \emph{American Journal of Political Science}, 68(1), 160-176.
}

\author{
  Shiyun Hu, Licheng Liu, Ye Wang, and Yiqing Xu
}

\encoding{UTF-8}
\name{fect}
\alias{fect}
\title{Fixed Effects Counterfactual Estimators}
\description{Implements counterfactual estimators in TSCS data analysis and statistical tools to test their identification
assumptions.}
\usage{fect(formula = NULL, data, Y, D, X = NULL,
            W = NULL, W.est = NULL, W.agg = NULL,
            group = NULL,
            na.rm = FALSE,
            index, force = "two-way",
            time.component.from = "notyettreated", em = TRUE,
            r = 0, lambda = NULL, nlambda = 10,
            CV = NULL, k = 20, cv.prop = 0.1, cv.method = "rolling",
            cv.nobs = 3, cv.donut = 1, cv.buffer = 1, criterion = "mspe",
            binary = FALSE, QR = FALSE,
            method = "fe",  se = FALSE, vartype = "bootstrap",
            para.error = "auto", cl = NULL,
            ci.method = "normal", quantile.CI = NULL,
            nboots = 200, alpha = 0.05,
            parallel = TRUE, cores = NULL, tol = 1e-5,
            max.iteration = 5000, seed = NULL,
            min.T0 = NULL, max.missing = NULL,
            proportion = 0.3, pre.periods = NULL,
            f.threshold = 0.5, tost.threshold = NULL,
            knots = NULL, degree = 2,
            group.fe = NULL,
            cfe = NULL,
            Z = NULL, gamma = NULL, Q = NULL, kappa = NULL,
            Q.type = NULL,
            Q.bspline.degree = NULL,
            Z.param = NULL, Q.param = NULL,
            balance.period = NULL, fill.missing = FALSE,
            placeboTest = FALSE, placebo.period = NULL,
            carryoverTest = FALSE, carryover.period = NULL, carryover.rm = NULL,
            loo = FALSE, permute = FALSE, m = 2,
            normalize = FALSE, keep.sims = FALSE,
            cm = FALSE,
            loading.bound = "none", gamma.loading = NULL,
            gamma.loading.grid = NULL,
            cv.rule = "1se")}
\arguments{
\item{formula}{an object of class "formula": a symbolic description of the model to be fitted, e.g, Y~D+X1+X2}
\item{data}{a data frame, can be a balanced or unbalanced panel data.}
\item{Y}{the outcome indicator.}
\item{D}{the treatment indicator. The treatment should be binary (0 and 1).}
\item{X}{time-varying covariates. Covariates that have perfect collinearity with specified fixed effects are dropped automatically.}
\item{W}{a string giving the column name of a weight variable. Convenience default that populates both \code{W.est} and \code{W.agg} when those are left \code{NULL}. Suitable for survey or sample weights, where the same column applies to both the outcome-model fit and the across-treated-obs aggregation.}

\item{W.est}{a string giving the column name of a weight variable that enters the outcome-model fit (the weighted least squares applied inside the IFE / MC / CFE solver). When \code{NULL}, falls back to \code{W}. Use this (with \code{W.agg = NULL}) when the weight reflects fit-side considerations and the estimand is the unweighted average ATT across treated cells.}

\item{W.agg}{a string giving the column name of a weight variable that enters the across-treated-obs aggregation (\code{att.on}, \code{est.avg}, \code{est.att}). When \code{NULL}, falls back to \code{W}. Use this (with \code{W.est = NULL}) when the user's estimand differs from "ATT for the treated units in the analysis sample" --- common cases include calibration weights to a target population or post-stratification weights that should adjust the summary but not the model fit.

A clean in-package solution for inverse-probability weights for confounding adjustment is under development for fect 3.0 (a cross-fit doubly-robust path); \code{W.agg} is not a substitute for that work and does not deliver the doubly-robust properties an IPW user expects.

In v2.3.1, \code{W.est} and \code{W.agg} (when both supplied) must point to the same column; truly distinct columns for fit vs. aggregation (e.g. combined survey x IPW designs) are scheduled for v2.4.0.}

\item{group}{the group indicator. If specified, the group-wise ATT will be estimated.}
\item{na.rm}{a logical flag indicating whether to list-wise delete missing observations. Default to FALSE. If \code{na.rm = FALSE}, it allows the situation when Y is missing but D is not missing for some observations. If \code{na.rm = TRUE}, it will list-wise delete observations whose Y, D, or X is missing.}
\item{index}{a character vector specifying the unit (first element) and time (second element) indicators. For most methods, must be of length 2. For \code{method = "cfe"}, additional elements (third, fourth, etc.) specify extra fixed-effect grouping variables. Every observation should be uniquely defined by the pair of the unit and time indicator.}
\item{force}{a string indicating whether unit or time or both fixed effects will be imposed. Must be one of the following, "none", "unit", "time", or "two-way". The default is "two-way".}
\item{time.component.from}{Controls which units provide the time-varying model components (time fixed effects, factor structure, temporal dynamics). Options are \code{"notyettreated"} (default) --- all units contribute during their pre-treatment periods, or \code{"nevertreated"} --- only never-treated units estimate the time components, which are then projected onto treated units.}
\item{em}{a logical flag indicating whether to use the EM algorithm for missing data in the estimation sample. Default is \code{TRUE}. Setting \code{em = FALSE} requires a complete estimation sample and is only compatible with \code{time.component.from = "nevertreated"}.}
\item{r}{an integer specifying the number of factors. If \code{CV = TRUE}, the cross validation procedure will select the optimal number of factors from \code{r} to 5.}
\item{lambda}{a single or sequence of positive numbers specifying the hyper-parameter sequence for matrix completion method. If \code{lambda} is a sequence and \code{CV = 1}, cross-validation will be performed.}
\item{nlambda}{an integer specifying the length of hyper-parameter sequence for matrix completion method. Default is \code{nlambda = 10}.}
\item{CV}{a logical flag indicating whether cross-validation will be performed to select the optimal number of factors or hyper-parameter in matrix completion algorithm. If \code{r} is not specified, the procedure will search through \code{r = 0} to \code{5}.}
\item{k}{an integer specifying number of cross-validation rounds. Default is \code{k = 20}.}
\item{cv.prop}{a numerical value specifying the proportion of testing set compared to sample size during the cross-validation procedure.}
\item{cv.method}{a string specifying the cross-validation masking strategy. One of \code{"rolling"} (default; standard time-series rolling-window CV), \code{"block"} (random scattered anchors with contiguous-block masking), or \code{"loo"} (leave-one-out, available for \code{fect_nevertreated}). The legacy aliases \code{"all_units"} (= \code{"block"}) and \code{"treated_units"} (block masking restricted to treated pre-treatment cells) are still accepted but emit a deprecation message; both will be replaced by the unified \code{(cv.method, cv.units)} API in v2.4.0.}
\item{cv.nobs}{an integer specifying the length of continuous observations within a unit in the testing set. Default is \code{cv.nobs = 3}.}
\item{cv.donut}{an integer specifying the length of removed observations at the head and tail of the continuous observations specified by \code{cv.nobs}. Used by block CV (\code{cv.method = "block"}, or the legacy aliases \code{"all_units"}/\code{"treated_units"}). Default is 1 (matches \code{cv.buffer} for rolling CV).}
\item{cv.buffer}{an integer specifying the length of past-side buffer cells masked from training (but not scored) immediately before each rolling-window holdout. Used only by \code{cv.method = "rolling"}; the future side is dropped from training by construction. Analogous to \code{cv.donut} for block CV but applied only on the past side. Default is 1.}
\item{criterion}{criterion used for model selection. Default is "mspe".
  \code{"mspe"} for the mean squared prediction error,
  \code{"gmspe"} for the geometric-mean squared prediction errors,
  \code{"moment"} for period-weighted residuals in test sets,
  \code{"pc"} for an information criterion method.}
\item{binary}{This version doesn't support this option.}
\item{QR}{This version doesn't support this option.}
\item{method}{a string specifying which imputation algorithm will be used.
  \code{"fe"}, \code{"ife"}, \code{"mc"},
  \code{"gsynth"}, or \code{"cfe"}. Default is \code{"fe"}.}
\item{se}{a logical flag indicating whether uncertainty estimates will be produced.}
\item{vartype}{a string specifying the type of variance estimator, e.g.
  \code{"bootstrap"}. Three values are supported:
  \code{"bootstrap"} (nonparametric cluster-bootstrap; the safe default),
  \code{"jackknife"} (leave-one-unit-out), and
  \code{"parametric"} (two-stage pseudo-treated parametric bootstrap).
  The \code{"parametric"} option is restricted to the
  gsynth-style regime: it requires
  \code{time.component.from = "nevertreated"}, no treatment reversal,
  and \code{method} not in \code{c("mc", "both")}.
  These three conditions correspond to Gates A, B, and C in the
  three-gate defense system (see \code{ARCHITECTURE.md}).
  For all other settings, \code{vartype = "bootstrap"} is recommended.}
\item{para.error}{a string specifying the residual-error model used by
  the parametric bootstrap path; sub-option of \code{vartype = "parametric"}
  (silently ignored otherwise). One of \code{"auto"}, \code{"ar"},
  \code{"empirical"}, or \code{"wild"}. Default \code{"auto"} resolves at
  fit time to \code{"empirical"} on a fully-observed panel and \code{"ar"}
  on a panel with missing cells; the resolved label is stored on
  \code{fit$para.error}. \code{"ar"} estimates an AR(1) error process from
  control residuals (works on any panel shape); \code{"empirical"}
  resamples residuals i.i.d. from the main-fit pool (requires fully-
  observed panel); \code{"wild"} applies unit-level Rademacher sign-flips
  over the empirical residual pool (requires fully-observed panel).}
\item{cl}{a string specifying the cluster column for cluster bootstrapping.
When \code{group.fe} is set to a single column and \code{cl} is unset,
\code{cl} auto-defaults to \code{group.fe[1]} --- the natural choice when
treatment varies at the group level (Bertrand, Duflo & Mullainathan 2004).
To override --- for example, to cluster at the unit level when group-level
FE is absorbed --- pass \code{cl = index[1]} (e.g., \code{cl = "id"}).
Note that \code{cl = NULL} does NOT disable clustering: the case bootstrap
always resamples units (which is unit-level clustering); the auto-default
only changes the resample unit to a coarser level. \code{cl = FALSE} is
rejected with a guiding error.}
\item{ci.method}{a string controlling how the bootstrap distribution
becomes the CIs reported in fect's \code{est.*} slots. Two values:
\code{"normal"} (default; Wald CI: \eqn{\hat\theta \pm z \cdot SE}) or
\code{"basic"} (reflected pivot:
\eqn{[2\hat\theta - q_{1-\alpha/2},\, 2\hat\theta - q_{\alpha/2}]}, the
literature-standard "percentile" CI per Davison & Hinkley (1997, §5.2.1)
and what \code{boot::boot.ci(type = "basic")} returns). For
\code{"percentile"}, \code{"bc"}, or \code{"bca"} CIs (typically wanted
on alternative estimands like \code{aptt} and \code{log.att}), call
\code{\link{estimand}} after fitting.}

\item{quantile.CI}{deprecated as of v2.4.2. Use \code{ci.method} instead.
\code{quantile.CI = FALSE} maps to \code{ci.method = "normal"};
\code{quantile.CI = TRUE} maps to \code{ci.method = "basic"}. Both
mappings still work but emit a one-time deprecation warning.}
\item{nboots}{an integer specifying the number of bootstrap runs. Ignored if \code{se=FALSE}. Default \code{200}, sufficient for the standard error and the normal-CI structure that \code{fect} ships in \code{est.att}/\code{att.avg}. For tail-quantile-based CI methods accessed via \code{estimand()} (\code{ci.method} \code{"basic"}, \code{"percentile"}, \code{"bc"}, \code{"bca"}), bump to \code{nboots = 1000} or higher (Efron 1987 Section 3; DiCiccio & Efron 1996 Section 4); \code{estimand()} emits a warning when called on under-replicated fits.}
\item{alpha}{the significance level for hypothesis tests and confidence intervals. Default \code{0.05}.}
\item{parallel}{controls which operations run in parallel. Accepted values:
\describe{
  \item{\code{TRUE}}{Enable parallel computing for both CV and bootstrap (default in \code{fect()}).}
  \item{\code{FALSE}}{Disable all parallel computing.}
  \item{\code{"cv"}}{Enable parallel CV only; bootstrap runs serially.}
  \item{\code{"boot"}}{Enable parallel bootstrap only; CV runs serially.}
  \item{\code{c("cv","boot")}}{Explicit form of \code{TRUE}: parallel for both.}
}
When \code{parallel = TRUE}, auto-enable thresholds apply: CV parallelism
engages only when \code{Nco * TT} exceeds the per-method threshold
(\code{ife = 20000}, \code{mc = 20000}, \code{cfe = 60000}). Explicit \code{"cv"} overrides
the threshold. Nested parallelism (calling \code{fect()} from within a
\code{future_lapply} or \code{foreach \%dopar\%} block) should use
\code{parallel = FALSE} to avoid deadlock.
When using \code{parallel} with \code{method = "mc"}, parallel CV computes
all candidate lambda values without early stopping (the serial path uses a
\code{break_check} short-circuit to skip lambdas with diminishing MSPE
returns). This guarantees numerical identity between serial and parallel
results but may compute a few extra lambda values compared to the serial
path. Use \code{parallel = FALSE} to preserve the short-circuit behavior.}
\item{cores}{an integer indicating the number of cores for parallel computing.}
\item{tol}{a positive number indicating the relative tolerance for the
EM update check (\code{||fit_new - fit_old||_F / ||fit_old||_F < tol}).
Default tightened from \code{1e-3} to \code{1e-5} in v2.4.3 because the
older default produced under-converged IFE/CFE point estimates that
shifted up to 40\% relative to the converged value (inference was
preserved, but the reported numbers were stopping-point-dependent).}
\item{max.iteration}{the maximal number of EM iterations. Default raised
from \code{1000} to \code{5000} in v2.4.3 to accommodate the tighter tol;
canonical IFE/CFE fits converge in 700-2000 iters at \code{tol = 1e-5}.
A \code{warning()} is emitted when the EM hits this cap without
satisfying the tol gate (under-convergence diagnostic).}
\item{seed}{an integer seed for random number generation.}
\item{min.T0}{an integer specifying the minimum number of pre-treatment periods for each treated unit.}
\item{max.missing}{an integer specifying the maximum number of missing observations allowed per unit.}
\item{proportion}{a numeric value specifying which pre-treatment periods are used for goodness-of-fit tests.}
\item{pre.periods}{a vector specifying the range of pre-treatment periods used for the goodness-of-fit test.}
\item{f.threshold}{a numeric threshold for an F-test in equivalence testing. Default \code{0.5}.}
\item{tost.threshold}{a numeric threshold for two-one-sided t-tests.}
\item{knots}{a numeric vector specifying knots (currently unused; reserved for future use).}
\item{degree}{an integer specifying the degree (currently unused; reserved for future use).}
\item{group.fe}{a character vector of column names naming additional simple
additive fixed-effect groupings to absorb (e.g., \code{group.fe = "state"}
when rows are counties and treatment varies at state level). Each entry must
be a column in \code{data}. Each entry must be \emph{nested in}
\code{index[1]} --- i.e., constant within each level of the unit identifier
--- otherwise an error is raised. When \code{group.fe} is set, \code{method
= "fe"} is silently routed to \code{method = "cfe"} (FE is a subset of CFE,
identical result); \code{method = "ife"}, \code{"mc"}, \code{"both"}, and
\code{"gsynth"} hard-error (use \code{method = "cfe"} with \code{r > 0} for
free latent factors with group-level FE). When \code{group.fe} has length 1
and \code{cl} is unset, \code{cl} auto-defaults to \code{group.fe[1]}. To
cluster at the unit level instead, pass \code{cl = index[1]} (e.g.,
\code{cl = "id"}); \code{cl = NULL} does NOT change behavior (the case
bootstrap always resamples units regardless of \code{cl}), and
\code{cl = FALSE} is rejected with a guiding error.}
\item{cfe}{a vector of lists specifying interactive fixed effects for \code{method="cfe"}.}
\item{Z}{a vector specifying the time-invariant covariates for the Z matrix.}
\item{gamma}{a vector specifying the time-varying covariates for the gamma matrix.}
\item{Q}{a vector specifying the time-varying covariates for the Q matrix.}
\item{kappa}{a vector specifying the time-invariant covariates for the kappa matrix.}
\item{Q.type}{a vector specifying the type of Q matrix.}
\item{Q.bspline.degree}{an integer specifying the degree used when \code{Q.type} includes \code{"bspline"} in \code{method="cfe"}. If NULL, a default degree is chosen based on the number of distinct time values.}
\item{Z.param}{a list specifying the parameters for the Z matrix.}
\item{Q.param}{a list specifying the parameters for the Q matrix.}
\item{balance.period}{a length-2 vector specifying a time range for a balanced sample.}
\item{fill.missing}{a logical flag indicating whether to allow missing observations in a balanced sample.}
\item{placeboTest}{a logical flag indicating whether to perform a placebo test.}
\item{placebo.period}{an integer or 2-element numeric vector specifying pseudo-treatment periods.}
\item{carryoverTest}{a logical flag for carryover tests.}
\item{carryover.period}{an integer or 2-element numeric vector specifying pseudo-carryover periods.}
\item{carryover.rm}{an integer specifying the range of post-treatment periods to treat as carryover.}
\item{loo}{a logical flag for leave-one-period-out goodness-of-fit tests.}
\item{permute}{a logical flag indicating whether to run a permutation test.}
\item{m}{an integer specifying the block length for permutation tests. Default \code{2}.}
\item{normalize}{a logical flag indicating whether to scale outcome and covariates.}
\item{keep.sims}{a logical flag indicating whether to save unit-time level bootstrap effects. Default \code{keep.sims = FALSE}. If \code{se = FALSE}, this argument is ignored.}
\item{cm}{a logical flag indicating whether to enable causal moderation analysis. When \code{TRUE}, the estimator decomposes the treatment effect into effect modification and causal moderation components. Currently available for \code{method = "fe"} and \code{method = "ife"}. Default is \code{FALSE}.}
\item{loading.bound}{a string controlling whether treated-unit factor loadings are bounded inside the convex hull of control loadings. \code{"none"} (default) reproduces standard GSC behavior. \code{"simplex"} constrains each treated unit's loading to be a non-negative convex combination of control loadings via an entropy-regularized simplex projection, ensuring the imputed counterfactual lies pointwise in the convex hull of factor-implied control outcomes. Currently applies only to \code{method = "ife"} or \code{method = "gsynth"} (equivalent forms), and requires \code{time.component.from = "nevertreated"}.}
\item{gamma.loading}{scalar regularization strength for the \code{"simplex"} projection. \code{NULL} (default) triggers 5-fold cross-validation over \code{gamma.loading.grid}. A numeric value is used directly. Ignored when \code{loading.bound = "none"}.}
\item{gamma.loading.grid}{a numeric vector of candidate \code{gamma.loading} values for cross-validation. \code{NULL} (default) uses \code{10^seq(-2, 2, length.out = 9)}. Ignored when \code{loading.bound = "none"} or when \code{gamma.loading} is supplied.}
\item{cv.rule}{a string selecting the cross-validation rule for choosing the number of factors \code{r} (or matrix-completion penalty \code{lambda}). One of:
\describe{
  \item{\code{"1se"} (default)}{The 1-SE rule (Breiman, Friedman, Olshen and Stone 1984; Hastie, Tibshirani and Friedman 2009, Section 7.10): pick the smallest \code{r} whose mean CV criterion is within one fold-SE of the minimum-CV-error \code{r}. Biases toward parsimony in a fold-aware way --- when CV is precise, it allows larger \code{r}; when CV is noisy, it gravitates to simpler models.}
  \item{\code{"min"}}{Pick the \code{r} that minimizes the mean CV criterion (no tolerance).}
  \item{\code{"1pct"}}{Legacy pre-2.3.0 heuristic: pick the smallest \code{r} within 1\% relative tolerance of the best mean CV criterion. Use this for byte-identical reproducibility of pre-2.3.0 fits.}
}
Ignored when \code{CV = FALSE}.}
}
\details{
  \code{fect} implements counterfactual estimators for TSCS data. It first imputes counterfactuals by fitting an outcome model using untreated observations, then estimates the individual treatment effect as the difference between observed and predicted outcomes. Finally, it computes average treatment effects on the treated (ATT) and period-specific ATTs. Placebo and equivalence tests help evaluate identification assumptions.
}
\value{
  \item{Y.dat}{T-by-N matrix of the outcome variable.}
  \item{D.dat}{T-by-N matrix of the treatment variable.}
  \item{I.dat}{T-by-N matrix of observation indicators (observed/missing).}
  \item{Y}{name of the outcome variable.}
  \item{D}{name of the treatment variable.}
  \item{X}{name of any time-varying covariates.}
  \item{W}{name of the weight variable.}
  \item{index}{name of the unit and time indicators.}
  \item{force}{specified fixed effects option.}
  \item{T}{number of time periods.}
  \item{N}{number of units.}
  \item{p}{number of time-varying observables.}
  \item{r.cv}{number of factors (selected by cross-validation if needed).}
  \item{lambda.cv}{optimal hyper-parameter for matrix completion, if applicable.}
  \item{beta}{coefficients for any covariates in an interactive fixed effects model.}
  \item{sigma2}{mean squared error.}
  \item{IC}{information criterion.}
  \item{est}{results of the fitted model.}
  \item{MSPE}{mean squared prediction error from cross-validation.}
  \item{CV.out}{results of the cross-validation procedure.}
  \item{niter}{number of iterations.}
  \item{factor}{estimated time-varying factors.}
  \item{lambda}{estimated loadings.}
  \item{lambda.tr}{estimated loadings for treated units.}
  \item{lambda.co}{estimated loadings for control units.}
  \item{mu}{estimated grand mean.}
  \item{xi}{estimated time fixed effects.}
  \item{alpha}{estimated unit fixed effects.}
  \item{alpha.tr}{estimated unit fixed effects for treated units.}
  \item{alpha.co}{estimated unit fixed effects for control units.}
  \item{validX}{logical indicating if valid covariates exist.}
  \item{validF}{logical indicating if factors exist.}
  \item{id}{vector of unit IDs.}
  \item{rawtime}{vector of time periods.}
  \item{obs.missing}{matrix indicating missingness patterns.}
  \item{Y.ct}{T-by-N matrix of predicted outcomes under no treatment.}
  \item{eff}{T-by-N matrix of estimated individual treatment effects.}
  \item{res}{residuals for observed values.}
  \item{eff.pre}{effects for treated units in pre-treatment periods.}
  \item{eff.pre.equiv}{pre-treatment effects under baseline (two-way FE) model.}
  \item{pre.sd}{by-period residual standard deviations for pre-treatment ATT.}
  \item{att.avg}{overall average treatment effect on the treated.}
  \item{att.avg.W}{weighted ATT.}
  \item{att.avg.unit}{by-unit average treatment effect on the treated.}
  \item{time}{time index for switch-on treatment effect.}
  \item{count}{count of observations for each switch-on effect time.}
  \item{att}{switch-on treatment effect.}
  \item{att.on.W}{weighted switch-on effect.}
  \item{time.off}{time index for switch-off treatment effect.}
  \item{att.off}{switch-off treatment effect.}
  \item{att.off.W}{weighted switch-off effect.}
  \item{count.off}{count for each switch-off period.}
  \item{att.placebo}{ATT for placebo periods.}
  \item{att.carryover}{ATT for carryover periods.}
  \item{eff.calendar}{ATT by calendar time.}
  \item{eff.calendar.fit}{loess-fitted ATT by calendar time.}
  \item{N.calandar}{number of treated observations each calendar period.}
  \item{balance.avg.att}{ATT for balanced sample.}
  \item{balance.att}{switch-on ATT for balanced sample.}
  \item{balance.time}{time index for balanced sample.}
  \item{balance.count}{count for each time in balanced sample.}
  \item{balance.att.placebo}{ATT for placebo period in balanced sample.}
  \item{group.att}{ATT for different groups.}
  \item{group.output}{list of switch-on treatment effects by group.}
  \item{est.att.avg}{inference for \code{att.avg}.}
  \item{est.att.avg.unit}{inference for \code{att.avg.unit}.}
  \item{est.att}{inference for \code{att}.}
  \item{est.att.W}{inference for weighted \code{att}.}
  \item{est.att.off}{inference for switch-off.}
  \item{est.att.off.W}{inference for weighted switch-off.}
  \item{est.placebo}{inference for placebo ATT.}
  \item{est.carryover}{inference for carryover ATT.}
  \item{est.eff.calendar}{inference for \code{eff.calendar}.}
  \item{est.eff.calendar.fit}{inference for \code{eff.calendar.fit}.}
  \item{est.balance.att}{inference for balanced sample switch-on.}
  \item{est.balance.avg}{inference for balanced sample average ATT.}
  \item{est.balance.placebo}{inference for balanced sample placebo.}
  \item{est.avg.W}{inference for \code{att.avg.W}.}
  \item{est.beta}{inference for \code{beta}.}
  \item{est.group.att}{inference for group-specific ATT.}
  \item{est.group.output}{inference for group output.}
  \item{att.avg.boot}{bootstrap draws for \code{att.avg}.}
  \item{att.avg.unit.boot}{bootstrap draws for \code{att.avg.unit}.}
  \item{att.count.boot}{bootstrap draws for \code{count}.}
  \item{att.off.boot}{bootstrap draws for \code{att.off}.}
  \item{att.off.count.boot}{bootstrap draws for \code{count.off}.}
  \item{att.placebo.boot}{bootstrap draws for \code{att.placebo}.}
  \item{att.carryover.boot}{bootstrap draws for \code{att.carryover}.}
  \item{balance.att.boot}{bootstrap draws for \code{balance.att}.}
  \item{att.bound}{equivalence confidence interval for pre-trend.}
  \item{att.off.bound}{equivalence confidence interval for switch-off.}
  \item{beta.boot}{bootstrap draws for \code{beta}.}
  \item{test.out}{F-test and equivalence test results for pre-treatment fit.}
  \item{loo.test.out}{leave-one-period-out test results.}
  \item{permute}{permutation test results.}
}
\author{
  Licheng Liu; Ye Wang; Yiqing Xu; Ziyi Liu
}
\references{
  Athey, S., Bayati, M., Doudchenko, N., Imbens, G., and Khosravi, K. (2021).
  Matrix completion methods for causal panel data models.
  \emph{Journal of the American Statistical Association, 116}(536), 1716-1730.

  Bai, J. (2009).
  Panel data models with interactive fixed effects.
  \emph{Econometrica, 77}(4), 1229-1279.

  Liu, L., Wang, Y., and Xu, Y. (2022).
  A Practical Guide to Counterfactual Estimators for Causal Inference with Time-Series Cross-Sectional Data.
  \emph{American Journal of Political Science, 68}(1), 160-176.

  Xu, Y. (2017).
  Generalized Synthetic Control Method: Causal Inference with Interactive Fixed Effects Models.
  \emph{Political Analysis, 25}(1), 57-76.
}
\seealso{
  \code{\link{plot.fect}}, \code{\link{print.fect}}
}
\examples{
library(fect)
data(simdata)
out <- fect(Y ~ D + X1 + X2, data = simdata,
            index = c("id","time"), force = "two-way",
            CV = TRUE, r = c(0, 5), se = 0, parallel = FALSE)
}

\name{did_wrapper}
\alias{did_wrapper}
\title{A Multi-Method Difference-in-Differences Estimator}
\description{
Implements multiple difference-in-differences (DiD) estimators under a unified interface, supporting overall ATT and event-study estimates across staggered adoption settings.
}
\usage{
did_wrapper(
  data,
  Y,
  D,
  X = NULL,
  index,
  method = c("twfe", "st", "iw", "cs_never", "cs_notyet", "didm"),
  se = c("default", "boot", "bootstrap", "jackknife"),
  nboots = 200,
  parallel = TRUE,
  core = NULL,
  time_to_treat_var = "Time_to_Treatment",
  treat_indicator = "treat",
  csdid.base_period = "universal",
  didm.effects = NA,
  didm.placebo = NA
)
}
\arguments{
  \item{data}{Input data frame.}
  \item{Y}{Outcome variable name (string).}
  \item{D}{Treatment indicator variable name (string).}
  \item{X}{Optional covariate vector for adjustment.}
  \item{index}{Character vector of unit and time variable names, e.g., \code{c("id", "time")}.}
  \item{method}{DiD method: \code{"twfe"}, \code{"st"}, \code{"iw"}, \code{"cs_never"}, \code{"cs_notyet"}, or \code{"didm"}.}
  \item{se}{Standard error method: \code{"default"}, \code{"boot"}, \code{"bootstrap"}, or \code{"jackknife"}.}
  \item{nboots}{Number of bootstrap replications (if applicable).}
  \item{parallel}{Logical; use parallel computation for bootstrapping.}
  \item{core}{Number of CPU cores to use if \code{parallel = TRUE}.}
  \item{time_to_treat_var}{Name of time-to-treatment variable; used internally.}
  \item{treat_indicator}{Name of treatment cohort indicator variable; used internally.}
  \item{csdid.base_period}{Baseline period choice for Callaway–Sant'Anna estimators.}
  \item{didm.effects}{Effects vector for \code{didm}; required if \code{method = "didm"}.}
  \item{didm.placebo}{Placebo vector for \code{didm}; required if \code{method = "didm"}.}
}
\details{
This function:
\enumerate{
  \item Drops always-treated units.
  \item Constructs event time and cohort variables.
  \item Computes ATT using the specified DiD method.
  \item Constructs event-study ATT curves.
  \item Optionally estimates bootstrap or jackknife standard errors.
}
Supported methods include TWFE, stacked DiD, interaction-weighted DiD, Callaway–Sant'Anna estimators, placebo DiD, and DIDmultiplegt.
}
\value{
A list of class \code{"did_wrapper"}:
\item{est.avg}{Data frame with overall ATT, standard error, confidence interval, and p-value.}
\item{est.att}{Event-study ATT estimates by relative period, including standard errors and confidence intervals.}
}
\author{
Rivka Lipkovitz
}
\examples{
\dontrun{
result_twfe <- did_wrapper(
  data   = df,
  Y      = "outcome",
  D      = "treat",
  index  = c("id", "time"),
  method = "twfe"
)
result_twfe$est.avg
result_twfe$est.att
}
}

\encoding{UTF-8}
\name{print.interFE}
\alias{print.interFE}
\title{Print Results}
\description{Print results of interactive fixed effects estimation.}
\usage{\method{print}{interFE}(x, \dots)}
\arguments{
  \item{x}{an \code{\link{interFE}} object.}
  \item{\dots}{other arguments.}
}
\author{
  Licheng Liu; Ye Wang; Yiqing Xu
}
\value{No return value.}
\references{
  Bai, J. (2009). 
  Panel data models with interactive fixed effects. 
  \emph{Econometrica, 77}(4), 1229-1279.
}
\seealso{
  \code{\link{interFE}}, \code{\link{fect}}
}
\examples{
library(fect)
data(simdata)
d <- simdata[-(1:150),] # remove the treated units
out <- interFE(Y ~ X1 + X2, data = d, index=c("id","time"),
               r = 2, force = "two-way", nboots = 50)
print(out)
}

\encoding{UTF-8}
\name{sim_region}
\docType{data}
\alias{sim_region}
\title{Simulated unbalanced panel with region-specific time effects}
\description{
A simulated unbalanced panel with 500 units belonging to 5 regions and 20 time periods.
The DGP includes region-specific linear time trends as the confounding source.
Treatment probability and timing depend on region, and units in higher-numbered
regions enter the panel later, creating unbalancedness.
Used to demonstrate additional fixed effects in the CFE estimator.
}
\format{A data frame with columns:
\describe{
  \item{time}{Time period (1 to 20)}
  \item{id}{Unit identifier (1 to 500)}
  \item{region}{Region indicator (1 to 5)}
  \item{Y}{Outcome variable}
  \item{D}{Treatment indicator (0/1)}
  \item{region_time}{Region-by-time interaction factor}
}
}
\keyword{datasets}

\encoding{UTF-8}
\name{sim_gsynth}
\docType{data}
\alias{sim_gsynth}
\title{Simulated data for Gsynth (no treatment reversal)}
\description{A simulated dataset with no treatment reversal, used to demonstrate the generalized synthetic control method (never-treated estimation regime).}
\format{dataframe}
\keyword{datasets}
\references{
  Xu, Y. (2017).
  Generalized Synthetic Control Method: Causal Inference with Interactive Fixed Effects Models.
  \emph{Political Analysis, 25}(1), 57-76.
}

\encoding{UTF-8}
\name{sim_trend}
\docType{data}
\alias{sim_trend}
\title{Simulated panel data with unit-specific sinusoidal time trends (block DID)}
\description{
A simulated balanced panel with 200 units and 50 time periods.
All 80 treated units receive treatment at period 41 (block DID design).
The DGP includes unit-specific amplitudes on a half-cycle sinusoidal time trend,
where treated units have systematically larger amplitudes.
}
\format{A data frame with 10000 rows and 4 columns:
\describe{
  \item{id}{Unit identifier (1 to 200)}
  \item{time}{Time period (1 to 50)}
  \item{Y}{Outcome variable}
  \item{D}{Treatment indicator (0/1)}
}
}
\keyword{datasets}

\encoding{UTF-8}
\name{simgsynth}
\docType{data}
\alias{simgsynth}
\title{Simulated data for Gsynth}
\description{A simulated dataset with no treatment reversal.}
\format{dataframe}
\keyword{datasets}
\references{
  Xu, Y. (2017). 
  Generalized Synthetic Control Method: Causal Inference with Interactive Fixed Effects Models. 
  \emph{Political Analysis, 25}(1), 57-76.
}

\encoding{UTF-8}
\name{sim_base}
\docType{data}
\alias{sim_base}
\title{Simulated data (no interactive fixed effects)}
\description{A simulated dataset with continuous outcomes and no interactive fixed effects (factor loadings set to zero). Same DGP as \code{simdata} but with the factor contribution removed from the outcome.}
\format{dataframe}
\keyword{datasets}
\references{
  Liu, L., Wang, Y., and Xu, Y. (2022).
  A Practical Guide to Counterfactual Estimators for Causal Inference with Time-Series Cross-Sectional Data.
  \emph{American Journal of Political Science, 68}(1), 160-176.
}

\encoding{UTF-8}
\name{turnout}
\docType{data}
\alias{turnout}
\title{EDR and Voter Turnout in the US}
\description{State-level voter turnout data.}
\format{dataframe}
\keyword{datasets}
\references{
  Springer, M. J. (2014). 
  How the States Shaped the Nation: American Electoral Institutions and Voter Turnout, 1920-2000. 
  University of Chicago Press.
  
  Xu, Y. (2017). 
  Generalized Synthetic Control Method: Causal Inference with Interactive Fixed Effects Models. 
  \emph{Political Analysis, 25}(1), 57-76.
}
